#!/usr/bin/env python
# 786
# Aldy source: __init__.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from .version import __version__  # noqa
