/* Generated by Cython 0.29.32 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/htslib_util.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/bgzf.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/cram.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/faidx.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/hfile.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/hts.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/kstring.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/sam.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/tbx.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/vcf.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/vcfutils.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/pysam_stream.h"
        ],
        "extra_compile_args": [
            "-Wno-unused-function"
        ],
        "include_dirs": [
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/samtools"
        ],
        "name": "aldy.indelpost.varaln",
        "sources": [
            "aldy/indelpost/varaln.pyx"
        ]
    },
    "module_name": "aldy.indelpost.varaln"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_32"
#define CYTHON_HEX_VERSION 0x001D20F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_HEX >= 0x07030900)
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__aldy__indelpost__varaln
#define __PYX_HAVE_API__aldy__indelpost__varaln
/* Early includes */
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "pythread.h"
#include <sys/types.h>
#include "stdarg.h"
#include "htslib/kstring.h"
#include "htslib_util.h"
#include "htslib/hfile.h"
#include "htslib/bgzf.h"
#include "htslib/hts.h"
#include "htslib/sam.h"
#include "htslib/faidx.h"
#include "htslib/tbx.h"
#include "htslib/vcf.h"
#include "htslib/vcfutils.h"
#include "htslib/cram.h"
#include "pysam_stream.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "aldy/indelpost/varaln.pyx",
  "stringsource",
  "array.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "libchtslib.pxd",
  "libcfaidx.pxd",
  "libcalignedsegment.pxd",
  "libcalignmentfile.pxd",
  "libcbcf.pxd",
  "aldy/indelpost/variant.pxd",
  "aldy/indelpost/contig.pxd",
  "aldy/indelpost/local_reference.pxd",
};

/*--- Type declarations ---*/
#ifndef _ARRAYARRAY_H
struct arrayobject;
typedef struct arrayobject arrayobject;
#endif
struct __pyx_obj_5pysam_10libchtslib_HTSFile;
struct __pyx_obj_5pysam_9libcfaidx_FastaFile;
struct __pyx_obj_5pysam_9libcfaidx_FastqProxy;
struct __pyx_obj_5pysam_9libcfaidx_FastxRecord;
struct __pyx_obj_5pysam_9libcfaidx_FastxFile;
struct __pyx_obj_5pysam_9libcfaidx_FastqFile;
struct __pyx_obj_5pysam_9libcfaidx_Fastafile;
struct __pyx_obj_5pysam_18libcalignedsegment_AlignedSegment;
struct __pyx_obj_5pysam_18libcalignedsegment_PileupColumn;
struct __pyx_obj_5pysam_18libcalignedsegment_PileupRead;
struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentHeader;
struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile;
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRow;
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowRegion;
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowHead;
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowAll;
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowAllRefs;
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowSelection;
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumn;
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumnRegion;
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumnAllRefs;
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumnAll;
struct __pyx_obj_5pysam_17libcalignmentfile_IndexedReads;
struct __pyx_obj_5pysam_7libcbcf_VariantHeader;
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderRecord;
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderRecords;
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderContigs;
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderSamples;
struct __pyx_obj_5pysam_7libcbcf_VariantContig;
struct __pyx_obj_5pysam_7libcbcf_VariantMetadata;
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderMetadata;
struct __pyx_obj_5pysam_7libcbcf_VariantRecord;
struct __pyx_obj_5pysam_7libcbcf_VariantRecordFilter;
struct __pyx_obj_5pysam_7libcbcf_VariantRecordFormat;
struct __pyx_obj_5pysam_7libcbcf_VariantRecordInfo;
struct __pyx_obj_5pysam_7libcbcf_VariantRecordSamples;
struct __pyx_obj_5pysam_7libcbcf_VariantRecordSample;
struct __pyx_obj_5pysam_7libcbcf_BaseIndex;
struct __pyx_obj_5pysam_7libcbcf_BCFIndex;
struct __pyx_obj_5pysam_7libcbcf_TabixIndex;
struct __pyx_obj_5pysam_7libcbcf_BaseIterator;
struct __pyx_obj_5pysam_7libcbcf_BCFIterator;
struct __pyx_obj_5pysam_7libcbcf_TabixIterator;
struct __pyx_obj_5pysam_7libcbcf_VariantFile;
struct __pyx_obj_4aldy_9indelpost_7variant_NullVariant;
struct __pyx_obj_4aldy_9indelpost_7variant_Variant;
struct __pyx_obj_4aldy_9indelpost_6contig_FailedContig;
struct __pyx_obj_4aldy_9indelpost_6contig_Contig;
struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference;
struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment;
struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup;
struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads;
struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr;
struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles;
struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr;
struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner;
struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr;
struct __pyx_opt_args_5pysam_9libcfaidx_10FastqProxy_get_quality_array;
struct __pyx_opt_args_5pysam_9libcfaidx_11FastxRecord_get_quality_array;

/* "pysam/libcfaidx.pxd":49
 *     cdef cython.str to_string(self)
 *     cdef cython.str tostring(self)
 *     cpdef array.array get_quality_array(self, int offset=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_5pysam_9libcfaidx_10FastqProxy_get_quality_array {
  int __pyx_n;
  int offset;
};

/* "pysam/libcfaidx.pxd":59
 *     cdef cython.str to_string(self)
 *     cdef cython.str tostring(self)
 *     cpdef array.array get_quality_array(self, int offset=*)             # <<<<<<<<<<<<<<
 * 
 * cdef class FastxFile:
 */
struct __pyx_opt_args_5pysam_9libcfaidx_11FastxRecord_get_quality_array {
  int __pyx_n;
  int offset;
};
struct __pyx_opt_args_5pysam_18libcalignedsegment_14AlignedSegment_set_tag;
struct __pyx_opt_args_5pysam_18libcalignedsegment_14AlignedSegment_get_tag;
struct __pyx_opt_args_5pysam_18libcalignedsegment_14AlignedSegment_tostring;

/* "pysam/libcalignedsegment.pxd":30
 * 
 * from pysam.libcalignmentfile cimport AlignmentFile, AlignmentHeader
 * ctypedef AlignmentFile AlignmentFile_t             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *__pyx_t_5pysam_18libcalignedsegment_AlignmentFile_t;

/* "pysam/libcalignedsegment.pxd":50
 *     # add an alignment tag with value to the AlignedSegment
 *     # an existing tag of the same name will be replaced.
 *     cpdef set_tag(self, tag, value, value_type=?, replace=?)             # <<<<<<<<<<<<<<
 * 
 *     # get an alignment tag from the AlignedSegment
 */
struct __pyx_opt_args_5pysam_18libcalignedsegment_14AlignedSegment_set_tag {
  int __pyx_n;
  PyObject *value_type;
  PyObject *replace;
};

/* "pysam/libcalignedsegment.pxd":53
 * 
 *     # get an alignment tag from the AlignedSegment
 *     cpdef get_tag(self, tag, with_value_type=?)             # <<<<<<<<<<<<<<
 * 
 *     # return true if tag exists
 */
struct __pyx_opt_args_5pysam_18libcalignedsegment_14AlignedSegment_get_tag {
  int __pyx_n;
  PyObject *with_value_type;
};

/* "pysam/libcalignedsegment.pxd":62
 * 
 *     # returns a valid sam alignment string (deprecated)
 *     cpdef tostring(self, htsfile=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_5pysam_18libcalignedsegment_14AlignedSegment_tostring {
  int __pyx_n;
  PyObject *htsfile;
};
struct __pyx_t_5pysam_17libcalignmentfile___iterdata;
typedef struct __pyx_t_5pysam_17libcalignmentfile___iterdata __pyx_t_5pysam_17libcalignmentfile___iterdata;
struct __pyx_opt_args_5pysam_17libcalignmentfile_14IteratorColumn__setup_iterator;

/* "pysam/libcalignmentfile.pxd":24
 * # Utility types
 * 
 * ctypedef struct __iterdata:             # <<<<<<<<<<<<<<
 *     htsFile * htsfile
 *     bam_hdr_t * header
 */
struct __pyx_t_5pysam_17libcalignmentfile___iterdata {
  htsFile *htsfile;
  bam_hdr_t *header;
  hts_itr_t *iter;
  faidx_t *fastafile;
  int tid;
  char *seq;
  int seq_len;
  int min_mapping_quality;
  int flag_require;
  int flag_filter;
  int compute_baq;
  int redo_baq;
  int ignore_orphans;
  int adjust_capq_threshold;
};

/* "pysam/libcalignmentfile.pxd":120
 *     cdef int cnext(self)
 *     cdef char * get_sequence(self)
 *     cdef _setup_iterator(self,             # <<<<<<<<<<<<<<
 *                          int tid,
 *                          int start,
 */
struct __pyx_opt_args_5pysam_17libcalignmentfile_14IteratorColumn__setup_iterator {
  int __pyx_n;
  int multiple_iterators;
};
struct __pyx_opt_args_4aldy_9indelpost_6varaln_16VariantAlignment___parse_pileup;

/* "aldy/indelpost/varaln.pxd":22
 *     cdef readonly UnsplicedLocalReference unspliced_local_reference
 * 
 *     cdef __parse_pileup(self, Contig contig=*, bint retargeted=*, bint skip_read_end_check=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_4aldy_9indelpost_6varaln_16VariantAlignment___parse_pileup {
  int __pyx_n;
  struct __pyx_obj_4aldy_9indelpost_6contig_Contig *contig;
  int retargeted;
  int skip_read_end_check;
};

/* "pysam/libchtslib.pxd":2680
 * 
 * 
 * cdef class HTSFile(object):             # <<<<<<<<<<<<<<
 *     cdef          htsFile *htsfile       # pointer to htsFile structure
 *     cdef          int64_t start_offset   # BGZF offset of first record
 */
struct __pyx_obj_5pysam_10libchtslib_HTSFile {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pysam_10libchtslib_HTSFile *__pyx_vtab;
  htsFile *htsfile;
  int64_t start_offset;
  PyObject *filename;
  PyObject *mode;
  PyObject *threads;
  PyObject *index_filename;
  int is_stream;
  int is_remote;
  int duplicate_filehandle;
};


/* "pysam/libcfaidx.pxd":37
 *                     int * dret)
 * 
 * cdef class FastaFile:             # <<<<<<<<<<<<<<
 *     cdef bint is_remote
 *     cdef object _filename, _references, _lengths, reference2length
 */
struct __pyx_obj_5pysam_9libcfaidx_FastaFile {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pysam_9libcfaidx_FastaFile *__pyx_vtab;
  int is_remote;
  PyObject *_filename;
  PyObject *_references;
  PyObject *_lengths;
  PyObject *reference2length;
  faidx_t *fastafile;
};


/* "pysam/libcfaidx.pxd":45
 * 
 * 
 * cdef class FastqProxy:             # <<<<<<<<<<<<<<
 *     cdef kseq_t * _delegate
 *     cdef cython.str to_string(self)
 */
struct __pyx_obj_5pysam_9libcfaidx_FastqProxy {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pysam_9libcfaidx_FastqProxy *__pyx_vtab;
  kseq_t *_delegate;
};


/* "pysam/libcfaidx.pxd":52
 * 
 * 
 * cdef class FastxRecord:             # <<<<<<<<<<<<<<
 *     """
 *     Python container for pysam.libcfaidx.FastqProxy with persistence.
 */
struct __pyx_obj_5pysam_9libcfaidx_FastxRecord {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxRecord *__pyx_vtab;
  PyObject *comment;
  PyObject *quality;
  PyObject *sequence;
  PyObject *name;
};


/* "pysam/libcfaidx.pxd":61
 *     cpdef array.array get_quality_array(self, int offset=*)
 * 
 * cdef class FastxFile:             # <<<<<<<<<<<<<<
 *     cdef object _filename
 *     cdef BGZF * fastqfile
 */
struct __pyx_obj_5pysam_9libcfaidx_FastxFile {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxFile *__pyx_vtab;
  PyObject *_filename;
  BGZF *fastqfile;
  kseq_t *entry;
  int persist;
  int is_remote;
};


/* "pysam/libcfaidx.pxd":73
 * 
 * # Compatibility Layer for pysam 0.8.1
 * cdef class FastqFile(FastxFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_5pysam_9libcfaidx_FastqFile {
  struct __pyx_obj_5pysam_9libcfaidx_FastxFile __pyx_base;
};


/* "pysam/libcfaidx.pxd":78
 * 
 * # Compatibility Layer for pysam < 0.8
 * cdef class Fastafile(FastaFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_5pysam_9libcfaidx_Fastafile {
  struct __pyx_obj_5pysam_9libcfaidx_FastaFile __pyx_base;
};


/* "pysam/libcalignedsegment.pxd":34
 * 
 * # Note: need to declare all C fields and methods here
 * cdef class AlignedSegment:             # <<<<<<<<<<<<<<
 * 
 *     # object that this AlignedSegment represents
 */
struct __pyx_obj_5pysam_18libcalignedsegment_AlignedSegment {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pysam_18libcalignedsegment_AlignedSegment *__pyx_vtab;
  bam1_t *_delegate;
  struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentHeader *header;
  PyObject *cache_query_qualities;
  PyObject *cache_query_alignment_qualities;
  PyObject *cache_query_sequence;
  PyObject *cache_query_alignment_sequence;
};


/* "pysam/libcalignedsegment.pxd":65
 * 
 * 
 * cdef class PileupColumn:             # <<<<<<<<<<<<<<
 *     cdef const bam_pileup1_t ** plp
 *     cdef int tid
 */
struct __pyx_obj_5pysam_18libcalignedsegment_PileupColumn {
  PyObject_HEAD
  bam_pileup1_t const **plp;
  int tid;
  int pos;
  int n_pu;
  struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentHeader *header;
  uint32_t min_base_quality;
  kstring_t buf;
  char *reference_sequence;
};


/* "pysam/libcalignedsegment.pxd":75
 *     cdef char * reference_sequence
 * 
 * cdef class PileupRead:             # <<<<<<<<<<<<<<
 *     cdef int32_t  _qpos
 *     cdef AlignedSegment _alignment
 */
struct __pyx_obj_5pysam_18libcalignedsegment_PileupRead {
  PyObject_HEAD
  int32_t _qpos;
  struct __pyx_obj_5pysam_18libcalignedsegment_AlignedSegment *_alignment;
  int _indel;
  int _level;
  uint32_t _is_del;
  uint32_t _is_head;
  uint32_t _is_tail;
  uint32_t _is_refskip;
};


/* "pysam/libcalignmentfile.pxd":41
 * 
 * 
 * cdef class AlignmentHeader(object):             # <<<<<<<<<<<<<<
 *     cdef bam_hdr_t *ptr
 * 
 */
struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentHeader {
  PyObject_HEAD
  bam_hdr_t *ptr;
};


/* "pysam/libcalignmentfile.pxd":44
 *     cdef bam_hdr_t *ptr
 * 
 * cdef class AlignmentFile(HTSFile):             # <<<<<<<<<<<<<<
 *     cdef readonly object reference_filename
 *     cdef readonly AlignmentHeader header
 */
struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile {
  struct __pyx_obj_5pysam_10libchtslib_HTSFile __pyx_base;
  PyObject *reference_filename;
  struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentHeader *header;
  hts_idx_t *index;
  bam1_t *b;
};


/* "pysam/libcalignmentfile.pxd":61
 * 
 * 
 * cdef class IteratorRow:             # <<<<<<<<<<<<<<
 *     cdef int retval
 *     cdef bam1_t * b
 */
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRow {
  PyObject_HEAD
  int retval;
  bam1_t *b;
  struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *samfile;
  htsFile *htsfile;
  hts_idx_t *index;
  struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentHeader *header;
  int owns_samfile;
};


/* "pysam/libcalignmentfile.pxd":71
 * 
 * 
 * cdef class IteratorRowRegion(IteratorRow):             # <<<<<<<<<<<<<<
 *     cdef hts_itr_t * iter
 *     cdef bam1_t * getCurrent(self)
 */
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowRegion {
  struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRow __pyx_base;
  struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowRegion *__pyx_vtab;
  hts_itr_t *iter;
};


/* "pysam/libcalignmentfile.pxd":77
 * 
 * 
 * cdef class IteratorRowHead(IteratorRow):             # <<<<<<<<<<<<<<
 *     cdef int max_rows
 *     cdef int current_row
 */
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowHead {
  struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRow __pyx_base;
  struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowHead *__pyx_vtab;
  int max_rows;
  int current_row;
};


/* "pysam/libcalignmentfile.pxd":84
 * 
 * 
 * cdef class IteratorRowAll(IteratorRow):             # <<<<<<<<<<<<<<
 *     cdef bam1_t * getCurrent(self)
 *     cdef int cnext(self)
 */
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowAll {
  struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRow __pyx_base;
  struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowAll *__pyx_vtab;
};


/* "pysam/libcalignmentfile.pxd":89
 * 
 * 
 * cdef class IteratorRowAllRefs(IteratorRow):             # <<<<<<<<<<<<<<
 *     cdef int         tid
 *     cdef IteratorRowRegion rowiter
 */
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowAllRefs {
  struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRow __pyx_base;
  int tid;
  struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowRegion *rowiter;
};


/* "pysam/libcalignmentfile.pxd":94
 * 
 * 
 * cdef class IteratorRowSelection(IteratorRow):             # <<<<<<<<<<<<<<
 *     cdef int current_pos
 *     cdef positions
 */
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowSelection {
  struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRow __pyx_base;
  struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowSelection *__pyx_vtab;
  int current_pos;
  PyObject *positions;
};


/* "pysam/libcalignmentfile.pxd":101
 * 
 * 
 * cdef class IteratorColumn:             # <<<<<<<<<<<<<<
 * 
 *     # result of the last plbuf_push
 */
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumn {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumn *__pyx_vtab;
  struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowRegion *iter;
  int tid;
  int pos;
  int n_plp;
  uint32_t min_base_quality;
  bam_pileup1_t const *plp;
  bam_mplp_t pileup_iter;
  __pyx_t_5pysam_17libcalignmentfile___iterdata iterdata;
  struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *samfile;
  struct __pyx_obj_5pysam_9libcfaidx_FastaFile *fastafile;
  PyObject *stepper;
  int max_depth;
  int ignore_overlaps;
};


/* "pysam/libcalignmentfile.pxd":133
 * 
 * 
 * cdef class IteratorColumnRegion(IteratorColumn):             # <<<<<<<<<<<<<<
 *     cdef int start
 *     cdef int stop
 */
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumnRegion {
  struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumn __pyx_base;
  int start;
  int stop;
  int truncate;
};


/* "pysam/libcalignmentfile.pxd":139
 * 
 * 
 * cdef class IteratorColumnAllRefs(IteratorColumn):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumnAllRefs {
  struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumn __pyx_base;
};


/* "pysam/libcalignmentfile.pxd":143
 * 
 * 
 * cdef class IteratorColumnAll(IteratorColumn):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumnAll {
  struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumn __pyx_base;
};


/* "pysam/libcalignmentfile.pxd":147
 * 
 * 
 * cdef class IndexedReads:             # <<<<<<<<<<<<<<
 *     cdef AlignmentFile samfile
 *     cdef htsFile * htsfile
 */
struct __pyx_obj_5pysam_17libcalignmentfile_IndexedReads {
  PyObject_HEAD
  struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *samfile;
  htsFile *htsfile;
  PyObject *index;
  int owns_samfile;
  struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentHeader *header;
};


/* "pysam/libcbcf.pxd":38
 * 
 * 
 * cdef class VariantHeader(object):             # <<<<<<<<<<<<<<
 *     cdef bcf_hdr_t *ptr
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantHeader {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pysam_7libcbcf_VariantHeader *__pyx_vtab;
  bcf_hdr_t *ptr;
};


/* "pysam/libcbcf.pxd":46
 * 
 * 
 * cdef class VariantHeaderRecord(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 *     cdef bcf_hrec_t *ptr
 */
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderRecord {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
  bcf_hrec_t *ptr;
};


/* "pysam/libcbcf.pxd":51
 * 
 * 
 * cdef class VariantHeaderRecords(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderRecords {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
};


/* "pysam/libcbcf.pxd":55
 * 
 * 
 * cdef class VariantHeaderContigs(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderContigs {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
};


/* "pysam/libcbcf.pxd":59
 * 
 * 
 * cdef class VariantHeaderSamples(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderSamples {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
};


/* "pysam/libcbcf.pxd":63
 * 
 * 
 * cdef class VariantContig(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 *     cdef int id
 */
struct __pyx_obj_5pysam_7libcbcf_VariantContig {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
  int id;
};


/* "pysam/libcbcf.pxd":68
 * 
 * 
 * cdef class VariantMetadata(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 *     cdef int type
 */
struct __pyx_obj_5pysam_7libcbcf_VariantMetadata {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
  int type;
  int id;
};


/* "pysam/libcbcf.pxd":74
 * 
 * 
 * cdef class VariantHeaderMetadata(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 *     cdef int32_t type
 */
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderMetadata {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
  int32_t type;
};


/* "pysam/libcbcf.pxd":79
 * 
 * 
 * cdef class VariantRecord(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 *     cdef bcf1_t *ptr
 */
struct __pyx_obj_5pysam_7libcbcf_VariantRecord {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
  bcf1_t *ptr;
};


/* "pysam/libcbcf.pxd":84
 * 
 * 
 * cdef class VariantRecordFilter(object):             # <<<<<<<<<<<<<<
 *     cdef VariantRecord record
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantRecordFilter {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantRecord *record;
};


/* "pysam/libcbcf.pxd":88
 * 
 * 
 * cdef class VariantRecordFormat(object):             # <<<<<<<<<<<<<<
 *     cdef VariantRecord record
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantRecordFormat {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantRecord *record;
};


/* "pysam/libcbcf.pxd":92
 * 
 * 
 * cdef class VariantRecordInfo(object):             # <<<<<<<<<<<<<<
 *     cdef VariantRecord record
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantRecordInfo {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantRecord *record;
};


/* "pysam/libcbcf.pxd":96
 * 
 * 
 * cdef class VariantRecordSamples(object):             # <<<<<<<<<<<<<<
 *     cdef VariantRecord record
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantRecordSamples {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantRecord *record;
};


/* "pysam/libcbcf.pxd":100
 * 
 * 
 * cdef class VariantRecordSample(object):             # <<<<<<<<<<<<<<
 *     cdef VariantRecord record
 *     cdef readonly int32_t index
 */
struct __pyx_obj_5pysam_7libcbcf_VariantRecordSample {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantRecord *record;
  int32_t index;
};


/* "pysam/libcbcf.pxd":105
 * 
 * 
 * cdef class BaseIndex(object):             # <<<<<<<<<<<<<<
 *     cdef tuple refs
 *     cdef dict refmap
 */
struct __pyx_obj_5pysam_7libcbcf_BaseIndex {
  PyObject_HEAD
  PyObject *refs;
  PyObject *refmap;
};


/* "pysam/libcbcf.pxd":110
 * 
 * 
 * cdef class BCFIndex(BaseIndex):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 *     cdef hts_idx_t *ptr
 */
struct __pyx_obj_5pysam_7libcbcf_BCFIndex {
  struct __pyx_obj_5pysam_7libcbcf_BaseIndex __pyx_base;
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
  hts_idx_t *ptr;
};


/* "pysam/libcbcf.pxd":115
 * 
 * 
 * cdef class TabixIndex(BaseIndex):             # <<<<<<<<<<<<<<
 *     cdef tbx_t *ptr
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_TabixIndex {
  struct __pyx_obj_5pysam_7libcbcf_BaseIndex __pyx_base;
  tbx_t *ptr;
};


/* "pysam/libcbcf.pxd":119
 * 
 * 
 * cdef class BaseIterator(object):             # <<<<<<<<<<<<<<
 *     cdef VariantFile bcf
 *     cdef hts_itr_t  *iter
 */
struct __pyx_obj_5pysam_7libcbcf_BaseIterator {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantFile *bcf;
  hts_itr_t *iter;
};


/* "pysam/libcbcf.pxd":124
 * 
 * 
 * cdef class BCFIterator(BaseIterator):             # <<<<<<<<<<<<<<
 *     cdef BCFIndex index
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_BCFIterator {
  struct __pyx_obj_5pysam_7libcbcf_BaseIterator __pyx_base;
  struct __pyx_obj_5pysam_7libcbcf_BCFIndex *index;
};


/* "pysam/libcbcf.pxd":128
 * 
 * 
 * cdef class TabixIterator(BaseIterator):             # <<<<<<<<<<<<<<
 *     cdef TabixIndex index
 *     cdef kstring_t line_buffer
 */
struct __pyx_obj_5pysam_7libcbcf_TabixIterator {
  struct __pyx_obj_5pysam_7libcbcf_BaseIterator __pyx_base;
  struct __pyx_obj_5pysam_7libcbcf_TabixIndex *index;
  kstring_t line_buffer;
};


/* "pysam/libcbcf.pxd":133
 * 
 * 
 * cdef class VariantFile(HTSFile):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader  header
 *     cdef readonly BaseIndex      index
 */
struct __pyx_obj_5pysam_7libcbcf_VariantFile {
  struct __pyx_obj_5pysam_10libchtslib_HTSFile __pyx_base;
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
  struct __pyx_obj_5pysam_7libcbcf_BaseIndex *index;
  int drop_samples;
  int is_reading;
  int header_written;
};


/* "variant.pxd":6
 * from pysam.libcbcf cimport VariantRecord, VariantFile
 * 
 * cdef class NullVariant:             # <<<<<<<<<<<<<<
 *     cdef readonly str chrom
 *     cdef readonly str ref, alt
 */
struct __pyx_obj_4aldy_9indelpost_7variant_NullVariant {
  PyObject_HEAD
  PyObject *chrom;
  PyObject *ref;
  PyObject *alt;
  int pos;
  struct __pyx_obj_5pysam_9libcfaidx_FastaFile *reference;
};


/* "variant.pxd":13
 * 
 * 
 * cdef class Variant:             # <<<<<<<<<<<<<<
 *     cdef public str chrom, ref, alt,
 *     cdef public int pos
 */
struct __pyx_obj_4aldy_9indelpost_7variant_Variant {
  PyObject_HEAD
  struct __pyx_vtabstruct_4aldy_9indelpost_7variant_Variant *__pyx_vtab;
  PyObject *chrom;
  PyObject *ref;
  PyObject *alt;
  int pos;
  struct __pyx_obj_5pysam_9libcfaidx_FastaFile *reference;
  PyObject *_chrom;
  struct __pyx_obj_5pysam_7libcbcf_VariantFile *vcf;
  struct __pyx_obj_5pysam_7libcbcf_VariantRecord *rec;
  struct __pyx_obj_5pysam_7libcbcf_VariantRecord *hit;
};


/* "contig.pxd":4
 * from .variant cimport Variant
 * 
 * cdef class FailedContig:             # <<<<<<<<<<<<<<
 *     cdef public bint target_not_found
 *     cdef public bint is_low_quality
 */
struct __pyx_obj_4aldy_9indelpost_6contig_FailedContig {
  PyObject_HEAD
  int target_not_found;
  int is_low_quality;
  int failed_anyway;
};


/* "contig.pxd":9
 *     cdef public bint failed_anyway
 * 
 * cdef class Contig:             # <<<<<<<<<<<<<<
 *     cdef Variant target
 *     cdef list pileup, targetpileup
 */
struct __pyx_obj_4aldy_9indelpost_6contig_Contig {
  PyObject_HEAD
  struct __pyx_obj_4aldy_9indelpost_7variant_Variant *target;
  PyObject *pileup;
  PyObject *targetpileup;
  int qc_passed;
  int failed;
  int donwsample_lim;
  PyObject *lt_genomic_index;
  PyObject *rt_genomic_index;
  int start;
  int end;
  PyObject *lt_reference_seq;
  PyObject *rt_reference_seq;
  PyObject *lt_target_block_reference_seq;
  PyObject *rt_target_block_reference_seq;
  int is_non_complex_at_target_pos;
  PyObject *target_ref;
  PyObject *target_alt;
  double low_consensus_thresh;
  int is_target_right_aligned;
  PyObject *contig_dict;
  PyObject *lt_consensus_seq;
  PyObject *rt_consensus_seq;
  PyObject *indel_seq;
  PyObject *lt_target_block_consensus_seq;
  PyObject *rt_target_block_consensus_seq;
  PyObject *lt_consensus_scores;
  PyObject *rt_consensus_scores;
  PyObject *mismatches;
  PyObject *non_target_indels;
  PyObject *gaps;
  PyObject *lt_target_block_consensus_scores;
  PyObject *rt_target_block_consensus_scores;
  int mapq;
  int lt_end_pos;
  double low_qual_mapping_rate;
  PyObject *qc_stats;
  PyObject *splice_pattern;
};


/* "local_reference.pxd":4
 * 
 * 
 * cdef class UnsplicedLocalReference:             # <<<<<<<<<<<<<<
 *     cdef:
 *         str chrom, ref, alt, unspliced_local_reference
 */
struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference {
  PyObject_HEAD
  PyObject *chrom;
  PyObject *ref;
  PyObject *alt;
  PyObject *unspliced_local_reference;
  int pos;
  int window;
  int local_ref_start;
  int ref_len;
  struct __pyx_obj_5pysam_9libcfaidx_FastaFile *reference;
  int left_len;
};


/* "aldy/indelpost/varaln.pxd":9
 * from .local_reference cimport UnsplicedLocalReference
 * 
 * cdef class VariantAlignment:             # <<<<<<<<<<<<<<
 *     cdef Variant target, __target, second_target
 *     cdef readonly AlignmentFile bam
 */
struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment {
  PyObject_HEAD
  struct __pyx_vtabstruct_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_vtab;
  struct __pyx_obj_4aldy_9indelpost_7variant_Variant *target;
  struct __pyx_obj_4aldy_9indelpost_7variant_Variant *__pyx___target;
  struct __pyx_obj_4aldy_9indelpost_7variant_Variant *second_target;
  struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *bam;
  int window;
  int retarget_window;
  int mapqthresh;
  int downsamplethresh;
  int basequalthresh;
  int match_score;
  int _observed_pos;
  int mismatch_penalty;
  int gap_open_penalty;
  int gap_extension_penalty;
  float retarget_cutoff;
  float __pyx___sample_factor;
  int exclude_duplicates;
  int exact_match_for_shiftable;
  int auto_adjust_extension_penalty;
  int no_realignment;
  PyObject *__pyx___pileup;
  PyObject *is_spurious_overhang;
  PyObject *is_complex_input;
  struct __pyx_obj_4aldy_9indelpost_6contig_Contig *contig;
  struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference *unspliced_local_reference;
};


/* "aldy/indelpost/varaln.pyx":171
 *         self.__pileup, self.contig = self.__parse_pileup()
 * 
 *     cdef __parse_pileup(self, Contig contig=None, bint retargeted=False, bint skip_read_end_check=False):             # <<<<<<<<<<<<<<
 *         """Dictize reads for target indel and make target indel template by consensus
 * 
 */
struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup {
  PyObject_HEAD
  int __pyx_v_target_pos;
};


/* "aldy/indelpost/varaln.pyx":549
 * 
 * 
 *     def fetch_reads(self, how='target'):             # <<<<<<<<<<<<<<
 *         """returns a `list <https://docs.python.org/3/library/stdtypes.html#list>`__
 *         of `pysam.AlignedSegment <https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignedSegment>`__ objects,
 */
struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads {
  PyObject_HEAD
  struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self;
};


/* "aldy/indelpost/varaln.pyx":565
 *         elif how == "non_target":
 *             pos, indel_len = self._observed_pos, len(self.target.indel_seq)
 *             r_pos = max(v.pos for v in self.target.generate_equivalents())             # <<<<<<<<<<<<<<
 *             margin = r_pos - pos
 *             del_len = indel_len if self.target.is_del else 0
 */
struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr {
  PyObject_HEAD
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads *__pyx_outer_scope;
  PyObject *__pyx_v_v;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "aldy/indelpost/varaln.pyx":576
 * 
 * 
 *     def count_alleles(             # <<<<<<<<<<<<<<
 *         self, fwrv=False, by_fragment=False, three_class=False, estimated_count=False, quality_window=None, quality_threshold=None
 *     ):
 */
struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles {
  PyObject_HEAD
  struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self;
};


/* "aldy/indelpost/varaln.pyx":603
 *         pos, indel_len = self._observed_pos, len(self.target.indel_seq)
 * 
 *         r_pos = max(v.pos for v in self.__target.generate_equivalents())             # <<<<<<<<<<<<<<
 * 
 *         margin = r_pos - pos
 */
struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr {
  PyObject_HEAD
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles *__pyx_outer_scope;
  PyObject *__pyx_v_v;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "aldy/indelpost/varaln.pyx":999
 *     return right_aligner(read, target)
 * 
 * def right_aligner(read, target):             # <<<<<<<<<<<<<<
 *     """Right align indels around splice site"""
 * 
 */
struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner {
  PyObject_HEAD
  PyObject *__pyx_v_right_aligned_vars;
};


/* "aldy/indelpost/varaln.pyx":1056
 *                                 ).generate_equivalents()
 * 
 *             diff = max(v.pos for v in right_aligned_vars) - ref_pos + 1             # <<<<<<<<<<<<<<
 *             if diff > 0:
 *                 new_cigar += [str(diff) + "M", str(event_len) + event, str(nxt_event_len - diff) + "M"]
 */
struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr {
  PyObject_HEAD
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner *__pyx_outer_scope;
  PyObject *__pyx_v_v;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};



/* "pysam/libchtslib.pxd":2680
 * 
 * 
 * cdef class HTSFile(object):             # <<<<<<<<<<<<<<
 *     cdef          htsFile *htsfile       # pointer to htsFile structure
 *     cdef          int64_t start_offset   # BGZF offset of first record
 */

struct __pyx_vtabstruct_5pysam_10libchtslib_HTSFile {
  htsFile *(*_open_htsfile)(struct __pyx_obj_5pysam_10libchtslib_HTSFile *);
};
static struct __pyx_vtabstruct_5pysam_10libchtslib_HTSFile *__pyx_vtabptr_5pysam_10libchtslib_HTSFile;


/* "pysam/libcfaidx.pxd":37
 *                     int * dret)
 * 
 * cdef class FastaFile:             # <<<<<<<<<<<<<<
 *     cdef bint is_remote
 *     cdef object _filename, _references, _lengths, reference2length
 */

struct __pyx_vtabstruct_5pysam_9libcfaidx_FastaFile {
  char *(*_fetch)(struct __pyx_obj_5pysam_9libcfaidx_FastaFile *, char *, int, int, int *);
};
static struct __pyx_vtabstruct_5pysam_9libcfaidx_FastaFile *__pyx_vtabptr_5pysam_9libcfaidx_FastaFile;


/* "pysam/libcfaidx.pxd":45
 * 
 * 
 * cdef class FastqProxy:             # <<<<<<<<<<<<<<
 *     cdef kseq_t * _delegate
 *     cdef cython.str to_string(self)
 */

struct __pyx_vtabstruct_5pysam_9libcfaidx_FastqProxy {
  PyObject *(*to_string)(struct __pyx_obj_5pysam_9libcfaidx_FastqProxy *);
  PyObject *(*tostring)(struct __pyx_obj_5pysam_9libcfaidx_FastqProxy *);
  arrayobject *(*get_quality_array)(struct __pyx_obj_5pysam_9libcfaidx_FastqProxy *, int __pyx_skip_dispatch, struct __pyx_opt_args_5pysam_9libcfaidx_10FastqProxy_get_quality_array *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5pysam_9libcfaidx_FastqProxy *__pyx_vtabptr_5pysam_9libcfaidx_FastqProxy;


/* "pysam/libcfaidx.pxd":52
 * 
 * 
 * cdef class FastxRecord:             # <<<<<<<<<<<<<<
 *     """
 *     Python container for pysam.libcfaidx.FastqProxy with persistence.
 */

struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxRecord {
  PyObject *(*to_string)(struct __pyx_obj_5pysam_9libcfaidx_FastxRecord *);
  PyObject *(*tostring)(struct __pyx_obj_5pysam_9libcfaidx_FastxRecord *);
  arrayobject *(*get_quality_array)(struct __pyx_obj_5pysam_9libcfaidx_FastxRecord *, int __pyx_skip_dispatch, struct __pyx_opt_args_5pysam_9libcfaidx_11FastxRecord_get_quality_array *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxRecord *__pyx_vtabptr_5pysam_9libcfaidx_FastxRecord;


/* "pysam/libcfaidx.pxd":61
 *     cpdef array.array get_quality_array(self, int offset=*)
 * 
 * cdef class FastxFile:             # <<<<<<<<<<<<<<
 *     cdef object _filename
 *     cdef BGZF * fastqfile
 */

struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxFile {
  kseq_t *(*getCurrent)(struct __pyx_obj_5pysam_9libcfaidx_FastxFile *);
  int (*cnext)(struct __pyx_obj_5pysam_9libcfaidx_FastxFile *);
};
static struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxFile *__pyx_vtabptr_5pysam_9libcfaidx_FastxFile;


/* "pysam/libcfaidx.pxd":73
 * 
 * # Compatibility Layer for pysam 0.8.1
 * cdef class FastqFile(FastxFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_5pysam_9libcfaidx_FastqFile {
  struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxFile __pyx_base;
};
static struct __pyx_vtabstruct_5pysam_9libcfaidx_FastqFile *__pyx_vtabptr_5pysam_9libcfaidx_FastqFile;


/* "pysam/libcfaidx.pxd":78
 * 
 * # Compatibility Layer for pysam < 0.8
 * cdef class Fastafile(FastaFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_5pysam_9libcfaidx_Fastafile {
  struct __pyx_vtabstruct_5pysam_9libcfaidx_FastaFile __pyx_base;
};
static struct __pyx_vtabstruct_5pysam_9libcfaidx_Fastafile *__pyx_vtabptr_5pysam_9libcfaidx_Fastafile;


/* "pysam/libcalignedsegment.pxd":34
 * 
 * # Note: need to declare all C fields and methods here
 * cdef class AlignedSegment:             # <<<<<<<<<<<<<<
 * 
 *     # object that this AlignedSegment represents
 */

struct __pyx_vtabstruct_5pysam_18libcalignedsegment_AlignedSegment {
  PyObject *(*set_tag)(struct __pyx_obj_5pysam_18libcalignedsegment_AlignedSegment *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5pysam_18libcalignedsegment_14AlignedSegment_set_tag *__pyx_optional_args);
  PyObject *(*get_tag)(struct __pyx_obj_5pysam_18libcalignedsegment_AlignedSegment *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5pysam_18libcalignedsegment_14AlignedSegment_get_tag *__pyx_optional_args);
  PyObject *(*has_tag)(struct __pyx_obj_5pysam_18libcalignedsegment_AlignedSegment *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*to_string)(struct __pyx_obj_5pysam_18libcalignedsegment_AlignedSegment *, int __pyx_skip_dispatch);
  PyObject *(*tostring)(struct __pyx_obj_5pysam_18libcalignedsegment_AlignedSegment *, int __pyx_skip_dispatch, struct __pyx_opt_args_5pysam_18libcalignedsegment_14AlignedSegment_tostring *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5pysam_18libcalignedsegment_AlignedSegment *__pyx_vtabptr_5pysam_18libcalignedsegment_AlignedSegment;


/* "pysam/libcalignmentfile.pxd":44
 *     cdef bam_hdr_t *ptr
 * 
 * cdef class AlignmentFile(HTSFile):             # <<<<<<<<<<<<<<
 *     cdef readonly object reference_filename
 *     cdef readonly AlignmentHeader header
 */

struct __pyx_vtabstruct_5pysam_17libcalignmentfile_AlignmentFile {
  struct __pyx_vtabstruct_5pysam_10libchtslib_HTSFile __pyx_base;
  bam1_t *(*getCurrent)(struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *);
  int (*cnext)(struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *);
  int (*write)(struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *, struct __pyx_obj_5pysam_18libcalignedsegment_AlignedSegment *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5pysam_17libcalignmentfile_AlignmentFile *__pyx_vtabptr_5pysam_17libcalignmentfile_AlignmentFile;


/* "pysam/libcalignmentfile.pxd":71
 * 
 * 
 * cdef class IteratorRowRegion(IteratorRow):             # <<<<<<<<<<<<<<
 *     cdef hts_itr_t * iter
 *     cdef bam1_t * getCurrent(self)
 */

struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowRegion {
  bam1_t *(*getCurrent)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowRegion *);
  int (*cnext)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowRegion *);
};
static struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowRegion *__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorRowRegion;


/* "pysam/libcalignmentfile.pxd":77
 * 
 * 
 * cdef class IteratorRowHead(IteratorRow):             # <<<<<<<<<<<<<<
 *     cdef int max_rows
 *     cdef int current_row
 */

struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowHead {
  bam1_t *(*getCurrent)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowHead *);
  int (*cnext)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowHead *);
};
static struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowHead *__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorRowHead;


/* "pysam/libcalignmentfile.pxd":84
 * 
 * 
 * cdef class IteratorRowAll(IteratorRow):             # <<<<<<<<<<<<<<
 *     cdef bam1_t * getCurrent(self)
 *     cdef int cnext(self)
 */

struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowAll {
  bam1_t *(*getCurrent)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowAll *);
  int (*cnext)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowAll *);
};
static struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowAll *__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorRowAll;


/* "pysam/libcalignmentfile.pxd":94
 * 
 * 
 * cdef class IteratorRowSelection(IteratorRow):             # <<<<<<<<<<<<<<
 *     cdef int current_pos
 *     cdef positions
 */

struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowSelection {
  bam1_t *(*getCurrent)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowSelection *);
  int (*cnext)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowSelection *);
};
static struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowSelection *__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorRowSelection;


/* "pysam/libcalignmentfile.pxd":101
 * 
 * 
 * cdef class IteratorColumn:             # <<<<<<<<<<<<<<
 * 
 *     # result of the last plbuf_push
 */

struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumn {
  int (*cnext)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumn *);
  char *(*get_sequence)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumn *);
  PyObject *(*_setup_iterator)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumn *, int, int, int, struct __pyx_opt_args_5pysam_17libcalignmentfile_14IteratorColumn__setup_iterator *__pyx_optional_args);
  PyObject *(*_setup_raw_rest_iterator)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumn *);
  PyObject *(*reset)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumn *, PyObject *, PyObject *, PyObject *);
  PyObject *(*_free_pileup_iter)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumn *);
  char *(*getSequence)(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumn *);
};
static struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumn *__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorColumn;


/* "pysam/libcalignmentfile.pxd":133
 * 
 * 
 * cdef class IteratorColumnRegion(IteratorColumn):             # <<<<<<<<<<<<<<
 *     cdef int start
 *     cdef int stop
 */

struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumnRegion {
  struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumn __pyx_base;
};
static struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumnRegion *__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorColumnRegion;


/* "pysam/libcalignmentfile.pxd":139
 * 
 * 
 * cdef class IteratorColumnAllRefs(IteratorColumn):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumnAllRefs {
  struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumn __pyx_base;
};
static struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumnAllRefs *__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorColumnAllRefs;


/* "pysam/libcalignmentfile.pxd":143
 * 
 * 
 * cdef class IteratorColumnAll(IteratorColumn):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumnAll {
  struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumn __pyx_base;
};
static struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumnAll *__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorColumnAll;


/* "pysam/libcbcf.pxd":38
 * 
 * 
 * cdef class VariantHeader(object):             # <<<<<<<<<<<<<<
 *     cdef bcf_hdr_t *ptr
 * 
 */

struct __pyx_vtabstruct_5pysam_7libcbcf_VariantHeader {
  PyObject *(*_add_sample)(struct __pyx_obj_5pysam_7libcbcf_VariantHeader *, PyObject *);
  PyObject *(*_hdr_sync)(struct __pyx_obj_5pysam_7libcbcf_VariantHeader *);
  PyObject *(*_subset_samples)(struct __pyx_obj_5pysam_7libcbcf_VariantHeader *, PyObject *);
};
static struct __pyx_vtabstruct_5pysam_7libcbcf_VariantHeader *__pyx_vtabptr_5pysam_7libcbcf_VariantHeader;


/* "pysam/libcbcf.pxd":133
 * 
 * 
 * cdef class VariantFile(HTSFile):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader  header
 *     cdef readonly BaseIndex      index
 */

struct __pyx_vtabstruct_5pysam_7libcbcf_VariantFile {
  struct __pyx_vtabstruct_5pysam_10libchtslib_HTSFile __pyx_base;
  int (*write)(struct __pyx_obj_5pysam_7libcbcf_VariantFile *, struct __pyx_obj_5pysam_7libcbcf_VariantRecord *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5pysam_7libcbcf_VariantFile *__pyx_vtabptr_5pysam_7libcbcf_VariantFile;


/* "variant.pxd":13
 * 
 * 
 * cdef class Variant:             # <<<<<<<<<<<<<<
 *     cdef public str chrom, ref, alt,
 *     cdef public int pos
 */

struct __pyx_vtabstruct_4aldy_9indelpost_7variant_Variant {
  PyObject *(*__pyx___validate)(struct __pyx_obj_4aldy_9indelpost_7variant_Variant *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4aldy_9indelpost_7variant_Variant *__pyx_vtabptr_4aldy_9indelpost_7variant_Variant;


/* "aldy/indelpost/varaln.pyx":41
 * random.seed(123)
 * 
 * cdef class VariantAlignment:             # <<<<<<<<<<<<<<
 *     """This class accepts the target indel as :class:`~indelpost.Variant`
 *     and the BAM file as `pysam.AlignmentFile <https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignmentFile>`__ .
 */

struct __pyx_vtabstruct_4aldy_9indelpost_6varaln_VariantAlignment {
  PyObject *(*__pyx___parse_pileup)(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *, struct __pyx_opt_args_4aldy_9indelpost_6varaln_16VariantAlignment___parse_pileup *__pyx_optional_args);
};
static struct __pyx_vtabstruct_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_vtabptr_4aldy_9indelpost_6varaln_VariantAlignment;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* py_abs.proto */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *num);
#define __Pyx_PyNumber_Absolute(x)\
    ((likely(PyLong_CheckExact(x))) ?\
         (likely(Py_SIZE(x) >= 0) ? (Py_INCREF(x), (x)) : __Pyx_PyLong_AbsNeg(x)) :\
         PyNumber_Absolute(x))
#else
#define __Pyx_PyNumber_Absolute(x)  PyNumber_Absolute(x)
#endif

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* dict_getitem_default.proto */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_SubtractCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_SubtractCObj(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_EqObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_EqObjC(op1, op2, floatval, inplace, zerodivision_check)\
    (PyObject_RichCompare(op1, op2, Py_EQ))
    #endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* ArrayAPI.proto */
#ifndef _ARRAYARRAY_H
#define _ARRAYARRAY_H
typedef struct arraydescr {
    int typecode;
    int itemsize;
    PyObject * (*getitem)(struct arrayobject *, Py_ssize_t);
    int (*setitem)(struct arrayobject *, Py_ssize_t, PyObject *);
#if PY_MAJOR_VERSION >= 3
    char *formats;
#endif
} arraydescr;
struct arrayobject {
    PyObject_HEAD
    Py_ssize_t ob_size;
    union {
        char *ob_item;
        float *as_floats;
        double *as_doubles;
        int *as_ints;
        unsigned int *as_uints;
        unsigned char *as_uchars;
        signed char *as_schars;
        char *as_chars;
        unsigned long *as_ulongs;
        long *as_longs;
#if PY_MAJOR_VERSION >= 3
        unsigned long long *as_ulonglongs;
        long long *as_longlongs;
#endif
        short *as_shorts;
        unsigned short *as_ushorts;
        Py_UNICODE *as_pyunicodes;
        void *as_voidptr;
    } data;
    Py_ssize_t allocated;
    struct arraydescr *ob_descr;
    PyObject *weakreflist;
#if PY_MAJOR_VERSION >= 3
        int ob_exports;
#endif
};
#ifndef NO_NEWARRAY_INLINE
static CYTHON_INLINE PyObject * newarrayobject(PyTypeObject *type, Py_ssize_t size,
    struct arraydescr *descr) {
    arrayobject *op;
    size_t nbytes;
    if (size < 0) {
        PyErr_BadInternalCall();
        return NULL;
    }
    nbytes = size * descr->itemsize;
    if (nbytes / descr->itemsize != (size_t)size) {
        return PyErr_NoMemory();
    }
    op = (arrayobject *) type->tp_alloc(type, 0);
    if (op == NULL) {
        return NULL;
    }
    op->ob_descr = descr;
    op->allocated = size;
    op->weakreflist = NULL;
    __Pyx_SET_SIZE(op, size);
    if (size <= 0) {
        op->data.ob_item = NULL;
    }
    else {
        op->data.ob_item = PyMem_NEW(char, nbytes);
        if (op->data.ob_item == NULL) {
            Py_DECREF(op);
            return PyErr_NoMemory();
        }
    }
    return (PyObject *) op;
}
#else
PyObject* newarrayobject(PyTypeObject *type, Py_ssize_t size,
    struct arraydescr *descr);
#endif
static CYTHON_INLINE int resize(arrayobject *self, Py_ssize_t n) {
    void *items = (void*) self->data.ob_item;
    PyMem_Resize(items, char, (size_t)(n * self->ob_descr->itemsize));
    if (items == NULL) {
        PyErr_NoMemory();
        return -1;
    }
    self->data.ob_item = (char*) items;
    __Pyx_SET_SIZE(self, n);
    self->allocated = n;
    return 0;
}
static CYTHON_INLINE int resize_smart(arrayobject *self, Py_ssize_t n) {
    void *items = (void*) self->data.ob_item;
    Py_ssize_t newsize;
    if (n < self->allocated && n*4 > self->allocated) {
        __Pyx_SET_SIZE(self, n);
        return 0;
    }
    newsize = n + (n / 2) + 1;
    if (newsize <= n) {
        PyErr_NoMemory();
        return -1;
    }
    PyMem_Resize(items, char, (size_t)(newsize * self->ob_descr->itemsize));
    if (items == NULL) {
        PyErr_NoMemory();
        return -1;
    }
    self->data.ob_item = (char*) items;
    __Pyx_SET_SIZE(self, n);
    self->allocated = newsize;
    return 0;
}
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_4aldy_9indelpost_6varaln_16VariantAlignment___parse_pileup(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self, struct __pyx_opt_args_4aldy_9indelpost_6varaln_16VariantAlignment___parse_pileup *__pyx_optional_args); /* proto*/

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cython' */

/* Module declarations from 'cpython.version' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'array' */

/* Module declarations from 'cpython.array' */
static PyTypeObject *__pyx_ptype_7cpython_5array_array = 0;
static CYTHON_INLINE int __pyx_f_7cpython_5array_extend_buffer(arrayobject *, char *, Py_ssize_t); /*proto*/

/* Module declarations from 'posix.types' */

/* Module declarations from 'pysam.libchtslib' */
static PyTypeObject *__pyx_ptype_5pysam_10libchtslib_HTSFile = 0;

/* Module declarations from 'pysam.libcfaidx' */
static PyTypeObject *__pyx_ptype_5pysam_9libcfaidx_FastaFile = 0;
static PyTypeObject *__pyx_ptype_5pysam_9libcfaidx_FastqProxy = 0;
static PyTypeObject *__pyx_ptype_5pysam_9libcfaidx_FastxRecord = 0;
static PyTypeObject *__pyx_ptype_5pysam_9libcfaidx_FastxFile = 0;
static PyTypeObject *__pyx_ptype_5pysam_9libcfaidx_FastqFile = 0;
static PyTypeObject *__pyx_ptype_5pysam_9libcfaidx_Fastafile = 0;

/* Module declarations from 'pysam.libcalignedsegment' */
static PyTypeObject *__pyx_ptype_5pysam_18libcalignedsegment_AlignedSegment = 0;
static PyTypeObject *__pyx_ptype_5pysam_18libcalignedsegment_PileupColumn = 0;
static PyTypeObject *__pyx_ptype_5pysam_18libcalignedsegment_PileupRead = 0;

/* Module declarations from 'pysam.libcalignmentfile' */
static PyTypeObject *__pyx_ptype_5pysam_17libcalignmentfile_AlignmentHeader = 0;
static PyTypeObject *__pyx_ptype_5pysam_17libcalignmentfile_AlignmentFile = 0;
static PyTypeObject *__pyx_ptype_5pysam_17libcalignmentfile_IteratorRow = 0;
static PyTypeObject *__pyx_ptype_5pysam_17libcalignmentfile_IteratorRowRegion = 0;
static PyTypeObject *__pyx_ptype_5pysam_17libcalignmentfile_IteratorRowHead = 0;
static PyTypeObject *__pyx_ptype_5pysam_17libcalignmentfile_IteratorRowAll = 0;
static PyTypeObject *__pyx_ptype_5pysam_17libcalignmentfile_IteratorRowAllRefs = 0;
static PyTypeObject *__pyx_ptype_5pysam_17libcalignmentfile_IteratorRowSelection = 0;
static PyTypeObject *__pyx_ptype_5pysam_17libcalignmentfile_IteratorColumn = 0;
static PyTypeObject *__pyx_ptype_5pysam_17libcalignmentfile_IteratorColumnRegion = 0;
static PyTypeObject *__pyx_ptype_5pysam_17libcalignmentfile_IteratorColumnAllRefs = 0;
static PyTypeObject *__pyx_ptype_5pysam_17libcalignmentfile_IteratorColumnAll = 0;
static PyTypeObject *__pyx_ptype_5pysam_17libcalignmentfile_IndexedReads = 0;

/* Module declarations from 'pysam.libcbcf' */
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantHeader = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantHeaderRecord = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantHeaderRecords = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantHeaderContigs = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantHeaderSamples = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantContig = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantMetadata = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantHeaderMetadata = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantRecord = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantRecordFilter = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantRecordFormat = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantRecordInfo = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantRecordSamples = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantRecordSample = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_BaseIndex = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_BCFIndex = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_TabixIndex = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_BaseIterator = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_BCFIterator = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_TabixIterator = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantFile = 0;

/* Module declarations from 'aldy.indelpost.variant' */
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_7variant_NullVariant = 0;
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_7variant_Variant = 0;

/* Module declarations from 'aldy.indelpost.contig' */
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_6contig_FailedContig = 0;
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_6contig_Contig = 0;

/* Module declarations from 'aldy.indelpost.local_reference' */
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_15local_reference_UnsplicedLocalReference = 0;

/* Module declarations from 'aldy.indelpost.pileup' */
static PyObject *(*__pyx_f_4aldy_9indelpost_6pileup_make_pileup)(struct __pyx_obj_4aldy_9indelpost_7variant_Variant *, struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *, struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference *, int, int, int, int); /*proto*/

/* Module declarations from 'aldy.indelpost.utilities' */
static PyObject *(*__pyx_f_4aldy_9indelpost_9utilities_split)(PyObject *, PyObject *, int, int, int, int); /*proto*/

/* Module declarations from 'aldy.indelpost.varaln' */
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_6varaln_VariantAlignment = 0;
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup = 0;
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads = 0;
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr = 0;
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles = 0;
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr = 0;
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner = 0;
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr = 0;
static int __pyx_f_4aldy_9indelpost_6varaln_count_as_non_target(PyObject *, int, int, int); /*proto*/
static PyObject *__pyx_f_4aldy_9indelpost_6varaln_preprocess_for_contig_construction(struct __pyx_obj_4aldy_9indelpost_7variant_Variant *, struct __pyx_obj_4aldy_9indelpost_7variant_Variant *, PyObject *, struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference *, int, int, int, int, int); /*proto*/
#define __Pyx_MODULE_NAME "aldy.indelpost.varaln"
extern int __pyx_module_is_main_aldy__indelpost__varaln;
int __pyx_module_is_main_aldy__indelpost__varaln = 0;

/* Implementation of 'aldy.indelpost.varaln' */
static PyObject *__pyx_builtin_sum;
static PyObject *__pyx_builtin_max;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_sorted;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_MemoryError;
static const char __pyx_k_A[] = "A";
static const char __pyx_k_D[] = "D";
static const char __pyx_k_H[] = "H";
static const char __pyx_k_I[] = "I";
static const char __pyx_k_M[] = "M";
static const char __pyx_k_N[] = "N";
static const char __pyx_k_P[] = "P";
static const char __pyx_k_S[] = "S";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_h[] = "h";
static const char __pyx_k_i[] = "i";
static const char __pyx_k__6[] = "";
static const char __pyx_k_hs[] = "hs";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_aln[] = "_aln";
static const char __pyx_k_alt[] = "alt";
static const char __pyx_k_bam[] = "bam";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_how[] = "how";
static const char __pyx_k_idx[] = "idx";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_pos[] = "pos";
static const char __pyx_k_ref[] = "ref";
static const char __pyx_k_res[] = "res";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_diff[] = "diff";
static const char __pyx_k_fwrv[] = "fwrv";
static const char __pyx_k_grid[] = "grid";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_read[] = "read";
static const char __pyx_k_seed[] = "seed";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_sort[] = "sort";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_align[] = "align";
static const char __pyx_k_chrom[] = "chrom";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_event[] = "event";
static const char __pyx_k_fetch[] = "fetch";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_local[] = "local";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_nxt_c[] = "nxt_c";
static const char __pyx_k_phase[] = "phase";
static const char __pyx_k_score[] = "score";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_contig[] = "contig";
static const char __pyx_k_failed[] = "failed";
static const char __pyx_k_greedy[] = "greedy";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_is_del[] = "is_del";
static const char __pyx_k_is_ins[] = "is_ins";
static const char __pyx_k_localn[] = "localn";
static const char __pyx_k_lt_len[] = "lt_len";
static const char __pyx_k_lt_ref[] = "lt_ref";
static const char __pyx_k_lt_seq[] = "lt_seq";
static const char __pyx_k_median[] = "median";
static const char __pyx_k_pileup[] = "pileup";
static const char __pyx_k_random[] = "random";
static const char __pyx_k_read_2[] = "_read";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_rt_len[] = "rt_len";
static const char __pyx_k_rt_ref[] = "rt_ref";
static const char __pyx_k_rt_seq[] = "rt_seq";
static const char __pyx_k_sample[] = "sample";
static const char __pyx_k_scores[] = "scores";
static const char __pyx_k_sorted[] = "sorted";
static const char __pyx_k_target[] = "target";
static const char __pyx_k_window[] = "window";
static const char __pyx_k_within[] = "within";
static const char __pyx_k_aligner[] = "aligner";
static const char __pyx_k_alleles[] = "alleles";
static const char __pyx_k_aln_end[] = "aln_end";
static const char __pyx_k_complex[] = "complex";
static const char __pyx_k_del_seq[] = "del_seq";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_ins_seq[] = "ins_seq";
static const char __pyx_k_lt_qual[] = "lt_qual";
static const char __pyx_k_partial[] = "partial";
static const char __pyx_k_ref_pos[] = "ref_pos";
static const char __pyx_k_ref_seq[] = "ref_seq";
static const char __pyx_k_rt_qual[] = "rt_qual";
static const char __pyx_k_setRead[] = "setRead";
static const char __pyx_k_best_res[] = "best_res";
static const char __pyx_k_contig_2[] = "_contig";
static const char __pyx_k_covering[] = "covering";
static const char __pyx_k_gap_open[] = "gap_open";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_is_dirty[] = "is_dirty";
static const char __pyx_k_is_indel[] = "is_indel";
static const char __pyx_k_lt_cigar[] = "lt_cigar";
static const char __pyx_k_lt_flank[] = "lt_flank";
static const char __pyx_k_read_end[] = "read_end";
static const char __pyx_k_read_seq[] = "read_seq";
static const char __pyx_k_ref_move[] = "ref_move";
static const char __pyx_k_retarget[] = "retarget";
static const char __pyx_k_rt_cigar[] = "rt_cigar";
static const char __pyx_k_rt_flank[] = "rt_flank";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_softclip[] = "softclip";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_aln_start[] = "aln_start";
static const char __pyx_k_candidate[] = "candidate";
static const char __pyx_k_cigar_lst[] = "cigar_lst";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_event_len[] = "event_len";
static const char __pyx_k_functools[] = "functools";
static const char __pyx_k_gappedaln[] = "gappedaln";
static const char __pyx_k_indel_len[] = "indel_len";
static const char __pyx_k_indel_seq[] = "indel_seq";
static const char __pyx_k_is_target[] = "is_target";
static const char __pyx_k_lt_median[] = "lt_median";
static const char __pyx_k_new_cigar[] = "new_cigar";
static const char __pyx_k_normalize[] = "normalize";
static const char __pyx_k_nxt_event[] = "nxt_event";
static const char __pyx_k_query_pos[] = "query_pos";
static const char __pyx_k_read_name[] = "read_name";
static const char __pyx_k_read_qual[] = "read_qual";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_ref_start[] = "ref_start";
static const char __pyx_k_reference[] = "reference";
static const char __pyx_k_responses[] = "responses";
static const char __pyx_k_rt_median[] = "rt_median";
static const char __pyx_k_skip_next[] = "skip_next";
static const char __pyx_k_unspliced[] = "unspliced";
static const char __pyx_k_utilities[] = "utilities";
static const char __pyx_k_centrality[] = "centrality";
static const char __pyx_k_cigar_list[] = "cigar_list";
static const char __pyx_k_contig_seq[] = "contig_seq";
static const char __pyx_k_is_reverse[] = "is_reverse";
static const char __pyx_k_lt_ref_len[] = "lt_ref_len";
static const char __pyx_k_lt_seq_len[] = "lt_seq_len";
static const char __pyx_k_non_target[] = "non_target";
static const char __pyx_k_prev_event[] = "prev_event";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_query_move[] = "query_move";
static const char __pyx_k_read_start[] = "read_start";
static const char __pyx_k_rt_aln_pos[] = "rt_aln_pos";
static const char __pyx_k_rt_ref_len[] = "rt_ref_len";
static const char __pyx_k_rt_seq_len[] = "rt_seq_len";
static const char __pyx_k_target_pos[] = "target_pos";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_best_params[] = "best_params";
static const char __pyx_k_by_fragment[] = "by_fragment";
static const char __pyx_k_grid_search[] = "grid_search";
static const char __pyx_k_is_covering[] = "is_covering";
static const char __pyx_k_mapq_thresh[] = "mapq_thresh";
static const char __pyx_k_match_score[] = "match_score";
static const char __pyx_k_most_common[] = "most_common";
static const char __pyx_k_orig_target[] = "orig_target";
static const char __pyx_k_split_cigar[] = "split_cigar";
static const char __pyx_k_three_class[] = "three_class";
static const char __pyx_k_cigar_string[] = "cigar_string";
static const char __pyx_k_make_aligner[] = "make_aligner";
static const char __pyx_k_padding_base[] = "padding_base";
static const char __pyx_k_relative_pos[] = "relative_pos";
static const char __pyx_k_undetermined[] = "undetermined";
static const char __pyx_k_updated_read[] = "updated_read";
static const char __pyx_k_bypass_search[] = "bypass_search";
static const char __pyx_k_cigar_updated[] = "cigar_updated";
static const char __pyx_k_count_alleles[] = "count_alleles";
static const char __pyx_k_count_repeats[] = "count_repeats";
static const char __pyx_k_gap_extension[] = "gap_extension";
static const char __pyx_k_generate_grid[] = "generate_grid";
static const char __pyx_k_get_gap_ptrn2[] = "get_gap_ptrn2";
static const char __pyx_k_is_gapped_aln[] = "is_gapped_aln";
static const char __pyx_k_nxt_event_len[] = "nxt_event_len";
static const char __pyx_k_qualitythresh[] = "qualitythresh";
static const char __pyx_k_qualitywindow[] = "qualitywindow";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_reference_end[] = "reference_end";
static const char __pyx_k_right_aligner[] = "right_aligner";
static const char __pyx_k_unspl_loc_ref[] = "unspl_loc_ref";
static const char __pyx_k_update_pileup[] = "update_pileup";
static const char __pyx_k_updated_reads[] = "updated_reads";
static const char __pyx_k_get_contig_seq[] = "get_contig_seq";
static const char __pyx_k_is_locally_ref[] = "is_locally_ref";
static const char __pyx_k_no_realignment[] = "no_realignment";
static const char __pyx_k_quality_window[] = "quality_window";
static const char __pyx_k_check_overhangs[] = "check_overhangs";
static const char __pyx_k_compare_contigs[] = "compare_contigs";
static const char __pyx_k_estimated_count[] = "estimated_count";
static const char __pyx_k_is_quality_read[] = "is_quality_read";
static const char __pyx_k_local_threshold[] = "local_threshold";
static const char __pyx_k_reference_start[] = "reference_start";
static const char __pyx_k_retarget_cutoff[] = "retarget_cutoff";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_skip_validation[] = "skip_validation";
static const char __pyx_k_VariantAlignment[] = "VariantAlignment";
static const char __pyx_k_covering_subread[] = "covering_subread";
static const char __pyx_k_gap_open_penalty[] = "gap_open_penalty";
static const char __pyx_k_is_perfect_match[] = "is_perfect_match";
static const char __pyx_k_is_reference_seq[] = "is_reference_seq";
static const char __pyx_k_mismatch_penalty[] = "mismatch_penalty";
static const char __pyx_k_relative_aln_pos[] = "relative_aln_pos";
static const char __pyx_k_update_read_info[] = "update_read_info";
static const char __pyx_k_quality_threshold[] = "quality_threshold";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_exclude_duplicates[] = "exclude_duplicates";
static const char __pyx_k_right_aligned_vars[] = "right_aligned_vars";
static const char __pyx_k_get_local_reference[] = "get_local_reference";
static const char __pyx_k_downsample_threshold[] = "downsample_threshold";
static const char __pyx_k_generate_equivalents[] = "generate_equivalents";
static const char __pyx_k_get_reference_length[] = "get_reference_length";
static const char __pyx_k_is_non_complex_indel[] = "is_non_complex_indel";
static const char __pyx_k_most_common_gap_ptrn[] = "most_common_gap_ptrn";
static const char __pyx_k_target_right_shifted[] = "target_right_shifted";
static const char __pyx_k_aldy_indelpost_varaln[] = "aldy.indelpost.varaln";
static const char __pyx_k_cinit___locals_lambda[] = "__cinit__.<locals>.<lambda>";
static const char __pyx_k_find_by_normalization[] = "find_by_normalization";
static const char __pyx_k_gap_extension_penalty[] = "gap_extension_penalty";
static const char __pyx_k_phase_nearby_variants[] = "phase_nearby_variants";
static const char __pyx_k_base_quality_threshold[] = "base_quality_threshold";
static const char __pyx_k_find_by_softclip_split[] = "find_by_softclip_split";
static const char __pyx_k_indel_repeat_threshold[] = "indel_repeat_threshold";
static const char __pyx_k_retarget_search_window[] = "retarget_search_window";
static const char __pyx_k_update_spliced_read_info[] = "update_spliced_read_info";
static const char __pyx_k_aldy_indelpost_varaln_pyx[] = "aldy/indelpost/varaln.pyx";
static const char __pyx_k_decompose_complex_variant[] = "decompose_complex_variant";
static const char __pyx_k_exact_match_for_shiftable[] = "exact_match_for_shiftable";
static const char __pyx_k_filter_spurious_overhangs[] = "filter_spurious_overhangs";
static const char __pyx_k_mapping_quality_threshold[] = "mapping_quality_threshold";
static const char __pyx_k_fetch_reads_locals_genexpr[] = "fetch_reads.<locals>.genexpr";
static const char __pyx_k_mutation_density_threshold[] = "mutation_density_threshold";
static const char __pyx_k_retarget_similarity_cutoff[] = "retarget_similarity_cutoff";
static const char __pyx_k_count_alleles_locals_genexpr[] = "count_alleles.<locals>.genexpr";
static const char __pyx_k_find_by_smith_waterman_realn[] = "find_by_smith_waterman_realn";
static const char __pyx_k_right_aligner_locals_genexpr[] = "right_aligner.<locals>.genexpr";
static const char __pyx_k_auto_adjust_extension_penalty[] = "auto_adjust_extension_penalty";
static const char __pyx_k_VariantAlignment___parse_pileup[] = "VariantAlignment.__parse_pileup.<locals>.<lambda>";
static const char __pyx_k_phasing_stragety_must_be_either[] = "phasing stragety must be either of local, greedy, complex";
static const char __pyx_k_fetch_stragety_must_be_either_of[] = "fetch stragety must be either of target, non_target, covering";
static const char __pyx_k_longest_common_substring_thresho[] = "longest_common_substring_threshold";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static PyObject *__pyx_n_u_A;
static PyObject *__pyx_n_u_D;
static PyObject *__pyx_n_u_H;
static PyObject *__pyx_n_u_I;
static PyObject *__pyx_n_u_M;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_u_N;
static PyObject *__pyx_n_u_P;
static PyObject *__pyx_n_u_S;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_VariantAlignment;
static PyObject *__pyx_n_s_VariantAlignment___parse_pileup;
static PyObject *__pyx_kp_u__6;
static PyObject *__pyx_n_s_aldy_indelpost_varaln;
static PyObject *__pyx_kp_s_aldy_indelpost_varaln_pyx;
static PyObject *__pyx_n_s_align;
static PyObject *__pyx_n_s_aligner;
static PyObject *__pyx_n_s_alleles;
static PyObject *__pyx_n_s_aln;
static PyObject *__pyx_n_u_aln_end;
static PyObject *__pyx_n_u_aln_start;
static PyObject *__pyx_n_s_alt;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_auto_adjust_extension_penalty;
static PyObject *__pyx_n_s_bam;
static PyObject *__pyx_n_s_base_quality_threshold;
static PyObject *__pyx_n_s_best_params;
static PyObject *__pyx_n_s_best_res;
static PyObject *__pyx_n_s_by_fragment;
static PyObject *__pyx_n_s_bypass_search;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_candidate;
static PyObject *__pyx_n_s_centrality;
static PyObject *__pyx_n_s_check_overhangs;
static PyObject *__pyx_n_s_chrom;
static PyObject *__pyx_n_u_cigar_list;
static PyObject *__pyx_n_s_cigar_lst;
static PyObject *__pyx_n_u_cigar_string;
static PyObject *__pyx_n_u_cigar_updated;
static PyObject *__pyx_n_s_cinit___locals_lambda;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_compare_contigs;
static PyObject *__pyx_n_u_complex;
static PyObject *__pyx_n_s_contig;
static PyObject *__pyx_n_s_contig_2;
static PyObject *__pyx_n_s_contig_seq;
static PyObject *__pyx_n_s_count_alleles;
static PyObject *__pyx_n_s_count_alleles_locals_genexpr;
static PyObject *__pyx_n_s_count_repeats;
static PyObject *__pyx_n_u_covering;
static PyObject *__pyx_n_u_covering_subread;
static PyObject *__pyx_n_s_decompose_complex_variant;
static PyObject *__pyx_n_s_del_seq;
static PyObject *__pyx_n_s_diff;
static PyObject *__pyx_n_s_downsample_threshold;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_estimated_count;
static PyObject *__pyx_n_s_event;
static PyObject *__pyx_n_s_event_len;
static PyObject *__pyx_n_s_exact_match_for_shiftable;
static PyObject *__pyx_n_s_exclude_duplicates;
static PyObject *__pyx_n_s_failed;
static PyObject *__pyx_n_s_fetch;
static PyObject *__pyx_n_s_fetch_reads_locals_genexpr;
static PyObject *__pyx_kp_u_fetch_stragety_must_be_either_of;
static PyObject *__pyx_n_s_filter_spurious_overhangs;
static PyObject *__pyx_n_s_find_by_normalization;
static PyObject *__pyx_n_s_find_by_smith_waterman_realn;
static PyObject *__pyx_n_s_find_by_softclip_split;
static PyObject *__pyx_n_s_functools;
static PyObject *__pyx_n_s_fwrv;
static PyObject *__pyx_n_s_gap_extension;
static PyObject *__pyx_n_s_gap_extension_penalty;
static PyObject *__pyx_n_s_gap_open;
static PyObject *__pyx_n_s_gap_open_penalty;
static PyObject *__pyx_n_s_gappedaln;
static PyObject *__pyx_n_s_generate_equivalents;
static PyObject *__pyx_n_s_generate_grid;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_get_contig_seq;
static PyObject *__pyx_n_s_get_gap_ptrn2;
static PyObject *__pyx_n_s_get_local_reference;
static PyObject *__pyx_n_s_get_reference_length;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_u_greedy;
static PyObject *__pyx_n_s_grid;
static PyObject *__pyx_n_s_grid_search;
static PyObject *__pyx_n_s_h;
static PyObject *__pyx_n_s_how;
static PyObject *__pyx_n_s_hs;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_idx;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_indel_len;
static PyObject *__pyx_n_s_indel_repeat_threshold;
static PyObject *__pyx_n_s_indel_seq;
static PyObject *__pyx_n_s_index;
static PyObject *__pyx_n_s_inf;
static PyObject *__pyx_n_s_ins_seq;
static PyObject *__pyx_n_u_is_covering;
static PyObject *__pyx_n_s_is_del;
static PyObject *__pyx_n_u_is_dirty;
static PyObject *__pyx_n_s_is_gapped_aln;
static PyObject *__pyx_n_s_is_indel;
static PyObject *__pyx_n_s_is_ins;
static PyObject *__pyx_n_s_is_locally_ref;
static PyObject *__pyx_n_u_is_locally_ref;
static PyObject *__pyx_n_s_is_non_complex_indel;
static PyObject *__pyx_n_s_is_perfect_match;
static PyObject *__pyx_n_s_is_quality_read;
static PyObject *__pyx_n_u_is_reference_seq;
static PyObject *__pyx_n_u_is_reverse;
static PyObject *__pyx_n_u_is_target;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_u_local;
static PyObject *__pyx_n_s_local_threshold;
static PyObject *__pyx_n_s_localn;
static PyObject *__pyx_n_s_longest_common_substring_thresho;
static PyObject *__pyx_n_u_lt_cigar;
static PyObject *__pyx_n_u_lt_flank;
static PyObject *__pyx_n_s_lt_len;
static PyObject *__pyx_n_s_lt_median;
static PyObject *__pyx_n_s_lt_qual;
static PyObject *__pyx_n_u_lt_qual;
static PyObject *__pyx_n_s_lt_ref;
static PyObject *__pyx_n_u_lt_ref;
static PyObject *__pyx_n_s_lt_ref_len;
static PyObject *__pyx_n_s_lt_seq;
static PyObject *__pyx_n_u_lt_seq;
static PyObject *__pyx_n_s_lt_seq_len;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_make_aligner;
static PyObject *__pyx_n_s_mapping_quality_threshold;
static PyObject *__pyx_n_s_mapq_thresh;
static PyObject *__pyx_n_s_match_score;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_median;
static PyObject *__pyx_n_s_mismatch_penalty;
static PyObject *__pyx_n_s_most_common;
static PyObject *__pyx_n_s_most_common_gap_ptrn;
static PyObject *__pyx_n_s_mutation_density_threshold;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_new_cigar;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_no_realignment;
static PyObject *__pyx_n_u_non_target;
static PyObject *__pyx_n_s_normalize;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_n_s_nxt_c;
static PyObject *__pyx_n_s_nxt_event;
static PyObject *__pyx_n_s_nxt_event_len;
static PyObject *__pyx_n_s_orig_target;
static PyObject *__pyx_n_s_padding_base;
static PyObject *__pyx_n_s_partial;
static PyObject *__pyx_n_s_phase;
static PyObject *__pyx_n_s_phase_nearby_variants;
static PyObject *__pyx_kp_u_phasing_stragety_must_be_either;
static PyObject *__pyx_n_s_pileup;
static PyObject *__pyx_n_s_pos;
static PyObject *__pyx_n_s_prev_event;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_quality_threshold;
static PyObject *__pyx_n_s_quality_window;
static PyObject *__pyx_n_s_qualitythresh;
static PyObject *__pyx_n_s_qualitywindow;
static PyObject *__pyx_n_s_query_move;
static PyObject *__pyx_n_s_query_pos;
static PyObject *__pyx_n_s_random;
static PyObject *__pyx_n_s_read;
static PyObject *__pyx_n_u_read;
static PyObject *__pyx_n_s_read_2;
static PyObject *__pyx_n_s_read_end;
static PyObject *__pyx_n_u_read_name;
static PyObject *__pyx_n_u_read_qual;
static PyObject *__pyx_n_s_read_seq;
static PyObject *__pyx_n_u_read_seq;
static PyObject *__pyx_n_s_read_start;
static PyObject *__pyx_n_u_read_start;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_ref;
static PyObject *__pyx_n_s_ref_move;
static PyObject *__pyx_n_s_ref_pos;
static PyObject *__pyx_n_s_ref_seq;
static PyObject *__pyx_n_u_ref_seq;
static PyObject *__pyx_n_s_ref_start;
static PyObject *__pyx_n_s_reference;
static PyObject *__pyx_n_s_reference_end;
static PyObject *__pyx_n_s_reference_start;
static PyObject *__pyx_n_s_relative_aln_pos;
static PyObject *__pyx_n_s_relative_pos;
static PyObject *__pyx_n_s_res;
static PyObject *__pyx_n_s_responses;
static PyObject *__pyx_n_s_retarget;
static PyObject *__pyx_n_s_retarget_cutoff;
static PyObject *__pyx_n_s_retarget_search_window;
static PyObject *__pyx_n_s_retarget_similarity_cutoff;
static PyObject *__pyx_n_s_right_aligned_vars;
static PyObject *__pyx_n_s_right_aligner;
static PyObject *__pyx_n_s_right_aligner_locals_genexpr;
static PyObject *__pyx_n_s_rt_aln_pos;
static PyObject *__pyx_n_u_rt_cigar;
static PyObject *__pyx_n_u_rt_flank;
static PyObject *__pyx_n_s_rt_len;
static PyObject *__pyx_n_s_rt_median;
static PyObject *__pyx_n_s_rt_qual;
static PyObject *__pyx_n_u_rt_qual;
static PyObject *__pyx_n_s_rt_ref;
static PyObject *__pyx_n_u_rt_ref;
static PyObject *__pyx_n_s_rt_ref_len;
static PyObject *__pyx_n_s_rt_seq;
static PyObject *__pyx_n_u_rt_seq;
static PyObject *__pyx_n_s_rt_seq_len;
static PyObject *__pyx_n_s_sample;
static PyObject *__pyx_n_s_score;
static PyObject *__pyx_n_s_scores;
static PyObject *__pyx_n_s_seed;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_setRead;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_skip_next;
static PyObject *__pyx_n_s_skip_validation;
static PyObject *__pyx_n_s_softclip;
static PyObject *__pyx_n_s_sort;
static PyObject *__pyx_n_s_sorted;
static PyObject *__pyx_n_s_split_cigar;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_target;
static PyObject *__pyx_n_u_target;
static PyObject *__pyx_n_s_target_pos;
static PyObject *__pyx_n_u_target_right_shifted;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_three_class;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_u_undetermined;
static PyObject *__pyx_n_s_unspl_loc_ref;
static PyObject *__pyx_n_s_unspliced;
static PyObject *__pyx_n_s_update_pileup;
static PyObject *__pyx_n_s_update_read_info;
static PyObject *__pyx_n_s_update_spliced_read_info;
static PyObject *__pyx_n_s_updated_read;
static PyObject *__pyx_n_s_updated_reads;
static PyObject *__pyx_n_s_utilities;
static PyObject *__pyx_n_s_window;
static PyObject *__pyx_n_s_within;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static int __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment___cinit__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self, struct __pyx_obj_4aldy_9indelpost_7variant_Variant *__pyx_v_target, struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *__pyx_v_bam, int __pyx_v_window, int __pyx_v_exclude_duplicates, int __pyx_v_retarget_search_window, float __pyx_v_retarget_similarity_cutoff, int __pyx_v_exact_match_for_shiftable, int __pyx_v_mapping_quality_threshold, int __pyx_v_downsample_threshold, int __pyx_v_base_quality_threshold, int __pyx_v_match_score, int __pyx_v_mismatch_penalty, int __pyx_v_gap_open_penalty, int __pyx_v_gap_extension_penalty, int __pyx_v_auto_adjust_extension_penalty, int __pyx_v_no_realignment); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda1(PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_2__eq__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static Py_hash_t __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_4__hash__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_6get_contig(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_8get_target_indel(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_11fetch_reads_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_10fetch_reads(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self, PyObject *__pyx_v_how); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_13count_alleles_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_12count_alleles(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self, PyObject *__pyx_v_fwrv, PyObject *__pyx_v_by_fragment, PyObject *__pyx_v_three_class, PyObject *__pyx_v_estimated_count, PyObject *__pyx_v_quality_window, PyObject *__pyx_v_quality_threshold); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_14phase(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self, PyObject *__pyx_v_how, PyObject *__pyx_v_local_threshold, PyObject *__pyx_v_longest_common_substring_threshold, PyObject *__pyx_v_indel_repeat_threshold, PyObject *__pyx_v_mutation_density_threshold); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_3bam___get__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_20is_spurious_overhang___get__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_16is_complex_input___get__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_6contig___get__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_25unspliced_local_reference___get__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_is_quality_read(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_pos, PyObject *__pyx_v_qualitywindow, PyObject *__pyx_v_qualitythresh); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_2is_locally_ref(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_pos); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_4centrality(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_target_pos); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_6update_spliced_read_info(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_target, PyObject *__pyx_v_orig_target, PyObject *__pyx_v_is_gapped_aln, PyObject *__pyx_v_window, PyObject *__pyx_v_match_score, PyObject *__pyx_v_mismatch_penalty, PyObject *__pyx_v_gap_open_penalty, PyObject *__pyx_v_gap_extension_penalty, PyObject *__pyx_v_unspl_loc_ref); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_13right_aligner_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_8right_aligner(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_target); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_10generate_grid(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_auto_adjust_extension_penalty, PyObject *__pyx_v_gap_open_penalty, PyObject *__pyx_v_gap_extension_penalty, PyObject *__pyx_v_target); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_12grid_search(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_target, PyObject *__pyx_v_pileup, PyObject *__pyx_v_window, PyObject *__pyx_v_mapq_thresh, PyObject *__pyx_v_within, PyObject *__pyx_v_retarget_cutoff, PyObject *__pyx_v_match_score, PyObject *__pyx_v_mismatch_penalty, PyObject *__pyx_v_grid, PyObject *__pyx_v_unspl_loc_ref, PyObject *__pyx_v_exact_match_for_shiftable); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_14is_perfect_match(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_aligner, PyObject *__pyx_v_contig_seq, PyObject *__pyx_v_read_seq); /* proto */
static int __pyx_pf_7cpython_5array_5array___getbuffer__(arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_7cpython_5array_5array_2__releasebuffer__(CYTHON_UNUSED arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln_VariantAlignment(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_get = {0, &__pyx_n_s_get, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_index = {0, &__pyx_n_s_index, 0, 0, 0};
static PyObject *__pyx_float_0_5;
static PyObject *__pyx_float_1_0;
static PyObject *__pyx_float_0_05;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_15;
static PyObject *__pyx_int_20;
static PyObject *__pyx_int_123;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_slice__5;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_codeobj__15;
static PyObject *__pyx_codeobj__17;
static PyObject *__pyx_codeobj__19;
static PyObject *__pyx_codeobj__21;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__29;
/* Late includes */

/* "aldy/indelpost/varaln.pyx":102
 *         True to only analyzed gap-aligned indels (default False)
 *     """
 *     def __cinit__(             # <<<<<<<<<<<<<<
 *         self,
 *         Variant target,
 */

/* Python wrapper */
static int __pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_4aldy_9indelpost_7variant_Variant *__pyx_v_target = 0;
  struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *__pyx_v_bam = 0;
  int __pyx_v_window;
  int __pyx_v_exclude_duplicates;
  int __pyx_v_retarget_search_window;
  float __pyx_v_retarget_similarity_cutoff;
  int __pyx_v_exact_match_for_shiftable;
  int __pyx_v_mapping_quality_threshold;
  int __pyx_v_downsample_threshold;
  int __pyx_v_base_quality_threshold;
  int __pyx_v_match_score;
  int __pyx_v_mismatch_penalty;
  int __pyx_v_gap_open_penalty;
  int __pyx_v_gap_extension_penalty;
  int __pyx_v_auto_adjust_extension_penalty;
  int __pyx_v_no_realignment;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_target,&__pyx_n_s_bam,&__pyx_n_s_window,&__pyx_n_s_exclude_duplicates,&__pyx_n_s_retarget_search_window,&__pyx_n_s_retarget_similarity_cutoff,&__pyx_n_s_exact_match_for_shiftable,&__pyx_n_s_mapping_quality_threshold,&__pyx_n_s_downsample_threshold,&__pyx_n_s_base_quality_threshold,&__pyx_n_s_match_score,&__pyx_n_s_mismatch_penalty,&__pyx_n_s_gap_open_penalty,&__pyx_n_s_gap_extension_penalty,&__pyx_n_s_auto_adjust_extension_penalty,&__pyx_n_s_no_realignment,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_target)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bam)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 16, 1); __PYX_ERR(0, 102, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_window);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exclude_duplicates);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_retarget_search_window);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_retarget_similarity_cutoff);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exact_match_for_shiftable);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mapping_quality_threshold);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_downsample_threshold);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_base_quality_threshold);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_match_score);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mismatch_penalty);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gap_open_penalty);
          if (value) { values[12] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gap_extension_penalty);
          if (value) { values[13] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_auto_adjust_extension_penalty);
          if (value) { values[14] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_no_realignment);
          if (value) { values[15] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 102, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)values[0]);
    __pyx_v_bam = ((struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *)values[1]);
    if (values[2]) {
      __pyx_v_window = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_window == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 106, __pyx_L3_error)
    } else {
      __pyx_v_window = ((int)50);
    }
    if (values[3]) {
      __pyx_v_exclude_duplicates = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_exclude_duplicates == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 107, __pyx_L3_error)
    } else {

      /* "aldy/indelpost/varaln.pyx":107
 *         AlignmentFile bam,
 *         int window=50,
 *         bint exclude_duplicates=True,             # <<<<<<<<<<<<<<
 *         int retarget_search_window=30,
 *         float retarget_similarity_cutoff=0.7,
 */
      __pyx_v_exclude_duplicates = ((int)1);
    }
    if (values[4]) {
      __pyx_v_retarget_search_window = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_retarget_search_window == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 108, __pyx_L3_error)
    } else {
      __pyx_v_retarget_search_window = ((int)30);
    }
    if (values[5]) {
      __pyx_v_retarget_similarity_cutoff = __pyx_PyFloat_AsFloat(values[5]); if (unlikely((__pyx_v_retarget_similarity_cutoff == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 109, __pyx_L3_error)
    } else {
      __pyx_v_retarget_similarity_cutoff = ((float)0.7);
    }
    if (values[6]) {
      __pyx_v_exact_match_for_shiftable = __Pyx_PyObject_IsTrue(values[6]); if (unlikely((__pyx_v_exact_match_for_shiftable == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L3_error)
    } else {

      /* "aldy/indelpost/varaln.pyx":110
 *         int retarget_search_window=30,
 *         float retarget_similarity_cutoff=0.7,
 *         bint exact_match_for_shiftable=True,             # <<<<<<<<<<<<<<
 *         int mapping_quality_threshold=1,
 *         int downsample_threshold=1000,
 */
      __pyx_v_exact_match_for_shiftable = ((int)1);
    }
    if (values[7]) {
      __pyx_v_mapping_quality_threshold = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_mapping_quality_threshold == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 111, __pyx_L3_error)
    } else {
      __pyx_v_mapping_quality_threshold = ((int)1);
    }
    if (values[8]) {
      __pyx_v_downsample_threshold = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_downsample_threshold == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 112, __pyx_L3_error)
    } else {
      __pyx_v_downsample_threshold = ((int)0x3E8);
    }
    if (values[9]) {
      __pyx_v_base_quality_threshold = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_base_quality_threshold == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 113, __pyx_L3_error)
    } else {
      __pyx_v_base_quality_threshold = ((int)20);
    }
    if (values[10]) {
      __pyx_v_match_score = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_match_score == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 114, __pyx_L3_error)
    } else {
      __pyx_v_match_score = ((int)3);
    }
    if (values[11]) {
      __pyx_v_mismatch_penalty = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_mismatch_penalty == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 115, __pyx_L3_error)
    } else {
      __pyx_v_mismatch_penalty = ((int)2);
    }
    if (values[12]) {
      __pyx_v_gap_open_penalty = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_gap_open_penalty == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 116, __pyx_L3_error)
    } else {
      __pyx_v_gap_open_penalty = ((int)3);
    }
    if (values[13]) {
      __pyx_v_gap_extension_penalty = __Pyx_PyInt_As_int(values[13]); if (unlikely((__pyx_v_gap_extension_penalty == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 117, __pyx_L3_error)
    } else {
      __pyx_v_gap_extension_penalty = ((int)1);
    }
    if (values[14]) {
      __pyx_v_auto_adjust_extension_penalty = __Pyx_PyObject_IsTrue(values[14]); if (unlikely((__pyx_v_auto_adjust_extension_penalty == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 118, __pyx_L3_error)
    } else {

      /* "aldy/indelpost/varaln.pyx":118
 *         int gap_open_penalty=3,
 *         int gap_extension_penalty=1,
 *         bint auto_adjust_extension_penalty=True,             # <<<<<<<<<<<<<<
 *         bint no_realignment=False,
 *     ):
 */
      __pyx_v_auto_adjust_extension_penalty = ((int)1);
    }
    if (values[15]) {
      __pyx_v_no_realignment = __Pyx_PyObject_IsTrue(values[15]); if (unlikely((__pyx_v_no_realignment == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 119, __pyx_L3_error)
    } else {

      /* "aldy/indelpost/varaln.pyx":119
 *         int gap_extension_penalty=1,
 *         bint auto_adjust_extension_penalty=True,
 *         bint no_realignment=False,             # <<<<<<<<<<<<<<
 *     ):
 * 
 */
      __pyx_v_no_realignment = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 102, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_target), __pyx_ptype_4aldy_9indelpost_7variant_Variant, 1, "target", 0))) __PYX_ERR(0, 104, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bam), __pyx_ptype_5pysam_17libcalignmentfile_AlignmentFile, 1, "bam", 0))) __PYX_ERR(0, 105, __pyx_L1_error)
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment___cinit__(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self), __pyx_v_target, __pyx_v_bam, __pyx_v_window, __pyx_v_exclude_duplicates, __pyx_v_retarget_search_window, __pyx_v_retarget_similarity_cutoff, __pyx_v_exact_match_for_shiftable, __pyx_v_mapping_quality_threshold, __pyx_v_downsample_threshold, __pyx_v_base_quality_threshold, __pyx_v_match_score, __pyx_v_mismatch_penalty, __pyx_v_gap_open_penalty, __pyx_v_gap_extension_penalty, __pyx_v_auto_adjust_extension_penalty, __pyx_v_no_realignment);

  /* "aldy/indelpost/varaln.pyx":102
 *         True to only analyzed gap-aligned indels (default False)
 *     """
 *     def __cinit__(             # <<<<<<<<<<<<<<
 *         self,
 *         Variant target,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":134
 * 
 *             decomposed_indels = [i for i in decomposed_variants if i.is_indel]
 *             decomposed_indels.sort(key=lambda x : len(x.indel_seq))             # <<<<<<<<<<<<<<
 *             self.__target = decomposed_indels[-1]
 *             self.target = self.__target
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_9__cinit___lambda(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6varaln_16VariantAlignment_9__cinit___lambda = {"lambda", (PyCFunction)__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_9__cinit___lambda, METH_O, 0};
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_9__cinit___lambda(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda(__pyx_self, ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_indel_seq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.__cinit__.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":102
 *         True to only analyzed gap-aligned indels (default False)
 *     """
 *     def __cinit__(             # <<<<<<<<<<<<<<
 *         self,
 *         Variant target,
 */

static int __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment___cinit__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self, struct __pyx_obj_4aldy_9indelpost_7variant_Variant *__pyx_v_target, struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *__pyx_v_bam, int __pyx_v_window, int __pyx_v_exclude_duplicates, int __pyx_v_retarget_search_window, float __pyx_v_retarget_similarity_cutoff, int __pyx_v_exact_match_for_shiftable, int __pyx_v_mapping_quality_threshold, int __pyx_v_downsample_threshold, int __pyx_v_base_quality_threshold, int __pyx_v_match_score, int __pyx_v_mismatch_penalty, int __pyx_v_gap_open_penalty, int __pyx_v_gap_extension_penalty, int __pyx_v_auto_adjust_extension_penalty, int __pyx_v_no_realignment) {
  PyObject *__pyx_v_second_target = NULL;
  int __pyx_v_is_complex_input;
  PyObject *__pyx_v_decomposed_variants = NULL;
  PyObject *__pyx_v_decomposed_indels = NULL;
  PyObject *__pyx_7genexpr__pyx_v_i = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  PyObject *(*__pyx_t_15)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "aldy/indelpost/varaln.pyx":122
 *     ):
 * 
 *         self.target, second_target = target, target             # <<<<<<<<<<<<<<
 * 
 *         is_complex_input = False
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_target);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)__pyx_v_target);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->target);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->target));
  __pyx_v_self->target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_second_target = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":124
 *         self.target, second_target = target, target
 * 
 *         is_complex_input = False             # <<<<<<<<<<<<<<
 *         if not target.is_non_complex_indel() and target.is_indel:
 *             is_complex_input = True
 */
  __pyx_v_is_complex_input = 0;

  /* "aldy/indelpost/varaln.pyx":125
 * 
 *         is_complex_input = False
 *         if not target.is_non_complex_indel() and target.is_indel:             # <<<<<<<<<<<<<<
 *             is_complex_input = True
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_target), __pyx_n_s_is_non_complex_indel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = ((!__pyx_t_5) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_3 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_target), __pyx_n_s_is_indel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "aldy/indelpost/varaln.pyx":126
 *         is_complex_input = False
 *         if not target.is_non_complex_indel() and target.is_indel:
 *             is_complex_input = True             # <<<<<<<<<<<<<<
 * 
 *             if auto_adjust_extension_penalty:
 */
    __pyx_v_is_complex_input = 1;

    /* "aldy/indelpost/varaln.pyx":128
 *             is_complex_input = True
 * 
 *             if auto_adjust_extension_penalty:             # <<<<<<<<<<<<<<
 *                 decomposed_variants = target.decompose_complex_variant(match_score, mismatch_penalty)
 *             else:
 */
    __pyx_t_3 = (__pyx_v_auto_adjust_extension_penalty != 0);
    if (__pyx_t_3) {

      /* "aldy/indelpost/varaln.pyx":129
 * 
 *             if auto_adjust_extension_penalty:
 *                 decomposed_variants = target.decompose_complex_variant(match_score, mismatch_penalty)             # <<<<<<<<<<<<<<
 *             else:
 *                 decomposed_variants = target.decompose_complex_variant(match_score, mismatch_penalty, gap_open_penalty, gap_extension_penalty)
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_target), __pyx_n_s_decompose_complex_variant); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_match_score); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_mismatch_penalty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_4, __pyx_t_7};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_4, __pyx_t_7};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 129, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_8) {
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
        __pyx_t_4 = 0;
        __pyx_t_7 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_decomposed_variants = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "aldy/indelpost/varaln.pyx":128
 *             is_complex_input = True
 * 
 *             if auto_adjust_extension_penalty:             # <<<<<<<<<<<<<<
 *                 decomposed_variants = target.decompose_complex_variant(match_score, mismatch_penalty)
 *             else:
 */
      goto __pyx_L6;
    }

    /* "aldy/indelpost/varaln.pyx":131
 *                 decomposed_variants = target.decompose_complex_variant(match_score, mismatch_penalty)
 *             else:
 *                 decomposed_variants = target.decompose_complex_variant(match_score, mismatch_penalty, gap_open_penalty, gap_extension_penalty)             # <<<<<<<<<<<<<<
 * 
 *             decomposed_indels = [i for i in decomposed_variants if i.is_indel]
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_target), __pyx_n_s_decompose_complex_variant); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_match_score); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_mismatch_penalty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_gap_open_penalty); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_gap_extension_penalty); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_11 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[5] = {__pyx_t_11, __pyx_t_10, __pyx_t_7, __pyx_t_4, __pyx_t_8};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[5] = {__pyx_t_11, __pyx_t_10, __pyx_t_7, __pyx_t_4, __pyx_t_8};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else
      #endif
      {
        __pyx_t_12 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 131, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (__pyx_t_11) {
          __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_11); __pyx_t_11 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_10);
        PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_9, __pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_9, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_9, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_12, 3+__pyx_t_9, __pyx_t_8);
        __pyx_t_10 = 0;
        __pyx_t_7 = 0;
        __pyx_t_4 = 0;
        __pyx_t_8 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_decomposed_variants = __pyx_t_2;
      __pyx_t_2 = 0;
    }
    __pyx_L6:;

    /* "aldy/indelpost/varaln.pyx":133
 *                 decomposed_variants = target.decompose_complex_variant(match_score, mismatch_penalty, gap_open_penalty, gap_extension_penalty)
 * 
 *             decomposed_indels = [i for i in decomposed_variants if i.is_indel]             # <<<<<<<<<<<<<<
 *             decomposed_indels.sort(key=lambda x : len(x.indel_seq))
 *             self.__target = decomposed_indels[-1]
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (likely(PyList_CheckExact(__pyx_v_decomposed_variants)) || PyTuple_CheckExact(__pyx_v_decomposed_variants)) {
        __pyx_t_1 = __pyx_v_decomposed_variants; __Pyx_INCREF(__pyx_t_1); __pyx_t_13 = 0;
        __pyx_t_14 = NULL;
      } else {
        __pyx_t_13 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_decomposed_variants); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_14 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 133, __pyx_L9_error)
      }
      for (;;) {
        if (likely(!__pyx_t_14)) {
          if (likely(PyList_CheckExact(__pyx_t_1))) {
            if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_12 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_13); __Pyx_INCREF(__pyx_t_12); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 133, __pyx_L9_error)
            #else
            __pyx_t_12 = PySequence_ITEM(__pyx_t_1, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 133, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_12);
            #endif
          } else {
            if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_12 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_13); __Pyx_INCREF(__pyx_t_12); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 133, __pyx_L9_error)
            #else
            __pyx_t_12 = PySequence_ITEM(__pyx_t_1, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 133, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_12);
            #endif
          }
        } else {
          __pyx_t_12 = __pyx_t_14(__pyx_t_1);
          if (unlikely(!__pyx_t_12)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 133, __pyx_L9_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_12);
        }
        __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_i, __pyx_t_12);
        __pyx_t_12 = 0;
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_7genexpr__pyx_v_i, __pyx_n_s_is_indel); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 133, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 133, __pyx_L9_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (__pyx_t_3) {
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_7genexpr__pyx_v_i))) __PYX_ERR(0, 133, __pyx_L9_error)
        }
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_i); __pyx_7genexpr__pyx_v_i = 0;
      goto __pyx_L13_exit_scope;
      __pyx_L9_error:;
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_i); __pyx_7genexpr__pyx_v_i = 0;
      goto __pyx_L1_error;
      __pyx_L13_exit_scope:;
    } /* exit inner scope */
    __pyx_v_decomposed_indels = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":134
 * 
 *             decomposed_indels = [i for i in decomposed_variants if i.is_indel]
 *             decomposed_indels.sort(key=lambda x : len(x.indel_seq))             # <<<<<<<<<<<<<<
 *             self.__target = decomposed_indels[-1]
 *             self.target = self.__target
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_decomposed_indels, __pyx_n_s_sort); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_12 = __Pyx_CyFunction_New(&__pyx_mdef_4aldy_9indelpost_6varaln_16VariantAlignment_9__cinit___lambda, 0, __pyx_n_s_cinit___locals_lambda, NULL, __pyx_n_s_aldy_indelpost_varaln, __pyx_d, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_key, __pyx_t_12) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "aldy/indelpost/varaln.pyx":135
 *             decomposed_indels = [i for i in decomposed_variants if i.is_indel]
 *             decomposed_indels.sort(key=lambda x : len(x.indel_seq))
 *             self.__target = decomposed_indels[-1]             # <<<<<<<<<<<<<<
 *             self.target = self.__target
 *             if len(decomposed_indels) > 1:
 */
    __pyx_t_12 = __Pyx_GetItemInt_List(__pyx_v_decomposed_indels, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (!(likely(((__pyx_t_12) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_12, __pyx_ptype_4aldy_9indelpost_7variant_Variant))))) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_12);
    __Pyx_GOTREF(__pyx_v_self->__pyx___target);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx___target));
    __pyx_v_self->__pyx___target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_12);
    __pyx_t_12 = 0;

    /* "aldy/indelpost/varaln.pyx":136
 *             decomposed_indels.sort(key=lambda x : len(x.indel_seq))
 *             self.__target = decomposed_indels[-1]
 *             self.target = self.__target             # <<<<<<<<<<<<<<
 *             if len(decomposed_indels) > 1:
 *                 second_target = decomposed_indels[-2]
 */
    __pyx_t_12 = ((PyObject *)__pyx_v_self->__pyx___target);
    __Pyx_INCREF(__pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_12);
    __Pyx_GOTREF(__pyx_v_self->target);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->target));
    __pyx_v_self->target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_12);
    __pyx_t_12 = 0;

    /* "aldy/indelpost/varaln.pyx":137
 *             self.__target = decomposed_indels[-1]
 *             self.target = self.__target
 *             if len(decomposed_indels) > 1:             # <<<<<<<<<<<<<<
 *                 second_target = decomposed_indels[-2]
 * 
 */
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_decomposed_indels); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 137, __pyx_L1_error)
    __pyx_t_3 = ((__pyx_t_13 > 1) != 0);
    if (__pyx_t_3) {

      /* "aldy/indelpost/varaln.pyx":138
 *             self.target = self.__target
 *             if len(decomposed_indels) > 1:
 *                 second_target = decomposed_indels[-2]             # <<<<<<<<<<<<<<
 * 
 *             #self.__target = max(decomposed_indels, key=lambda l : len(l.indel_seq))
 */
      __pyx_t_12 = __Pyx_GetItemInt_List(__pyx_v_decomposed_indels, -2L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF_SET(__pyx_v_second_target, __pyx_t_12);
      __pyx_t_12 = 0;

      /* "aldy/indelpost/varaln.pyx":137
 *             self.__target = decomposed_indels[-1]
 *             self.target = self.__target
 *             if len(decomposed_indels) > 1:             # <<<<<<<<<<<<<<
 *                 second_target = decomposed_indels[-2]
 * 
 */
    }

    /* "aldy/indelpost/varaln.pyx":125
 * 
 *         is_complex_input = False
 *         if not target.is_non_complex_indel() and target.is_indel:             # <<<<<<<<<<<<<<
 *             is_complex_input = True
 * 
 */
    goto __pyx_L3;
  }

  /* "aldy/indelpost/varaln.pyx":143
 *             #self.target = self.__target
 *         else:
 *             self.__target = target.normalize()             # <<<<<<<<<<<<<<
 * 
 *         self.bam = bam
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_target), __pyx_n_s_normalize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_12 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(((__pyx_t_12) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_12, __pyx_ptype_4aldy_9indelpost_7variant_Variant))))) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_12);
    __Pyx_GOTREF(__pyx_v_self->__pyx___target);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx___target));
    __pyx_v_self->__pyx___target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_12);
    __pyx_t_12 = 0;
  }
  __pyx_L3:;

  /* "aldy/indelpost/varaln.pyx":145
 *             self.__target = target.normalize()
 * 
 *         self.bam = bam             # <<<<<<<<<<<<<<
 *         self.window = window
 *         self.exclude_duplicates = exclude_duplicates
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_bam));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_bam));
  __Pyx_GOTREF(__pyx_v_self->bam);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->bam));
  __pyx_v_self->bam = __pyx_v_bam;

  /* "aldy/indelpost/varaln.pyx":146
 * 
 *         self.bam = bam
 *         self.window = window             # <<<<<<<<<<<<<<
 *         self.exclude_duplicates = exclude_duplicates
 *         self.retarget_window = retarget_search_window
 */
  __pyx_v_self->window = __pyx_v_window;

  /* "aldy/indelpost/varaln.pyx":147
 *         self.bam = bam
 *         self.window = window
 *         self.exclude_duplicates = exclude_duplicates             # <<<<<<<<<<<<<<
 *         self.retarget_window = retarget_search_window
 *         self.retarget_cutoff = retarget_similarity_cutoff
 */
  __pyx_v_self->exclude_duplicates = __pyx_v_exclude_duplicates;

  /* "aldy/indelpost/varaln.pyx":148
 *         self.window = window
 *         self.exclude_duplicates = exclude_duplicates
 *         self.retarget_window = retarget_search_window             # <<<<<<<<<<<<<<
 *         self.retarget_cutoff = retarget_similarity_cutoff
 *         self.exact_match_for_shiftable = exact_match_for_shiftable
 */
  __pyx_v_self->retarget_window = __pyx_v_retarget_search_window;

  /* "aldy/indelpost/varaln.pyx":149
 *         self.exclude_duplicates = exclude_duplicates
 *         self.retarget_window = retarget_search_window
 *         self.retarget_cutoff = retarget_similarity_cutoff             # <<<<<<<<<<<<<<
 *         self.exact_match_for_shiftable = exact_match_for_shiftable
 *         self.mapqthresh = mapping_quality_threshold
 */
  __pyx_v_self->retarget_cutoff = __pyx_v_retarget_similarity_cutoff;

  /* "aldy/indelpost/varaln.pyx":150
 *         self.retarget_window = retarget_search_window
 *         self.retarget_cutoff = retarget_similarity_cutoff
 *         self.exact_match_for_shiftable = exact_match_for_shiftable             # <<<<<<<<<<<<<<
 *         self.mapqthresh = mapping_quality_threshold
 *         self.downsamplethresh = downsample_threshold
 */
  __pyx_v_self->exact_match_for_shiftable = __pyx_v_exact_match_for_shiftable;

  /* "aldy/indelpost/varaln.pyx":151
 *         self.retarget_cutoff = retarget_similarity_cutoff
 *         self.exact_match_for_shiftable = exact_match_for_shiftable
 *         self.mapqthresh = mapping_quality_threshold             # <<<<<<<<<<<<<<
 *         self.downsamplethresh = downsample_threshold
 *         self.basequalthresh = base_quality_threshold
 */
  __pyx_v_self->mapqthresh = __pyx_v_mapping_quality_threshold;

  /* "aldy/indelpost/varaln.pyx":152
 *         self.exact_match_for_shiftable = exact_match_for_shiftable
 *         self.mapqthresh = mapping_quality_threshold
 *         self.downsamplethresh = downsample_threshold             # <<<<<<<<<<<<<<
 *         self.basequalthresh = base_quality_threshold
 *         self.match_score = match_score
 */
  __pyx_v_self->downsamplethresh = __pyx_v_downsample_threshold;

  /* "aldy/indelpost/varaln.pyx":153
 *         self.mapqthresh = mapping_quality_threshold
 *         self.downsamplethresh = downsample_threshold
 *         self.basequalthresh = base_quality_threshold             # <<<<<<<<<<<<<<
 *         self.match_score = match_score
 *         self.mismatch_penalty = mismatch_penalty
 */
  __pyx_v_self->basequalthresh = __pyx_v_base_quality_threshold;

  /* "aldy/indelpost/varaln.pyx":154
 *         self.downsamplethresh = downsample_threshold
 *         self.basequalthresh = base_quality_threshold
 *         self.match_score = match_score             # <<<<<<<<<<<<<<
 *         self.mismatch_penalty = mismatch_penalty
 *         self.gap_open_penalty = gap_open_penalty
 */
  __pyx_v_self->match_score = __pyx_v_match_score;

  /* "aldy/indelpost/varaln.pyx":155
 *         self.basequalthresh = base_quality_threshold
 *         self.match_score = match_score
 *         self.mismatch_penalty = mismatch_penalty             # <<<<<<<<<<<<<<
 *         self.gap_open_penalty = gap_open_penalty
 *         self.gap_extension_penalty = gap_extension_penalty
 */
  __pyx_v_self->mismatch_penalty = __pyx_v_mismatch_penalty;

  /* "aldy/indelpost/varaln.pyx":156
 *         self.match_score = match_score
 *         self.mismatch_penalty = mismatch_penalty
 *         self.gap_open_penalty = gap_open_penalty             # <<<<<<<<<<<<<<
 *         self.gap_extension_penalty = gap_extension_penalty
 *         self.auto_adjust_extension_penalty = auto_adjust_extension_penalty
 */
  __pyx_v_self->gap_open_penalty = __pyx_v_gap_open_penalty;

  /* "aldy/indelpost/varaln.pyx":157
 *         self.mismatch_penalty = mismatch_penalty
 *         self.gap_open_penalty = gap_open_penalty
 *         self.gap_extension_penalty = gap_extension_penalty             # <<<<<<<<<<<<<<
 *         self.auto_adjust_extension_penalty = auto_adjust_extension_penalty
 *         self.no_realignment = no_realignment
 */
  __pyx_v_self->gap_extension_penalty = __pyx_v_gap_extension_penalty;

  /* "aldy/indelpost/varaln.pyx":158
 *         self.gap_open_penalty = gap_open_penalty
 *         self.gap_extension_penalty = gap_extension_penalty
 *         self.auto_adjust_extension_penalty = auto_adjust_extension_penalty             # <<<<<<<<<<<<<<
 *         self.no_realignment = no_realignment
 *         self.is_complex_input = is_complex_input
 */
  __pyx_v_self->auto_adjust_extension_penalty = __pyx_v_auto_adjust_extension_penalty;

  /* "aldy/indelpost/varaln.pyx":159
 *         self.gap_extension_penalty = gap_extension_penalty
 *         self.auto_adjust_extension_penalty = auto_adjust_extension_penalty
 *         self.no_realignment = no_realignment             # <<<<<<<<<<<<<<
 *         self.is_complex_input = is_complex_input
 *         self.second_target = second_target
 */
  __pyx_v_self->no_realignment = __pyx_v_no_realignment;

  /* "aldy/indelpost/varaln.pyx":160
 *         self.auto_adjust_extension_penalty = auto_adjust_extension_penalty
 *         self.no_realignment = no_realignment
 *         self.is_complex_input = is_complex_input             # <<<<<<<<<<<<<<
 *         self.second_target = second_target
 *         self.unspliced_local_reference = UnsplicedLocalReference(
 */
  __pyx_t_12 = __Pyx_PyBool_FromLong(__pyx_v_is_complex_input); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_12);
  __Pyx_GOTREF(__pyx_v_self->is_complex_input);
  __Pyx_DECREF(__pyx_v_self->is_complex_input);
  __pyx_v_self->is_complex_input = __pyx_t_12;
  __pyx_t_12 = 0;

  /* "aldy/indelpost/varaln.pyx":161
 *         self.no_realignment = no_realignment
 *         self.is_complex_input = is_complex_input
 *         self.second_target = second_target             # <<<<<<<<<<<<<<
 *         self.unspliced_local_reference = UnsplicedLocalReference(
 *                                             self.__target.chrom,
 */
  if (!(likely(((__pyx_v_second_target) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_second_target, __pyx_ptype_4aldy_9indelpost_7variant_Variant))))) __PYX_ERR(0, 161, __pyx_L1_error)
  __pyx_t_12 = __pyx_v_second_target;
  __Pyx_INCREF(__pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_12);
  __Pyx_GOTREF(__pyx_v_self->second_target);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->second_target));
  __pyx_v_self->second_target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_12);
  __pyx_t_12 = 0;

  /* "aldy/indelpost/varaln.pyx":164
 *         self.unspliced_local_reference = UnsplicedLocalReference(
 *                                             self.__target.chrom,
 *                                             self.__target.pos,             # <<<<<<<<<<<<<<
 *                                             self.__target.reference.get_reference_length(self.__target.chrom),
 *                                             self.window,
 */
  __pyx_t_12 = __Pyx_PyInt_From_int(__pyx_v_self->__pyx___target->pos); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);

  /* "aldy/indelpost/varaln.pyx":165
 *                                             self.__target.chrom,
 *                                             self.__target.pos,
 *                                             self.__target.reference.get_reference_length(self.__target.chrom),             # <<<<<<<<<<<<<<
 *                                             self.window,
 *                                             self.__target.reference
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx___target->reference), __pyx_n_s_get_reference_length); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_v_self->__pyx___target->chrom) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self->__pyx___target->chrom);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":166
 *                                             self.__target.pos,
 *                                             self.__target.reference.get_reference_length(self.__target.chrom),
 *                                             self.window,             # <<<<<<<<<<<<<<
 *                                             self.__target.reference
 *                                          )
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->window); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "aldy/indelpost/varaln.pyx":162
 *         self.is_complex_input = is_complex_input
 *         self.second_target = second_target
 *         self.unspliced_local_reference = UnsplicedLocalReference(             # <<<<<<<<<<<<<<
 *                                             self.__target.chrom,
 *                                             self.__target.pos,
 */
  __pyx_t_8 = PyTuple_New(5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_self->__pyx___target->chrom);
  __Pyx_GIVEREF(__pyx_v_self->__pyx___target->chrom);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_self->__pyx___target->chrom);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target->reference));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target->reference));
  PyTuple_SET_ITEM(__pyx_t_8, 4, ((PyObject *)__pyx_v_self->__pyx___target->reference));
  __pyx_t_12 = 0;
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4aldy_9indelpost_15local_reference_UnsplicedLocalReference), __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->unspliced_local_reference);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
  __pyx_v_self->unspliced_local_reference = ((struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":169
 *                                             self.__target.reference
 *                                          )
 *         self.__pileup, self.contig = self.__parse_pileup()             # <<<<<<<<<<<<<<
 * 
 *     cdef __parse_pileup(self, Contig contig=None, bint retargeted=False, bint skip_read_end_check=False):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self->__pyx_vtab)->__pyx___parse_pileup(__pyx_v_self, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 169, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_1);
    #else
    __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_12 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_15 = Py_TYPE(__pyx_t_12)->tp_iternext;
    index = 0; __pyx_t_8 = __pyx_t_15(__pyx_t_12); if (unlikely(!__pyx_t_8)) goto __pyx_L15_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_8);
    index = 1; __pyx_t_1 = __pyx_t_15(__pyx_t_12); if (unlikely(!__pyx_t_1)) goto __pyx_L15_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_12), 2) < 0) __PYX_ERR(0, 169, __pyx_L1_error)
    __pyx_t_15 = NULL;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    goto __pyx_L16_unpacking_done;
    __pyx_L15_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_15 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 169, __pyx_L1_error)
    __pyx_L16_unpacking_done:;
  }
  if (!(likely(PyList_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_8)->tp_name), 0))) __PYX_ERR(0, 169, __pyx_L1_error)
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_4aldy_9indelpost_6contig_Contig))))) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_8);
  __Pyx_GOTREF(__pyx_v_self->__pyx___pileup);
  __Pyx_DECREF(__pyx_v_self->__pyx___pileup);
  __pyx_v_self->__pyx___pileup = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->contig);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->contig));
  __pyx_v_self->contig = ((struct __pyx_obj_4aldy_9indelpost_6contig_Contig *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":102
 *         True to only analyzed gap-aligned indels (default False)
 *     """
 *     def __cinit__(             # <<<<<<<<<<<<<<
 *         self,
 *         Variant target,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_second_target);
  __Pyx_XDECREF(__pyx_v_decomposed_variants);
  __Pyx_XDECREF(__pyx_v_decomposed_indels);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_i);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":468
 *                         except:
 *                             target_pos = self.__target.pos
 *                             indels.sort(key=lambda x : abs(x.pos - target_pos))             # <<<<<<<<<<<<<<
 * 
 *                         return self.__parse_pileup(contig=None, retargeted=False, skip_read_end_check=True)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_14__parse_pileup_lambda1(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6varaln_16VariantAlignment_14__parse_pileup_lambda1 = {"lambda1", (PyCFunction)__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_14__parse_pileup_lambda1, METH_O, 0};
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_14__parse_pileup_lambda1(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda1 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda1(__pyx_self, ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda1(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup *__pyx_cur_scope;
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda1", 0);
  __pyx_outer_scope = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_target_pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Subtract(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyNumber_Absolute(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.__parse_pileup.lambda1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":171
 *         self.__pileup, self.contig = self.__parse_pileup()
 * 
 *     cdef __parse_pileup(self, Contig contig=None, bint retargeted=False, bint skip_read_end_check=False):             # <<<<<<<<<<<<<<
 *         """Dictize reads for target indel and make target indel template by consensus
 * 
 */

static PyObject *__pyx_f_4aldy_9indelpost_6varaln_16VariantAlignment___parse_pileup(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self, struct __pyx_opt_args_4aldy_9indelpost_6varaln_16VariantAlignment___parse_pileup *__pyx_optional_args) {
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup *__pyx_cur_scope;
  struct __pyx_obj_4aldy_9indelpost_6contig_Contig *__pyx_v_contig = ((struct __pyx_obj_4aldy_9indelpost_6contig_Contig *)Py_None);
  int __pyx_v_retargeted = ((int)0);
  int __pyx_v_skip_read_end_check = ((int)0);
  PyObject *__pyx_v_pileup = 0;
  PyObject *__pyx_v_read_end_evidence_only = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_sample_factor = NULL;
  PyObject *__pyx_v_exptension_penalty_used = NULL;
  long __pyx_v_within;
  PyObject *__pyx_v_grid = NULL;
  PyObject *__pyx_v_ans = NULL;
  PyObject *__pyx_v_intron = NULL;
  PyObject *__pyx_v_overhangs = NULL;
  PyObject *__pyx_v_non_spurious_overhangs = NULL;
  PyObject *__pyx_v_res = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_retarget_reads = NULL;
  struct __pyx_obj_4aldy_9indelpost_6contig_Contig *__pyx_v_orig_contig = NULL;
  double __pyx_v_cutoff;
  PyObject *__pyx_v_target = NULL;
  PyObject *__pyx_v_nontarget = NULL;
  PyObject *__pyx_v_target_pileup = NULL;
  PyObject *__pyx_v_newly_identified = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_expected_ptrn = NULL;
  PyObject *__pyx_v_indels = NULL;
  PyObject *__pyx_v_contig_seq = NULL;
  PyObject *__pyx_v_alinger = NULL;
  PyObject *__pyx_v_new_one = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_read = NULL;
  PyObject *__pyx_8genexpr2__pyx_v_read = NULL;
  PyObject *__pyx_8genexpr3__pyx_v_read = NULL;
  PyObject *__pyx_8genexpr4__pyx_v_read = NULL;
  PyObject *__pyx_8genexpr5__pyx_v_read = NULL;
  PyObject *__pyx_8genexpr6__pyx_v_i = NULL;
  PyObject *__pyx_8genexpr7__pyx_v_d = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  float __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *(*__pyx_t_14)(PyObject *);
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  struct __pyx_opt_args_4aldy_9indelpost_6varaln_16VariantAlignment___parse_pileup __pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  Py_ssize_t __pyx_t_22;
  int __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  PyObject *(*__pyx_t_25)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__parse_pileup", 0);
  __pyx_cur_scope = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup *)__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup(__pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 171, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_contig = __pyx_optional_args->contig;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_retargeted = __pyx_optional_args->retargeted;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_skip_read_end_check = __pyx_optional_args->skip_read_end_check;
        }
      }
    }
  }
  __Pyx_INCREF((PyObject *)__pyx_v_contig);

  /* "aldy/indelpost/varaln.pyx":182
 *         cdef list pileup
 * 
 *         read_end_evidence_only = False             # <<<<<<<<<<<<<<
 * 
 *         # equivalence search
 */
  __Pyx_INCREF(Py_False);
  __pyx_v_read_end_evidence_only = Py_False;

  /* "aldy/indelpost/varaln.pyx":185
 * 
 *         # equivalence search
 *         if retargeted:             # <<<<<<<<<<<<<<
 *             pileup = self.__pileup
 *             sample_factor = self.__sample_factor
 */
  __pyx_t_1 = (__pyx_v_retargeted != 0);
  if (__pyx_t_1) {

    /* "aldy/indelpost/varaln.pyx":186
 *         # equivalence search
 *         if retargeted:
 *             pileup = self.__pileup             # <<<<<<<<<<<<<<
 *             sample_factor = self.__sample_factor
 *         else:
 */
    __pyx_t_2 = __pyx_v_self->__pyx___pileup;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_pileup = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":187
 *         if retargeted:
 *             pileup = self.__pileup
 *             sample_factor = self.__sample_factor             # <<<<<<<<<<<<<<
 *         else:
 *             pileup, self.__sample_factor = make_pileup(
 */
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx___sample_factor); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_sample_factor = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":185
 * 
 *         # equivalence search
 *         if retargeted:             # <<<<<<<<<<<<<<
 *             pileup = self.__pileup
 *             sample_factor = self.__sample_factor
 */
    goto __pyx_L3;
  }

  /* "aldy/indelpost/varaln.pyx":189
 *             sample_factor = self.__sample_factor
 *         else:
 *             pileup, self.__sample_factor = make_pileup(             # <<<<<<<<<<<<<<
 *                 self.__target,
 *                 self.bam,
 */
  /*else*/ {

    /* "aldy/indelpost/varaln.pyx":190
 *         else:
 *             pileup, self.__sample_factor = make_pileup(
 *                 self.__target,             # <<<<<<<<<<<<<<
 *                 self.bam,
 *                 self.unspliced_local_reference,
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_self->__pyx___target);
    __Pyx_INCREF(__pyx_t_2);

    /* "aldy/indelpost/varaln.pyx":191
 *             pileup, self.__sample_factor = make_pileup(
 *                 self.__target,
 *                 self.bam,             # <<<<<<<<<<<<<<
 *                 self.unspliced_local_reference,
 *                 exclude_duplicates=self.exclude_duplicates,
 */
    __pyx_t_3 = ((PyObject *)__pyx_v_self->bam);
    __Pyx_INCREF(__pyx_t_3);

    /* "aldy/indelpost/varaln.pyx":192
 *                 self.__target,
 *                 self.bam,
 *                 self.unspliced_local_reference,             # <<<<<<<<<<<<<<
 *                 exclude_duplicates=self.exclude_duplicates,
 *                 window=self.window,
 */
    __pyx_t_4 = ((PyObject *)__pyx_v_self->unspliced_local_reference);
    __Pyx_INCREF(__pyx_t_4);

    /* "aldy/indelpost/varaln.pyx":189
 *             sample_factor = self.__sample_factor
 *         else:
 *             pileup, self.__sample_factor = make_pileup(             # <<<<<<<<<<<<<<
 *                 self.__target,
 *                 self.bam,
 */
    __pyx_t_5 = __pyx_f_4aldy_9indelpost_6pileup_make_pileup(((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_2), ((struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *)__pyx_t_3), ((struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference *)__pyx_t_4), __pyx_v_self->exclude_duplicates, __pyx_v_self->window, __pyx_v_self->downsamplethresh, __pyx_v_self->basequalthresh); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(__pyx_t_5 != Py_None)) {
      PyObject* sequence = __pyx_t_5;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 189, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 189, __pyx_L1_error)
    }
    if (!(likely(PyList_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 189, __pyx_L1_error)
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_6 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 189, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_pileup = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_v_self->__pyx___sample_factor = __pyx_t_6;

    /* "aldy/indelpost/varaln.pyx":205
 *             self._observed_pos,
 *             read_end_evidence_only
 *             ) = find_by_normalization(             # <<<<<<<<<<<<<<
 *                 self.__target,
 *                 pileup,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_find_by_normalization); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 205, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "aldy/indelpost/varaln.pyx":208
 *                 self.__target,
 *                 pileup,
 *                 self.window,             # <<<<<<<<<<<<<<
 *                 self.match_score,
 *                 self.mismatch_penalty,
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->window); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "aldy/indelpost/varaln.pyx":209
 *                 pileup,
 *                 self.window,
 *                 self.match_score,             # <<<<<<<<<<<<<<
 *                 self.mismatch_penalty,
 *                 self.gap_open_penalty,
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->match_score); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "aldy/indelpost/varaln.pyx":210
 *                 self.window,
 *                 self.match_score,
 *                 self.mismatch_penalty,             # <<<<<<<<<<<<<<
 *                 self.gap_open_penalty,
 *                 self.gap_extension_penalty,
 */
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_self->mismatch_penalty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "aldy/indelpost/varaln.pyx":211
 *                 self.match_score,
 *                 self.mismatch_penalty,
 *                 self.gap_open_penalty,             # <<<<<<<<<<<<<<
 *                 self.gap_extension_penalty,
 *                 self.basequalthresh,
 */
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_self->gap_open_penalty); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "aldy/indelpost/varaln.pyx":212
 *                 self.mismatch_penalty,
 *                 self.gap_open_penalty,
 *                 self.gap_extension_penalty,             # <<<<<<<<<<<<<<
 *                 self.basequalthresh,
 *             )
 */
    __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_self->gap_extension_penalty); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);

    /* "aldy/indelpost/varaln.pyx":213
 *                 self.gap_open_penalty,
 *                 self.gap_extension_penalty,
 *                 self.basequalthresh,             # <<<<<<<<<<<<<<
 *             )
 * 
 */
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_self->basequalthresh); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = NULL;
    __pyx_t_12 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_12 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[9] = {__pyx_t_11, ((PyObject *)__pyx_v_self->__pyx___target), __pyx_v_pileup, __pyx_t_4, __pyx_t_2, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_12, 8+__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 205, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[9] = {__pyx_t_11, ((PyObject *)__pyx_v_self->__pyx___target), __pyx_v_pileup, __pyx_t_4, __pyx_t_2, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_12, 8+__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 205, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else
    #endif
    {
      __pyx_t_13 = PyTuple_New(8+__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 205, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      if (__pyx_t_11) {
        __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_11); __pyx_t_11 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
      PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_12, ((PyObject *)__pyx_v_self->__pyx___target));
      __Pyx_INCREF(__pyx_v_pileup);
      __Pyx_GIVEREF(__pyx_v_pileup);
      PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_12, __pyx_v_pileup);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_12, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_13, 3+__pyx_t_12, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_13, 4+__pyx_t_12, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_13, 5+__pyx_t_12, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_13, 6+__pyx_t_12, __pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_13, 7+__pyx_t_12, __pyx_t_10);
      __pyx_t_4 = 0;
      __pyx_t_2 = 0;
      __pyx_t_7 = 0;
      __pyx_t_8 = 0;
      __pyx_t_9 = 0;
      __pyx_t_10 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_13, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 205, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
      PyObject* sequence = __pyx_t_5;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 5)) {
        if (size > 5) __Pyx_RaiseTooManyValuesError(5);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 200, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_13 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 2); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 3); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 4); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_13 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 2); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 3); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 4); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_13);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      #else
      {
        Py_ssize_t i;
        PyObject** temps[5] = {&__pyx_t_3,&__pyx_t_13,&__pyx_t_10,&__pyx_t_9,&__pyx_t_8};
        for (i=0; i < 5; i++) {
          PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 200, __pyx_L1_error)
          __Pyx_GOTREF(item);
          *(temps[i]) = item;
        }
      }
      #endif
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      Py_ssize_t index = -1;
      PyObject** temps[5] = {&__pyx_t_3,&__pyx_t_13,&__pyx_t_10,&__pyx_t_9,&__pyx_t_8};
      __pyx_t_7 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_14 = Py_TYPE(__pyx_t_7)->tp_iternext;
      for (index=0; index < 5; index++) {
        PyObject* item = __pyx_t_14(__pyx_t_7); if (unlikely(!item)) goto __pyx_L4_unpacking_failed;
        __Pyx_GOTREF(item);
        *(temps[index]) = item;
      }
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_7), 5) < 0) __PYX_ERR(0, 200, __pyx_L1_error)
      __pyx_t_14 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5_unpacking_done;
      __pyx_L4_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_14 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 200, __pyx_L1_error)
      __pyx_L5_unpacking_done:;
    }

    /* "aldy/indelpost/varaln.pyx":200
 * 
 *             (
 *             self.__target,             # <<<<<<<<<<<<<<
 *             pileup,
 *             exptension_penalty_used,
 */
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_4aldy_9indelpost_7variant_Variant))))) __PYX_ERR(0, 200, __pyx_L1_error)
    if (!(likely(PyList_CheckExact(__pyx_t_13))||((__pyx_t_13) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_13)->tp_name), 0))) __PYX_ERR(0, 200, __pyx_L1_error)
    __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_9); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 200, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->__pyx___target);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx___target));
    __pyx_v_self->__pyx___target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_pileup, ((PyObject*)__pyx_t_13));
    __pyx_t_13 = 0;
    __pyx_v_exptension_penalty_used = __pyx_t_10;
    __pyx_t_10 = 0;

    /* "aldy/indelpost/varaln.pyx":203
 *             pileup,
 *             exptension_penalty_used,
 *             self._observed_pos,             # <<<<<<<<<<<<<<
 *             read_end_evidence_only
 *             ) = find_by_normalization(
 */
    __pyx_v_self->_observed_pos = __pyx_t_12;
    __Pyx_DECREF_SET(__pyx_v_read_end_evidence_only, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "aldy/indelpost/varaln.pyx":216
 *             )
 * 
 *             if skip_read_end_check:             # <<<<<<<<<<<<<<
 *                 read_end_evidence_only = False
 * 
 */
    __pyx_t_1 = (__pyx_v_skip_read_end_check != 0);
    if (__pyx_t_1) {

      /* "aldy/indelpost/varaln.pyx":217
 * 
 *             if skip_read_end_check:
 *                 read_end_evidence_only = False             # <<<<<<<<<<<<<<
 * 
 *             if self.target != self.__target:
 */
      __Pyx_INCREF(Py_False);
      __Pyx_DECREF_SET(__pyx_v_read_end_evidence_only, Py_False);

      /* "aldy/indelpost/varaln.pyx":216
 *             )
 * 
 *             if skip_read_end_check:             # <<<<<<<<<<<<<<
 *                 read_end_evidence_only = False
 * 
 */
    }

    /* "aldy/indelpost/varaln.pyx":219
 *                 read_end_evidence_only = False
 * 
 *             if self.target != self.__target:             # <<<<<<<<<<<<<<
 *                     self.__target, pileup = update_pileup(
 *                         pileup,
 */
    __pyx_t_5 = PyObject_RichCompare(((PyObject *)__pyx_v_self->target), ((PyObject *)__pyx_v_self->__pyx___target), Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 219, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_1) {

      /* "aldy/indelpost/varaln.pyx":220
 * 
 *             if self.target != self.__target:
 *                     self.__target, pileup = update_pileup(             # <<<<<<<<<<<<<<
 *                         pileup,
 *                         self.__target,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_update_pileup); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "aldy/indelpost/varaln.pyx":223
 *                         pileup,
 *                         self.__target,
 *                         self.window,             # <<<<<<<<<<<<<<
 *                         self.match_score,
 *                         self.mismatch_penalty,
 */
      __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_self->window); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 223, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "aldy/indelpost/varaln.pyx":224
 *                         self.__target,
 *                         self.window,
 *                         self.match_score,             # <<<<<<<<<<<<<<
 *                         self.mismatch_penalty,
 *                         self.gap_open_penalty,
 */
      __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_self->match_score); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);

      /* "aldy/indelpost/varaln.pyx":225
 *                         self.window,
 *                         self.match_score,
 *                         self.mismatch_penalty,             # <<<<<<<<<<<<<<
 *                         self.gap_open_penalty,
 *                         self.gap_extension_penalty,
 */
      __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_self->mismatch_penalty); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);

      /* "aldy/indelpost/varaln.pyx":226
 *                         self.match_score,
 *                         self.mismatch_penalty,
 *                         self.gap_open_penalty,             # <<<<<<<<<<<<<<
 *                         self.gap_extension_penalty,
 *                         self.basequalthresh,
 */
      __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_self->gap_open_penalty); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 226, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);

      /* "aldy/indelpost/varaln.pyx":227
 *                         self.mismatch_penalty,
 *                         self.gap_open_penalty,
 *                         self.gap_extension_penalty,             # <<<<<<<<<<<<<<
 *                         self.basequalthresh,
 *                         bypass_search=True
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->gap_extension_penalty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 227, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "aldy/indelpost/varaln.pyx":228
 *                         self.gap_open_penalty,
 *                         self.gap_extension_penalty,
 *                         self.basequalthresh,             # <<<<<<<<<<<<<<
 *                         bypass_search=True
 *                     )
 */
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_self->basequalthresh); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 228, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "aldy/indelpost/varaln.pyx":220
 * 
 *             if self.target != self.__target:
 *                     self.__target, pileup = update_pileup(             # <<<<<<<<<<<<<<
 *                         pileup,
 *                         self.__target,
 */
      __pyx_t_2 = PyTuple_New(8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_pileup);
      __Pyx_GIVEREF(__pyx_v_pileup);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_pileup);
      __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
      PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self->__pyx___target));
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_13);
      PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_13);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_9 = 0;
      __pyx_t_10 = 0;
      __pyx_t_13 = 0;
      __pyx_t_3 = 0;
      __pyx_t_7 = 0;

      /* "aldy/indelpost/varaln.pyx":229
 *                         self.gap_extension_penalty,
 *                         self.basequalthresh,
 *                         bypass_search=True             # <<<<<<<<<<<<<<
 *                     )
 * 
 */
      __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 229, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_bypass_search, Py_True) < 0) __PYX_ERR(0, 229, __pyx_L1_error)

      /* "aldy/indelpost/varaln.pyx":220
 * 
 *             if self.target != self.__target:
 *                     self.__target, pileup = update_pileup(             # <<<<<<<<<<<<<<
 *                         pileup,
 *                         self.__target,
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
        PyObject* sequence = __pyx_t_3;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 220, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_2);
        #else
        __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 220, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 220, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 220, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_14 = Py_TYPE(__pyx_t_5)->tp_iternext;
        index = 0; __pyx_t_7 = __pyx_t_14(__pyx_t_5); if (unlikely(!__pyx_t_7)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        index = 1; __pyx_t_2 = __pyx_t_14(__pyx_t_5); if (unlikely(!__pyx_t_2)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_5), 2) < 0) __PYX_ERR(0, 220, __pyx_L1_error)
        __pyx_t_14 = NULL;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L9_unpacking_done;
        __pyx_L8_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_14 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 220, __pyx_L1_error)
        __pyx_L9_unpacking_done:;
      }
      if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_4aldy_9indelpost_7variant_Variant))))) __PYX_ERR(0, 220, __pyx_L1_error)
      if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 220, __pyx_L1_error)
      __Pyx_GIVEREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_v_self->__pyx___target);
      __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx___target));
      __pyx_v_self->__pyx___target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_DECREF_SET(__pyx_v_pileup, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "aldy/indelpost/varaln.pyx":219
 *                 read_end_evidence_only = False
 * 
 *             if self.target != self.__target:             # <<<<<<<<<<<<<<
 *                     self.__target, pileup = update_pileup(
 *                         pileup,
 */
    }

    /* "aldy/indelpost/varaln.pyx":235
 *                 self.__target,
 *                 preprocess_for_contig_construction(
 *                     self.__target,             # <<<<<<<<<<<<<<
 *                     self.target,
 *                     pileup,
 */
    __pyx_t_3 = ((PyObject *)__pyx_v_self->__pyx___target);
    __Pyx_INCREF(__pyx_t_3);

    /* "aldy/indelpost/varaln.pyx":236
 *                 preprocess_for_contig_construction(
 *                     self.__target,
 *                     self.target,             # <<<<<<<<<<<<<<
 *                     pileup,
 *                     self.unspliced_local_reference,
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_self->target);
    __Pyx_INCREF(__pyx_t_2);

    /* "aldy/indelpost/varaln.pyx":238
 *                     self.target,
 *                     pileup,
 *                     self.unspliced_local_reference,             # <<<<<<<<<<<<<<
 *                     self.window,
 *                     self.match_score,
 */
    __pyx_t_7 = ((PyObject *)__pyx_v_self->unspliced_local_reference);
    __Pyx_INCREF(__pyx_t_7);

    /* "aldy/indelpost/varaln.pyx":243
 *                     self.mismatch_penalty,
 *                     self.gap_open_penalty,
 *                     exptension_penalty_used,             # <<<<<<<<<<<<<<
 *                 ),
 *                 self.unspliced_local_reference,
 */
    __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_exptension_penalty_used); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 243, __pyx_L1_error)

    /* "aldy/indelpost/varaln.pyx":234
 *             contig = Contig(
 *                 self.__target,
 *                 preprocess_for_contig_construction(             # <<<<<<<<<<<<<<
 *                     self.__target,
 *                     self.target,
 */
    __pyx_t_5 = __pyx_f_4aldy_9indelpost_6varaln_preprocess_for_contig_construction(((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_3), ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_2), __pyx_v_pileup, ((struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference *)__pyx_t_7), __pyx_v_self->window, __pyx_v_self->match_score, __pyx_v_self->mismatch_penalty, __pyx_v_self->gap_open_penalty, __pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "aldy/indelpost/varaln.pyx":246
 *                 ),
 *                 self.unspliced_local_reference,
 *                 self.basequalthresh,             # <<<<<<<<<<<<<<
 *                 self.mapqthresh,
 *             )
 */
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_self->basequalthresh); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "aldy/indelpost/varaln.pyx":247
 *                 self.unspliced_local_reference,
 *                 self.basequalthresh,
 *                 self.mapqthresh,             # <<<<<<<<<<<<<<
 *             )
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->mapqthresh); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "aldy/indelpost/varaln.pyx":232
 *                     )
 * 
 *             contig = Contig(             # <<<<<<<<<<<<<<
 *                 self.__target,
 *                 preprocess_for_contig_construction(
 */
    __pyx_t_3 = PyTuple_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self->__pyx___target));
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __Pyx_INCREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
    PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)__pyx_v_self->unspliced_local_reference));
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_t_2);
    __pyx_t_5 = 0;
    __pyx_t_7 = 0;
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4aldy_9indelpost_6contig_Contig), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_contig, ((struct __pyx_obj_4aldy_9indelpost_6contig_Contig *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":250
 *             )
 * 
 *             self.is_spurious_overhang = False             # <<<<<<<<<<<<<<
 *             if contig.failed and not self.no_realignment:
 *                 within = self.retarget_window
 */
    __Pyx_INCREF(Py_False);
    __Pyx_GIVEREF(Py_False);
    __Pyx_GOTREF(__pyx_v_self->is_spurious_overhang);
    __Pyx_DECREF(__pyx_v_self->is_spurious_overhang);
    __pyx_v_self->is_spurious_overhang = Py_False;

    /* "aldy/indelpost/varaln.pyx":251
 * 
 *             self.is_spurious_overhang = False
 *             if contig.failed and not self.no_realignment:             # <<<<<<<<<<<<<<
 *                 within = self.retarget_window
 * 
 */
    __pyx_t_15 = (__pyx_v_contig->failed != 0);
    if (__pyx_t_15) {
    } else {
      __pyx_t_1 = __pyx_t_15;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_15 = ((!(__pyx_v_self->no_realignment != 0)) != 0);
    __pyx_t_1 = __pyx_t_15;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_1) {

      /* "aldy/indelpost/varaln.pyx":252
 *             self.is_spurious_overhang = False
 *             if contig.failed and not self.no_realignment:
 *                 within = self.retarget_window             # <<<<<<<<<<<<<<
 * 
 *                 grid = generate_grid(self.auto_adjust_extension_penalty,
 */
      __pyx_t_12 = __pyx_v_self->retarget_window;
      __pyx_v_within = __pyx_t_12;

      /* "aldy/indelpost/varaln.pyx":254
 *                 within = self.retarget_window
 * 
 *                 grid = generate_grid(self.auto_adjust_extension_penalty,             # <<<<<<<<<<<<<<
 *                                      self.gap_open_penalty,
 *                                      self.gap_extension_penalty,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_generate_grid); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 254, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->auto_adjust_extension_penalty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 254, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "aldy/indelpost/varaln.pyx":255
 * 
 *                 grid = generate_grid(self.auto_adjust_extension_penalty,
 *                                      self.gap_open_penalty,             # <<<<<<<<<<<<<<
 *                                      self.gap_extension_penalty,
 *                                      self.__target,
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->gap_open_penalty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "aldy/indelpost/varaln.pyx":256
 *                 grid = generate_grid(self.auto_adjust_extension_penalty,
 *                                      self.gap_open_penalty,
 *                                      self.gap_extension_penalty,             # <<<<<<<<<<<<<<
 *                                      self.__target,
 *                        )
 */
      __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_self->gap_extension_penalty); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 256, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);

      /* "aldy/indelpost/varaln.pyx":257
 *                                      self.gap_open_penalty,
 *                                      self.gap_extension_penalty,
 *                                      self.__target,             # <<<<<<<<<<<<<<
 *                        )
 * 
 */
      __pyx_t_10 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_12 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[5] = {__pyx_t_10, __pyx_t_7, __pyx_t_5, __pyx_t_13, ((PyObject *)__pyx_v_self->__pyx___target)};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 254, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[5] = {__pyx_t_10, __pyx_t_7, __pyx_t_5, __pyx_t_13, ((PyObject *)__pyx_v_self->__pyx___target)};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 254, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(4+__pyx_t_12); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 254, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_12, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_12, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_12, __pyx_t_13);
        __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
        PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_12, ((PyObject *)__pyx_v_self->__pyx___target));
        __pyx_t_7 = 0;
        __pyx_t_5 = 0;
        __pyx_t_13 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 254, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_grid = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "aldy/indelpost/varaln.pyx":260
 *                        )
 * 
 *                 ans = check_overhangs(pileup)             # <<<<<<<<<<<<<<
 * 
 *                 if ans:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_check_overhangs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_9, __pyx_v_pileup) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_pileup);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_ans = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "aldy/indelpost/varaln.pyx":262
 *                 ans = check_overhangs(pileup)
 * 
 *                 if ans:             # <<<<<<<<<<<<<<
 *                     intron, overhangs = ans[0], ans[1]
 *                     non_spurious_overhangs = filter_spurious_overhangs(
 */
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_ans); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 262, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "aldy/indelpost/varaln.pyx":263
 * 
 *                 if ans:
 *                     intron, overhangs = ans[0], ans[1]             # <<<<<<<<<<<<<<
 *                     non_spurious_overhangs = filter_spurious_overhangs(
 *                         self.__target,
 */
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_ans, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 263, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_ans, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 263, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_v_intron = __pyx_t_2;
        __pyx_t_2 = 0;
        __pyx_v_overhangs = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "aldy/indelpost/varaln.pyx":264
 *                 if ans:
 *                     intron, overhangs = ans[0], ans[1]
 *                     non_spurious_overhangs = filter_spurious_overhangs(             # <<<<<<<<<<<<<<
 *                         self.__target,
 *                         intron,
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_filter_spurious_overhangs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 264, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);

        /* "aldy/indelpost/varaln.pyx":268
 *                         intron,
 *                         overhangs,
 *                         self.match_score,             # <<<<<<<<<<<<<<
 *                         self.mismatch_penalty,
 *                         self.gap_open_penalty,
 */
        __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_self->match_score); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 268, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);

        /* "aldy/indelpost/varaln.pyx":269
 *                         overhangs,
 *                         self.match_score,
 *                         self.mismatch_penalty,             # <<<<<<<<<<<<<<
 *                         self.gap_open_penalty,
 *                         self.gap_extension_penalty,
 */
        __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_self->mismatch_penalty); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 269, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);

        /* "aldy/indelpost/varaln.pyx":270
 *                         self.match_score,
 *                         self.mismatch_penalty,
 *                         self.gap_open_penalty,             # <<<<<<<<<<<<<<
 *                         self.gap_extension_penalty,
 *                     )
 */
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->gap_open_penalty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 270, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);

        /* "aldy/indelpost/varaln.pyx":271
 *                         self.mismatch_penalty,
 *                         self.gap_open_penalty,
 *                         self.gap_extension_penalty,             # <<<<<<<<<<<<<<
 *                     )
 *                     if not non_spurious_overhangs:
 */
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_self->gap_extension_penalty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 271, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = NULL;
        __pyx_t_12 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_12 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[8] = {__pyx_t_10, ((PyObject *)__pyx_v_self->__pyx___target), __pyx_v_intron, __pyx_v_overhangs, __pyx_t_9, __pyx_t_13, __pyx_t_5, __pyx_t_7};
          __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_12, 7+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 264, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[8] = {__pyx_t_10, ((PyObject *)__pyx_v_self->__pyx___target), __pyx_v_intron, __pyx_v_overhangs, __pyx_t_9, __pyx_t_13, __pyx_t_5, __pyx_t_7};
          __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_12, 7+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 264, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(7+__pyx_t_12); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 264, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_10) {
            __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_10); __pyx_t_10 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_12, ((PyObject *)__pyx_v_self->__pyx___target));
          __Pyx_INCREF(__pyx_v_intron);
          __Pyx_GIVEREF(__pyx_v_intron);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_12, __pyx_v_intron);
          __Pyx_INCREF(__pyx_v_overhangs);
          __Pyx_GIVEREF(__pyx_v_overhangs);
          PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_12, __pyx_v_overhangs);
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_12, __pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_13);
          PyTuple_SET_ITEM(__pyx_t_8, 4+__pyx_t_12, __pyx_t_13);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_8, 5+__pyx_t_12, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_8, 6+__pyx_t_12, __pyx_t_7);
          __pyx_t_9 = 0;
          __pyx_t_13 = 0;
          __pyx_t_5 = 0;
          __pyx_t_7 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 264, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_non_spurious_overhangs = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "aldy/indelpost/varaln.pyx":273
 *                         self.gap_extension_penalty,
 *                     )
 *                     if not non_spurious_overhangs:             # <<<<<<<<<<<<<<
 *                         contig = Contig(self.__target, [], self.unspliced_local_reference, self.basequalthresh, self.mapqthresh)
 *                         self.is_spurious_overhang = True
 */
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_non_spurious_overhangs); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 273, __pyx_L1_error)
        __pyx_t_15 = ((!__pyx_t_1) != 0);
        if (__pyx_t_15) {

          /* "aldy/indelpost/varaln.pyx":274
 *                     )
 *                     if not non_spurious_overhangs:
 *                         contig = Contig(self.__target, [], self.unspliced_local_reference, self.basequalthresh, self.mapqthresh)             # <<<<<<<<<<<<<<
 *                         self.is_spurious_overhang = True
 *                         return pileup, contig
 */
          __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 274, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->basequalthresh); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 274, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_self->mapqthresh); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 274, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 274, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
          PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_v_self->__pyx___target));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_3);
          __Pyx_INCREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
          PyTuple_SET_ITEM(__pyx_t_7, 2, ((PyObject *)__pyx_v_self->unspliced_local_reference));
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_8);
          __pyx_t_3 = 0;
          __pyx_t_2 = 0;
          __pyx_t_8 = 0;
          __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4aldy_9indelpost_6contig_Contig), __pyx_t_7, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 274, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF_SET(__pyx_v_contig, ((struct __pyx_obj_4aldy_9indelpost_6contig_Contig *)__pyx_t_8));
          __pyx_t_8 = 0;

          /* "aldy/indelpost/varaln.pyx":275
 *                     if not non_spurious_overhangs:
 *                         contig = Contig(self.__target, [], self.unspliced_local_reference, self.basequalthresh, self.mapqthresh)
 *                         self.is_spurious_overhang = True             # <<<<<<<<<<<<<<
 *                         return pileup, contig
 *                     else:
 */
          __Pyx_INCREF(Py_True);
          __Pyx_GIVEREF(Py_True);
          __Pyx_GOTREF(__pyx_v_self->is_spurious_overhang);
          __Pyx_DECREF(__pyx_v_self->is_spurious_overhang);
          __pyx_v_self->is_spurious_overhang = Py_True;

          /* "aldy/indelpost/varaln.pyx":276
 *                         contig = Contig(self.__target, [], self.unspliced_local_reference, self.basequalthresh, self.mapqthresh)
 *                         self.is_spurious_overhang = True
 *                         return pileup, contig             # <<<<<<<<<<<<<<
 *                     else:
 *                         res = grid_search(
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 276, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_INCREF(__pyx_v_pileup);
          __Pyx_GIVEREF(__pyx_v_pileup);
          PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_pileup);
          __Pyx_INCREF(((PyObject *)__pyx_v_contig));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_contig));
          PyTuple_SET_ITEM(__pyx_t_8, 1, ((PyObject *)__pyx_v_contig));
          __pyx_r = __pyx_t_8;
          __pyx_t_8 = 0;
          goto __pyx_L0;

          /* "aldy/indelpost/varaln.pyx":273
 *                         self.gap_extension_penalty,
 *                     )
 *                     if not non_spurious_overhangs:             # <<<<<<<<<<<<<<
 *                         contig = Contig(self.__target, [], self.unspliced_local_reference, self.basequalthresh, self.mapqthresh)
 *                         self.is_spurious_overhang = True
 */
        }

        /* "aldy/indelpost/varaln.pyx":278
 *                         return pileup, contig
 *                     else:
 *                         res = grid_search(             # <<<<<<<<<<<<<<
 *                             self.__target,
 *                             non_spurious_overhangs,
 */
        /*else*/ {
          __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_grid_search); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 278, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);

          /* "aldy/indelpost/varaln.pyx":281
 *                             self.__target,
 *                             non_spurious_overhangs,
 *                             self.window,             # <<<<<<<<<<<<<<
 *                             self.mapqthresh,
 *                             within,
 */
          __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->window); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 281, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);

          /* "aldy/indelpost/varaln.pyx":282
 *                             non_spurious_overhangs,
 *                             self.window,
 *                             self.mapqthresh,             # <<<<<<<<<<<<<<
 *                             within,
 *                             self.retarget_cutoff,
 */
          __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->mapqthresh); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 282, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);

          /* "aldy/indelpost/varaln.pyx":283
 *                             self.window,
 *                             self.mapqthresh,
 *                             within,             # <<<<<<<<<<<<<<
 *                             self.retarget_cutoff,
 *                             self.match_score,
 */
          __pyx_t_5 = __Pyx_PyInt_From_long(__pyx_v_within); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 283, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);

          /* "aldy/indelpost/varaln.pyx":284
 *                             self.mapqthresh,
 *                             within,
 *                             self.retarget_cutoff,             # <<<<<<<<<<<<<<
 *                             self.match_score,
 *                             self.mismatch_penalty,
 */
          __pyx_t_13 = PyFloat_FromDouble(__pyx_v_self->retarget_cutoff); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 284, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);

          /* "aldy/indelpost/varaln.pyx":285
 *                             within,
 *                             self.retarget_cutoff,
 *                             self.match_score,             # <<<<<<<<<<<<<<
 *                             self.mismatch_penalty,
 *                             grid,
 */
          __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_self->match_score); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 285, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);

          /* "aldy/indelpost/varaln.pyx":286
 *                             self.retarget_cutoff,
 *                             self.match_score,
 *                             self.mismatch_penalty,             # <<<<<<<<<<<<<<
 *                             grid,
 *                             self.unspliced_local_reference,
 */
          __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_self->mismatch_penalty); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 286, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);

          /* "aldy/indelpost/varaln.pyx":289
 *                             grid,
 *                             self.unspliced_local_reference,
 *                             self.exact_match_for_shiftable,             # <<<<<<<<<<<<<<
 *                         )
 * 
 */
          __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->exact_match_for_shiftable); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_11 = NULL;
          __pyx_t_12 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_11)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_12 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[12] = {__pyx_t_11, ((PyObject *)__pyx_v_self->__pyx___target), __pyx_v_non_spurious_overhangs, __pyx_t_2, __pyx_t_3, __pyx_t_5, __pyx_t_13, __pyx_t_9, __pyx_t_10, __pyx_v_grid, ((PyObject *)__pyx_v_self->unspliced_local_reference), __pyx_t_4};
            __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_12, 11+__pyx_t_12); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 278, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[12] = {__pyx_t_11, ((PyObject *)__pyx_v_self->__pyx___target), __pyx_v_non_spurious_overhangs, __pyx_t_2, __pyx_t_3, __pyx_t_5, __pyx_t_13, __pyx_t_9, __pyx_t_10, __pyx_v_grid, ((PyObject *)__pyx_v_self->unspliced_local_reference), __pyx_t_4};
            __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_12, 11+__pyx_t_12); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 278, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else
          #endif
          {
            __pyx_t_16 = PyTuple_New(11+__pyx_t_12); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 278, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            if (__pyx_t_11) {
              __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_11); __pyx_t_11 = NULL;
            }
            __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
            PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_12, ((PyObject *)__pyx_v_self->__pyx___target));
            __Pyx_INCREF(__pyx_v_non_spurious_overhangs);
            __Pyx_GIVEREF(__pyx_v_non_spurious_overhangs);
            PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_12, __pyx_v_non_spurious_overhangs);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_12, __pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_16, 3+__pyx_t_12, __pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_16, 4+__pyx_t_12, __pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_13);
            PyTuple_SET_ITEM(__pyx_t_16, 5+__pyx_t_12, __pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_9);
            PyTuple_SET_ITEM(__pyx_t_16, 6+__pyx_t_12, __pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_10);
            PyTuple_SET_ITEM(__pyx_t_16, 7+__pyx_t_12, __pyx_t_10);
            __Pyx_INCREF(__pyx_v_grid);
            __Pyx_GIVEREF(__pyx_v_grid);
            PyTuple_SET_ITEM(__pyx_t_16, 8+__pyx_t_12, __pyx_v_grid);
            __Pyx_INCREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
            PyTuple_SET_ITEM(__pyx_t_16, 9+__pyx_t_12, ((PyObject *)__pyx_v_self->unspliced_local_reference));
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_16, 10+__pyx_t_12, __pyx_t_4);
            __pyx_t_2 = 0;
            __pyx_t_3 = 0;
            __pyx_t_5 = 0;
            __pyx_t_13 = 0;
            __pyx_t_9 = 0;
            __pyx_t_10 = 0;
            __pyx_t_4 = 0;
            __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_16, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 278, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_v_res = __pyx_t_8;
          __pyx_t_8 = 0;

          /* "aldy/indelpost/varaln.pyx":292
 *                         )
 * 
 *                         if res:             # <<<<<<<<<<<<<<
 *                             self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]
 *                         else:
 */
          __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_v_res); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 292, __pyx_L1_error)
          if (__pyx_t_15) {

            /* "aldy/indelpost/varaln.pyx":293
 * 
 *                         if res:
 *                             self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]             # <<<<<<<<<<<<<<
 *                         else:
 *                             contig = Contig(self.__target, [], self.unspliced_local_reference, self.basequalthresh, self.mapqthresh)
 */
            __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_res, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 293, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 293, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_res, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 293, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_17 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_17 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 293, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __pyx_v_self->gap_open_penalty = __pyx_t_12;
            __pyx_v_self->gap_extension_penalty = __pyx_t_17;

            /* "aldy/indelpost/varaln.pyx":292
 *                         )
 * 
 *                         if res:             # <<<<<<<<<<<<<<
 *                             self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]
 *                         else:
 */
            goto __pyx_L15;
          }

          /* "aldy/indelpost/varaln.pyx":295
 *                             self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]
 *                         else:
 *                             contig = Contig(self.__target, [], self.unspliced_local_reference, self.basequalthresh, self.mapqthresh)             # <<<<<<<<<<<<<<
 *                             self.is_spurious_overhang = True
 *                             return pileup, contig
 */
          /*else*/ {
            __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 295, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_self->basequalthresh); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 295, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_self->mapqthresh); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 295, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 295, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
            PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_self->__pyx___target));
            __Pyx_GIVEREF(__pyx_t_8);
            PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_8);
            __Pyx_INCREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
            PyTuple_SET_ITEM(__pyx_t_4, 2, ((PyObject *)__pyx_v_self->unspliced_local_reference));
            __Pyx_GIVEREF(__pyx_t_7);
            PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_16);
            PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_t_16);
            __pyx_t_8 = 0;
            __pyx_t_7 = 0;
            __pyx_t_16 = 0;
            __pyx_t_16 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4aldy_9indelpost_6contig_Contig), __pyx_t_4, NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 295, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF_SET(__pyx_v_contig, ((struct __pyx_obj_4aldy_9indelpost_6contig_Contig *)__pyx_t_16));
            __pyx_t_16 = 0;

            /* "aldy/indelpost/varaln.pyx":296
 *                         else:
 *                             contig = Contig(self.__target, [], self.unspliced_local_reference, self.basequalthresh, self.mapqthresh)
 *                             self.is_spurious_overhang = True             # <<<<<<<<<<<<<<
 *                             return pileup, contig
 *                 else:
 */
            __Pyx_INCREF(Py_True);
            __Pyx_GIVEREF(Py_True);
            __Pyx_GOTREF(__pyx_v_self->is_spurious_overhang);
            __Pyx_DECREF(__pyx_v_self->is_spurious_overhang);
            __pyx_v_self->is_spurious_overhang = Py_True;

            /* "aldy/indelpost/varaln.pyx":297
 *                             contig = Contig(self.__target, [], self.unspliced_local_reference, self.basequalthresh, self.mapqthresh)
 *                             self.is_spurious_overhang = True
 *                             return pileup, contig             # <<<<<<<<<<<<<<
 *                 else:
 *                     res = grid_search(
 */
            __Pyx_XDECREF(__pyx_r);
            __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 297, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            __Pyx_INCREF(__pyx_v_pileup);
            __Pyx_GIVEREF(__pyx_v_pileup);
            PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_v_pileup);
            __Pyx_INCREF(((PyObject *)__pyx_v_contig));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_contig));
            PyTuple_SET_ITEM(__pyx_t_16, 1, ((PyObject *)__pyx_v_contig));
            __pyx_r = __pyx_t_16;
            __pyx_t_16 = 0;
            goto __pyx_L0;
          }
          __pyx_L15:;
        }

        /* "aldy/indelpost/varaln.pyx":262
 *                 ans = check_overhangs(pileup)
 * 
 *                 if ans:             # <<<<<<<<<<<<<<
 *                     intron, overhangs = ans[0], ans[1]
 *                     non_spurious_overhangs = filter_spurious_overhangs(
 */
        goto __pyx_L13;
      }

      /* "aldy/indelpost/varaln.pyx":299
 *                             return pileup, contig
 *                 else:
 *                     res = grid_search(             # <<<<<<<<<<<<<<
 *                         self.__target,
 *                         pileup,
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_grid_search); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 299, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);

        /* "aldy/indelpost/varaln.pyx":302
 *                         self.__target,
 *                         pileup,
 *                         self.window,             # <<<<<<<<<<<<<<
 *                         self.mapqthresh,
 *                         within,
 */
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_self->window); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 302, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "aldy/indelpost/varaln.pyx":303
 *                         pileup,
 *                         self.window,
 *                         self.mapqthresh,             # <<<<<<<<<<<<<<
 *                         within,
 *                         self.retarget_cutoff,
 */
        __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_self->mapqthresh); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 303, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);

        /* "aldy/indelpost/varaln.pyx":304
 *                         self.window,
 *                         self.mapqthresh,
 *                         within,             # <<<<<<<<<<<<<<
 *                         self.retarget_cutoff,
 *                         self.match_score,
 */
        __pyx_t_10 = __Pyx_PyInt_From_long(__pyx_v_within); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 304, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);

        /* "aldy/indelpost/varaln.pyx":305
 *                         self.mapqthresh,
 *                         within,
 *                         self.retarget_cutoff,             # <<<<<<<<<<<<<<
 *                         self.match_score,
 *                         self.mismatch_penalty,
 */
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->retarget_cutoff); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 305, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);

        /* "aldy/indelpost/varaln.pyx":306
 *                         within,
 *                         self.retarget_cutoff,
 *                         self.match_score,             # <<<<<<<<<<<<<<
 *                         self.mismatch_penalty,
 *                         grid,
 */
        __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_self->match_score); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 306, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);

        /* "aldy/indelpost/varaln.pyx":307
 *                         self.retarget_cutoff,
 *                         self.match_score,
 *                         self.mismatch_penalty,             # <<<<<<<<<<<<<<
 *                         grid,
 *                         self.unspliced_local_reference,
 */
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->mismatch_penalty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 307, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);

        /* "aldy/indelpost/varaln.pyx":310
 *                         grid,
 *                         self.unspliced_local_reference,
 *                         self.exact_match_for_shiftable,             # <<<<<<<<<<<<<<
 *                     )
 * 
 */
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->exact_match_for_shiftable); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_17 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_17 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[12] = {__pyx_t_2, ((PyObject *)__pyx_v_self->__pyx___target), __pyx_v_pileup, __pyx_t_7, __pyx_t_8, __pyx_t_10, __pyx_t_9, __pyx_t_13, __pyx_t_5, __pyx_v_grid, ((PyObject *)__pyx_v_self->unspliced_local_reference), __pyx_t_3};
          __pyx_t_16 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_17, 11+__pyx_t_17); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 299, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[12] = {__pyx_t_2, ((PyObject *)__pyx_v_self->__pyx___target), __pyx_v_pileup, __pyx_t_7, __pyx_t_8, __pyx_t_10, __pyx_t_9, __pyx_t_13, __pyx_t_5, __pyx_v_grid, ((PyObject *)__pyx_v_self->unspliced_local_reference), __pyx_t_3};
          __pyx_t_16 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_17, 11+__pyx_t_17); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 299, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_11 = PyTuple_New(11+__pyx_t_17); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 299, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
          PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_17, ((PyObject *)__pyx_v_self->__pyx___target));
          __Pyx_INCREF(__pyx_v_pileup);
          __Pyx_GIVEREF(__pyx_v_pileup);
          PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_17, __pyx_v_pileup);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_17, __pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_17, __pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_11, 4+__pyx_t_17, __pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_11, 5+__pyx_t_17, __pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_13);
          PyTuple_SET_ITEM(__pyx_t_11, 6+__pyx_t_17, __pyx_t_13);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_11, 7+__pyx_t_17, __pyx_t_5);
          __Pyx_INCREF(__pyx_v_grid);
          __Pyx_GIVEREF(__pyx_v_grid);
          PyTuple_SET_ITEM(__pyx_t_11, 8+__pyx_t_17, __pyx_v_grid);
          __Pyx_INCREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
          PyTuple_SET_ITEM(__pyx_t_11, 9+__pyx_t_17, ((PyObject *)__pyx_v_self->unspliced_local_reference));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_11, 10+__pyx_t_17, __pyx_t_3);
          __pyx_t_7 = 0;
          __pyx_t_8 = 0;
          __pyx_t_10 = 0;
          __pyx_t_9 = 0;
          __pyx_t_13 = 0;
          __pyx_t_5 = 0;
          __pyx_t_3 = 0;
          __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_11, NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 299, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v_res = __pyx_t_16;
        __pyx_t_16 = 0;

        /* "aldy/indelpost/varaln.pyx":313
 *                     )
 * 
 *                     if res:             # <<<<<<<<<<<<<<
 *                         self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]
 * 
 */
        __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_v_res); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 313, __pyx_L1_error)
        if (__pyx_t_15) {

          /* "aldy/indelpost/varaln.pyx":314
 * 
 *                     if res:
 *                         self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]             # <<<<<<<<<<<<<<
 * 
 *                 # if retargeted successfully -> make contig based on the retarget
 */
          __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_res, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 314, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_17 = __Pyx_PyInt_As_int(__pyx_t_16); if (unlikely((__pyx_t_17 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 314, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_res, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 314, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_16); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 314, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_v_self->gap_open_penalty = __pyx_t_17;
          __pyx_v_self->gap_extension_penalty = __pyx_t_12;

          /* "aldy/indelpost/varaln.pyx":313
 *                     )
 * 
 *                     if res:             # <<<<<<<<<<<<<<
 *                         self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]
 * 
 */
        }
      }
      __pyx_L13:;

      /* "aldy/indelpost/varaln.pyx":317
 * 
 *                 # if retargeted successfully -> make contig based on the retarget
 *                 if res:             # <<<<<<<<<<<<<<
 *                     self.__target, retarget_reads = res[0], res[1]
 * 
 */
      __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_v_res); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 317, __pyx_L1_error)
      if (__pyx_t_15) {

        /* "aldy/indelpost/varaln.pyx":318
 *                 # if retargeted successfully -> make contig based on the retarget
 *                 if res:
 *                     self.__target, retarget_reads = res[0], res[1]             # <<<<<<<<<<<<<<
 * 
 *                     self.__target, self.__pileup = update_pileup(
 */
        __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_res, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 318, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        if (!(likely(((__pyx_t_16) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_16, __pyx_ptype_4aldy_9indelpost_7variant_Variant))))) __PYX_ERR(0, 318, __pyx_L1_error)
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_res, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 318, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_16);
        __Pyx_GOTREF(__pyx_v_self->__pyx___target);
        __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx___target));
        __pyx_v_self->__pyx___target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_16);
        __pyx_t_16 = 0;
        __pyx_v_retarget_reads = __pyx_t_4;
        __pyx_t_4 = 0;

        /* "aldy/indelpost/varaln.pyx":320
 *                     self.__target, retarget_reads = res[0], res[1]
 * 
 *                     self.__target, self.__pileup = update_pileup(             # <<<<<<<<<<<<<<
 *                         pileup,
 *                         self.__target,
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_update_pileup); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);

        /* "aldy/indelpost/varaln.pyx":323
 *                         pileup,
 *                         self.__target,
 *                         self.window,             # <<<<<<<<<<<<<<
 *                         self.match_score,
 *                         self.mismatch_penalty,
 */
        __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_self->window); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 323, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);

        /* "aldy/indelpost/varaln.pyx":324
 *                         self.__target,
 *                         self.window,
 *                         self.match_score,             # <<<<<<<<<<<<<<
 *                         self.mismatch_penalty,
 *                         self.gap_open_penalty,
 */
        __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_self->match_score); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 324, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);

        /* "aldy/indelpost/varaln.pyx":325
 *                         self.window,
 *                         self.match_score,
 *                         self.mismatch_penalty,             # <<<<<<<<<<<<<<
 *                         self.gap_open_penalty,
 *                         self.gap_extension_penalty,
 */
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->mismatch_penalty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 325, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);

        /* "aldy/indelpost/varaln.pyx":326
 *                         self.match_score,
 *                         self.mismatch_penalty,
 *                         self.gap_open_penalty,             # <<<<<<<<<<<<<<
 *                         self.gap_extension_penalty,
 *                         self.basequalthresh,
 */
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->gap_open_penalty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 326, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);

        /* "aldy/indelpost/varaln.pyx":327
 *                         self.mismatch_penalty,
 *                         self.gap_open_penalty,
 *                         self.gap_extension_penalty,             # <<<<<<<<<<<<<<
 *                         self.basequalthresh,
 *                         bypass_search=True,
 */
        __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_self->gap_extension_penalty); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 327, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);

        /* "aldy/indelpost/varaln.pyx":328
 *                         self.gap_open_penalty,
 *                         self.gap_extension_penalty,
 *                         self.basequalthresh,             # <<<<<<<<<<<<<<
 *                         bypass_search=True,
 *                     )
 */
        __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_self->basequalthresh); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 328, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);

        /* "aldy/indelpost/varaln.pyx":320
 *                     self.__target, retarget_reads = res[0], res[1]
 * 
 *                     self.__target, self.__pileup = update_pileup(             # <<<<<<<<<<<<<<
 *                         pileup,
 *                         self.__target,
 */
        __pyx_t_10 = PyTuple_New(8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_INCREF(__pyx_v_pileup);
        __Pyx_GIVEREF(__pyx_v_pileup);
        PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_pileup);
        __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
        PyTuple_SET_ITEM(__pyx_t_10, 1, ((PyObject *)__pyx_v_self->__pyx___target));
        __Pyx_GIVEREF(__pyx_t_16);
        PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_16);
        __Pyx_GIVEREF(__pyx_t_11);
        PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_11);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_10, 5, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_10, 6, __pyx_t_13);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_10, 7, __pyx_t_9);
        __pyx_t_16 = 0;
        __pyx_t_11 = 0;
        __pyx_t_3 = 0;
        __pyx_t_5 = 0;
        __pyx_t_13 = 0;
        __pyx_t_9 = 0;

        /* "aldy/indelpost/varaln.pyx":329
 *                         self.gap_extension_penalty,
 *                         self.basequalthresh,
 *                         bypass_search=True,             # <<<<<<<<<<<<<<
 *                     )
 * 
 */
        __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 329, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_bypass_search, Py_True) < 0) __PYX_ERR(0, 329, __pyx_L1_error)

        /* "aldy/indelpost/varaln.pyx":320
 *                     self.__target, retarget_reads = res[0], res[1]
 * 
 *                     self.__target, self.__pileup = update_pileup(             # <<<<<<<<<<<<<<
 *                         pileup,
 *                         self.__target,
 */
        __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_10, __pyx_t_9); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_13))) || (PyList_CheckExact(__pyx_t_13))) {
          PyObject* sequence = __pyx_t_13;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 320, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          #else
          __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 320, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 320, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_4 = PyObject_GetIter(__pyx_t_13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 320, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __pyx_t_14 = Py_TYPE(__pyx_t_4)->tp_iternext;
          index = 0; __pyx_t_9 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_9)) goto __pyx_L18_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_9);
          index = 1; __pyx_t_10 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_10)) goto __pyx_L18_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_10);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_4), 2) < 0) __PYX_ERR(0, 320, __pyx_L1_error)
          __pyx_t_14 = NULL;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          goto __pyx_L19_unpacking_done;
          __pyx_L18_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_14 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 320, __pyx_L1_error)
          __pyx_L19_unpacking_done:;
        }
        if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_4aldy_9indelpost_7variant_Variant))))) __PYX_ERR(0, 320, __pyx_L1_error)
        if (!(likely(PyList_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_10)->tp_name), 0))) __PYX_ERR(0, 320, __pyx_L1_error)
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GOTREF(__pyx_v_self->__pyx___target);
        __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx___target));
        __pyx_v_self->__pyx___target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_9);
        __pyx_t_9 = 0;
        __Pyx_GIVEREF(__pyx_t_10);
        __Pyx_GOTREF(__pyx_v_self->__pyx___pileup);
        __Pyx_DECREF(__pyx_v_self->__pyx___pileup);
        __pyx_v_self->__pyx___pileup = ((PyObject*)__pyx_t_10);
        __pyx_t_10 = 0;

        /* "aldy/indelpost/varaln.pyx":335
 *                         self.__target,
 *                         preprocess_for_contig_construction(
 *                             self.__target,             # <<<<<<<<<<<<<<
 *                             self.__target,
 *                             self.__pileup,
 */
        __pyx_t_13 = ((PyObject *)__pyx_v_self->__pyx___target);
        __Pyx_INCREF(__pyx_t_13);

        /* "aldy/indelpost/varaln.pyx":336
 *                         preprocess_for_contig_construction(
 *                             self.__target,
 *                             self.__target,             # <<<<<<<<<<<<<<
 *                             self.__pileup,
 *                             self.unspliced_local_reference,
 */
        __pyx_t_10 = ((PyObject *)__pyx_v_self->__pyx___target);
        __Pyx_INCREF(__pyx_t_10);

        /* "aldy/indelpost/varaln.pyx":337
 *                             self.__target,
 *                             self.__target,
 *                             self.__pileup,             # <<<<<<<<<<<<<<
 *                             self.unspliced_local_reference,
 *                             self.window,
 */
        __pyx_t_9 = __pyx_v_self->__pyx___pileup;
        __Pyx_INCREF(__pyx_t_9);

        /* "aldy/indelpost/varaln.pyx":338
 *                             self.__target,
 *                             self.__pileup,
 *                             self.unspliced_local_reference,             # <<<<<<<<<<<<<<
 *                             self.window,
 *                             self.match_score,
 */
        __pyx_t_4 = ((PyObject *)__pyx_v_self->unspliced_local_reference);
        __Pyx_INCREF(__pyx_t_4);

        /* "aldy/indelpost/varaln.pyx":334
 *                     contig = Contig(
 *                         self.__target,
 *                         preprocess_for_contig_construction(             # <<<<<<<<<<<<<<
 *                             self.__target,
 *                             self.__target,
 */
        __pyx_t_5 = __pyx_f_4aldy_9indelpost_6varaln_preprocess_for_contig_construction(((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_13), ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_10), ((PyObject*)__pyx_t_9), ((struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference *)__pyx_t_4), __pyx_v_self->window, __pyx_v_self->match_score, __pyx_v_self->mismatch_penalty, __pyx_v_self->gap_open_penalty, __pyx_v_self->gap_extension_penalty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "aldy/indelpost/varaln.pyx":346
 *                         ),
 *                         self.unspliced_local_reference,
 *                         self.basequalthresh,             # <<<<<<<<<<<<<<
 *                         self.mapqthresh
 *                     )
 */
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->basequalthresh); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 346, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);

        /* "aldy/indelpost/varaln.pyx":347
 *                         self.unspliced_local_reference,
 *                         self.basequalthresh,
 *                         self.mapqthresh             # <<<<<<<<<<<<<<
 *                     )
 * 
 */
        __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_self->mapqthresh); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 347, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);

        /* "aldy/indelpost/varaln.pyx":332
 *                     )
 * 
 *                     contig = Contig(             # <<<<<<<<<<<<<<
 *                         self.__target,
 *                         preprocess_for_contig_construction(
 */
        __pyx_t_10 = PyTuple_New(5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 332, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
        PyTuple_SET_ITEM(__pyx_t_10, 0, ((PyObject *)__pyx_v_self->__pyx___target));
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_5);
        __Pyx_INCREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
        PyTuple_SET_ITEM(__pyx_t_10, 2, ((PyObject *)__pyx_v_self->unspliced_local_reference));
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_t_9);
        __pyx_t_5 = 0;
        __pyx_t_4 = 0;
        __pyx_t_9 = 0;
        __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4aldy_9indelpost_6contig_Contig), __pyx_t_10, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 332, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF_SET(__pyx_v_contig, ((struct __pyx_obj_4aldy_9indelpost_6contig_Contig *)__pyx_t_9));
        __pyx_t_9 = 0;

        /* "aldy/indelpost/varaln.pyx":351
 * 
 *                     # 2nd-pass using the retarget
 *                     return self.__parse_pileup(contig=contig, retargeted=True)             # <<<<<<<<<<<<<<
 * 
 *                 # no target in this pileup
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_18.__pyx_n = 2;
        __pyx_t_18.contig = __pyx_v_contig;
        __pyx_t_18.retargeted = 1;
        __pyx_t_9 = ((struct __pyx_vtabstruct_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self->__pyx_vtab)->__pyx___parse_pileup(__pyx_v_self, &__pyx_t_18); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 351, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_r = __pyx_t_9;
        __pyx_t_9 = 0;
        goto __pyx_L0;

        /* "aldy/indelpost/varaln.pyx":317
 * 
 *                 # if retargeted successfully -> make contig based on the retarget
 *                 if res:             # <<<<<<<<<<<<<<
 *                     self.__target, retarget_reads = res[0], res[1]
 * 
 */
      }

      /* "aldy/indelpost/varaln.pyx":355
 *                 # no target in this pileup
 *                 else:
 *                     if self.is_complex_input:             # <<<<<<<<<<<<<<
 *                         try:
 *                             self.__target = self.second_target
 */
      /*else*/ {
        __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_v_self->is_complex_input); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 355, __pyx_L1_error)
        if (__pyx_t_15) {

          /* "aldy/indelpost/varaln.pyx":356
 *                 else:
 *                     if self.is_complex_input:
 *                         try:             # <<<<<<<<<<<<<<
 *                             self.__target = self.second_target
 *                             self.target = self.second_target
 */
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            /*try:*/ {

              /* "aldy/indelpost/varaln.pyx":357
 *                     if self.is_complex_input:
 *                         try:
 *                             self.__target = self.second_target             # <<<<<<<<<<<<<<
 *                             self.target = self.second_target
 *                             self.is_complex_input = False
 */
              __pyx_t_9 = ((PyObject *)__pyx_v_self->second_target);
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_GIVEREF(__pyx_t_9);
              __Pyx_GOTREF(__pyx_v_self->__pyx___target);
              __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx___target));
              __pyx_v_self->__pyx___target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_9);
              __pyx_t_9 = 0;

              /* "aldy/indelpost/varaln.pyx":358
 *                         try:
 *                             self.__target = self.second_target
 *                             self.target = self.second_target             # <<<<<<<<<<<<<<
 *                             self.is_complex_input = False
 *                             return self.__parse_pileup(contig=None, retargeted=False, skip_read_end_check=True)
 */
              __pyx_t_9 = ((PyObject *)__pyx_v_self->second_target);
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_GIVEREF(__pyx_t_9);
              __Pyx_GOTREF(__pyx_v_self->target);
              __Pyx_DECREF(((PyObject *)__pyx_v_self->target));
              __pyx_v_self->target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_9);
              __pyx_t_9 = 0;

              /* "aldy/indelpost/varaln.pyx":359
 *                             self.__target = self.second_target
 *                             self.target = self.second_target
 *                             self.is_complex_input = False             # <<<<<<<<<<<<<<
 *                             return self.__parse_pileup(contig=None, retargeted=False, skip_read_end_check=True)
 *                         except:
 */
              __Pyx_INCREF(Py_False);
              __Pyx_GIVEREF(Py_False);
              __Pyx_GOTREF(__pyx_v_self->is_complex_input);
              __Pyx_DECREF(__pyx_v_self->is_complex_input);
              __pyx_v_self->is_complex_input = Py_False;

              /* "aldy/indelpost/varaln.pyx":360
 *                             self.target = self.second_target
 *                             self.is_complex_input = False
 *                             return self.__parse_pileup(contig=None, retargeted=False, skip_read_end_check=True)             # <<<<<<<<<<<<<<
 *                         except:
 *                              pileup, contig
 */
              __Pyx_XDECREF(__pyx_r);
              __pyx_t_18.__pyx_n = 3;
              __pyx_t_18.contig = ((struct __pyx_obj_4aldy_9indelpost_6contig_Contig *)Py_None);
              __pyx_t_18.retargeted = 0;
              __pyx_t_18.skip_read_end_check = 1;
              __pyx_t_9 = ((struct __pyx_vtabstruct_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self->__pyx_vtab)->__pyx___parse_pileup(__pyx_v_self, &__pyx_t_18); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 360, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_r = __pyx_t_9;
              __pyx_t_9 = 0;
              goto __pyx_L25_try_return;

              /* "aldy/indelpost/varaln.pyx":356
 *                 else:
 *                     if self.is_complex_input:
 *                         try:             # <<<<<<<<<<<<<<
 *                             self.__target = self.second_target
 *                             self.target = self.second_target
 */
            }
            __pyx_L21_error:;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

            /* "aldy/indelpost/varaln.pyx":361
 *                             self.is_complex_input = False
 *                             return self.__parse_pileup(contig=None, retargeted=False, skip_read_end_check=True)
 *                         except:             # <<<<<<<<<<<<<<
 *                              pileup, contig
 *                     else:
 */
            /*except:*/ {
              __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.__parse_pileup", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_10, &__pyx_t_4) < 0) __PYX_ERR(0, 361, __pyx_L23_except_error)
              __Pyx_GOTREF(__pyx_t_9);
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_GOTREF(__pyx_t_4);

              /* "aldy/indelpost/varaln.pyx":362
 *                             return self.__parse_pileup(contig=None, retargeted=False, skip_read_end_check=True)
 *                         except:
 *                              pileup, contig             # <<<<<<<<<<<<<<
 *                     else:
 *                         return pileup, contig
 */
              __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 362, __pyx_L23_except_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_INCREF(__pyx_v_pileup);
              __Pyx_GIVEREF(__pyx_v_pileup);
              PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_pileup);
              __Pyx_INCREF(((PyObject *)__pyx_v_contig));
              __Pyx_GIVEREF(((PyObject *)__pyx_v_contig));
              PyTuple_SET_ITEM(__pyx_t_5, 1, ((PyObject *)__pyx_v_contig));
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              goto __pyx_L22_exception_handled;
            }
            __pyx_L23_except_error:;

            /* "aldy/indelpost/varaln.pyx":356
 *                 else:
 *                     if self.is_complex_input:
 *                         try:             # <<<<<<<<<<<<<<
 *                             self.__target = self.second_target
 *                             self.target = self.second_target
 */
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            goto __pyx_L1_error;
            __pyx_L25_try_return:;
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            goto __pyx_L0;
            __pyx_L22_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
          }

          /* "aldy/indelpost/varaln.pyx":355
 *                 # no target in this pileup
 *                 else:
 *                     if self.is_complex_input:             # <<<<<<<<<<<<<<
 *                         try:
 *                             self.__target = self.second_target
 */
          goto __pyx_L20;
        }

        /* "aldy/indelpost/varaln.pyx":364
 *                              pileup, contig
 *                     else:
 *                         return pileup, contig             # <<<<<<<<<<<<<<
 * 
 *         #_target = [read for read in pileup if read["is_target"]]
 */
        /*else*/ {
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 364, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_INCREF(__pyx_v_pileup);
          __Pyx_GIVEREF(__pyx_v_pileup);
          PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_pileup);
          __Pyx_INCREF(((PyObject *)__pyx_v_contig));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_contig));
          PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_contig));
          __pyx_r = __pyx_t_4;
          __pyx_t_4 = 0;
          goto __pyx_L0;
        }
        __pyx_L20:;
      }

      /* "aldy/indelpost/varaln.pyx":251
 * 
 *             self.is_spurious_overhang = False
 *             if contig.failed and not self.no_realignment:             # <<<<<<<<<<<<<<
 *                 within = self.retarget_window
 * 
 */
    }
  }
  __pyx_L3:;

  /* "aldy/indelpost/varaln.pyx":376
 * 
 *         # soft-clip realn & SW realn
 *         if contig.qc_passed and not self.no_realignment:             # <<<<<<<<<<<<<<
 * 
 *             orig_contig = contig
 */
  __pyx_t_1 = (__pyx_v_contig->qc_passed != 0);
  if (__pyx_t_1) {
  } else {
    __pyx_t_15 = __pyx_t_1;
    goto __pyx_L30_bool_binop_done;
  }
  __pyx_t_1 = ((!(__pyx_v_self->no_realignment != 0)) != 0);
  __pyx_t_15 = __pyx_t_1;
  __pyx_L30_bool_binop_done:;
  if (__pyx_t_15) {

    /* "aldy/indelpost/varaln.pyx":378
 *         if contig.qc_passed and not self.no_realignment:
 * 
 *             orig_contig = contig             # <<<<<<<<<<<<<<
 * 
 *             # realign reads that are not through retarget path
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_contig));
    __pyx_v_orig_contig = __pyx_v_contig;

    /* "aldy/indelpost/varaln.pyx":381
 * 
 *             # realign reads that are not through retarget path
 *             if not retargeted:             # <<<<<<<<<<<<<<
 *                cutoff = 1.0
 *                within = 30
 */
    __pyx_t_15 = ((!(__pyx_v_retargeted != 0)) != 0);
    if (__pyx_t_15) {

      /* "aldy/indelpost/varaln.pyx":382
 *             # realign reads that are not through retarget path
 *             if not retargeted:
 *                cutoff = 1.0             # <<<<<<<<<<<<<<
 *                within = 30
 * 
 */
      __pyx_v_cutoff = 1.0;

      /* "aldy/indelpost/varaln.pyx":383
 *             if not retargeted:
 *                cutoff = 1.0
 *                within = 30             # <<<<<<<<<<<<<<
 * 
 *                target = [read for read in pileup if read["is_target"]]
 */
      __pyx_v_within = 30;

      /* "aldy/indelpost/varaln.pyx":385
 *                within = 30
 * 
 *                target = [read for read in pileup if read["is_target"]]             # <<<<<<<<<<<<<<
 *                nontarget = [read for read in pileup if not read["is_target"]]
 * 
 */
      { /* enter inner scope */
        __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 385, __pyx_L35_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(__pyx_v_pileup == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
          __PYX_ERR(0, 385, __pyx_L35_error)
        }
        __pyx_t_10 = __pyx_v_pileup; __Pyx_INCREF(__pyx_t_10); __pyx_t_22 = 0;
        for (;;) {
          if (__pyx_t_22 >= PyList_GET_SIZE(__pyx_t_10)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_9 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_22); __Pyx_INCREF(__pyx_t_9); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 385, __pyx_L35_error)
          #else
          __pyx_t_9 = PySequence_ITEM(__pyx_t_10, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 385, __pyx_L35_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
          __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_read, __pyx_t_9);
          __pyx_t_9 = 0;
          __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_8genexpr1__pyx_v_read, __pyx_n_u_is_target); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 385, __pyx_L35_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 385, __pyx_L35_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (__pyx_t_15) {
            if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_8genexpr1__pyx_v_read))) __PYX_ERR(0, 385, __pyx_L35_error)
          }
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_read); __pyx_8genexpr1__pyx_v_read = 0;
        goto __pyx_L39_exit_scope;
        __pyx_L35_error:;
        __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_read); __pyx_8genexpr1__pyx_v_read = 0;
        goto __pyx_L1_error;
        __pyx_L39_exit_scope:;
      } /* exit inner scope */
      __pyx_v_target = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "aldy/indelpost/varaln.pyx":386
 * 
 *                target = [read for read in pileup if read["is_target"]]
 *                nontarget = [read for read in pileup if not read["is_target"]]             # <<<<<<<<<<<<<<
 * 
 *                grid = generate_grid(self.auto_adjust_extension_penalty,
 */
      { /* enter inner scope */
        __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 386, __pyx_L42_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(__pyx_v_pileup == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
          __PYX_ERR(0, 386, __pyx_L42_error)
        }
        __pyx_t_10 = __pyx_v_pileup; __Pyx_INCREF(__pyx_t_10); __pyx_t_22 = 0;
        for (;;) {
          if (__pyx_t_22 >= PyList_GET_SIZE(__pyx_t_10)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_9 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_22); __Pyx_INCREF(__pyx_t_9); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 386, __pyx_L42_error)
          #else
          __pyx_t_9 = PySequence_ITEM(__pyx_t_10, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 386, __pyx_L42_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
          __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_read, __pyx_t_9);
          __pyx_t_9 = 0;
          __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_8genexpr2__pyx_v_read, __pyx_n_u_is_target); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 386, __pyx_L42_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 386, __pyx_L42_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_1 = ((!__pyx_t_15) != 0);
          if (__pyx_t_1) {
            if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_8genexpr2__pyx_v_read))) __PYX_ERR(0, 386, __pyx_L42_error)
          }
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_read); __pyx_8genexpr2__pyx_v_read = 0;
        goto __pyx_L46_exit_scope;
        __pyx_L42_error:;
        __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_read); __pyx_8genexpr2__pyx_v_read = 0;
        goto __pyx_L1_error;
        __pyx_L46_exit_scope:;
      } /* exit inner scope */
      __pyx_v_nontarget = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "aldy/indelpost/varaln.pyx":388
 *                nontarget = [read for read in pileup if not read["is_target"]]
 * 
 *                grid = generate_grid(self.auto_adjust_extension_penalty,             # <<<<<<<<<<<<<<
 *                                     self.gap_open_penalty,
 *                                     self.gap_extension_penalty,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_generate_grid); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = __Pyx_PyBool_FromLong(__pyx_v_self->auto_adjust_extension_penalty); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);

      /* "aldy/indelpost/varaln.pyx":389
 * 
 *                grid = generate_grid(self.auto_adjust_extension_penalty,
 *                                     self.gap_open_penalty,             # <<<<<<<<<<<<<<
 *                                     self.gap_extension_penalty,
 *                                     self.__target,
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->gap_open_penalty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "aldy/indelpost/varaln.pyx":390
 *                grid = generate_grid(self.auto_adjust_extension_penalty,
 *                                     self.gap_open_penalty,
 *                                     self.gap_extension_penalty,             # <<<<<<<<<<<<<<
 *                                     self.__target,
 *                       )
 */
      __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_self->gap_extension_penalty); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 390, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);

      /* "aldy/indelpost/varaln.pyx":391
 *                                     self.gap_open_penalty,
 *                                     self.gap_extension_penalty,
 *                                     self.__target,             # <<<<<<<<<<<<<<
 *                       )
 * 
 */
      __pyx_t_3 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_12 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_t_9, __pyx_t_5, __pyx_t_13, ((PyObject *)__pyx_v_self->__pyx___target)};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 388, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_t_9, __pyx_t_5, __pyx_t_13, ((PyObject *)__pyx_v_self->__pyx___target)};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 388, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(4+__pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 388, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_12, __pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_12, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_12, __pyx_t_13);
        __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
        PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_12, ((PyObject *)__pyx_v_self->__pyx___target));
        __pyx_t_9 = 0;
        __pyx_t_5 = 0;
        __pyx_t_13 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_11, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 388, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF_SET(__pyx_v_grid, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "aldy/indelpost/varaln.pyx":394
 *                       )
 * 
 *                res = grid_search(             # <<<<<<<<<<<<<<
 *                        self.__target,
 *                        nontarget,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_grid_search); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);

      /* "aldy/indelpost/varaln.pyx":397
 *                        self.__target,
 *                        nontarget,
 *                        self.window,             # <<<<<<<<<<<<<<
 *                        self.mapqthresh,
 *                        within,
 */
      __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_self->window); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 397, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);

      /* "aldy/indelpost/varaln.pyx":398
 *                        nontarget,
 *                        self.window,
 *                        self.mapqthresh,             # <<<<<<<<<<<<<<
 *                        within,
 *                        cutoff,
 */
      __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_self->mapqthresh); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);

      /* "aldy/indelpost/varaln.pyx":399
 *                        self.window,
 *                        self.mapqthresh,
 *                        within,             # <<<<<<<<<<<<<<
 *                        cutoff,
 *                        self.match_score,
 */
      __pyx_t_5 = __Pyx_PyInt_From_long(__pyx_v_within); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 399, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "aldy/indelpost/varaln.pyx":400
 *                        self.mapqthresh,
 *                        within,
 *                        cutoff,             # <<<<<<<<<<<<<<
 *                        self.match_score,
 *                        self.mismatch_penalty,
 */
      __pyx_t_9 = PyFloat_FromDouble(__pyx_v_cutoff); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 400, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);

      /* "aldy/indelpost/varaln.pyx":401
 *                        within,
 *                        cutoff,
 *                        self.match_score,             # <<<<<<<<<<<<<<
 *                        self.mismatch_penalty,
 *                        grid,
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->match_score); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 401, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "aldy/indelpost/varaln.pyx":402
 *                        cutoff,
 *                        self.match_score,
 *                        self.mismatch_penalty,             # <<<<<<<<<<<<<<
 *                        grid,
 *                        self.unspliced_local_reference,
 */
      __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_self->mismatch_penalty); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 402, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);

      /* "aldy/indelpost/varaln.pyx":405
 *                        grid,
 *                        self.unspliced_local_reference,
 *                        self.exact_match_for_shiftable,             # <<<<<<<<<<<<<<
 *                    )
 * 
 */
      __pyx_t_8 = __Pyx_PyBool_FromLong(__pyx_v_self->exact_match_for_shiftable); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 405, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_12 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[12] = {__pyx_t_7, ((PyObject *)__pyx_v_self->__pyx___target), __pyx_v_nontarget, __pyx_t_11, __pyx_t_13, __pyx_t_5, __pyx_t_9, __pyx_t_3, __pyx_t_16, __pyx_v_grid, ((PyObject *)__pyx_v_self->unspliced_local_reference), __pyx_t_8};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_12, 11+__pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 394, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[12] = {__pyx_t_7, ((PyObject *)__pyx_v_self->__pyx___target), __pyx_v_nontarget, __pyx_t_11, __pyx_t_13, __pyx_t_5, __pyx_t_9, __pyx_t_3, __pyx_t_16, __pyx_v_grid, ((PyObject *)__pyx_v_self->unspliced_local_reference), __pyx_t_8};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_12, 11+__pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 394, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else
      #endif
      {
        __pyx_t_2 = PyTuple_New(11+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
        PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_12, ((PyObject *)__pyx_v_self->__pyx___target));
        __Pyx_INCREF(__pyx_v_nontarget);
        __Pyx_GIVEREF(__pyx_v_nontarget);
        PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_12, __pyx_v_nontarget);
        __Pyx_GIVEREF(__pyx_t_11);
        PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_12, __pyx_t_11);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_2, 3+__pyx_t_12, __pyx_t_13);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_2, 4+__pyx_t_12, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_2, 5+__pyx_t_12, __pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_2, 6+__pyx_t_12, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_16);
        PyTuple_SET_ITEM(__pyx_t_2, 7+__pyx_t_12, __pyx_t_16);
        __Pyx_INCREF(__pyx_v_grid);
        __Pyx_GIVEREF(__pyx_v_grid);
        PyTuple_SET_ITEM(__pyx_t_2, 8+__pyx_t_12, __pyx_v_grid);
        __Pyx_INCREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
        PyTuple_SET_ITEM(__pyx_t_2, 9+__pyx_t_12, ((PyObject *)__pyx_v_self->unspliced_local_reference));
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_2, 10+__pyx_t_12, __pyx_t_8);
        __pyx_t_11 = 0;
        __pyx_t_13 = 0;
        __pyx_t_5 = 0;
        __pyx_t_9 = 0;
        __pyx_t_3 = 0;
        __pyx_t_16 = 0;
        __pyx_t_8 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_2, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF_SET(__pyx_v_res, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "aldy/indelpost/varaln.pyx":408
 *                    )
 * 
 *                if res:             # <<<<<<<<<<<<<<
 *                    nontarget = [read for read in nontarget if read not in res[1]]
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_res); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 408, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "aldy/indelpost/varaln.pyx":409
 * 
 *                if res:
 *                    nontarget = [read for read in nontarget if read not in res[1]]             # <<<<<<<<<<<<<<
 * 
 *                    pileup = target + res[1] + nontarget
 */
        { /* enter inner scope */
          __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 409, __pyx_L50_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_10 = __pyx_v_nontarget; __Pyx_INCREF(__pyx_t_10); __pyx_t_22 = 0;
          for (;;) {
            if (__pyx_t_22 >= PyList_GET_SIZE(__pyx_t_10)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_22); __Pyx_INCREF(__pyx_t_2); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 409, __pyx_L50_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_10, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 409, __pyx_L50_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
            __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_read, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_res, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 409, __pyx_L50_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_8genexpr3__pyx_v_read, __pyx_t_2, Py_NE)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 409, __pyx_L50_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_15 = (__pyx_t_1 != 0);
            if (__pyx_t_15) {
              if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_8genexpr3__pyx_v_read))) __PYX_ERR(0, 409, __pyx_L50_error)
            }
          }
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_read); __pyx_8genexpr3__pyx_v_read = 0;
          goto __pyx_L54_exit_scope;
          __pyx_L50_error:;
          __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_read); __pyx_8genexpr3__pyx_v_read = 0;
          goto __pyx_L1_error;
          __pyx_L54_exit_scope:;
        } /* exit inner scope */
        __Pyx_DECREF_SET(__pyx_v_nontarget, ((PyObject*)__pyx_t_4));
        __pyx_t_4 = 0;

        /* "aldy/indelpost/varaln.pyx":411
 *                    nontarget = [read for read in nontarget if read not in res[1]]
 * 
 *                    pileup = target + res[1] + nontarget             # <<<<<<<<<<<<<<
 *                    self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]
 * 
 */
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_res, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 411, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = PyNumber_Add(__pyx_v_target, __pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 411, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Add(__pyx_t_10, __pyx_v_nontarget); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 411, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 411, __pyx_L1_error)
        __Pyx_DECREF_SET(__pyx_v_pileup, ((PyObject*)__pyx_t_4));
        __pyx_t_4 = 0;

        /* "aldy/indelpost/varaln.pyx":412
 * 
 *                    pileup = target + res[1] + nontarget
 *                    self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]             # <<<<<<<<<<<<<<
 * 
 *                    self.__target, pileup = update_pileup(
 */
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_res, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 412, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_res, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 412, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_17 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_17 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v_self->gap_open_penalty = __pyx_t_12;
        __pyx_v_self->gap_extension_penalty = __pyx_t_17;

        /* "aldy/indelpost/varaln.pyx":414
 *                    self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]
 * 
 *                    self.__target, pileup = update_pileup(             # <<<<<<<<<<<<<<
 *                        pileup,
 *                        self.__target,
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_update_pileup); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 414, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);

        /* "aldy/indelpost/varaln.pyx":417
 *                        pileup,
 *                        self.__target,
 *                        self.window,             # <<<<<<<<<<<<<<
 *                        self.match_score,
 *                        self.mismatch_penalty,
 */
        __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_self->window); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 417, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);

        /* "aldy/indelpost/varaln.pyx":418
 *                        self.__target,
 *                        self.window,
 *                        self.match_score,             # <<<<<<<<<<<<<<
 *                        self.mismatch_penalty,
 *                        self.gap_open_penalty,
 */
        __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->match_score); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 418, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);

        /* "aldy/indelpost/varaln.pyx":419
 *                        self.window,
 *                        self.match_score,
 *                        self.mismatch_penalty,             # <<<<<<<<<<<<<<
 *                        self.gap_open_penalty,
 *                        self.gap_extension_penalty,
 */
        __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_self->mismatch_penalty); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 419, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);

        /* "aldy/indelpost/varaln.pyx":420
 *                        self.match_score,
 *                        self.mismatch_penalty,
 *                        self.gap_open_penalty,             # <<<<<<<<<<<<<<
 *                        self.gap_extension_penalty,
 *                        self.basequalthresh,
 */
        __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_self->gap_open_penalty); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 420, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);

        /* "aldy/indelpost/varaln.pyx":421
 *                        self.mismatch_penalty,
 *                        self.gap_open_penalty,
 *                        self.gap_extension_penalty,             # <<<<<<<<<<<<<<
 *                        self.basequalthresh,
 *                        bypass_search=True,
 */
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->gap_extension_penalty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 421, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);

        /* "aldy/indelpost/varaln.pyx":422
 *                        self.gap_open_penalty,
 *                        self.gap_extension_penalty,
 *                        self.basequalthresh,             # <<<<<<<<<<<<<<
 *                        bypass_search=True,
 *                    )
 */
        __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_self->basequalthresh); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 422, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);

        /* "aldy/indelpost/varaln.pyx":414
 *                    self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]
 * 
 *                    self.__target, pileup = update_pileup(             # <<<<<<<<<<<<<<
 *                        pileup,
 *                        self.__target,
 */
        __pyx_t_5 = PyTuple_New(8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 414, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_v_pileup);
        __Pyx_GIVEREF(__pyx_v_pileup);
        PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_pileup);
        __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
        PyTuple_SET_ITEM(__pyx_t_5, 1, ((PyObject *)__pyx_v_self->__pyx___target));
        __Pyx_GIVEREF(__pyx_t_10);
        PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_16);
        PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_t_16);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_5, 6, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_5, 7, __pyx_t_9);
        __pyx_t_10 = 0;
        __pyx_t_2 = 0;
        __pyx_t_8 = 0;
        __pyx_t_16 = 0;
        __pyx_t_3 = 0;
        __pyx_t_9 = 0;

        /* "aldy/indelpost/varaln.pyx":423
 *                        self.gap_extension_penalty,
 *                        self.basequalthresh,
 *                        bypass_search=True,             # <<<<<<<<<<<<<<
 *                    )
 * 
 */
        __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 423, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_bypass_search, Py_True) < 0) __PYX_ERR(0, 423, __pyx_L1_error)

        /* "aldy/indelpost/varaln.pyx":414
 *                    self.gap_open_penalty, self.gap_extension_penalty = res[2], res[3]
 * 
 *                    self.__target, pileup = update_pileup(             # <<<<<<<<<<<<<<
 *                        pileup,
 *                        self.__target,
 */
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, __pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
          PyObject* sequence = __pyx_t_3;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 414, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_5);
          #else
          __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 414, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 414, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 414, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_14 = Py_TYPE(__pyx_t_4)->tp_iternext;
          index = 0; __pyx_t_9 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_9)) goto __pyx_L55_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_9);
          index = 1; __pyx_t_5 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L55_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_5);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_4), 2) < 0) __PYX_ERR(0, 414, __pyx_L1_error)
          __pyx_t_14 = NULL;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          goto __pyx_L56_unpacking_done;
          __pyx_L55_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_14 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 414, __pyx_L1_error)
          __pyx_L56_unpacking_done:;
        }
        if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_4aldy_9indelpost_7variant_Variant))))) __PYX_ERR(0, 414, __pyx_L1_error)
        if (!(likely(PyList_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 414, __pyx_L1_error)
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GOTREF(__pyx_v_self->__pyx___target);
        __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx___target));
        __pyx_v_self->__pyx___target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_9);
        __pyx_t_9 = 0;
        __Pyx_DECREF_SET(__pyx_v_pileup, ((PyObject*)__pyx_t_5));
        __pyx_t_5 = 0;

        /* "aldy/indelpost/varaln.pyx":427
 * 
 *                    #equivalent but different position
 *                    if self.__target == res[0]:             # <<<<<<<<<<<<<<
 *                        self.__target = res[0]
 * 
 */
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_res, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 427, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = PyObject_RichCompare(((PyObject *)__pyx_v_self->__pyx___target), __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 427, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 427, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (__pyx_t_15) {

          /* "aldy/indelpost/varaln.pyx":428
 *                    #equivalent but different position
 *                    if self.__target == res[0]:
 *                        self.__target = res[0]             # <<<<<<<<<<<<<<
 * 
 *                else:
 */
          __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_res, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 428, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_4aldy_9indelpost_7variant_Variant))))) __PYX_ERR(0, 428, __pyx_L1_error)
          __Pyx_GIVEREF(__pyx_t_5);
          __Pyx_GOTREF(__pyx_v_self->__pyx___target);
          __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx___target));
          __pyx_v_self->__pyx___target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_5);
          __pyx_t_5 = 0;

          /* "aldy/indelpost/varaln.pyx":427
 * 
 *                    #equivalent but different position
 *                    if self.__target == res[0]:             # <<<<<<<<<<<<<<
 *                        self.__target = res[0]
 * 
 */
        }

        /* "aldy/indelpost/varaln.pyx":408
 *                    )
 * 
 *                if res:             # <<<<<<<<<<<<<<
 *                    nontarget = [read for read in nontarget if read not in res[1]]
 * 
 */
        goto __pyx_L47;
      }

      /* "aldy/indelpost/varaln.pyx":431
 * 
 *                else:
 *                   pileup = target + nontarget             # <<<<<<<<<<<<<<
 * 
 *             if self.__target.count_repeats() == 0:
 */
      /*else*/ {
        __pyx_t_5 = PyNumber_Add(__pyx_v_target, __pyx_v_nontarget); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 431, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF_SET(__pyx_v_pileup, ((PyObject*)__pyx_t_5));
        __pyx_t_5 = 0;
      }
      __pyx_L47:;

      /* "aldy/indelpost/varaln.pyx":381
 * 
 *             # realign reads that are not through retarget path
 *             if not retargeted:             # <<<<<<<<<<<<<<
 *                cutoff = 1.0
 *                within = 30
 */
    }

    /* "aldy/indelpost/varaln.pyx":433
 *                   pileup = target + nontarget
 * 
 *             if self.__target.count_repeats() == 0:             # <<<<<<<<<<<<<<
 *                 pileup = find_by_softclip_split(self.__target, contig, pileup)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx___target), __pyx_n_s_count_repeats); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_5, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_15) {

      /* "aldy/indelpost/varaln.pyx":434
 * 
 *             if self.__target.count_repeats() == 0:
 *                 pileup = find_by_softclip_split(self.__target, contig, pileup)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_find_by_softclip_split); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 434, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = NULL;
      __pyx_t_17 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_17 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[4] = {__pyx_t_9, ((PyObject *)__pyx_v_self->__pyx___target), ((PyObject *)__pyx_v_contig), __pyx_v_pileup};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[4] = {__pyx_t_9, ((PyObject *)__pyx_v_self->__pyx___target), ((PyObject *)__pyx_v_contig), __pyx_v_pileup};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_4 = PyTuple_New(3+__pyx_t_17); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 434, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (__pyx_t_9) {
          __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_9); __pyx_t_9 = NULL;
        }
        __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
        PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_17, ((PyObject *)__pyx_v_self->__pyx___target));
        __Pyx_INCREF(((PyObject *)__pyx_v_contig));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_contig));
        PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_17, ((PyObject *)__pyx_v_contig));
        __Pyx_INCREF(__pyx_v_pileup);
        __Pyx_GIVEREF(__pyx_v_pileup);
        PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_17, __pyx_v_pileup);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (!(likely(PyList_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 434, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_pileup, ((PyObject*)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "aldy/indelpost/varaln.pyx":433
 *                   pileup = target + nontarget
 * 
 *             if self.__target.count_repeats() == 0:             # <<<<<<<<<<<<<<
 *                 pileup = find_by_softclip_split(self.__target, contig, pileup)
 * 
 */
    }

    /* "aldy/indelpost/varaln.pyx":437
 * 
 * 
 *             if read_end_evidence_only:             # <<<<<<<<<<<<<<
 *                 target_pileup = [read for read in pileup if read["is_target"]]
 * 
 */
    __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_v_read_end_evidence_only); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 437, __pyx_L1_error)
    if (__pyx_t_15) {

      /* "aldy/indelpost/varaln.pyx":438
 * 
 *             if read_end_evidence_only:
 *                 target_pileup = [read for read in pileup if read["is_target"]]             # <<<<<<<<<<<<<<
 * 
 *             pileup = find_by_smith_waterman_realn(
 */
      { /* enter inner scope */
        __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 438, __pyx_L62_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__pyx_v_pileup == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
          __PYX_ERR(0, 438, __pyx_L62_error)
        }
        __pyx_t_5 = __pyx_v_pileup; __Pyx_INCREF(__pyx_t_5); __pyx_t_22 = 0;
        for (;;) {
          if (__pyx_t_22 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_22); __Pyx_INCREF(__pyx_t_4); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 438, __pyx_L62_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 438, __pyx_L62_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
          __Pyx_XDECREF_SET(__pyx_8genexpr4__pyx_v_read, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_8genexpr4__pyx_v_read, __pyx_n_u_is_target); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 438, __pyx_L62_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 438, __pyx_L62_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (__pyx_t_15) {
            if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_8genexpr4__pyx_v_read))) __PYX_ERR(0, 438, __pyx_L62_error)
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_read); __pyx_8genexpr4__pyx_v_read = 0;
        goto __pyx_L66_exit_scope;
        __pyx_L62_error:;
        __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_read); __pyx_8genexpr4__pyx_v_read = 0;
        goto __pyx_L1_error;
        __pyx_L66_exit_scope:;
      } /* exit inner scope */
      __pyx_v_target_pileup = ((PyObject*)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "aldy/indelpost/varaln.pyx":437
 * 
 * 
 *             if read_end_evidence_only:             # <<<<<<<<<<<<<<
 *                 target_pileup = [read for read in pileup if read["is_target"]]
 * 
 */
    }

    /* "aldy/indelpost/varaln.pyx":440
 *                 target_pileup = [read for read in pileup if read["is_target"]]
 * 
 *             pileup = find_by_smith_waterman_realn(             # <<<<<<<<<<<<<<
 *                 self.__target,
 *                 contig,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_find_by_smith_waterman_realn); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "aldy/indelpost/varaln.pyx":444
 *                 contig,
 *                 pileup,
 *                 self.match_score,             # <<<<<<<<<<<<<<
 *                 self.mismatch_penalty,
 *                 self.gap_open_penalty,
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->match_score); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "aldy/indelpost/varaln.pyx":445
 *                 pileup,
 *                 self.match_score,
 *                 self.mismatch_penalty,             # <<<<<<<<<<<<<<
 *                 self.gap_open_penalty,
 *                 self.gap_extension_penalty,
 */
    __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_self->mismatch_penalty); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);

    /* "aldy/indelpost/varaln.pyx":446
 *                 self.match_score,
 *                 self.mismatch_penalty,
 *                 self.gap_open_penalty,             # <<<<<<<<<<<<<<
 *                 self.gap_extension_penalty,
 *                 self.basequalthresh
 */
    __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_self->gap_open_penalty); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);

    /* "aldy/indelpost/varaln.pyx":447
 *                 self.mismatch_penalty,
 *                 self.gap_open_penalty,
 *                 self.gap_extension_penalty,             # <<<<<<<<<<<<<<
 *                 self.basequalthresh
 *             )
 */
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_self->gap_extension_penalty); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "aldy/indelpost/varaln.pyx":448
 *                 self.gap_open_penalty,
 *                 self.gap_extension_penalty,
 *                 self.basequalthresh             # <<<<<<<<<<<<<<
 *             )
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->basequalthresh); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 448, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_10 = NULL;
    __pyx_t_17 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_17 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[9] = {__pyx_t_10, ((PyObject *)__pyx_v_self->__pyx___target), ((PyObject *)__pyx_v_contig), __pyx_v_pileup, __pyx_t_4, __pyx_t_9, __pyx_t_16, __pyx_t_8, __pyx_t_2};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_17, 8+__pyx_t_17); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[9] = {__pyx_t_10, ((PyObject *)__pyx_v_self->__pyx___target), ((PyObject *)__pyx_v_contig), __pyx_v_pileup, __pyx_t_4, __pyx_t_9, __pyx_t_16, __pyx_t_8, __pyx_t_2};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_17, 8+__pyx_t_17); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_13 = PyTuple_New(8+__pyx_t_17); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
      PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_17, ((PyObject *)__pyx_v_self->__pyx___target));
      __Pyx_INCREF(((PyObject *)__pyx_v_contig));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_contig));
      PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_17, ((PyObject *)__pyx_v_contig));
      __Pyx_INCREF(__pyx_v_pileup);
      __Pyx_GIVEREF(__pyx_v_pileup);
      PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_17, __pyx_v_pileup);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_13, 3+__pyx_t_17, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_13, 4+__pyx_t_17, __pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_16);
      PyTuple_SET_ITEM(__pyx_t_13, 5+__pyx_t_17, __pyx_t_16);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_13, 6+__pyx_t_17, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_13, 7+__pyx_t_17, __pyx_t_2);
      __pyx_t_4 = 0;
      __pyx_t_9 = 0;
      __pyx_t_16 = 0;
      __pyx_t_8 = 0;
      __pyx_t_2 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "aldy/indelpost/varaln.pyx":440
 *                 target_pileup = [read for read in pileup if read["is_target"]]
 * 
 *             pileup = find_by_smith_waterman_realn(             # <<<<<<<<<<<<<<
 *                 self.__target,
 *                 contig,
 */
    if (!(likely(PyList_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_pileup, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":451
 *             )
 * 
 *             if read_end_evidence_only:             # <<<<<<<<<<<<<<
 *                 newly_identified = [read for read in pileup if read["is_target"] and not read in target_pileup]
 *                 if newly_identified:
 */
    __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_v_read_end_evidence_only); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 451, __pyx_L1_error)
    if (__pyx_t_15) {

      /* "aldy/indelpost/varaln.pyx":452
 * 
 *             if read_end_evidence_only:
 *                 newly_identified = [read for read in pileup if read["is_target"] and not read in target_pileup]             # <<<<<<<<<<<<<<
 *                 if newly_identified:
 *                     expected_ptrn = most_common_gap_ptrn(newly_identified)
 */
      { /* enter inner scope */
        __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 452, __pyx_L70_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__pyx_v_pileup == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
          __PYX_ERR(0, 452, __pyx_L70_error)
        }
        __pyx_t_5 = __pyx_v_pileup; __Pyx_INCREF(__pyx_t_5); __pyx_t_22 = 0;
        for (;;) {
          if (__pyx_t_22 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_13 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_22); __Pyx_INCREF(__pyx_t_13); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 452, __pyx_L70_error)
          #else
          __pyx_t_13 = PySequence_ITEM(__pyx_t_5, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 452, __pyx_L70_error)
          __Pyx_GOTREF(__pyx_t_13);
          #endif
          __Pyx_XDECREF_SET(__pyx_8genexpr5__pyx_v_read, __pyx_t_13);
          __pyx_t_13 = 0;
          __pyx_t_13 = __Pyx_PyObject_Dict_GetItem(__pyx_8genexpr5__pyx_v_read, __pyx_n_u_is_target); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 452, __pyx_L70_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 452, __pyx_L70_error)
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (__pyx_t_1) {
          } else {
            __pyx_t_15 = __pyx_t_1;
            goto __pyx_L74_bool_binop_done;
          }
          if (unlikely(!__pyx_v_target_pileup)) { __Pyx_RaiseUnboundLocalError("target_pileup"); __PYX_ERR(0, 452, __pyx_L70_error) }
          __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_8genexpr5__pyx_v_read, __pyx_v_target_pileup, Py_NE)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 452, __pyx_L70_error)
          __pyx_t_23 = (__pyx_t_1 != 0);
          __pyx_t_15 = __pyx_t_23;
          __pyx_L74_bool_binop_done:;
          if (__pyx_t_15) {
            if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_8genexpr5__pyx_v_read))) __PYX_ERR(0, 452, __pyx_L70_error)
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_read); __pyx_8genexpr5__pyx_v_read = 0;
        goto __pyx_L76_exit_scope;
        __pyx_L70_error:;
        __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_read); __pyx_8genexpr5__pyx_v_read = 0;
        goto __pyx_L1_error;
        __pyx_L76_exit_scope:;
      } /* exit inner scope */
      __pyx_v_newly_identified = ((PyObject*)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "aldy/indelpost/varaln.pyx":453
 *             if read_end_evidence_only:
 *                 newly_identified = [read for read in pileup if read["is_target"] and not read in target_pileup]
 *                 if newly_identified:             # <<<<<<<<<<<<<<
 *                     expected_ptrn = most_common_gap_ptrn(newly_identified)
 * 
 */
      __pyx_t_15 = (PyList_GET_SIZE(__pyx_v_newly_identified) != 0);
      if (__pyx_t_15) {

        /* "aldy/indelpost/varaln.pyx":454
 *                 newly_identified = [read for read in pileup if read["is_target"] and not read in target_pileup]
 *                 if newly_identified:
 *                     expected_ptrn = most_common_gap_ptrn(newly_identified)             # <<<<<<<<<<<<<<
 * 
 *                     indels = []
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_most_common_gap_ptrn); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 454, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_13 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_3 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_13, __pyx_v_newly_identified) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_newly_identified);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 454, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_expected_ptrn = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "aldy/indelpost/varaln.pyx":456
 *                     expected_ptrn = most_common_gap_ptrn(newly_identified)
 * 
 *                     indels = []             # <<<<<<<<<<<<<<
 *                     contig_seq = contig.get_contig_seq()
 *                     alinger = make_aligner(contig_seq, self.match_score, self.mismatch_penalty)
 */
        __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 456, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_v_indels = ((PyObject*)__pyx_t_3);
        __pyx_t_3 = 0;

        /* "aldy/indelpost/varaln.pyx":457
 * 
 *                     indels = []
 *                     contig_seq = contig.get_contig_seq()             # <<<<<<<<<<<<<<
 *                     alinger = make_aligner(contig_seq, self.match_score, self.mismatch_penalty)
 *                     for new_one in newly_identified:
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_contig), __pyx_n_s_get_contig_seq); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 457, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_13 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_3 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 457, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_contig_seq = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "aldy/indelpost/varaln.pyx":458
 *                     indels = []
 *                     contig_seq = contig.get_contig_seq()
 *                     alinger = make_aligner(contig_seq, self.match_score, self.mismatch_penalty)             # <<<<<<<<<<<<<<
 *                     for new_one in newly_identified:
 *                         if "N" not in new_one["cigar_string"] and is_perfect_match(alinger, contig_seq, new_one["read_seq"]):
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_make_aligner); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 458, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_self->match_score); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 458, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->mismatch_penalty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 458, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_8 = NULL;
        __pyx_t_17 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_17 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_v_contig_seq, __pyx_t_13, __pyx_t_2};
          __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 458, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_v_contig_seq, __pyx_t_13, __pyx_t_2};
          __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 458, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        } else
        #endif
        {
          __pyx_t_16 = PyTuple_New(3+__pyx_t_17); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 458, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (__pyx_t_8) {
            __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_8); __pyx_t_8 = NULL;
          }
          __Pyx_INCREF(__pyx_v_contig_seq);
          __Pyx_GIVEREF(__pyx_v_contig_seq);
          PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_17, __pyx_v_contig_seq);
          __Pyx_GIVEREF(__pyx_t_13);
          PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_17, __pyx_t_13);
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_17, __pyx_t_2);
          __pyx_t_13 = 0;
          __pyx_t_2 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_16, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 458, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_alinger = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "aldy/indelpost/varaln.pyx":459
 *                     contig_seq = contig.get_contig_seq()
 *                     alinger = make_aligner(contig_seq, self.match_score, self.mismatch_penalty)
 *                     for new_one in newly_identified:             # <<<<<<<<<<<<<<
 *                         if "N" not in new_one["cigar_string"] and is_perfect_match(alinger, contig_seq, new_one["read_seq"]):
 *                             indels += [i[-1] for i in new_one["I"]] + [d[-1] for d in new_one["D"]]
 */
        __pyx_t_3 = __pyx_v_newly_identified; __Pyx_INCREF(__pyx_t_3); __pyx_t_22 = 0;
        for (;;) {
          if (__pyx_t_22 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_22); __Pyx_INCREF(__pyx_t_5); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 459, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 459, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
          __Pyx_XDECREF_SET(__pyx_v_new_one, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "aldy/indelpost/varaln.pyx":460
 *                     alinger = make_aligner(contig_seq, self.match_score, self.mismatch_penalty)
 *                     for new_one in newly_identified:
 *                         if "N" not in new_one["cigar_string"] and is_perfect_match(alinger, contig_seq, new_one["read_seq"]):             # <<<<<<<<<<<<<<
 *                             indels += [i[-1] for i in new_one["I"]] + [d[-1] for d in new_one["D"]]
 * 
 */
          __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_new_one, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 460, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_23 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_N, __pyx_t_5, Py_NE)); if (unlikely(__pyx_t_23 < 0)) __PYX_ERR(0, 460, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_1 = (__pyx_t_23 != 0);
          if (__pyx_t_1) {
          } else {
            __pyx_t_15 = __pyx_t_1;
            goto __pyx_L81_bool_binop_done;
          }
          __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_is_perfect_match); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 460, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_new_one, __pyx_n_u_read_seq); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 460, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_13 = NULL;
          __pyx_t_17 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_16))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_16);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_16, function);
              __pyx_t_17 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_16)) {
            PyObject *__pyx_temp[4] = {__pyx_t_13, __pyx_v_alinger, __pyx_v_contig_seq, __pyx_t_2};
            __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_16, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 460, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_16)) {
            PyObject *__pyx_temp[4] = {__pyx_t_13, __pyx_v_alinger, __pyx_v_contig_seq, __pyx_t_2};
            __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_16, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 460, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_8 = PyTuple_New(3+__pyx_t_17); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 460, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            if (__pyx_t_13) {
              __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_13); __pyx_t_13 = NULL;
            }
            __Pyx_INCREF(__pyx_v_alinger);
            __Pyx_GIVEREF(__pyx_v_alinger);
            PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_17, __pyx_v_alinger);
            __Pyx_INCREF(__pyx_v_contig_seq);
            __Pyx_GIVEREF(__pyx_v_contig_seq);
            PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_17, __pyx_v_contig_seq);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_17, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_t_8, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 460, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 460, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_15 = __pyx_t_1;
          __pyx_L81_bool_binop_done:;
          if (__pyx_t_15) {

            /* "aldy/indelpost/varaln.pyx":461
 *                     for new_one in newly_identified:
 *                         if "N" not in new_one["cigar_string"] and is_perfect_match(alinger, contig_seq, new_one["read_seq"]):
 *                             indels += [i[-1] for i in new_one["I"]] + [d[-1] for d in new_one["D"]]             # <<<<<<<<<<<<<<
 * 
 *                     if indels:
 */
            { /* enter inner scope */
              __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 461, __pyx_L85_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_16 = __Pyx_PyObject_Dict_GetItem(__pyx_v_new_one, __pyx_n_u_I); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 461, __pyx_L85_error)
              __Pyx_GOTREF(__pyx_t_16);
              if (likely(PyList_CheckExact(__pyx_t_16)) || PyTuple_CheckExact(__pyx_t_16)) {
                __pyx_t_8 = __pyx_t_16; __Pyx_INCREF(__pyx_t_8); __pyx_t_24 = 0;
                __pyx_t_25 = NULL;
              } else {
                __pyx_t_24 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_16); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 461, __pyx_L85_error)
                __Pyx_GOTREF(__pyx_t_8);
                __pyx_t_25 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 461, __pyx_L85_error)
              }
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              for (;;) {
                if (likely(!__pyx_t_25)) {
                  if (likely(PyList_CheckExact(__pyx_t_8))) {
                    if (__pyx_t_24 >= PyList_GET_SIZE(__pyx_t_8)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_16 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_24); __Pyx_INCREF(__pyx_t_16); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 461, __pyx_L85_error)
                    #else
                    __pyx_t_16 = PySequence_ITEM(__pyx_t_8, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 461, __pyx_L85_error)
                    __Pyx_GOTREF(__pyx_t_16);
                    #endif
                  } else {
                    if (__pyx_t_24 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_16 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_24); __Pyx_INCREF(__pyx_t_16); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 461, __pyx_L85_error)
                    #else
                    __pyx_t_16 = PySequence_ITEM(__pyx_t_8, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 461, __pyx_L85_error)
                    __Pyx_GOTREF(__pyx_t_16);
                    #endif
                  }
                } else {
                  __pyx_t_16 = __pyx_t_25(__pyx_t_8);
                  if (unlikely(!__pyx_t_16)) {
                    PyObject* exc_type = PyErr_Occurred();
                    if (exc_type) {
                      if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                      else __PYX_ERR(0, 461, __pyx_L85_error)
                    }
                    break;
                  }
                  __Pyx_GOTREF(__pyx_t_16);
                }
                __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_i, __pyx_t_16);
                __pyx_t_16 = 0;
                __pyx_t_16 = __Pyx_GetItemInt(__pyx_8genexpr6__pyx_v_i, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 461, __pyx_L85_error)
                __Pyx_GOTREF(__pyx_t_16);
                if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_16))) __PYX_ERR(0, 461, __pyx_L85_error)
                __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              }
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_i); __pyx_8genexpr6__pyx_v_i = 0;
              goto __pyx_L88_exit_scope;
              __pyx_L85_error:;
              __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_i); __pyx_8genexpr6__pyx_v_i = 0;
              goto __pyx_L1_error;
              __pyx_L88_exit_scope:;
            } /* exit inner scope */
            { /* enter inner scope */
              __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 461, __pyx_L91_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_16 = __Pyx_PyObject_Dict_GetItem(__pyx_v_new_one, __pyx_n_u_D); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 461, __pyx_L91_error)
              __Pyx_GOTREF(__pyx_t_16);
              if (likely(PyList_CheckExact(__pyx_t_16)) || PyTuple_CheckExact(__pyx_t_16)) {
                __pyx_t_2 = __pyx_t_16; __Pyx_INCREF(__pyx_t_2); __pyx_t_24 = 0;
                __pyx_t_25 = NULL;
              } else {
                __pyx_t_24 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 461, __pyx_L91_error)
                __Pyx_GOTREF(__pyx_t_2);
                __pyx_t_25 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 461, __pyx_L91_error)
              }
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              for (;;) {
                if (likely(!__pyx_t_25)) {
                  if (likely(PyList_CheckExact(__pyx_t_2))) {
                    if (__pyx_t_24 >= PyList_GET_SIZE(__pyx_t_2)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_16 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_24); __Pyx_INCREF(__pyx_t_16); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 461, __pyx_L91_error)
                    #else
                    __pyx_t_16 = PySequence_ITEM(__pyx_t_2, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 461, __pyx_L91_error)
                    __Pyx_GOTREF(__pyx_t_16);
                    #endif
                  } else {
                    if (__pyx_t_24 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_16 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_24); __Pyx_INCREF(__pyx_t_16); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 461, __pyx_L91_error)
                    #else
                    __pyx_t_16 = PySequence_ITEM(__pyx_t_2, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 461, __pyx_L91_error)
                    __Pyx_GOTREF(__pyx_t_16);
                    #endif
                  }
                } else {
                  __pyx_t_16 = __pyx_t_25(__pyx_t_2);
                  if (unlikely(!__pyx_t_16)) {
                    PyObject* exc_type = PyErr_Occurred();
                    if (exc_type) {
                      if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                      else __PYX_ERR(0, 461, __pyx_L91_error)
                    }
                    break;
                  }
                  __Pyx_GOTREF(__pyx_t_16);
                }
                __Pyx_XDECREF_SET(__pyx_8genexpr7__pyx_v_d, __pyx_t_16);
                __pyx_t_16 = 0;
                __pyx_t_16 = __Pyx_GetItemInt(__pyx_8genexpr7__pyx_v_d, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 461, __pyx_L91_error)
                __Pyx_GOTREF(__pyx_t_16);
                if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_t_16))) __PYX_ERR(0, 461, __pyx_L91_error)
                __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              }
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_d); __pyx_8genexpr7__pyx_v_d = 0;
              goto __pyx_L94_exit_scope;
              __pyx_L91_error:;
              __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_d); __pyx_8genexpr7__pyx_v_d = 0;
              goto __pyx_L1_error;
              __pyx_L94_exit_scope:;
            } /* exit inner scope */
            __pyx_t_2 = PyNumber_Add(__pyx_t_5, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 461, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_indels, __pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 461, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF_SET(__pyx_v_indels, ((PyObject*)__pyx_t_8));
            __pyx_t_8 = 0;

            /* "aldy/indelpost/varaln.pyx":460
 *                     alinger = make_aligner(contig_seq, self.match_score, self.mismatch_penalty)
 *                     for new_one in newly_identified:
 *                         if "N" not in new_one["cigar_string"] and is_perfect_match(alinger, contig_seq, new_one["read_seq"]):             # <<<<<<<<<<<<<<
 *                             indels += [i[-1] for i in new_one["I"]] + [d[-1] for d in new_one["D"]]
 * 
 */
          }

          /* "aldy/indelpost/varaln.pyx":459
 *                     contig_seq = contig.get_contig_seq()
 *                     alinger = make_aligner(contig_seq, self.match_score, self.mismatch_penalty)
 *                     for new_one in newly_identified:             # <<<<<<<<<<<<<<
 *                         if "N" not in new_one["cigar_string"] and is_perfect_match(alinger, contig_seq, new_one["read_seq"]):
 *                             indels += [i[-1] for i in new_one["I"]] + [d[-1] for d in new_one["D"]]
 */
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "aldy/indelpost/varaln.pyx":463
 *                             indels += [i[-1] for i in new_one["I"]] + [d[-1] for d in new_one["D"]]
 * 
 *                     if indels:             # <<<<<<<<<<<<<<
 *                         try:
 *                             self.__target = most_common(indels)
 */
        __pyx_t_15 = (PyList_GET_SIZE(__pyx_v_indels) != 0);
        if (__pyx_t_15) {

          /* "aldy/indelpost/varaln.pyx":464
 * 
 *                     if indels:
 *                         try:             # <<<<<<<<<<<<<<
 *                             self.__target = most_common(indels)
 *                         except:
 */
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_19);
            /*try:*/ {

              /* "aldy/indelpost/varaln.pyx":465
 *                     if indels:
 *                         try:
 *                             self.__target = most_common(indels)             # <<<<<<<<<<<<<<
 *                         except:
 *                             target_pos = self.__target.pos
 */
              __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_most_common); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 465, __pyx_L96_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_2 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
                __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
                if (likely(__pyx_t_2)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
                  __Pyx_INCREF(__pyx_t_2);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_8, function);
                }
              }
              __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_2, __pyx_v_indels) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_indels);
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 465, __pyx_L96_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_4aldy_9indelpost_7variant_Variant))))) __PYX_ERR(0, 465, __pyx_L96_error)
              __Pyx_GIVEREF(__pyx_t_3);
              __Pyx_GOTREF(__pyx_v_self->__pyx___target);
              __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx___target));
              __pyx_v_self->__pyx___target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_3);
              __pyx_t_3 = 0;

              /* "aldy/indelpost/varaln.pyx":464
 * 
 *                     if indels:
 *                         try:             # <<<<<<<<<<<<<<
 *                             self.__target = most_common(indels)
 *                         except:
 */
            }
            __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
            __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            goto __pyx_L101_try_end;
            __pyx_L96_error:;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

            /* "aldy/indelpost/varaln.pyx":466
 *                         try:
 *                             self.__target = most_common(indels)
 *                         except:             # <<<<<<<<<<<<<<
 *                             target_pos = self.__target.pos
 *                             indels.sort(key=lambda x : abs(x.pos - target_pos))
 */
            /*except:*/ {
              __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.__parse_pileup", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_8, &__pyx_t_2) < 0) __PYX_ERR(0, 466, __pyx_L98_except_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_GOTREF(__pyx_t_2);

              /* "aldy/indelpost/varaln.pyx":467
 *                             self.__target = most_common(indels)
 *                         except:
 *                             target_pos = self.__target.pos             # <<<<<<<<<<<<<<
 *                             indels.sort(key=lambda x : abs(x.pos - target_pos))
 * 
 */
              __pyx_t_17 = __pyx_v_self->__pyx___target->pos;
              __pyx_cur_scope->__pyx_v_target_pos = __pyx_t_17;

              /* "aldy/indelpost/varaln.pyx":468
 *                         except:
 *                             target_pos = self.__target.pos
 *                             indels.sort(key=lambda x : abs(x.pos - target_pos))             # <<<<<<<<<<<<<<
 * 
 *                         return self.__parse_pileup(contig=None, retargeted=False, skip_read_end_check=True)
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_indels, __pyx_n_s_sort); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 468, __pyx_L98_except_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_16 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 468, __pyx_L98_except_error)
              __Pyx_GOTREF(__pyx_t_16);
              __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_4aldy_9indelpost_6varaln_16VariantAlignment_14__parse_pileup_lambda1, 0, __pyx_n_s_VariantAlignment___parse_pileup, ((PyObject*)__pyx_cur_scope), __pyx_n_s_aldy_indelpost_varaln, __pyx_d, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 468, __pyx_L98_except_error)
              __Pyx_GOTREF(__pyx_t_13);
              if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_key, __pyx_t_13) < 0) __PYX_ERR(0, 468, __pyx_L98_except_error)
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_16); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 468, __pyx_L98_except_error)
              __Pyx_GOTREF(__pyx_t_13);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              goto __pyx_L97_exception_handled;
            }
            __pyx_L98_except_error:;

            /* "aldy/indelpost/varaln.pyx":464
 * 
 *                     if indels:
 *                         try:             # <<<<<<<<<<<<<<
 *                             self.__target = most_common(indels)
 *                         except:
 */
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
            goto __pyx_L1_error;
            __pyx_L97_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
            __pyx_L101_try_end:;
          }

          /* "aldy/indelpost/varaln.pyx":470
 *                             indels.sort(key=lambda x : abs(x.pos - target_pos))
 * 
 *                         return self.__parse_pileup(contig=None, retargeted=False, skip_read_end_check=True)             # <<<<<<<<<<<<<<
 * 
 *             contig = Contig(
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_18.__pyx_n = 3;
          __pyx_t_18.contig = ((struct __pyx_obj_4aldy_9indelpost_6contig_Contig *)Py_None);
          __pyx_t_18.retargeted = 0;
          __pyx_t_18.skip_read_end_check = 1;
          __pyx_t_2 = ((struct __pyx_vtabstruct_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self->__pyx_vtab)->__pyx___parse_pileup(__pyx_v_self, &__pyx_t_18); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 470, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_r = __pyx_t_2;
          __pyx_t_2 = 0;
          goto __pyx_L0;

          /* "aldy/indelpost/varaln.pyx":463
 *                             indels += [i[-1] for i in new_one["I"]] + [d[-1] for d in new_one["D"]]
 * 
 *                     if indels:             # <<<<<<<<<<<<<<
 *                         try:
 *                             self.__target = most_common(indels)
 */
        }

        /* "aldy/indelpost/varaln.pyx":453
 *             if read_end_evidence_only:
 *                 newly_identified = [read for read in pileup if read["is_target"] and not read in target_pileup]
 *                 if newly_identified:             # <<<<<<<<<<<<<<
 *                     expected_ptrn = most_common_gap_ptrn(newly_identified)
 * 
 */
      }

      /* "aldy/indelpost/varaln.pyx":451
 *             )
 * 
 *             if read_end_evidence_only:             # <<<<<<<<<<<<<<
 *                 newly_identified = [read for read in pileup if read["is_target"] and not read in target_pileup]
 *                 if newly_identified:
 */
    }

    /* "aldy/indelpost/varaln.pyx":475
 *                 self.__target,
 *                 preprocess_for_contig_construction(
 *                     self.__target,             # <<<<<<<<<<<<<<
 *                     self.target,
 *                     pileup,
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_self->__pyx___target);
    __Pyx_INCREF(__pyx_t_2);

    /* "aldy/indelpost/varaln.pyx":476
 *                 preprocess_for_contig_construction(
 *                     self.__target,
 *                     self.target,             # <<<<<<<<<<<<<<
 *                     pileup,
 *                     self.unspliced_local_reference,
 */
    __pyx_t_8 = ((PyObject *)__pyx_v_self->target);
    __Pyx_INCREF(__pyx_t_8);

    /* "aldy/indelpost/varaln.pyx":478
 *                     self.target,
 *                     pileup,
 *                     self.unspliced_local_reference,             # <<<<<<<<<<<<<<
 *                     self.window,
 *                     self.match_score,
 */
    __pyx_t_3 = ((PyObject *)__pyx_v_self->unspliced_local_reference);
    __Pyx_INCREF(__pyx_t_3);

    /* "aldy/indelpost/varaln.pyx":474
 *             contig = Contig(
 *                 self.__target,
 *                 preprocess_for_contig_construction(             # <<<<<<<<<<<<<<
 *                     self.__target,
 *                     self.target,
 */
    __pyx_t_13 = __pyx_f_4aldy_9indelpost_6varaln_preprocess_for_contig_construction(((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_2), ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)__pyx_t_8), __pyx_v_pileup, ((struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference *)__pyx_t_3), __pyx_v_self->window, __pyx_v_self->match_score, __pyx_v_self->mismatch_penalty, __pyx_v_self->gap_open_penalty, __pyx_v_self->gap_extension_penalty); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":486
 *                 ),
 *                 self.unspliced_local_reference,
 *                 self.basequalthresh,             # <<<<<<<<<<<<<<
 *                 self.mapqthresh,
 *             )
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->basequalthresh); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "aldy/indelpost/varaln.pyx":487
 *                 self.unspliced_local_reference,
 *                 self.basequalthresh,
 *                 self.mapqthresh,             # <<<<<<<<<<<<<<
 *             )
 * 
 */
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_self->mapqthresh); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "aldy/indelpost/varaln.pyx":472
 *                         return self.__parse_pileup(contig=None, retargeted=False, skip_read_end_check=True)
 * 
 *             contig = Contig(             # <<<<<<<<<<<<<<
 *                 self.__target,
 *                 preprocess_for_contig_construction(
 */
    __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self->__pyx___target));
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_13);
    __Pyx_INCREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
    PyTuple_SET_ITEM(__pyx_t_2, 2, ((PyObject *)__pyx_v_self->unspliced_local_reference));
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_8);
    __pyx_t_13 = 0;
    __pyx_t_3 = 0;
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4aldy_9indelpost_6contig_Contig), __pyx_t_2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_contig, ((struct __pyx_obj_4aldy_9indelpost_6contig_Contig *)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "aldy/indelpost/varaln.pyx":490
 *             )
 * 
 *             contig = compare_contigs(orig_contig, contig, self.__target.pos)             # <<<<<<<<<<<<<<
 * 
 *         return pileup, contig
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_compare_contigs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->__pyx___target->pos); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = NULL;
    __pyx_t_17 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_17 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[4] = {__pyx_t_13, ((PyObject *)__pyx_v_orig_contig), ((PyObject *)__pyx_v_contig), __pyx_t_3};
      __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[4] = {__pyx_t_13, ((PyObject *)__pyx_v_orig_contig), ((PyObject *)__pyx_v_contig), __pyx_t_3};
      __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_16 = PyTuple_New(3+__pyx_t_17); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      if (__pyx_t_13) {
        __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_13); __pyx_t_13 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_orig_contig));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_orig_contig));
      PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_17, ((PyObject *)__pyx_v_orig_contig));
      __Pyx_INCREF(((PyObject *)__pyx_v_contig));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_contig));
      PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_17, ((PyObject *)__pyx_v_contig));
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_17, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_16, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(((__pyx_t_8) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_8, __pyx_ptype_4aldy_9indelpost_6contig_Contig))))) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_contig, ((struct __pyx_obj_4aldy_9indelpost_6contig_Contig *)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "aldy/indelpost/varaln.pyx":376
 * 
 *         # soft-clip realn & SW realn
 *         if contig.qc_passed and not self.no_realignment:             # <<<<<<<<<<<<<<
 * 
 *             orig_contig = contig
 */
  }

  /* "aldy/indelpost/varaln.pyx":492
 *             contig = compare_contigs(orig_contig, contig, self.__target.pos)
 * 
 *         return pileup, contig             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_pileup);
  __Pyx_GIVEREF(__pyx_v_pileup);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_pileup);
  __Pyx_INCREF(((PyObject *)__pyx_v_contig));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_contig));
  PyTuple_SET_ITEM(__pyx_t_8, 1, ((PyObject *)__pyx_v_contig));
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "aldy/indelpost/varaln.pyx":171
 *         self.__pileup, self.contig = self.__parse_pileup()
 * 
 *     cdef __parse_pileup(self, Contig contig=None, bint retargeted=False, bint skip_read_end_check=False):             # <<<<<<<<<<<<<<
 *         """Dictize reads for target indel and make target indel template by consensus
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.__parse_pileup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pileup);
  __Pyx_XDECREF(__pyx_v_read_end_evidence_only);
  __Pyx_XDECREF(__pyx_v_sample_factor);
  __Pyx_XDECREF(__pyx_v_exptension_penalty_used);
  __Pyx_XDECREF(__pyx_v_grid);
  __Pyx_XDECREF(__pyx_v_ans);
  __Pyx_XDECREF(__pyx_v_intron);
  __Pyx_XDECREF(__pyx_v_overhangs);
  __Pyx_XDECREF(__pyx_v_non_spurious_overhangs);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_retarget_reads);
  __Pyx_XDECREF((PyObject *)__pyx_v_orig_contig);
  __Pyx_XDECREF(__pyx_v_target);
  __Pyx_XDECREF(__pyx_v_nontarget);
  __Pyx_XDECREF(__pyx_v_target_pileup);
  __Pyx_XDECREF(__pyx_v_newly_identified);
  __Pyx_XDECREF(__pyx_v_expected_ptrn);
  __Pyx_XDECREF(__pyx_v_indels);
  __Pyx_XDECREF(__pyx_v_contig_seq);
  __Pyx_XDECREF(__pyx_v_alinger);
  __Pyx_XDECREF(__pyx_v_new_one);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_read);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_read);
  __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_read);
  __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_read);
  __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_read);
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_i);
  __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_d);
  __Pyx_XDECREF((PyObject *)__pyx_v_contig);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":495
 * 
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         # alignment equivalence
 *         my_contig, other_contig = self.contig, other.contig
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_3__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_3__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_2__eq__(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_2__eq__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_4aldy_9indelpost_6contig_Contig *__pyx_v_my_contig = NULL;
  PyObject *__pyx_v_other_contig = NULL;
  PyObject *__pyx_v_phasing_mode = NULL;
  PyObject *__pyx_v_my_phased = NULL;
  PyObject *__pyx_v_other_phased = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);

  /* "aldy/indelpost/varaln.pyx":497
 *     def __eq__(self, other):
 *         # alignment equivalence
 *         my_contig, other_contig = self.contig, other.contig             # <<<<<<<<<<<<<<
 * 
 *         if my_contig.failed or other_contig.failed:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->contig);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_contig); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_my_contig = ((struct __pyx_obj_4aldy_9indelpost_6contig_Contig *)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_other_contig = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":499
 *         my_contig, other_contig = self.contig, other.contig
 * 
 *         if my_contig.failed or other_contig.failed:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  __pyx_t_4 = (__pyx_v_my_contig->failed != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_other_contig, __pyx_n_s_failed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "aldy/indelpost/varaln.pyx":500
 * 
 *         if my_contig.failed or other_contig.failed:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         # check eq in phased form
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":499
 *         my_contig, other_contig = self.contig, other.contig
 * 
 *         if my_contig.failed or other_contig.failed:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  }

  /* "aldy/indelpost/varaln.pyx":503
 * 
 *         # check eq in phased form
 *         phasing_mode = "local"             # <<<<<<<<<<<<<<
 *         my_phased, other_phased = self.phase(how=phasing_mode), other.phase(how=phasing_mode)
 * 
 */
  __Pyx_INCREF(__pyx_n_u_local);
  __pyx_v_phasing_mode = __pyx_n_u_local;

  /* "aldy/indelpost/varaln.pyx":504
 *         # check eq in phased form
 *         phasing_mode = "local"
 *         my_phased, other_phased = self.phase(how=phasing_mode), other.phase(how=phasing_mode)             # <<<<<<<<<<<<<<
 * 
 *         return (my_phased == other_phased)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_phase); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_how, __pyx_v_phasing_mode) < 0) __PYX_ERR(0, 504, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_phase); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_how, __pyx_v_phasing_mode) < 0) __PYX_ERR(0, 504, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_my_phased = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_v_other_phased = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "aldy/indelpost/varaln.pyx":506
 *         my_phased, other_phased = self.phase(how=phasing_mode), other.phase(how=phasing_mode)
 * 
 *         return (my_phased == other_phased)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = PyObject_RichCompare(__pyx_v_my_phased, __pyx_v_other_phased, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 506, __pyx_L1_error)
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "aldy/indelpost/varaln.pyx":495
 * 
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         # alignment equivalence
 *         my_contig, other_contig = self.contig, other.contig
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_my_contig);
  __Pyx_XDECREF(__pyx_v_other_contig);
  __Pyx_XDECREF(__pyx_v_phasing_mode);
  __Pyx_XDECREF(__pyx_v_my_phased);
  __Pyx_XDECREF(__pyx_v_other_phased);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":509
 * 
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         hashable = self.phase(how="local")
 *         return hash(hashable)
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_5__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_5__hash__(PyObject *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_4__hash__(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_4__hash__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self) {
  PyObject *__pyx_v_hashable = NULL;
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_hash_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "aldy/indelpost/varaln.pyx":510
 * 
 *     def __hash__(self):
 *         hashable = self.phase(how="local")             # <<<<<<<<<<<<<<
 *         return hash(hashable)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_phase); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_how, __pyx_n_u_local) < 0) __PYX_ERR(0, 510, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_hashable = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "aldy/indelpost/varaln.pyx":511
 *     def __hash__(self):
 *         hashable = self.phase(how="local")
 *         return hash(hashable)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = PyObject_Hash(__pyx_v_hashable); if (unlikely(__pyx_t_4 == ((Py_hash_t)-1))) __PYX_ERR(0, 511, __pyx_L1_error)
  __pyx_r = __pyx_t_4;
  goto __pyx_L0;

  /* "aldy/indelpost/varaln.pyx":509
 * 
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         hashable = self.phase(how="local")
 *         return hash(hashable)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_hashable);
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":514
 * 
 * 
 *     def get_contig(self):             # <<<<<<<<<<<<<<
 *         """returns :class:`~indelpost.Contig` object built from reads supporting the target indel.
 *         :class:`~indelpost.FailedContig` is returned if contig assembly is not successful.
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_7get_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_16VariantAlignment_6get_contig[] = "VariantAlignment.get_contig(self)\nreturns :class:`~indelpost.Contig` object built from reads supporting the target indel.\n        :class:`~indelpost.FailedContig` is returned if contig assembly is not successful.\n        ";
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_7get_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_contig (wrapper)", 0);
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_6get_contig(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_6get_contig(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self) {
  struct __pyx_obj_4aldy_9indelpost_6contig_Contig *__pyx_v_contig = NULL;
  struct __pyx_obj_4aldy_9indelpost_6contig_FailedContig *__pyx_v_failed = NULL;
  PyObject *__pyx_v_alt_cnt = NULL;
  PyObject *__pyx_v_dirty_target_pileup = NULL;
  PyObject *__pyx_8genexpr8__pyx_v_read = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_contig", 0);

  /* "aldy/indelpost/varaln.pyx":518
 *         :class:`~indelpost.FailedContig` is returned if contig assembly is not successful.
 *         """
 *         contig = self.contig             # <<<<<<<<<<<<<<
 *         if contig and not contig.failed:
 *             return contig
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->contig);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_contig = ((struct __pyx_obj_4aldy_9indelpost_6contig_Contig *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":519
 *         """
 *         contig = self.contig
 *         if contig and not contig.failed:             # <<<<<<<<<<<<<<
 *             return contig
 *         else:
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_contig)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 519, __pyx_L1_error)
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = ((!(__pyx_v_contig->failed != 0)) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "aldy/indelpost/varaln.pyx":520
 *         contig = self.contig
 *         if contig and not contig.failed:
 *             return contig             # <<<<<<<<<<<<<<
 *         else:
 *             failed = FailedContig()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_contig));
    __pyx_r = ((PyObject *)__pyx_v_contig);
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":519
 *         """
 *         contig = self.contig
 *         if contig and not contig.failed:             # <<<<<<<<<<<<<<
 *             return contig
 *         else:
 */
  }

  /* "aldy/indelpost/varaln.pyx":522
 *             return contig
 *         else:
 *             failed = FailedContig()             # <<<<<<<<<<<<<<
 *             alt_cnt = self.count_alleles()[1]
 *             if alt_cnt:
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_4aldy_9indelpost_6contig_FailedContig)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_failed = ((struct __pyx_obj_4aldy_9indelpost_6contig_FailedContig *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "aldy/indelpost/varaln.pyx":523
 *         else:
 *             failed = FailedContig()
 *             alt_cnt = self.count_alleles()[1]             # <<<<<<<<<<<<<<
 *             if alt_cnt:
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_count_alleles); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_alt_cnt = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "aldy/indelpost/varaln.pyx":524
 *             failed = FailedContig()
 *             alt_cnt = self.count_alleles()[1]
 *             if alt_cnt:             # <<<<<<<<<<<<<<
 * 
 *                 dirty_target_pileup = [read["is_dirty"] for read in self.__pileup if read["is_target"]]
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_alt_cnt); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 524, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "aldy/indelpost/varaln.pyx":526
 *             if alt_cnt:
 * 
 *                 dirty_target_pileup = [read["is_dirty"] for read in self.__pileup if read["is_target"]]             # <<<<<<<<<<<<<<
 *                 if sum(dirty_target_pileup) == len(dirty_target_pileup):
 *                     failed.is_low_quality = True
 */
      { /* enter inner scope */
        __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 526, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(__pyx_v_self->__pyx___pileup == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
          __PYX_ERR(0, 526, __pyx_L9_error)
        }
        __pyx_t_1 = __pyx_v_self->__pyx___pileup; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
        for (;;) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 526, __pyx_L9_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 526, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
          __Pyx_XDECREF_SET(__pyx_8genexpr8__pyx_v_read, __pyx_t_5);
          __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_8genexpr8__pyx_v_read, __pyx_n_u_is_target); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 526, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 526, __pyx_L9_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (__pyx_t_2) {
            __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_8genexpr8__pyx_v_read, __pyx_n_u_is_dirty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 526, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 526, __pyx_L9_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_read); __pyx_8genexpr8__pyx_v_read = 0;
        goto __pyx_L13_exit_scope;
        __pyx_L9_error:;
        __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_read); __pyx_8genexpr8__pyx_v_read = 0;
        goto __pyx_L1_error;
        __pyx_L13_exit_scope:;
      } /* exit inner scope */
      __pyx_v_dirty_target_pileup = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "aldy/indelpost/varaln.pyx":527
 * 
 *                 dirty_target_pileup = [read["is_dirty"] for read in self.__pileup if read["is_target"]]
 *                 if sum(dirty_target_pileup) == len(dirty_target_pileup):             # <<<<<<<<<<<<<<
 *                     failed.is_low_quality = True
 *                 else:
 */
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_v_dirty_target_pileup); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 527, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = PyList_GET_SIZE(__pyx_v_dirty_target_pileup); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 527, __pyx_L1_error)
      __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 527, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 527, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 527, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_2) {

        /* "aldy/indelpost/varaln.pyx":528
 *                 dirty_target_pileup = [read["is_dirty"] for read in self.__pileup if read["is_target"]]
 *                 if sum(dirty_target_pileup) == len(dirty_target_pileup):
 *                     failed.is_low_quality = True             # <<<<<<<<<<<<<<
 *                 else:
 *                     failed.failed_anyway = True
 */
        __pyx_v_failed->is_low_quality = 1;

        /* "aldy/indelpost/varaln.pyx":527
 * 
 *                 dirty_target_pileup = [read["is_dirty"] for read in self.__pileup if read["is_target"]]
 *                 if sum(dirty_target_pileup) == len(dirty_target_pileup):             # <<<<<<<<<<<<<<
 *                     failed.is_low_quality = True
 *                 else:
 */
        goto __pyx_L14;
      }

      /* "aldy/indelpost/varaln.pyx":530
 *                     failed.is_low_quality = True
 *                 else:
 *                     failed.failed_anyway = True             # <<<<<<<<<<<<<<
 *             else:
 *                 failed.target_not_found = True
 */
      /*else*/ {
        __pyx_v_failed->failed_anyway = 1;
      }
      __pyx_L14:;

      /* "aldy/indelpost/varaln.pyx":524
 *             failed = FailedContig()
 *             alt_cnt = self.count_alleles()[1]
 *             if alt_cnt:             # <<<<<<<<<<<<<<
 * 
 *                 dirty_target_pileup = [read["is_dirty"] for read in self.__pileup if read["is_target"]]
 */
      goto __pyx_L6;
    }

    /* "aldy/indelpost/varaln.pyx":532
 *                     failed.failed_anyway = True
 *             else:
 *                 failed.target_not_found = True             # <<<<<<<<<<<<<<
 * 
 *             return failed
 */
    /*else*/ {
      __pyx_v_failed->target_not_found = 1;
    }
    __pyx_L6:;

    /* "aldy/indelpost/varaln.pyx":534
 *                 failed.target_not_found = True
 * 
 *             return failed             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_failed));
    __pyx_r = ((PyObject *)__pyx_v_failed);
    goto __pyx_L0;
  }

  /* "aldy/indelpost/varaln.pyx":514
 * 
 * 
 *     def get_contig(self):             # <<<<<<<<<<<<<<
 *         """returns :class:`~indelpost.Contig` object built from reads supporting the target indel.
 *         :class:`~indelpost.FailedContig` is returned if contig assembly is not successful.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.get_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_contig);
  __Pyx_XDECREF((PyObject *)__pyx_v_failed);
  __Pyx_XDECREF(__pyx_v_alt_cnt);
  __Pyx_XDECREF(__pyx_v_dirty_target_pileup);
  __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_read);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":537
 * 
 * 
 *     def get_target_indel(self):             # <<<<<<<<<<<<<<
 *         """returns :class:`~indelpost.Variant` object representing the actual target indel analyzed.
 *         The target indel may be different from the input indel due to the internal realignment.
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_9get_target_indel(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_16VariantAlignment_8get_target_indel[] = "VariantAlignment.get_target_indel(self)\nreturns :class:`~indelpost.Variant` object representing the actual target indel analyzed.\n        The target indel may be different from the input indel due to the internal realignment.\n        :class:`~indelpost.NullVariant` is returned when no target is found.\n        ";
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_9get_target_indel(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_target_indel (wrapper)", 0);
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_8get_target_indel(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_8get_target_indel(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self) {
  PyObject *__pyx_v_alt_cnt = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_target_indel", 0);

  /* "aldy/indelpost/varaln.pyx":542
 *         :class:`~indelpost.NullVariant` is returned when no target is found.
 *         """
 *         alt_cnt = self.count_alleles()[1]             # <<<<<<<<<<<<<<
 *         if alt_cnt:
 *             return self.__target
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_count_alleles); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_alt_cnt = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":543
 *         """
 *         alt_cnt = self.count_alleles()[1]
 *         if alt_cnt:             # <<<<<<<<<<<<<<
 *             return self.__target
 *         else:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_alt_cnt); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 543, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "aldy/indelpost/varaln.pyx":544
 *         alt_cnt = self.count_alleles()[1]
 *         if alt_cnt:
 *             return self.__target             # <<<<<<<<<<<<<<
 *         else:
 *             return NullVariant(self.__target.chrom, self.__target.pos, self.__target.reference)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
    __pyx_r = ((PyObject *)__pyx_v_self->__pyx___target);
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":543
 *         """
 *         alt_cnt = self.count_alleles()[1]
 *         if alt_cnt:             # <<<<<<<<<<<<<<
 *             return self.__target
 *         else:
 */
  }

  /* "aldy/indelpost/varaln.pyx":546
 *             return self.__target
 *         else:
 *             return NullVariant(self.__target.chrom, self.__target.pos, self.__target.reference)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->__pyx___target->pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_self->__pyx___target->chrom);
    __Pyx_GIVEREF(__pyx_v_self->__pyx___target->chrom);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx___target->chrom);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target->reference));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target->reference));
    PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->__pyx___target->reference));
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4aldy_9indelpost_7variant_NullVariant), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "aldy/indelpost/varaln.pyx":537
 * 
 * 
 *     def get_target_indel(self):             # <<<<<<<<<<<<<<
 *         """returns :class:`~indelpost.Variant` object representing the actual target indel analyzed.
 *         The target indel may be different from the input indel due to the internal realignment.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.get_target_indel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_alt_cnt);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":549
 * 
 * 
 *     def fetch_reads(self, how='target'):             # <<<<<<<<<<<<<<
 *         """returns a `list <https://docs.python.org/3/library/stdtypes.html#list>`__
 *         of `pysam.AlignedSegment <https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignedSegment>`__ objects,
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_11fetch_reads(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_16VariantAlignment_10fetch_reads[] = "VariantAlignment.fetch_reads(self, how=u'target')\nreturns a `list <https://docs.python.org/3/library/stdtypes.html#list>`__\n        of `pysam.AlignedSegment <https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignedSegment>`__ objects,\n        depending on the strategy specified.\n\n        Parameters\n        ----------\n            how : string\n                specifies the read fetch strategy. \"target\" (default) to retrieve reads supporting the target.\n                \"non_target\" for reads w/o target. \"covering\" for reads covering the locus w/ and w/o the target.\n        ";
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_11fetch_reads(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_how = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fetch_reads (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_how,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_n_u_target);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_how);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "fetch_reads") < 0)) __PYX_ERR(0, 549, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_how = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("fetch_reads", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 549, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.fetch_reads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_10fetch_reads(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self), __pyx_v_how);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_4aldy_9indelpost_6varaln_16VariantAlignment_11fetch_reads_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "aldy/indelpost/varaln.pyx":565
 *         elif how == "non_target":
 *             pos, indel_len = self._observed_pos, len(self.target.indel_seq)
 *             r_pos = max(v.pos for v in self.target.generate_equivalents())             # <<<<<<<<<<<<<<
 *             margin = r_pos - pos
 *             del_len = indel_len if self.target.is_del else 0
 */

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_11fetch_reads_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr *)__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr(__pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 565, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_4aldy_9indelpost_6varaln_16VariantAlignment_11fetch_reads_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_fetch_reads_locals_genexpr, __pyx_n_s_aldy_indelpost_varaln); if (unlikely(!gen)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.fetch_reads.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_4aldy_9indelpost_6varaln_16VariantAlignment_11fetch_reads_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr *__pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 565, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 565, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->target), __pyx_n_s_generate_equivalents); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 565, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 565, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 565, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 565, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_v, __pyx_n_s_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_4;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_5;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_5 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 565, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":549
 * 
 * 
 *     def fetch_reads(self, how='target'):             # <<<<<<<<<<<<<<
 *         """returns a `list <https://docs.python.org/3/library/stdtypes.html#list>`__
 *         of `pysam.AlignedSegment <https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignedSegment>`__ objects,
 */

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_10fetch_reads(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self, PyObject *__pyx_v_how) {
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads *__pyx_cur_scope;
  PyObject *__pyx_v_pos = NULL;
  Py_ssize_t __pyx_v_indel_len;
  PyObject *__pyx_v_r_pos = NULL;
  PyObject *__pyx_v_margin = NULL;
  Py_ssize_t __pyx_v_del_len;
  PyObject *__pyx_v_targets = NULL;
  PyObject *__pyx_8genexpr9__pyx_v_read = NULL;
  PyObject *__pyx_gb_4aldy_9indelpost_6varaln_16VariantAlignment_11fetch_reads_2generator = 0;
  PyObject *__pyx_9genexpr11__pyx_v_read = NULL;
  PyObject *__pyx_9genexpr12__pyx_v_read = NULL;
  PyObject *__pyx_9genexpr13__pyx_v_read = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fetch_reads", 0);
  __pyx_cur_scope = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads *)__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads(__pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 549, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "aldy/indelpost/varaln.pyx":561
 *         """
 * 
 *         if how == "target":             # <<<<<<<<<<<<<<
 *             return [read["read"] for read in self.__pileup if read["is_target"]]
 *         elif how == "non_target":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_how, __pyx_n_u_target, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 561, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "aldy/indelpost/varaln.pyx":562
 * 
 *         if how == "target":
 *             return [read["read"] for read in self.__pileup if read["is_target"]]             # <<<<<<<<<<<<<<
 *         elif how == "non_target":
 *             pos, indel_len = self._observed_pos, len(self.target.indel_seq)
 */
    __Pyx_XDECREF(__pyx_r);
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 562, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__pyx_cur_scope->__pyx_v_self->__pyx___pileup == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 562, __pyx_L6_error)
      }
      __pyx_t_3 = __pyx_cur_scope->__pyx_v_self->__pyx___pileup; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
      for (;;) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 562, __pyx_L6_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 562, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
        __Pyx_XDECREF_SET(__pyx_8genexpr9__pyx_v_read, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_8genexpr9__pyx_v_read, __pyx_n_u_is_target); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 562, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 562, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (__pyx_t_1) {
          __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_8genexpr9__pyx_v_read, __pyx_n_u_read); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 562, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 562, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_read); __pyx_8genexpr9__pyx_v_read = 0;
      goto __pyx_L10_exit_scope;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_read); __pyx_8genexpr9__pyx_v_read = 0;
      goto __pyx_L1_error;
      __pyx_L10_exit_scope:;
    } /* exit inner scope */
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":561
 *         """
 * 
 *         if how == "target":             # <<<<<<<<<<<<<<
 *             return [read["read"] for read in self.__pileup if read["is_target"]]
 *         elif how == "non_target":
 */
  }

  /* "aldy/indelpost/varaln.pyx":563
 *         if how == "target":
 *             return [read["read"] for read in self.__pileup if read["is_target"]]
 *         elif how == "non_target":             # <<<<<<<<<<<<<<
 *             pos, indel_len = self._observed_pos, len(self.target.indel_seq)
 *             r_pos = max(v.pos for v in self.target.generate_equivalents())
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_how, __pyx_n_u_non_target, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 563, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "aldy/indelpost/varaln.pyx":564
 *             return [read["read"] for read in self.__pileup if read["is_target"]]
 *         elif how == "non_target":
 *             pos, indel_len = self._observed_pos, len(self.target.indel_seq)             # <<<<<<<<<<<<<<
 *             r_pos = max(v.pos for v in self.target.generate_equivalents())
 *             margin = r_pos - pos
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_self->_observed_pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self->target), __pyx_n_s_indel_seq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_pos = __pyx_t_2;
    __pyx_t_2 = 0;
    __pyx_v_indel_len = __pyx_t_4;

    /* "aldy/indelpost/varaln.pyx":565
 *         elif how == "non_target":
 *             pos, indel_len = self._observed_pos, len(self.target.indel_seq)
 *             r_pos = max(v.pos for v in self.target.generate_equivalents())             # <<<<<<<<<<<<<<
 *             margin = r_pos - pos
 *             del_len = indel_len if self.target.is_del else 0
 */
    __pyx_t_2 = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_11fetch_reads_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_r_pos = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":566
 *             pos, indel_len = self._observed_pos, len(self.target.indel_seq)
 *             r_pos = max(v.pos for v in self.target.generate_equivalents())
 *             margin = r_pos - pos             # <<<<<<<<<<<<<<
 *             del_len = indel_len if self.target.is_del else 0
 *             targets = [read["read_name"] for read in self.__pileup if read["is_target"]]
 */
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_r_pos, __pyx_v_pos); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_margin = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":567
 *             r_pos = max(v.pos for v in self.target.generate_equivalents())
 *             margin = r_pos - pos
 *             del_len = indel_len if self.target.is_del else 0             # <<<<<<<<<<<<<<
 *             targets = [read["read_name"] for read in self.__pileup if read["is_target"]]
 *             return [read["read"] for read in self.__pileup if count_as_non_target(read, pos, del_len, margin) and not read["read_name"] in targets]
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self->target), __pyx_n_s_is_del); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 567, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {
      __pyx_t_4 = __pyx_v_indel_len;
    } else {
      __pyx_t_4 = 0;
    }
    __pyx_v_del_len = __pyx_t_4;

    /* "aldy/indelpost/varaln.pyx":568
 *             margin = r_pos - pos
 *             del_len = indel_len if self.target.is_del else 0
 *             targets = [read["read_name"] for read in self.__pileup if read["is_target"]]             # <<<<<<<<<<<<<<
 *             return [read["read"] for read in self.__pileup if count_as_non_target(read, pos, del_len, margin) and not read["read_name"] in targets]
 *         elif how == "covering":
 */
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 568, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__pyx_cur_scope->__pyx_v_self->__pyx___pileup == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 568, __pyx_L13_error)
      }
      __pyx_t_2 = __pyx_cur_scope->__pyx_v_self->__pyx___pileup; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
      for (;;) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 568, __pyx_L13_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
        __Pyx_XDECREF_SET(__pyx_9genexpr11__pyx_v_read, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_9genexpr11__pyx_v_read, __pyx_n_u_is_target); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 568, __pyx_L13_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (__pyx_t_1) {
          __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_9genexpr11__pyx_v_read, __pyx_n_u_read_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 568, __pyx_L13_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_read); __pyx_9genexpr11__pyx_v_read = 0;
      goto __pyx_L17_exit_scope;
      __pyx_L13_error:;
      __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_read); __pyx_9genexpr11__pyx_v_read = 0;
      goto __pyx_L1_error;
      __pyx_L17_exit_scope:;
    } /* exit inner scope */
    __pyx_v_targets = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":569
 *             del_len = indel_len if self.target.is_del else 0
 *             targets = [read["read_name"] for read in self.__pileup if read["is_target"]]
 *             return [read["read"] for read in self.__pileup if count_as_non_target(read, pos, del_len, margin) and not read["read_name"] in targets]             # <<<<<<<<<<<<<<
 *         elif how == "covering":
 *             return [read["read"] for read in self.__pileup if read["is_covering"]]
 */
    __Pyx_XDECREF(__pyx_r);
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 569, __pyx_L20_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__pyx_cur_scope->__pyx_v_self->__pyx___pileup == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 569, __pyx_L20_error)
      }
      __pyx_t_2 = __pyx_cur_scope->__pyx_v_self->__pyx___pileup; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
      for (;;) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 569, __pyx_L20_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 569, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
        __Pyx_XDECREF_SET(__pyx_9genexpr12__pyx_v_read, __pyx_t_5);
        __pyx_t_5 = 0;
        if (!(likely(PyDict_CheckExact(__pyx_9genexpr12__pyx_v_read))||((__pyx_9genexpr12__pyx_v_read) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_9genexpr12__pyx_v_read)->tp_name), 0))) __PYX_ERR(0, 569, __pyx_L20_error)
        __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_pos); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 569, __pyx_L20_error)
        __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_margin); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 569, __pyx_L20_error)
        __pyx_t_8 = (__pyx_f_4aldy_9indelpost_6varaln_count_as_non_target(((PyObject*)__pyx_9genexpr12__pyx_v_read), __pyx_t_6, __pyx_v_del_len, __pyx_t_7) != 0);
        if (__pyx_t_8) {
        } else {
          __pyx_t_1 = __pyx_t_8;
          goto __pyx_L24_bool_binop_done;
        }
        __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_9genexpr12__pyx_v_read, __pyx_n_u_read_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 569, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_t_5, __pyx_v_targets, Py_NE)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 569, __pyx_L20_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_9 = (__pyx_t_8 != 0);
        __pyx_t_1 = __pyx_t_9;
        __pyx_L24_bool_binop_done:;
        if (__pyx_t_1) {
          __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_9genexpr12__pyx_v_read, __pyx_n_u_read); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 569, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 569, __pyx_L20_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_read); __pyx_9genexpr12__pyx_v_read = 0;
      goto __pyx_L26_exit_scope;
      __pyx_L20_error:;
      __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_read); __pyx_9genexpr12__pyx_v_read = 0;
      goto __pyx_L1_error;
      __pyx_L26_exit_scope:;
    } /* exit inner scope */
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":563
 *         if how == "target":
 *             return [read["read"] for read in self.__pileup if read["is_target"]]
 *         elif how == "non_target":             # <<<<<<<<<<<<<<
 *             pos, indel_len = self._observed_pos, len(self.target.indel_seq)
 *             r_pos = max(v.pos for v in self.target.generate_equivalents())
 */
  }

  /* "aldy/indelpost/varaln.pyx":570
 *             targets = [read["read_name"] for read in self.__pileup if read["is_target"]]
 *             return [read["read"] for read in self.__pileup if count_as_non_target(read, pos, del_len, margin) and not read["read_name"] in targets]
 *         elif how == "covering":             # <<<<<<<<<<<<<<
 *             return [read["read"] for read in self.__pileup if read["is_covering"]]
 *         else:
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_how, __pyx_n_u_covering, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 570, __pyx_L1_error)
  if (likely(__pyx_t_1)) {

    /* "aldy/indelpost/varaln.pyx":571
 *             return [read["read"] for read in self.__pileup if count_as_non_target(read, pos, del_len, margin) and not read["read_name"] in targets]
 *         elif how == "covering":
 *             return [read["read"] for read in self.__pileup if read["is_covering"]]             # <<<<<<<<<<<<<<
 *         else:
 *             raise Exception("fetch stragety must be either of target, non_target, covering")
 */
    __Pyx_XDECREF(__pyx_r);
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 571, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__pyx_cur_scope->__pyx_v_self->__pyx___pileup == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 571, __pyx_L29_error)
      }
      __pyx_t_2 = __pyx_cur_scope->__pyx_v_self->__pyx___pileup; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
      for (;;) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 571, __pyx_L29_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 571, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
        __Pyx_XDECREF_SET(__pyx_9genexpr13__pyx_v_read, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_9genexpr13__pyx_v_read, __pyx_n_u_is_covering); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 571, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 571, __pyx_L29_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (__pyx_t_1) {
          __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_9genexpr13__pyx_v_read, __pyx_n_u_read); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 571, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 571, __pyx_L29_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_read); __pyx_9genexpr13__pyx_v_read = 0;
      goto __pyx_L33_exit_scope;
      __pyx_L29_error:;
      __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_read); __pyx_9genexpr13__pyx_v_read = 0;
      goto __pyx_L1_error;
      __pyx_L33_exit_scope:;
    } /* exit inner scope */
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":570
 *             targets = [read["read_name"] for read in self.__pileup if read["is_target"]]
 *             return [read["read"] for read in self.__pileup if count_as_non_target(read, pos, del_len, margin) and not read["read_name"] in targets]
 *         elif how == "covering":             # <<<<<<<<<<<<<<
 *             return [read["read"] for read in self.__pileup if read["is_covering"]]
 *         else:
 */
  }

  /* "aldy/indelpost/varaln.pyx":573
 *             return [read["read"] for read in self.__pileup if read["is_covering"]]
 *         else:
 *             raise Exception("fetch stragety must be either of target, non_target, covering")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 573, __pyx_L1_error)
  }

  /* "aldy/indelpost/varaln.pyx":549
 * 
 * 
 *     def fetch_reads(self, how='target'):             # <<<<<<<<<<<<<<
 *         """returns a `list <https://docs.python.org/3/library/stdtypes.html#list>`__
 *         of `pysam.AlignedSegment <https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignedSegment>`__ objects,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.fetch_reads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pos);
  __Pyx_XDECREF(__pyx_v_r_pos);
  __Pyx_XDECREF(__pyx_v_margin);
  __Pyx_XDECREF(__pyx_v_targets);
  __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_read);
  __Pyx_XDECREF(__pyx_gb_4aldy_9indelpost_6varaln_16VariantAlignment_11fetch_reads_2generator);
  __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_read);
  __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_read);
  __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_read);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":576
 * 
 * 
 *     def count_alleles(             # <<<<<<<<<<<<<<
 *         self, fwrv=False, by_fragment=False, three_class=False, estimated_count=False, quality_window=None, quality_threshold=None
 *     ):
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_13count_alleles(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_16VariantAlignment_12count_alleles[] = "VariantAlignment.count_alleles(self, fwrv=False, by_fragment=False, three_class=False, estimated_count=False, quality_window=None, quality_threshold=None)\nreturns a `tuple <https://docs.python.org/3/library/stdtypes.html#tuple>`__ of\n        read counts:  (#non_target reads, #target reads).\n\n        Parameters\n        ----------\n            fwrv : bool\n                breaks down to the forward and reverse read counts.\n                ( (non_target-fw, non_target-rv), (target-fw, target-rv) ).\n            by_fragment : bool\n                counts by fragment. Overlapping fw and rv reads are counted as one.\n            three_class : bool\n                breaks down to (reference, non_reference_non_target, target) counts.\n            estimated_count : bool\n                True to return estimated count when the coverage is higher than :attr:`~indelpost.VariantAlignment.downsample_threshold`.\n            quality_window : integer\n                specifies the range of base call quality filter. indel pos +/- quality_window.\n            quality_threshold : integer\n                filters reads with the median base call quality in indel pos +/- quality_window < quality_threshold.\n        ";
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_13count_alleles(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fwrv = 0;
  PyObject *__pyx_v_by_fragment = 0;
  PyObject *__pyx_v_three_class = 0;
  PyObject *__pyx_v_estimated_count = 0;
  PyObject *__pyx_v_quality_window = 0;
  PyObject *__pyx_v_quality_threshold = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("count_alleles (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fwrv,&__pyx_n_s_by_fragment,&__pyx_n_s_three_class,&__pyx_n_s_estimated_count,&__pyx_n_s_quality_window,&__pyx_n_s_quality_threshold,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "aldy/indelpost/varaln.pyx":577
 * 
 *     def count_alleles(
 *         self, fwrv=False, by_fragment=False, three_class=False, estimated_count=False, quality_window=None, quality_threshold=None             # <<<<<<<<<<<<<<
 *     ):
 *         """returns a `tuple <https://docs.python.org/3/library/stdtypes.html#tuple>`__ of
 */
    values[0] = ((PyObject *)Py_False);
    values[1] = ((PyObject *)Py_False);
    values[2] = ((PyObject *)Py_False);
    values[3] = ((PyObject *)Py_False);
    values[4] = ((PyObject *)Py_None);
    values[5] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fwrv);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_by_fragment);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_three_class);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_estimated_count);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_quality_window);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_quality_threshold);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "count_alleles") < 0)) __PYX_ERR(0, 576, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fwrv = values[0];
    __pyx_v_by_fragment = values[1];
    __pyx_v_three_class = values[2];
    __pyx_v_estimated_count = values[3];
    __pyx_v_quality_window = values[4];
    __pyx_v_quality_threshold = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("count_alleles", 0, 0, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 576, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.count_alleles", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_12count_alleles(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self), __pyx_v_fwrv, __pyx_v_by_fragment, __pyx_v_three_class, __pyx_v_estimated_count, __pyx_v_quality_window, __pyx_v_quality_threshold);

  /* "aldy/indelpost/varaln.pyx":576
 * 
 * 
 *     def count_alleles(             # <<<<<<<<<<<<<<
 *         self, fwrv=False, by_fragment=False, three_class=False, estimated_count=False, quality_window=None, quality_threshold=None
 *     ):
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_4aldy_9indelpost_6varaln_16VariantAlignment_13count_alleles_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "aldy/indelpost/varaln.pyx":603
 *         pos, indel_len = self._observed_pos, len(self.target.indel_seq)
 * 
 *         r_pos = max(v.pos for v in self.__target.generate_equivalents())             # <<<<<<<<<<<<<<
 * 
 *         margin = r_pos - pos
 */

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_13count_alleles_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr *)__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr(__pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 603, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_4aldy_9indelpost_6varaln_16VariantAlignment_13count_alleles_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_count_alleles_locals_genexpr, __pyx_n_s_aldy_indelpost_varaln); if (unlikely(!gen)) __PYX_ERR(0, 603, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.count_alleles.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_4aldy_9indelpost_6varaln_16VariantAlignment_13count_alleles_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr *__pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 603, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 603, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->__pyx___target), __pyx_n_s_generate_equivalents); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 603, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 603, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 603, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 603, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 603, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 603, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_v, __pyx_n_s_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_4;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_5;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_5 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 603, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":576
 * 
 * 
 *     def count_alleles(             # <<<<<<<<<<<<<<
 *         self, fwrv=False, by_fragment=False, three_class=False, estimated_count=False, quality_window=None, quality_threshold=None
 *     ):
 */

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_12count_alleles(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self, PyObject *__pyx_v_fwrv, PyObject *__pyx_v_by_fragment, PyObject *__pyx_v_three_class, PyObject *__pyx_v_estimated_count, PyObject *__pyx_v_quality_window, PyObject *__pyx_v_quality_threshold) {
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles *__pyx_cur_scope;
  PyObject *__pyx_v_read = 0;
  PyObject *__pyx_v_pos = NULL;
  Py_ssize_t __pyx_v_indel_len;
  PyObject *__pyx_v_r_pos = NULL;
  PyObject *__pyx_v_margin = NULL;
  Py_ssize_t __pyx_v_del_len;
  PyObject *__pyx_v_reads = NULL;
  PyObject *__pyx_v_fw_target = NULL;
  PyObject *__pyx_v_rv_target = NULL;
  PyObject *__pyx_v_fw_non_target = NULL;
  PyObject *__pyx_v_rv_non_target = NULL;
  PyObject *__pyx_v_rw_target = NULL;
  PyObject *__pyx_v_fwrv_target = NULL;
  PyObject *__pyx_v_est = NULL;
  PyObject *__pyx_v_fw_ref = NULL;
  PyObject *__pyx_v_fw_non_ref_non_target = NULL;
  PyObject *__pyx_v_rv_ref = NULL;
  PyObject *__pyx_v_rv_non_ref_non_target = NULL;
  Py_ssize_t __pyx_v_fwrv_ref;
  Py_ssize_t __pyx_v_fwrv_non_ref_non_target;
  PyObject *__pyx_v_fwrv_non_target = NULL;
  PyObject *__pyx_gb_4aldy_9indelpost_6varaln_16VariantAlignment_13count_alleles_2generator1 = 0;
  PyObject *__pyx_9genexpr15__pyx_v_read = NULL;
  PyObject *__pyx_9genexpr16__pyx_v_read = NULL;
  PyObject *__pyx_9genexpr17__pyx_v_read = NULL;
  PyObject *__pyx_9genexpr18__pyx_v_read = NULL;
  PyObject *__pyx_9genexpr19__pyx_v_read = NULL;
  PyObject *__pyx_9genexpr20__pyx_v_read = NULL;
  PyObject *__pyx_9genexpr21__pyx_v_read = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_alleles", 0);
  __pyx_cur_scope = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles *)__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles(__pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 576, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "aldy/indelpost/varaln.pyx":601
 *         cdef dict read
 * 
 *         pos, indel_len = self._observed_pos, len(self.target.indel_seq)             # <<<<<<<<<<<<<<
 * 
 *         r_pos = max(v.pos for v in self.__target.generate_equivalents())
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_self->_observed_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self->target), __pyx_n_s_indel_seq); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_pos = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_indel_len = __pyx_t_3;

  /* "aldy/indelpost/varaln.pyx":603
 *         pos, indel_len = self._observed_pos, len(self.target.indel_seq)
 * 
 *         r_pos = max(v.pos for v in self.__target.generate_equivalents())             # <<<<<<<<<<<<<<
 * 
 *         margin = r_pos - pos
 */
  __pyx_t_1 = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_13count_alleles_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_r_pos = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":605
 *         r_pos = max(v.pos for v in self.__target.generate_equivalents())
 * 
 *         margin = r_pos - pos             # <<<<<<<<<<<<<<
 * 
 *         del_len = indel_len if self.target.is_del else 0
 */
  __pyx_t_2 = PyNumber_Subtract(__pyx_v_r_pos, __pyx_v_pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_margin = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":607
 *         margin = r_pos - pos
 * 
 *         del_len = indel_len if self.target.is_del else 0             # <<<<<<<<<<<<<<
 * 
 *         reads = self.__pileup
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self->target), __pyx_n_s_is_del); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 607, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {
    __pyx_t_3 = __pyx_v_indel_len;
  } else {
    __pyx_t_3 = 0;
  }
  __pyx_v_del_len = __pyx_t_3;

  /* "aldy/indelpost/varaln.pyx":609
 *         del_len = indel_len if self.target.is_del else 0
 * 
 *         reads = self.__pileup             # <<<<<<<<<<<<<<
 *         if quality_window and quality_threshold:
 *             reads = [
 */
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_self->__pyx___pileup;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_reads = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":610
 * 
 *         reads = self.__pileup
 *         if quality_window and quality_threshold:             # <<<<<<<<<<<<<<
 *             reads = [
 *                 read
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_quality_window); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 610, __pyx_L1_error)
  if (__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_quality_threshold); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 610, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "aldy/indelpost/varaln.pyx":611
 *         reads = self.__pileup
 *         if quality_window and quality_threshold:
 *             reads = [             # <<<<<<<<<<<<<<
 *                 read
 *                 for read in reads
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 611, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "aldy/indelpost/varaln.pyx":613
 *             reads = [
 *                 read
 *                 for read in reads             # <<<<<<<<<<<<<<
 *                 if is_quality_read(read, pos, quality_window, quality_threshold)
 *             ]
 */
      if (unlikely(__pyx_v_reads == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 613, __pyx_L8_error)
      }
      __pyx_t_1 = __pyx_v_reads; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
      for (;;) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 613, __pyx_L8_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 613, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        if (!(likely(PyDict_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 613, __pyx_L8_error)
        __Pyx_XDECREF_SET(__pyx_9genexpr15__pyx_v_read, ((PyObject*)__pyx_t_6));
        __pyx_t_6 = 0;

        /* "aldy/indelpost/varaln.pyx":614
 *                 read
 *                 for read in reads
 *                 if is_quality_read(read, pos, quality_window, quality_threshold)             # <<<<<<<<<<<<<<
 *             ]
 * 
 */
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_is_quality_read); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 614, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_9 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_9genexpr15__pyx_v_read, __pyx_v_pos, __pyx_v_quality_window, __pyx_v_quality_threshold};
          __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L8_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_6);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_9genexpr15__pyx_v_read, __pyx_v_pos, __pyx_v_quality_window, __pyx_v_quality_threshold};
          __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L8_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_6);
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 614, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_8) {
            __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
          }
          __Pyx_INCREF(__pyx_9genexpr15__pyx_v_read);
          __Pyx_GIVEREF(__pyx_9genexpr15__pyx_v_read);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_9genexpr15__pyx_v_read);
          __Pyx_INCREF(__pyx_v_pos);
          __Pyx_GIVEREF(__pyx_v_pos);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_v_pos);
          __Pyx_INCREF(__pyx_v_quality_window);
          __Pyx_GIVEREF(__pyx_v_quality_window);
          PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_v_quality_window);
          __Pyx_INCREF(__pyx_v_quality_threshold);
          __Pyx_GIVEREF(__pyx_v_quality_threshold);
          PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_9, __pyx_v_quality_threshold);
          __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 614, __pyx_L8_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_4) {

          /* "aldy/indelpost/varaln.pyx":612
 *         if quality_window and quality_threshold:
 *             reads = [
 *                 read             # <<<<<<<<<<<<<<
 *                 for read in reads
 *                 if is_quality_read(read, pos, quality_window, quality_threshold)
 */
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_9genexpr15__pyx_v_read))) __PYX_ERR(0, 611, __pyx_L8_error)

          /* "aldy/indelpost/varaln.pyx":614
 *                 read
 *                 for read in reads
 *                 if is_quality_read(read, pos, quality_window, quality_threshold)             # <<<<<<<<<<<<<<
 *             ]
 * 
 */
        }

        /* "aldy/indelpost/varaln.pyx":613
 *             reads = [
 *                 read
 *                 for read in reads             # <<<<<<<<<<<<<<
 *                 if is_quality_read(read, pos, quality_window, quality_threshold)
 *             ]
 */
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_read); __pyx_9genexpr15__pyx_v_read = 0;
      goto __pyx_L12_exit_scope;
      __pyx_L8_error:;
      __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_read); __pyx_9genexpr15__pyx_v_read = 0;
      goto __pyx_L1_error;
      __pyx_L12_exit_scope:;
    } /* exit inner scope */
    __Pyx_DECREF_SET(__pyx_v_reads, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":610
 * 
 *         reads = self.__pileup
 *         if quality_window and quality_threshold:             # <<<<<<<<<<<<<<
 *             reads = [
 *                 read
 */
  }

  /* "aldy/indelpost/varaln.pyx":617
 *             ]
 * 
 *         if three_class:             # <<<<<<<<<<<<<<
 *             for read in reads:
 *                 read["is_locally_ref"] = is_locally_ref(read, pos)
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_three_class); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 617, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "aldy/indelpost/varaln.pyx":618
 * 
 *         if three_class:
 *             for read in reads:             # <<<<<<<<<<<<<<
 *                 read["is_locally_ref"] = is_locally_ref(read, pos)
 * 
 */
    if (unlikely(__pyx_v_reads == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 618, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_v_reads; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 618, __pyx_L1_error)
      #else
      __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 618, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_read, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "aldy/indelpost/varaln.pyx":619
 *         if three_class:
 *             for read in reads:
 *                 read["is_locally_ref"] = is_locally_ref(read, pos)             # <<<<<<<<<<<<<<
 * 
 *         fw_target = [
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_is_locally_ref); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 619, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_read, __pyx_v_pos};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_read, __pyx_v_pos};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 619, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_INCREF(__pyx_v_read);
        __Pyx_GIVEREF(__pyx_v_read);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_read);
        __Pyx_INCREF(__pyx_v_pos);
        __Pyx_GIVEREF(__pyx_v_pos);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_v_pos);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(__pyx_v_read == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 619, __pyx_L1_error)
      }
      if (unlikely(PyDict_SetItem(__pyx_v_read, __pyx_n_u_is_locally_ref, __pyx_t_1) < 0)) __PYX_ERR(0, 619, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "aldy/indelpost/varaln.pyx":618
 * 
 *         if three_class:
 *             for read in reads:             # <<<<<<<<<<<<<<
 *                 read["is_locally_ref"] = is_locally_ref(read, pos)
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":617
 *             ]
 * 
 *         if three_class:             # <<<<<<<<<<<<<<
 *             for read in reads:
 *                 read["is_locally_ref"] = is_locally_ref(read, pos)
 */
  }

  /* "aldy/indelpost/varaln.pyx":621
 *                 read["is_locally_ref"] = is_locally_ref(read, pos)
 * 
 *         fw_target = [             # <<<<<<<<<<<<<<
 *             read["read_name"]
 *             for read in reads
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 621, __pyx_L18_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "aldy/indelpost/varaln.pyx":623
 *         fw_target = [
 *             read["read_name"]
 *             for read in reads             # <<<<<<<<<<<<<<
 *             if read["is_target"] and not read["is_reverse"]
 *         ]
 */
    if (unlikely(__pyx_v_reads == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 623, __pyx_L18_error)
    }
    __pyx_t_1 = __pyx_v_reads; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 623, __pyx_L18_error)
      #else
      __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 623, __pyx_L18_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      if (!(likely(PyDict_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 623, __pyx_L18_error)
      __Pyx_XDECREF_SET(__pyx_9genexpr16__pyx_v_read, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "aldy/indelpost/varaln.pyx":624
 *             read["read_name"]
 *             for read in reads
 *             if read["is_target"] and not read["is_reverse"]             # <<<<<<<<<<<<<<
 *         ]
 *         rv_target = [
 */
      if (unlikely(__pyx_9genexpr16__pyx_v_read == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 624, __pyx_L18_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr16__pyx_v_read, __pyx_n_u_is_target); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 624, __pyx_L18_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 624, __pyx_L18_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (__pyx_t_5) {
      } else {
        __pyx_t_4 = __pyx_t_5;
        goto __pyx_L22_bool_binop_done;
      }
      if (unlikely(__pyx_9genexpr16__pyx_v_read == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 624, __pyx_L18_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr16__pyx_v_read, __pyx_n_u_is_reverse); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 624, __pyx_L18_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 624, __pyx_L18_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_11 = ((!__pyx_t_5) != 0);
      __pyx_t_4 = __pyx_t_11;
      __pyx_L22_bool_binop_done:;
      if (__pyx_t_4) {

        /* "aldy/indelpost/varaln.pyx":622
 * 
 *         fw_target = [
 *             read["read_name"]             # <<<<<<<<<<<<<<
 *             for read in reads
 *             if read["is_target"] and not read["is_reverse"]
 */
        if (unlikely(__pyx_9genexpr16__pyx_v_read == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 622, __pyx_L18_error)
        }
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr16__pyx_v_read, __pyx_n_u_read_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 622, __pyx_L18_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 621, __pyx_L18_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "aldy/indelpost/varaln.pyx":624
 *             read["read_name"]
 *             for read in reads
 *             if read["is_target"] and not read["is_reverse"]             # <<<<<<<<<<<<<<
 *         ]
 *         rv_target = [
 */
      }

      /* "aldy/indelpost/varaln.pyx":623
 *         fw_target = [
 *             read["read_name"]
 *             for read in reads             # <<<<<<<<<<<<<<
 *             if read["is_target"] and not read["is_reverse"]
 *         ]
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_read); __pyx_9genexpr16__pyx_v_read = 0;
    goto __pyx_L24_exit_scope;
    __pyx_L18_error:;
    __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_read); __pyx_9genexpr16__pyx_v_read = 0;
    goto __pyx_L1_error;
    __pyx_L24_exit_scope:;
  } /* exit inner scope */
  __pyx_v_fw_target = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":626
 *             if read["is_target"] and not read["is_reverse"]
 *         ]
 *         rv_target = [             # <<<<<<<<<<<<<<
 *             read["read_name"]
 *             for read in reads
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 626, __pyx_L27_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "aldy/indelpost/varaln.pyx":628
 *         rv_target = [
 *             read["read_name"]
 *             for read in reads             # <<<<<<<<<<<<<<
 *             if read["is_target"] and read["is_reverse"]
 *         ]
 */
    if (unlikely(__pyx_v_reads == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 628, __pyx_L27_error)
    }
    __pyx_t_1 = __pyx_v_reads; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 628, __pyx_L27_error)
      #else
      __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 628, __pyx_L27_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      if (!(likely(PyDict_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 628, __pyx_L27_error)
      __Pyx_XDECREF_SET(__pyx_9genexpr17__pyx_v_read, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "aldy/indelpost/varaln.pyx":629
 *             read["read_name"]
 *             for read in reads
 *             if read["is_target"] and read["is_reverse"]             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
      if (unlikely(__pyx_9genexpr17__pyx_v_read == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 629, __pyx_L27_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr17__pyx_v_read, __pyx_n_u_is_target); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 629, __pyx_L27_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 629, __pyx_L27_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (__pyx_t_11) {
      } else {
        __pyx_t_4 = __pyx_t_11;
        goto __pyx_L31_bool_binop_done;
      }
      if (unlikely(__pyx_9genexpr17__pyx_v_read == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 629, __pyx_L27_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr17__pyx_v_read, __pyx_n_u_is_reverse); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 629, __pyx_L27_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 629, __pyx_L27_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_4 = __pyx_t_11;
      __pyx_L31_bool_binop_done:;
      if (__pyx_t_4) {

        /* "aldy/indelpost/varaln.pyx":627
 *         ]
 *         rv_target = [
 *             read["read_name"]             # <<<<<<<<<<<<<<
 *             for read in reads
 *             if read["is_target"] and read["is_reverse"]
 */
        if (unlikely(__pyx_9genexpr17__pyx_v_read == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 627, __pyx_L27_error)
        }
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr17__pyx_v_read, __pyx_n_u_read_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 627, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 626, __pyx_L27_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "aldy/indelpost/varaln.pyx":629
 *             read["read_name"]
 *             for read in reads
 *             if read["is_target"] and read["is_reverse"]             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
      }

      /* "aldy/indelpost/varaln.pyx":628
 *         rv_target = [
 *             read["read_name"]
 *             for read in reads             # <<<<<<<<<<<<<<
 *             if read["is_target"] and read["is_reverse"]
 *         ]
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_read); __pyx_9genexpr17__pyx_v_read = 0;
    goto __pyx_L33_exit_scope;
    __pyx_L27_error:;
    __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_read); __pyx_9genexpr17__pyx_v_read = 0;
    goto __pyx_L1_error;
    __pyx_L33_exit_scope:;
  } /* exit inner scope */
  __pyx_v_rv_target = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":632
 *         ]
 * 
 *         fw_non_target = [             # <<<<<<<<<<<<<<
 *             read["read_name"]
 *             for read in reads
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 632, __pyx_L36_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "aldy/indelpost/varaln.pyx":634
 *         fw_non_target = [
 *             read["read_name"]
 *             for read in reads             # <<<<<<<<<<<<<<
 *             if count_as_non_target(read, pos, del_len, margin) and not read["is_reverse"]
 *         ]
 */
    if (unlikely(__pyx_v_reads == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 634, __pyx_L36_error)
    }
    __pyx_t_1 = __pyx_v_reads; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 634, __pyx_L36_error)
      #else
      __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 634, __pyx_L36_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      if (!(likely(PyDict_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 634, __pyx_L36_error)
      __Pyx_XDECREF_SET(__pyx_9genexpr18__pyx_v_read, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "aldy/indelpost/varaln.pyx":635
 *             read["read_name"]
 *             for read in reads
 *             if count_as_non_target(read, pos, del_len, margin) and not read["is_reverse"]             # <<<<<<<<<<<<<<
 *         ]
 *         rv_non_target = [
 */
      __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_v_pos); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 635, __pyx_L36_error)
      __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_margin); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 635, __pyx_L36_error)
      __pyx_t_11 = (__pyx_f_4aldy_9indelpost_6varaln_count_as_non_target(__pyx_9genexpr18__pyx_v_read, __pyx_t_9, __pyx_v_del_len, __pyx_t_12) != 0);
      if (__pyx_t_11) {
      } else {
        __pyx_t_4 = __pyx_t_11;
        goto __pyx_L40_bool_binop_done;
      }
      if (unlikely(__pyx_9genexpr18__pyx_v_read == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 635, __pyx_L36_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr18__pyx_v_read, __pyx_n_u_is_reverse); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 635, __pyx_L36_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 635, __pyx_L36_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_5 = ((!__pyx_t_11) != 0);
      __pyx_t_4 = __pyx_t_5;
      __pyx_L40_bool_binop_done:;
      if (__pyx_t_4) {

        /* "aldy/indelpost/varaln.pyx":633
 * 
 *         fw_non_target = [
 *             read["read_name"]             # <<<<<<<<<<<<<<
 *             for read in reads
 *             if count_as_non_target(read, pos, del_len, margin) and not read["is_reverse"]
 */
        if (unlikely(__pyx_9genexpr18__pyx_v_read == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 633, __pyx_L36_error)
        }
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr18__pyx_v_read, __pyx_n_u_read_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 633, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 632, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "aldy/indelpost/varaln.pyx":635
 *             read["read_name"]
 *             for read in reads
 *             if count_as_non_target(read, pos, del_len, margin) and not read["is_reverse"]             # <<<<<<<<<<<<<<
 *         ]
 *         rv_non_target = [
 */
      }

      /* "aldy/indelpost/varaln.pyx":634
 *         fw_non_target = [
 *             read["read_name"]
 *             for read in reads             # <<<<<<<<<<<<<<
 *             if count_as_non_target(read, pos, del_len, margin) and not read["is_reverse"]
 *         ]
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_read); __pyx_9genexpr18__pyx_v_read = 0;
    goto __pyx_L42_exit_scope;
    __pyx_L36_error:;
    __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_read); __pyx_9genexpr18__pyx_v_read = 0;
    goto __pyx_L1_error;
    __pyx_L42_exit_scope:;
  } /* exit inner scope */
  __pyx_v_fw_non_target = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":637
 *             if count_as_non_target(read, pos, del_len, margin) and not read["is_reverse"]
 *         ]
 *         rv_non_target = [             # <<<<<<<<<<<<<<
 *             read["read_name"]
 *             for read in reads
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 637, __pyx_L45_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "aldy/indelpost/varaln.pyx":639
 *         rv_non_target = [
 *             read["read_name"]
 *             for read in reads             # <<<<<<<<<<<<<<
 *             if count_as_non_target(read, pos, del_len, margin) and read["is_reverse"]
 *         ]
 */
    if (unlikely(__pyx_v_reads == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 639, __pyx_L45_error)
    }
    __pyx_t_1 = __pyx_v_reads; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 639, __pyx_L45_error)
      #else
      __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 639, __pyx_L45_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      if (!(likely(PyDict_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 639, __pyx_L45_error)
      __Pyx_XDECREF_SET(__pyx_9genexpr19__pyx_v_read, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "aldy/indelpost/varaln.pyx":640
 *             read["read_name"]
 *             for read in reads
 *             if count_as_non_target(read, pos, del_len, margin) and read["is_reverse"]             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
      __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_pos); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 640, __pyx_L45_error)
      __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_v_margin); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 640, __pyx_L45_error)
      __pyx_t_5 = (__pyx_f_4aldy_9indelpost_6varaln_count_as_non_target(__pyx_9genexpr19__pyx_v_read, __pyx_t_12, __pyx_v_del_len, __pyx_t_9) != 0);
      if (__pyx_t_5) {
      } else {
        __pyx_t_4 = __pyx_t_5;
        goto __pyx_L49_bool_binop_done;
      }
      if (unlikely(__pyx_9genexpr19__pyx_v_read == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 640, __pyx_L45_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr19__pyx_v_read, __pyx_n_u_is_reverse); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 640, __pyx_L45_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 640, __pyx_L45_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_4 = __pyx_t_5;
      __pyx_L49_bool_binop_done:;
      if (__pyx_t_4) {

        /* "aldy/indelpost/varaln.pyx":638
 *         ]
 *         rv_non_target = [
 *             read["read_name"]             # <<<<<<<<<<<<<<
 *             for read in reads
 *             if count_as_non_target(read, pos, del_len, margin) and read["is_reverse"]
 */
        if (unlikely(__pyx_9genexpr19__pyx_v_read == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 638, __pyx_L45_error)
        }
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr19__pyx_v_read, __pyx_n_u_read_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 638, __pyx_L45_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 637, __pyx_L45_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "aldy/indelpost/varaln.pyx":640
 *             read["read_name"]
 *             for read in reads
 *             if count_as_non_target(read, pos, del_len, margin) and read["is_reverse"]             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
      }

      /* "aldy/indelpost/varaln.pyx":639
 *         rv_non_target = [
 *             read["read_name"]
 *             for read in reads             # <<<<<<<<<<<<<<
 *             if count_as_non_target(read, pos, del_len, margin) and read["is_reverse"]
 *         ]
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_read); __pyx_9genexpr19__pyx_v_read = 0;
    goto __pyx_L51_exit_scope;
    __pyx_L45_error:;
    __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_read); __pyx_9genexpr19__pyx_v_read = 0;
    goto __pyx_L1_error;
    __pyx_L51_exit_scope:;
  } /* exit inner scope */
  __pyx_v_rv_non_target = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":643
 *         ]
 * 
 *         fw_target = set(fw_target)             # <<<<<<<<<<<<<<
 *         rw_target = set(rv_target)
 *         fwrv_target = fw_target | rw_target
 */
  __pyx_t_2 = PySet_New(__pyx_v_fw_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_fw_target, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":644
 * 
 *         fw_target = set(fw_target)
 *         rw_target = set(rv_target)             # <<<<<<<<<<<<<<
 *         fwrv_target = fw_target | rw_target
 * 
 */
  __pyx_t_2 = PySet_New(__pyx_v_rv_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_rw_target = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":645
 *         fw_target = set(fw_target)
 *         rw_target = set(rv_target)
 *         fwrv_target = fw_target | rw_target             # <<<<<<<<<<<<<<
 * 
 *         fw_non_target = set(fw_non_target) - fwrv_target
 */
  __pyx_t_2 = PyNumber_Or(__pyx_v_fw_target, __pyx_v_rw_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_fwrv_target = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":647
 *         fwrv_target = fw_target | rw_target
 * 
 *         fw_non_target = set(fw_non_target) - fwrv_target             # <<<<<<<<<<<<<<
 *         rv_non_target = set(rv_non_target) - fwrv_target
 * 
 */
  __pyx_t_2 = PySet_New(__pyx_v_fw_non_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Subtract(__pyx_t_2, __pyx_v_fwrv_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_fw_non_target, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":648
 * 
 *         fw_non_target = set(fw_non_target) - fwrv_target
 *         rv_non_target = set(rv_non_target) - fwrv_target             # <<<<<<<<<<<<<<
 * 
 *         est = self.__sample_factor if estimated_count else 1
 */
  __pyx_t_1 = PySet_New(__pyx_v_rv_non_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Subtract(__pyx_t_1, __pyx_v_fwrv_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_rv_non_target, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":650
 *         rv_non_target = set(rv_non_target) - fwrv_target
 * 
 *         est = self.__sample_factor if estimated_count else 1             # <<<<<<<<<<<<<<
 * 
 *         if three_class:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_estimated_count); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 650, __pyx_L1_error)
  if (__pyx_t_4) {
    __pyx_t_1 = PyFloat_FromDouble(__pyx_cur_scope->__pyx_v_self->__pyx___sample_factor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_t_1;
    __pyx_t_1 = 0;
  } else {
    __Pyx_INCREF(__pyx_int_1);
    __pyx_t_2 = __pyx_int_1;
  }
  __pyx_v_est = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":652
 *         est = self.__sample_factor if estimated_count else 1
 * 
 *         if three_class:             # <<<<<<<<<<<<<<
 *             fw_ref = [
 *                 read["read_name"]
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_three_class); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 652, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "aldy/indelpost/varaln.pyx":653
 * 
 *         if three_class:
 *             fw_ref = [             # <<<<<<<<<<<<<<
 *                 read["read_name"]
 *                 for read in reads if read["is_locally_ref"] and not read["is_reverse"] and read["read_name"] in fw_non_target
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 653, __pyx_L55_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "aldy/indelpost/varaln.pyx":655
 *             fw_ref = [
 *                 read["read_name"]
 *                 for read in reads if read["is_locally_ref"] and not read["is_reverse"] and read["read_name"] in fw_non_target             # <<<<<<<<<<<<<<
 *             ]
 * 
 */
      if (unlikely(__pyx_v_reads == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 655, __pyx_L55_error)
      }
      __pyx_t_1 = __pyx_v_reads; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
      for (;;) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 655, __pyx_L55_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 655, __pyx_L55_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        if (!(likely(PyDict_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 655, __pyx_L55_error)
        __Pyx_XDECREF_SET(__pyx_9genexpr20__pyx_v_read, ((PyObject*)__pyx_t_6));
        __pyx_t_6 = 0;
        if (unlikely(__pyx_9genexpr20__pyx_v_read == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 655, __pyx_L55_error)
        }
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr20__pyx_v_read, __pyx_n_u_is_locally_ref); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 655, __pyx_L55_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 655, __pyx_L55_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_5) {
        } else {
          __pyx_t_4 = __pyx_t_5;
          goto __pyx_L59_bool_binop_done;
        }
        if (unlikely(__pyx_9genexpr20__pyx_v_read == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 655, __pyx_L55_error)
        }
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr20__pyx_v_read, __pyx_n_u_is_reverse); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 655, __pyx_L55_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 655, __pyx_L55_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_11 = ((!__pyx_t_5) != 0);
        if (__pyx_t_11) {
        } else {
          __pyx_t_4 = __pyx_t_11;
          goto __pyx_L59_bool_binop_done;
        }
        if (unlikely(__pyx_9genexpr20__pyx_v_read == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 655, __pyx_L55_error)
        }
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr20__pyx_v_read, __pyx_n_u_read_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 655, __pyx_L55_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_11 = (__Pyx_PySequence_ContainsTF(__pyx_t_6, __pyx_v_fw_non_target, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 655, __pyx_L55_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_5 = (__pyx_t_11 != 0);
        __pyx_t_4 = __pyx_t_5;
        __pyx_L59_bool_binop_done:;
        if (__pyx_t_4) {

          /* "aldy/indelpost/varaln.pyx":654
 *         if three_class:
 *             fw_ref = [
 *                 read["read_name"]             # <<<<<<<<<<<<<<
 *                 for read in reads if read["is_locally_ref"] and not read["is_reverse"] and read["read_name"] in fw_non_target
 *             ]
 */
          if (unlikely(__pyx_9genexpr20__pyx_v_read == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 654, __pyx_L55_error)
          }
          __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr20__pyx_v_read, __pyx_n_u_read_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 654, __pyx_L55_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 653, __pyx_L55_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "aldy/indelpost/varaln.pyx":655
 *             fw_ref = [
 *                 read["read_name"]
 *                 for read in reads if read["is_locally_ref"] and not read["is_reverse"] and read["read_name"] in fw_non_target             # <<<<<<<<<<<<<<
 *             ]
 * 
 */
        }
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_read); __pyx_9genexpr20__pyx_v_read = 0;
      goto __pyx_L62_exit_scope;
      __pyx_L55_error:;
      __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_read); __pyx_9genexpr20__pyx_v_read = 0;
      goto __pyx_L1_error;
      __pyx_L62_exit_scope:;
    } /* exit inner scope */
    __pyx_v_fw_ref = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":658
 *             ]
 * 
 *             fw_ref = set(fw_ref)             # <<<<<<<<<<<<<<
 *             fw_non_ref_non_target = fw_non_target - fw_ref
 * 
 */
    __pyx_t_2 = PySet_New(__pyx_v_fw_ref); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 658, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_fw_ref, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":659
 * 
 *             fw_ref = set(fw_ref)
 *             fw_non_ref_non_target = fw_non_target - fw_ref             # <<<<<<<<<<<<<<
 * 
 *             rv_ref = [
 */
    __pyx_t_2 = PyNumber_Subtract(__pyx_v_fw_non_target, __pyx_v_fw_ref); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 659, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_fw_non_ref_non_target = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":661
 *             fw_non_ref_non_target = fw_non_target - fw_ref
 * 
 *             rv_ref = [             # <<<<<<<<<<<<<<
 *                 read["read_name"]
 *                 for read in reads if read["is_locally_ref"] and read["is_reverse"] and read["read_name"] in rv_non_target
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 661, __pyx_L65_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "aldy/indelpost/varaln.pyx":663
 *             rv_ref = [
 *                 read["read_name"]
 *                 for read in reads if read["is_locally_ref"] and read["is_reverse"] and read["read_name"] in rv_non_target             # <<<<<<<<<<<<<<
 *             ]
 *             rv_ref = set(rv_ref)
 */
      if (unlikely(__pyx_v_reads == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 663, __pyx_L65_error)
      }
      __pyx_t_1 = __pyx_v_reads; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
      for (;;) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 663, __pyx_L65_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 663, __pyx_L65_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        if (!(likely(PyDict_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 663, __pyx_L65_error)
        __Pyx_XDECREF_SET(__pyx_9genexpr21__pyx_v_read, ((PyObject*)__pyx_t_6));
        __pyx_t_6 = 0;
        if (unlikely(__pyx_9genexpr21__pyx_v_read == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 663, __pyx_L65_error)
        }
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr21__pyx_v_read, __pyx_n_u_is_locally_ref); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 663, __pyx_L65_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 663, __pyx_L65_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_5) {
        } else {
          __pyx_t_4 = __pyx_t_5;
          goto __pyx_L69_bool_binop_done;
        }
        if (unlikely(__pyx_9genexpr21__pyx_v_read == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 663, __pyx_L65_error)
        }
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr21__pyx_v_read, __pyx_n_u_is_reverse); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 663, __pyx_L65_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 663, __pyx_L65_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_5) {
        } else {
          __pyx_t_4 = __pyx_t_5;
          goto __pyx_L69_bool_binop_done;
        }
        if (unlikely(__pyx_9genexpr21__pyx_v_read == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 663, __pyx_L65_error)
        }
        __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr21__pyx_v_read, __pyx_n_u_read_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 663, __pyx_L65_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_t_6, __pyx_v_rv_non_target, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 663, __pyx_L65_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_11 = (__pyx_t_5 != 0);
        __pyx_t_4 = __pyx_t_11;
        __pyx_L69_bool_binop_done:;
        if (__pyx_t_4) {

          /* "aldy/indelpost/varaln.pyx":662
 * 
 *             rv_ref = [
 *                 read["read_name"]             # <<<<<<<<<<<<<<
 *                 for read in reads if read["is_locally_ref"] and read["is_reverse"] and read["read_name"] in rv_non_target
 *             ]
 */
          if (unlikely(__pyx_9genexpr21__pyx_v_read == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 662, __pyx_L65_error)
          }
          __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr21__pyx_v_read, __pyx_n_u_read_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 662, __pyx_L65_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 661, __pyx_L65_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "aldy/indelpost/varaln.pyx":663
 *             rv_ref = [
 *                 read["read_name"]
 *                 for read in reads if read["is_locally_ref"] and read["is_reverse"] and read["read_name"] in rv_non_target             # <<<<<<<<<<<<<<
 *             ]
 *             rv_ref = set(rv_ref)
 */
        }
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_read); __pyx_9genexpr21__pyx_v_read = 0;
      goto __pyx_L72_exit_scope;
      __pyx_L65_error:;
      __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_read); __pyx_9genexpr21__pyx_v_read = 0;
      goto __pyx_L1_error;
      __pyx_L72_exit_scope:;
    } /* exit inner scope */
    __pyx_v_rv_ref = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":665
 *                 for read in reads if read["is_locally_ref"] and read["is_reverse"] and read["read_name"] in rv_non_target
 *             ]
 *             rv_ref = set(rv_ref)             # <<<<<<<<<<<<<<
 *             rv_non_ref_non_target = rv_non_target - rv_ref
 * 
 */
    __pyx_t_2 = PySet_New(__pyx_v_rv_ref); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_rv_ref, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":666
 *             ]
 *             rv_ref = set(rv_ref)
 *             rv_non_ref_non_target = rv_non_target - rv_ref             # <<<<<<<<<<<<<<
 * 
 *             if fwrv:
 */
    __pyx_t_2 = PyNumber_Subtract(__pyx_v_rv_non_target, __pyx_v_rv_ref); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_rv_non_ref_non_target = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":668
 *             rv_non_ref_non_target = rv_non_target - rv_ref
 * 
 *             if fwrv:             # <<<<<<<<<<<<<<
 *                 return (
 *                     (
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_fwrv); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 668, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "aldy/indelpost/varaln.pyx":669
 * 
 *             if fwrv:
 *                 return (             # <<<<<<<<<<<<<<
 *                     (
 *                         int(len(fw_ref) * est),
 */
      __Pyx_XDECREF(__pyx_r);

      /* "aldy/indelpost/varaln.pyx":671
 *                 return (
 *                     (
 *                         int(len(fw_ref) * est),             # <<<<<<<<<<<<<<
 *                         int(len(rv_ref) * est),
 *                     ),
 */
      __pyx_t_3 = PyObject_Length(__pyx_v_fw_ref); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 671, __pyx_L1_error)
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyNumber_Multiply(__pyx_t_2, __pyx_v_est); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "aldy/indelpost/varaln.pyx":672
 *                     (
 *                         int(len(fw_ref) * est),
 *                         int(len(rv_ref) * est),             # <<<<<<<<<<<<<<
 *                     ),
 *                     (
 */
      __pyx_t_3 = PyObject_Length(__pyx_v_rv_ref); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 672, __pyx_L1_error)
      __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 672, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PyNumber_Multiply(__pyx_t_1, __pyx_v_est); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 672, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 672, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "aldy/indelpost/varaln.pyx":671
 *                 return (
 *                     (
 *                         int(len(fw_ref) * est),             # <<<<<<<<<<<<<<
 *                         int(len(rv_ref) * est),
 *                     ),
 */
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
      __pyx_t_2 = 0;
      __pyx_t_1 = 0;

      /* "aldy/indelpost/varaln.pyx":675
 *                     ),
 *                     (
 *                         int(len(fw_non_ref_non_target) * est),             # <<<<<<<<<<<<<<
 *                         int(len(rv_non_ref_non_target) * est),
 *                     ),
 */
      __pyx_t_3 = PyObject_Length(__pyx_v_fw_non_ref_non_target); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 675, __pyx_L1_error)
      __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = PyNumber_Multiply(__pyx_t_1, __pyx_v_est); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "aldy/indelpost/varaln.pyx":676
 *                     (
 *                         int(len(fw_non_ref_non_target) * est),
 *                         int(len(rv_non_ref_non_target) * est),             # <<<<<<<<<<<<<<
 *                     ),
 *                     (
 */
      __pyx_t_3 = PyObject_Length(__pyx_v_rv_non_ref_non_target); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 676, __pyx_L1_error)
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 676, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = PyNumber_Multiply(__pyx_t_2, __pyx_v_est); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 676, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 676, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "aldy/indelpost/varaln.pyx":675
 *                     ),
 *                     (
 *                         int(len(fw_non_ref_non_target) * est),             # <<<<<<<<<<<<<<
 *                         int(len(rv_non_ref_non_target) * est),
 *                     ),
 */
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_2);
      __pyx_t_1 = 0;
      __pyx_t_2 = 0;

      /* "aldy/indelpost/varaln.pyx":679
 *                     ),
 *                     (
 *                         int(len(fw_target) * est),             # <<<<<<<<<<<<<<
 *                         int(len(rv_target) * est),
 *                     ),
 */
      __pyx_t_3 = PyObject_Length(__pyx_v_fw_target); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 679, __pyx_L1_error)
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyNumber_Multiply(__pyx_t_2, __pyx_v_est); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "aldy/indelpost/varaln.pyx":680
 *                     (
 *                         int(len(fw_target) * est),
 *                         int(len(rv_target) * est),             # <<<<<<<<<<<<<<
 *                     ),
 *                 )
 */
      __pyx_t_3 = PyList_GET_SIZE(__pyx_v_rv_target); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 680, __pyx_L1_error)
      __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 680, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = PyNumber_Multiply(__pyx_t_1, __pyx_v_est); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 680, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 680, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "aldy/indelpost/varaln.pyx":679
 *                     ),
 *                     (
 *                         int(len(fw_target) * est),             # <<<<<<<<<<<<<<
 *                         int(len(rv_target) * est),
 *                     ),
 */
      __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_1);
      __pyx_t_2 = 0;
      __pyx_t_1 = 0;

      /* "aldy/indelpost/varaln.pyx":670
 *             if fwrv:
 *                 return (
 *                     (             # <<<<<<<<<<<<<<
 *                         int(len(fw_ref) * est),
 *                         int(len(rv_ref) * est),
 */
      __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 670, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "aldy/indelpost/varaln.pyx":668
 *             rv_non_ref_non_target = rv_non_target - rv_ref
 * 
 *             if fwrv:             # <<<<<<<<<<<<<<
 *                 return (
 *                     (
 */
    }

    /* "aldy/indelpost/varaln.pyx":685
 * 
 *             else:
 *                 if by_fragment:             # <<<<<<<<<<<<<<
 *                     fwrv_ref = len(fw_ref | rv_ref)
 *                     fwrv_non_ref_non_target = len(fw_non_ref_non_target | rv_non_ref_non_target)
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_by_fragment); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 685, __pyx_L1_error)
      if (__pyx_t_4) {

        /* "aldy/indelpost/varaln.pyx":686
 *             else:
 *                 if by_fragment:
 *                     fwrv_ref = len(fw_ref | rv_ref)             # <<<<<<<<<<<<<<
 *                     fwrv_non_ref_non_target = len(fw_non_ref_non_target | rv_non_ref_non_target)
 *                     fwrv_target = len(fw_target | rw_target)
 */
        __pyx_t_1 = PyNumber_Or(__pyx_v_fw_ref, __pyx_v_rv_ref); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 686, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 686, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v_fwrv_ref = __pyx_t_3;

        /* "aldy/indelpost/varaln.pyx":687
 *                 if by_fragment:
 *                     fwrv_ref = len(fw_ref | rv_ref)
 *                     fwrv_non_ref_non_target = len(fw_non_ref_non_target | rv_non_ref_non_target)             # <<<<<<<<<<<<<<
 *                     fwrv_target = len(fw_target | rw_target)
 *                 else:
 */
        __pyx_t_1 = PyNumber_Or(__pyx_v_fw_non_ref_non_target, __pyx_v_rv_non_ref_non_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 687, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 687, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v_fwrv_non_ref_non_target = __pyx_t_3;

        /* "aldy/indelpost/varaln.pyx":688
 *                     fwrv_ref = len(fw_ref | rv_ref)
 *                     fwrv_non_ref_non_target = len(fw_non_ref_non_target | rv_non_ref_non_target)
 *                     fwrv_target = len(fw_target | rw_target)             # <<<<<<<<<<<<<<
 *                 else:
 *                     fwrv_ref = len(fw_ref) + len(rv_ref)
 */
        __pyx_t_1 = PyNumber_Or(__pyx_v_fw_target, __pyx_v_rw_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 688, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 688, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 688, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF_SET(__pyx_v_fwrv_target, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "aldy/indelpost/varaln.pyx":685
 * 
 *             else:
 *                 if by_fragment:             # <<<<<<<<<<<<<<
 *                     fwrv_ref = len(fw_ref | rv_ref)
 *                     fwrv_non_ref_non_target = len(fw_non_ref_non_target | rv_non_ref_non_target)
 */
        goto __pyx_L74;
      }

      /* "aldy/indelpost/varaln.pyx":690
 *                     fwrv_target = len(fw_target | rw_target)
 *                 else:
 *                     fwrv_ref = len(fw_ref) + len(rv_ref)             # <<<<<<<<<<<<<<
 *                     fwrv_non_ref_non_target =  len(fw_non_ref_non_target) + len(rv_non_ref_non_target)
 *                     fwrv_target = len(fw_target) + len(rw_target)
 */
      /*else*/ {
        __pyx_t_3 = PyObject_Length(__pyx_v_fw_ref); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 690, __pyx_L1_error)
        __pyx_t_13 = PyObject_Length(__pyx_v_rv_ref); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 690, __pyx_L1_error)
        __pyx_v_fwrv_ref = (__pyx_t_3 + __pyx_t_13);

        /* "aldy/indelpost/varaln.pyx":691
 *                 else:
 *                     fwrv_ref = len(fw_ref) + len(rv_ref)
 *                     fwrv_non_ref_non_target =  len(fw_non_ref_non_target) + len(rv_non_ref_non_target)             # <<<<<<<<<<<<<<
 *                     fwrv_target = len(fw_target) + len(rw_target)
 * 
 */
        __pyx_t_13 = PyObject_Length(__pyx_v_fw_non_ref_non_target); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 691, __pyx_L1_error)
        __pyx_t_3 = PyObject_Length(__pyx_v_rv_non_ref_non_target); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 691, __pyx_L1_error)
        __pyx_v_fwrv_non_ref_non_target = (__pyx_t_13 + __pyx_t_3);

        /* "aldy/indelpost/varaln.pyx":692
 *                     fwrv_ref = len(fw_ref) + len(rv_ref)
 *                     fwrv_non_ref_non_target =  len(fw_non_ref_non_target) + len(rv_non_ref_non_target)
 *                     fwrv_target = len(fw_target) + len(rw_target)             # <<<<<<<<<<<<<<
 * 
 *                 return (int(fwrv_ref), int(fwrv_non_ref_non_target), int(fwrv_target))
 */
        __pyx_t_3 = PyObject_Length(__pyx_v_fw_target); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 692, __pyx_L1_error)
        __pyx_t_13 = PySet_GET_SIZE(__pyx_v_rw_target); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 692, __pyx_L1_error)
        __pyx_t_1 = PyInt_FromSsize_t((__pyx_t_3 + __pyx_t_13)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 692, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF_SET(__pyx_v_fwrv_target, __pyx_t_1);
        __pyx_t_1 = 0;
      }
      __pyx_L74:;

      /* "aldy/indelpost/varaln.pyx":694
 *                     fwrv_target = len(fw_target) + len(rw_target)
 * 
 *                 return (int(fwrv_ref), int(fwrv_non_ref_non_target), int(fwrv_target))             # <<<<<<<<<<<<<<
 * 
 *         if fwrv:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_fwrv_ref); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyInt_Type)), __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_fwrv_non_ref_non_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyInt_Type)), __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_v_fwrv_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
      __pyx_t_7 = 0;
      __pyx_t_10 = 0;
      __pyx_t_1 = 0;
      __pyx_r = __pyx_t_6;
      __pyx_t_6 = 0;
      goto __pyx_L0;
    }

    /* "aldy/indelpost/varaln.pyx":652
 *         est = self.__sample_factor if estimated_count else 1
 * 
 *         if three_class:             # <<<<<<<<<<<<<<
 *             fw_ref = [
 *                 read["read_name"]
 */
  }

  /* "aldy/indelpost/varaln.pyx":696
 *                 return (int(fwrv_ref), int(fwrv_non_ref_non_target), int(fwrv_target))
 * 
 *         if fwrv:             # <<<<<<<<<<<<<<
 *             return (
 *                 (
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_fwrv); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 696, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "aldy/indelpost/varaln.pyx":697
 * 
 *         if fwrv:
 *             return (             # <<<<<<<<<<<<<<
 *                 (
 *                     int(len(fw_non_target) * est),
 */
    __Pyx_XDECREF(__pyx_r);

    /* "aldy/indelpost/varaln.pyx":699
 *             return (
 *                 (
 *                     int(len(fw_non_target) * est),             # <<<<<<<<<<<<<<
 *                     int(len(rv_non_target) * est),
 *                 ),
 */
    __pyx_t_13 = PyObject_Length(__pyx_v_fw_non_target); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 699, __pyx_L1_error)
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyNumber_Multiply(__pyx_t_6, __pyx_v_est); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyNumber_Int(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "aldy/indelpost/varaln.pyx":700
 *                 (
 *                     int(len(fw_non_target) * est),
 *                     int(len(rv_non_target) * est),             # <<<<<<<<<<<<<<
 *                 ),
 *                 (
 */
    __pyx_t_13 = PyObject_Length(__pyx_v_rv_non_target); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 700, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = PyNumber_Multiply(__pyx_t_1, __pyx_v_est); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "aldy/indelpost/varaln.pyx":699
 *             return (
 *                 (
 *                     int(len(fw_non_target) * est),             # <<<<<<<<<<<<<<
 *                     int(len(rv_non_target) * est),
 *                 ),
 */
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;

    /* "aldy/indelpost/varaln.pyx":703
 *                 ),
 *                 (
 *                     int(len(fw_target) * est),             # <<<<<<<<<<<<<<
 *                     int(len(rv_target) * est),
 *                 ),
 */
    __pyx_t_13 = PyObject_Length(__pyx_v_fw_target); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 703, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyNumber_Multiply(__pyx_t_1, __pyx_v_est); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "aldy/indelpost/varaln.pyx":704
 *                 (
 *                     int(len(fw_target) * est),
 *                     int(len(rv_target) * est),             # <<<<<<<<<<<<<<
 *                 ),
 *             )
 */
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_rv_target); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 704, __pyx_L1_error)
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = PyNumber_Multiply(__pyx_t_6, __pyx_v_est); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyNumber_Int(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "aldy/indelpost/varaln.pyx":703
 *                 ),
 *                 (
 *                     int(len(fw_target) * est),             # <<<<<<<<<<<<<<
 *                     int(len(rv_target) * est),
 *                 ),
 */
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);
    __pyx_t_1 = 0;
    __pyx_t_6 = 0;

    /* "aldy/indelpost/varaln.pyx":698
 *         if fwrv:
 *             return (
 *                 (             # <<<<<<<<<<<<<<
 *                     int(len(fw_non_target) * est),
 *                     int(len(rv_non_target) * est),
 */
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_7);
    __pyx_t_10 = 0;
    __pyx_t_7 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":696
 *                 return (int(fwrv_ref), int(fwrv_non_ref_non_target), int(fwrv_target))
 * 
 *         if fwrv:             # <<<<<<<<<<<<<<
 *             return (
 *                 (
 */
  }

  /* "aldy/indelpost/varaln.pyx":708
 *             )
 *         else:
 *             if by_fragment:             # <<<<<<<<<<<<<<
 *                 fwrv_non_target = len(fw_non_target | rv_non_target)
 *                 fwrv_target = len(fw_target | rw_target)
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_by_fragment); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 708, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "aldy/indelpost/varaln.pyx":709
 *         else:
 *             if by_fragment:
 *                 fwrv_non_target = len(fw_non_target | rv_non_target)             # <<<<<<<<<<<<<<
 *                 fwrv_target = len(fw_target | rw_target)
 *             else:
 */
      __pyx_t_6 = PyNumber_Or(__pyx_v_fw_non_target, __pyx_v_rv_non_target); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 709, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_13 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 709, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 709, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_v_fwrv_non_target = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "aldy/indelpost/varaln.pyx":710
 *             if by_fragment:
 *                 fwrv_non_target = len(fw_non_target | rv_non_target)
 *                 fwrv_target = len(fw_target | rw_target)             # <<<<<<<<<<<<<<
 *             else:
 *                 fwrv_non_target = len(fw_non_target) + len(rv_non_target)
 */
      __pyx_t_6 = PyNumber_Or(__pyx_v_fw_target, __pyx_v_rw_target); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 710, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_13 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 710, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 710, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_fwrv_target, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "aldy/indelpost/varaln.pyx":708
 *             )
 *         else:
 *             if by_fragment:             # <<<<<<<<<<<<<<
 *                 fwrv_non_target = len(fw_non_target | rv_non_target)
 *                 fwrv_target = len(fw_target | rw_target)
 */
      goto __pyx_L76;
    }

    /* "aldy/indelpost/varaln.pyx":712
 *                 fwrv_target = len(fw_target | rw_target)
 *             else:
 *                 fwrv_non_target = len(fw_non_target) + len(rv_non_target)             # <<<<<<<<<<<<<<
 *                 fwrv_target = len(fw_target) + len(rw_target)
 * 
 */
    /*else*/ {
      __pyx_t_13 = PyObject_Length(__pyx_v_fw_non_target); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 712, __pyx_L1_error)
      __pyx_t_3 = PyObject_Length(__pyx_v_rv_non_target); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 712, __pyx_L1_error)
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_13 + __pyx_t_3)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_v_fwrv_non_target = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "aldy/indelpost/varaln.pyx":713
 *             else:
 *                 fwrv_non_target = len(fw_non_target) + len(rv_non_target)
 *                 fwrv_target = len(fw_target) + len(rw_target)             # <<<<<<<<<<<<<<
 * 
 *             return (
 */
      __pyx_t_3 = PyObject_Length(__pyx_v_fw_target); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 713, __pyx_L1_error)
      __pyx_t_13 = PySet_GET_SIZE(__pyx_v_rw_target); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 713, __pyx_L1_error)
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_3 + __pyx_t_13)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 713, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_fwrv_target, __pyx_t_6);
      __pyx_t_6 = 0;
    }
    __pyx_L76:;

    /* "aldy/indelpost/varaln.pyx":715
 *                 fwrv_target = len(fw_target) + len(rw_target)
 * 
 *             return (             # <<<<<<<<<<<<<<
 *                 int(fwrv_non_target * est),
 *                 int(fwrv_target * est),
 */
    __Pyx_XDECREF(__pyx_r);

    /* "aldy/indelpost/varaln.pyx":716
 * 
 *             return (
 *                 int(fwrv_non_target * est),             # <<<<<<<<<<<<<<
 *                 int(fwrv_target * est),
 *             )
 */
    __pyx_t_6 = PyNumber_Multiply(__pyx_v_fwrv_non_target, __pyx_v_est); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 716, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyNumber_Int(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 716, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "aldy/indelpost/varaln.pyx":717
 *             return (
 *                 int(fwrv_non_target * est),
 *                 int(fwrv_target * est),             # <<<<<<<<<<<<<<
 *             )
 * 
 */
    __pyx_t_6 = PyNumber_Multiply(__pyx_v_fwrv_target, __pyx_v_est); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_10 = __Pyx_PyNumber_Int(__pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "aldy/indelpost/varaln.pyx":716
 * 
 *             return (
 *                 int(fwrv_non_target * est),             # <<<<<<<<<<<<<<
 *                 int(fwrv_target * est),
 *             )
 */
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 716, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_10);
    __pyx_t_7 = 0;
    __pyx_t_10 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* "aldy/indelpost/varaln.pyx":576
 * 
 * 
 *     def count_alleles(             # <<<<<<<<<<<<<<
 *         self, fwrv=False, by_fragment=False, three_class=False, estimated_count=False, quality_window=None, quality_threshold=None
 *     ):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.count_alleles", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_read);
  __Pyx_XDECREF(__pyx_v_pos);
  __Pyx_XDECREF(__pyx_v_r_pos);
  __Pyx_XDECREF(__pyx_v_margin);
  __Pyx_XDECREF(__pyx_v_reads);
  __Pyx_XDECREF(__pyx_v_fw_target);
  __Pyx_XDECREF(__pyx_v_rv_target);
  __Pyx_XDECREF(__pyx_v_fw_non_target);
  __Pyx_XDECREF(__pyx_v_rv_non_target);
  __Pyx_XDECREF(__pyx_v_rw_target);
  __Pyx_XDECREF(__pyx_v_fwrv_target);
  __Pyx_XDECREF(__pyx_v_est);
  __Pyx_XDECREF(__pyx_v_fw_ref);
  __Pyx_XDECREF(__pyx_v_fw_non_ref_non_target);
  __Pyx_XDECREF(__pyx_v_rv_ref);
  __Pyx_XDECREF(__pyx_v_rv_non_ref_non_target);
  __Pyx_XDECREF(__pyx_v_fwrv_non_target);
  __Pyx_XDECREF(__pyx_gb_4aldy_9indelpost_6varaln_16VariantAlignment_13count_alleles_2generator1);
  __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_read);
  __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_read);
  __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_read);
  __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_read);
  __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_read);
  __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_read);
  __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_read);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":721
 * 
 * 
 *     def phase(             # <<<<<<<<<<<<<<
 *         self,
 *         how="local",
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_15phase(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_16VariantAlignment_14phase[] = "VariantAlignment.phase(self, how=u'local', local_threshold=20, longest_common_substring_threshold=15, indel_repeat_threshold=None, mutation_density_threshold=0.05)\nreturns a :class:`~indelpost.Variant` object represeting a phased target indel.\n        :class:`~indelpost.NullVariant` is returned if no target is found.\n        Phasing is limited to the exon where the target indel is located for RNA-Seq.\n        Refer to :meth:`~indelpost.Contig.get_phasables()` to retrieve phasable variants across exons.\n\n        Parameters\n        ----------\n        how : string\n            - \"local\" (default) phasing by recursive elimination of longest common substrings (LCS) and locality  scoring (see local_threshold).\n            - \"greedy\" phase all phasable events.\n            - \"complex\" phasing by the \"local\" option + exclusivity check. The exclusivity check removes phasable events that are also observed in non-target reads such as nearby homozygous germline events.\n        local_threshold : integer\n            local (and complex) phasing method checks if phasable SNVs are locally clustered around the indel.\n            For i-th base apart from indel, the score gains one if mismatch and, if match, loses\n            1*min(j/local_thresh, 1) where j = i for short indels (< 10-nt) and 0.6*i for longer indels.\n        longest_common_substring_threshold : integer\n            removes common substrings between the reference and contig assembled from target reads that are longer than longest_common_substring_threshold (default 15).\n        indel_repeat_threshold : integer\n            do not phase indels that repeat more than indel_repeat_threshold (default None)\n        mutation_density_threshold : float\n            do not phase if the pileup contains too many mutations (possibly error-prone dirty region).\n            In non-target reads, #non-ref bases/#all bases > mutation_density_threshold (default 0.05)\n        ";
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_15phase(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_how = 0;
  PyObject *__pyx_v_local_threshold = 0;
  PyObject *__pyx_v_longest_common_substring_threshold = 0;
  PyObject *__pyx_v_indel_repeat_threshold = 0;
  PyObject *__pyx_v_mutation_density_threshold = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("phase (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_how,&__pyx_n_s_local_threshold,&__pyx_n_s_longest_common_substring_thresho,&__pyx_n_s_indel_repeat_threshold,&__pyx_n_s_mutation_density_threshold,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[0] = ((PyObject *)__pyx_n_u_local);
    values[1] = ((PyObject *)__pyx_int_20);
    values[2] = ((PyObject *)__pyx_int_15);

    /* "aldy/indelpost/varaln.pyx":726
 *         local_threshold=20,
 *         longest_common_substring_threshold=15,
 *         indel_repeat_threshold=None,             # <<<<<<<<<<<<<<
 *         mutation_density_threshold=0.05,
 *     ):
 */
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)__pyx_float_0_05);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_how);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_local_threshold);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_longest_common_substring_thresho);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indel_repeat_threshold);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mutation_density_threshold);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "phase") < 0)) __PYX_ERR(0, 721, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_how = values[0];
    __pyx_v_local_threshold = values[1];
    __pyx_v_longest_common_substring_threshold = values[2];
    __pyx_v_indel_repeat_threshold = values[3];
    __pyx_v_mutation_density_threshold = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("phase", 0, 0, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 721, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.phase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_14phase(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self), __pyx_v_how, __pyx_v_local_threshold, __pyx_v_longest_common_substring_threshold, __pyx_v_indel_repeat_threshold, __pyx_v_mutation_density_threshold);

  /* "aldy/indelpost/varaln.pyx":721
 * 
 * 
 *     def phase(             # <<<<<<<<<<<<<<
 *         self,
 *         how="local",
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_14phase(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self, PyObject *__pyx_v_how, PyObject *__pyx_v_local_threshold, PyObject *__pyx_v_longest_common_substring_threshold, PyObject *__pyx_v_indel_repeat_threshold, PyObject *__pyx_v_mutation_density_threshold) {
  int __pyx_v_hard;
  int __pyx_v_to_complex;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("phase", 0);
  __Pyx_INCREF(__pyx_v_indel_repeat_threshold);

  /* "aldy/indelpost/varaln.pyx":752
 *             In non-target reads, #non-ref bases/#all bases > mutation_density_threshold (default 0.05)
 *         """
 *         if how == "complex":             # <<<<<<<<<<<<<<
 *             hard, to_complex = False, True
 *         elif how == "greedy":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_how, __pyx_n_u_complex, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 752, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "aldy/indelpost/varaln.pyx":753
 *         """
 *         if how == "complex":
 *             hard, to_complex = False, True             # <<<<<<<<<<<<<<
 *         elif how == "greedy":
 *             hard, to_complex = True, False
 */
    __pyx_t_1 = 0;
    __pyx_t_2 = 1;
    __pyx_v_hard = __pyx_t_1;
    __pyx_v_to_complex = __pyx_t_2;

    /* "aldy/indelpost/varaln.pyx":752
 *             In non-target reads, #non-ref bases/#all bases > mutation_density_threshold (default 0.05)
 *         """
 *         if how == "complex":             # <<<<<<<<<<<<<<
 *             hard, to_complex = False, True
 *         elif how == "greedy":
 */
    goto __pyx_L3;
  }

  /* "aldy/indelpost/varaln.pyx":754
 *         if how == "complex":
 *             hard, to_complex = False, True
 *         elif how == "greedy":             # <<<<<<<<<<<<<<
 *             hard, to_complex = True, False
 *         elif how == "local":
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_how, __pyx_n_u_greedy, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 754, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "aldy/indelpost/varaln.pyx":755
 *             hard, to_complex = False, True
 *         elif how == "greedy":
 *             hard, to_complex = True, False             # <<<<<<<<<<<<<<
 *         elif how == "local":
 *             hard, to_complex = False, False
 */
    __pyx_t_2 = 1;
    __pyx_t_1 = 0;
    __pyx_v_hard = __pyx_t_2;
    __pyx_v_to_complex = __pyx_t_1;

    /* "aldy/indelpost/varaln.pyx":754
 *         if how == "complex":
 *             hard, to_complex = False, True
 *         elif how == "greedy":             # <<<<<<<<<<<<<<
 *             hard, to_complex = True, False
 *         elif how == "local":
 */
    goto __pyx_L3;
  }

  /* "aldy/indelpost/varaln.pyx":756
 *         elif how == "greedy":
 *             hard, to_complex = True, False
 *         elif how == "local":             # <<<<<<<<<<<<<<
 *             hard, to_complex = False, False
 *         else:
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_how, __pyx_n_u_local, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 756, __pyx_L1_error)
  if (likely(__pyx_t_1)) {

    /* "aldy/indelpost/varaln.pyx":757
 *             hard, to_complex = True, False
 *         elif how == "local":
 *             hard, to_complex = False, False             # <<<<<<<<<<<<<<
 *         else:
 *             raise Exception("phasing stragety must be either of local, greedy, complex")
 */
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_v_hard = __pyx_t_1;
    __pyx_v_to_complex = __pyx_t_2;

    /* "aldy/indelpost/varaln.pyx":756
 *         elif how == "greedy":
 *             hard, to_complex = True, False
 *         elif how == "local":             # <<<<<<<<<<<<<<
 *             hard, to_complex = False, False
 *         else:
 */
    goto __pyx_L3;
  }

  /* "aldy/indelpost/varaln.pyx":759
 *             hard, to_complex = False, False
 *         else:
 *             raise Exception("phasing stragety must be either of local, greedy, complex")             # <<<<<<<<<<<<<<
 * 
 *         if indel_repeat_threshold is None:
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 759, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 759, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "aldy/indelpost/varaln.pyx":761
 *             raise Exception("phasing stragety must be either of local, greedy, complex")
 * 
 *         if indel_repeat_threshold is None:             # <<<<<<<<<<<<<<
 *             indel_repeat_threshold = np.inf
 * 
 */
  __pyx_t_2 = (__pyx_v_indel_repeat_threshold == Py_None);
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "aldy/indelpost/varaln.pyx":762
 * 
 *         if indel_repeat_threshold is None:
 *             indel_repeat_threshold = np.inf             # <<<<<<<<<<<<<<
 * 
 *         return phase_nearby_variants(
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 762, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_inf); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 762, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_indel_repeat_threshold, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "aldy/indelpost/varaln.pyx":761
 *             raise Exception("phasing stragety must be either of local, greedy, complex")
 * 
 *         if indel_repeat_threshold is None:             # <<<<<<<<<<<<<<
 *             indel_repeat_threshold = np.inf
 * 
 */
  }

  /* "aldy/indelpost/varaln.pyx":764
 *             indel_repeat_threshold = np.inf
 * 
 *         return phase_nearby_variants(             # <<<<<<<<<<<<<<
 *             self.__target,
 *             self.contig,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_phase_nearby_variants); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "aldy/indelpost/varaln.pyx":768
 *             self.contig,
 *             self.__pileup,
 *             self.basequalthresh,             # <<<<<<<<<<<<<<
 *             local_threshold,
 *             longest_common_substring_threshold,
 */
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->basequalthresh); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "aldy/indelpost/varaln.pyx":773
 *             indel_repeat_threshold,
 *             mutation_density_threshold,
 *             hard,             # <<<<<<<<<<<<<<
 *             to_complex
 *         )
 */
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_hard); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "aldy/indelpost/varaln.pyx":774
 *             mutation_density_threshold,
 *             hard,
 *             to_complex             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_to_complex); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[11] = {__pyx_t_8, ((PyObject *)__pyx_v_self->__pyx___target), ((PyObject *)__pyx_v_self->contig), __pyx_v_self->__pyx___pileup, __pyx_t_5, __pyx_v_local_threshold, __pyx_v_longest_common_substring_threshold, __pyx_v_indel_repeat_threshold, __pyx_v_mutation_density_threshold, __pyx_t_6, __pyx_t_7};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 10+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 764, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[11] = {__pyx_t_8, ((PyObject *)__pyx_v_self->__pyx___target), ((PyObject *)__pyx_v_self->contig), __pyx_v_self->__pyx___pileup, __pyx_t_5, __pyx_v_local_threshold, __pyx_v_longest_common_substring_threshold, __pyx_v_indel_repeat_threshold, __pyx_v_mutation_density_threshold, __pyx_t_6, __pyx_t_7};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 10+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 764, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_10 = PyTuple_New(10+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 764, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_8) {
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx___target));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx___target));
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, ((PyObject *)__pyx_v_self->__pyx___target));
    __Pyx_INCREF(((PyObject *)__pyx_v_self->contig));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self->contig));
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, ((PyObject *)__pyx_v_self->contig));
    __Pyx_INCREF(__pyx_v_self->__pyx___pileup);
    __Pyx_GIVEREF(__pyx_v_self->__pyx___pileup);
    PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_v_self->__pyx___pileup);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_9, __pyx_t_5);
    __Pyx_INCREF(__pyx_v_local_threshold);
    __Pyx_GIVEREF(__pyx_v_local_threshold);
    PyTuple_SET_ITEM(__pyx_t_10, 4+__pyx_t_9, __pyx_v_local_threshold);
    __Pyx_INCREF(__pyx_v_longest_common_substring_threshold);
    __Pyx_GIVEREF(__pyx_v_longest_common_substring_threshold);
    PyTuple_SET_ITEM(__pyx_t_10, 5+__pyx_t_9, __pyx_v_longest_common_substring_threshold);
    __Pyx_INCREF(__pyx_v_indel_repeat_threshold);
    __Pyx_GIVEREF(__pyx_v_indel_repeat_threshold);
    PyTuple_SET_ITEM(__pyx_t_10, 6+__pyx_t_9, __pyx_v_indel_repeat_threshold);
    __Pyx_INCREF(__pyx_v_mutation_density_threshold);
    __Pyx_GIVEREF(__pyx_v_mutation_density_threshold);
    PyTuple_SET_ITEM(__pyx_t_10, 7+__pyx_t_9, __pyx_v_mutation_density_threshold);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_10, 8+__pyx_t_9, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_10, 9+__pyx_t_9, __pyx_t_7);
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_7 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 764, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "aldy/indelpost/varaln.pyx":721
 * 
 * 
 *     def phase(             # <<<<<<<<<<<<<<
 *         self,
 *         how="local",
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.phase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_indel_repeat_threshold);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pxd":11
 * cdef class VariantAlignment:
 *     cdef Variant target, __target, second_target
 *     cdef readonly AlignmentFile bam             # <<<<<<<<<<<<<<
 *     cdef int window, retarget_window, mapqthresh
 *     cdef int downsamplethresh, basequalthresh, match_score, _observed_pos
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_3bam_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_3bam_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_3bam___get__(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_3bam___get__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->bam));
  __pyx_r = ((PyObject *)__pyx_v_self->bam);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pxd":18
 *     cdef bint exclude_duplicates, exact_match_for_shiftable, auto_adjust_extension_penalty, no_realignment
 *     cdef list __pileup
 *     cdef readonly is_spurious_overhang, is_complex_input             # <<<<<<<<<<<<<<
 *     cdef readonly Contig contig
 *     cdef readonly UnsplicedLocalReference unspliced_local_reference
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_20is_spurious_overhang_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_20is_spurious_overhang_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_20is_spurious_overhang___get__(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_20is_spurious_overhang___get__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->is_spurious_overhang);
  __pyx_r = __pyx_v_self->is_spurious_overhang;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_16is_complex_input_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_16is_complex_input_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_16is_complex_input___get__(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_16is_complex_input___get__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->is_complex_input);
  __pyx_r = __pyx_v_self->is_complex_input;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pxd":19
 *     cdef list __pileup
 *     cdef readonly is_spurious_overhang, is_complex_input
 *     cdef readonly Contig contig             # <<<<<<<<<<<<<<
 *     cdef readonly UnsplicedLocalReference unspliced_local_reference
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_6contig_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_6contig_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_6contig___get__(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_6contig___get__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->contig));
  __pyx_r = ((PyObject *)__pyx_v_self->contig);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pxd":20
 *     cdef readonly is_spurious_overhang, is_complex_input
 *     cdef readonly Contig contig
 *     cdef readonly UnsplicedLocalReference unspliced_local_reference             # <<<<<<<<<<<<<<
 * 
 *     cdef __parse_pileup(self, Contig contig=*, bint retargeted=*, bint skip_read_end_check=*)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_25unspliced_local_reference_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_25unspliced_local_reference_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_25unspliced_local_reference___get__(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_25unspliced_local_reference___get__(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->unspliced_local_reference));
  __pyx_r = ((PyObject *)__pyx_v_self->unspliced_local_reference);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_16VariantAlignment_16__reduce_cython__[] = "VariantAlignment.__reduce_cython__(self)";
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_16__reduce_cython__(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_16VariantAlignment_18__setstate_cython__[] = "VariantAlignment.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_18__setstate_cython__(((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_16VariantAlignment_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("aldy.indelpost.varaln.VariantAlignment.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":778
 * 
 * 
 * def is_quality_read(read, pos, qualitywindow, qualitythresh):             # <<<<<<<<<<<<<<
 * 
 *     try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_1is_quality_read(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_is_quality_read[] = "is_quality_read(read, pos, qualitywindow, qualitythresh)";
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6varaln_1is_quality_read = {"is_quality_read", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6varaln_1is_quality_read, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4aldy_9indelpost_6varaln_is_quality_read};
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_1is_quality_read(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_read = 0;
  PyObject *__pyx_v_pos = 0;
  PyObject *__pyx_v_qualitywindow = 0;
  PyObject *__pyx_v_qualitythresh = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_quality_read (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_read,&__pyx_n_s_pos,&__pyx_n_s_qualitywindow,&__pyx_n_s_qualitythresh,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_read)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pos)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_quality_read", 1, 4, 4, 1); __PYX_ERR(0, 778, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_qualitywindow)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_quality_read", 1, 4, 4, 2); __PYX_ERR(0, 778, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_qualitythresh)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_quality_read", 1, 4, 4, 3); __PYX_ERR(0, 778, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "is_quality_read") < 0)) __PYX_ERR(0, 778, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_read = values[0];
    __pyx_v_pos = values[1];
    __pyx_v_qualitywindow = values[2];
    __pyx_v_qualitythresh = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_quality_read", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 778, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.is_quality_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_is_quality_read(__pyx_self, __pyx_v_read, __pyx_v_pos, __pyx_v_qualitywindow, __pyx_v_qualitythresh);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_is_quality_read(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_pos, PyObject *__pyx_v_qualitywindow, PyObject *__pyx_v_qualitythresh) {
  PyObject *__pyx_v_lt_qual = NULL;
  PyObject *__pyx_v_rt_qual = NULL;
  PyObject *__pyx_v_lt_median = NULL;
  PyObject *__pyx_v_rt_median = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_quality_read", 0);

  /* "aldy/indelpost/varaln.pyx":780
 * def is_quality_read(read, pos, qualitywindow, qualitythresh):
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         lt_qual, rt_qual = read["lt_qual"], read["rt_qual"]
 *     except:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "aldy/indelpost/varaln.pyx":781
 * 
 *     try:
 *         lt_qual, rt_qual = read["lt_qual"], read["rt_qual"]             # <<<<<<<<<<<<<<
 *     except:
 *         lt_qual, rt_qual = split(
 */
      __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_lt_qual); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 781, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_rt_qual); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 781, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_lt_qual = __pyx_t_4;
      __pyx_t_4 = 0;
      __pyx_v_rt_qual = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "aldy/indelpost/varaln.pyx":780
 * def is_quality_read(read, pos, qualitywindow, qualitythresh):
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         lt_qual, rt_qual = read["lt_qual"], read["rt_qual"]
 *     except:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "aldy/indelpost/varaln.pyx":782
 *     try:
 *         lt_qual, rt_qual = read["lt_qual"], read["rt_qual"]
 *     except:             # <<<<<<<<<<<<<<
 *         lt_qual, rt_qual = split(
 *             read["read_qual"],
 */
    /*except:*/ {
      __Pyx_AddTraceback("aldy.indelpost.varaln.is_quality_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_6) < 0) __PYX_ERR(0, 782, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_6);

      /* "aldy/indelpost/varaln.pyx":784
 *     except:
 *         lt_qual, rt_qual = split(
 *             read["read_qual"],             # <<<<<<<<<<<<<<
 *             read["cigar_string"],
 *             pos,
 */
      __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_qual); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 784, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "aldy/indelpost/varaln.pyx":785
 *         lt_qual, rt_qual = split(
 *             read["read_qual"],
 *             read["cigar_string"],             # <<<<<<<<<<<<<<
 *             pos,
 *             read["read_start"],
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 785, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_8)->tp_name), 0))) __PYX_ERR(0, 785, __pyx_L5_except_error)

      /* "aldy/indelpost/varaln.pyx":786
 *             read["read_qual"],
 *             read["cigar_string"],
 *             pos,             # <<<<<<<<<<<<<<
 *             read["read_start"],
 *             is_for_ref=False,
 */
      __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_v_pos); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L5_except_error)

      /* "aldy/indelpost/varaln.pyx":787
 *             read["cigar_string"],
 *             pos,
 *             read["read_start"],             # <<<<<<<<<<<<<<
 *             is_for_ref=False,
 *             reverse=False,
 */
      __pyx_t_10 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_start); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 787, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_t_10); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 787, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "aldy/indelpost/varaln.pyx":783
 *         lt_qual, rt_qual = read["lt_qual"], read["rt_qual"]
 *     except:
 *         lt_qual, rt_qual = split(             # <<<<<<<<<<<<<<
 *             read["read_qual"],
 *             read["cigar_string"],
 */
      __pyx_t_10 = __pyx_f_4aldy_9indelpost_9utilities_split(__pyx_t_7, ((PyObject*)__pyx_t_8), __pyx_t_9, __pyx_t_11, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 783, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (likely(__pyx_t_10 != Py_None)) {
        PyObject* sequence = __pyx_t_10;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 783, __pyx_L5_except_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        #else
        __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 783, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 783, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      } else {
        __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 783, __pyx_L5_except_error)
      }
      __Pyx_XDECREF_SET(__pyx_v_lt_qual, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_v_rt_qual, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L4_exception_handled;
    }
    __pyx_L5_except_error:;

    /* "aldy/indelpost/varaln.pyx":780
 * def is_quality_read(read, pos, qualitywindow, qualitythresh):
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         lt_qual, rt_qual = read["lt_qual"], read["rt_qual"]
 *     except:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "aldy/indelpost/varaln.pyx":792
 *         )
 * 
 *     if lt_qual and rt_qual:             # <<<<<<<<<<<<<<
 *         lt_median = np.median(lt_qual[-min(len(lt_qual), qualitywindow) :])
 *         rt_median = np.median(rt_qual[: min(len(rt_qual), qualitywindow)])
 */
  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_lt_qual); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 792, __pyx_L1_error)
  if (__pyx_t_13) {
  } else {
    __pyx_t_12 = __pyx_t_13;
    goto __pyx_L12_bool_binop_done;
  }
  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_rt_qual); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 792, __pyx_L1_error)
  __pyx_t_12 = __pyx_t_13;
  __pyx_L12_bool_binop_done:;
  if (__pyx_t_12) {

    /* "aldy/indelpost/varaln.pyx":793
 * 
 *     if lt_qual and rt_qual:
 *         lt_median = np.median(lt_qual[-min(len(lt_qual), qualitywindow) :])             # <<<<<<<<<<<<<<
 *         rt_median = np.median(rt_qual[: min(len(rt_qual), qualitywindow)])
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_v_qualitywindow);
    __pyx_t_4 = __pyx_v_qualitywindow;
    __pyx_t_14 = PyObject_Length(__pyx_v_lt_qual); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 793, __pyx_L1_error)
    __pyx_t_7 = PyInt_FromSsize_t(__pyx_t_14); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_4, __pyx_t_7, Py_LT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_12) {
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_10 = __pyx_t_4;
    } else {
      __pyx_t_8 = PyInt_FromSsize_t(__pyx_t_14); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 793, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_10 = __pyx_t_8;
      __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Negative(__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_GetSlice(__pyx_v_lt_qual, 0, 0, &__pyx_t_4, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_lt_median = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "aldy/indelpost/varaln.pyx":794
 *     if lt_qual and rt_qual:
 *         lt_median = np.median(lt_qual[-min(len(lt_qual), qualitywindow) :])
 *         rt_median = np.median(rt_qual[: min(len(rt_qual), qualitywindow)])             # <<<<<<<<<<<<<<
 * 
 *         return lt_median > qualitythresh and rt_median > qualitythresh
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_median); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_v_qualitywindow);
    __pyx_t_5 = __pyx_v_qualitywindow;
    __pyx_t_14 = PyObject_Length(__pyx_v_rt_qual); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 794, __pyx_L1_error)
    __pyx_t_8 = PyInt_FromSsize_t(__pyx_t_14); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = PyObject_RichCompare(__pyx_t_5, __pyx_t_8, Py_LT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_12) {
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_4 = __pyx_t_5;
    } else {
      __pyx_t_7 = PyInt_FromSsize_t(__pyx_t_14); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 794, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = __pyx_t_7;
      __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_rt_qual, 0, 0, NULL, &__pyx_t_4, NULL, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_v_rt_median = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "aldy/indelpost/varaln.pyx":796
 *         rt_median = np.median(rt_qual[: min(len(rt_qual), qualitywindow)])
 * 
 *         return lt_median > qualitythresh and rt_median > qualitythresh             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = PyObject_RichCompare(__pyx_v_lt_median, __pyx_v_qualitythresh, Py_GT); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 796, __pyx_L1_error)
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 796, __pyx_L1_error)
    if (__pyx_t_12) {
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_10);
      __pyx_t_6 = __pyx_t_10;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_10 = PyObject_RichCompare(__pyx_v_rt_median, __pyx_v_qualitythresh, Py_GT); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_INCREF(__pyx_t_10);
    __pyx_t_6 = __pyx_t_10;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_L14_bool_binop_done:;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":792
 *         )
 * 
 *     if lt_qual and rt_qual:             # <<<<<<<<<<<<<<
 *         lt_median = np.median(lt_qual[-min(len(lt_qual), qualitywindow) :])
 *         rt_median = np.median(rt_qual[: min(len(rt_qual), qualitywindow)])
 */
  }

  /* "aldy/indelpost/varaln.pyx":778
 * 
 * 
 * def is_quality_read(read, pos, qualitywindow, qualitythresh):             # <<<<<<<<<<<<<<
 * 
 *     try:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("aldy.indelpost.varaln.is_quality_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_lt_qual);
  __Pyx_XDECREF(__pyx_v_rt_qual);
  __Pyx_XDECREF(__pyx_v_lt_median);
  __Pyx_XDECREF(__pyx_v_rt_median);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":799
 * 
 * 
 * def is_locally_ref(read, pos):             # <<<<<<<<<<<<<<
 *     if read["is_reference_seq"]:
 *         return True
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_3is_locally_ref(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_2is_locally_ref[] = "is_locally_ref(read, pos)";
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6varaln_3is_locally_ref = {"is_locally_ref", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6varaln_3is_locally_ref, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4aldy_9indelpost_6varaln_2is_locally_ref};
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_3is_locally_ref(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_read = 0;
  PyObject *__pyx_v_pos = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_locally_ref (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_read,&__pyx_n_s_pos,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_read)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pos)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_locally_ref", 1, 2, 2, 1); __PYX_ERR(0, 799, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "is_locally_ref") < 0)) __PYX_ERR(0, 799, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_read = values[0];
    __pyx_v_pos = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_locally_ref", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 799, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.is_locally_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_2is_locally_ref(__pyx_self, __pyx_v_read, __pyx_v_pos);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_2is_locally_ref(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_pos) {
  PyObject *__pyx_v_lt_seq = NULL;
  PyObject *__pyx_v_rt_seq = NULL;
  PyObject *__pyx_v_lt_ref = NULL;
  PyObject *__pyx_v_rt_ref = NULL;
  Py_ssize_t __pyx_v_lt_seq_len;
  Py_ssize_t __pyx_v_lt_ref_len;
  PyObject *__pyx_v_lt_len = NULL;
  Py_ssize_t __pyx_v_rt_seq_len;
  Py_ssize_t __pyx_v_rt_ref_len;
  Py_ssize_t __pyx_v_rt_len;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  long __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_locally_ref", 0);

  /* "aldy/indelpost/varaln.pyx":800
 * 
 * def is_locally_ref(read, pos):
 *     if read["is_reference_seq"]:             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_is_reference_seq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "aldy/indelpost/varaln.pyx":801
 * def is_locally_ref(read, pos):
 *     if read["is_reference_seq"]:
 *         return True             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":800
 * 
 * def is_locally_ref(read, pos):
 *     if read["is_reference_seq"]:             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  }

  /* "aldy/indelpost/varaln.pyx":804
 * 
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         lt_seq, rt_seq = read["lt_seq"], read["rt_seq"]
 *     except:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "aldy/indelpost/varaln.pyx":805
 * 
 *     try:
 *         lt_seq, rt_seq = read["lt_seq"], read["rt_seq"]             # <<<<<<<<<<<<<<
 *     except:
 *         lt_seq, rt_seq = split(read["read_seq"],
 */
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_lt_seq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 805, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_rt_seq); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 805, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_v_lt_seq = __pyx_t_1;
      __pyx_t_1 = 0;
      __pyx_v_rt_seq = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "aldy/indelpost/varaln.pyx":804
 * 
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         lt_seq, rt_seq = read["lt_seq"], read["rt_seq"]
 *     except:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "aldy/indelpost/varaln.pyx":806
 *     try:
 *         lt_seq, rt_seq = read["lt_seq"], read["rt_seq"]
 *     except:             # <<<<<<<<<<<<<<
 *         lt_seq, rt_seq = split(read["read_seq"],
 *             read["cigar_string"],
 */
    /*except:*/ {
      __Pyx_AddTraceback("aldy.indelpost.varaln.is_locally_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_1, &__pyx_t_7) < 0) __PYX_ERR(0, 806, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_7);

      /* "aldy/indelpost/varaln.pyx":807
 *         lt_seq, rt_seq = read["lt_seq"], read["rt_seq"]
 *     except:
 *         lt_seq, rt_seq = split(read["read_seq"],             # <<<<<<<<<<<<<<
 *             read["cigar_string"],
 *             pos,
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_seq); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 807, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "aldy/indelpost/varaln.pyx":808
 *     except:
 *         lt_seq, rt_seq = split(read["read_seq"],
 *             read["cigar_string"],             # <<<<<<<<<<<<<<
 *             pos,
 *             read["read_start"],
 */
      __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 808, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_9))||((__pyx_t_9) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_9)->tp_name), 0))) __PYX_ERR(0, 808, __pyx_L6_except_error)

      /* "aldy/indelpost/varaln.pyx":809
 *         lt_seq, rt_seq = split(read["read_seq"],
 *             read["cigar_string"],
 *             pos,             # <<<<<<<<<<<<<<
 *             read["read_start"],
 *             is_for_ref=False,
 */
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_v_pos); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 809, __pyx_L6_except_error)

      /* "aldy/indelpost/varaln.pyx":810
 *             read["cigar_string"],
 *             pos,
 *             read["read_start"],             # <<<<<<<<<<<<<<
 *             is_for_ref=False,
 *             reverse=False,
 */
      __pyx_t_11 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_start); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_11); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 810, __pyx_L6_except_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "aldy/indelpost/varaln.pyx":807
 *         lt_seq, rt_seq = read["lt_seq"], read["rt_seq"]
 *     except:
 *         lt_seq, rt_seq = split(read["read_seq"],             # <<<<<<<<<<<<<<
 *             read["cigar_string"],
 *             pos,
 */
      __pyx_t_11 = __pyx_f_4aldy_9indelpost_9utilities_split(__pyx_t_8, ((PyObject*)__pyx_t_9), __pyx_t_10, __pyx_t_12, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 807, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (likely(__pyx_t_11 != Py_None)) {
        PyObject* sequence = __pyx_t_11;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 807, __pyx_L6_except_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 807, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 807, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      } else {
        __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 807, __pyx_L6_except_error)
      }
      __Pyx_XDECREF_SET(__pyx_v_lt_seq, __pyx_t_9);
      __pyx_t_9 = 0;
      __Pyx_XDECREF_SET(__pyx_v_rt_seq, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L5_exception_handled;
    }
    __pyx_L6_except_error:;

    /* "aldy/indelpost/varaln.pyx":804
 * 
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         lt_seq, rt_seq = read["lt_seq"], read["rt_seq"]
 *     except:
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    __pyx_L9_try_end:;
  }

  /* "aldy/indelpost/varaln.pyx":815
 *         )
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         lt_ref, rt_ref = read["lt_ref"], read["rt_ref"]
 *     except:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "aldy/indelpost/varaln.pyx":816
 * 
 *     try:
 *         lt_ref, rt_ref = read["lt_ref"], read["rt_ref"]             # <<<<<<<<<<<<<<
 *     except:
 *         lt_ref, rt_ref = split(read["ref_seq"],
 */
      __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_lt_ref); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 816, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_rt_ref); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 816, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_lt_ref = __pyx_t_7;
      __pyx_t_7 = 0;
      __pyx_v_rt_ref = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "aldy/indelpost/varaln.pyx":815
 *         )
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         lt_ref, rt_ref = read["lt_ref"], read["rt_ref"]
 *     except:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L17_try_end;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "aldy/indelpost/varaln.pyx":817
 *     try:
 *         lt_ref, rt_ref = read["lt_ref"], read["rt_ref"]
 *     except:             # <<<<<<<<<<<<<<
 *         lt_ref, rt_ref = split(read["ref_seq"],
 *             read["cigar_string"],
 */
    /*except:*/ {
      __Pyx_AddTraceback("aldy.indelpost.varaln.is_locally_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 817, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_6);

      /* "aldy/indelpost/varaln.pyx":818
 *         lt_ref, rt_ref = read["lt_ref"], read["rt_ref"]
 *     except:
 *         lt_ref, rt_ref = split(read["ref_seq"],             # <<<<<<<<<<<<<<
 *             read["cigar_string"],
 *             pos,
 */
      __pyx_t_11 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_ref_seq); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 818, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_11);

      /* "aldy/indelpost/varaln.pyx":819
 *     except:
 *         lt_ref, rt_ref = split(read["ref_seq"],
 *             read["cigar_string"],             # <<<<<<<<<<<<<<
 *             pos,
 *             read["aln_start"],
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 819, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_8)->tp_name), 0))) __PYX_ERR(0, 819, __pyx_L14_except_error)

      /* "aldy/indelpost/varaln.pyx":820
 *         lt_ref, rt_ref = split(read["ref_seq"],
 *             read["cigar_string"],
 *             pos,             # <<<<<<<<<<<<<<
 *             read["aln_start"],
 *             is_for_ref=True,
 */
      __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_pos); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 820, __pyx_L14_except_error)

      /* "aldy/indelpost/varaln.pyx":821
 *             read["cigar_string"],
 *             pos,
 *             read["aln_start"],             # <<<<<<<<<<<<<<
 *             is_for_ref=True,
 *             reverse=False,
 */
      __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 821, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_9); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 821, __pyx_L14_except_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "aldy/indelpost/varaln.pyx":818
 *         lt_ref, rt_ref = read["lt_ref"], read["rt_ref"]
 *     except:
 *         lt_ref, rt_ref = split(read["ref_seq"],             # <<<<<<<<<<<<<<
 *             read["cigar_string"],
 *             pos,
 */
      __pyx_t_9 = __pyx_f_4aldy_9indelpost_9utilities_split(__pyx_t_11, ((PyObject*)__pyx_t_8), __pyx_t_12, __pyx_t_10, 1, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 818, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (likely(__pyx_t_9 != Py_None)) {
        PyObject* sequence = __pyx_t_9;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 818, __pyx_L14_except_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_11 = PyTuple_GET_ITEM(sequence, 1); 
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_11);
        #else
        __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 818, __pyx_L14_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_11 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 818, __pyx_L14_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        #endif
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 818, __pyx_L14_except_error)
      }
      __Pyx_XDECREF_SET(__pyx_v_lt_ref, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_v_rt_ref, __pyx_t_11);
      __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L13_exception_handled;
    }
    __pyx_L14_except_error:;

    /* "aldy/indelpost/varaln.pyx":815
 *         )
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         lt_ref, rt_ref = read["lt_ref"], read["rt_ref"]
 *     except:
 */
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L13_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    __pyx_L17_try_end:;
  }

  /* "aldy/indelpost/varaln.pyx":828
 * 
 * 
 *     lt_seq_len = len(lt_seq)             # <<<<<<<<<<<<<<
 *     lt_ref_len = len(lt_ref)
 *     #clipped
 */
  __pyx_t_13 = PyObject_Length(__pyx_v_lt_seq); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 828, __pyx_L1_error)
  __pyx_v_lt_seq_len = __pyx_t_13;

  /* "aldy/indelpost/varaln.pyx":829
 * 
 *     lt_seq_len = len(lt_seq)
 *     lt_ref_len = len(lt_ref)             # <<<<<<<<<<<<<<
 *     #clipped
 *     if not lt_ref_len:
 */
  __pyx_t_13 = PyObject_Length(__pyx_v_lt_ref); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 829, __pyx_L1_error)
  __pyx_v_lt_ref_len = __pyx_t_13;

  /* "aldy/indelpost/varaln.pyx":831
 *     lt_ref_len = len(lt_ref)
 *     #clipped
 *     if not lt_ref_len:             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_lt_ref_len != 0)) != 0);
  if (__pyx_t_2) {

    /* "aldy/indelpost/varaln.pyx":832
 *     #clipped
 *     if not lt_ref_len:
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     lt_len = min(5, lt_seq_len, lt_ref_len)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":831
 *     lt_ref_len = len(lt_ref)
 *     #clipped
 *     if not lt_ref_len:             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "aldy/indelpost/varaln.pyx":834
 *         return False
 * 
 *     lt_len = min(5, lt_seq_len, lt_ref_len)             # <<<<<<<<<<<<<<
 * 
 *     rt_seq_len = len(rt_seq)
 */
  __pyx_t_13 = __pyx_v_lt_seq_len;
  __pyx_t_14 = __pyx_v_lt_ref_len;
  __pyx_t_15 = 5;
  if (((__pyx_t_13 < __pyx_t_15) != 0)) {
    __pyx_t_16 = __pyx_t_13;
  } else {
    __pyx_t_16 = __pyx_t_15;
  }
  __pyx_t_17 = __pyx_t_16;
  if (((__pyx_t_14 < __pyx_t_17) != 0)) {
    __pyx_t_16 = __pyx_t_14;
  } else {
    __pyx_t_16 = __pyx_t_17;
  }
  __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_16); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_lt_len = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "aldy/indelpost/varaln.pyx":836
 *     lt_len = min(5, lt_seq_len, lt_ref_len)
 * 
 *     rt_seq_len = len(rt_seq)             # <<<<<<<<<<<<<<
 *     rt_ref_len = len(rt_ref)
 *     if not rt_ref_len:
 */
  __pyx_t_16 = PyObject_Length(__pyx_v_rt_seq); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 836, __pyx_L1_error)
  __pyx_v_rt_seq_len = __pyx_t_16;

  /* "aldy/indelpost/varaln.pyx":837
 * 
 *     rt_seq_len = len(rt_seq)
 *     rt_ref_len = len(rt_ref)             # <<<<<<<<<<<<<<
 *     if not rt_ref_len:
 *         return False
 */
  __pyx_t_16 = PyObject_Length(__pyx_v_rt_ref); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 837, __pyx_L1_error)
  __pyx_v_rt_ref_len = __pyx_t_16;

  /* "aldy/indelpost/varaln.pyx":838
 *     rt_seq_len = len(rt_seq)
 *     rt_ref_len = len(rt_ref)
 *     if not rt_ref_len:             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_rt_ref_len != 0)) != 0);
  if (__pyx_t_2) {

    /* "aldy/indelpost/varaln.pyx":839
 *     rt_ref_len = len(rt_ref)
 *     if not rt_ref_len:
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     rt_len = min(5, rt_seq_len, rt_ref_len)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":838
 *     rt_seq_len = len(rt_seq)
 *     rt_ref_len = len(rt_ref)
 *     if not rt_ref_len:             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "aldy/indelpost/varaln.pyx":841
 *         return False
 * 
 *     rt_len = min(5, rt_seq_len, rt_ref_len)             # <<<<<<<<<<<<<<
 * 
 *     if lt_seq[-lt_len:] == lt_ref[-lt_len:] and rt_seq[:rt_len] == rt_ref[:rt_len]:
 */
  __pyx_t_16 = __pyx_v_rt_seq_len;
  __pyx_t_13 = __pyx_v_rt_ref_len;
  __pyx_t_15 = 5;
  if (((__pyx_t_16 < __pyx_t_15) != 0)) {
    __pyx_t_14 = __pyx_t_16;
  } else {
    __pyx_t_14 = __pyx_t_15;
  }
  __pyx_t_17 = __pyx_t_14;
  if (((__pyx_t_13 < __pyx_t_17) != 0)) {
    __pyx_t_14 = __pyx_t_13;
  } else {
    __pyx_t_14 = __pyx_t_17;
  }
  __pyx_v_rt_len = __pyx_t_14;

  /* "aldy/indelpost/varaln.pyx":843
 *     rt_len = min(5, rt_seq_len, rt_ref_len)
 * 
 *     if lt_seq[-lt_len:] == lt_ref[-lt_len:] and rt_seq[:rt_len] == rt_ref[:rt_len]:             # <<<<<<<<<<<<<<
 *         return True
 *     else:
 */
  __pyx_t_6 = PyNumber_Negative(__pyx_v_lt_len); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetSlice(__pyx_v_lt_seq, 0, 0, &__pyx_t_6, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyNumber_Negative(__pyx_v_lt_len); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_lt_ref, 0, 0, &__pyx_t_6, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyObject_RichCompare(__pyx_t_7, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_18 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_18 < 0)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_18) {
  } else {
    __pyx_t_2 = __pyx_t_18;
    goto __pyx_L23_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_PyObject_GetSlice(__pyx_v_rt_seq, 0, __pyx_v_rt_len, NULL, NULL, NULL, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_rt_ref, 0, __pyx_v_rt_len, NULL, NULL, NULL, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyObject_RichCompare(__pyx_t_6, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_18 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_18 < 0)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_2 = __pyx_t_18;
  __pyx_L23_bool_binop_done:;
  if (__pyx_t_2) {

    /* "aldy/indelpost/varaln.pyx":844
 * 
 *     if lt_seq[-lt_len:] == lt_ref[-lt_len:] and rt_seq[:rt_len] == rt_ref[:rt_len]:
 *         return True             # <<<<<<<<<<<<<<
 *     else:
 *         return False
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":843
 *     rt_len = min(5, rt_seq_len, rt_ref_len)
 * 
 *     if lt_seq[-lt_len:] == lt_ref[-lt_len:] and rt_seq[:rt_len] == rt_ref[:rt_len]:             # <<<<<<<<<<<<<<
 *         return True
 *     else:
 */
  }

  /* "aldy/indelpost/varaln.pyx":846
 *         return True
 *     else:
 *         return False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;
  }

  /* "aldy/indelpost/varaln.pyx":799
 * 
 * 
 * def is_locally_ref(read, pos):             # <<<<<<<<<<<<<<
 *     if read["is_reference_seq"]:
 *         return True
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("aldy.indelpost.varaln.is_locally_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_lt_seq);
  __Pyx_XDECREF(__pyx_v_rt_seq);
  __Pyx_XDECREF(__pyx_v_lt_ref);
  __Pyx_XDECREF(__pyx_v_rt_ref);
  __Pyx_XDECREF(__pyx_v_lt_len);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":849
 * 
 * 
 * cdef bint count_as_non_target(dict read, int pos, int del_len, int margin):             # <<<<<<<<<<<<<<
 *     if read["is_target"]:
 *         return False
 */

static int __pyx_f_4aldy_9indelpost_6varaln_count_as_non_target(PyObject *__pyx_v_read, int __pyx_v_pos, int __pyx_v_del_len, int __pyx_v_margin) {
  int __pyx_v_aln_start;
  int __pyx_v_aln_end;
  PyObject *__pyx_v_covering_subread = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_as_non_target", 0);

  /* "aldy/indelpost/varaln.pyx":850
 * 
 * cdef bint count_as_non_target(dict read, int pos, int del_len, int margin):
 *     if read["is_target"]:             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  if (unlikely(__pyx_v_read == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 850, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_read, __pyx_n_u_is_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "aldy/indelpost/varaln.pyx":851
 * cdef bint count_as_non_target(dict read, int pos, int del_len, int margin):
 *     if read["is_target"]:
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cdef int aln_start = read["aln_start"]
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":850
 * 
 * cdef bint count_as_non_target(dict read, int pos, int del_len, int margin):
 *     if read["is_target"]:             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "aldy/indelpost/varaln.pyx":853
 *         return False
 * 
 *     cdef int aln_start = read["aln_start"]             # <<<<<<<<<<<<<<
 *     cdef int aln_end = read["aln_end"]
 * 
 */
  if (unlikely(__pyx_v_read == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 853, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_read, __pyx_n_u_aln_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_aln_start = __pyx_t_3;

  /* "aldy/indelpost/varaln.pyx":854
 * 
 *     cdef int aln_start = read["aln_start"]
 *     cdef int aln_end = read["aln_end"]             # <<<<<<<<<<<<<<
 * 
 *     # undetermined reads
 */
  if (unlikely(__pyx_v_read == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 854, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_read, __pyx_n_u_aln_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_aln_end = __pyx_t_3;

  /* "aldy/indelpost/varaln.pyx":857
 * 
 *     # undetermined reads
 *     if read.get("undetermined", False):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  if (unlikely(__pyx_v_read == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
    __PYX_ERR(0, 857, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_v_read, __pyx_n_u_undetermined, Py_False); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "aldy/indelpost/varaln.pyx":858
 *     # undetermined reads
 *     if read.get("undetermined", False):
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     if read["is_covering"]:
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":857
 * 
 *     # undetermined reads
 *     if read.get("undetermined", False):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "aldy/indelpost/varaln.pyx":860
 *         return False
 * 
 *     if read["is_covering"]:             # <<<<<<<<<<<<<<
 *         covering_subread = read["covering_subread"]
 *         if covering_subread[1] <= pos + margin:
 */
  if (unlikely(__pyx_v_read == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 860, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_read, __pyx_n_u_is_covering); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 860, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "aldy/indelpost/varaln.pyx":861
 * 
 *     if read["is_covering"]:
 *         covering_subread = read["covering_subread"]             # <<<<<<<<<<<<<<
 *         if covering_subread[1] <= pos + margin:
 *             return False
 */
    if (unlikely(__pyx_v_read == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 861, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_read, __pyx_n_u_covering_subread); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_covering_subread = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "aldy/indelpost/varaln.pyx":862
 *     if read["is_covering"]:
 *         covering_subread = read["covering_subread"]
 *         if covering_subread[1] <= pos + margin:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_covering_subread, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_pos + __pyx_v_margin)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_LE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_2) {

      /* "aldy/indelpost/varaln.pyx":863
 *         covering_subread = read["covering_subread"]
 *         if covering_subread[1] <= pos + margin:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         if pos < aln_start or aln_end < pos:
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "aldy/indelpost/varaln.pyx":862
 *     if read["is_covering"]:
 *         covering_subread = read["covering_subread"]
 *         if covering_subread[1] <= pos + margin:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    }

    /* "aldy/indelpost/varaln.pyx":865
 *             return False
 * 
 *         if pos < aln_start or aln_end < pos:             # <<<<<<<<<<<<<<
 *             return False
 *     else:
 */
    __pyx_t_6 = ((__pyx_v_pos < __pyx_v_aln_start) != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_2 = __pyx_t_6;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_6 = ((__pyx_v_aln_end < __pyx_v_pos) != 0);
    __pyx_t_2 = __pyx_t_6;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_2) {

      /* "aldy/indelpost/varaln.pyx":866
 * 
 *         if pos < aln_start or aln_end < pos:
 *             return False             # <<<<<<<<<<<<<<
 *     else:
 *         if aln_end < pos:
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "aldy/indelpost/varaln.pyx":865
 *             return False
 * 
 *         if pos < aln_start or aln_end < pos:             # <<<<<<<<<<<<<<
 *             return False
 *     else:
 */
    }

    /* "aldy/indelpost/varaln.pyx":860
 *         return False
 * 
 *     if read["is_covering"]:             # <<<<<<<<<<<<<<
 *         covering_subread = read["covering_subread"]
 *         if covering_subread[1] <= pos + margin:
 */
    goto __pyx_L5;
  }

  /* "aldy/indelpost/varaln.pyx":868
 *             return False
 *     else:
 *         if aln_end < pos:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  /*else*/ {
    __pyx_t_2 = ((__pyx_v_aln_end < __pyx_v_pos) != 0);
    if (__pyx_t_2) {

      /* "aldy/indelpost/varaln.pyx":869
 *     else:
 *         if aln_end < pos:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         if del_len:
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "aldy/indelpost/varaln.pyx":868
 *             return False
 *     else:
 *         if aln_end < pos:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    }

    /* "aldy/indelpost/varaln.pyx":871
 *             return False
 * 
 *         if del_len:             # <<<<<<<<<<<<<<
 *             if pos + del_len < aln_start:
 *                 return False
 */
    __pyx_t_2 = (__pyx_v_del_len != 0);
    if (__pyx_t_2) {

      /* "aldy/indelpost/varaln.pyx":872
 * 
 *         if del_len:
 *             if pos + del_len < aln_start:             # <<<<<<<<<<<<<<
 *                 return False
 *         else:
 */
      __pyx_t_2 = (((__pyx_v_pos + __pyx_v_del_len) < __pyx_v_aln_start) != 0);
      if (__pyx_t_2) {

        /* "aldy/indelpost/varaln.pyx":873
 *         if del_len:
 *             if pos + del_len < aln_start:
 *                 return False             # <<<<<<<<<<<<<<
 *         else:
 *             return False
 */
        __pyx_r = 0;
        goto __pyx_L0;

        /* "aldy/indelpost/varaln.pyx":872
 * 
 *         if del_len:
 *             if pos + del_len < aln_start:             # <<<<<<<<<<<<<<
 *                 return False
 *         else:
 */
      }

      /* "aldy/indelpost/varaln.pyx":871
 *             return False
 * 
 *         if del_len:             # <<<<<<<<<<<<<<
 *             if pos + del_len < aln_start:
 *                 return False
 */
      goto __pyx_L11;
    }

    /* "aldy/indelpost/varaln.pyx":875
 *                 return False
 *         else:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *     return True
 */
    /*else*/ {
      __pyx_r = 0;
      goto __pyx_L0;
    }
    __pyx_L11:;
  }
  __pyx_L5:;

  /* "aldy/indelpost/varaln.pyx":877
 *             return False
 * 
 *     return True             # <<<<<<<<<<<<<<
 * 
 * def centrality(read, target_pos):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "aldy/indelpost/varaln.pyx":849
 * 
 * 
 * cdef bint count_as_non_target(dict read, int pos, int del_len, int margin):             # <<<<<<<<<<<<<<
 *     if read["is_target"]:
 *         return False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("aldy.indelpost.varaln.count_as_non_target", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_covering_subread);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":879
 *     return True
 * 
 * def centrality(read, target_pos):             # <<<<<<<<<<<<<<
 *     relative_pos = relative_aln_pos(read["ref_seq"], read["cigar_list"], read["aln_start"], target_pos)
 *     return abs(0.5 - relative_pos)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_5centrality(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_4centrality[] = "centrality(read, target_pos)";
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6varaln_5centrality = {"centrality", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6varaln_5centrality, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4aldy_9indelpost_6varaln_4centrality};
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_5centrality(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_read = 0;
  PyObject *__pyx_v_target_pos = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("centrality (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_read,&__pyx_n_s_target_pos,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_read)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_target_pos)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("centrality", 1, 2, 2, 1); __PYX_ERR(0, 879, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "centrality") < 0)) __PYX_ERR(0, 879, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_read = values[0];
    __pyx_v_target_pos = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("centrality", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 879, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.centrality", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_4centrality(__pyx_self, __pyx_v_read, __pyx_v_target_pos);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_4centrality(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_target_pos) {
  PyObject *__pyx_v_relative_pos = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("centrality", 0);

  /* "aldy/indelpost/varaln.pyx":880
 * 
 * def centrality(read, target_pos):
 *     relative_pos = relative_aln_pos(read["ref_seq"], read["cigar_list"], read["aln_start"], target_pos)             # <<<<<<<<<<<<<<
 *     return abs(0.5 - relative_pos)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_relative_aln_pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_ref_seq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_start); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_v_target_pos};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 880, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_v_target_pos};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 880, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 880, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_5);
    __Pyx_INCREF(__pyx_v_target_pos);
    __Pyx_GIVEREF(__pyx_v_target_pos);
    PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_7, __pyx_v_target_pos);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 880, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_relative_pos = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":881
 * def centrality(read, target_pos):
 *     relative_pos = relative_aln_pos(read["ref_seq"], read["cigar_list"], read["aln_start"], target_pos)
 *     return abs(0.5 - relative_pos)             # <<<<<<<<<<<<<<
 * 
 * cdef list preprocess_for_contig_construction(
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyFloat_SubtractCObj(__pyx_float_0_5, __pyx_v_relative_pos, 0.5, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyNumber_Absolute(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "aldy/indelpost/varaln.pyx":879
 *     return True
 * 
 * def centrality(read, target_pos):             # <<<<<<<<<<<<<<
 *     relative_pos = relative_aln_pos(read["ref_seq"], read["cigar_list"], read["aln_start"], target_pos)
 *     return abs(0.5 - relative_pos)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("aldy.indelpost.varaln.centrality", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_relative_pos);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":883
 *     return abs(0.5 - relative_pos)
 * 
 * cdef list preprocess_for_contig_construction(             # <<<<<<<<<<<<<<
 *     Variant target,
 *     Variant orig_target,
 */

static PyObject *__pyx_f_4aldy_9indelpost_6varaln_preprocess_for_contig_construction(struct __pyx_obj_4aldy_9indelpost_7variant_Variant *__pyx_v_target, struct __pyx_obj_4aldy_9indelpost_7variant_Variant *__pyx_v_orig_target, PyObject *__pyx_v_pileup, struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference *__pyx_v_unspl_loc_ref, int __pyx_v_window, int __pyx_v_match_score, int __pyx_v_mismatch_penalty, int __pyx_v_gap_open_penalty, int __pyx_v_gap_extension_penalty) {
  CYTHON_UNUSED int __pyx_v_clips;
  int __pyx_v_nonclips;
  PyObject *__pyx_v_targetpileup = NULL;
  PyObject *__pyx_v_clipped_targetpileup = NULL;
  PyObject *__pyx_v_nonclipped_targetpileup = NULL;
  PyObject *__pyx_v_unspl_ref_seq = NULL;
  PyObject *__pyx_v_unspl_lt_len = NULL;
  PyObject *__pyx_v_unspl_aligner = NULL;
  PyObject *__pyx_v_unspl_start = NULL;
  int __pyx_v_is_gapped_aln;
  PyObject *__pyx_v__targetpileup = NULL;
  PyObject *__pyx_9genexpr22__pyx_v_read = NULL;
  PyObject *__pyx_9genexpr23__pyx_v_read = NULL;
  PyObject *__pyx_9genexpr24__pyx_v_read = NULL;
  PyObject *__pyx_9genexpr25__pyx_v_read = NULL;
  PyObject *__pyx_9genexpr26__pyx_v_read = NULL;
  PyObject *__pyx_9genexpr27__pyx_v_read = NULL;
  PyObject *__pyx_9genexpr28__pyx_v_read = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("preprocess_for_contig_construction", 0);

  /* "aldy/indelpost/varaln.pyx":898
 *     cdef int clips, nonclips
 * 
 *     if not pileup:             # <<<<<<<<<<<<<<
 *         return pileup
 * 
 */
  __pyx_t_1 = (__pyx_v_pileup != Py_None)&&(PyList_GET_SIZE(__pyx_v_pileup) != 0);
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "aldy/indelpost/varaln.pyx":899
 * 
 *     if not pileup:
 *         return pileup             # <<<<<<<<<<<<<<
 * 
 *     targetpileup = [read for read in pileup if read["is_target"] and not read["is_dirty"]]
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_pileup);
    __pyx_r = __pyx_v_pileup;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":898
 *     cdef int clips, nonclips
 * 
 *     if not pileup:             # <<<<<<<<<<<<<<
 *         return pileup
 * 
 */
  }

  /* "aldy/indelpost/varaln.pyx":901
 *         return pileup
 * 
 *     targetpileup = [read for read in pileup if read["is_target"] and not read["is_dirty"]]             # <<<<<<<<<<<<<<
 * 
 *     if not targetpileup:
 */
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 901, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__pyx_v_pileup == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 901, __pyx_L6_error)
    }
    __pyx_t_4 = __pyx_v_pileup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    for (;;) {
      if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 901, __pyx_L6_error)
      #else
      __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 901, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      if (!(likely(PyDict_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 901, __pyx_L6_error)
      __Pyx_XDECREF_SET(__pyx_9genexpr22__pyx_v_read, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;
      if (unlikely(__pyx_9genexpr22__pyx_v_read == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 901, __pyx_L6_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr22__pyx_v_read, __pyx_n_u_is_target); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 901, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 901, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (__pyx_t_1) {
      } else {
        __pyx_t_2 = __pyx_t_1;
        goto __pyx_L10_bool_binop_done;
      }
      if (unlikely(__pyx_9genexpr22__pyx_v_read == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 901, __pyx_L6_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr22__pyx_v_read, __pyx_n_u_is_dirty); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 901, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 901, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = ((!__pyx_t_1) != 0);
      __pyx_t_2 = __pyx_t_7;
      __pyx_L10_bool_binop_done:;
      if (__pyx_t_2) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_9genexpr22__pyx_v_read))) __PYX_ERR(0, 901, __pyx_L6_error)
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr22__pyx_v_read); __pyx_9genexpr22__pyx_v_read = 0;
    goto __pyx_L12_exit_scope;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_9genexpr22__pyx_v_read); __pyx_9genexpr22__pyx_v_read = 0;
    goto __pyx_L1_error;
    __pyx_L12_exit_scope:;
  } /* exit inner scope */
  __pyx_v_targetpileup = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "aldy/indelpost/varaln.pyx":903
 *     targetpileup = [read for read in pileup if read["is_target"] and not read["is_dirty"]]
 * 
 *     if not targetpileup:             # <<<<<<<<<<<<<<
 *         return targetpileup
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_targetpileup); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 903, __pyx_L1_error)
  __pyx_t_7 = ((!__pyx_t_2) != 0);
  if (__pyx_t_7) {

    /* "aldy/indelpost/varaln.pyx":904
 * 
 *     if not targetpileup:
 *         return targetpileup             # <<<<<<<<<<<<<<
 * 
 *     clipped_targetpileup = [
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PyList_CheckExact(__pyx_v_targetpileup))||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_targetpileup)->tp_name), 0))) __PYX_ERR(0, 904, __pyx_L1_error)
    __Pyx_INCREF(__pyx_v_targetpileup);
    __pyx_r = ((PyObject*)__pyx_v_targetpileup);
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":903
 *     targetpileup = [read for read in pileup if read["is_target"] and not read["is_dirty"]]
 * 
 *     if not targetpileup:             # <<<<<<<<<<<<<<
 *         return targetpileup
 * 
 */
  }

  /* "aldy/indelpost/varaln.pyx":906
 *         return targetpileup
 * 
 *     clipped_targetpileup = [             # <<<<<<<<<<<<<<
 *         read for read in targetpileup if "S" in read["cigar_string"]
 *     ]
 */
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 906, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "aldy/indelpost/varaln.pyx":907
 * 
 *     clipped_targetpileup = [
 *         read for read in targetpileup if "S" in read["cigar_string"]             # <<<<<<<<<<<<<<
 *     ]
 *     nonclipped_targetpileup = [
 */
    if (likely(PyList_CheckExact(__pyx_v_targetpileup)) || PyTuple_CheckExact(__pyx_v_targetpileup)) {
      __pyx_t_4 = __pyx_v_targetpileup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_targetpileup); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 907, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 907, __pyx_L16_error)
    }
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 907, __pyx_L16_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 907, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 907, __pyx_L16_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 907, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_8(__pyx_t_4);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 907, __pyx_L16_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      if (!(likely(PyDict_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 907, __pyx_L16_error)
      __Pyx_XDECREF_SET(__pyx_9genexpr23__pyx_v_read, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;
      if (unlikely(__pyx_9genexpr23__pyx_v_read == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 907, __pyx_L16_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr23__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 907, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_S, __pyx_t_6, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 907, __pyx_L16_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_2 = (__pyx_t_7 != 0);
      if (__pyx_t_2) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_9genexpr23__pyx_v_read))) __PYX_ERR(0, 906, __pyx_L16_error)
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_read); __pyx_9genexpr23__pyx_v_read = 0;
    goto __pyx_L20_exit_scope;
    __pyx_L16_error:;
    __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_read); __pyx_9genexpr23__pyx_v_read = 0;
    goto __pyx_L1_error;
    __pyx_L20_exit_scope:;
  } /* exit inner scope */
  __pyx_v_clipped_targetpileup = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "aldy/indelpost/varaln.pyx":909
 *         read for read in targetpileup if "S" in read["cigar_string"]
 *     ]
 *     nonclipped_targetpileup = [             # <<<<<<<<<<<<<<
 *         read for read in targetpileup if not "S" in read["cigar_string"]
 *         and (read.get("lt_cigar", None) and read.get("rt_cigar", None))
 */
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 909, __pyx_L23_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "aldy/indelpost/varaln.pyx":910
 *     ]
 *     nonclipped_targetpileup = [
 *         read for read in targetpileup if not "S" in read["cigar_string"]             # <<<<<<<<<<<<<<
 *         and (read.get("lt_cigar", None) and read.get("rt_cigar", None))
 *     ]
 */
    if (likely(PyList_CheckExact(__pyx_v_targetpileup)) || PyTuple_CheckExact(__pyx_v_targetpileup)) {
      __pyx_t_4 = __pyx_v_targetpileup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_targetpileup); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 910, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 910, __pyx_L23_error)
    }
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 910, __pyx_L23_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 910, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 910, __pyx_L23_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 910, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_8(__pyx_t_4);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 910, __pyx_L23_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      if (!(likely(PyDict_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 910, __pyx_L23_error)
      __Pyx_XDECREF_SET(__pyx_9genexpr24__pyx_v_read, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "aldy/indelpost/varaln.pyx":911
 *     nonclipped_targetpileup = [
 *         read for read in targetpileup if not "S" in read["cigar_string"]
 *         and (read.get("lt_cigar", None) and read.get("rt_cigar", None))             # <<<<<<<<<<<<<<
 *     ]
 * 
 */
      if (unlikely(__pyx_9genexpr24__pyx_v_read == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 910, __pyx_L23_error)
      }

      /* "aldy/indelpost/varaln.pyx":910
 *     ]
 *     nonclipped_targetpileup = [
 *         read for read in targetpileup if not "S" in read["cigar_string"]             # <<<<<<<<<<<<<<
 *         and (read.get("lt_cigar", None) and read.get("rt_cigar", None))
 *     ]
 */
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_9genexpr24__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 910, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_S, __pyx_t_6, Py_NE)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 910, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_1 = (__pyx_t_7 != 0);
      if (__pyx_t_1) {
      } else {
        __pyx_t_2 = __pyx_t_1;
        goto __pyx_L27_bool_binop_done;
      }

      /* "aldy/indelpost/varaln.pyx":911
 *     nonclipped_targetpileup = [
 *         read for read in targetpileup if not "S" in read["cigar_string"]
 *         and (read.get("lt_cigar", None) and read.get("rt_cigar", None))             # <<<<<<<<<<<<<<
 *     ]
 * 
 */
      if (unlikely(__pyx_9genexpr24__pyx_v_read == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
        __PYX_ERR(0, 911, __pyx_L23_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItemDefault(__pyx_9genexpr24__pyx_v_read, __pyx_n_u_lt_cigar, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 911, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 911, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (__pyx_t_1) {
      } else {
        __pyx_t_2 = __pyx_t_1;
        goto __pyx_L27_bool_binop_done;
      }
      if (unlikely(__pyx_9genexpr24__pyx_v_read == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
        __PYX_ERR(0, 911, __pyx_L23_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItemDefault(__pyx_9genexpr24__pyx_v_read, __pyx_n_u_rt_cigar, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 911, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 911, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_2 = __pyx_t_1;
      __pyx_L27_bool_binop_done:;

      /* "aldy/indelpost/varaln.pyx":910
 *     ]
 *     nonclipped_targetpileup = [
 *         read for read in targetpileup if not "S" in read["cigar_string"]             # <<<<<<<<<<<<<<
 *         and (read.get("lt_cigar", None) and read.get("rt_cigar", None))
 *     ]
 */
      if (__pyx_t_2) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_9genexpr24__pyx_v_read))) __PYX_ERR(0, 909, __pyx_L23_error)
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_read); __pyx_9genexpr24__pyx_v_read = 0;
    goto __pyx_L30_exit_scope;
    __pyx_L23_error:;
    __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_read); __pyx_9genexpr24__pyx_v_read = 0;
    goto __pyx_L1_error;
    __pyx_L30_exit_scope:;
  } /* exit inner scope */
  __pyx_v_nonclipped_targetpileup = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "aldy/indelpost/varaln.pyx":914
 *     ]
 * 
 *     clips = len(clipped_targetpileup)             # <<<<<<<<<<<<<<
 *     nonclips = len(nonclipped_targetpileup)
 * 
 */
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_clipped_targetpileup); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 914, __pyx_L1_error)
  __pyx_v_clips = __pyx_t_5;

  /* "aldy/indelpost/varaln.pyx":915
 * 
 *     clips = len(clipped_targetpileup)
 *     nonclips = len(nonclipped_targetpileup)             # <<<<<<<<<<<<<<
 * 
 *     if target == orig_target and nonclips > 9:
 */
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_nonclipped_targetpileup); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 915, __pyx_L1_error)
  __pyx_v_nonclips = __pyx_t_5;

  /* "aldy/indelpost/varaln.pyx":917
 *     nonclips = len(nonclipped_targetpileup)
 * 
 *     if target == orig_target and nonclips > 9:             # <<<<<<<<<<<<<<
 *         random.seed(123)
 *         targetpileup = random.sample(nonclipped_targetpileup, 10)
 */
  __pyx_t_3 = PyObject_RichCompare(((PyObject *)__pyx_v_target), ((PyObject *)__pyx_v_orig_target), Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 917, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L32_bool_binop_done;
  }
  __pyx_t_1 = ((__pyx_v_nonclips > 9) != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L32_bool_binop_done:;
  if (__pyx_t_2) {

    /* "aldy/indelpost/varaln.pyx":918
 * 
 *     if target == orig_target and nonclips > 9:
 *         random.seed(123)             # <<<<<<<<<<<<<<
 *         targetpileup = random.sample(nonclipped_targetpileup, 10)
 *         targetpileup = [right_aligner(read, target) for read in targetpileup]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_random); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 918, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_seed); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 918, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_int_123) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_int_123);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 918, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":919
 *     if target == orig_target and nonclips > 9:
 *         random.seed(123)
 *         targetpileup = random.sample(nonclipped_targetpileup, 10)             # <<<<<<<<<<<<<<
 *         targetpileup = [right_aligner(read, target) for read in targetpileup]
 *     else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_random); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 919, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_sample); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 919, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_nonclipped_targetpileup, __pyx_int_10};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_nonclipped_targetpileup, __pyx_int_10};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 919, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_nonclipped_targetpileup);
      __Pyx_GIVEREF(__pyx_v_nonclipped_targetpileup);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_nonclipped_targetpileup);
      __Pyx_INCREF(__pyx_int_10);
      __Pyx_GIVEREF(__pyx_int_10);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_int_10);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_targetpileup, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":920
 *         random.seed(123)
 *         targetpileup = random.sample(nonclipped_targetpileup, 10)
 *         targetpileup = [right_aligner(read, target) for read in targetpileup]             # <<<<<<<<<<<<<<
 *     else:
 *         targetpileup = sorted(targetpileup, key=partial(centrality, target_pos=target.pos))
 */
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L36_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (likely(PyList_CheckExact(__pyx_v_targetpileup)) || PyTuple_CheckExact(__pyx_v_targetpileup)) {
        __pyx_t_4 = __pyx_v_targetpileup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_targetpileup); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 920, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 920, __pyx_L36_error)
      }
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_10); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 920, __pyx_L36_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 920, __pyx_L36_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          } else {
            if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_10); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 920, __pyx_L36_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 920, __pyx_L36_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          }
        } else {
          __pyx_t_10 = __pyx_t_8(__pyx_t_4);
          if (unlikely(!__pyx_t_10)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 920, __pyx_L36_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_10);
        }
        if (!(likely(PyDict_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_10)->tp_name), 0))) __PYX_ERR(0, 920, __pyx_L36_error)
        __Pyx_XDECREF_SET(__pyx_9genexpr25__pyx_v_read, ((PyObject*)__pyx_t_10));
        __pyx_t_10 = 0;
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_right_aligner); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 920, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_11 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_9 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_9genexpr25__pyx_v_read, ((PyObject *)__pyx_v_target)};
          __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 920, __pyx_L36_error)
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_GOTREF(__pyx_t_10);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_9genexpr25__pyx_v_read, ((PyObject *)__pyx_v_target)};
          __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 920, __pyx_L36_error)
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_GOTREF(__pyx_t_10);
        } else
        #endif
        {
          __pyx_t_12 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 920, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_12);
          if (__pyx_t_11) {
            __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_11); __pyx_t_11 = NULL;
          }
          __Pyx_INCREF(__pyx_9genexpr25__pyx_v_read);
          __Pyx_GIVEREF(__pyx_9genexpr25__pyx_v_read);
          PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_9, __pyx_9genexpr25__pyx_v_read);
          __Pyx_INCREF(((PyObject *)__pyx_v_target));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_target));
          PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_9, ((PyObject *)__pyx_v_target));
          __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_12, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 920, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 920, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_read); __pyx_9genexpr25__pyx_v_read = 0;
      goto __pyx_L39_exit_scope;
      __pyx_L36_error:;
      __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_read); __pyx_9genexpr25__pyx_v_read = 0;
      goto __pyx_L1_error;
      __pyx_L39_exit_scope:;
    } /* exit inner scope */
    __Pyx_DECREF_SET(__pyx_v_targetpileup, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":917
 *     nonclips = len(nonclipped_targetpileup)
 * 
 *     if target == orig_target and nonclips > 9:             # <<<<<<<<<<<<<<
 *         random.seed(123)
 *         targetpileup = random.sample(nonclipped_targetpileup, 10)
 */
    goto __pyx_L31;
  }

  /* "aldy/indelpost/varaln.pyx":922
 *         targetpileup = [right_aligner(read, target) for read in targetpileup]
 *     else:
 *         targetpileup = sorted(targetpileup, key=partial(centrality, target_pos=target.pos))             # <<<<<<<<<<<<<<
 * 
 *         unspl_ref_seq, unspl_lt_len = get_local_reference(orig_target, pileup, window, unspl_loc_ref, unspliced=True)
 */
  /*else*/ {
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_targetpileup);
    __Pyx_GIVEREF(__pyx_v_targetpileup);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_targetpileup);
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_partial); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_centrality); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_12 = PyTuple_New(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_target->pos); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_target_pos, __pyx_t_11) < 0) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_12, __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_key, __pyx_t_11) < 0) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_targetpileup, __pyx_t_11);
    __pyx_t_11 = 0;

    /* "aldy/indelpost/varaln.pyx":924
 *         targetpileup = sorted(targetpileup, key=partial(centrality, target_pos=target.pos))
 * 
 *         unspl_ref_seq, unspl_lt_len = get_local_reference(orig_target, pileup, window, unspl_loc_ref, unspliced=True)             # <<<<<<<<<<<<<<
 *         unspl_aligner = make_aligner(unspl_ref_seq, match_score, mismatch_penalty)
 *         unspl_start = orig_target.pos + 1 - unspl_lt_len
 */
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_get_local_reference); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_window); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)__pyx_v_orig_target));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_orig_target));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_orig_target));
    __Pyx_INCREF(__pyx_v_pileup);
    __Pyx_GIVEREF(__pyx_v_pileup);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_pileup);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_4);
    __Pyx_INCREF(((PyObject *)__pyx_v_unspl_loc_ref));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_unspl_loc_ref));
    PyTuple_SET_ITEM(__pyx_t_3, 3, ((PyObject *)__pyx_v_unspl_loc_ref));
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_unspliced, Py_True) < 0) __PYX_ERR(0, 924, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
      PyObject* sequence = __pyx_t_6;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 924, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 924, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 924, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_11 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 924, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_13 = Py_TYPE(__pyx_t_11)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_13(__pyx_t_11); if (unlikely(!__pyx_t_4)) goto __pyx_L40_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_3 = __pyx_t_13(__pyx_t_11); if (unlikely(!__pyx_t_3)) goto __pyx_L40_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_11), 2) < 0) __PYX_ERR(0, 924, __pyx_L1_error)
      __pyx_t_13 = NULL;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L41_unpacking_done;
      __pyx_L40_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_13 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 924, __pyx_L1_error)
      __pyx_L41_unpacking_done:;
    }
    __pyx_v_unspl_ref_seq = __pyx_t_4;
    __pyx_t_4 = 0;
    __pyx_v_unspl_lt_len = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":925
 * 
 *         unspl_ref_seq, unspl_lt_len = get_local_reference(orig_target, pileup, window, unspl_loc_ref, unspliced=True)
 *         unspl_aligner = make_aligner(unspl_ref_seq, match_score, mismatch_penalty)             # <<<<<<<<<<<<<<
 *         unspl_start = orig_target.pos + 1 - unspl_lt_len
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_make_aligner); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 925, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_match_score); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 925, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_mismatch_penalty); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 925, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_12, __pyx_v_unspl_ref_seq, __pyx_t_4, __pyx_t_11};
      __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 925, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_12, __pyx_v_unspl_ref_seq, __pyx_t_4, __pyx_t_11};
      __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 925, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 925, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_12) {
        __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_12); __pyx_t_12 = NULL;
      }
      __Pyx_INCREF(__pyx_v_unspl_ref_seq);
      __Pyx_GIVEREF(__pyx_v_unspl_ref_seq);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_unspl_ref_seq);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_t_11);
      __pyx_t_4 = 0;
      __pyx_t_11 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 925, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_unspl_aligner = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "aldy/indelpost/varaln.pyx":926
 *         unspl_ref_seq, unspl_lt_len = get_local_reference(orig_target, pileup, window, unspl_loc_ref, unspliced=True)
 *         unspl_aligner = make_aligner(unspl_ref_seq, match_score, mismatch_penalty)
 *         unspl_start = orig_target.pos + 1 - unspl_lt_len             # <<<<<<<<<<<<<<
 * 
 *         is_gapped_aln = False
 */
    __pyx_t_6 = __Pyx_PyInt_From_long((__pyx_v_orig_target->pos + 1)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_6, __pyx_v_unspl_lt_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_unspl_start = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":928
 *         unspl_start = orig_target.pos + 1 - unspl_lt_len
 * 
 *         is_gapped_aln = False             # <<<<<<<<<<<<<<
 *         targetpileup = [
 *             update_spliced_read_info(
 */
    __pyx_v_is_gapped_aln = 0;

    /* "aldy/indelpost/varaln.pyx":929
 * 
 *         is_gapped_aln = False
 *         targetpileup = [             # <<<<<<<<<<<<<<
 *             update_spliced_read_info(
 *                 read,
 */
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 929, __pyx_L44_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "aldy/indelpost/varaln.pyx":953
 *                 unspl_start,
 *             )
 *             for read in targetpileup             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
      if (likely(PyList_CheckExact(__pyx_v_targetpileup)) || PyTuple_CheckExact(__pyx_v_targetpileup)) {
        __pyx_t_6 = __pyx_v_targetpileup; __Pyx_INCREF(__pyx_t_6); __pyx_t_5 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_5 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_targetpileup); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 953, __pyx_L44_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 953, __pyx_L44_error)
      }
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_6))) {
            if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_5); __Pyx_INCREF(__pyx_t_10); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 953, __pyx_L44_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_6, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 953, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          } else {
            if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_5); __Pyx_INCREF(__pyx_t_10); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 953, __pyx_L44_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_6, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 953, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          }
        } else {
          __pyx_t_10 = __pyx_t_8(__pyx_t_6);
          if (unlikely(!__pyx_t_10)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 953, __pyx_L44_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_10);
        }
        if (!(likely(PyDict_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_10)->tp_name), 0))) __PYX_ERR(0, 953, __pyx_L44_error)
        __Pyx_XDECREF_SET(__pyx_9genexpr26__pyx_v_read, ((PyObject*)__pyx_t_10));
        __pyx_t_10 = 0;

        /* "aldy/indelpost/varaln.pyx":942
 *                 unspl_loc_ref,
 *             )
 *             if "N" in read["cigar_string"] else             # <<<<<<<<<<<<<<
 *             update_read_info(
 *                 read,
 */
        if (unlikely(__pyx_9genexpr26__pyx_v_read == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 942, __pyx_L44_error)
        }
        __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_9genexpr26__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 942, __pyx_L44_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_N, __pyx_t_11, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 942, __pyx_L44_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if ((__pyx_t_2 != 0)) {

          /* "aldy/indelpost/varaln.pyx":930
 *         is_gapped_aln = False
 *         targetpileup = [
 *             update_spliced_read_info(             # <<<<<<<<<<<<<<
 *                 read,
 *                 target,
 */
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_update_spliced_read_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 930, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_4);

          /* "aldy/indelpost/varaln.pyx":934
 *                 target,
 *                 orig_target,
 *                 is_gapped_aln,             # <<<<<<<<<<<<<<
 *                 window,
 *                 match_score,
 */
          __pyx_t_12 = __Pyx_PyBool_FromLong(__pyx_v_is_gapped_aln); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 934, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_12);

          /* "aldy/indelpost/varaln.pyx":935
 *                 orig_target,
 *                 is_gapped_aln,
 *                 window,             # <<<<<<<<<<<<<<
 *                 match_score,
 *                 mismatch_penalty,
 */
          __pyx_t_14 = __Pyx_PyInt_From_int(__pyx_v_window); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 935, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_14);

          /* "aldy/indelpost/varaln.pyx":936
 *                 is_gapped_aln,
 *                 window,
 *                 match_score,             # <<<<<<<<<<<<<<
 *                 mismatch_penalty,
 *                 gap_open_penalty,
 */
          __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_match_score); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 936, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_15);

          /* "aldy/indelpost/varaln.pyx":937
 *                 window,
 *                 match_score,
 *                 mismatch_penalty,             # <<<<<<<<<<<<<<
 *                 gap_open_penalty,
 *                 gap_extension_penalty,
 */
          __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_mismatch_penalty); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 937, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_16);

          /* "aldy/indelpost/varaln.pyx":938
 *                 match_score,
 *                 mismatch_penalty,
 *                 gap_open_penalty,             # <<<<<<<<<<<<<<
 *                 gap_extension_penalty,
 *                 unspl_loc_ref,
 */
          __pyx_t_17 = __Pyx_PyInt_From_int(__pyx_v_gap_open_penalty); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 938, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_17);

          /* "aldy/indelpost/varaln.pyx":939
 *                 mismatch_penalty,
 *                 gap_open_penalty,
 *                 gap_extension_penalty,             # <<<<<<<<<<<<<<
 *                 unspl_loc_ref,
 *             )
 */
          __pyx_t_18 = __Pyx_PyInt_From_int(__pyx_v_gap_extension_penalty); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 939, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_18);

          /* "aldy/indelpost/varaln.pyx":940
 *                 gap_open_penalty,
 *                 gap_extension_penalty,
 *                 unspl_loc_ref,             # <<<<<<<<<<<<<<
 *             )
 *             if "N" in read["cigar_string"] else
 */
          __pyx_t_19 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_19)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_19);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[11] = {__pyx_t_19, __pyx_9genexpr26__pyx_v_read, ((PyObject *)__pyx_v_target), ((PyObject *)__pyx_v_orig_target), __pyx_t_12, __pyx_t_14, __pyx_t_15, __pyx_t_16, __pyx_t_17, __pyx_t_18, ((PyObject *)__pyx_v_unspl_loc_ref)};
            __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 10+__pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 930, __pyx_L44_error)
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[11] = {__pyx_t_19, __pyx_9genexpr26__pyx_v_read, ((PyObject *)__pyx_v_target), ((PyObject *)__pyx_v_orig_target), __pyx_t_12, __pyx_t_14, __pyx_t_15, __pyx_t_16, __pyx_t_17, __pyx_t_18, ((PyObject *)__pyx_v_unspl_loc_ref)};
            __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 10+__pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 930, __pyx_L44_error)
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          } else
          #endif
          {
            __pyx_t_20 = PyTuple_New(10+__pyx_t_9); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 930, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_20);
            if (__pyx_t_19) {
              __Pyx_GIVEREF(__pyx_t_19); PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_19); __pyx_t_19 = NULL;
            }
            __Pyx_INCREF(__pyx_9genexpr26__pyx_v_read);
            __Pyx_GIVEREF(__pyx_9genexpr26__pyx_v_read);
            PyTuple_SET_ITEM(__pyx_t_20, 0+__pyx_t_9, __pyx_9genexpr26__pyx_v_read);
            __Pyx_INCREF(((PyObject *)__pyx_v_target));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_target));
            PyTuple_SET_ITEM(__pyx_t_20, 1+__pyx_t_9, ((PyObject *)__pyx_v_target));
            __Pyx_INCREF(((PyObject *)__pyx_v_orig_target));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_orig_target));
            PyTuple_SET_ITEM(__pyx_t_20, 2+__pyx_t_9, ((PyObject *)__pyx_v_orig_target));
            __Pyx_GIVEREF(__pyx_t_12);
            PyTuple_SET_ITEM(__pyx_t_20, 3+__pyx_t_9, __pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_14);
            PyTuple_SET_ITEM(__pyx_t_20, 4+__pyx_t_9, __pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_15);
            PyTuple_SET_ITEM(__pyx_t_20, 5+__pyx_t_9, __pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_16);
            PyTuple_SET_ITEM(__pyx_t_20, 6+__pyx_t_9, __pyx_t_16);
            __Pyx_GIVEREF(__pyx_t_17);
            PyTuple_SET_ITEM(__pyx_t_20, 7+__pyx_t_9, __pyx_t_17);
            __Pyx_GIVEREF(__pyx_t_18);
            PyTuple_SET_ITEM(__pyx_t_20, 8+__pyx_t_9, __pyx_t_18);
            __Pyx_INCREF(((PyObject *)__pyx_v_unspl_loc_ref));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_unspl_loc_ref));
            PyTuple_SET_ITEM(__pyx_t_20, 9+__pyx_t_9, ((PyObject *)__pyx_v_unspl_loc_ref));
            __pyx_t_12 = 0;
            __pyx_t_14 = 0;
            __pyx_t_15 = 0;
            __pyx_t_16 = 0;
            __pyx_t_17 = 0;
            __pyx_t_18 = 0;
            __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_20, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 930, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_10 = __pyx_t_11;
          __pyx_t_11 = 0;
        } else {

          /* "aldy/indelpost/varaln.pyx":943
 *             )
 *             if "N" in read["cigar_string"] else
 *             update_read_info(             # <<<<<<<<<<<<<<
 *                 read,
 *                 target,
 */
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_update_read_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 943, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_4);

          /* "aldy/indelpost/varaln.pyx":946
 *                 read,
 *                 target,
 *                 is_gapped_aln,             # <<<<<<<<<<<<<<
 *                 gap_open_penalty,
 *                 gap_extension_penalty,
 */
          __pyx_t_20 = __Pyx_PyBool_FromLong(__pyx_v_is_gapped_aln); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 946, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_20);

          /* "aldy/indelpost/varaln.pyx":947
 *                 target,
 *                 is_gapped_aln,
 *                 gap_open_penalty,             # <<<<<<<<<<<<<<
 *                 gap_extension_penalty,
 *                 unspl_aligner,
 */
          __pyx_t_18 = __Pyx_PyInt_From_int(__pyx_v_gap_open_penalty); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 947, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_18);

          /* "aldy/indelpost/varaln.pyx":948
 *                 is_gapped_aln,
 *                 gap_open_penalty,
 *                 gap_extension_penalty,             # <<<<<<<<<<<<<<
 *                 unspl_aligner,
 *                 unspl_ref_seq,
 */
          __pyx_t_17 = __Pyx_PyInt_From_int(__pyx_v_gap_extension_penalty); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 948, __pyx_L44_error)
          __Pyx_GOTREF(__pyx_t_17);

          /* "aldy/indelpost/varaln.pyx":951
 *                 unspl_aligner,
 *                 unspl_ref_seq,
 *                 unspl_start,             # <<<<<<<<<<<<<<
 *             )
 *             for read in targetpileup
 */
          __pyx_t_16 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_16)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_16);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[9] = {__pyx_t_16, __pyx_9genexpr26__pyx_v_read, ((PyObject *)__pyx_v_target), __pyx_t_20, __pyx_t_18, __pyx_t_17, __pyx_v_unspl_aligner, __pyx_v_unspl_ref_seq, __pyx_v_unspl_start};
            __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 8+__pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 943, __pyx_L44_error)
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[9] = {__pyx_t_16, __pyx_9genexpr26__pyx_v_read, ((PyObject *)__pyx_v_target), __pyx_t_20, __pyx_t_18, __pyx_t_17, __pyx_v_unspl_aligner, __pyx_v_unspl_ref_seq, __pyx_v_unspl_start};
            __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 8+__pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 943, __pyx_L44_error)
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          } else
          #endif
          {
            __pyx_t_15 = PyTuple_New(8+__pyx_t_9); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 943, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_15);
            if (__pyx_t_16) {
              __Pyx_GIVEREF(__pyx_t_16); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_16); __pyx_t_16 = NULL;
            }
            __Pyx_INCREF(__pyx_9genexpr26__pyx_v_read);
            __Pyx_GIVEREF(__pyx_9genexpr26__pyx_v_read);
            PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_9, __pyx_9genexpr26__pyx_v_read);
            __Pyx_INCREF(((PyObject *)__pyx_v_target));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_target));
            PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_9, ((PyObject *)__pyx_v_target));
            __Pyx_GIVEREF(__pyx_t_20);
            PyTuple_SET_ITEM(__pyx_t_15, 2+__pyx_t_9, __pyx_t_20);
            __Pyx_GIVEREF(__pyx_t_18);
            PyTuple_SET_ITEM(__pyx_t_15, 3+__pyx_t_9, __pyx_t_18);
            __Pyx_GIVEREF(__pyx_t_17);
            PyTuple_SET_ITEM(__pyx_t_15, 4+__pyx_t_9, __pyx_t_17);
            __Pyx_INCREF(__pyx_v_unspl_aligner);
            __Pyx_GIVEREF(__pyx_v_unspl_aligner);
            PyTuple_SET_ITEM(__pyx_t_15, 5+__pyx_t_9, __pyx_v_unspl_aligner);
            __Pyx_INCREF(__pyx_v_unspl_ref_seq);
            __Pyx_GIVEREF(__pyx_v_unspl_ref_seq);
            PyTuple_SET_ITEM(__pyx_t_15, 6+__pyx_t_9, __pyx_v_unspl_ref_seq);
            __Pyx_INCREF(__pyx_v_unspl_start);
            __Pyx_GIVEREF(__pyx_v_unspl_start);
            PyTuple_SET_ITEM(__pyx_t_15, 7+__pyx_t_9, __pyx_v_unspl_start);
            __pyx_t_20 = 0;
            __pyx_t_18 = 0;
            __pyx_t_17 = 0;
            __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_15, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 943, __pyx_L44_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_10 = __pyx_t_11;
          __pyx_t_11 = 0;
        }
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 929, __pyx_L44_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "aldy/indelpost/varaln.pyx":953
 *                 unspl_start,
 *             )
 *             for read in targetpileup             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_9genexpr26__pyx_v_read); __pyx_9genexpr26__pyx_v_read = 0;
      goto __pyx_L47_exit_scope;
      __pyx_L44_error:;
      __Pyx_XDECREF(__pyx_9genexpr26__pyx_v_read); __pyx_9genexpr26__pyx_v_read = 0;
      goto __pyx_L1_error;
      __pyx_L47_exit_scope:;
    } /* exit inner scope */
    __Pyx_DECREF_SET(__pyx_v_targetpileup, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":956
 *         ]
 * 
 *         targetpileup = [read for read in targetpileup if read is not None and (read.get("lt_cigar", None) and read.get("rt_cigar", None))]             # <<<<<<<<<<<<<<
 * 
 *         _targetpileup = [
 */
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 956, __pyx_L50_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (likely(PyList_CheckExact(__pyx_v_targetpileup)) || PyTuple_CheckExact(__pyx_v_targetpileup)) {
        __pyx_t_6 = __pyx_v_targetpileup; __Pyx_INCREF(__pyx_t_6); __pyx_t_5 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_5 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_targetpileup); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 956, __pyx_L50_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 956, __pyx_L50_error)
      }
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_6))) {
            if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_5); __Pyx_INCREF(__pyx_t_10); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 956, __pyx_L50_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_6, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 956, __pyx_L50_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          } else {
            if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_5); __Pyx_INCREF(__pyx_t_10); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 956, __pyx_L50_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_6, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 956, __pyx_L50_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          }
        } else {
          __pyx_t_10 = __pyx_t_8(__pyx_t_6);
          if (unlikely(!__pyx_t_10)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 956, __pyx_L50_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_10);
        }
        if (!(likely(PyDict_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_10)->tp_name), 0))) __PYX_ERR(0, 956, __pyx_L50_error)
        __Pyx_XDECREF_SET(__pyx_9genexpr27__pyx_v_read, ((PyObject*)__pyx_t_10));
        __pyx_t_10 = 0;
        __pyx_t_1 = (__pyx_9genexpr27__pyx_v_read != ((PyObject*)Py_None));
        __pyx_t_7 = (__pyx_t_1 != 0);
        if (__pyx_t_7) {
        } else {
          __pyx_t_2 = __pyx_t_7;
          goto __pyx_L54_bool_binop_done;
        }
        if (unlikely(__pyx_9genexpr27__pyx_v_read == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
          __PYX_ERR(0, 956, __pyx_L50_error)
        }
        __pyx_t_10 = __Pyx_PyDict_GetItemDefault(__pyx_9genexpr27__pyx_v_read, __pyx_n_u_lt_cigar, Py_None); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 956, __pyx_L50_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 956, __pyx_L50_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (__pyx_t_7) {
        } else {
          __pyx_t_2 = __pyx_t_7;
          goto __pyx_L54_bool_binop_done;
        }
        if (unlikely(__pyx_9genexpr27__pyx_v_read == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
          __PYX_ERR(0, 956, __pyx_L50_error)
        }
        __pyx_t_10 = __Pyx_PyDict_GetItemDefault(__pyx_9genexpr27__pyx_v_read, __pyx_n_u_rt_cigar, Py_None); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 956, __pyx_L50_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 956, __pyx_L50_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_2 = __pyx_t_7;
        __pyx_L54_bool_binop_done:;
        if (__pyx_t_2) {
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_9genexpr27__pyx_v_read))) __PYX_ERR(0, 956, __pyx_L50_error)
        }
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_9genexpr27__pyx_v_read); __pyx_9genexpr27__pyx_v_read = 0;
      goto __pyx_L57_exit_scope;
      __pyx_L50_error:;
      __Pyx_XDECREF(__pyx_9genexpr27__pyx_v_read); __pyx_9genexpr27__pyx_v_read = 0;
      goto __pyx_L1_error;
      __pyx_L57_exit_scope:;
    } /* exit inner scope */
    __Pyx_DECREF_SET(__pyx_v_targetpileup, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":958
 *         targetpileup = [read for read in targetpileup if read is not None and (read.get("lt_cigar", None) and read.get("rt_cigar", None))]
 * 
 *         _targetpileup = [             # <<<<<<<<<<<<<<
 *             read for read in targetpileup if read.get("cigar_updated", False)
 *         ]
 */
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 958, __pyx_L60_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "aldy/indelpost/varaln.pyx":959
 * 
 *         _targetpileup = [
 *             read for read in targetpileup if read.get("cigar_updated", False)             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
      if (likely(PyList_CheckExact(__pyx_v_targetpileup)) || PyTuple_CheckExact(__pyx_v_targetpileup)) {
        __pyx_t_6 = __pyx_v_targetpileup; __Pyx_INCREF(__pyx_t_6); __pyx_t_5 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_5 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_targetpileup); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 959, __pyx_L60_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 959, __pyx_L60_error)
      }
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_6))) {
            if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_5); __Pyx_INCREF(__pyx_t_10); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 959, __pyx_L60_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_6, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 959, __pyx_L60_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          } else {
            if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_5); __Pyx_INCREF(__pyx_t_10); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 959, __pyx_L60_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_6, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 959, __pyx_L60_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          }
        } else {
          __pyx_t_10 = __pyx_t_8(__pyx_t_6);
          if (unlikely(!__pyx_t_10)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 959, __pyx_L60_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_10);
        }
        if (!(likely(PyDict_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_10)->tp_name), 0))) __PYX_ERR(0, 959, __pyx_L60_error)
        __Pyx_XDECREF_SET(__pyx_9genexpr28__pyx_v_read, ((PyObject*)__pyx_t_10));
        __pyx_t_10 = 0;
        if (unlikely(__pyx_9genexpr28__pyx_v_read == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
          __PYX_ERR(0, 959, __pyx_L60_error)
        }
        __pyx_t_10 = __Pyx_PyDict_GetItemDefault(__pyx_9genexpr28__pyx_v_read, __pyx_n_u_cigar_updated, Py_False); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 959, __pyx_L60_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 959, __pyx_L60_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (__pyx_t_2) {
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_9genexpr28__pyx_v_read))) __PYX_ERR(0, 958, __pyx_L60_error)
        }
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_9genexpr28__pyx_v_read); __pyx_9genexpr28__pyx_v_read = 0;
      goto __pyx_L64_exit_scope;
      __pyx_L60_error:;
      __Pyx_XDECREF(__pyx_9genexpr28__pyx_v_read); __pyx_9genexpr28__pyx_v_read = 0;
      goto __pyx_L1_error;
      __pyx_L64_exit_scope:;
    } /* exit inner scope */
    __pyx_v__targetpileup = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":962
 *         ]
 * 
 *         if _targetpileup:             # <<<<<<<<<<<<<<
 *             targetpileup = _targetpileup
 *         else:
 */
    __pyx_t_2 = (PyList_GET_SIZE(__pyx_v__targetpileup) != 0);
    if (__pyx_t_2) {

      /* "aldy/indelpost/varaln.pyx":963
 * 
 *         if _targetpileup:
 *             targetpileup = _targetpileup             # <<<<<<<<<<<<<<
 *         else:
 *             return targetpileup
 */
      __Pyx_INCREF(__pyx_v__targetpileup);
      __Pyx_DECREF_SET(__pyx_v_targetpileup, __pyx_v__targetpileup);

      /* "aldy/indelpost/varaln.pyx":962
 *         ]
 * 
 *         if _targetpileup:             # <<<<<<<<<<<<<<
 *             targetpileup = _targetpileup
 *         else:
 */
      goto __pyx_L65;
    }

    /* "aldy/indelpost/varaln.pyx":965
 *             targetpileup = _targetpileup
 *         else:
 *             return targetpileup             # <<<<<<<<<<<<<<
 * 
 *     return targetpileup
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      if (!(likely(PyList_CheckExact(__pyx_v_targetpileup))||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_targetpileup)->tp_name), 0))) __PYX_ERR(0, 965, __pyx_L1_error)
      __Pyx_INCREF(__pyx_v_targetpileup);
      __pyx_r = ((PyObject*)__pyx_v_targetpileup);
      goto __pyx_L0;
    }
    __pyx_L65:;
  }
  __pyx_L31:;

  /* "aldy/indelpost/varaln.pyx":967
 *             return targetpileup
 * 
 *     return targetpileup             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyList_CheckExact(__pyx_v_targetpileup))||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_targetpileup)->tp_name), 0))) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_INCREF(__pyx_v_targetpileup);
  __pyx_r = ((PyObject*)__pyx_v_targetpileup);
  goto __pyx_L0;

  /* "aldy/indelpost/varaln.pyx":883
 *     return abs(0.5 - relative_pos)
 * 
 * cdef list preprocess_for_contig_construction(             # <<<<<<<<<<<<<<
 *     Variant target,
 *     Variant orig_target,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("aldy.indelpost.varaln.preprocess_for_contig_construction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_targetpileup);
  __Pyx_XDECREF(__pyx_v_clipped_targetpileup);
  __Pyx_XDECREF(__pyx_v_nonclipped_targetpileup);
  __Pyx_XDECREF(__pyx_v_unspl_ref_seq);
  __Pyx_XDECREF(__pyx_v_unspl_lt_len);
  __Pyx_XDECREF(__pyx_v_unspl_aligner);
  __Pyx_XDECREF(__pyx_v_unspl_start);
  __Pyx_XDECREF(__pyx_v__targetpileup);
  __Pyx_XDECREF(__pyx_9genexpr22__pyx_v_read);
  __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_read);
  __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_read);
  __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_read);
  __Pyx_XDECREF(__pyx_9genexpr26__pyx_v_read);
  __Pyx_XDECREF(__pyx_9genexpr27__pyx_v_read);
  __Pyx_XDECREF(__pyx_9genexpr28__pyx_v_read);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":970
 * 
 * 
 * def update_spliced_read_info(             # <<<<<<<<<<<<<<
 *     read,
 *     target,
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_7update_spliced_read_info(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_6update_spliced_read_info[] = "update_spliced_read_info(read, target, orig_target, is_gapped_aln, window, match_score, mismatch_penalty, gap_open_penalty, gap_extension_penalty, unspl_loc_ref)";
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6varaln_7update_spliced_read_info = {"update_spliced_read_info", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6varaln_7update_spliced_read_info, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4aldy_9indelpost_6varaln_6update_spliced_read_info};
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_7update_spliced_read_info(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_read = 0;
  PyObject *__pyx_v_target = 0;
  PyObject *__pyx_v_orig_target = 0;
  PyObject *__pyx_v_is_gapped_aln = 0;
  PyObject *__pyx_v_window = 0;
  PyObject *__pyx_v_match_score = 0;
  PyObject *__pyx_v_mismatch_penalty = 0;
  PyObject *__pyx_v_gap_open_penalty = 0;
  PyObject *__pyx_v_gap_extension_penalty = 0;
  PyObject *__pyx_v_unspl_loc_ref = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update_spliced_read_info (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_read,&__pyx_n_s_target,&__pyx_n_s_orig_target,&__pyx_n_s_is_gapped_aln,&__pyx_n_s_window,&__pyx_n_s_match_score,&__pyx_n_s_mismatch_penalty,&__pyx_n_s_gap_open_penalty,&__pyx_n_s_gap_extension_penalty,&__pyx_n_s_unspl_loc_ref,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_read)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_target)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("update_spliced_read_info", 1, 10, 10, 1); __PYX_ERR(0, 970, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_orig_target)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("update_spliced_read_info", 1, 10, 10, 2); __PYX_ERR(0, 970, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_is_gapped_aln)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("update_spliced_read_info", 1, 10, 10, 3); __PYX_ERR(0, 970, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_window)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("update_spliced_read_info", 1, 10, 10, 4); __PYX_ERR(0, 970, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_match_score)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("update_spliced_read_info", 1, 10, 10, 5); __PYX_ERR(0, 970, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mismatch_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("update_spliced_read_info", 1, 10, 10, 6); __PYX_ERR(0, 970, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gap_open_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("update_spliced_read_info", 1, 10, 10, 7); __PYX_ERR(0, 970, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gap_extension_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("update_spliced_read_info", 1, 10, 10, 8); __PYX_ERR(0, 970, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_unspl_loc_ref)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("update_spliced_read_info", 1, 10, 10, 9); __PYX_ERR(0, 970, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "update_spliced_read_info") < 0)) __PYX_ERR(0, 970, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_read = values[0];
    __pyx_v_target = values[1];
    __pyx_v_orig_target = values[2];
    __pyx_v_is_gapped_aln = values[3];
    __pyx_v_window = values[4];
    __pyx_v_match_score = values[5];
    __pyx_v_mismatch_penalty = values[6];
    __pyx_v_gap_open_penalty = values[7];
    __pyx_v_gap_extension_penalty = values[8];
    __pyx_v_unspl_loc_ref = values[9];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("update_spliced_read_info", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 970, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.update_spliced_read_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_6update_spliced_read_info(__pyx_self, __pyx_v_read, __pyx_v_target, __pyx_v_orig_target, __pyx_v_is_gapped_aln, __pyx_v_window, __pyx_v_match_score, __pyx_v_mismatch_penalty, __pyx_v_gap_open_penalty, __pyx_v_gap_extension_penalty, __pyx_v_unspl_loc_ref);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_6update_spliced_read_info(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_target, PyObject *__pyx_v_orig_target, PyObject *__pyx_v_is_gapped_aln, PyObject *__pyx_v_window, PyObject *__pyx_v_match_score, PyObject *__pyx_v_mismatch_penalty, PyObject *__pyx_v_gap_open_penalty, PyObject *__pyx_v_gap_extension_penalty, PyObject *__pyx_v_unspl_loc_ref) {
  PyObject *__pyx_v_ref_seq = NULL;
  PyObject *__pyx_v_lt_len = NULL;
  PyObject *__pyx_v_aligner = NULL;
  PyObject *__pyx_v_ref_start = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update_spliced_read_info", 0);
  __Pyx_INCREF(__pyx_v_read);

  /* "aldy/indelpost/varaln.pyx":982
 *     unspl_loc_ref,
 * ):
 *     ref_seq, lt_len = get_local_reference(orig_target, [read], window, unspl_loc_ref)             # <<<<<<<<<<<<<<
 *     aligner = make_aligner(ref_seq, match_score, mismatch_penalty)
 *     ref_start = orig_target.pos + 1 - lt_len
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_get_local_reference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_read);
  __Pyx_GIVEREF(__pyx_v_read);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_read);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_v_orig_target, __pyx_t_3, __pyx_v_window, __pyx_v_unspl_loc_ref};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 982, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_v_orig_target, __pyx_t_3, __pyx_v_window, __pyx_v_unspl_loc_ref};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 982, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(4+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_orig_target);
    __Pyx_GIVEREF(__pyx_v_orig_target);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_orig_target);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_window);
    __Pyx_GIVEREF(__pyx_v_window);
    PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_window);
    __Pyx_INCREF(__pyx_v_unspl_loc_ref);
    __Pyx_GIVEREF(__pyx_v_unspl_loc_ref);
    PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_5, __pyx_v_unspl_loc_ref);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 982, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_6);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_7(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_6 = __pyx_t_7(__pyx_t_3); if (unlikely(!__pyx_t_6)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_3), 2) < 0) __PYX_ERR(0, 982, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 982, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_ref_seq = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_lt_len = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "aldy/indelpost/varaln.pyx":983
 * ):
 *     ref_seq, lt_len = get_local_reference(orig_target, [read], window, unspl_loc_ref)
 *     aligner = make_aligner(ref_seq, match_score, mismatch_penalty)             # <<<<<<<<<<<<<<
 *     ref_start = orig_target.pos + 1 - lt_len
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_make_aligner); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_ref_seq, __pyx_v_match_score, __pyx_v_mismatch_penalty};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_ref_seq, __pyx_v_match_score, __pyx_v_mismatch_penalty};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_ref_seq);
    __Pyx_GIVEREF(__pyx_v_ref_seq);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_5, __pyx_v_ref_seq);
    __Pyx_INCREF(__pyx_v_match_score);
    __Pyx_GIVEREF(__pyx_v_match_score);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_5, __pyx_v_match_score);
    __Pyx_INCREF(__pyx_v_mismatch_penalty);
    __Pyx_GIVEREF(__pyx_v_mismatch_penalty);
    PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_5, __pyx_v_mismatch_penalty);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_aligner = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":984
 *     ref_seq, lt_len = get_local_reference(orig_target, [read], window, unspl_loc_ref)
 *     aligner = make_aligner(ref_seq, match_score, mismatch_penalty)
 *     ref_start = orig_target.pos + 1 - lt_len             # <<<<<<<<<<<<<<
 * 
 *     read = update_read_info(
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_orig_target, __pyx_n_s_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Subtract(__pyx_t_6, __pyx_v_lt_len); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_ref_start = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":986
 *     ref_start = orig_target.pos + 1 - lt_len
 * 
 *     read = update_read_info(             # <<<<<<<<<<<<<<
 *         read,
 *         target,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_update_read_info); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "aldy/indelpost/varaln.pyx":994
 *         aligner,
 *         ref_seq,
 *         ref_start             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[9] = {__pyx_t_3, __pyx_v_read, __pyx_v_target, __pyx_v_is_gapped_aln, __pyx_v_gap_open_penalty, __pyx_v_gap_extension_penalty, __pyx_v_aligner, __pyx_v_ref_seq, __pyx_v_ref_start};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 8+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[9] = {__pyx_t_3, __pyx_v_read, __pyx_v_target, __pyx_v_is_gapped_aln, __pyx_v_gap_open_penalty, __pyx_v_gap_extension_penalty, __pyx_v_aligner, __pyx_v_ref_seq, __pyx_v_ref_start};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 8+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(8+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 986, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_read);
    __Pyx_GIVEREF(__pyx_v_read);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_5, __pyx_v_read);
    __Pyx_INCREF(__pyx_v_target);
    __Pyx_GIVEREF(__pyx_v_target);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_5, __pyx_v_target);
    __Pyx_INCREF(__pyx_v_is_gapped_aln);
    __Pyx_GIVEREF(__pyx_v_is_gapped_aln);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_5, __pyx_v_is_gapped_aln);
    __Pyx_INCREF(__pyx_v_gap_open_penalty);
    __Pyx_GIVEREF(__pyx_v_gap_open_penalty);
    PyTuple_SET_ITEM(__pyx_t_2, 3+__pyx_t_5, __pyx_v_gap_open_penalty);
    __Pyx_INCREF(__pyx_v_gap_extension_penalty);
    __Pyx_GIVEREF(__pyx_v_gap_extension_penalty);
    PyTuple_SET_ITEM(__pyx_t_2, 4+__pyx_t_5, __pyx_v_gap_extension_penalty);
    __Pyx_INCREF(__pyx_v_aligner);
    __Pyx_GIVEREF(__pyx_v_aligner);
    PyTuple_SET_ITEM(__pyx_t_2, 5+__pyx_t_5, __pyx_v_aligner);
    __Pyx_INCREF(__pyx_v_ref_seq);
    __Pyx_GIVEREF(__pyx_v_ref_seq);
    PyTuple_SET_ITEM(__pyx_t_2, 6+__pyx_t_5, __pyx_v_ref_seq);
    __Pyx_INCREF(__pyx_v_ref_start);
    __Pyx_GIVEREF(__pyx_v_ref_start);
    PyTuple_SET_ITEM(__pyx_t_2, 7+__pyx_t_5, __pyx_v_ref_start);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF_SET(__pyx_v_read, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":997
 *     )
 * 
 *     return right_aligner(read, target)             # <<<<<<<<<<<<<<
 * 
 * def right_aligner(read, target):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_right_aligner); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_read, __pyx_v_target};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_read, __pyx_v_target};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_read);
    __Pyx_GIVEREF(__pyx_v_read);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_5, __pyx_v_read);
    __Pyx_INCREF(__pyx_v_target);
    __Pyx_GIVEREF(__pyx_v_target);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_5, __pyx_v_target);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "aldy/indelpost/varaln.pyx":970
 * 
 * 
 * def update_spliced_read_info(             # <<<<<<<<<<<<<<
 *     read,
 *     target,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("aldy.indelpost.varaln.update_spliced_read_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ref_seq);
  __Pyx_XDECREF(__pyx_v_lt_len);
  __Pyx_XDECREF(__pyx_v_aligner);
  __Pyx_XDECREF(__pyx_v_ref_start);
  __Pyx_XDECREF(__pyx_v_read);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":999
 *     return right_aligner(read, target)
 * 
 * def right_aligner(read, target):             # <<<<<<<<<<<<<<
 *     """Right align indels around splice site"""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_9right_aligner(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_8right_aligner[] = "right_aligner(read, target)\nRight align indels around splice site";
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6varaln_9right_aligner = {"right_aligner", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6varaln_9right_aligner, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4aldy_9indelpost_6varaln_8right_aligner};
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_9right_aligner(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_read = 0;
  PyObject *__pyx_v_target = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("right_aligner (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_read,&__pyx_n_s_target,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_read)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_target)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("right_aligner", 1, 2, 2, 1); __PYX_ERR(0, 999, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "right_aligner") < 0)) __PYX_ERR(0, 999, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_read = values[0];
    __pyx_v_target = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("right_aligner", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 999, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.right_aligner", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_8right_aligner(__pyx_self, __pyx_v_read, __pyx_v_target);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_4aldy_9indelpost_6varaln_13right_aligner_2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "aldy/indelpost/varaln.pyx":1056
 *                                 ).generate_equivalents()
 * 
 *             diff = max(v.pos for v in right_aligned_vars) - ref_pos + 1             # <<<<<<<<<<<<<<
 *             if diff > 0:
 *                 new_cigar += [str(diff) + "M", str(event_len) + event, str(nxt_event_len - diff) + "M"]
 */

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_13right_aligner_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr *)__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr(__pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1056, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_4aldy_9indelpost_6varaln_13right_aligner_2generator2, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_right_aligner_locals_genexpr, __pyx_n_s_aldy_indelpost_varaln); if (unlikely(!gen)) __PYX_ERR(0, 1056, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.right_aligner.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_4aldy_9indelpost_6varaln_13right_aligner_2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr *__pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1056, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_right_aligned_vars)) { __Pyx_RaiseClosureNameError("right_aligned_vars"); __PYX_ERR(0, 1056, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_right_aligned_vars)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_right_aligned_vars)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_right_aligned_vars; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_right_aligned_vars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1056, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1056, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1056, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1056, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1056, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1056, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_v, __pyx_n_s_pos); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1056, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":999
 *     return right_aligner(read, target)
 * 
 * def right_aligner(read, target):             # <<<<<<<<<<<<<<
 *     """Right align indels around splice site"""
 * 
 */

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_8right_aligner(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_target) {
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner *__pyx_cur_scope;
  PyObject *__pyx_v_cigar_lst = NULL;
  PyObject *__pyx_v_query_pos = NULL;
  PyObject *__pyx_v_ref_pos = NULL;
  PyObject *__pyx_v_new_cigar = NULL;
  PyObject *__pyx_v_prev_event = NULL;
  int __pyx_v_skip_next;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_v_event = NULL;
  PyObject *__pyx_v_event_len = NULL;
  PyObject *__pyx_v_query_move = NULL;
  PyObject *__pyx_v_ref_move = NULL;
  PyObject *__pyx_v_nxt_c = NULL;
  PyObject *__pyx_v_nxt_event = NULL;
  PyObject *__pyx_v_nxt_event_len = NULL;
  PyObject *__pyx_v_chrom = NULL;
  PyObject *__pyx_v_reference = NULL;
  PyObject *__pyx_v_padding_base = NULL;
  PyObject *__pyx_v_ins_seq = NULL;
  PyObject *__pyx_v_ref = NULL;
  PyObject *__pyx_v_alt = NULL;
  PyObject *__pyx_v_del_seq = NULL;
  PyObject *__pyx_v_diff = NULL;
  PyObject *__pyx_v_rt_aln_pos = NULL;
  Py_ssize_t __pyx_v_indel_len;
  PyObject *__pyx_gb_4aldy_9indelpost_6varaln_13right_aligner_2generator2 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_t_18;
  PyObject *(*__pyx_t_19)(PyObject *);
  int __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("right_aligner", 0);
  __pyx_cur_scope = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner *)__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner(__pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 999, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "aldy/indelpost/varaln.pyx":1003
 * 
 *     if (
 *         "N" not in read["cigar_string"]             # <<<<<<<<<<<<<<
 *         or (
 *             "I" not in read["cigar_string"]
 */
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_N, __pyx_t_2, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }

  /* "aldy/indelpost/varaln.pyx":1005
 *         "N" not in read["cigar_string"]
 *         or (
 *             "I" not in read["cigar_string"]             # <<<<<<<<<<<<<<
 *             and "D" not in read["cigar_string"]
 *         )
 */
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_I, __pyx_t_2, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = (__pyx_t_4 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }

  /* "aldy/indelpost/varaln.pyx":1006
 *         or (
 *             "I" not in read["cigar_string"]
 *             and "D" not in read["cigar_string"]             # <<<<<<<<<<<<<<
 *         )
 *     ):
 */
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_D, __pyx_t_2, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;

  /* "aldy/indelpost/varaln.pyx":1002
 *     """Right align indels around splice site"""
 * 
 *     if (             # <<<<<<<<<<<<<<
 *         "N" not in read["cigar_string"]
 *         or (
 */
  if (__pyx_t_1) {

    /* "aldy/indelpost/varaln.pyx":1009
 *         )
 *     ):
 *         return read             # <<<<<<<<<<<<<<
 * 
 *     cigar_lst = read["cigar_list"]
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_read);
    __pyx_r = __pyx_v_read;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":1002
 *     """Right align indels around splice site"""
 * 
 *     if (             # <<<<<<<<<<<<<<
 *         "N" not in read["cigar_string"]
 *         or (
 */
  }

  /* "aldy/indelpost/varaln.pyx":1011
 *         return read
 * 
 *     cigar_lst = read["cigar_list"]             # <<<<<<<<<<<<<<
 * 
 *     query_pos = 0
 */
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_list); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_cigar_lst = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":1013
 *     cigar_lst = read["cigar_list"]
 * 
 *     query_pos = 0             # <<<<<<<<<<<<<<
 *     ref_pos = read["read_start"]
 *     new_cigar = []
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_query_pos = __pyx_int_0;

  /* "aldy/indelpost/varaln.pyx":1014
 * 
 *     query_pos = 0
 *     ref_pos = read["read_start"]             # <<<<<<<<<<<<<<
 *     new_cigar = []
 *     prev_event = "A"
 */
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_start); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_ref_pos = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":1015
 *     query_pos = 0
 *     ref_pos = read["read_start"]
 *     new_cigar = []             # <<<<<<<<<<<<<<
 *     prev_event = "A"
 *     skip_next = False
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_new_cigar = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":1016
 *     ref_pos = read["read_start"]
 *     new_cigar = []
 *     prev_event = "A"             # <<<<<<<<<<<<<<
 *     skip_next = False
 *     for i, c in enumerate(cigar_lst):
 */
  __Pyx_INCREF(__pyx_n_u_A);
  __pyx_v_prev_event = __pyx_n_u_A;

  /* "aldy/indelpost/varaln.pyx":1017
 *     new_cigar = []
 *     prev_event = "A"
 *     skip_next = False             # <<<<<<<<<<<<<<
 *     for i, c in enumerate(cigar_lst):
 *         event, event_len = c[-1], int(c[:-1])
 */
  __pyx_v_skip_next = 0;

  /* "aldy/indelpost/varaln.pyx":1018
 *     prev_event = "A"
 *     skip_next = False
 *     for i, c in enumerate(cigar_lst):             # <<<<<<<<<<<<<<
 *         event, event_len = c[-1], int(c[:-1])
 * 
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_2 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_cigar_lst)) || PyTuple_CheckExact(__pyx_v_cigar_lst)) {
    __pyx_t_5 = __pyx_v_cigar_lst; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_cigar_lst); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1018, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1018, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1018, __pyx_L1_error)
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1018, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1018, __pyx_L1_error)
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1018, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      }
    } else {
      __pyx_t_8 = __pyx_t_7(__pyx_t_5);
      if (unlikely(!__pyx_t_8)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1018, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_8);
    }
    __Pyx_XDECREF_SET(__pyx_v_c, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_8 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1018, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_8;
    __pyx_t_8 = 0;

    /* "aldy/indelpost/varaln.pyx":1019
 *     skip_next = False
 *     for i, c in enumerate(cigar_lst):
 *         event, event_len = c[-1], int(c[:-1])             # <<<<<<<<<<<<<<
 * 
 *         if event_len < 0:
 */
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_c, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1019, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetSlice(__pyx_v_c, 0, -1L, NULL, NULL, &__pyx_slice__5, 0, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1019, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyNumber_Int(__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1019, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF_SET(__pyx_v_event, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_XDECREF_SET(__pyx_v_event_len, __pyx_t_10);
    __pyx_t_10 = 0;

    /* "aldy/indelpost/varaln.pyx":1021
 *         event, event_len = c[-1], int(c[:-1])
 * 
 *         if event_len < 0:             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
    __pyx_t_10 = PyObject_RichCompare(__pyx_v_event_len, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1021, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1021, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (__pyx_t_1) {

      /* "aldy/indelpost/varaln.pyx":1022
 * 
 *         if event_len < 0:
 *             return None             # <<<<<<<<<<<<<<
 * 
 *         query_move = 0 if event in ("D", "N", "H", "P") else event_len
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "aldy/indelpost/varaln.pyx":1021
 *         event, event_len = c[-1], int(c[:-1])
 * 
 *         if event_len < 0:             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
    }

    /* "aldy/indelpost/varaln.pyx":1024
 *             return None
 * 
 *         query_move = 0 if event in ("D", "N", "H", "P") else event_len             # <<<<<<<<<<<<<<
 *         ref_move = 0 if event in ("I", "H", "P") else event_len
 * 
 */
    __Pyx_INCREF(__pyx_v_event);
    __pyx_t_8 = __pyx_v_event;
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_8, __pyx_n_u_D, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1024, __pyx_L1_error)
    if (!__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_8, __pyx_n_u_N, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1024, __pyx_L1_error)
    if (!__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_8, __pyx_n_u_H, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1024, __pyx_L1_error)
    if (!__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_8, __pyx_n_u_P, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_4;
    __pyx_L10_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if ((__pyx_t_1 != 0)) {
      __Pyx_INCREF(__pyx_int_0);
      __pyx_t_10 = __pyx_int_0;
    } else {
      __Pyx_INCREF(__pyx_v_event_len);
      __pyx_t_10 = __pyx_v_event_len;
    }
    __Pyx_XDECREF_SET(__pyx_v_query_move, __pyx_t_10);
    __pyx_t_10 = 0;

    /* "aldy/indelpost/varaln.pyx":1025
 * 
 *         query_move = 0 if event in ("D", "N", "H", "P") else event_len
 *         ref_move = 0 if event in ("I", "H", "P") else event_len             # <<<<<<<<<<<<<<
 * 
 *         if event in ("I", "D") and prev_event == "N":
 */
    __Pyx_INCREF(__pyx_v_event);
    __pyx_t_8 = __pyx_v_event;
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_8, __pyx_n_u_I, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1025, __pyx_L1_error)
    if (!__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_8, __pyx_n_u_H, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1025, __pyx_L1_error)
    if (!__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_8, __pyx_n_u_P, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1025, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_4;
    __pyx_L14_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if ((__pyx_t_1 != 0)) {
      __Pyx_INCREF(__pyx_int_0);
      __pyx_t_10 = __pyx_int_0;
    } else {
      __Pyx_INCREF(__pyx_v_event_len);
      __pyx_t_10 = __pyx_v_event_len;
    }
    __Pyx_XDECREF_SET(__pyx_v_ref_move, __pyx_t_10);
    __pyx_t_10 = 0;

    /* "aldy/indelpost/varaln.pyx":1027
 *         ref_move = 0 if event in ("I", "H", "P") else event_len
 * 
 *         if event in ("I", "D") and prev_event == "N":             # <<<<<<<<<<<<<<
 *             try:
 *                 nxt_c = cigar_lst[i + 1]
 */
    __Pyx_INCREF(__pyx_v_event);
    __pyx_t_10 = __pyx_v_event;
    __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_10, __pyx_n_u_I, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1027, __pyx_L1_error)
    if (!__pyx_t_3) {
    } else {
      __pyx_t_4 = __pyx_t_3;
      goto __pyx_L20_bool_binop_done;
    }
    __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_10, __pyx_n_u_D, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1027, __pyx_L1_error)
    __pyx_t_4 = __pyx_t_3;
    __pyx_L20_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L18_bool_binop_done;
    }
    __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_prev_event, __pyx_n_u_N, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1027, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_3;
    __pyx_L18_bool_binop_done:;
    if (__pyx_t_1) {

      /* "aldy/indelpost/varaln.pyx":1028
 * 
 *         if event in ("I", "D") and prev_event == "N":
 *             try:             # <<<<<<<<<<<<<<
 *                 nxt_c = cigar_lst[i + 1]
 *                 nxt_event, nxt_event_len = nxt_c[-1], int(nxt_c[:-1])
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_13);
        /*try:*/ {

          /* "aldy/indelpost/varaln.pyx":1029
 *         if event in ("I", "D") and prev_event == "N":
 *             try:
 *                 nxt_c = cigar_lst[i + 1]             # <<<<<<<<<<<<<<
 *                 nxt_event, nxt_event_len = nxt_c[-1], int(nxt_c[:-1])
 *                 if nxt_event != "M":
 */
          __pyx_t_10 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1029, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_cigar_lst, __pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1029, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF_SET(__pyx_v_nxt_c, __pyx_t_8);
          __pyx_t_8 = 0;

          /* "aldy/indelpost/varaln.pyx":1030
 *             try:
 *                 nxt_c = cigar_lst[i + 1]
 *                 nxt_event, nxt_event_len = nxt_c[-1], int(nxt_c[:-1])             # <<<<<<<<<<<<<<
 *                 if nxt_event != "M":
 *                     raise Exception
 */
          __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_nxt_c, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1030, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_10 = __Pyx_PyObject_GetSlice(__pyx_v_nxt_c, 0, -1L, NULL, NULL, &__pyx_slice__5, 0, 1, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1030, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_9 = __Pyx_PyNumber_Int(__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1030, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF_SET(__pyx_v_nxt_event, __pyx_t_8);
          __pyx_t_8 = 0;
          __Pyx_XDECREF_SET(__pyx_v_nxt_event_len, __pyx_t_9);
          __pyx_t_9 = 0;

          /* "aldy/indelpost/varaln.pyx":1031
 *                 nxt_c = cigar_lst[i + 1]
 *                 nxt_event, nxt_event_len = nxt_c[-1], int(nxt_c[:-1])
 *                 if nxt_event != "M":             # <<<<<<<<<<<<<<
 *                     raise Exception
 *             except:
 */
          __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_nxt_event, __pyx_n_u_M, Py_NE)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1031, __pyx_L22_error)
          if (unlikely(__pyx_t_1)) {

            /* "aldy/indelpost/varaln.pyx":1032
 *                 nxt_event, nxt_event_len = nxt_c[-1], int(nxt_c[:-1])
 *                 if nxt_event != "M":
 *                     raise Exception             # <<<<<<<<<<<<<<
 *             except:
 *                 return None
 */
            __Pyx_Raise(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), 0, 0, 0);
            __PYX_ERR(0, 1032, __pyx_L22_error)

            /* "aldy/indelpost/varaln.pyx":1031
 *                 nxt_c = cigar_lst[i + 1]
 *                 nxt_event, nxt_event_len = nxt_c[-1], int(nxt_c[:-1])
 *                 if nxt_event != "M":             # <<<<<<<<<<<<<<
 *                     raise Exception
 *             except:
 */
          }

          /* "aldy/indelpost/varaln.pyx":1028
 * 
 *         if event in ("I", "D") and prev_event == "N":
 *             try:             # <<<<<<<<<<<<<<
 *                 nxt_c = cigar_lst[i + 1]
 *                 nxt_event, nxt_event_len = nxt_c[-1], int(nxt_c[:-1])
 */
        }
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        goto __pyx_L29_try_end;
        __pyx_L22_error:;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "aldy/indelpost/varaln.pyx":1033
 *                 if nxt_event != "M":
 *                     raise Exception
 *             except:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */
        /*except:*/ {
          __Pyx_AddTraceback("aldy.indelpost.varaln.right_aligner", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(0, 1033, __pyx_L24_except_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GOTREF(__pyx_t_10);

          /* "aldy/indelpost/varaln.pyx":1034
 *                     raise Exception
 *             except:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *             chrom, reference = target.chrom, target.reference
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_r = Py_None; __Pyx_INCREF(Py_None);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L25_except_return;
        }
        __pyx_L24_except_error:;

        /* "aldy/indelpost/varaln.pyx":1028
 * 
 *         if event in ("I", "D") and prev_event == "N":
 *             try:             # <<<<<<<<<<<<<<
 *                 nxt_c = cigar_lst[i + 1]
 *                 nxt_event, nxt_event_len = nxt_c[-1], int(nxt_c[:-1])
 */
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
        goto __pyx_L1_error;
        __pyx_L25_except_return:;
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
        goto __pyx_L0;
        __pyx_L29_try_end:;
      }

      /* "aldy/indelpost/varaln.pyx":1036
 *                 return None
 * 
 *             chrom, reference = target.chrom, target.reference             # <<<<<<<<<<<<<<
 *             padding_base = reference.fetch(chrom, ref_pos - 2, ref_pos - 1)
 *             if event == "I":
 */
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_target, __pyx_n_s_chrom); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1036, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_target, __pyx_n_s_reference); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1036, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_v_chrom, __pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_XDECREF_SET(__pyx_v_reference, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "aldy/indelpost/varaln.pyx":1037
 * 
 *             chrom, reference = target.chrom, target.reference
 *             padding_base = reference.fetch(chrom, ref_pos - 2, ref_pos - 1)             # <<<<<<<<<<<<<<
 *             if event == "I":
 *                 ins_seq = read["read_seq"][query_pos: query_pos + event_len]
 */
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_fetch); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = __Pyx_PyInt_SubtractObjC(__pyx_v_ref_pos, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_14 = __Pyx_PyInt_SubtractObjC(__pyx_v_ref_pos, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_15 = NULL;
      __pyx_t_16 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_15)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_15);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_16 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[4] = {__pyx_t_15, __pyx_v_chrom, __pyx_t_9, __pyx_t_14};
        __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_16, 3+__pyx_t_16); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1037, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[4] = {__pyx_t_15, __pyx_v_chrom, __pyx_t_9, __pyx_t_14};
        __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_16, 3+__pyx_t_16); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1037, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      } else
      #endif
      {
        __pyx_t_17 = PyTuple_New(3+__pyx_t_16); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1037, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        if (__pyx_t_15) {
          __Pyx_GIVEREF(__pyx_t_15); PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_15); __pyx_t_15 = NULL;
        }
        __Pyx_INCREF(__pyx_v_chrom);
        __Pyx_GIVEREF(__pyx_v_chrom);
        PyTuple_SET_ITEM(__pyx_t_17, 0+__pyx_t_16, __pyx_v_chrom);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_17, 1+__pyx_t_16, __pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_14);
        PyTuple_SET_ITEM(__pyx_t_17, 2+__pyx_t_16, __pyx_t_14);
        __pyx_t_9 = 0;
        __pyx_t_14 = 0;
        __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_17, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1037, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF_SET(__pyx_v_padding_base, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "aldy/indelpost/varaln.pyx":1038
 *             chrom, reference = target.chrom, target.reference
 *             padding_base = reference.fetch(chrom, ref_pos - 2, ref_pos - 1)
 *             if event == "I":             # <<<<<<<<<<<<<<
 *                 ins_seq = read["read_seq"][query_pos: query_pos + event_len]
 *                 ref = padding_base
 */
      __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_event, __pyx_n_u_I, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1038, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "aldy/indelpost/varaln.pyx":1039
 *             padding_base = reference.fetch(chrom, ref_pos - 2, ref_pos - 1)
 *             if event == "I":
 *                 ins_seq = read["read_seq"][query_pos: query_pos + event_len]             # <<<<<<<<<<<<<<
 *                 ref = padding_base
 *                 alt = padding_base + ins_seq
 */
        __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_seq); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_10 = PyNumber_Add(__pyx_v_query_pos, __pyx_v_event_len); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_17 = __Pyx_PyObject_GetSlice(__pyx_t_8, 0, 0, &__pyx_v_query_pos, &__pyx_t_10, NULL, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF_SET(__pyx_v_ins_seq, __pyx_t_17);
        __pyx_t_17 = 0;

        /* "aldy/indelpost/varaln.pyx":1040
 *             if event == "I":
 *                 ins_seq = read["read_seq"][query_pos: query_pos + event_len]
 *                 ref = padding_base             # <<<<<<<<<<<<<<
 *                 alt = padding_base + ins_seq
 *             else:
 */
        __Pyx_INCREF(__pyx_v_padding_base);
        __Pyx_XDECREF_SET(__pyx_v_ref, __pyx_v_padding_base);

        /* "aldy/indelpost/varaln.pyx":1041
 *                 ins_seq = read["read_seq"][query_pos: query_pos + event_len]
 *                 ref = padding_base
 *                 alt = padding_base + ins_seq             # <<<<<<<<<<<<<<
 *             else:
 *                 del_seq = reference.fetch(chrom, ref_pos - 1, ref_pos - 1 + event_len)
 */
        __pyx_t_17 = PyNumber_Add(__pyx_v_padding_base, __pyx_v_ins_seq); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_XDECREF_SET(__pyx_v_alt, __pyx_t_17);
        __pyx_t_17 = 0;

        /* "aldy/indelpost/varaln.pyx":1038
 *             chrom, reference = target.chrom, target.reference
 *             padding_base = reference.fetch(chrom, ref_pos - 2, ref_pos - 1)
 *             if event == "I":             # <<<<<<<<<<<<<<
 *                 ins_seq = read["read_seq"][query_pos: query_pos + event_len]
 *                 ref = padding_base
 */
        goto __pyx_L33;
      }

      /* "aldy/indelpost/varaln.pyx":1043
 *                 alt = padding_base + ins_seq
 *             else:
 *                 del_seq = reference.fetch(chrom, ref_pos - 1, ref_pos - 1 + event_len)             # <<<<<<<<<<<<<<
 *                 ref = padding_base + del_seq
 *                 alt = padding_base
 */
      /*else*/ {
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_fetch); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1043, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_8 = __Pyx_PyInt_SubtractObjC(__pyx_v_ref_pos, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1043, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_14 = __Pyx_PyInt_SubtractObjC(__pyx_v_ref_pos, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1043, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_9 = PyNumber_Add(__pyx_t_14, __pyx_v_event_len); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1043, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_14 = NULL;
        __pyx_t_16 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_14)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_14);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_16 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[4] = {__pyx_t_14, __pyx_v_chrom, __pyx_t_8, __pyx_t_9};
          __pyx_t_17 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_16, 3+__pyx_t_16); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1043, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[4] = {__pyx_t_14, __pyx_v_chrom, __pyx_t_8, __pyx_t_9};
          __pyx_t_17 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_16, 3+__pyx_t_16); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1043, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        } else
        #endif
        {
          __pyx_t_15 = PyTuple_New(3+__pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1043, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          if (__pyx_t_14) {
            __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_14); __pyx_t_14 = NULL;
          }
          __Pyx_INCREF(__pyx_v_chrom);
          __Pyx_GIVEREF(__pyx_v_chrom);
          PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_16, __pyx_v_chrom);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_16, __pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_15, 2+__pyx_t_16, __pyx_t_9);
          __pyx_t_8 = 0;
          __pyx_t_9 = 0;
          __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_15, NULL); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1043, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF_SET(__pyx_v_del_seq, __pyx_t_17);
        __pyx_t_17 = 0;

        /* "aldy/indelpost/varaln.pyx":1044
 *             else:
 *                 del_seq = reference.fetch(chrom, ref_pos - 1, ref_pos - 1 + event_len)
 *                 ref = padding_base + del_seq             # <<<<<<<<<<<<<<
 *                 alt = padding_base
 * 
 */
        __pyx_t_17 = PyNumber_Add(__pyx_v_padding_base, __pyx_v_del_seq); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1044, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_XDECREF_SET(__pyx_v_ref, __pyx_t_17);
        __pyx_t_17 = 0;

        /* "aldy/indelpost/varaln.pyx":1045
 *                 del_seq = reference.fetch(chrom, ref_pos - 1, ref_pos - 1 + event_len)
 *                 ref = padding_base + del_seq
 *                 alt = padding_base             # <<<<<<<<<<<<<<
 * 
 *             right_aligned_vars = Variant(
 */
        __Pyx_INCREF(__pyx_v_padding_base);
        __Pyx_XDECREF_SET(__pyx_v_alt, __pyx_v_padding_base);
      }
      __pyx_L33:;

      /* "aldy/indelpost/varaln.pyx":1049
 *             right_aligned_vars = Variant(
 *                                     chrom,
 *                                     ref_pos - 1,             # <<<<<<<<<<<<<<
 *                                     ref,
 *                                     alt,
 */
      __pyx_t_10 = __Pyx_PyInt_SubtractObjC(__pyx_v_ref_pos, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);

      /* "aldy/indelpost/varaln.pyx":1047
 *                 alt = padding_base
 * 
 *             right_aligned_vars = Variant(             # <<<<<<<<<<<<<<
 *                                     chrom,
 *                                     ref_pos - 1,
 */
      __pyx_t_15 = PyTuple_New(5); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1047, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_INCREF(__pyx_v_chrom);
      __Pyx_GIVEREF(__pyx_v_chrom);
      PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v_chrom);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_10);
      __Pyx_INCREF(__pyx_v_ref);
      __Pyx_GIVEREF(__pyx_v_ref);
      PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_ref);
      __Pyx_INCREF(__pyx_v_alt);
      __Pyx_GIVEREF(__pyx_v_alt);
      PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_v_alt);
      __Pyx_INCREF(__pyx_v_reference);
      __Pyx_GIVEREF(__pyx_v_reference);
      PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_v_reference);
      __pyx_t_10 = 0;

      /* "aldy/indelpost/varaln.pyx":1053
 *                                     alt,
 *                                     reference,
 *                                     skip_validation=True             # <<<<<<<<<<<<<<
 *                                 ).generate_equivalents()
 * 
 */
      __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1053, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_skip_validation, Py_True) < 0) __PYX_ERR(0, 1053, __pyx_L1_error)

      /* "aldy/indelpost/varaln.pyx":1047
 *                 alt = padding_base
 * 
 *             right_aligned_vars = Variant(             # <<<<<<<<<<<<<<
 *                                     chrom,
 *                                     ref_pos - 1,
 */
      __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_4aldy_9indelpost_7variant_Variant), __pyx_t_15, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1047, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "aldy/indelpost/varaln.pyx":1054
 *                                     reference,
 *                                     skip_validation=True
 *                                 ).generate_equivalents()             # <<<<<<<<<<<<<<
 * 
 *             diff = max(v.pos for v in right_aligned_vars) - ref_pos + 1
 */
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_generate_equivalents); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_17 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_right_aligned_vars);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_right_aligned_vars, __pyx_t_17);
      __Pyx_GIVEREF(__pyx_t_17);
      __pyx_t_17 = 0;

      /* "aldy/indelpost/varaln.pyx":1056
 *                                 ).generate_equivalents()
 * 
 *             diff = max(v.pos for v in right_aligned_vars) - ref_pos + 1             # <<<<<<<<<<<<<<
 *             if diff > 0:
 *                 new_cigar += [str(diff) + "M", str(event_len) + event, str(nxt_event_len - diff) + "M"]
 */
      __pyx_t_17 = __pyx_pf_4aldy_9indelpost_6varaln_13right_aligner_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1056, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_17); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1056, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_17 = PyNumber_Subtract(__pyx_t_10, __pyx_v_ref_pos); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1056, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyInt_AddObjC(__pyx_t_17, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1056, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_XDECREF_SET(__pyx_v_diff, __pyx_t_10);
      __pyx_t_10 = 0;

      /* "aldy/indelpost/varaln.pyx":1057
 * 
 *             diff = max(v.pos for v in right_aligned_vars) - ref_pos + 1
 *             if diff > 0:             # <<<<<<<<<<<<<<
 *                 new_cigar += [str(diff) + "M", str(event_len) + event, str(nxt_event_len - diff) + "M"]
 *             else:
 */
      __pyx_t_10 = PyObject_RichCompare(__pyx_v_diff, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1057, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1057, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (__pyx_t_1) {

        /* "aldy/indelpost/varaln.pyx":1058
 *             diff = max(v.pos for v in right_aligned_vars) - ref_pos + 1
 *             if diff > 0:
 *                 new_cigar += [str(diff) + "M", str(event_len) + event, str(nxt_event_len - diff) + "M"]             # <<<<<<<<<<<<<<
 *             else:
 *                 return None
 */
        __pyx_t_10 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_diff); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_17 = __Pyx_PyUnicode_Concat(__pyx_t_10, __pyx_n_u_M); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_event_len); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_9 = PyNumber_Add(__pyx_t_10, __pyx_v_event); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = PyNumber_Subtract(__pyx_v_nxt_event_len, __pyx_v_diff); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_15 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_10); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = __Pyx_PyUnicode_Concat(__pyx_t_15, __pyx_n_u_M); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_15 = PyList_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_GIVEREF(__pyx_t_17);
        PyList_SET_ITEM(__pyx_t_15, 0, __pyx_t_17);
        __Pyx_GIVEREF(__pyx_t_9);
        PyList_SET_ITEM(__pyx_t_15, 1, __pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_10);
        PyList_SET_ITEM(__pyx_t_15, 2, __pyx_t_10);
        __pyx_t_17 = 0;
        __pyx_t_9 = 0;
        __pyx_t_10 = 0;
        __pyx_t_10 = PyNumber_InPlaceAdd(__pyx_v_new_cigar, __pyx_t_15); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF_SET(__pyx_v_new_cigar, ((PyObject*)__pyx_t_10));
        __pyx_t_10 = 0;

        /* "aldy/indelpost/varaln.pyx":1057
 * 
 *             diff = max(v.pos for v in right_aligned_vars) - ref_pos + 1
 *             if diff > 0:             # <<<<<<<<<<<<<<
 *                 new_cigar += [str(diff) + "M", str(event_len) + event, str(nxt_event_len - diff) + "M"]
 *             else:
 */
        goto __pyx_L34;
      }

      /* "aldy/indelpost/varaln.pyx":1060
 *                 new_cigar += [str(diff) + "M", str(event_len) + event, str(nxt_event_len - diff) + "M"]
 *             else:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *             ref_pos += query_move + nxt_event_len
 */
      /*else*/ {
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L0;
      }
      __pyx_L34:;

      /* "aldy/indelpost/varaln.pyx":1062
 *                 return None
 * 
 *             ref_pos += query_move + nxt_event_len             # <<<<<<<<<<<<<<
 *             query_pos += ref_move + nxt_event_len
 *             skip_next = True
 */
      __pyx_t_10 = PyNumber_Add(__pyx_v_query_move, __pyx_v_nxt_event_len); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1062, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_15 = PyNumber_InPlaceAdd(__pyx_v_ref_pos, __pyx_t_10); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1062, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF_SET(__pyx_v_ref_pos, __pyx_t_15);
      __pyx_t_15 = 0;

      /* "aldy/indelpost/varaln.pyx":1063
 * 
 *             ref_pos += query_move + nxt_event_len
 *             query_pos += ref_move + nxt_event_len             # <<<<<<<<<<<<<<
 *             skip_next = True
 * 
 */
      __pyx_t_15 = PyNumber_Add(__pyx_v_ref_move, __pyx_v_nxt_event_len); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1063, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_10 = PyNumber_InPlaceAdd(__pyx_v_query_pos, __pyx_t_15); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1063, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF_SET(__pyx_v_query_pos, __pyx_t_10);
      __pyx_t_10 = 0;

      /* "aldy/indelpost/varaln.pyx":1064
 *             ref_pos += query_move + nxt_event_len
 *             query_pos += ref_move + nxt_event_len
 *             skip_next = True             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
      __pyx_v_skip_next = 1;

      /* "aldy/indelpost/varaln.pyx":1027
 *         ref_move = 0 if event in ("I", "H", "P") else event_len
 * 
 *         if event in ("I", "D") and prev_event == "N":             # <<<<<<<<<<<<<<
 *             try:
 *                 nxt_c = cigar_lst[i + 1]
 */
      goto __pyx_L17;
    }

    /* "aldy/indelpost/varaln.pyx":1067
 * 
 *         else:
 *             if skip_next:             # <<<<<<<<<<<<<<
 *                 skip_next = False
 *             else:
 */
    /*else*/ {
      __pyx_t_1 = (__pyx_v_skip_next != 0);
      if (__pyx_t_1) {

        /* "aldy/indelpost/varaln.pyx":1068
 *         else:
 *             if skip_next:
 *                 skip_next = False             # <<<<<<<<<<<<<<
 *             else:
 *                 query_pos += query_move
 */
        __pyx_v_skip_next = 0;

        /* "aldy/indelpost/varaln.pyx":1067
 * 
 *         else:
 *             if skip_next:             # <<<<<<<<<<<<<<
 *                 skip_next = False
 *             else:
 */
        goto __pyx_L35;
      }

      /* "aldy/indelpost/varaln.pyx":1070
 *                 skip_next = False
 *             else:
 *                 query_pos += query_move             # <<<<<<<<<<<<<<
 *                 ref_pos += ref_move
 *                 new_cigar.append(c)
 */
      /*else*/ {
        __pyx_t_10 = PyNumber_InPlaceAdd(__pyx_v_query_pos, __pyx_v_query_move); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1070, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF_SET(__pyx_v_query_pos, __pyx_t_10);
        __pyx_t_10 = 0;

        /* "aldy/indelpost/varaln.pyx":1071
 *             else:
 *                 query_pos += query_move
 *                 ref_pos += ref_move             # <<<<<<<<<<<<<<
 *                 new_cigar.append(c)
 * 
 */
        __pyx_t_10 = PyNumber_InPlaceAdd(__pyx_v_ref_pos, __pyx_v_ref_move); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1071, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF_SET(__pyx_v_ref_pos, __pyx_t_10);
        __pyx_t_10 = 0;

        /* "aldy/indelpost/varaln.pyx":1072
 *                 query_pos += query_move
 *                 ref_pos += ref_move
 *                 new_cigar.append(c)             # <<<<<<<<<<<<<<
 * 
 *         prev_event = event
 */
        __pyx_t_18 = __Pyx_PyList_Append(__pyx_v_new_cigar, __pyx_v_c); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 1072, __pyx_L1_error)
      }
      __pyx_L35:;
    }
    __pyx_L17:;

    /* "aldy/indelpost/varaln.pyx":1074
 *                 new_cigar.append(c)
 * 
 *         prev_event = event             # <<<<<<<<<<<<<<
 * 
 *     read["cigar_list"] = new_cigar
 */
    __Pyx_INCREF(__pyx_v_event);
    __Pyx_DECREF_SET(__pyx_v_prev_event, __pyx_v_event);

    /* "aldy/indelpost/varaln.pyx":1018
 *     prev_event = "A"
 *     skip_next = False
 *     for i, c in enumerate(cigar_lst):             # <<<<<<<<<<<<<<
 *         event, event_len = c[-1], int(c[:-1])
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":1076
 *         prev_event = event
 * 
 *     read["cigar_list"] = new_cigar             # <<<<<<<<<<<<<<
 *     read["cigar_string"] = "".join(new_cigar)
 * 
 */
  if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_cigar_list, __pyx_v_new_cigar) < 0)) __PYX_ERR(0, 1076, __pyx_L1_error)

  /* "aldy/indelpost/varaln.pyx":1077
 * 
 *     read["cigar_list"] = new_cigar
 *     read["cigar_string"] = "".join(new_cigar)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __pyx_t_2 = PyUnicode_Join(__pyx_kp_u__6, __pyx_v_new_cigar); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_cigar_string, __pyx_t_2) < 0)) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":1079
 *     read["cigar_string"] = "".join(new_cigar)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if target in right_aligned_vars:
 *             rt_aln_pos = target.pos + diff
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_12, &__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_13);
    __Pyx_XGOTREF(__pyx_t_12);
    __Pyx_XGOTREF(__pyx_t_11);
    /*try:*/ {

      /* "aldy/indelpost/varaln.pyx":1080
 * 
 *     try:
 *         if target in right_aligned_vars:             # <<<<<<<<<<<<<<
 *             rt_aln_pos = target.pos + diff
 *             read["lt_cigar"], read["rt_cigar"] = split_cigar(read["cigar_string"], rt_aln_pos, read["read_start"])
 */
      if (unlikely(!__pyx_cur_scope->__pyx_v_right_aligned_vars)) { __Pyx_RaiseUnboundLocalError("right_aligned_vars"); __PYX_ERR(0, 1080, __pyx_L36_error) }
      __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_target, __pyx_cur_scope->__pyx_v_right_aligned_vars, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1080, __pyx_L36_error)
      __pyx_t_3 = (__pyx_t_1 != 0);
      if (__pyx_t_3) {

        /* "aldy/indelpost/varaln.pyx":1081
 *     try:
 *         if target in right_aligned_vars:
 *             rt_aln_pos = target.pos + diff             # <<<<<<<<<<<<<<
 *             read["lt_cigar"], read["rt_cigar"] = split_cigar(read["cigar_string"], rt_aln_pos, read["read_start"])
 *             read["lt_flank"], read["rt_flank"] = split(
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_target, __pyx_n_s_pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (unlikely(!__pyx_v_diff)) { __Pyx_RaiseUnboundLocalError("diff"); __PYX_ERR(0, 1081, __pyx_L36_error) }
        __pyx_t_5 = PyNumber_Add(__pyx_t_2, __pyx_v_diff); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1081, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_rt_aln_pos = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "aldy/indelpost/varaln.pyx":1082
 *         if target in right_aligned_vars:
 *             rt_aln_pos = target.pos + diff
 *             read["lt_cigar"], read["rt_cigar"] = split_cigar(read["cigar_string"], rt_aln_pos, read["read_start"])             # <<<<<<<<<<<<<<
 *             read["lt_flank"], read["rt_flank"] = split(
 *                                                       read["read_seq"],
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_split_cigar); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1082, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_10 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1082, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_15 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_start); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1082, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_15);
        __pyx_t_9 = NULL;
        __pyx_t_16 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_16 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_9, __pyx_t_10, __pyx_v_rt_aln_pos, __pyx_t_15};
          __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_16, 3+__pyx_t_16); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1082, __pyx_L36_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[4] = {__pyx_t_9, __pyx_t_10, __pyx_v_rt_aln_pos, __pyx_t_15};
          __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_16, 3+__pyx_t_16); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1082, __pyx_L36_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        } else
        #endif
        {
          __pyx_t_17 = PyTuple_New(3+__pyx_t_16); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1082, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_17);
          if (__pyx_t_9) {
            __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_9); __pyx_t_9 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_17, 0+__pyx_t_16, __pyx_t_10);
          __Pyx_INCREF(__pyx_v_rt_aln_pos);
          __Pyx_GIVEREF(__pyx_v_rt_aln_pos);
          PyTuple_SET_ITEM(__pyx_t_17, 1+__pyx_t_16, __pyx_v_rt_aln_pos);
          __Pyx_GIVEREF(__pyx_t_15);
          PyTuple_SET_ITEM(__pyx_t_17, 2+__pyx_t_16, __pyx_t_15);
          __pyx_t_10 = 0;
          __pyx_t_15 = 0;
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_17, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1082, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
          PyObject* sequence = __pyx_t_5;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 1082, __pyx_L36_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_17 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_17 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_17);
          #else
          __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1082, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_17 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1082, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_17);
          #endif
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_15 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1082, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_19 = Py_TYPE(__pyx_t_15)->tp_iternext;
          index = 0; __pyx_t_2 = __pyx_t_19(__pyx_t_15); if (unlikely(!__pyx_t_2)) goto __pyx_L43_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_2);
          index = 1; __pyx_t_17 = __pyx_t_19(__pyx_t_15); if (unlikely(!__pyx_t_17)) goto __pyx_L43_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_17);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_15), 2) < 0) __PYX_ERR(0, 1082, __pyx_L36_error)
          __pyx_t_19 = NULL;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          goto __pyx_L44_unpacking_done;
          __pyx_L43_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_19 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 1082, __pyx_L36_error)
          __pyx_L44_unpacking_done:;
        }
        if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_lt_cigar, __pyx_t_2) < 0)) __PYX_ERR(0, 1082, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_rt_cigar, __pyx_t_17) < 0)) __PYX_ERR(0, 1082, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

        /* "aldy/indelpost/varaln.pyx":1084
 *             read["lt_cigar"], read["rt_cigar"] = split_cigar(read["cigar_string"], rt_aln_pos, read["read_start"])
 *             read["lt_flank"], read["rt_flank"] = split(
 *                                                       read["read_seq"],             # <<<<<<<<<<<<<<
 *                                                       read["cigar_string"],
 *                                                       rt_aln_pos,
 */
        __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_seq); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1084, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_5);

        /* "aldy/indelpost/varaln.pyx":1085
 *             read["lt_flank"], read["rt_flank"] = split(
 *                                                       read["read_seq"],
 *                                                       read["cigar_string"],             # <<<<<<<<<<<<<<
 *                                                       rt_aln_pos,
 *                                                       read["read_start"],
 */
        __pyx_t_17 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1085, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_17);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_17))||((__pyx_t_17) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_17)->tp_name), 0))) __PYX_ERR(0, 1085, __pyx_L36_error)

        /* "aldy/indelpost/varaln.pyx":1086
 *                                                       read["read_seq"],
 *                                                       read["cigar_string"],
 *                                                       rt_aln_pos,             # <<<<<<<<<<<<<<
 *                                                       read["read_start"],
 *                                                       is_for_ref=False,
 */
        __pyx_t_16 = __Pyx_PyInt_As_int(__pyx_v_rt_aln_pos); if (unlikely((__pyx_t_16 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1086, __pyx_L36_error)

        /* "aldy/indelpost/varaln.pyx":1087
 *                                                       read["cigar_string"],
 *                                                       rt_aln_pos,
 *                                                       read["read_start"],             # <<<<<<<<<<<<<<
 *                                                       is_for_ref=False,
 *                                                       reverse=False
 */
        __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_start); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1087, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_20 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_20 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1087, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "aldy/indelpost/varaln.pyx":1083
 *             rt_aln_pos = target.pos + diff
 *             read["lt_cigar"], read["rt_cigar"] = split_cigar(read["cigar_string"], rt_aln_pos, read["read_start"])
 *             read["lt_flank"], read["rt_flank"] = split(             # <<<<<<<<<<<<<<
 *                                                       read["read_seq"],
 *                                                       read["cigar_string"],
 */
        __pyx_t_2 = __pyx_f_4aldy_9indelpost_9utilities_split(__pyx_t_5, ((PyObject*)__pyx_t_17), __pyx_t_16, __pyx_t_20, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1083, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        if (likely(__pyx_t_2 != Py_None)) {
          PyObject* sequence = __pyx_t_2;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 1083, __pyx_L36_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_17 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
          __Pyx_INCREF(__pyx_t_17);
          __Pyx_INCREF(__pyx_t_5);
          #else
          __pyx_t_17 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1083, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1083, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        } else {
          __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 1083, __pyx_L36_error)
        }
        if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_lt_flank, __pyx_t_17) < 0)) __PYX_ERR(0, 1083, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_rt_flank, __pyx_t_5) < 0)) __PYX_ERR(0, 1083, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "aldy/indelpost/varaln.pyx":1092
 *                                                  )
 *             read["lt_qual"], read["rt_qual"] = split(
 *                                                     read["read_qual"],             # <<<<<<<<<<<<<<
 *                                                     read["cigar_string"],
 *                                                     rt_aln_pos,
 */
        __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_qual); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1092, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_2);

        /* "aldy/indelpost/varaln.pyx":1093
 *             read["lt_qual"], read["rt_qual"] = split(
 *                                                     read["read_qual"],
 *                                                     read["cigar_string"],             # <<<<<<<<<<<<<<
 *                                                     rt_aln_pos,
 *                                                     read["read_start"],
 */
        __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1093, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 1093, __pyx_L36_error)

        /* "aldy/indelpost/varaln.pyx":1094
 *                                                     read["read_qual"],
 *                                                     read["cigar_string"],
 *                                                     rt_aln_pos,             # <<<<<<<<<<<<<<
 *                                                     read["read_start"],
 *                                                     is_for_ref=False,
 */
        __pyx_t_20 = __Pyx_PyInt_As_int(__pyx_v_rt_aln_pos); if (unlikely((__pyx_t_20 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1094, __pyx_L36_error)

        /* "aldy/indelpost/varaln.pyx":1095
 *                                                     read["cigar_string"],
 *                                                     rt_aln_pos,
 *                                                     read["read_start"],             # <<<<<<<<<<<<<<
 *                                                     is_for_ref=False,
 *                                                     reverse=False
 */
        __pyx_t_17 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_start); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1095, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_17);
        __pyx_t_16 = __Pyx_PyInt_As_int(__pyx_t_17); if (unlikely((__pyx_t_16 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1095, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

        /* "aldy/indelpost/varaln.pyx":1091
 *                                                       reverse=False
 *                                                  )
 *             read["lt_qual"], read["rt_qual"] = split(             # <<<<<<<<<<<<<<
 *                                                     read["read_qual"],
 *                                                     read["cigar_string"],
 */
        __pyx_t_17 = __pyx_f_4aldy_9indelpost_9utilities_split(__pyx_t_2, ((PyObject*)__pyx_t_5), __pyx_t_20, __pyx_t_16, 0, 0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1091, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (likely(__pyx_t_17 != Py_None)) {
          PyObject* sequence = __pyx_t_17;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 1091, __pyx_L36_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_2);
          #else
          __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1091, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1091, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        } else {
          __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 1091, __pyx_L36_error)
        }
        if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_lt_qual, __pyx_t_5) < 0)) __PYX_ERR(0, 1091, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_rt_qual, __pyx_t_2) < 0)) __PYX_ERR(0, 1091, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "aldy/indelpost/varaln.pyx":1100
 *                                                )
 *             read["lt_ref"], read["rt_ref"] = split(
 *                                                     read["ref_seq"],             # <<<<<<<<<<<<<<
 *                                                     read["cigar_string"],
 *                                                     rt_aln_pos, read["aln_start"],
 */
        __pyx_t_17 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_ref_seq); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1100, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_17);

        /* "aldy/indelpost/varaln.pyx":1101
 *             read["lt_ref"], read["rt_ref"] = split(
 *                                                     read["ref_seq"],
 *                                                     read["cigar_string"],             # <<<<<<<<<<<<<<
 *                                                     rt_aln_pos, read["aln_start"],
 *                                                     is_for_ref=True,
 */
        __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1101, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 1101, __pyx_L36_error)

        /* "aldy/indelpost/varaln.pyx":1102
 *                                                     read["ref_seq"],
 *                                                     read["cigar_string"],
 *                                                     rt_aln_pos, read["aln_start"],             # <<<<<<<<<<<<<<
 *                                                     is_for_ref=True,
 *                                                     reverse=False
 */
        __pyx_t_16 = __Pyx_PyInt_As_int(__pyx_v_rt_aln_pos); if (unlikely((__pyx_t_16 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1102, __pyx_L36_error)
        __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_start); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1102, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_20 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_20 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1102, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "aldy/indelpost/varaln.pyx":1099
 *                                                     reverse=False
 *                                                )
 *             read["lt_ref"], read["rt_ref"] = split(             # <<<<<<<<<<<<<<
 *                                                     read["ref_seq"],
 *                                                     read["cigar_string"],
 */
        __pyx_t_5 = __pyx_f_4aldy_9indelpost_9utilities_split(__pyx_t_17, ((PyObject*)__pyx_t_2), __pyx_t_16, __pyx_t_20, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1099, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (likely(__pyx_t_5 != Py_None)) {
          PyObject* sequence = __pyx_t_5;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 1099, __pyx_L36_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_17 = PyTuple_GET_ITEM(sequence, 1); 
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_17);
          #else
          __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1099, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_17 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1099, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_17);
          #endif
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else {
          __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 1099, __pyx_L36_error)
        }
        if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_lt_ref, __pyx_t_2) < 0)) __PYX_ERR(0, 1099, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_rt_ref, __pyx_t_17) < 0)) __PYX_ERR(0, 1099, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

        /* "aldy/indelpost/varaln.pyx":1106
 *                                                     reverse=False
 *                                              )
 *             read["target_right_shifted"] = rt_aln_pos             # <<<<<<<<<<<<<<
 * 
 *             indel_len = len(target.indel_seq)
 */
        if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_target_right_shifted, __pyx_v_rt_aln_pos) < 0)) __PYX_ERR(0, 1106, __pyx_L36_error)

        /* "aldy/indelpost/varaln.pyx":1108
 *             read["target_right_shifted"] = rt_aln_pos
 * 
 *             indel_len = len(target.indel_seq)             # <<<<<<<<<<<<<<
 *             if target.is_ins:
 *                 read["rt_flank"] = read["rt_flank"][indel_len :]
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_target, __pyx_n_s_indel_seq); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1108, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyObject_Length(__pyx_t_5); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1108, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_indel_len = __pyx_t_6;

        /* "aldy/indelpost/varaln.pyx":1109
 * 
 *             indel_len = len(target.indel_seq)
 *             if target.is_ins:             # <<<<<<<<<<<<<<
 *                 read["rt_flank"] = read["rt_flank"][indel_len :]
 *                 read["rt_qual"] = read["rt_qual"][indel_len :]
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_target, __pyx_n_s_is_ins); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1109, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1109, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (__pyx_t_3) {

          /* "aldy/indelpost/varaln.pyx":1110
 *             indel_len = len(target.indel_seq)
 *             if target.is_ins:
 *                 read["rt_flank"] = read["rt_flank"][indel_len :]             # <<<<<<<<<<<<<<
 *                 read["rt_qual"] = read["rt_qual"][indel_len :]
 *             else:
 */
          __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_rt_flank); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1110, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_17 = __Pyx_PyObject_GetSlice(__pyx_t_5, __pyx_v_indel_len, 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1110, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_rt_flank, __pyx_t_17) < 0)) __PYX_ERR(0, 1110, __pyx_L36_error)
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

          /* "aldy/indelpost/varaln.pyx":1111
 *             if target.is_ins:
 *                 read["rt_flank"] = read["rt_flank"][indel_len :]
 *                 read["rt_qual"] = read["rt_qual"][indel_len :]             # <<<<<<<<<<<<<<
 *             else:
 *                 read["rt_ref"] = read["rt_ref"][indel_len :]
 */
          __pyx_t_17 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_rt_qual); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1111, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_t_17, __pyx_v_indel_len, 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1111, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_rt_qual, __pyx_t_5) < 0)) __PYX_ERR(0, 1111, __pyx_L36_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "aldy/indelpost/varaln.pyx":1109
 * 
 *             indel_len = len(target.indel_seq)
 *             if target.is_ins:             # <<<<<<<<<<<<<<
 *                 read["rt_flank"] = read["rt_flank"][indel_len :]
 *                 read["rt_qual"] = read["rt_qual"][indel_len :]
 */
          goto __pyx_L45;
        }

        /* "aldy/indelpost/varaln.pyx":1113
 *                 read["rt_qual"] = read["rt_qual"][indel_len :]
 *             else:
 *                 read["rt_ref"] = read["rt_ref"][indel_len :]             # <<<<<<<<<<<<<<
 *         else:
 *             read["lt_cigar"], read["rt_cigar"] = split_cigar(read["cigar_string"], target.pos, read["read_start"])
 */
        /*else*/ {
          __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_rt_ref); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1113, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_17 = __Pyx_PyObject_GetSlice(__pyx_t_5, __pyx_v_indel_len, 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1113, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_rt_ref, __pyx_t_17) < 0)) __PYX_ERR(0, 1113, __pyx_L36_error)
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        }
        __pyx_L45:;

        /* "aldy/indelpost/varaln.pyx":1080
 * 
 *     try:
 *         if target in right_aligned_vars:             # <<<<<<<<<<<<<<
 *             rt_aln_pos = target.pos + diff
 *             read["lt_cigar"], read["rt_cigar"] = split_cigar(read["cigar_string"], rt_aln_pos, read["read_start"])
 */
        goto __pyx_L42;
      }

      /* "aldy/indelpost/varaln.pyx":1115
 *                 read["rt_ref"] = read["rt_ref"][indel_len :]
 *         else:
 *             read["lt_cigar"], read["rt_cigar"] = split_cigar(read["cigar_string"], target.pos, read["read_start"])             # <<<<<<<<<<<<<<
 *     except:
 *         pass
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_split_cigar); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1115, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1115, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_target, __pyx_n_s_pos); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1115, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_15);
        __pyx_t_10 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_start); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1115, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_9 = NULL;
        __pyx_t_20 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_20 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[4] = {__pyx_t_9, __pyx_t_2, __pyx_t_15, __pyx_t_10};
          __pyx_t_17 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_20, 3+__pyx_t_20); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1115, __pyx_L36_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[4] = {__pyx_t_9, __pyx_t_2, __pyx_t_15, __pyx_t_10};
          __pyx_t_17 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_20, 3+__pyx_t_20); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1115, __pyx_L36_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(3+__pyx_t_20); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1115, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_9) {
            __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_9); __pyx_t_9 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_20, __pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_15);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_20, __pyx_t_15);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_20, __pyx_t_10);
          __pyx_t_2 = 0;
          __pyx_t_15 = 0;
          __pyx_t_10 = 0;
          __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1115, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_17))) || (PyList_CheckExact(__pyx_t_17))) {
          PyObject* sequence = __pyx_t_17;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 1115, __pyx_L36_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          #else
          __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1115, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1115, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_10 = PyObject_GetIter(__pyx_t_17); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1115, __pyx_L36_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __pyx_t_19 = Py_TYPE(__pyx_t_10)->tp_iternext;
          index = 0; __pyx_t_5 = __pyx_t_19(__pyx_t_10); if (unlikely(!__pyx_t_5)) goto __pyx_L46_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_5);
          index = 1; __pyx_t_8 = __pyx_t_19(__pyx_t_10); if (unlikely(!__pyx_t_8)) goto __pyx_L46_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_8);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_10), 2) < 0) __PYX_ERR(0, 1115, __pyx_L36_error)
          __pyx_t_19 = NULL;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L47_unpacking_done;
          __pyx_L46_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_19 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 1115, __pyx_L36_error)
          __pyx_L47_unpacking_done:;
        }
        if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_lt_cigar, __pyx_t_5) < 0)) __PYX_ERR(0, 1115, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_rt_cigar, __pyx_t_8) < 0)) __PYX_ERR(0, 1115, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __pyx_L42:;

      /* "aldy/indelpost/varaln.pyx":1079
 *     read["cigar_string"] = "".join(new_cigar)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if target in right_aligned_vars:
 *             rt_aln_pos = target.pos + diff
 */
    }
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L41_try_end;
    __pyx_L36_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "aldy/indelpost/varaln.pyx":1116
 *         else:
 *             read["lt_cigar"], read["rt_cigar"] = split_cigar(read["cigar_string"], target.pos, read["read_start"])
 *     except:             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
    /*except:*/ {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L37_exception_handled;
    }
    __pyx_L37_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_13);
    __Pyx_XGIVEREF(__pyx_t_12);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_12, __pyx_t_11);
    __pyx_L41_try_end:;
  }

  /* "aldy/indelpost/varaln.pyx":1119
 *         pass
 * 
 *     return read             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_read);
  __pyx_r = __pyx_v_read;
  goto __pyx_L0;

  /* "aldy/indelpost/varaln.pyx":999
 *     return right_aligner(read, target)
 * 
 * def right_aligner(read, target):             # <<<<<<<<<<<<<<
 *     """Right align indels around splice site"""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_AddTraceback("aldy.indelpost.varaln.right_aligner", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cigar_lst);
  __Pyx_XDECREF(__pyx_v_query_pos);
  __Pyx_XDECREF(__pyx_v_ref_pos);
  __Pyx_XDECREF(__pyx_v_new_cigar);
  __Pyx_XDECREF(__pyx_v_prev_event);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XDECREF(__pyx_v_event);
  __Pyx_XDECREF(__pyx_v_event_len);
  __Pyx_XDECREF(__pyx_v_query_move);
  __Pyx_XDECREF(__pyx_v_ref_move);
  __Pyx_XDECREF(__pyx_v_nxt_c);
  __Pyx_XDECREF(__pyx_v_nxt_event);
  __Pyx_XDECREF(__pyx_v_nxt_event_len);
  __Pyx_XDECREF(__pyx_v_chrom);
  __Pyx_XDECREF(__pyx_v_reference);
  __Pyx_XDECREF(__pyx_v_padding_base);
  __Pyx_XDECREF(__pyx_v_ins_seq);
  __Pyx_XDECREF(__pyx_v_ref);
  __Pyx_XDECREF(__pyx_v_alt);
  __Pyx_XDECREF(__pyx_v_del_seq);
  __Pyx_XDECREF(__pyx_v_diff);
  __Pyx_XDECREF(__pyx_v_rt_aln_pos);
  __Pyx_XDECREF(__pyx_gb_4aldy_9indelpost_6varaln_13right_aligner_2generator2);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":1122
 * 
 * 
 * def generate_grid (auto_adjust_extension_penalty,             # <<<<<<<<<<<<<<
 *                    gap_open_penalty,
 *                    gap_extension_penalty,
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_11generate_grid(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_10generate_grid[] = "generate_grid(auto_adjust_extension_penalty, gap_open_penalty, gap_extension_penalty, target)";
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6varaln_11generate_grid = {"generate_grid", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6varaln_11generate_grid, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4aldy_9indelpost_6varaln_10generate_grid};
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_11generate_grid(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_auto_adjust_extension_penalty = 0;
  PyObject *__pyx_v_gap_open_penalty = 0;
  PyObject *__pyx_v_gap_extension_penalty = 0;
  PyObject *__pyx_v_target = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("generate_grid (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_auto_adjust_extension_penalty,&__pyx_n_s_gap_open_penalty,&__pyx_n_s_gap_extension_penalty,&__pyx_n_s_target,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_auto_adjust_extension_penalty)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gap_open_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("generate_grid", 1, 4, 4, 1); __PYX_ERR(0, 1122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gap_extension_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("generate_grid", 1, 4, 4, 2); __PYX_ERR(0, 1122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_target)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("generate_grid", 1, 4, 4, 3); __PYX_ERR(0, 1122, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "generate_grid") < 0)) __PYX_ERR(0, 1122, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_auto_adjust_extension_penalty = values[0];
    __pyx_v_gap_open_penalty = values[1];
    __pyx_v_gap_extension_penalty = values[2];
    __pyx_v_target = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("generate_grid", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1122, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.generate_grid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_10generate_grid(__pyx_self, __pyx_v_auto_adjust_extension_penalty, __pyx_v_gap_open_penalty, __pyx_v_gap_extension_penalty, __pyx_v_target);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_10generate_grid(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_auto_adjust_extension_penalty, PyObject *__pyx_v_gap_open_penalty, PyObject *__pyx_v_gap_extension_penalty, PyObject *__pyx_v_target) {
  PyObject *__pyx_v_grid = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("generate_grid", 0);

  /* "aldy/indelpost/varaln.pyx":1127
 *                    target,
 * ):
 *     if auto_adjust_extension_penalty:             # <<<<<<<<<<<<<<
 *         if (gap_open_penalty, gap_extension_penalty) != (3, 1):
 *             if len(target.indel_seq) < 20:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_auto_adjust_extension_penalty); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1127, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "aldy/indelpost/varaln.pyx":1128
 * ):
 *     if auto_adjust_extension_penalty:
 *         if (gap_open_penalty, gap_extension_penalty) != (3, 1):             # <<<<<<<<<<<<<<
 *             if len(target.indel_seq) < 20:
 *                 grid = [(gap_open_penalty, gap_extension_penalty),
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_gap_open_penalty);
    __Pyx_GIVEREF(__pyx_v_gap_open_penalty);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_gap_open_penalty);
    __Pyx_INCREF(__pyx_v_gap_extension_penalty);
    __Pyx_GIVEREF(__pyx_v_gap_extension_penalty);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_gap_extension_penalty);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_tuple__7, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1128, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1128, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {

      /* "aldy/indelpost/varaln.pyx":1129
 *     if auto_adjust_extension_penalty:
 *         if (gap_open_penalty, gap_extension_penalty) != (3, 1):
 *             if len(target.indel_seq) < 20:             # <<<<<<<<<<<<<<
 *                 grid = [(gap_open_penalty, gap_extension_penalty),
 *                             (3, 1), (3, 0), (5, 1), (5, 0), (4, 1), (4, 0)
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_target, __pyx_n_s_indel_seq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1129, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_1 = ((__pyx_t_4 < 20) != 0);
      if (__pyx_t_1) {

        /* "aldy/indelpost/varaln.pyx":1130
 *         if (gap_open_penalty, gap_extension_penalty) != (3, 1):
 *             if len(target.indel_seq) < 20:
 *                 grid = [(gap_open_penalty, gap_extension_penalty),             # <<<<<<<<<<<<<<
 *                             (3, 1), (3, 0), (5, 1), (5, 0), (4, 1), (4, 0)
 *                        ]
 */
        __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1130, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_v_gap_open_penalty);
        __Pyx_GIVEREF(__pyx_v_gap_open_penalty);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_gap_open_penalty);
        __Pyx_INCREF(__pyx_v_gap_extension_penalty);
        __Pyx_GIVEREF(__pyx_v_gap_extension_penalty);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_gap_extension_penalty);

        /* "aldy/indelpost/varaln.pyx":1131
 *             if len(target.indel_seq) < 20:
 *                 grid = [(gap_open_penalty, gap_extension_penalty),
 *                             (3, 1), (3, 0), (5, 1), (5, 0), (4, 1), (4, 0)             # <<<<<<<<<<<<<<
 *                        ]
 *             else:
 */
        __pyx_t_2 = PyList_New(7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1130, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_3);
        PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
        __Pyx_INCREF(__pyx_tuple__7);
        __Pyx_GIVEREF(__pyx_tuple__7);
        PyList_SET_ITEM(__pyx_t_2, 1, __pyx_tuple__7);
        __Pyx_INCREF(__pyx_tuple__8);
        __Pyx_GIVEREF(__pyx_tuple__8);
        PyList_SET_ITEM(__pyx_t_2, 2, __pyx_tuple__8);
        __Pyx_INCREF(__pyx_tuple__9);
        __Pyx_GIVEREF(__pyx_tuple__9);
        PyList_SET_ITEM(__pyx_t_2, 3, __pyx_tuple__9);
        __Pyx_INCREF(__pyx_tuple__10);
        __Pyx_GIVEREF(__pyx_tuple__10);
        PyList_SET_ITEM(__pyx_t_2, 4, __pyx_tuple__10);
        __Pyx_INCREF(__pyx_tuple__11);
        __Pyx_GIVEREF(__pyx_tuple__11);
        PyList_SET_ITEM(__pyx_t_2, 5, __pyx_tuple__11);
        __Pyx_INCREF(__pyx_tuple__12);
        __Pyx_GIVEREF(__pyx_tuple__12);
        PyList_SET_ITEM(__pyx_t_2, 6, __pyx_tuple__12);
        __pyx_t_3 = 0;
        __pyx_v_grid = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;

        /* "aldy/indelpost/varaln.pyx":1129
 *     if auto_adjust_extension_penalty:
 *         if (gap_open_penalty, gap_extension_penalty) != (3, 1):
 *             if len(target.indel_seq) < 20:             # <<<<<<<<<<<<<<
 *                 grid = [(gap_open_penalty, gap_extension_penalty),
 *                             (3, 1), (3, 0), (5, 1), (5, 0), (4, 1), (4, 0)
 */
        goto __pyx_L5;
      }

      /* "aldy/indelpost/varaln.pyx":1134
 *                        ]
 *             else:
 *                 grid = [(gap_open_penalty, gap_extension_penalty),             # <<<<<<<<<<<<<<
 *                             (3, 0), (3, 1), (5, 1), (5, 0), (4, 1), (4, 0)
 *                        ]
 */
      /*else*/ {
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1134, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_v_gap_open_penalty);
        __Pyx_GIVEREF(__pyx_v_gap_open_penalty);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_gap_open_penalty);
        __Pyx_INCREF(__pyx_v_gap_extension_penalty);
        __Pyx_GIVEREF(__pyx_v_gap_extension_penalty);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_gap_extension_penalty);

        /* "aldy/indelpost/varaln.pyx":1135
 *             else:
 *                 grid = [(gap_open_penalty, gap_extension_penalty),
 *                             (3, 0), (3, 1), (5, 1), (5, 0), (4, 1), (4, 0)             # <<<<<<<<<<<<<<
 *                        ]
 *         else:
 */
        __pyx_t_3 = PyList_New(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1134, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_2);
        PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
        __Pyx_INCREF(__pyx_tuple__8);
        __Pyx_GIVEREF(__pyx_tuple__8);
        PyList_SET_ITEM(__pyx_t_3, 1, __pyx_tuple__8);
        __Pyx_INCREF(__pyx_tuple__7);
        __Pyx_GIVEREF(__pyx_tuple__7);
        PyList_SET_ITEM(__pyx_t_3, 2, __pyx_tuple__7);
        __Pyx_INCREF(__pyx_tuple__9);
        __Pyx_GIVEREF(__pyx_tuple__9);
        PyList_SET_ITEM(__pyx_t_3, 3, __pyx_tuple__9);
        __Pyx_INCREF(__pyx_tuple__10);
        __Pyx_GIVEREF(__pyx_tuple__10);
        PyList_SET_ITEM(__pyx_t_3, 4, __pyx_tuple__10);
        __Pyx_INCREF(__pyx_tuple__11);
        __Pyx_GIVEREF(__pyx_tuple__11);
        PyList_SET_ITEM(__pyx_t_3, 5, __pyx_tuple__11);
        __Pyx_INCREF(__pyx_tuple__12);
        __Pyx_GIVEREF(__pyx_tuple__12);
        PyList_SET_ITEM(__pyx_t_3, 6, __pyx_tuple__12);
        __pyx_t_2 = 0;
        __pyx_v_grid = ((PyObject*)__pyx_t_3);
        __pyx_t_3 = 0;
      }
      __pyx_L5:;

      /* "aldy/indelpost/varaln.pyx":1128
 * ):
 *     if auto_adjust_extension_penalty:
 *         if (gap_open_penalty, gap_extension_penalty) != (3, 1):             # <<<<<<<<<<<<<<
 *             if len(target.indel_seq) < 20:
 *                 grid = [(gap_open_penalty, gap_extension_penalty),
 */
      goto __pyx_L4;
    }

    /* "aldy/indelpost/varaln.pyx":1138
 *                        ]
 *         else:
 *             if len(target.indel_seq) < 20:             # <<<<<<<<<<<<<<
 *                 grid = [(3, 1), (3, 0), (5, 1), (5, 0), (4, 1), (4, 0)]
 *             else:
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_target, __pyx_n_s_indel_seq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1138, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_1 = ((__pyx_t_4 < 20) != 0);
      if (__pyx_t_1) {

        /* "aldy/indelpost/varaln.pyx":1139
 *         else:
 *             if len(target.indel_seq) < 20:
 *                 grid = [(3, 1), (3, 0), (5, 1), (5, 0), (4, 1), (4, 0)]             # <<<<<<<<<<<<<<
 *             else:
 *                 grid = [(3, 0), (3, 1), (5, 1), (5, 0), (4, 1), (4, 0)]
 */
        __pyx_t_3 = PyList_New(6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1139, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_tuple__7);
        __Pyx_GIVEREF(__pyx_tuple__7);
        PyList_SET_ITEM(__pyx_t_3, 0, __pyx_tuple__7);
        __Pyx_INCREF(__pyx_tuple__8);
        __Pyx_GIVEREF(__pyx_tuple__8);
        PyList_SET_ITEM(__pyx_t_3, 1, __pyx_tuple__8);
        __Pyx_INCREF(__pyx_tuple__9);
        __Pyx_GIVEREF(__pyx_tuple__9);
        PyList_SET_ITEM(__pyx_t_3, 2, __pyx_tuple__9);
        __Pyx_INCREF(__pyx_tuple__10);
        __Pyx_GIVEREF(__pyx_tuple__10);
        PyList_SET_ITEM(__pyx_t_3, 3, __pyx_tuple__10);
        __Pyx_INCREF(__pyx_tuple__11);
        __Pyx_GIVEREF(__pyx_tuple__11);
        PyList_SET_ITEM(__pyx_t_3, 4, __pyx_tuple__11);
        __Pyx_INCREF(__pyx_tuple__12);
        __Pyx_GIVEREF(__pyx_tuple__12);
        PyList_SET_ITEM(__pyx_t_3, 5, __pyx_tuple__12);
        __pyx_v_grid = ((PyObject*)__pyx_t_3);
        __pyx_t_3 = 0;

        /* "aldy/indelpost/varaln.pyx":1138
 *                        ]
 *         else:
 *             if len(target.indel_seq) < 20:             # <<<<<<<<<<<<<<
 *                 grid = [(3, 1), (3, 0), (5, 1), (5, 0), (4, 1), (4, 0)]
 *             else:
 */
        goto __pyx_L6;
      }

      /* "aldy/indelpost/varaln.pyx":1141
 *                 grid = [(3, 1), (3, 0), (5, 1), (5, 0), (4, 1), (4, 0)]
 *             else:
 *                 grid = [(3, 0), (3, 1), (5, 1), (5, 0), (4, 1), (4, 0)]             # <<<<<<<<<<<<<<
 *     else:
 *         grid = [(gap_open_penalty, gap_extension_penalty)]
 */
      /*else*/ {
        __pyx_t_3 = PyList_New(6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1141, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_tuple__8);
        __Pyx_GIVEREF(__pyx_tuple__8);
        PyList_SET_ITEM(__pyx_t_3, 0, __pyx_tuple__8);
        __Pyx_INCREF(__pyx_tuple__7);
        __Pyx_GIVEREF(__pyx_tuple__7);
        PyList_SET_ITEM(__pyx_t_3, 1, __pyx_tuple__7);
        __Pyx_INCREF(__pyx_tuple__9);
        __Pyx_GIVEREF(__pyx_tuple__9);
        PyList_SET_ITEM(__pyx_t_3, 2, __pyx_tuple__9);
        __Pyx_INCREF(__pyx_tuple__10);
        __Pyx_GIVEREF(__pyx_tuple__10);
        PyList_SET_ITEM(__pyx_t_3, 3, __pyx_tuple__10);
        __Pyx_INCREF(__pyx_tuple__11);
        __Pyx_GIVEREF(__pyx_tuple__11);
        PyList_SET_ITEM(__pyx_t_3, 4, __pyx_tuple__11);
        __Pyx_INCREF(__pyx_tuple__12);
        __Pyx_GIVEREF(__pyx_tuple__12);
        PyList_SET_ITEM(__pyx_t_3, 5, __pyx_tuple__12);
        __pyx_v_grid = ((PyObject*)__pyx_t_3);
        __pyx_t_3 = 0;
      }
      __pyx_L6:;
    }
    __pyx_L4:;

    /* "aldy/indelpost/varaln.pyx":1127
 *                    target,
 * ):
 *     if auto_adjust_extension_penalty:             # <<<<<<<<<<<<<<
 *         if (gap_open_penalty, gap_extension_penalty) != (3, 1):
 *             if len(target.indel_seq) < 20:
 */
    goto __pyx_L3;
  }

  /* "aldy/indelpost/varaln.pyx":1143
 *                 grid = [(3, 0), (3, 1), (5, 1), (5, 0), (4, 1), (4, 0)]
 *     else:
 *         grid = [(gap_open_penalty, gap_extension_penalty)]             # <<<<<<<<<<<<<<
 * 
 *     return grid
 */
  /*else*/ {
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_gap_open_penalty);
    __Pyx_GIVEREF(__pyx_v_gap_open_penalty);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_gap_open_penalty);
    __Pyx_INCREF(__pyx_v_gap_extension_penalty);
    __Pyx_GIVEREF(__pyx_v_gap_extension_penalty);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_gap_extension_penalty);
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_v_grid = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "aldy/indelpost/varaln.pyx":1145
 *         grid = [(gap_open_penalty, gap_extension_penalty)]
 * 
 *     return grid             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_grid);
  __pyx_r = __pyx_v_grid;
  goto __pyx_L0;

  /* "aldy/indelpost/varaln.pyx":1122
 * 
 * 
 * def generate_grid (auto_adjust_extension_penalty,             # <<<<<<<<<<<<<<
 *                    gap_open_penalty,
 *                    gap_extension_penalty,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("aldy.indelpost.varaln.generate_grid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_grid);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":1148
 * 
 * 
 * def grid_search(             # <<<<<<<<<<<<<<
 *     target,
 *     pileup,
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_13grid_search(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_12grid_search[] = "grid_search(target, pileup, window, mapq_thresh, within, retarget_cutoff, match_score, mismatch_penalty, grid, unspl_loc_ref, exact_match_for_shiftable)";
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6varaln_13grid_search = {"grid_search", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6varaln_13grid_search, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4aldy_9indelpost_6varaln_12grid_search};
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_13grid_search(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_target = 0;
  PyObject *__pyx_v_pileup = 0;
  PyObject *__pyx_v_window = 0;
  PyObject *__pyx_v_mapq_thresh = 0;
  PyObject *__pyx_v_within = 0;
  PyObject *__pyx_v_retarget_cutoff = 0;
  PyObject *__pyx_v_match_score = 0;
  PyObject *__pyx_v_mismatch_penalty = 0;
  PyObject *__pyx_v_grid = 0;
  PyObject *__pyx_v_unspl_loc_ref = 0;
  PyObject *__pyx_v_exact_match_for_shiftable = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("grid_search (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_target,&__pyx_n_s_pileup,&__pyx_n_s_window,&__pyx_n_s_mapq_thresh,&__pyx_n_s_within,&__pyx_n_s_retarget_cutoff,&__pyx_n_s_match_score,&__pyx_n_s_mismatch_penalty,&__pyx_n_s_grid,&__pyx_n_s_unspl_loc_ref,&__pyx_n_s_exact_match_for_shiftable,0};
    PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_target)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pileup)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grid_search", 1, 11, 11, 1); __PYX_ERR(0, 1148, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_window)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grid_search", 1, 11, 11, 2); __PYX_ERR(0, 1148, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mapq_thresh)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grid_search", 1, 11, 11, 3); __PYX_ERR(0, 1148, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_within)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grid_search", 1, 11, 11, 4); __PYX_ERR(0, 1148, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_retarget_cutoff)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grid_search", 1, 11, 11, 5); __PYX_ERR(0, 1148, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_match_score)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grid_search", 1, 11, 11, 6); __PYX_ERR(0, 1148, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mismatch_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grid_search", 1, 11, 11, 7); __PYX_ERR(0, 1148, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_grid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grid_search", 1, 11, 11, 8); __PYX_ERR(0, 1148, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_unspl_loc_ref)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grid_search", 1, 11, 11, 9); __PYX_ERR(0, 1148, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exact_match_for_shiftable)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("grid_search", 1, 11, 11, 10); __PYX_ERR(0, 1148, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "grid_search") < 0)) __PYX_ERR(0, 1148, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 11) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
    }
    __pyx_v_target = values[0];
    __pyx_v_pileup = values[1];
    __pyx_v_window = values[2];
    __pyx_v_mapq_thresh = values[3];
    __pyx_v_within = values[4];
    __pyx_v_retarget_cutoff = values[5];
    __pyx_v_match_score = values[6];
    __pyx_v_mismatch_penalty = values[7];
    __pyx_v_grid = values[8];
    __pyx_v_unspl_loc_ref = values[9];
    __pyx_v_exact_match_for_shiftable = values[10];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("grid_search", 1, 11, 11, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1148, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.grid_search", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_12grid_search(__pyx_self, __pyx_v_target, __pyx_v_pileup, __pyx_v_window, __pyx_v_mapq_thresh, __pyx_v_within, __pyx_v_retarget_cutoff, __pyx_v_match_score, __pyx_v_mismatch_penalty, __pyx_v_grid, __pyx_v_unspl_loc_ref, __pyx_v_exact_match_for_shiftable);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_12grid_search(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_target, PyObject *__pyx_v_pileup, PyObject *__pyx_v_window, PyObject *__pyx_v_mapq_thresh, PyObject *__pyx_v_within, PyObject *__pyx_v_retarget_cutoff, PyObject *__pyx_v_match_score, PyObject *__pyx_v_mismatch_penalty, PyObject *__pyx_v_grid, PyObject *__pyx_v_unspl_loc_ref, PyObject *__pyx_v_exact_match_for_shiftable) {
  PyObject *__pyx_v_h = NULL;
  PyObject *__pyx_v_responses = NULL;
  PyObject *__pyx_v_scores = NULL;
  PyObject *__pyx_v_hs = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_v_score = NULL;
  PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_best_res = NULL;
  PyObject *__pyx_v_best_params = NULL;
  int __pyx_v_is_gapped_aln;
  PyObject *__pyx_v_candidate = NULL;
  PyObject *__pyx_v_gap_open_penalty = NULL;
  PyObject *__pyx_v_gap_extension_penalty = NULL;
  PyObject *__pyx_v_updated_reads = NULL;
  PyObject *__pyx_v_read = NULL;
  PyObject *__pyx_v_aligner = NULL;
  PyObject *__pyx_v_ref_seq = NULL;
  PyObject *__pyx_v_ref_start = NULL;
  PyObject *__pyx_v_updated_read = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  PyObject *(*__pyx_t_13)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("grid_search", 0);

  /* "aldy/indelpost/varaln.pyx":1162
 * ):
 *     # grid = [(gap.open, gap.ext)]
 *     h = 0             # <<<<<<<<<<<<<<
 *     responses, scores, hs = [], [], []
 *     while h < len(grid):
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_h = __pyx_int_0;

  /* "aldy/indelpost/varaln.pyx":1163
 *     # grid = [(gap.open, gap.ext)]
 *     h = 0
 *     responses, scores, hs = [], [], []             # <<<<<<<<<<<<<<
 *     while h < len(grid):
 *         res = retarget(
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_responses = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_scores = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_v_hs = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "aldy/indelpost/varaln.pyx":1164
 *     h = 0
 *     responses, scores, hs = [], [], []
 *     while h < len(grid):             # <<<<<<<<<<<<<<
 *         res = retarget(
 *             target,
 */
  while (1) {
    __pyx_t_4 = PyObject_Length(__pyx_v_grid); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1164, __pyx_L1_error)
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_h, __pyx_t_3, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1164, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1164, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!__pyx_t_5) break;

    /* "aldy/indelpost/varaln.pyx":1165
 *     responses, scores, hs = [], [], []
 *     while h < len(grid):
 *         res = retarget(             # <<<<<<<<<<<<<<
 *             target,
 *             pileup,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_retarget); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "aldy/indelpost/varaln.pyx":1174
 *             match_score,
 *             mismatch_penalty,
 *             grid[h][0],             # <<<<<<<<<<<<<<
 *             grid[h][1],
 *             unspl_loc_ref,
 */
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_grid, __pyx_v_h); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "aldy/indelpost/varaln.pyx":1175
 *             mismatch_penalty,
 *             grid[h][0],
 *             grid[h][1],             # <<<<<<<<<<<<<<
 *             unspl_loc_ref,
 *             exact_match_for_shiftable,
 */
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_grid, __pyx_v_h); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "aldy/indelpost/varaln.pyx":1177
 *             grid[h][1],
 *             unspl_loc_ref,
 *             exact_match_for_shiftable,             # <<<<<<<<<<<<<<
 *         )
 * 
 */
    __pyx_t_1 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[13] = {__pyx_t_1, __pyx_v_target, __pyx_v_pileup, __pyx_v_window, __pyx_v_mapq_thresh, __pyx_v_within, __pyx_v_retarget_cutoff, __pyx_v_match_score, __pyx_v_mismatch_penalty, __pyx_t_6, __pyx_t_7, __pyx_v_unspl_loc_ref, __pyx_v_exact_match_for_shiftable};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 12+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1165, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[13] = {__pyx_t_1, __pyx_v_target, __pyx_v_pileup, __pyx_v_window, __pyx_v_mapq_thresh, __pyx_v_within, __pyx_v_retarget_cutoff, __pyx_v_match_score, __pyx_v_mismatch_penalty, __pyx_t_6, __pyx_t_7, __pyx_v_unspl_loc_ref, __pyx_v_exact_match_for_shiftable};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 12+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1165, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(12+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1165, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_v_target);
      __Pyx_GIVEREF(__pyx_v_target);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_target);
      __Pyx_INCREF(__pyx_v_pileup);
      __Pyx_GIVEREF(__pyx_v_pileup);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_pileup);
      __Pyx_INCREF(__pyx_v_window);
      __Pyx_GIVEREF(__pyx_v_window);
      PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_v_window);
      __Pyx_INCREF(__pyx_v_mapq_thresh);
      __Pyx_GIVEREF(__pyx_v_mapq_thresh);
      PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_v_mapq_thresh);
      __Pyx_INCREF(__pyx_v_within);
      __Pyx_GIVEREF(__pyx_v_within);
      PyTuple_SET_ITEM(__pyx_t_9, 4+__pyx_t_8, __pyx_v_within);
      __Pyx_INCREF(__pyx_v_retarget_cutoff);
      __Pyx_GIVEREF(__pyx_v_retarget_cutoff);
      PyTuple_SET_ITEM(__pyx_t_9, 5+__pyx_t_8, __pyx_v_retarget_cutoff);
      __Pyx_INCREF(__pyx_v_match_score);
      __Pyx_GIVEREF(__pyx_v_match_score);
      PyTuple_SET_ITEM(__pyx_t_9, 6+__pyx_t_8, __pyx_v_match_score);
      __Pyx_INCREF(__pyx_v_mismatch_penalty);
      __Pyx_GIVEREF(__pyx_v_mismatch_penalty);
      PyTuple_SET_ITEM(__pyx_t_9, 7+__pyx_t_8, __pyx_v_mismatch_penalty);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_9, 8+__pyx_t_8, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_9, 9+__pyx_t_8, __pyx_t_7);
      __Pyx_INCREF(__pyx_v_unspl_loc_ref);
      __Pyx_GIVEREF(__pyx_v_unspl_loc_ref);
      PyTuple_SET_ITEM(__pyx_t_9, 10+__pyx_t_8, __pyx_v_unspl_loc_ref);
      __Pyx_INCREF(__pyx_v_exact_match_for_shiftable);
      __Pyx_GIVEREF(__pyx_v_exact_match_for_shiftable);
      PyTuple_SET_ITEM(__pyx_t_9, 11+__pyx_t_8, __pyx_v_exact_match_for_shiftable);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1165, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_res, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":1180
 *         )
 * 
 *         if res:             # <<<<<<<<<<<<<<
 *             score = res[2]
 *             responses.append(res)
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_res); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1180, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "aldy/indelpost/varaln.pyx":1181
 * 
 *         if res:
 *             score = res[2]             # <<<<<<<<<<<<<<
 *             responses.append(res)
 *             #scores.append(score)
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_res, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_score, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "aldy/indelpost/varaln.pyx":1182
 *         if res:
 *             score = res[2]
 *             responses.append(res)             # <<<<<<<<<<<<<<
 *             #scores.append(score)
 *             hs.append(h)
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_responses, __pyx_v_res); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1182, __pyx_L1_error)

      /* "aldy/indelpost/varaln.pyx":1184
 *             responses.append(res)
 *             #scores.append(score)
 *             hs.append(h)             # <<<<<<<<<<<<<<
 * 
 *             # exact match
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_hs, __pyx_v_h); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1184, __pyx_L1_error)

      /* "aldy/indelpost/varaln.pyx":1187
 * 
 *             # exact match
 *             if score == 1.0:             # <<<<<<<<<<<<<<
 *                 scores.append(score * len(res[1]))
 *                 # cnt exact hit?
 */
      __pyx_t_2 = __Pyx_PyFloat_EqObjC(__pyx_v_score, __pyx_float_1_0, 1.0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1187, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_5) {

        /* "aldy/indelpost/varaln.pyx":1188
 *             # exact match
 *             if score == 1.0:
 *                 scores.append(score * len(res[1]))             # <<<<<<<<<<<<<<
 *                 # cnt exact hit?
 *                 #break
 */
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_res, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1188, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1188, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1188, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = PyNumber_Multiply(__pyx_v_score, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1188, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_scores, __pyx_t_3); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1188, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "aldy/indelpost/varaln.pyx":1187
 * 
 *             # exact match
 *             if score == 1.0:             # <<<<<<<<<<<<<<
 *                 scores.append(score * len(res[1]))
 *                 # cnt exact hit?
 */
        goto __pyx_L6;
      }

      /* "aldy/indelpost/varaln.pyx":1192
 *                 #break
 *             else:
 *                 scores.append(score)             # <<<<<<<<<<<<<<
 *         h += 1
 * 
 */
      /*else*/ {
        __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_scores, __pyx_v_score); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1192, __pyx_L1_error)
      }
      __pyx_L6:;

      /* "aldy/indelpost/varaln.pyx":1180
 *         )
 * 
 *         if res:             # <<<<<<<<<<<<<<
 *             score = res[2]
 *             responses.append(res)
 */
    }

    /* "aldy/indelpost/varaln.pyx":1193
 *             else:
 *                 scores.append(score)
 *         h += 1             # <<<<<<<<<<<<<<
 * 
 *     if responses:
 */
    __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_v_h, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_h, __pyx_t_3);
    __pyx_t_3 = 0;
  }

  /* "aldy/indelpost/varaln.pyx":1195
 *         h += 1
 * 
 *     if responses:             # <<<<<<<<<<<<<<
 *         idx = scores.index(max(scores))
 *         best_res = responses[idx]
 */
  __pyx_t_5 = (PyList_GET_SIZE(__pyx_v_responses) != 0);
  if (__pyx_t_5) {

    /* "aldy/indelpost/varaln.pyx":1196
 * 
 *     if responses:
 *         idx = scores.index(max(scores))             # <<<<<<<<<<<<<<
 *         best_res = responses[idx]
 *         best_params = grid[hs[idx]]
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_v_scores); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyList_Type_index, __pyx_v_scores, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_idx = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":1197
 *     if responses:
 *         idx = scores.index(max(scores))
 *         best_res = responses[idx]             # <<<<<<<<<<<<<<
 *         best_params = grid[hs[idx]]
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_responses, __pyx_v_idx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_best_res = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":1198
 *         idx = scores.index(max(scores))
 *         best_res = responses[idx]
 *         best_params = grid[hs[idx]]             # <<<<<<<<<<<<<<
 * 
 *         is_gapped_aln=False # to be removed
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_hs, __pyx_v_idx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_grid, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_best_params = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":1200
 *         best_params = grid[hs[idx]]
 * 
 *         is_gapped_aln=False # to be removed             # <<<<<<<<<<<<<<
 * 
 *         candidate = best_res[0]
 */
    __pyx_v_is_gapped_aln = 0;

    /* "aldy/indelpost/varaln.pyx":1202
 *         is_gapped_aln=False # to be removed
 * 
 *         candidate = best_res[0]             # <<<<<<<<<<<<<<
 * 
 *         gap_open_penalty, gap_extension_penalty = best_params[0], best_params[1]
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_best_res, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_candidate = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "aldy/indelpost/varaln.pyx":1204
 *         candidate = best_res[0]
 * 
 *         gap_open_penalty, gap_extension_penalty = best_params[0], best_params[1]             # <<<<<<<<<<<<<<
 * 
 *         updated_reads = []
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_best_params, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_best_params, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_gap_open_penalty = __pyx_t_3;
    __pyx_t_3 = 0;
    __pyx_v_gap_extension_penalty = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":1206
 *         gap_open_penalty, gap_extension_penalty = best_params[0], best_params[1]
 * 
 *         updated_reads = []             # <<<<<<<<<<<<<<
 *         for read, aligner, ref_seq, ref_start in zip(best_res[1], best_res[5], best_res[3], best_res[4]):
 * 
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_updated_reads = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "aldy/indelpost/varaln.pyx":1207
 * 
 *         updated_reads = []
 *         for read, aligner, ref_seq, ref_start in zip(best_res[1], best_res[5], best_res[3], best_res[4]):             # <<<<<<<<<<<<<<
 * 
 *             updated_read = update_read_info(
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_best_res, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_best_res, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_best_res, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_best_res, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_7);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_9 = 0;
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_6, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
      __pyx_t_6 = __pyx_t_7; __Pyx_INCREF(__pyx_t_6); __pyx_t_4 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1207, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_11 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1207, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_4); __Pyx_INCREF(__pyx_t_7); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1207, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_6, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1207, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_4); __Pyx_INCREF(__pyx_t_7); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1207, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_6, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1207, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_11(__pyx_t_6);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1207, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
        PyObject* sequence = __pyx_t_7;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 4)) {
          if (size > 4) __Pyx_RaiseTooManyValuesError(4);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1207, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 2); 
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 3); 
        } else {
          __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
          __pyx_t_2 = PyList_GET_ITEM(sequence, 2); 
          __pyx_t_1 = PyList_GET_ITEM(sequence, 3); 
        }
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        #else
        {
          Py_ssize_t i;
          PyObject** temps[4] = {&__pyx_t_9,&__pyx_t_3,&__pyx_t_2,&__pyx_t_1};
          for (i=0; i < 4; i++) {
            PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 1207, __pyx_L1_error)
            __Pyx_GOTREF(item);
            *(temps[i]) = item;
          }
        }
        #endif
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        Py_ssize_t index = -1;
        PyObject** temps[4] = {&__pyx_t_9,&__pyx_t_3,&__pyx_t_2,&__pyx_t_1};
        __pyx_t_12 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1207, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_13 = Py_TYPE(__pyx_t_12)->tp_iternext;
        for (index=0; index < 4; index++) {
          PyObject* item = __pyx_t_13(__pyx_t_12); if (unlikely(!item)) goto __pyx_L10_unpacking_failed;
          __Pyx_GOTREF(item);
          *(temps[index]) = item;
        }
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_12), 4) < 0) __PYX_ERR(0, 1207, __pyx_L1_error)
        __pyx_t_13 = NULL;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        goto __pyx_L11_unpacking_done;
        __pyx_L10_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_13 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1207, __pyx_L1_error)
        __pyx_L11_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_read, __pyx_t_9);
      __pyx_t_9 = 0;
      __Pyx_XDECREF_SET(__pyx_v_aligner, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ref_seq, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ref_start, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "aldy/indelpost/varaln.pyx":1209
 *         for read, aligner, ref_seq, ref_start in zip(best_res[1], best_res[5], best_res[3], best_res[4]):
 * 
 *             updated_read = update_read_info(             # <<<<<<<<<<<<<<
 *                                 read,
 *                                 candidate,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_update_read_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "aldy/indelpost/varaln.pyx":1212
 *                                 read,
 *                                 candidate,
 *                                 is_gapped_aln,             # <<<<<<<<<<<<<<
 *                                 gap_open_penalty,
 *                                 gap_extension_penalty,
 */
      __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_is_gapped_aln); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1212, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "aldy/indelpost/varaln.pyx":1217
 *                                 aligner,
 *                                 ref_seq,
 *                                 ref_start,             # <<<<<<<<<<<<<<
 *                             )
 * 
 */
      __pyx_t_3 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[9] = {__pyx_t_3, __pyx_v_read, __pyx_v_candidate, __pyx_t_2, __pyx_v_gap_open_penalty, __pyx_v_gap_extension_penalty, __pyx_v_aligner, __pyx_v_ref_seq, __pyx_v_ref_start};
        __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 8+__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1209, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[9] = {__pyx_t_3, __pyx_v_read, __pyx_v_candidate, __pyx_t_2, __pyx_v_gap_open_penalty, __pyx_v_gap_extension_penalty, __pyx_v_aligner, __pyx_v_ref_seq, __pyx_v_ref_start};
        __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 8+__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1209, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(8+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1209, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_INCREF(__pyx_v_read);
        __Pyx_GIVEREF(__pyx_v_read);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_read);
        __Pyx_INCREF(__pyx_v_candidate);
        __Pyx_GIVEREF(__pyx_v_candidate);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_candidate);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_2);
        __Pyx_INCREF(__pyx_v_gap_open_penalty);
        __Pyx_GIVEREF(__pyx_v_gap_open_penalty);
        PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_v_gap_open_penalty);
        __Pyx_INCREF(__pyx_v_gap_extension_penalty);
        __Pyx_GIVEREF(__pyx_v_gap_extension_penalty);
        PyTuple_SET_ITEM(__pyx_t_9, 4+__pyx_t_8, __pyx_v_gap_extension_penalty);
        __Pyx_INCREF(__pyx_v_aligner);
        __Pyx_GIVEREF(__pyx_v_aligner);
        PyTuple_SET_ITEM(__pyx_t_9, 5+__pyx_t_8, __pyx_v_aligner);
        __Pyx_INCREF(__pyx_v_ref_seq);
        __Pyx_GIVEREF(__pyx_v_ref_seq);
        PyTuple_SET_ITEM(__pyx_t_9, 6+__pyx_t_8, __pyx_v_ref_seq);
        __Pyx_INCREF(__pyx_v_ref_start);
        __Pyx_GIVEREF(__pyx_v_ref_start);
        PyTuple_SET_ITEM(__pyx_t_9, 7+__pyx_t_8, __pyx_v_ref_start);
        __pyx_t_2 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1209, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_updated_read, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "aldy/indelpost/varaln.pyx":1220
 *                             )
 * 
 *             updated_reads.append(updated_read)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_updated_reads, __pyx_v_updated_read); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1220, __pyx_L1_error)

      /* "aldy/indelpost/varaln.pyx":1207
 * 
 *         updated_reads = []
 *         for read, aligner, ref_seq, ref_start in zip(best_res[1], best_res[5], best_res[3], best_res[4]):             # <<<<<<<<<<<<<<
 * 
 *             updated_read = update_read_info(
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "aldy/indelpost/varaln.pyx":1223
 * 
 * 
 *         return candidate, updated_reads, gap_open_penalty, gap_extension_penalty             # <<<<<<<<<<<<<<
 *     else:
 *         return None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_candidate);
    __Pyx_GIVEREF(__pyx_v_candidate);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_candidate);
    __Pyx_INCREF(__pyx_v_updated_reads);
    __Pyx_GIVEREF(__pyx_v_updated_reads);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_updated_reads);
    __Pyx_INCREF(__pyx_v_gap_open_penalty);
    __Pyx_GIVEREF(__pyx_v_gap_open_penalty);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_gap_open_penalty);
    __Pyx_INCREF(__pyx_v_gap_extension_penalty);
    __Pyx_GIVEREF(__pyx_v_gap_extension_penalty);
    PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_v_gap_extension_penalty);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "aldy/indelpost/varaln.pyx":1195
 *         h += 1
 * 
 *     if responses:             # <<<<<<<<<<<<<<
 *         idx = scores.index(max(scores))
 *         best_res = responses[idx]
 */
  }

  /* "aldy/indelpost/varaln.pyx":1225
 *         return candidate, updated_reads, gap_open_penalty, gap_extension_penalty
 *     else:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }

  /* "aldy/indelpost/varaln.pyx":1148
 * 
 * 
 * def grid_search(             # <<<<<<<<<<<<<<
 *     target,
 *     pileup,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("aldy.indelpost.varaln.grid_search", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_h);
  __Pyx_XDECREF(__pyx_v_responses);
  __Pyx_XDECREF(__pyx_v_scores);
  __Pyx_XDECREF(__pyx_v_hs);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_score);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_best_res);
  __Pyx_XDECREF(__pyx_v_best_params);
  __Pyx_XDECREF(__pyx_v_candidate);
  __Pyx_XDECREF(__pyx_v_gap_open_penalty);
  __Pyx_XDECREF(__pyx_v_gap_extension_penalty);
  __Pyx_XDECREF(__pyx_v_updated_reads);
  __Pyx_XDECREF(__pyx_v_read);
  __Pyx_XDECREF(__pyx_v_aligner);
  __Pyx_XDECREF(__pyx_v_ref_seq);
  __Pyx_XDECREF(__pyx_v_ref_start);
  __Pyx_XDECREF(__pyx_v_updated_read);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/varaln.pyx":1228
 * 
 * 
 * def is_perfect_match(aligner, contig_seq, read_seq):             # <<<<<<<<<<<<<<
 *     aligner.setRead(read_seq)
 *     _aln = aligner.align(gap_open=len(read_seq), gap_extension=len(read_seq))
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_15is_perfect_match(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6varaln_14is_perfect_match[] = "is_perfect_match(aligner, contig_seq, read_seq)";
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6varaln_15is_perfect_match = {"is_perfect_match", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6varaln_15is_perfect_match, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4aldy_9indelpost_6varaln_14is_perfect_match};
static PyObject *__pyx_pw_4aldy_9indelpost_6varaln_15is_perfect_match(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_aligner = 0;
  PyObject *__pyx_v_contig_seq = 0;
  PyObject *__pyx_v_read_seq = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_perfect_match (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_aligner,&__pyx_n_s_contig_seq,&__pyx_n_s_read_seq,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aligner)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_contig_seq)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_perfect_match", 1, 3, 3, 1); __PYX_ERR(0, 1228, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_read_seq)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_perfect_match", 1, 3, 3, 2); __PYX_ERR(0, 1228, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "is_perfect_match") < 0)) __PYX_ERR(0, 1228, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_aligner = values[0];
    __pyx_v_contig_seq = values[1];
    __pyx_v_read_seq = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_perfect_match", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1228, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.varaln.is_perfect_match", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6varaln_14is_perfect_match(__pyx_self, __pyx_v_aligner, __pyx_v_contig_seq, __pyx_v_read_seq);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6varaln_14is_perfect_match(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_aligner, PyObject *__pyx_v_contig_seq, PyObject *__pyx_v_read_seq) {
  PyObject *__pyx_v__aln = NULL;
  PyObject *__pyx_v__contig = NULL;
  PyObject *__pyx_v__read = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_perfect_match", 0);

  /* "aldy/indelpost/varaln.pyx":1229
 * 
 * def is_perfect_match(aligner, contig_seq, read_seq):
 *     aligner.setRead(read_seq)             # <<<<<<<<<<<<<<
 *     _aln = aligner.align(gap_open=len(read_seq), gap_extension=len(read_seq))
 *     _contig = contig_seq[_aln.reference_start : _aln.reference_end]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_aligner, __pyx_n_s_setRead); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_read_seq) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_read_seq);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":1230
 * def is_perfect_match(aligner, contig_seq, read_seq):
 *     aligner.setRead(read_seq)
 *     _aln = aligner.align(gap_open=len(read_seq), gap_extension=len(read_seq))             # <<<<<<<<<<<<<<
 *     _contig = contig_seq[_aln.reference_start : _aln.reference_end]
 *     _read = read_seq[_aln.read_start : _aln.read_end]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_aligner, __pyx_n_s_align); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_Length(__pyx_v_read_seq); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1230, __pyx_L1_error)
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_gap_open, __pyx_t_3) < 0) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = PyObject_Length(__pyx_v_read_seq); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1230, __pyx_L1_error)
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_gap_extension, __pyx_t_3) < 0) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v__aln = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "aldy/indelpost/varaln.pyx":1231
 *     aligner.setRead(read_seq)
 *     _aln = aligner.align(gap_open=len(read_seq), gap_extension=len(read_seq))
 *     _contig = contig_seq[_aln.reference_start : _aln.reference_end]             # <<<<<<<<<<<<<<
 *     _read = read_seq[_aln.read_start : _aln.read_end]
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v__aln, __pyx_n_s_reference_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v__aln, __pyx_n_s_reference_end); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_contig_seq, 0, 0, &__pyx_t_3, &__pyx_t_2, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v__contig = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":1232
 *     _aln = aligner.align(gap_open=len(read_seq), gap_extension=len(read_seq))
 *     _contig = contig_seq[_aln.reference_start : _aln.reference_end]
 *     _read = read_seq[_aln.read_start : _aln.read_end]             # <<<<<<<<<<<<<<
 * 
 *     return _contig == _read
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v__aln, __pyx_n_s_read_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v__aln, __pyx_n_s_read_end); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, &__pyx_t_1, &__pyx_t_2, NULL, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v__read = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "aldy/indelpost/varaln.pyx":1234
 *     _read = read_seq[_aln.read_start : _aln.read_end]
 * 
 *     return _contig == _read             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v__contig, __pyx_v__read, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1234, __pyx_L1_error)
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "aldy/indelpost/varaln.pyx":1228
 * 
 * 
 * def is_perfect_match(aligner, contig_seq, read_seq):             # <<<<<<<<<<<<<<
 *     aligner.setRead(read_seq)
 *     _aln = aligner.align(gap_open=len(read_seq), gap_extension=len(read_seq))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("aldy.indelpost.varaln.is_perfect_match", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__aln);
  __Pyx_XDECREF(__pyx_v__contig);
  __Pyx_XDECREF(__pyx_v__read);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":93
 *             __data_union data
 * 
 *         def __getbuffer__(self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_7cpython_5array_5array_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_7cpython_5array_5array_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7cpython_5array_5array___getbuffer__(((arrayobject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7cpython_5array_5array___getbuffer__(arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info, CYTHON_UNUSED int __pyx_v_flags) {
  PyObject *__pyx_v_item_count = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  char __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "array.pxd":98
 *             # In particular strided access is always provided regardless
 *             # of flags
 *             item_count = Py_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *             info.suboffsets = NULL
 */
  __pyx_t_1 = PyInt_FromSsize_t(Py_SIZE(((PyObject *)__pyx_v_self))); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_item_count = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "array.pxd":100
 *             item_count = Py_SIZE(self)
 * 
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.buf = self.data.as_chars
 *             info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "array.pxd":101
 * 
 *             info.suboffsets = NULL
 *             info.buf = self.data.as_chars             # <<<<<<<<<<<<<<
 *             info.readonly = 0
 *             info.ndim = 1
 */
  __pyx_t_2 = __pyx_v_self->data.as_chars;
  __pyx_v_info->buf = __pyx_t_2;

  /* "array.pxd":102
 *             info.suboffsets = NULL
 *             info.buf = self.data.as_chars
 *             info.readonly = 0             # <<<<<<<<<<<<<<
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 */
  __pyx_v_info->readonly = 0;

  /* "array.pxd":103
 *             info.buf = self.data.as_chars
 *             info.readonly = 0
 *             info.ndim = 1             # <<<<<<<<<<<<<<
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 *             info.len = info.itemsize * item_count
 */
  __pyx_v_info->ndim = 1;

  /* "array.pxd":104
 *             info.readonly = 0
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)             # <<<<<<<<<<<<<<
 *             info.len = info.itemsize * item_count
 * 
 */
  __pyx_t_3 = __pyx_v_self->ob_descr->itemsize;
  __pyx_v_info->itemsize = __pyx_t_3;

  /* "array.pxd":105
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 *             info.len = info.itemsize * item_count             # <<<<<<<<<<<<<<
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 */
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_info->itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_1, __pyx_v_item_count); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_info->len = __pyx_t_5;

  /* "array.pxd":107
 *             info.len = info.itemsize * item_count
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)             # <<<<<<<<<<<<<<
 *             if not info.shape:
 *                 raise MemoryError()
 */
  __pyx_v_info->shape = ((Py_ssize_t *)PyObject_Malloc(((sizeof(Py_ssize_t)) + 2)));

  /* "array.pxd":108
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 */
  __pyx_t_6 = ((!(__pyx_v_info->shape != 0)) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "array.pxd":109
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 *             info.shape[0] = item_count      # constant regardless of resizing
 *             info.strides = &info.itemsize
 */
    PyErr_NoMemory(); __PYX_ERR(2, 109, __pyx_L1_error)

    /* "array.pxd":108
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 */
  }

  /* "array.pxd":110
 *             if not info.shape:
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing             # <<<<<<<<<<<<<<
 *             info.strides = &info.itemsize
 * 
 */
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_v_item_count); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 110, __pyx_L1_error)
  (__pyx_v_info->shape[0]) = __pyx_t_5;

  /* "array.pxd":111
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 *             info.strides = &info.itemsize             # <<<<<<<<<<<<<<
 * 
 *             info.format = <char*> (info.shape + 1)
 */
  __pyx_v_info->strides = (&__pyx_v_info->itemsize);

  /* "array.pxd":113
 *             info.strides = &info.itemsize
 * 
 *             info.format = <char*> (info.shape + 1)             # <<<<<<<<<<<<<<
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0
 */
  __pyx_v_info->format = ((char *)(__pyx_v_info->shape + 1));

  /* "array.pxd":114
 * 
 *             info.format = <char*> (info.shape + 1)
 *             info.format[0] = self.ob_descr.typecode             # <<<<<<<<<<<<<<
 *             info.format[1] = 0
 *             info.obj = self
 */
  __pyx_t_7 = __pyx_v_self->ob_descr->typecode;
  (__pyx_v_info->format[0]) = __pyx_t_7;

  /* "array.pxd":115
 *             info.format = <char*> (info.shape + 1)
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0             # <<<<<<<<<<<<<<
 *             info.obj = self
 * 
 */
  (__pyx_v_info->format[1]) = 0;

  /* "array.pxd":116
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(self, Py_buffer* info):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "array.pxd":93
 *             __data_union data
 * 
 *         def __getbuffer__(self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cpython.array.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_item_count);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":118
 *             info.obj = self
 * 
 *         def __releasebuffer__(self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             PyObject_Free(info.shape)
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_7cpython_5array_5array_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_7cpython_5array_5array_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_7cpython_5array_5array_2__releasebuffer__(((arrayobject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7cpython_5array_5array_2__releasebuffer__(CYTHON_UNUSED arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "array.pxd":119
 * 
 *         def __releasebuffer__(self, Py_buffer* info):
 *             PyObject_Free(info.shape)             # <<<<<<<<<<<<<<
 * 
 *     array newarrayobject(PyTypeObject* type, Py_ssize_t size, arraydescr *descr)
 */
  PyObject_Free(__pyx_v_info->shape);

  /* "array.pxd":118
 *             info.obj = self
 * 
 *         def __releasebuffer__(self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             PyObject_Free(info.shape)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "array.pxd":130
 * 
 * 
 * cdef inline array clone(array template, Py_ssize_t length, bint zero):             # <<<<<<<<<<<<<<
 *     """ fast creation of a new array, given a template array.
 *     type will be same as template.
 */

static CYTHON_INLINE arrayobject *__pyx_f_7cpython_5array_clone(arrayobject *__pyx_v_template, Py_ssize_t __pyx_v_length, int __pyx_v_zero) {
  arrayobject *__pyx_v_op = 0;
  arrayobject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);

  /* "array.pxd":134
 *     type will be same as template.
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)             # <<<<<<<<<<<<<<
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 */
  __pyx_t_1 = ((PyObject *)newarrayobject(Py_TYPE(((PyObject *)__pyx_v_template)), __pyx_v_length, __pyx_v_template->ob_descr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "array.pxd":135
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:             # <<<<<<<<<<<<<<
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op
 */
  __pyx_t_3 = (__pyx_v_zero != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (((PyObject *)__pyx_v_op) != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "array.pxd":136
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 *     return op
 * 
 */
    (void)(memset(__pyx_v_op->data.as_chars, 0, (__pyx_v_length * __pyx_v_op->ob_descr->itemsize)));

    /* "array.pxd":135
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:             # <<<<<<<<<<<<<<
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op
 */
  }

  /* "array.pxd":137
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op             # <<<<<<<<<<<<<<
 * 
 * cdef inline array copy(array self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_op));
  __pyx_r = __pyx_v_op;
  goto __pyx_L0;

  /* "array.pxd":130
 * 
 * 
 * cdef inline array clone(array template, Py_ssize_t length, bint zero):             # <<<<<<<<<<<<<<
 *     """ fast creation of a new array, given a template array.
 *     type will be same as template.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.array.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":139
 *     return op
 * 
 * cdef inline array copy(array self):             # <<<<<<<<<<<<<<
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 */

static CYTHON_INLINE arrayobject *__pyx_f_7cpython_5array_copy(arrayobject *__pyx_v_self) {
  arrayobject *__pyx_v_op = 0;
  arrayobject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "array.pxd":141
 * cdef inline array copy(array self):
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)             # <<<<<<<<<<<<<<
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)
 *     return op
 */
  __pyx_t_1 = ((PyObject *)newarrayobject(Py_TYPE(((PyObject *)__pyx_v_self)), Py_SIZE(((PyObject *)__pyx_v_self)), __pyx_v_self->ob_descr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "array.pxd":142
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 *     return op
 * 
 */
  (void)(memcpy(__pyx_v_op->data.as_chars, __pyx_v_self->data.as_chars, (Py_SIZE(((PyObject *)__pyx_v_op)) * __pyx_v_op->ob_descr->itemsize)));

  /* "array.pxd":143
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)
 *     return op             # <<<<<<<<<<<<<<
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_op));
  __pyx_r = __pyx_v_op;
  goto __pyx_L0;

  /* "array.pxd":139
 *     return op
 * 
 * cdef inline array copy(array self):             # <<<<<<<<<<<<<<
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.array.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":145
 *     return op
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:             # <<<<<<<<<<<<<<
 *     """ efficient appending of new stuff of same type
 *     (e.g. of same array type)
 */

static CYTHON_INLINE int __pyx_f_7cpython_5array_extend_buffer(arrayobject *__pyx_v_self, char *__pyx_v_stuff, Py_ssize_t __pyx_v_n) {
  Py_ssize_t __pyx_v_itemsize;
  Py_ssize_t __pyx_v_origsize;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extend_buffer", 0);

  /* "array.pxd":149
 *     (e.g. of same array type)
 *     n: number of elements (not number of bytes!) """
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)
 */
  __pyx_t_1 = __pyx_v_self->ob_descr->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "array.pxd":150
 *     n: number of elements (not number of bytes!) """
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize
 *     cdef Py_ssize_t origsize = Py_SIZE(self)             # <<<<<<<<<<<<<<
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 */
  __pyx_v_origsize = Py_SIZE(((PyObject *)__pyx_v_self));

  /* "array.pxd":151
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)             # <<<<<<<<<<<<<<
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 *     return 0
 */
  __pyx_t_1 = resize_smart(__pyx_v_self, (__pyx_v_origsize + __pyx_v_n)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 151, __pyx_L1_error)

  /* "array.pxd":152
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  (void)(memcpy((__pyx_v_self->data.as_chars + (__pyx_v_origsize * __pyx_v_itemsize)), __pyx_v_stuff, (__pyx_v_n * __pyx_v_itemsize)));

  /* "array.pxd":153
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef inline int extend(array self, array other) except -1:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "array.pxd":145
 *     return op
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:             # <<<<<<<<<<<<<<
 *     """ efficient appending of new stuff of same type
 *     (e.g. of same array type)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cpython.array.extend_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":155
 *     return 0
 * 
 * cdef inline int extend(array self, array other) except -1:             # <<<<<<<<<<<<<<
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 */

static CYTHON_INLINE int __pyx_f_7cpython_5array_extend(arrayobject *__pyx_v_self, arrayobject *__pyx_v_other) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extend", 0);

  /* "array.pxd":157
 * cdef inline int extend(array self, array other) except -1:
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:             # <<<<<<<<<<<<<<
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 */
  __pyx_t_1 = ((__pyx_v_self->ob_descr->typecode != __pyx_v_other->ob_descr->typecode) != 0);
  if (__pyx_t_1) {

    /* "array.pxd":158
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 *         PyErr_BadArgument()             # <<<<<<<<<<<<<<
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 */
    __pyx_t_2 = PyErr_BadArgument(); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(2, 158, __pyx_L1_error)

    /* "array.pxd":157
 * cdef inline int extend(array self, array other) except -1:
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:             # <<<<<<<<<<<<<<
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 */
  }

  /* "array.pxd":159
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))             # <<<<<<<<<<<<<<
 * 
 * cdef inline void zero(array self):
 */
  __pyx_t_2 = __pyx_f_7cpython_5array_extend_buffer(__pyx_v_self, __pyx_v_other->data.as_chars, Py_SIZE(((PyObject *)__pyx_v_other))); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(2, 159, __pyx_L1_error)
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "array.pxd":155
 *     return 0
 * 
 * cdef inline int extend(array self, array other) except -1:             # <<<<<<<<<<<<<<
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cpython.array.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":161
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

static CYTHON_INLINE void __pyx_f_7cpython_5array_zero(arrayobject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zero", 0);

  /* "array.pxd":163
 * cdef inline void zero(array self):
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 */
  (void)(memset(__pyx_v_self->data.as_chars, 0, (Py_SIZE(((PyObject *)__pyx_v_self)) * __pyx_v_self->ob_descr->itemsize)));

  /* "array.pxd":161
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}
static struct __pyx_vtabstruct_4aldy_9indelpost_6varaln_VariantAlignment __pyx_vtable_4aldy_9indelpost_6varaln_VariantAlignment;

static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln_VariantAlignment(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)o);
  p->__pyx_vtab = __pyx_vtabptr_4aldy_9indelpost_6varaln_VariantAlignment;
  p->target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)Py_None); Py_INCREF(Py_None);
  p->__pyx___target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)Py_None); Py_INCREF(Py_None);
  p->second_target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)Py_None); Py_INCREF(Py_None);
  p->bam = ((struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *)Py_None); Py_INCREF(Py_None);
  p->__pyx___pileup = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->is_spurious_overhang = Py_None; Py_INCREF(Py_None);
  p->is_complex_input = Py_None; Py_INCREF(Py_None);
  p->contig = ((struct __pyx_obj_4aldy_9indelpost_6contig_Contig *)Py_None); Py_INCREF(Py_None);
  p->unspliced_local_reference = ((struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_4aldy_9indelpost_6varaln_VariantAlignment(PyObject *o) {
  struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *p = (struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->target);
  Py_CLEAR(p->__pyx___target);
  Py_CLEAR(p->second_target);
  Py_CLEAR(p->bam);
  Py_CLEAR(p->__pyx___pileup);
  Py_CLEAR(p->is_spurious_overhang);
  Py_CLEAR(p->is_complex_input);
  Py_CLEAR(p->contig);
  Py_CLEAR(p->unspliced_local_reference);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_4aldy_9indelpost_6varaln_VariantAlignment(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *p = (struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)o;
  if (p->target) {
    e = (*v)(((PyObject *)p->target), a); if (e) return e;
  }
  if (p->__pyx___target) {
    e = (*v)(((PyObject *)p->__pyx___target), a); if (e) return e;
  }
  if (p->second_target) {
    e = (*v)(((PyObject *)p->second_target), a); if (e) return e;
  }
  if (p->bam) {
    e = (*v)(((PyObject *)p->bam), a); if (e) return e;
  }
  if (p->__pyx___pileup) {
    e = (*v)(p->__pyx___pileup, a); if (e) return e;
  }
  if (p->is_spurious_overhang) {
    e = (*v)(p->is_spurious_overhang, a); if (e) return e;
  }
  if (p->is_complex_input) {
    e = (*v)(p->is_complex_input, a); if (e) return e;
  }
  if (p->contig) {
    e = (*v)(((PyObject *)p->contig), a); if (e) return e;
  }
  if (p->unspliced_local_reference) {
    e = (*v)(((PyObject *)p->unspliced_local_reference), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4aldy_9indelpost_6varaln_VariantAlignment(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *p = (struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)o;
  tmp = ((PyObject*)p->target);
  p->target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx___target);
  p->__pyx___target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->second_target);
  p->second_target = ((struct __pyx_obj_4aldy_9indelpost_7variant_Variant *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->bam);
  p->bam = ((struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx___pileup);
  p->__pyx___pileup = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->is_spurious_overhang);
  p->is_spurious_overhang = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->is_complex_input);
  p->is_complex_input = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->contig);
  p->contig = ((struct __pyx_obj_4aldy_9indelpost_6contig_Contig *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->unspliced_local_reference);
  p->unspliced_local_reference = ((struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_tp_richcompare_4aldy_9indelpost_6varaln_VariantAlignment(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_3__eq__(o1, o2);
    }
    case Py_NE: {
      PyObject *ret;
      ret = __pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_3__eq__(o1, o2);
      if (likely(ret && ret != Py_NotImplemented)) {
        int b = __Pyx_PyObject_IsTrue(ret); Py_DECREF(ret);
        if (unlikely(b < 0)) return NULL;
        ret = (b) ? Py_False : Py_True;
        Py_INCREF(ret);
      }
      return ret;
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_4aldy_9indelpost_6varaln_16VariantAlignment_bam(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_3bam_1__get__(o);
}

static PyObject *__pyx_getprop_4aldy_9indelpost_6varaln_16VariantAlignment_is_spurious_overhang(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_20is_spurious_overhang_1__get__(o);
}

static PyObject *__pyx_getprop_4aldy_9indelpost_6varaln_16VariantAlignment_is_complex_input(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_16is_complex_input_1__get__(o);
}

static PyObject *__pyx_getprop_4aldy_9indelpost_6varaln_16VariantAlignment_contig(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_6contig_1__get__(o);
}

static PyObject *__pyx_getprop_4aldy_9indelpost_6varaln_16VariantAlignment_unspliced_local_reference(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_25unspliced_local_reference_1__get__(o);
}

static PyMethodDef __pyx_methods_4aldy_9indelpost_6varaln_VariantAlignment[] = {
  {"get_contig", (PyCFunction)__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_7get_contig, METH_NOARGS, __pyx_doc_4aldy_9indelpost_6varaln_16VariantAlignment_6get_contig},
  {"get_target_indel", (PyCFunction)__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_9get_target_indel, METH_NOARGS, __pyx_doc_4aldy_9indelpost_6varaln_16VariantAlignment_8get_target_indel},
  {"fetch_reads", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_11fetch_reads, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4aldy_9indelpost_6varaln_16VariantAlignment_10fetch_reads},
  {"count_alleles", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_13count_alleles, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4aldy_9indelpost_6varaln_16VariantAlignment_12count_alleles},
  {"phase", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_15phase, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4aldy_9indelpost_6varaln_16VariantAlignment_14phase},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_17__reduce_cython__, METH_NOARGS, __pyx_doc_4aldy_9indelpost_6varaln_16VariantAlignment_16__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_19__setstate_cython__, METH_O, __pyx_doc_4aldy_9indelpost_6varaln_16VariantAlignment_18__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_4aldy_9indelpost_6varaln_VariantAlignment[] = {
  {(char *)"bam", __pyx_getprop_4aldy_9indelpost_6varaln_16VariantAlignment_bam, 0, (char *)0, 0},
  {(char *)"is_spurious_overhang", __pyx_getprop_4aldy_9indelpost_6varaln_16VariantAlignment_is_spurious_overhang, 0, (char *)0, 0},
  {(char *)"is_complex_input", __pyx_getprop_4aldy_9indelpost_6varaln_16VariantAlignment_is_complex_input, 0, (char *)0, 0},
  {(char *)"contig", __pyx_getprop_4aldy_9indelpost_6varaln_16VariantAlignment_contig, 0, (char *)0, 0},
  {(char *)"unspliced_local_reference", __pyx_getprop_4aldy_9indelpost_6varaln_16VariantAlignment_unspliced_local_reference, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_4aldy_9indelpost_6varaln_VariantAlignment = {
  PyVarObject_HEAD_INIT(0, 0)
  "aldy.indelpost.varaln.VariantAlignment", /*tp_name*/
  sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4aldy_9indelpost_6varaln_VariantAlignment, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_4aldy_9indelpost_6varaln_16VariantAlignment_5__hash__, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "This class accepts the target indel as :class:`~indelpost.Variant`\n    and the BAM file as `pysam.AlignmentFile <https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignmentFile>`__ .\n    Upon creation, the target indel is searched through realignment.\n    Evaluation of the equality between :class:`~indelpost.VariantAlignment` objects triggers local-phasing around the indels\n    to test the alignment identity.\n\n\n    Parameters\n    ----------\n    variant : Variant\n        :class:`~indelpost.Variant` object representing the target indel.\n\n    bam : pysam.AlignmentFile\n        BAM file supplied as\n        `pysam.AlignmentFile <https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignmentFile>`__ object.\n\n    window : integer\n        analyzes region input_indel_pos +/- window (defalt 50).\n\n    exclude_duplicates : bool\n        True (default) to exclude reads marked duplicate by\n        `MarkDuplicate <https://broadinstitute.github.io/picard/command-line-overview.html#MarkDuplicates>`__.\n\n    downsample_threshold : integer\n        downsamples to the threshold if the covearge at the input locus exceeds the threshold (default to 1000).\n\n    mapping_quality_threshold : integer\n        reads with a mapping quality (MAPQ) < the threshold (default 1) will be filtered.\n\n    base_quality_threshold : integer\n        non-reference base-calls with a quality score < the threshold (default 20) will be masked as \"N\" in the analysis.\n\n    retarget_search_window : integer\n        attempts to re-target in input_indel_pos +/- window (defalt 30) if the exact match is not found after realignment.\n\n    retarget_similarity_cutoff : float\n        re-targets to the indel sequence with a highest\n        `Ratcliff/Obershelp similarity score <https://docs.python.org/3/library/difflib.html#difflib.get_close_matches>`__ > cutoff (default 0.7).\n\n    exact_match_for_shiftable : bool\n        True to require exact (equivalent)  match for indels that are equ""ivalently alignable by shifting the position (default True)\n\n    match_score : integer\n        score (default 3) for matched bases in Smith-Waterman local alignment.\n\n    mismatch_penalty : interger\n        penalty (default 2) for mismatched bases in Smith-Waterman local alignment.\n\n    gap_open_penalty : integer\n        penalty (default 3) to create a gap in Smith-Waterman local alignment.\n\n    gap_extension_penalty : integer\n        penalty (default 1) to expend a created gap by one base in Smith-Waterman local alignment.\n\n    auto_adjust_extension_penalty : bool\n        True (default) to auto-adjust the gap open and extend penalties to find the input indel by realignment.\n\n    no_realignment : bool\n        True to only analyzed gap-aligned indels (default False)\n    ", /*tp_doc*/
  __pyx_tp_traverse_4aldy_9indelpost_6varaln_VariantAlignment, /*tp_traverse*/
  __pyx_tp_clear_4aldy_9indelpost_6varaln_VariantAlignment, /*tp_clear*/
  __pyx_tp_richcompare_4aldy_9indelpost_6varaln_VariantAlignment, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_4aldy_9indelpost_6varaln_VariantAlignment, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_4aldy_9indelpost_6varaln_VariantAlignment, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4aldy_9indelpost_6varaln_VariantAlignment, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup *__pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup[8];
static int __pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup = 0;

static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup)))) {
    o = (PyObject*)__pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup[--__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup];
    memset(o, 0, sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup)))) {
    __pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup[__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup++] = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup = {
  PyVarObject_HEAD_INIT(0, 0)
  "aldy.indelpost.varaln.__pyx_scope_struct____pyx___parse_pileup", /*tp_name*/
  sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads *__pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads[8];
static int __pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads = 0;

static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads)))) {
    o = (PyObject*)__pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads[--__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads];
    memset(o, 0, sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads(PyObject *o) {
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads)))) {
    __pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads[__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads++] = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads = {
  PyVarObject_HEAD_INIT(0, 0)
  "aldy.indelpost.varaln.__pyx_scope_struct_1_fetch_reads", /*tp_name*/
  sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads, /*tp_traverse*/
  __pyx_tp_clear_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr *__pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr[8];
static int __pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr = 0;

static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr)))) {
    o = (PyObject*)__pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr[--__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr(PyObject *o) {
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_v);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr)))) {
    __pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr[__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr++] = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "aldy.indelpost.varaln.__pyx_scope_struct_2_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles *__pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles[8];
static int __pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles = 0;

static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles)))) {
    o = (PyObject*)__pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles[--__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles];
    memset(o, 0, sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles(PyObject *o) {
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles)))) {
    __pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles[__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles++] = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles = {
  PyVarObject_HEAD_INIT(0, 0)
  "aldy.indelpost.varaln.__pyx_scope_struct_3_count_alleles", /*tp_name*/
  sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles, /*tp_traverse*/
  __pyx_tp_clear_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr *__pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr[8];
static int __pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr = 0;

static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr)))) {
    o = (PyObject*)__pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr[--__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr(PyObject *o) {
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_v);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr)))) {
    __pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr[__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr++] = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "aldy.indelpost.varaln.__pyx_scope_struct_4_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner *__pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner[8];
static int __pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner = 0;

static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner)))) {
    o = (PyObject*)__pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner[--__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner];
    memset(o, 0, sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner(PyObject *o) {
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_right_aligned_vars);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner)))) {
    __pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner[__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner++] = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner *)o;
  if (p->__pyx_v_right_aligned_vars) {
    e = (*v)(p->__pyx_v_right_aligned_vars, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner *)o;
  tmp = ((PyObject*)p->__pyx_v_right_aligned_vars);
  p->__pyx_v_right_aligned_vars = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner = {
  PyVarObject_HEAD_INIT(0, 0)
  "aldy.indelpost.varaln.__pyx_scope_struct_5_right_aligner", /*tp_name*/
  sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner, /*tp_traverse*/
  __pyx_tp_clear_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr *__pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr[8];
static int __pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr = 0;

static PyObject *__pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr)))) {
    o = (PyObject*)__pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr[--__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr(PyObject *o) {
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_v);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr)))) {
    __pyx_freelist_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr[__pyx_freecount_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr++] = ((struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr *p = (struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "aldy.indelpost.varaln.__pyx_scope_struct_6_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_varaln(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_varaln},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "varaln",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_u_A, __pyx_k_A, sizeof(__pyx_k_A), 0, 1, 0, 1},
  {&__pyx_n_u_D, __pyx_k_D, sizeof(__pyx_k_D), 0, 1, 0, 1},
  {&__pyx_n_u_H, __pyx_k_H, sizeof(__pyx_k_H), 0, 1, 0, 1},
  {&__pyx_n_u_I, __pyx_k_I, sizeof(__pyx_k_I), 0, 1, 0, 1},
  {&__pyx_n_u_M, __pyx_k_M, sizeof(__pyx_k_M), 0, 1, 0, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_u_N, __pyx_k_N, sizeof(__pyx_k_N), 0, 1, 0, 1},
  {&__pyx_n_u_P, __pyx_k_P, sizeof(__pyx_k_P), 0, 1, 0, 1},
  {&__pyx_n_u_S, __pyx_k_S, sizeof(__pyx_k_S), 0, 1, 0, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_VariantAlignment, __pyx_k_VariantAlignment, sizeof(__pyx_k_VariantAlignment), 0, 0, 1, 1},
  {&__pyx_n_s_VariantAlignment___parse_pileup, __pyx_k_VariantAlignment___parse_pileup, sizeof(__pyx_k_VariantAlignment___parse_pileup), 0, 0, 1, 1},
  {&__pyx_kp_u__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 1, 0, 0},
  {&__pyx_n_s_aldy_indelpost_varaln, __pyx_k_aldy_indelpost_varaln, sizeof(__pyx_k_aldy_indelpost_varaln), 0, 0, 1, 1},
  {&__pyx_kp_s_aldy_indelpost_varaln_pyx, __pyx_k_aldy_indelpost_varaln_pyx, sizeof(__pyx_k_aldy_indelpost_varaln_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_align, __pyx_k_align, sizeof(__pyx_k_align), 0, 0, 1, 1},
  {&__pyx_n_s_aligner, __pyx_k_aligner, sizeof(__pyx_k_aligner), 0, 0, 1, 1},
  {&__pyx_n_s_alleles, __pyx_k_alleles, sizeof(__pyx_k_alleles), 0, 0, 1, 1},
  {&__pyx_n_s_aln, __pyx_k_aln, sizeof(__pyx_k_aln), 0, 0, 1, 1},
  {&__pyx_n_u_aln_end, __pyx_k_aln_end, sizeof(__pyx_k_aln_end), 0, 1, 0, 1},
  {&__pyx_n_u_aln_start, __pyx_k_aln_start, sizeof(__pyx_k_aln_start), 0, 1, 0, 1},
  {&__pyx_n_s_alt, __pyx_k_alt, sizeof(__pyx_k_alt), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_auto_adjust_extension_penalty, __pyx_k_auto_adjust_extension_penalty, sizeof(__pyx_k_auto_adjust_extension_penalty), 0, 0, 1, 1},
  {&__pyx_n_s_bam, __pyx_k_bam, sizeof(__pyx_k_bam), 0, 0, 1, 1},
  {&__pyx_n_s_base_quality_threshold, __pyx_k_base_quality_threshold, sizeof(__pyx_k_base_quality_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_best_params, __pyx_k_best_params, sizeof(__pyx_k_best_params), 0, 0, 1, 1},
  {&__pyx_n_s_best_res, __pyx_k_best_res, sizeof(__pyx_k_best_res), 0, 0, 1, 1},
  {&__pyx_n_s_by_fragment, __pyx_k_by_fragment, sizeof(__pyx_k_by_fragment), 0, 0, 1, 1},
  {&__pyx_n_s_bypass_search, __pyx_k_bypass_search, sizeof(__pyx_k_bypass_search), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_candidate, __pyx_k_candidate, sizeof(__pyx_k_candidate), 0, 0, 1, 1},
  {&__pyx_n_s_centrality, __pyx_k_centrality, sizeof(__pyx_k_centrality), 0, 0, 1, 1},
  {&__pyx_n_s_check_overhangs, __pyx_k_check_overhangs, sizeof(__pyx_k_check_overhangs), 0, 0, 1, 1},
  {&__pyx_n_s_chrom, __pyx_k_chrom, sizeof(__pyx_k_chrom), 0, 0, 1, 1},
  {&__pyx_n_u_cigar_list, __pyx_k_cigar_list, sizeof(__pyx_k_cigar_list), 0, 1, 0, 1},
  {&__pyx_n_s_cigar_lst, __pyx_k_cigar_lst, sizeof(__pyx_k_cigar_lst), 0, 0, 1, 1},
  {&__pyx_n_u_cigar_string, __pyx_k_cigar_string, sizeof(__pyx_k_cigar_string), 0, 1, 0, 1},
  {&__pyx_n_u_cigar_updated, __pyx_k_cigar_updated, sizeof(__pyx_k_cigar_updated), 0, 1, 0, 1},
  {&__pyx_n_s_cinit___locals_lambda, __pyx_k_cinit___locals_lambda, sizeof(__pyx_k_cinit___locals_lambda), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_compare_contigs, __pyx_k_compare_contigs, sizeof(__pyx_k_compare_contigs), 0, 0, 1, 1},
  {&__pyx_n_u_complex, __pyx_k_complex, sizeof(__pyx_k_complex), 0, 1, 0, 1},
  {&__pyx_n_s_contig, __pyx_k_contig, sizeof(__pyx_k_contig), 0, 0, 1, 1},
  {&__pyx_n_s_contig_2, __pyx_k_contig_2, sizeof(__pyx_k_contig_2), 0, 0, 1, 1},
  {&__pyx_n_s_contig_seq, __pyx_k_contig_seq, sizeof(__pyx_k_contig_seq), 0, 0, 1, 1},
  {&__pyx_n_s_count_alleles, __pyx_k_count_alleles, sizeof(__pyx_k_count_alleles), 0, 0, 1, 1},
  {&__pyx_n_s_count_alleles_locals_genexpr, __pyx_k_count_alleles_locals_genexpr, sizeof(__pyx_k_count_alleles_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_count_repeats, __pyx_k_count_repeats, sizeof(__pyx_k_count_repeats), 0, 0, 1, 1},
  {&__pyx_n_u_covering, __pyx_k_covering, sizeof(__pyx_k_covering), 0, 1, 0, 1},
  {&__pyx_n_u_covering_subread, __pyx_k_covering_subread, sizeof(__pyx_k_covering_subread), 0, 1, 0, 1},
  {&__pyx_n_s_decompose_complex_variant, __pyx_k_decompose_complex_variant, sizeof(__pyx_k_decompose_complex_variant), 0, 0, 1, 1},
  {&__pyx_n_s_del_seq, __pyx_k_del_seq, sizeof(__pyx_k_del_seq), 0, 0, 1, 1},
  {&__pyx_n_s_diff, __pyx_k_diff, sizeof(__pyx_k_diff), 0, 0, 1, 1},
  {&__pyx_n_s_downsample_threshold, __pyx_k_downsample_threshold, sizeof(__pyx_k_downsample_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_estimated_count, __pyx_k_estimated_count, sizeof(__pyx_k_estimated_count), 0, 0, 1, 1},
  {&__pyx_n_s_event, __pyx_k_event, sizeof(__pyx_k_event), 0, 0, 1, 1},
  {&__pyx_n_s_event_len, __pyx_k_event_len, sizeof(__pyx_k_event_len), 0, 0, 1, 1},
  {&__pyx_n_s_exact_match_for_shiftable, __pyx_k_exact_match_for_shiftable, sizeof(__pyx_k_exact_match_for_shiftable), 0, 0, 1, 1},
  {&__pyx_n_s_exclude_duplicates, __pyx_k_exclude_duplicates, sizeof(__pyx_k_exclude_duplicates), 0, 0, 1, 1},
  {&__pyx_n_s_failed, __pyx_k_failed, sizeof(__pyx_k_failed), 0, 0, 1, 1},
  {&__pyx_n_s_fetch, __pyx_k_fetch, sizeof(__pyx_k_fetch), 0, 0, 1, 1},
  {&__pyx_n_s_fetch_reads_locals_genexpr, __pyx_k_fetch_reads_locals_genexpr, sizeof(__pyx_k_fetch_reads_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_kp_u_fetch_stragety_must_be_either_of, __pyx_k_fetch_stragety_must_be_either_of, sizeof(__pyx_k_fetch_stragety_must_be_either_of), 0, 1, 0, 0},
  {&__pyx_n_s_filter_spurious_overhangs, __pyx_k_filter_spurious_overhangs, sizeof(__pyx_k_filter_spurious_overhangs), 0, 0, 1, 1},
  {&__pyx_n_s_find_by_normalization, __pyx_k_find_by_normalization, sizeof(__pyx_k_find_by_normalization), 0, 0, 1, 1},
  {&__pyx_n_s_find_by_smith_waterman_realn, __pyx_k_find_by_smith_waterman_realn, sizeof(__pyx_k_find_by_smith_waterman_realn), 0, 0, 1, 1},
  {&__pyx_n_s_find_by_softclip_split, __pyx_k_find_by_softclip_split, sizeof(__pyx_k_find_by_softclip_split), 0, 0, 1, 1},
  {&__pyx_n_s_functools, __pyx_k_functools, sizeof(__pyx_k_functools), 0, 0, 1, 1},
  {&__pyx_n_s_fwrv, __pyx_k_fwrv, sizeof(__pyx_k_fwrv), 0, 0, 1, 1},
  {&__pyx_n_s_gap_extension, __pyx_k_gap_extension, sizeof(__pyx_k_gap_extension), 0, 0, 1, 1},
  {&__pyx_n_s_gap_extension_penalty, __pyx_k_gap_extension_penalty, sizeof(__pyx_k_gap_extension_penalty), 0, 0, 1, 1},
  {&__pyx_n_s_gap_open, __pyx_k_gap_open, sizeof(__pyx_k_gap_open), 0, 0, 1, 1},
  {&__pyx_n_s_gap_open_penalty, __pyx_k_gap_open_penalty, sizeof(__pyx_k_gap_open_penalty), 0, 0, 1, 1},
  {&__pyx_n_s_gappedaln, __pyx_k_gappedaln, sizeof(__pyx_k_gappedaln), 0, 0, 1, 1},
  {&__pyx_n_s_generate_equivalents, __pyx_k_generate_equivalents, sizeof(__pyx_k_generate_equivalents), 0, 0, 1, 1},
  {&__pyx_n_s_generate_grid, __pyx_k_generate_grid, sizeof(__pyx_k_generate_grid), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_get_contig_seq, __pyx_k_get_contig_seq, sizeof(__pyx_k_get_contig_seq), 0, 0, 1, 1},
  {&__pyx_n_s_get_gap_ptrn2, __pyx_k_get_gap_ptrn2, sizeof(__pyx_k_get_gap_ptrn2), 0, 0, 1, 1},
  {&__pyx_n_s_get_local_reference, __pyx_k_get_local_reference, sizeof(__pyx_k_get_local_reference), 0, 0, 1, 1},
  {&__pyx_n_s_get_reference_length, __pyx_k_get_reference_length, sizeof(__pyx_k_get_reference_length), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_u_greedy, __pyx_k_greedy, sizeof(__pyx_k_greedy), 0, 1, 0, 1},
  {&__pyx_n_s_grid, __pyx_k_grid, sizeof(__pyx_k_grid), 0, 0, 1, 1},
  {&__pyx_n_s_grid_search, __pyx_k_grid_search, sizeof(__pyx_k_grid_search), 0, 0, 1, 1},
  {&__pyx_n_s_h, __pyx_k_h, sizeof(__pyx_k_h), 0, 0, 1, 1},
  {&__pyx_n_s_how, __pyx_k_how, sizeof(__pyx_k_how), 0, 0, 1, 1},
  {&__pyx_n_s_hs, __pyx_k_hs, sizeof(__pyx_k_hs), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_idx, __pyx_k_idx, sizeof(__pyx_k_idx), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_indel_len, __pyx_k_indel_len, sizeof(__pyx_k_indel_len), 0, 0, 1, 1},
  {&__pyx_n_s_indel_repeat_threshold, __pyx_k_indel_repeat_threshold, sizeof(__pyx_k_indel_repeat_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_indel_seq, __pyx_k_indel_seq, sizeof(__pyx_k_indel_seq), 0, 0, 1, 1},
  {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
  {&__pyx_n_s_ins_seq, __pyx_k_ins_seq, sizeof(__pyx_k_ins_seq), 0, 0, 1, 1},
  {&__pyx_n_u_is_covering, __pyx_k_is_covering, sizeof(__pyx_k_is_covering), 0, 1, 0, 1},
  {&__pyx_n_s_is_del, __pyx_k_is_del, sizeof(__pyx_k_is_del), 0, 0, 1, 1},
  {&__pyx_n_u_is_dirty, __pyx_k_is_dirty, sizeof(__pyx_k_is_dirty), 0, 1, 0, 1},
  {&__pyx_n_s_is_gapped_aln, __pyx_k_is_gapped_aln, sizeof(__pyx_k_is_gapped_aln), 0, 0, 1, 1},
  {&__pyx_n_s_is_indel, __pyx_k_is_indel, sizeof(__pyx_k_is_indel), 0, 0, 1, 1},
  {&__pyx_n_s_is_ins, __pyx_k_is_ins, sizeof(__pyx_k_is_ins), 0, 0, 1, 1},
  {&__pyx_n_s_is_locally_ref, __pyx_k_is_locally_ref, sizeof(__pyx_k_is_locally_ref), 0, 0, 1, 1},
  {&__pyx_n_u_is_locally_ref, __pyx_k_is_locally_ref, sizeof(__pyx_k_is_locally_ref), 0, 1, 0, 1},
  {&__pyx_n_s_is_non_complex_indel, __pyx_k_is_non_complex_indel, sizeof(__pyx_k_is_non_complex_indel), 0, 0, 1, 1},
  {&__pyx_n_s_is_perfect_match, __pyx_k_is_perfect_match, sizeof(__pyx_k_is_perfect_match), 0, 0, 1, 1},
  {&__pyx_n_s_is_quality_read, __pyx_k_is_quality_read, sizeof(__pyx_k_is_quality_read), 0, 0, 1, 1},
  {&__pyx_n_u_is_reference_seq, __pyx_k_is_reference_seq, sizeof(__pyx_k_is_reference_seq), 0, 1, 0, 1},
  {&__pyx_n_u_is_reverse, __pyx_k_is_reverse, sizeof(__pyx_k_is_reverse), 0, 1, 0, 1},
  {&__pyx_n_u_is_target, __pyx_k_is_target, sizeof(__pyx_k_is_target), 0, 1, 0, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_u_local, __pyx_k_local, sizeof(__pyx_k_local), 0, 1, 0, 1},
  {&__pyx_n_s_local_threshold, __pyx_k_local_threshold, sizeof(__pyx_k_local_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_localn, __pyx_k_localn, sizeof(__pyx_k_localn), 0, 0, 1, 1},
  {&__pyx_n_s_longest_common_substring_thresho, __pyx_k_longest_common_substring_thresho, sizeof(__pyx_k_longest_common_substring_thresho), 0, 0, 1, 1},
  {&__pyx_n_u_lt_cigar, __pyx_k_lt_cigar, sizeof(__pyx_k_lt_cigar), 0, 1, 0, 1},
  {&__pyx_n_u_lt_flank, __pyx_k_lt_flank, sizeof(__pyx_k_lt_flank), 0, 1, 0, 1},
  {&__pyx_n_s_lt_len, __pyx_k_lt_len, sizeof(__pyx_k_lt_len), 0, 0, 1, 1},
  {&__pyx_n_s_lt_median, __pyx_k_lt_median, sizeof(__pyx_k_lt_median), 0, 0, 1, 1},
  {&__pyx_n_s_lt_qual, __pyx_k_lt_qual, sizeof(__pyx_k_lt_qual), 0, 0, 1, 1},
  {&__pyx_n_u_lt_qual, __pyx_k_lt_qual, sizeof(__pyx_k_lt_qual), 0, 1, 0, 1},
  {&__pyx_n_s_lt_ref, __pyx_k_lt_ref, sizeof(__pyx_k_lt_ref), 0, 0, 1, 1},
  {&__pyx_n_u_lt_ref, __pyx_k_lt_ref, sizeof(__pyx_k_lt_ref), 0, 1, 0, 1},
  {&__pyx_n_s_lt_ref_len, __pyx_k_lt_ref_len, sizeof(__pyx_k_lt_ref_len), 0, 0, 1, 1},
  {&__pyx_n_s_lt_seq, __pyx_k_lt_seq, sizeof(__pyx_k_lt_seq), 0, 0, 1, 1},
  {&__pyx_n_u_lt_seq, __pyx_k_lt_seq, sizeof(__pyx_k_lt_seq), 0, 1, 0, 1},
  {&__pyx_n_s_lt_seq_len, __pyx_k_lt_seq_len, sizeof(__pyx_k_lt_seq_len), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_make_aligner, __pyx_k_make_aligner, sizeof(__pyx_k_make_aligner), 0, 0, 1, 1},
  {&__pyx_n_s_mapping_quality_threshold, __pyx_k_mapping_quality_threshold, sizeof(__pyx_k_mapping_quality_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_mapq_thresh, __pyx_k_mapq_thresh, sizeof(__pyx_k_mapq_thresh), 0, 0, 1, 1},
  {&__pyx_n_s_match_score, __pyx_k_match_score, sizeof(__pyx_k_match_score), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_median, __pyx_k_median, sizeof(__pyx_k_median), 0, 0, 1, 1},
  {&__pyx_n_s_mismatch_penalty, __pyx_k_mismatch_penalty, sizeof(__pyx_k_mismatch_penalty), 0, 0, 1, 1},
  {&__pyx_n_s_most_common, __pyx_k_most_common, sizeof(__pyx_k_most_common), 0, 0, 1, 1},
  {&__pyx_n_s_most_common_gap_ptrn, __pyx_k_most_common_gap_ptrn, sizeof(__pyx_k_most_common_gap_ptrn), 0, 0, 1, 1},
  {&__pyx_n_s_mutation_density_threshold, __pyx_k_mutation_density_threshold, sizeof(__pyx_k_mutation_density_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_new_cigar, __pyx_k_new_cigar, sizeof(__pyx_k_new_cigar), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_no_realignment, __pyx_k_no_realignment, sizeof(__pyx_k_no_realignment), 0, 0, 1, 1},
  {&__pyx_n_u_non_target, __pyx_k_non_target, sizeof(__pyx_k_non_target), 0, 1, 0, 1},
  {&__pyx_n_s_normalize, __pyx_k_normalize, sizeof(__pyx_k_normalize), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_n_s_nxt_c, __pyx_k_nxt_c, sizeof(__pyx_k_nxt_c), 0, 0, 1, 1},
  {&__pyx_n_s_nxt_event, __pyx_k_nxt_event, sizeof(__pyx_k_nxt_event), 0, 0, 1, 1},
  {&__pyx_n_s_nxt_event_len, __pyx_k_nxt_event_len, sizeof(__pyx_k_nxt_event_len), 0, 0, 1, 1},
  {&__pyx_n_s_orig_target, __pyx_k_orig_target, sizeof(__pyx_k_orig_target), 0, 0, 1, 1},
  {&__pyx_n_s_padding_base, __pyx_k_padding_base, sizeof(__pyx_k_padding_base), 0, 0, 1, 1},
  {&__pyx_n_s_partial, __pyx_k_partial, sizeof(__pyx_k_partial), 0, 0, 1, 1},
  {&__pyx_n_s_phase, __pyx_k_phase, sizeof(__pyx_k_phase), 0, 0, 1, 1},
  {&__pyx_n_s_phase_nearby_variants, __pyx_k_phase_nearby_variants, sizeof(__pyx_k_phase_nearby_variants), 0, 0, 1, 1},
  {&__pyx_kp_u_phasing_stragety_must_be_either, __pyx_k_phasing_stragety_must_be_either, sizeof(__pyx_k_phasing_stragety_must_be_either), 0, 1, 0, 0},
  {&__pyx_n_s_pileup, __pyx_k_pileup, sizeof(__pyx_k_pileup), 0, 0, 1, 1},
  {&__pyx_n_s_pos, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 0, 1, 1},
  {&__pyx_n_s_prev_event, __pyx_k_prev_event, sizeof(__pyx_k_prev_event), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_quality_threshold, __pyx_k_quality_threshold, sizeof(__pyx_k_quality_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_quality_window, __pyx_k_quality_window, sizeof(__pyx_k_quality_window), 0, 0, 1, 1},
  {&__pyx_n_s_qualitythresh, __pyx_k_qualitythresh, sizeof(__pyx_k_qualitythresh), 0, 0, 1, 1},
  {&__pyx_n_s_qualitywindow, __pyx_k_qualitywindow, sizeof(__pyx_k_qualitywindow), 0, 0, 1, 1},
  {&__pyx_n_s_query_move, __pyx_k_query_move, sizeof(__pyx_k_query_move), 0, 0, 1, 1},
  {&__pyx_n_s_query_pos, __pyx_k_query_pos, sizeof(__pyx_k_query_pos), 0, 0, 1, 1},
  {&__pyx_n_s_random, __pyx_k_random, sizeof(__pyx_k_random), 0, 0, 1, 1},
  {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
  {&__pyx_n_u_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 1, 0, 1},
  {&__pyx_n_s_read_2, __pyx_k_read_2, sizeof(__pyx_k_read_2), 0, 0, 1, 1},
  {&__pyx_n_s_read_end, __pyx_k_read_end, sizeof(__pyx_k_read_end), 0, 0, 1, 1},
  {&__pyx_n_u_read_name, __pyx_k_read_name, sizeof(__pyx_k_read_name), 0, 1, 0, 1},
  {&__pyx_n_u_read_qual, __pyx_k_read_qual, sizeof(__pyx_k_read_qual), 0, 1, 0, 1},
  {&__pyx_n_s_read_seq, __pyx_k_read_seq, sizeof(__pyx_k_read_seq), 0, 0, 1, 1},
  {&__pyx_n_u_read_seq, __pyx_k_read_seq, sizeof(__pyx_k_read_seq), 0, 1, 0, 1},
  {&__pyx_n_s_read_start, __pyx_k_read_start, sizeof(__pyx_k_read_start), 0, 0, 1, 1},
  {&__pyx_n_u_read_start, __pyx_k_read_start, sizeof(__pyx_k_read_start), 0, 1, 0, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_ref, __pyx_k_ref, sizeof(__pyx_k_ref), 0, 0, 1, 1},
  {&__pyx_n_s_ref_move, __pyx_k_ref_move, sizeof(__pyx_k_ref_move), 0, 0, 1, 1},
  {&__pyx_n_s_ref_pos, __pyx_k_ref_pos, sizeof(__pyx_k_ref_pos), 0, 0, 1, 1},
  {&__pyx_n_s_ref_seq, __pyx_k_ref_seq, sizeof(__pyx_k_ref_seq), 0, 0, 1, 1},
  {&__pyx_n_u_ref_seq, __pyx_k_ref_seq, sizeof(__pyx_k_ref_seq), 0, 1, 0, 1},
  {&__pyx_n_s_ref_start, __pyx_k_ref_start, sizeof(__pyx_k_ref_start), 0, 0, 1, 1},
  {&__pyx_n_s_reference, __pyx_k_reference, sizeof(__pyx_k_reference), 0, 0, 1, 1},
  {&__pyx_n_s_reference_end, __pyx_k_reference_end, sizeof(__pyx_k_reference_end), 0, 0, 1, 1},
  {&__pyx_n_s_reference_start, __pyx_k_reference_start, sizeof(__pyx_k_reference_start), 0, 0, 1, 1},
  {&__pyx_n_s_relative_aln_pos, __pyx_k_relative_aln_pos, sizeof(__pyx_k_relative_aln_pos), 0, 0, 1, 1},
  {&__pyx_n_s_relative_pos, __pyx_k_relative_pos, sizeof(__pyx_k_relative_pos), 0, 0, 1, 1},
  {&__pyx_n_s_res, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
  {&__pyx_n_s_responses, __pyx_k_responses, sizeof(__pyx_k_responses), 0, 0, 1, 1},
  {&__pyx_n_s_retarget, __pyx_k_retarget, sizeof(__pyx_k_retarget), 0, 0, 1, 1},
  {&__pyx_n_s_retarget_cutoff, __pyx_k_retarget_cutoff, sizeof(__pyx_k_retarget_cutoff), 0, 0, 1, 1},
  {&__pyx_n_s_retarget_search_window, __pyx_k_retarget_search_window, sizeof(__pyx_k_retarget_search_window), 0, 0, 1, 1},
  {&__pyx_n_s_retarget_similarity_cutoff, __pyx_k_retarget_similarity_cutoff, sizeof(__pyx_k_retarget_similarity_cutoff), 0, 0, 1, 1},
  {&__pyx_n_s_right_aligned_vars, __pyx_k_right_aligned_vars, sizeof(__pyx_k_right_aligned_vars), 0, 0, 1, 1},
  {&__pyx_n_s_right_aligner, __pyx_k_right_aligner, sizeof(__pyx_k_right_aligner), 0, 0, 1, 1},
  {&__pyx_n_s_right_aligner_locals_genexpr, __pyx_k_right_aligner_locals_genexpr, sizeof(__pyx_k_right_aligner_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_rt_aln_pos, __pyx_k_rt_aln_pos, sizeof(__pyx_k_rt_aln_pos), 0, 0, 1, 1},
  {&__pyx_n_u_rt_cigar, __pyx_k_rt_cigar, sizeof(__pyx_k_rt_cigar), 0, 1, 0, 1},
  {&__pyx_n_u_rt_flank, __pyx_k_rt_flank, sizeof(__pyx_k_rt_flank), 0, 1, 0, 1},
  {&__pyx_n_s_rt_len, __pyx_k_rt_len, sizeof(__pyx_k_rt_len), 0, 0, 1, 1},
  {&__pyx_n_s_rt_median, __pyx_k_rt_median, sizeof(__pyx_k_rt_median), 0, 0, 1, 1},
  {&__pyx_n_s_rt_qual, __pyx_k_rt_qual, sizeof(__pyx_k_rt_qual), 0, 0, 1, 1},
  {&__pyx_n_u_rt_qual, __pyx_k_rt_qual, sizeof(__pyx_k_rt_qual), 0, 1, 0, 1},
  {&__pyx_n_s_rt_ref, __pyx_k_rt_ref, sizeof(__pyx_k_rt_ref), 0, 0, 1, 1},
  {&__pyx_n_u_rt_ref, __pyx_k_rt_ref, sizeof(__pyx_k_rt_ref), 0, 1, 0, 1},
  {&__pyx_n_s_rt_ref_len, __pyx_k_rt_ref_len, sizeof(__pyx_k_rt_ref_len), 0, 0, 1, 1},
  {&__pyx_n_s_rt_seq, __pyx_k_rt_seq, sizeof(__pyx_k_rt_seq), 0, 0, 1, 1},
  {&__pyx_n_u_rt_seq, __pyx_k_rt_seq, sizeof(__pyx_k_rt_seq), 0, 1, 0, 1},
  {&__pyx_n_s_rt_seq_len, __pyx_k_rt_seq_len, sizeof(__pyx_k_rt_seq_len), 0, 0, 1, 1},
  {&__pyx_n_s_sample, __pyx_k_sample, sizeof(__pyx_k_sample), 0, 0, 1, 1},
  {&__pyx_n_s_score, __pyx_k_score, sizeof(__pyx_k_score), 0, 0, 1, 1},
  {&__pyx_n_s_scores, __pyx_k_scores, sizeof(__pyx_k_scores), 0, 0, 1, 1},
  {&__pyx_n_s_seed, __pyx_k_seed, sizeof(__pyx_k_seed), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_setRead, __pyx_k_setRead, sizeof(__pyx_k_setRead), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_skip_next, __pyx_k_skip_next, sizeof(__pyx_k_skip_next), 0, 0, 1, 1},
  {&__pyx_n_s_skip_validation, __pyx_k_skip_validation, sizeof(__pyx_k_skip_validation), 0, 0, 1, 1},
  {&__pyx_n_s_softclip, __pyx_k_softclip, sizeof(__pyx_k_softclip), 0, 0, 1, 1},
  {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
  {&__pyx_n_s_sorted, __pyx_k_sorted, sizeof(__pyx_k_sorted), 0, 0, 1, 1},
  {&__pyx_n_s_split_cigar, __pyx_k_split_cigar, sizeof(__pyx_k_split_cigar), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_target, __pyx_k_target, sizeof(__pyx_k_target), 0, 0, 1, 1},
  {&__pyx_n_u_target, __pyx_k_target, sizeof(__pyx_k_target), 0, 1, 0, 1},
  {&__pyx_n_s_target_pos, __pyx_k_target_pos, sizeof(__pyx_k_target_pos), 0, 0, 1, 1},
  {&__pyx_n_u_target_right_shifted, __pyx_k_target_right_shifted, sizeof(__pyx_k_target_right_shifted), 0, 1, 0, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_three_class, __pyx_k_three_class, sizeof(__pyx_k_three_class), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_u_undetermined, __pyx_k_undetermined, sizeof(__pyx_k_undetermined), 0, 1, 0, 1},
  {&__pyx_n_s_unspl_loc_ref, __pyx_k_unspl_loc_ref, sizeof(__pyx_k_unspl_loc_ref), 0, 0, 1, 1},
  {&__pyx_n_s_unspliced, __pyx_k_unspliced, sizeof(__pyx_k_unspliced), 0, 0, 1, 1},
  {&__pyx_n_s_update_pileup, __pyx_k_update_pileup, sizeof(__pyx_k_update_pileup), 0, 0, 1, 1},
  {&__pyx_n_s_update_read_info, __pyx_k_update_read_info, sizeof(__pyx_k_update_read_info), 0, 0, 1, 1},
  {&__pyx_n_s_update_spliced_read_info, __pyx_k_update_spliced_read_info, sizeof(__pyx_k_update_spliced_read_info), 0, 0, 1, 1},
  {&__pyx_n_s_updated_read, __pyx_k_updated_read, sizeof(__pyx_k_updated_read), 0, 0, 1, 1},
  {&__pyx_n_s_updated_reads, __pyx_k_updated_reads, sizeof(__pyx_k_updated_reads), 0, 0, 1, 1},
  {&__pyx_n_s_utilities, __pyx_k_utilities, sizeof(__pyx_k_utilities), 0, 0, 1, 1},
  {&__pyx_n_s_window, __pyx_k_window, sizeof(__pyx_k_window), 0, 0, 1, 1},
  {&__pyx_n_s_within, __pyx_k_within, sizeof(__pyx_k_within), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 527, __pyx_L1_error)
  __pyx_builtin_max = __Pyx_GetBuiltinName(__pyx_n_s_max); if (!__pyx_builtin_max) __PYX_ERR(0, 565, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_sorted = __Pyx_GetBuiltinName(__pyx_n_s_sorted); if (!__pyx_builtin_sorted) __PYX_ERR(0, 922, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 1018, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 1207, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(2, 109, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "aldy/indelpost/varaln.pyx":573
 *             return [read["read"] for read in self.__pileup if read["is_covering"]]
 *         else:
 *             raise Exception("fetch stragety must be either of target, non_target, covering")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_fetch_stragety_must_be_either_of); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "aldy/indelpost/varaln.pyx":759
 *             hard, to_complex = False, False
 *         else:
 *             raise Exception("phasing stragety must be either of local, greedy, complex")             # <<<<<<<<<<<<<<
 * 
 *         if indel_repeat_threshold is None:
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_phasing_stragety_must_be_either); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "aldy/indelpost/varaln.pyx":1019
 *     skip_next = False
 *     for i, c in enumerate(cigar_lst):
 *         event, event_len = c[-1], int(c[:-1])             # <<<<<<<<<<<<<<
 * 
 *         if event_len < 0:
 */
  __pyx_slice__5 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__5)) __PYX_ERR(0, 1019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__5);
  __Pyx_GIVEREF(__pyx_slice__5);

  /* "aldy/indelpost/varaln.pyx":1128
 * ):
 *     if auto_adjust_extension_penalty:
 *         if (gap_open_penalty, gap_extension_penalty) != (3, 1):             # <<<<<<<<<<<<<<
 *             if len(target.indel_seq) < 20:
 *                 grid = [(gap_open_penalty, gap_extension_penalty),
 */
  __pyx_tuple__7 = PyTuple_Pack(2, __pyx_int_3, __pyx_int_1); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 1128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "aldy/indelpost/varaln.pyx":1131
 *             if len(target.indel_seq) < 20:
 *                 grid = [(gap_open_penalty, gap_extension_penalty),
 *                             (3, 1), (3, 0), (5, 1), (5, 0), (4, 1), (4, 0)             # <<<<<<<<<<<<<<
 *                        ]
 *             else:
 */
  __pyx_tuple__8 = PyTuple_Pack(2, __pyx_int_3, __pyx_int_0); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __pyx_tuple__9 = PyTuple_Pack(2, __pyx_int_5, __pyx_int_1); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __pyx_tuple__10 = PyTuple_Pack(2, __pyx_int_5, __pyx_int_0); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_tuple__11 = PyTuple_Pack(2, __pyx_int_4, __pyx_int_1); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  __pyx_tuple__12 = PyTuple_Pack(2, __pyx_int_4, __pyx_int_0); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "aldy/indelpost/varaln.pyx":39
 * from pysam.libcalignmentfile cimport AlignmentFile
 * 
 * random.seed(123)             # <<<<<<<<<<<<<<
 * 
 * cdef class VariantAlignment:
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_int_123); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "aldy/indelpost/varaln.pyx":778
 * 
 * 
 * def is_quality_read(read, pos, qualitywindow, qualitythresh):             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __pyx_tuple__14 = PyTuple_Pack(8, __pyx_n_s_read, __pyx_n_s_pos, __pyx_n_s_qualitywindow, __pyx_n_s_qualitythresh, __pyx_n_s_lt_qual, __pyx_n_s_rt_qual, __pyx_n_s_lt_median, __pyx_n_s_rt_median); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_codeobj__15 = (PyObject*)__Pyx_PyCode_New(4, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__14, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_varaln_pyx, __pyx_n_s_is_quality_read, 778, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__15)) __PYX_ERR(0, 778, __pyx_L1_error)

  /* "aldy/indelpost/varaln.pyx":799
 * 
 * 
 * def is_locally_ref(read, pos):             # <<<<<<<<<<<<<<
 *     if read["is_reference_seq"]:
 *         return True
 */
  __pyx_tuple__16 = PyTuple_Pack(12, __pyx_n_s_read, __pyx_n_s_pos, __pyx_n_s_lt_seq, __pyx_n_s_rt_seq, __pyx_n_s_lt_ref, __pyx_n_s_rt_ref, __pyx_n_s_lt_seq_len, __pyx_n_s_lt_ref_len, __pyx_n_s_lt_len, __pyx_n_s_rt_seq_len, __pyx_n_s_rt_ref_len, __pyx_n_s_rt_len); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(2, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_varaln_pyx, __pyx_n_s_is_locally_ref, 799, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(0, 799, __pyx_L1_error)

  /* "aldy/indelpost/varaln.pyx":879
 *     return True
 * 
 * def centrality(read, target_pos):             # <<<<<<<<<<<<<<
 *     relative_pos = relative_aln_pos(read["ref_seq"], read["cigar_list"], read["aln_start"], target_pos)
 *     return abs(0.5 - relative_pos)
 */
  __pyx_tuple__18 = PyTuple_Pack(3, __pyx_n_s_read, __pyx_n_s_target_pos, __pyx_n_s_relative_pos); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_varaln_pyx, __pyx_n_s_centrality, 879, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 879, __pyx_L1_error)

  /* "aldy/indelpost/varaln.pyx":970
 * 
 * 
 * def update_spliced_read_info(             # <<<<<<<<<<<<<<
 *     read,
 *     target,
 */
  __pyx_tuple__20 = PyTuple_Pack(14, __pyx_n_s_read, __pyx_n_s_target, __pyx_n_s_orig_target, __pyx_n_s_is_gapped_aln, __pyx_n_s_window, __pyx_n_s_match_score, __pyx_n_s_mismatch_penalty, __pyx_n_s_gap_open_penalty, __pyx_n_s_gap_extension_penalty, __pyx_n_s_unspl_loc_ref, __pyx_n_s_ref_seq, __pyx_n_s_lt_len, __pyx_n_s_aligner, __pyx_n_s_ref_start); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(10, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_varaln_pyx, __pyx_n_s_update_spliced_read_info, 970, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 970, __pyx_L1_error)

  /* "aldy/indelpost/varaln.pyx":999
 *     return right_aligner(read, target)
 * 
 * def right_aligner(read, target):             # <<<<<<<<<<<<<<
 *     """Right align indels around splice site"""
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(30, __pyx_n_s_read, __pyx_n_s_target, __pyx_n_s_cigar_lst, __pyx_n_s_query_pos, __pyx_n_s_ref_pos, __pyx_n_s_new_cigar, __pyx_n_s_prev_event, __pyx_n_s_skip_next, __pyx_n_s_i, __pyx_n_s_c, __pyx_n_s_event, __pyx_n_s_event_len, __pyx_n_s_query_move, __pyx_n_s_ref_move, __pyx_n_s_nxt_c, __pyx_n_s_nxt_event, __pyx_n_s_nxt_event_len, __pyx_n_s_chrom, __pyx_n_s_reference, __pyx_n_s_padding_base, __pyx_n_s_ins_seq, __pyx_n_s_ref, __pyx_n_s_alt, __pyx_n_s_del_seq, __pyx_n_s_right_aligned_vars, __pyx_n_s_diff, __pyx_n_s_rt_aln_pos, __pyx_n_s_indel_len, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(2, 0, 30, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_varaln_pyx, __pyx_n_s_right_aligner, 999, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 999, __pyx_L1_error)

  /* "aldy/indelpost/varaln.pyx":1122
 * 
 * 
 * def generate_grid (auto_adjust_extension_penalty,             # <<<<<<<<<<<<<<
 *                    gap_open_penalty,
 *                    gap_extension_penalty,
 */
  __pyx_tuple__24 = PyTuple_Pack(5, __pyx_n_s_auto_adjust_extension_penalty, __pyx_n_s_gap_open_penalty, __pyx_n_s_gap_extension_penalty, __pyx_n_s_target, __pyx_n_s_grid); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_varaln_pyx, __pyx_n_s_generate_grid, 1122, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 1122, __pyx_L1_error)

  /* "aldy/indelpost/varaln.pyx":1148
 * 
 * 
 * def grid_search(             # <<<<<<<<<<<<<<
 *     target,
 *     pileup,
 */
  __pyx_tuple__26 = PyTuple_Pack(30, __pyx_n_s_target, __pyx_n_s_pileup, __pyx_n_s_window, __pyx_n_s_mapq_thresh, __pyx_n_s_within, __pyx_n_s_retarget_cutoff, __pyx_n_s_match_score, __pyx_n_s_mismatch_penalty, __pyx_n_s_grid, __pyx_n_s_unspl_loc_ref, __pyx_n_s_exact_match_for_shiftable, __pyx_n_s_h, __pyx_n_s_responses, __pyx_n_s_scores, __pyx_n_s_hs, __pyx_n_s_res, __pyx_n_s_score, __pyx_n_s_idx, __pyx_n_s_best_res, __pyx_n_s_best_params, __pyx_n_s_is_gapped_aln, __pyx_n_s_candidate, __pyx_n_s_gap_open_penalty, __pyx_n_s_gap_extension_penalty, __pyx_n_s_updated_reads, __pyx_n_s_read, __pyx_n_s_aligner, __pyx_n_s_ref_seq, __pyx_n_s_ref_start, __pyx_n_s_updated_read); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(11, 0, 30, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_varaln_pyx, __pyx_n_s_grid_search, 1148, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 1148, __pyx_L1_error)

  /* "aldy/indelpost/varaln.pyx":1228
 * 
 * 
 * def is_perfect_match(aligner, contig_seq, read_seq):             # <<<<<<<<<<<<<<
 *     aligner.setRead(read_seq)
 *     _aln = aligner.align(gap_open=len(read_seq), gap_extension=len(read_seq))
 */
  __pyx_tuple__28 = PyTuple_Pack(6, __pyx_n_s_aligner, __pyx_n_s_contig_seq, __pyx_n_s_read_seq, __pyx_n_s_aln, __pyx_n_s_contig_2, __pyx_n_s_read_2); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 1228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(3, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_varaln_pyx, __pyx_n_s_is_perfect_match, 1228, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 1228, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_get.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyList_Type_index.type = (PyObject*)&PyList_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_5 = PyFloat_FromDouble(0.5); if (unlikely(!__pyx_float_0_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1_0 = PyFloat_FromDouble(1.0); if (unlikely(!__pyx_float_1_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_05 = PyFloat_FromDouble(0.05); if (unlikely(!__pyx_float_0_05)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_15 = PyInt_FromLong(15); if (unlikely(!__pyx_int_15)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_20 = PyInt_FromLong(20); if (unlikely(!__pyx_int_20)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_123 = PyInt_FromLong(123); if (unlikely(!__pyx_int_123)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_4aldy_9indelpost_6varaln_VariantAlignment = &__pyx_vtable_4aldy_9indelpost_6varaln_VariantAlignment;
  __pyx_vtable_4aldy_9indelpost_6varaln_VariantAlignment.__pyx___parse_pileup = (PyObject *(*)(struct __pyx_obj_4aldy_9indelpost_6varaln_VariantAlignment *, struct __pyx_opt_args_4aldy_9indelpost_6varaln_16VariantAlignment___parse_pileup *__pyx_optional_args))__pyx_f_4aldy_9indelpost_6varaln_16VariantAlignment___parse_pileup;
  if (PyType_Ready(&__pyx_type_4aldy_9indelpost_6varaln_VariantAlignment) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_4aldy_9indelpost_6varaln_VariantAlignment.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4aldy_9indelpost_6varaln_VariantAlignment.tp_dictoffset && __pyx_type_4aldy_9indelpost_6varaln_VariantAlignment.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4aldy_9indelpost_6varaln_VariantAlignment.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_4aldy_9indelpost_6varaln_VariantAlignment.tp_dict, __pyx_vtabptr_4aldy_9indelpost_6varaln_VariantAlignment) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_VariantAlignment, (PyObject *)&__pyx_type_4aldy_9indelpost_6varaln_VariantAlignment) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_4aldy_9indelpost_6varaln_VariantAlignment) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __pyx_ptype_4aldy_9indelpost_6varaln_VariantAlignment = &__pyx_type_4aldy_9indelpost_6varaln_VariantAlignment;
  if (PyType_Ready(&__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup) < 0) __PYX_ERR(0, 171, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup.tp_dictoffset && __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup = &__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct____pyx___parse_pileup;
  if (PyType_Ready(&__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads) < 0) __PYX_ERR(0, 549, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads.tp_dictoffset && __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads = &__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_1_fetch_reads;
  if (PyType_Ready(&__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 565, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr.tp_dictoffset && __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr = &__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_2_genexpr;
  if (PyType_Ready(&__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles) < 0) __PYX_ERR(0, 576, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles.tp_dictoffset && __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles = &__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_3_count_alleles;
  if (PyType_Ready(&__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr) < 0) __PYX_ERR(0, 603, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr.tp_dictoffset && __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr = &__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_4_genexpr;
  if (PyType_Ready(&__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner) < 0) __PYX_ERR(0, 999, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner.tp_dictoffset && __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner = &__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_5_right_aligner;
  if (PyType_Ready(&__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr) < 0) __PYX_ERR(0, 1056, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr.tp_dictoffset && __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr = &__pyx_type_4aldy_9indelpost_6varaln___pyx_scope_struct_6_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(5, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("array"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_5array_array = __Pyx_ImportType(__pyx_t_1, "array", "array", sizeof(arrayobject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_5array_array) __PYX_ERR(2, 58, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pysam.libchtslib"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 2680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pysam_10libchtslib_HTSFile = __Pyx_ImportType(__pyx_t_1, "pysam.libchtslib", "HTSFile", sizeof(struct __pyx_obj_5pysam_10libchtslib_HTSFile), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_10libchtslib_HTSFile) __PYX_ERR(6, 2680, __pyx_L1_error)
  __pyx_vtabptr_5pysam_10libchtslib_HTSFile = (struct __pyx_vtabstruct_5pysam_10libchtslib_HTSFile*)__Pyx_GetVtable(__pyx_ptype_5pysam_10libchtslib_HTSFile->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_10libchtslib_HTSFile)) __PYX_ERR(6, 2680, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pysam.libcfaidx"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pysam_9libcfaidx_FastaFile = __Pyx_ImportType(__pyx_t_1, "pysam.libcfaidx", "FastaFile", sizeof(struct __pyx_obj_5pysam_9libcfaidx_FastaFile), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_9libcfaidx_FastaFile) __PYX_ERR(7, 37, __pyx_L1_error)
  __pyx_vtabptr_5pysam_9libcfaidx_FastaFile = (struct __pyx_vtabstruct_5pysam_9libcfaidx_FastaFile*)__Pyx_GetVtable(__pyx_ptype_5pysam_9libcfaidx_FastaFile->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_9libcfaidx_FastaFile)) __PYX_ERR(7, 37, __pyx_L1_error)
  __pyx_ptype_5pysam_9libcfaidx_FastqProxy = __Pyx_ImportType(__pyx_t_1, "pysam.libcfaidx", "FastqProxy", sizeof(struct __pyx_obj_5pysam_9libcfaidx_FastqProxy), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_9libcfaidx_FastqProxy) __PYX_ERR(7, 45, __pyx_L1_error)
  __pyx_vtabptr_5pysam_9libcfaidx_FastqProxy = (struct __pyx_vtabstruct_5pysam_9libcfaidx_FastqProxy*)__Pyx_GetVtable(__pyx_ptype_5pysam_9libcfaidx_FastqProxy->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_9libcfaidx_FastqProxy)) __PYX_ERR(7, 45, __pyx_L1_error)
  __pyx_ptype_5pysam_9libcfaidx_FastxRecord = __Pyx_ImportType(__pyx_t_1, "pysam.libcfaidx", "FastxRecord", sizeof(struct __pyx_obj_5pysam_9libcfaidx_FastxRecord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_9libcfaidx_FastxRecord) __PYX_ERR(7, 52, __pyx_L1_error)
  __pyx_vtabptr_5pysam_9libcfaidx_FastxRecord = (struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxRecord*)__Pyx_GetVtable(__pyx_ptype_5pysam_9libcfaidx_FastxRecord->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_9libcfaidx_FastxRecord)) __PYX_ERR(7, 52, __pyx_L1_error)
  __pyx_ptype_5pysam_9libcfaidx_FastxFile = __Pyx_ImportType(__pyx_t_1, "pysam.libcfaidx", "FastxFile", sizeof(struct __pyx_obj_5pysam_9libcfaidx_FastxFile), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_9libcfaidx_FastxFile) __PYX_ERR(7, 61, __pyx_L1_error)
  __pyx_vtabptr_5pysam_9libcfaidx_FastxFile = (struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxFile*)__Pyx_GetVtable(__pyx_ptype_5pysam_9libcfaidx_FastxFile->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_9libcfaidx_FastxFile)) __PYX_ERR(7, 61, __pyx_L1_error)
  __pyx_ptype_5pysam_9libcfaidx_FastqFile = __Pyx_ImportType(__pyx_t_1, "pysam.libcfaidx", "FastqFile", sizeof(struct __pyx_obj_5pysam_9libcfaidx_FastqFile), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_9libcfaidx_FastqFile) __PYX_ERR(7, 73, __pyx_L1_error)
  __pyx_vtabptr_5pysam_9libcfaidx_FastqFile = (struct __pyx_vtabstruct_5pysam_9libcfaidx_FastqFile*)__Pyx_GetVtable(__pyx_ptype_5pysam_9libcfaidx_FastqFile->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_9libcfaidx_FastqFile)) __PYX_ERR(7, 73, __pyx_L1_error)
  __pyx_ptype_5pysam_9libcfaidx_Fastafile = __Pyx_ImportType(__pyx_t_1, "pysam.libcfaidx", "Fastafile", sizeof(struct __pyx_obj_5pysam_9libcfaidx_Fastafile), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_9libcfaidx_Fastafile) __PYX_ERR(7, 78, __pyx_L1_error)
  __pyx_vtabptr_5pysam_9libcfaidx_Fastafile = (struct __pyx_vtabstruct_5pysam_9libcfaidx_Fastafile*)__Pyx_GetVtable(__pyx_ptype_5pysam_9libcfaidx_Fastafile->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_9libcfaidx_Fastafile)) __PYX_ERR(7, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pysam.libcalignedsegment"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pysam_18libcalignedsegment_AlignedSegment = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignedsegment", "AlignedSegment", sizeof(struct __pyx_obj_5pysam_18libcalignedsegment_AlignedSegment), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_18libcalignedsegment_AlignedSegment) __PYX_ERR(8, 34, __pyx_L1_error)
  __pyx_vtabptr_5pysam_18libcalignedsegment_AlignedSegment = (struct __pyx_vtabstruct_5pysam_18libcalignedsegment_AlignedSegment*)__Pyx_GetVtable(__pyx_ptype_5pysam_18libcalignedsegment_AlignedSegment->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_18libcalignedsegment_AlignedSegment)) __PYX_ERR(8, 34, __pyx_L1_error)
  __pyx_ptype_5pysam_18libcalignedsegment_PileupColumn = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignedsegment", "PileupColumn", sizeof(struct __pyx_obj_5pysam_18libcalignedsegment_PileupColumn), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_18libcalignedsegment_PileupColumn) __PYX_ERR(8, 65, __pyx_L1_error)
  __pyx_ptype_5pysam_18libcalignedsegment_PileupRead = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignedsegment", "PileupRead", sizeof(struct __pyx_obj_5pysam_18libcalignedsegment_PileupRead), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_18libcalignedsegment_PileupRead) __PYX_ERR(8, 75, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pysam.libcalignmentfile"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pysam_17libcalignmentfile_AlignmentHeader = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignmentfile", "AlignmentHeader", sizeof(struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentHeader), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_17libcalignmentfile_AlignmentHeader) __PYX_ERR(9, 41, __pyx_L1_error)
  __pyx_ptype_5pysam_17libcalignmentfile_AlignmentFile = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignmentfile", "AlignmentFile", sizeof(struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_17libcalignmentfile_AlignmentFile) __PYX_ERR(9, 44, __pyx_L1_error)
  __pyx_vtabptr_5pysam_17libcalignmentfile_AlignmentFile = (struct __pyx_vtabstruct_5pysam_17libcalignmentfile_AlignmentFile*)__Pyx_GetVtable(__pyx_ptype_5pysam_17libcalignmentfile_AlignmentFile->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_17libcalignmentfile_AlignmentFile)) __PYX_ERR(9, 44, __pyx_L1_error)
  __pyx_ptype_5pysam_17libcalignmentfile_IteratorRow = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignmentfile", "IteratorRow", sizeof(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRow), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_17libcalignmentfile_IteratorRow) __PYX_ERR(9, 61, __pyx_L1_error)
  __pyx_ptype_5pysam_17libcalignmentfile_IteratorRowRegion = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignmentfile", "IteratorRowRegion", sizeof(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowRegion), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_17libcalignmentfile_IteratorRowRegion) __PYX_ERR(9, 71, __pyx_L1_error)
  __pyx_vtabptr_5pysam_17libcalignmentfile_IteratorRowRegion = (struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowRegion*)__Pyx_GetVtable(__pyx_ptype_5pysam_17libcalignmentfile_IteratorRowRegion->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorRowRegion)) __PYX_ERR(9, 71, __pyx_L1_error)
  __pyx_ptype_5pysam_17libcalignmentfile_IteratorRowHead = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignmentfile", "IteratorRowHead", sizeof(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowHead), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_17libcalignmentfile_IteratorRowHead) __PYX_ERR(9, 77, __pyx_L1_error)
  __pyx_vtabptr_5pysam_17libcalignmentfile_IteratorRowHead = (struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowHead*)__Pyx_GetVtable(__pyx_ptype_5pysam_17libcalignmentfile_IteratorRowHead->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorRowHead)) __PYX_ERR(9, 77, __pyx_L1_error)
  __pyx_ptype_5pysam_17libcalignmentfile_IteratorRowAll = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignmentfile", "IteratorRowAll", sizeof(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowAll), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_17libcalignmentfile_IteratorRowAll) __PYX_ERR(9, 84, __pyx_L1_error)
  __pyx_vtabptr_5pysam_17libcalignmentfile_IteratorRowAll = (struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowAll*)__Pyx_GetVtable(__pyx_ptype_5pysam_17libcalignmentfile_IteratorRowAll->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorRowAll)) __PYX_ERR(9, 84, __pyx_L1_error)
  __pyx_ptype_5pysam_17libcalignmentfile_IteratorRowAllRefs = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignmentfile", "IteratorRowAllRefs", sizeof(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowAllRefs), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_17libcalignmentfile_IteratorRowAllRefs) __PYX_ERR(9, 89, __pyx_L1_error)
  __pyx_ptype_5pysam_17libcalignmentfile_IteratorRowSelection = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignmentfile", "IteratorRowSelection", sizeof(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorRowSelection), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_17libcalignmentfile_IteratorRowSelection) __PYX_ERR(9, 94, __pyx_L1_error)
  __pyx_vtabptr_5pysam_17libcalignmentfile_IteratorRowSelection = (struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorRowSelection*)__Pyx_GetVtable(__pyx_ptype_5pysam_17libcalignmentfile_IteratorRowSelection->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorRowSelection)) __PYX_ERR(9, 94, __pyx_L1_error)
  __pyx_ptype_5pysam_17libcalignmentfile_IteratorColumn = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignmentfile", "IteratorColumn", sizeof(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumn), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_17libcalignmentfile_IteratorColumn) __PYX_ERR(9, 101, __pyx_L1_error)
  __pyx_vtabptr_5pysam_17libcalignmentfile_IteratorColumn = (struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumn*)__Pyx_GetVtable(__pyx_ptype_5pysam_17libcalignmentfile_IteratorColumn->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorColumn)) __PYX_ERR(9, 101, __pyx_L1_error)
  __pyx_ptype_5pysam_17libcalignmentfile_IteratorColumnRegion = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignmentfile", "IteratorColumnRegion", sizeof(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumnRegion), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_17libcalignmentfile_IteratorColumnRegion) __PYX_ERR(9, 133, __pyx_L1_error)
  __pyx_vtabptr_5pysam_17libcalignmentfile_IteratorColumnRegion = (struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumnRegion*)__Pyx_GetVtable(__pyx_ptype_5pysam_17libcalignmentfile_IteratorColumnRegion->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorColumnRegion)) __PYX_ERR(9, 133, __pyx_L1_error)
  __pyx_ptype_5pysam_17libcalignmentfile_IteratorColumnAllRefs = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignmentfile", "IteratorColumnAllRefs", sizeof(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumnAllRefs), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_17libcalignmentfile_IteratorColumnAllRefs) __PYX_ERR(9, 139, __pyx_L1_error)
  __pyx_vtabptr_5pysam_17libcalignmentfile_IteratorColumnAllRefs = (struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumnAllRefs*)__Pyx_GetVtable(__pyx_ptype_5pysam_17libcalignmentfile_IteratorColumnAllRefs->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorColumnAllRefs)) __PYX_ERR(9, 139, __pyx_L1_error)
  __pyx_ptype_5pysam_17libcalignmentfile_IteratorColumnAll = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignmentfile", "IteratorColumnAll", sizeof(struct __pyx_obj_5pysam_17libcalignmentfile_IteratorColumnAll), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_17libcalignmentfile_IteratorColumnAll) __PYX_ERR(9, 143, __pyx_L1_error)
  __pyx_vtabptr_5pysam_17libcalignmentfile_IteratorColumnAll = (struct __pyx_vtabstruct_5pysam_17libcalignmentfile_IteratorColumnAll*)__Pyx_GetVtable(__pyx_ptype_5pysam_17libcalignmentfile_IteratorColumnAll->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_17libcalignmentfile_IteratorColumnAll)) __PYX_ERR(9, 143, __pyx_L1_error)
  __pyx_ptype_5pysam_17libcalignmentfile_IndexedReads = __Pyx_ImportType(__pyx_t_1, "pysam.libcalignmentfile", "IndexedReads", sizeof(struct __pyx_obj_5pysam_17libcalignmentfile_IndexedReads), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_17libcalignmentfile_IndexedReads) __PYX_ERR(9, 147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pysam.libcbcf"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pysam_7libcbcf_VariantHeader = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantHeader", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantHeader), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantHeader) __PYX_ERR(10, 38, __pyx_L1_error)
  __pyx_vtabptr_5pysam_7libcbcf_VariantHeader = (struct __pyx_vtabstruct_5pysam_7libcbcf_VariantHeader*)__Pyx_GetVtable(__pyx_ptype_5pysam_7libcbcf_VariantHeader->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_7libcbcf_VariantHeader)) __PYX_ERR(10, 38, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantHeaderRecord = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantHeaderRecord", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantHeaderRecord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantHeaderRecord) __PYX_ERR(10, 46, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantHeaderRecords = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantHeaderRecords", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantHeaderRecords), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantHeaderRecords) __PYX_ERR(10, 51, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantHeaderContigs = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantHeaderContigs", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantHeaderContigs), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantHeaderContigs) __PYX_ERR(10, 55, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantHeaderSamples = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantHeaderSamples", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantHeaderSamples), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantHeaderSamples) __PYX_ERR(10, 59, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantContig = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantContig", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantContig), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantContig) __PYX_ERR(10, 63, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantMetadata = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantMetadata", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantMetadata), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantMetadata) __PYX_ERR(10, 68, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantHeaderMetadata = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantHeaderMetadata", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantHeaderMetadata), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantHeaderMetadata) __PYX_ERR(10, 74, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantRecord = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantRecord", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantRecord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantRecord) __PYX_ERR(10, 79, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantRecordFilter = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantRecordFilter", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantRecordFilter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantRecordFilter) __PYX_ERR(10, 84, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantRecordFormat = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantRecordFormat", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantRecordFormat), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantRecordFormat) __PYX_ERR(10, 88, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantRecordInfo = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantRecordInfo", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantRecordInfo), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantRecordInfo) __PYX_ERR(10, 92, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantRecordSamples = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantRecordSamples", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantRecordSamples), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantRecordSamples) __PYX_ERR(10, 96, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantRecordSample = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantRecordSample", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantRecordSample), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantRecordSample) __PYX_ERR(10, 100, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_BaseIndex = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "BaseIndex", sizeof(struct __pyx_obj_5pysam_7libcbcf_BaseIndex), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_BaseIndex) __PYX_ERR(10, 105, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_BCFIndex = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "BCFIndex", sizeof(struct __pyx_obj_5pysam_7libcbcf_BCFIndex), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_BCFIndex) __PYX_ERR(10, 110, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_TabixIndex = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "TabixIndex", sizeof(struct __pyx_obj_5pysam_7libcbcf_TabixIndex), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_TabixIndex) __PYX_ERR(10, 115, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_BaseIterator = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "BaseIterator", sizeof(struct __pyx_obj_5pysam_7libcbcf_BaseIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_BaseIterator) __PYX_ERR(10, 119, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_BCFIterator = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "BCFIterator", sizeof(struct __pyx_obj_5pysam_7libcbcf_BCFIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_BCFIterator) __PYX_ERR(10, 124, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_TabixIterator = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "TabixIterator", sizeof(struct __pyx_obj_5pysam_7libcbcf_TabixIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_TabixIterator) __PYX_ERR(10, 128, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantFile = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantFile", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantFile), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantFile) __PYX_ERR(10, 133, __pyx_L1_error)
  __pyx_vtabptr_5pysam_7libcbcf_VariantFile = (struct __pyx_vtabstruct_5pysam_7libcbcf_VariantFile*)__Pyx_GetVtable(__pyx_ptype_5pysam_7libcbcf_VariantFile->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_7libcbcf_VariantFile)) __PYX_ERR(10, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("aldy.indelpost.variant"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4aldy_9indelpost_7variant_NullVariant = __Pyx_ImportType(__pyx_t_1, "aldy.indelpost.variant", "NullVariant", sizeof(struct __pyx_obj_4aldy_9indelpost_7variant_NullVariant), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4aldy_9indelpost_7variant_NullVariant) __PYX_ERR(11, 6, __pyx_L1_error)
  __pyx_ptype_4aldy_9indelpost_7variant_Variant = __Pyx_ImportType(__pyx_t_1, "aldy.indelpost.variant", "Variant", sizeof(struct __pyx_obj_4aldy_9indelpost_7variant_Variant), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4aldy_9indelpost_7variant_Variant) __PYX_ERR(11, 13, __pyx_L1_error)
  __pyx_vtabptr_4aldy_9indelpost_7variant_Variant = (struct __pyx_vtabstruct_4aldy_9indelpost_7variant_Variant*)__Pyx_GetVtable(__pyx_ptype_4aldy_9indelpost_7variant_Variant->tp_dict); if (unlikely(!__pyx_vtabptr_4aldy_9indelpost_7variant_Variant)) __PYX_ERR(11, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("aldy.indelpost.contig"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4aldy_9indelpost_6contig_FailedContig = __Pyx_ImportType(__pyx_t_1, "aldy.indelpost.contig", "FailedContig", sizeof(struct __pyx_obj_4aldy_9indelpost_6contig_FailedContig), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4aldy_9indelpost_6contig_FailedContig) __PYX_ERR(12, 4, __pyx_L1_error)
  __pyx_ptype_4aldy_9indelpost_6contig_Contig = __Pyx_ImportType(__pyx_t_1, "aldy.indelpost.contig", "Contig", sizeof(struct __pyx_obj_4aldy_9indelpost_6contig_Contig), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4aldy_9indelpost_6contig_Contig) __PYX_ERR(12, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("aldy.indelpost.local_reference"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4aldy_9indelpost_15local_reference_UnsplicedLocalReference = __Pyx_ImportType(__pyx_t_1, "aldy.indelpost.local_reference", "UnsplicedLocalReference", sizeof(struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4aldy_9indelpost_15local_reference_UnsplicedLocalReference) __PYX_ERR(13, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("aldy.indelpost.pileup"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "make_pileup", (void (**)(void))&__pyx_f_4aldy_9indelpost_6pileup_make_pileup, "PyObject *(struct __pyx_obj_4aldy_9indelpost_7variant_Variant *, struct __pyx_obj_5pysam_17libcalignmentfile_AlignmentFile *, struct __pyx_obj_4aldy_9indelpost_15local_reference_UnsplicedLocalReference *, int, int, int, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("aldy.indelpost.utilities"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "split", (void (**)(void))&__pyx_f_4aldy_9indelpost_9utilities_split, "PyObject *(PyObject *, PyObject *, int, int, int, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initvaraln(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initvaraln(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_varaln(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_varaln(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_varaln(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'varaln' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_varaln(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("varaln", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_aldy__indelpost__varaln) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "aldy.indelpost.varaln")) {
      if (unlikely(PyDict_SetItemString(modules, "aldy.indelpost.varaln", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "aldy/indelpost/varaln.pyx":5
 * 
 * cimport cython
 * import random             # <<<<<<<<<<<<<<
 * import numpy as np
 * from functools import partial
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_random, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_random, __pyx_t_1) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":6
 * cimport cython
 * import random
 * import numpy as np             # <<<<<<<<<<<<<<
 * from functools import partial
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":7
 * import random
 * import numpy as np
 * from functools import partial             # <<<<<<<<<<<<<<
 * 
 * from .pileup import (
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_partial);
  __Pyx_GIVEREF(__pyx_n_s_partial);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_partial);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_functools, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_partial); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_partial, __pyx_t_1) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":10
 * 
 * from .pileup import (
 *     update_pileup,             # <<<<<<<<<<<<<<
 *     retarget,
 *     update_read_info,
 */
  __pyx_t_2 = PyList_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_update_pileup);
  __Pyx_GIVEREF(__pyx_n_s_update_pileup);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_update_pileup);
  __Pyx_INCREF(__pyx_n_s_retarget);
  __Pyx_GIVEREF(__pyx_n_s_retarget);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_retarget);
  __Pyx_INCREF(__pyx_n_s_update_read_info);
  __Pyx_GIVEREF(__pyx_n_s_update_read_info);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_update_read_info);
  __Pyx_INCREF(__pyx_n_s_check_overhangs);
  __Pyx_GIVEREF(__pyx_n_s_check_overhangs);
  PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_check_overhangs);
  __Pyx_INCREF(__pyx_n_s_filter_spurious_overhangs);
  __Pyx_GIVEREF(__pyx_n_s_filter_spurious_overhangs);
  PyList_SET_ITEM(__pyx_t_2, 4, __pyx_n_s_filter_spurious_overhangs);

  /* "aldy/indelpost/varaln.pyx":9
 * from functools import partial
 * 
 * from .pileup import (             # <<<<<<<<<<<<<<
 *     update_pileup,
 *     retarget,
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_pileup, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_update_pileup); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_update_pileup, __pyx_t_2) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_retarget); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_retarget, __pyx_t_2) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_update_read_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_update_read_info, __pyx_t_2) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_check_overhangs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_check_overhangs, __pyx_t_2) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_filter_spurious_overhangs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_filter_spurious_overhangs, __pyx_t_2) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":16
 *     filter_spurious_overhangs,
 * )
 * from .gappedaln import find_by_normalization             # <<<<<<<<<<<<<<
 * from .softclip import find_by_softclip_split
 * from .localn import find_by_smith_waterman_realn, make_aligner
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_find_by_normalization);
  __Pyx_GIVEREF(__pyx_n_s_find_by_normalization);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_find_by_normalization);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_gappedaln, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_find_by_normalization); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_find_by_normalization, __pyx_t_1) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":17
 * )
 * from .gappedaln import find_by_normalization
 * from .softclip import find_by_softclip_split             # <<<<<<<<<<<<<<
 * from .localn import find_by_smith_waterman_realn, make_aligner
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_find_by_softclip_split);
  __Pyx_GIVEREF(__pyx_n_s_find_by_softclip_split);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_find_by_softclip_split);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_softclip, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_find_by_softclip_split); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_find_by_softclip_split, __pyx_t_2) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":18
 * from .gappedaln import find_by_normalization
 * from .softclip import find_by_softclip_split
 * from .localn import find_by_smith_waterman_realn, make_aligner             # <<<<<<<<<<<<<<
 * 
 * from .alleles import phase_nearby_variants
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_find_by_smith_waterman_realn);
  __Pyx_GIVEREF(__pyx_n_s_find_by_smith_waterman_realn);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_find_by_smith_waterman_realn);
  __Pyx_INCREF(__pyx_n_s_make_aligner);
  __Pyx_GIVEREF(__pyx_n_s_make_aligner);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_make_aligner);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_localn, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_find_by_smith_waterman_realn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_find_by_smith_waterman_realn, __pyx_t_1) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_make_aligner); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_make_aligner, __pyx_t_1) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":20
 * from .localn import find_by_smith_waterman_realn, make_aligner
 * 
 * from .alleles import phase_nearby_variants             # <<<<<<<<<<<<<<
 * from .contig import compare_contigs
 * from .utilities import (
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_phase_nearby_variants);
  __Pyx_GIVEREF(__pyx_n_s_phase_nearby_variants);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_phase_nearby_variants);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_alleles, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_phase_nearby_variants); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_phase_nearby_variants, __pyx_t_2) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":21
 * 
 * from .alleles import phase_nearby_variants
 * from .contig import compare_contigs             # <<<<<<<<<<<<<<
 * from .utilities import (
 *     get_local_reference,
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_compare_contigs);
  __Pyx_GIVEREF(__pyx_n_s_compare_contigs);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_compare_contigs);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_contig, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_compare_contigs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_compare_contigs, __pyx_t_1) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/varaln.pyx":23
 * from .contig import compare_contigs
 * from .utilities import (
 *     get_local_reference,             # <<<<<<<<<<<<<<
 *     relative_aln_pos,
 *     split_cigar,
 */
  __pyx_t_2 = PyList_New(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_get_local_reference);
  __Pyx_GIVEREF(__pyx_n_s_get_local_reference);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_get_local_reference);
  __Pyx_INCREF(__pyx_n_s_relative_aln_pos);
  __Pyx_GIVEREF(__pyx_n_s_relative_aln_pos);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_relative_aln_pos);
  __Pyx_INCREF(__pyx_n_s_split_cigar);
  __Pyx_GIVEREF(__pyx_n_s_split_cigar);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_split_cigar);
  __Pyx_INCREF(__pyx_n_s_most_common_gap_ptrn);
  __Pyx_GIVEREF(__pyx_n_s_most_common_gap_ptrn);
  PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_most_common_gap_ptrn);
  __Pyx_INCREF(__pyx_n_s_get_gap_ptrn2);
  __Pyx_GIVEREF(__pyx_n_s_get_gap_ptrn2);
  PyList_SET_ITEM(__pyx_t_2, 4, __pyx_n_s_get_gap_ptrn2);
  __Pyx_INCREF(__pyx_n_s_most_common);
  __Pyx_GIVEREF(__pyx_n_s_most_common);
  PyList_SET_ITEM(__pyx_t_2, 5, __pyx_n_s_most_common);

  /* "aldy/indelpost/varaln.pyx":22
 * from .alleles import phase_nearby_variants
 * from .contig import compare_contigs
 * from .utilities import (             # <<<<<<<<<<<<<<
 *     get_local_reference,
 *     relative_aln_pos,
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_utilities, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_get_local_reference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_local_reference, __pyx_t_2) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_relative_aln_pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_relative_aln_pos, __pyx_t_2) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_split_cigar); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_split_cigar, __pyx_t_2) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_most_common_gap_ptrn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_most_common_gap_ptrn, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_get_gap_ptrn2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_gap_ptrn2, __pyx_t_2) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_most_common); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_most_common, __pyx_t_2) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":39
 * from pysam.libcalignmentfile cimport AlignmentFile
 * 
 * random.seed(123)             # <<<<<<<<<<<<<<
 * 
 * cdef class VariantAlignment:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_random); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_seed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":778
 * 
 * 
 * def is_quality_read(read, pos, qualitywindow, qualitythresh):             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6varaln_1is_quality_read, NULL, __pyx_n_s_aldy_indelpost_varaln); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_quality_read, __pyx_t_1) < 0) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":799
 * 
 * 
 * def is_locally_ref(read, pos):             # <<<<<<<<<<<<<<
 *     if read["is_reference_seq"]:
 *         return True
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6varaln_3is_locally_ref, NULL, __pyx_n_s_aldy_indelpost_varaln); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_locally_ref, __pyx_t_1) < 0) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":879
 *     return True
 * 
 * def centrality(read, target_pos):             # <<<<<<<<<<<<<<
 *     relative_pos = relative_aln_pos(read["ref_seq"], read["cigar_list"], read["aln_start"], target_pos)
 *     return abs(0.5 - relative_pos)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6varaln_5centrality, NULL, __pyx_n_s_aldy_indelpost_varaln); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_centrality, __pyx_t_1) < 0) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":970
 * 
 * 
 * def update_spliced_read_info(             # <<<<<<<<<<<<<<
 *     read,
 *     target,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6varaln_7update_spliced_read_info, NULL, __pyx_n_s_aldy_indelpost_varaln); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_update_spliced_read_info, __pyx_t_1) < 0) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":999
 *     return right_aligner(read, target)
 * 
 * def right_aligner(read, target):             # <<<<<<<<<<<<<<
 *     """Right align indels around splice site"""
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6varaln_9right_aligner, NULL, __pyx_n_s_aldy_indelpost_varaln); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_right_aligner, __pyx_t_1) < 0) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":1122
 * 
 * 
 * def generate_grid (auto_adjust_extension_penalty,             # <<<<<<<<<<<<<<
 *                    gap_open_penalty,
 *                    gap_extension_penalty,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6varaln_11generate_grid, NULL, __pyx_n_s_aldy_indelpost_varaln); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_generate_grid, __pyx_t_1) < 0) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":1148
 * 
 * 
 * def grid_search(             # <<<<<<<<<<<<<<
 *     target,
 *     pileup,
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6varaln_13grid_search, NULL, __pyx_n_s_aldy_indelpost_varaln); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_grid_search, __pyx_t_1) < 0) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":1228
 * 
 * 
 * def is_perfect_match(aligner, contig_seq, read_seq):             # <<<<<<<<<<<<<<
 *     aligner.setRead(read_seq)
 *     _aln = aligner.align(gap_open=len(read_seq), gap_extension=len(read_seq))
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6varaln_15is_perfect_match, NULL, __pyx_n_s_aldy_indelpost_varaln); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_perfect_match, __pyx_t_1) < 0) __PYX_ERR(0, 1228, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/varaln.pyx":1
 * #cython: embedsignature=True             # <<<<<<<<<<<<<<
 * #cython: profile=False
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "array.pxd":161
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init aldy.indelpost.varaln", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init aldy.indelpost.varaln");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* py_abs */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *n) {
    if (likely(Py_SIZE(n) == -1)) {
        return PyLong_FromLong(((PyLongObject*)n)->ob_digit[0]);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    {
        PyObject *copy = _PyLong_Copy((PyLongObject*)n);
        if (likely(copy)) {
            __Pyx_SET_SIZE(copy, -Py_SIZE(copy));
        }
        return copy;
    }
#else
    return PyNumber_Negative(n);
#endif
}
#endif

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* dict_getitem_default */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value) {
    PyObject* value;
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (unlikely(PyErr_Occurred()))
            return NULL;
        value = default_value;
    }
    Py_INCREF(value);
    if ((1));
#else
    if (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key)) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            value = default_value;
        }
        Py_INCREF(value);
    }
#endif
    else {
        if (default_value == Py_None)
            value = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_get, d, key);
        else
            value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_get, d, key, default_value);
    }
    return value;
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyFloatBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_SubtractCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double a = floatval;
    double b, result;
    (void)inplace;
    (void)zerodivision_check;
    if (likely(PyFloat_CheckExact(op2))) {
        b = PyFloat_AS_DOUBLE(op2);
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        b = (double) PyInt_AS_LONG(op2);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op2))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op2)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op2);
        switch (size) {
            case  0: b = 0.0; break;
            case -1: b = -(double) digits[0]; break;
            case  1: b = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    b = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    b = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            b = -b;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            b = PyLong_AsDouble(op2);
            if (unlikely(b == -1.0 && PyErr_Occurred())) return NULL;
            
        }
    } else {
        return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
    }
        
        PyFPE_START_PROTECT("subtract", return NULL)
        result = a - b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* ObjectGetItem */
  #if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* PyFloatBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_EqObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double b = floatval;
    double a;
    (void)inplace;
    (void)zerodivision_check;
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    if (likely(PyFloat_CheckExact(op1))) {
        a = PyFloat_AS_DOUBLE(op1);
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        a = (double) PyInt_AS_LONG(op1);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op1))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        switch (size) {
            case  0: a = 0.0; break;
            case -1: a = -(double) digits[0]; break;
            case  1: a = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    a = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            return (
                PyFloat_Type.tp_richcompare(op2, op1, Py_EQ));
        }
    } else {
        return (
            PyObject_RichCompare(op1, op2, Py_EQ));
    }
        if (a == b) {
            Py_RETURN_TRUE;
        } else {
            Py_RETURN_FALSE;
        }
}
#endif

/* PyObject_GenericGetAttrNoDict */
    #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
    #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
    static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyErrExceptionMatches */
    #if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyObjectGetAttrStrNoError */
    static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
    static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
    #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
    static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* Import */
    static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
    static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CLineInTraceback */
    #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
    static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
    #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
    #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
    static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
    static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
    #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* SwapException */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectGetMethod */
    static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
    static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
    #include <structmember.h>
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
    static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
    #ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
    static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
    static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
    #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
    static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
