/* Generated by Cython 0.29.32 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/htslib_util.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/bgzf.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/cram.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/faidx.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/hfile.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/hts.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/kstring.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/sam.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/tbx.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/vcf.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib/htslib/vcfutils.h",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/pysam_stream.h"
        ],
        "extra_compile_args": [
            "-Wno-unused-function"
        ],
        "include_dirs": [
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/htslib",
            "/opt/hostedtoolcache/Python/3.7.15/x64/lib/python3.7/site-packages/pysam/include/samtools"
        ],
        "name": "aldy.indelpost.localn",
        "sources": [
            "aldy/indelpost/localn.pyx"
        ]
    },
    "module_name": "aldy.indelpost.localn"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_32"
#define CYTHON_HEX_VERSION 0x001D20F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_HEX >= 0x07030900)
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__aldy__indelpost__localn
#define __PYX_HAVE_API__aldy__indelpost__localn
/* Early includes */
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "pythread.h"
#include <sys/types.h>
#include "stdarg.h"
#include "htslib/kstring.h"
#include "htslib_util.h"
#include "htslib/hfile.h"
#include "htslib/bgzf.h"
#include "htslib/hts.h"
#include "htslib/sam.h"
#include "htslib/faidx.h"
#include "htslib/tbx.h"
#include "htslib/vcf.h"
#include "htslib/vcfutils.h"
#include "htslib/cram.h"
#include "pysam_stream.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "aldy/indelpost/localn.pyx",
  "array.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "libchtslib.pxd",
  "libcfaidx.pxd",
  "libcbcf.pxd",
  "aldy/indelpost/variant.pxd",
};

/*--- Type declarations ---*/
#ifndef _ARRAYARRAY_H
struct arrayobject;
typedef struct arrayobject arrayobject;
#endif
struct __pyx_obj_5pysam_10libchtslib_HTSFile;
struct __pyx_obj_5pysam_9libcfaidx_FastaFile;
struct __pyx_obj_5pysam_9libcfaidx_FastqProxy;
struct __pyx_obj_5pysam_9libcfaidx_FastxRecord;
struct __pyx_obj_5pysam_9libcfaidx_FastxFile;
struct __pyx_obj_5pysam_9libcfaidx_FastqFile;
struct __pyx_obj_5pysam_9libcfaidx_Fastafile;
struct __pyx_obj_5pysam_7libcbcf_VariantHeader;
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderRecord;
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderRecords;
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderContigs;
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderSamples;
struct __pyx_obj_5pysam_7libcbcf_VariantContig;
struct __pyx_obj_5pysam_7libcbcf_VariantMetadata;
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderMetadata;
struct __pyx_obj_5pysam_7libcbcf_VariantRecord;
struct __pyx_obj_5pysam_7libcbcf_VariantRecordFilter;
struct __pyx_obj_5pysam_7libcbcf_VariantRecordFormat;
struct __pyx_obj_5pysam_7libcbcf_VariantRecordInfo;
struct __pyx_obj_5pysam_7libcbcf_VariantRecordSamples;
struct __pyx_obj_5pysam_7libcbcf_VariantRecordSample;
struct __pyx_obj_5pysam_7libcbcf_BaseIndex;
struct __pyx_obj_5pysam_7libcbcf_BCFIndex;
struct __pyx_obj_5pysam_7libcbcf_TabixIndex;
struct __pyx_obj_5pysam_7libcbcf_BaseIterator;
struct __pyx_obj_5pysam_7libcbcf_BCFIterator;
struct __pyx_obj_5pysam_7libcbcf_TabixIterator;
struct __pyx_obj_5pysam_7libcbcf_VariantFile;
struct __pyx_obj_4aldy_9indelpost_7variant_NullVariant;
struct __pyx_obj_4aldy_9indelpost_7variant_Variant;
struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn;
struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr;
struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr;
struct __pyx_opt_args_5pysam_9libcfaidx_10FastqProxy_get_quality_array;
struct __pyx_opt_args_5pysam_9libcfaidx_11FastxRecord_get_quality_array;

/* "pysam/libcfaidx.pxd":49
 *     cdef cython.str to_string(self)
 *     cdef cython.str tostring(self)
 *     cpdef array.array get_quality_array(self, int offset=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_5pysam_9libcfaidx_10FastqProxy_get_quality_array {
  int __pyx_n;
  int offset;
};

/* "pysam/libcfaidx.pxd":59
 *     cdef cython.str to_string(self)
 *     cdef cython.str tostring(self)
 *     cpdef array.array get_quality_array(self, int offset=*)             # <<<<<<<<<<<<<<
 * 
 * cdef class FastxFile:
 */
struct __pyx_opt_args_5pysam_9libcfaidx_11FastxRecord_get_quality_array {
  int __pyx_n;
  int offset;
};

/* "pysam/libchtslib.pxd":2680
 * 
 * 
 * cdef class HTSFile(object):             # <<<<<<<<<<<<<<
 *     cdef          htsFile *htsfile       # pointer to htsFile structure
 *     cdef          int64_t start_offset   # BGZF offset of first record
 */
struct __pyx_obj_5pysam_10libchtslib_HTSFile {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pysam_10libchtslib_HTSFile *__pyx_vtab;
  htsFile *htsfile;
  int64_t start_offset;
  PyObject *filename;
  PyObject *mode;
  PyObject *threads;
  PyObject *index_filename;
  int is_stream;
  int is_remote;
  int duplicate_filehandle;
};


/* "pysam/libcfaidx.pxd":37
 *                     int * dret)
 * 
 * cdef class FastaFile:             # <<<<<<<<<<<<<<
 *     cdef bint is_remote
 *     cdef object _filename, _references, _lengths, reference2length
 */
struct __pyx_obj_5pysam_9libcfaidx_FastaFile {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pysam_9libcfaidx_FastaFile *__pyx_vtab;
  int is_remote;
  PyObject *_filename;
  PyObject *_references;
  PyObject *_lengths;
  PyObject *reference2length;
  faidx_t *fastafile;
};


/* "pysam/libcfaidx.pxd":45
 * 
 * 
 * cdef class FastqProxy:             # <<<<<<<<<<<<<<
 *     cdef kseq_t * _delegate
 *     cdef cython.str to_string(self)
 */
struct __pyx_obj_5pysam_9libcfaidx_FastqProxy {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pysam_9libcfaidx_FastqProxy *__pyx_vtab;
  kseq_t *_delegate;
};


/* "pysam/libcfaidx.pxd":52
 * 
 * 
 * cdef class FastxRecord:             # <<<<<<<<<<<<<<
 *     """
 *     Python container for pysam.libcfaidx.FastqProxy with persistence.
 */
struct __pyx_obj_5pysam_9libcfaidx_FastxRecord {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxRecord *__pyx_vtab;
  PyObject *comment;
  PyObject *quality;
  PyObject *sequence;
  PyObject *name;
};


/* "pysam/libcfaidx.pxd":61
 *     cpdef array.array get_quality_array(self, int offset=*)
 * 
 * cdef class FastxFile:             # <<<<<<<<<<<<<<
 *     cdef object _filename
 *     cdef BGZF * fastqfile
 */
struct __pyx_obj_5pysam_9libcfaidx_FastxFile {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxFile *__pyx_vtab;
  PyObject *_filename;
  BGZF *fastqfile;
  kseq_t *entry;
  int persist;
  int is_remote;
};


/* "pysam/libcfaidx.pxd":73
 * 
 * # Compatibility Layer for pysam 0.8.1
 * cdef class FastqFile(FastxFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_5pysam_9libcfaidx_FastqFile {
  struct __pyx_obj_5pysam_9libcfaidx_FastxFile __pyx_base;
};


/* "pysam/libcfaidx.pxd":78
 * 
 * # Compatibility Layer for pysam < 0.8
 * cdef class Fastafile(FastaFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_5pysam_9libcfaidx_Fastafile {
  struct __pyx_obj_5pysam_9libcfaidx_FastaFile __pyx_base;
};


/* "pysam/libcbcf.pxd":38
 * 
 * 
 * cdef class VariantHeader(object):             # <<<<<<<<<<<<<<
 *     cdef bcf_hdr_t *ptr
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantHeader {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pysam_7libcbcf_VariantHeader *__pyx_vtab;
  bcf_hdr_t *ptr;
};


/* "pysam/libcbcf.pxd":46
 * 
 * 
 * cdef class VariantHeaderRecord(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 *     cdef bcf_hrec_t *ptr
 */
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderRecord {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
  bcf_hrec_t *ptr;
};


/* "pysam/libcbcf.pxd":51
 * 
 * 
 * cdef class VariantHeaderRecords(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderRecords {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
};


/* "pysam/libcbcf.pxd":55
 * 
 * 
 * cdef class VariantHeaderContigs(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderContigs {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
};


/* "pysam/libcbcf.pxd":59
 * 
 * 
 * cdef class VariantHeaderSamples(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderSamples {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
};


/* "pysam/libcbcf.pxd":63
 * 
 * 
 * cdef class VariantContig(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 *     cdef int id
 */
struct __pyx_obj_5pysam_7libcbcf_VariantContig {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
  int id;
};


/* "pysam/libcbcf.pxd":68
 * 
 * 
 * cdef class VariantMetadata(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 *     cdef int type
 */
struct __pyx_obj_5pysam_7libcbcf_VariantMetadata {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
  int type;
  int id;
};


/* "pysam/libcbcf.pxd":74
 * 
 * 
 * cdef class VariantHeaderMetadata(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 *     cdef int32_t type
 */
struct __pyx_obj_5pysam_7libcbcf_VariantHeaderMetadata {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
  int32_t type;
};


/* "pysam/libcbcf.pxd":79
 * 
 * 
 * cdef class VariantRecord(object):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 *     cdef bcf1_t *ptr
 */
struct __pyx_obj_5pysam_7libcbcf_VariantRecord {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
  bcf1_t *ptr;
};


/* "pysam/libcbcf.pxd":84
 * 
 * 
 * cdef class VariantRecordFilter(object):             # <<<<<<<<<<<<<<
 *     cdef VariantRecord record
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantRecordFilter {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantRecord *record;
};


/* "pysam/libcbcf.pxd":88
 * 
 * 
 * cdef class VariantRecordFormat(object):             # <<<<<<<<<<<<<<
 *     cdef VariantRecord record
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantRecordFormat {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantRecord *record;
};


/* "pysam/libcbcf.pxd":92
 * 
 * 
 * cdef class VariantRecordInfo(object):             # <<<<<<<<<<<<<<
 *     cdef VariantRecord record
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantRecordInfo {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantRecord *record;
};


/* "pysam/libcbcf.pxd":96
 * 
 * 
 * cdef class VariantRecordSamples(object):             # <<<<<<<<<<<<<<
 *     cdef VariantRecord record
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_VariantRecordSamples {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantRecord *record;
};


/* "pysam/libcbcf.pxd":100
 * 
 * 
 * cdef class VariantRecordSample(object):             # <<<<<<<<<<<<<<
 *     cdef VariantRecord record
 *     cdef readonly int32_t index
 */
struct __pyx_obj_5pysam_7libcbcf_VariantRecordSample {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantRecord *record;
  int32_t index;
};


/* "pysam/libcbcf.pxd":105
 * 
 * 
 * cdef class BaseIndex(object):             # <<<<<<<<<<<<<<
 *     cdef tuple refs
 *     cdef dict refmap
 */
struct __pyx_obj_5pysam_7libcbcf_BaseIndex {
  PyObject_HEAD
  PyObject *refs;
  PyObject *refmap;
};


/* "pysam/libcbcf.pxd":110
 * 
 * 
 * cdef class BCFIndex(BaseIndex):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader header
 *     cdef hts_idx_t *ptr
 */
struct __pyx_obj_5pysam_7libcbcf_BCFIndex {
  struct __pyx_obj_5pysam_7libcbcf_BaseIndex __pyx_base;
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
  hts_idx_t *ptr;
};


/* "pysam/libcbcf.pxd":115
 * 
 * 
 * cdef class TabixIndex(BaseIndex):             # <<<<<<<<<<<<<<
 *     cdef tbx_t *ptr
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_TabixIndex {
  struct __pyx_obj_5pysam_7libcbcf_BaseIndex __pyx_base;
  tbx_t *ptr;
};


/* "pysam/libcbcf.pxd":119
 * 
 * 
 * cdef class BaseIterator(object):             # <<<<<<<<<<<<<<
 *     cdef VariantFile bcf
 *     cdef hts_itr_t  *iter
 */
struct __pyx_obj_5pysam_7libcbcf_BaseIterator {
  PyObject_HEAD
  struct __pyx_obj_5pysam_7libcbcf_VariantFile *bcf;
  hts_itr_t *iter;
};


/* "pysam/libcbcf.pxd":124
 * 
 * 
 * cdef class BCFIterator(BaseIterator):             # <<<<<<<<<<<<<<
 *     cdef BCFIndex index
 * 
 */
struct __pyx_obj_5pysam_7libcbcf_BCFIterator {
  struct __pyx_obj_5pysam_7libcbcf_BaseIterator __pyx_base;
  struct __pyx_obj_5pysam_7libcbcf_BCFIndex *index;
};


/* "pysam/libcbcf.pxd":128
 * 
 * 
 * cdef class TabixIterator(BaseIterator):             # <<<<<<<<<<<<<<
 *     cdef TabixIndex index
 *     cdef kstring_t line_buffer
 */
struct __pyx_obj_5pysam_7libcbcf_TabixIterator {
  struct __pyx_obj_5pysam_7libcbcf_BaseIterator __pyx_base;
  struct __pyx_obj_5pysam_7libcbcf_TabixIndex *index;
  kstring_t line_buffer;
};


/* "pysam/libcbcf.pxd":133
 * 
 * 
 * cdef class VariantFile(HTSFile):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader  header
 *     cdef readonly BaseIndex      index
 */
struct __pyx_obj_5pysam_7libcbcf_VariantFile {
  struct __pyx_obj_5pysam_10libchtslib_HTSFile __pyx_base;
  struct __pyx_obj_5pysam_7libcbcf_VariantHeader *header;
  struct __pyx_obj_5pysam_7libcbcf_BaseIndex *index;
  int drop_samples;
  int is_reading;
  int header_written;
};


/* "variant.pxd":6
 * from pysam.libcbcf cimport VariantRecord, VariantFile
 * 
 * cdef class NullVariant:             # <<<<<<<<<<<<<<
 *     cdef readonly str chrom
 *     cdef readonly str ref, alt
 */
struct __pyx_obj_4aldy_9indelpost_7variant_NullVariant {
  PyObject_HEAD
  PyObject *chrom;
  PyObject *ref;
  PyObject *alt;
  int pos;
  struct __pyx_obj_5pysam_9libcfaidx_FastaFile *reference;
};


/* "variant.pxd":13
 * 
 * 
 * cdef class Variant:             # <<<<<<<<<<<<<<
 *     cdef public str chrom, ref, alt,
 *     cdef public int pos
 */
struct __pyx_obj_4aldy_9indelpost_7variant_Variant {
  PyObject_HEAD
  struct __pyx_vtabstruct_4aldy_9indelpost_7variant_Variant *__pyx_vtab;
  PyObject *chrom;
  PyObject *ref;
  PyObject *alt;
  int pos;
  struct __pyx_obj_5pysam_9libcfaidx_FastaFile *reference;
  PyObject *_chrom;
  struct __pyx_obj_5pysam_7libcbcf_VariantFile *vcf;
  struct __pyx_obj_5pysam_7libcbcf_VariantRecord *rec;
  struct __pyx_obj_5pysam_7libcbcf_VariantRecord *hit;
};


/* "aldy/indelpost/localn.pyx":138
 * 
 * 
 * def is_worth_realn(read, target_indel, qual_lim=23):             # <<<<<<<<<<<<<<
 * 
 *     if read["covering_subread"]:
 */
struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn {
  PyObject_HEAD
  PyObject *__pyx_v_mismatches;
};


/* "aldy/indelpost/localn.pyx":199
 *     if mismatches:
 *         if is_lefty:
 *             lt_most_pos = min(var[0] for var in mismatches)             # <<<<<<<<<<<<<<
 *             with_end_mismatches = abs(lt_most_pos - read["aln_start"]) < 4
 *         else:
 */
struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *__pyx_outer_scope;
  PyObject *__pyx_v_var;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "aldy/indelpost/localn.pyx":202
 *             with_end_mismatches = abs(lt_most_pos - read["aln_start"]) < 4
 *         else:
 *             rt_most_pos = max(var[0] for var in mismatches)             # <<<<<<<<<<<<<<
 *             with_end_mismatches = abs(rt_most_pos - read["aln_end"]) < 4
 * 
 */
struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr {
  PyObject_HEAD
  struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *__pyx_outer_scope;
  PyObject *__pyx_v_var;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};



/* "pysam/libchtslib.pxd":2680
 * 
 * 
 * cdef class HTSFile(object):             # <<<<<<<<<<<<<<
 *     cdef          htsFile *htsfile       # pointer to htsFile structure
 *     cdef          int64_t start_offset   # BGZF offset of first record
 */

struct __pyx_vtabstruct_5pysam_10libchtslib_HTSFile {
  htsFile *(*_open_htsfile)(struct __pyx_obj_5pysam_10libchtslib_HTSFile *);
};
static struct __pyx_vtabstruct_5pysam_10libchtslib_HTSFile *__pyx_vtabptr_5pysam_10libchtslib_HTSFile;


/* "pysam/libcfaidx.pxd":37
 *                     int * dret)
 * 
 * cdef class FastaFile:             # <<<<<<<<<<<<<<
 *     cdef bint is_remote
 *     cdef object _filename, _references, _lengths, reference2length
 */

struct __pyx_vtabstruct_5pysam_9libcfaidx_FastaFile {
  char *(*_fetch)(struct __pyx_obj_5pysam_9libcfaidx_FastaFile *, char *, int, int, int *);
};
static struct __pyx_vtabstruct_5pysam_9libcfaidx_FastaFile *__pyx_vtabptr_5pysam_9libcfaidx_FastaFile;


/* "pysam/libcfaidx.pxd":45
 * 
 * 
 * cdef class FastqProxy:             # <<<<<<<<<<<<<<
 *     cdef kseq_t * _delegate
 *     cdef cython.str to_string(self)
 */

struct __pyx_vtabstruct_5pysam_9libcfaidx_FastqProxy {
  PyObject *(*to_string)(struct __pyx_obj_5pysam_9libcfaidx_FastqProxy *);
  PyObject *(*tostring)(struct __pyx_obj_5pysam_9libcfaidx_FastqProxy *);
  arrayobject *(*get_quality_array)(struct __pyx_obj_5pysam_9libcfaidx_FastqProxy *, int __pyx_skip_dispatch, struct __pyx_opt_args_5pysam_9libcfaidx_10FastqProxy_get_quality_array *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5pysam_9libcfaidx_FastqProxy *__pyx_vtabptr_5pysam_9libcfaidx_FastqProxy;


/* "pysam/libcfaidx.pxd":52
 * 
 * 
 * cdef class FastxRecord:             # <<<<<<<<<<<<<<
 *     """
 *     Python container for pysam.libcfaidx.FastqProxy with persistence.
 */

struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxRecord {
  PyObject *(*to_string)(struct __pyx_obj_5pysam_9libcfaidx_FastxRecord *);
  PyObject *(*tostring)(struct __pyx_obj_5pysam_9libcfaidx_FastxRecord *);
  arrayobject *(*get_quality_array)(struct __pyx_obj_5pysam_9libcfaidx_FastxRecord *, int __pyx_skip_dispatch, struct __pyx_opt_args_5pysam_9libcfaidx_11FastxRecord_get_quality_array *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxRecord *__pyx_vtabptr_5pysam_9libcfaidx_FastxRecord;


/* "pysam/libcfaidx.pxd":61
 *     cpdef array.array get_quality_array(self, int offset=*)
 * 
 * cdef class FastxFile:             # <<<<<<<<<<<<<<
 *     cdef object _filename
 *     cdef BGZF * fastqfile
 */

struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxFile {
  kseq_t *(*getCurrent)(struct __pyx_obj_5pysam_9libcfaidx_FastxFile *);
  int (*cnext)(struct __pyx_obj_5pysam_9libcfaidx_FastxFile *);
};
static struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxFile *__pyx_vtabptr_5pysam_9libcfaidx_FastxFile;


/* "pysam/libcfaidx.pxd":73
 * 
 * # Compatibility Layer for pysam 0.8.1
 * cdef class FastqFile(FastxFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_5pysam_9libcfaidx_FastqFile {
  struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxFile __pyx_base;
};
static struct __pyx_vtabstruct_5pysam_9libcfaidx_FastqFile *__pyx_vtabptr_5pysam_9libcfaidx_FastqFile;


/* "pysam/libcfaidx.pxd":78
 * 
 * # Compatibility Layer for pysam < 0.8
 * cdef class Fastafile(FastaFile):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_5pysam_9libcfaidx_Fastafile {
  struct __pyx_vtabstruct_5pysam_9libcfaidx_FastaFile __pyx_base;
};
static struct __pyx_vtabstruct_5pysam_9libcfaidx_Fastafile *__pyx_vtabptr_5pysam_9libcfaidx_Fastafile;


/* "pysam/libcbcf.pxd":38
 * 
 * 
 * cdef class VariantHeader(object):             # <<<<<<<<<<<<<<
 *     cdef bcf_hdr_t *ptr
 * 
 */

struct __pyx_vtabstruct_5pysam_7libcbcf_VariantHeader {
  PyObject *(*_add_sample)(struct __pyx_obj_5pysam_7libcbcf_VariantHeader *, PyObject *);
  PyObject *(*_hdr_sync)(struct __pyx_obj_5pysam_7libcbcf_VariantHeader *);
  PyObject *(*_subset_samples)(struct __pyx_obj_5pysam_7libcbcf_VariantHeader *, PyObject *);
};
static struct __pyx_vtabstruct_5pysam_7libcbcf_VariantHeader *__pyx_vtabptr_5pysam_7libcbcf_VariantHeader;


/* "pysam/libcbcf.pxd":133
 * 
 * 
 * cdef class VariantFile(HTSFile):             # <<<<<<<<<<<<<<
 *     cdef readonly VariantHeader  header
 *     cdef readonly BaseIndex      index
 */

struct __pyx_vtabstruct_5pysam_7libcbcf_VariantFile {
  struct __pyx_vtabstruct_5pysam_10libchtslib_HTSFile __pyx_base;
  int (*write)(struct __pyx_obj_5pysam_7libcbcf_VariantFile *, struct __pyx_obj_5pysam_7libcbcf_VariantRecord *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5pysam_7libcbcf_VariantFile *__pyx_vtabptr_5pysam_7libcbcf_VariantFile;


/* "variant.pxd":13
 * 
 * 
 * cdef class Variant:             # <<<<<<<<<<<<<<
 *     cdef public str chrom, ref, alt,
 *     cdef public int pos
 */

struct __pyx_vtabstruct_4aldy_9indelpost_7variant_Variant {
  PyObject *(*__pyx___validate)(struct __pyx_obj_4aldy_9indelpost_7variant_Variant *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4aldy_9indelpost_7variant_Variant *__pyx_vtabptr_4aldy_9indelpost_7variant_Variant;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* py_abs.proto */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *num);
#define __Pyx_PyNumber_Absolute(x)\
    ((likely(PyLong_CheckExact(x))) ?\
         (likely(Py_SIZE(x) >= 0) ? (Py_INCREF(x), (x)) : __Pyx_PyLong_AbsNeg(x)) :\
         PyNumber_Absolute(x))
#else
#define __Pyx_PyNumber_Absolute(x)  PyNumber_Absolute(x)
#endif

/* PyUnicode_Substring.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Substring(
            PyObject* text, Py_ssize_t start, Py_ssize_t stop);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* ArrayAPI.proto */
#ifndef _ARRAYARRAY_H
#define _ARRAYARRAY_H
typedef struct arraydescr {
    int typecode;
    int itemsize;
    PyObject * (*getitem)(struct arrayobject *, Py_ssize_t);
    int (*setitem)(struct arrayobject *, Py_ssize_t, PyObject *);
#if PY_MAJOR_VERSION >= 3
    char *formats;
#endif
} arraydescr;
struct arrayobject {
    PyObject_HEAD
    Py_ssize_t ob_size;
    union {
        char *ob_item;
        float *as_floats;
        double *as_doubles;
        int *as_ints;
        unsigned int *as_uints;
        unsigned char *as_uchars;
        signed char *as_schars;
        char *as_chars;
        unsigned long *as_ulongs;
        long *as_longs;
#if PY_MAJOR_VERSION >= 3
        unsigned long long *as_ulonglongs;
        long long *as_longlongs;
#endif
        short *as_shorts;
        unsigned short *as_ushorts;
        Py_UNICODE *as_pyunicodes;
        void *as_voidptr;
    } data;
    Py_ssize_t allocated;
    struct arraydescr *ob_descr;
    PyObject *weakreflist;
#if PY_MAJOR_VERSION >= 3
        int ob_exports;
#endif
};
#ifndef NO_NEWARRAY_INLINE
static CYTHON_INLINE PyObject * newarrayobject(PyTypeObject *type, Py_ssize_t size,
    struct arraydescr *descr) {
    arrayobject *op;
    size_t nbytes;
    if (size < 0) {
        PyErr_BadInternalCall();
        return NULL;
    }
    nbytes = size * descr->itemsize;
    if (nbytes / descr->itemsize != (size_t)size) {
        return PyErr_NoMemory();
    }
    op = (arrayobject *) type->tp_alloc(type, 0);
    if (op == NULL) {
        return NULL;
    }
    op->ob_descr = descr;
    op->allocated = size;
    op->weakreflist = NULL;
    __Pyx_SET_SIZE(op, size);
    if (size <= 0) {
        op->data.ob_item = NULL;
    }
    else {
        op->data.ob_item = PyMem_NEW(char, nbytes);
        if (op->data.ob_item == NULL) {
            Py_DECREF(op);
            return PyErr_NoMemory();
        }
    }
    return (PyObject *) op;
}
#else
PyObject* newarrayobject(PyTypeObject *type, Py_ssize_t size,
    struct arraydescr *descr);
#endif
static CYTHON_INLINE int resize(arrayobject *self, Py_ssize_t n) {
    void *items = (void*) self->data.ob_item;
    PyMem_Resize(items, char, (size_t)(n * self->ob_descr->itemsize));
    if (items == NULL) {
        PyErr_NoMemory();
        return -1;
    }
    self->data.ob_item = (char*) items;
    __Pyx_SET_SIZE(self, n);
    self->allocated = n;
    return 0;
}
static CYTHON_INLINE int resize_smart(arrayobject *self, Py_ssize_t n) {
    void *items = (void*) self->data.ob_item;
    Py_ssize_t newsize;
    if (n < self->allocated && n*4 > self->allocated) {
        __Pyx_SET_SIZE(self, n);
        return 0;
    }
    newsize = n + (n / 2) + 1;
    if (newsize <= n) {
        PyErr_NoMemory();
        return -1;
    }
    PyMem_Resize(items, char, (size_t)(newsize * self->ob_descr->itemsize));
    if (items == NULL) {
        PyErr_NoMemory();
        return -1;
    }
    self->data.ob_item = (char*) items;
    __Pyx_SET_SIZE(self, n);
    self->allocated = newsize;
    return 0;
}
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libc.stdint' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cython' */

/* Module declarations from 'cpython.version' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'array' */

/* Module declarations from 'cpython.array' */
static PyTypeObject *__pyx_ptype_7cpython_5array_array = 0;
static CYTHON_INLINE int __pyx_f_7cpython_5array_extend_buffer(arrayobject *, char *, Py_ssize_t); /*proto*/

/* Module declarations from 'posix.types' */

/* Module declarations from 'pysam.libchtslib' */
static PyTypeObject *__pyx_ptype_5pysam_10libchtslib_HTSFile = 0;

/* Module declarations from 'pysam.libcfaidx' */
static PyTypeObject *__pyx_ptype_5pysam_9libcfaidx_FastaFile = 0;
static PyTypeObject *__pyx_ptype_5pysam_9libcfaidx_FastqProxy = 0;
static PyTypeObject *__pyx_ptype_5pysam_9libcfaidx_FastxRecord = 0;
static PyTypeObject *__pyx_ptype_5pysam_9libcfaidx_FastxFile = 0;
static PyTypeObject *__pyx_ptype_5pysam_9libcfaidx_FastqFile = 0;
static PyTypeObject *__pyx_ptype_5pysam_9libcfaidx_Fastafile = 0;

/* Module declarations from 'pysam.libcbcf' */
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantHeader = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantHeaderRecord = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantHeaderRecords = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantHeaderContigs = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantHeaderSamples = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantContig = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantMetadata = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantHeaderMetadata = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantRecord = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantRecordFilter = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantRecordFormat = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantRecordInfo = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantRecordSamples = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantRecordSample = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_BaseIndex = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_BCFIndex = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_TabixIndex = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_BaseIterator = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_BCFIterator = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_TabixIterator = 0;
static PyTypeObject *__pyx_ptype_5pysam_7libcbcf_VariantFile = 0;

/* Module declarations from 'aldy.indelpost.variant' */
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_7variant_NullVariant = 0;
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_7variant_Variant = 0;

/* Module declarations from 'aldy.indelpost.utilities' */
static PyObject *(*__pyx_f_4aldy_9indelpost_9utilities_split)(PyObject *, PyObject *, int, int, int, int); /*proto*/
static int (*__pyx_f_4aldy_9indelpost_9utilities_count_lowqual_non_ref_bases)(PyObject *, PyObject *, arrayobject *, PyObject *, int); /*proto*/

/* Module declarations from 'aldy.indelpost.localn' */
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn = 0;
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_ptype_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr = 0;
static int __pyx_f_4aldy_9indelpost_6localn_is_covering_target(PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, int, int, int, int, int, int); /*proto*/
#define __Pyx_MODULE_NAME "aldy.indelpost.localn"
extern int __pyx_module_is_main_aldy__indelpost__localn;
int __pyx_module_is_main_aldy__indelpost__localn = 0;

/* Implementation of 'aldy.indelpost.localn' */
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_min;
static PyObject *__pyx_builtin_max;
static PyObject *__pyx_builtin_MemoryError;
static const char __pyx_k_D[] = "D";
static const char __pyx_k_I[] = "I";
static const char __pyx_k_S[] = "S";
static const char __pyx_k_a[] = "a";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_q[] = "q";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_v[] = "v";
static const char __pyx_k__3[] = "";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_re[] = "re";
static const char __pyx_k_SSW[] = "SSW";
static const char __pyx_k_alt[] = "alt";
static const char __pyx_k_cnt[] = "cnt";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_pos[] = "pos";
static const char __pyx_k_ref[] = "ref";
static const char __pyx_k_seq[] = "seq";
static const char __pyx_k_snv[] = "snv";
static const char __pyx_k_var[] = "var";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mapq[] = "mapq";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_read[] = "read";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_snvs[] = "snvs";
static const char __pyx_k_span[] = "span";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_CIGAR[] = "CIGAR";
static const char __pyx_k_align[] = "align";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_event[] = "event";
static const char __pyx_k_indel[] = "indel";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_quals[] = "quals";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_sswpy[] = "sswpy";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_token[] = "token";
static const char __pyx_k_upper[] = "upper";
static const char __pyx_k_contig[] = "contig";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_indels[] = "indels";
static const char __pyx_k_is_ins[] = "is_ins";
static const char __pyx_k_lt_len[] = "lt_len";
static const char __pyx_k_lt_pos[] = "lt_pos";
static const char __pyx_k_lt_ref[] = "lt_ref";
static const char __pyx_k_lt_seq[] = "lt_seq";
static const char __pyx_k_pileup[] = "pileup";
static const char __pyx_k_rt_len[] = "rt_len";
static const char __pyx_k_rt_pos[] = "rt_pos";
static const char __pyx_k_rt_ref[] = "rt_ref";
static const char __pyx_k_rt_seq[] = "rt_seq";
static const char __pyx_k_aligner[] = "aligner";
static const char __pyx_k_aln_end[] = "aln_end";
static const char __pyx_k_compile[] = "compile";
static const char __pyx_k_del_pos[] = "del_pos";
static const char __pyx_k_del_seq[] = "del_seq";
static const char __pyx_k_findall[] = "findall";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_is_done[] = "is_done";
static const char __pyx_k_is_left[] = "is_left";
static const char __pyx_k_lt_diff[] = "lt_diff";
static const char __pyx_k_lt_qual[] = "lt_qual";
static const char __pyx_k_mid_seq[] = "mid_seq";
static const char __pyx_k_mut_aln[] = "mut_aln";
static const char __pyx_k_mut_ref[] = "mut_ref";
static const char __pyx_k_ref_aln[] = "ref_aln";
static const char __pyx_k_ref_end[] = "ref_end";
static const char __pyx_k_ref_idx[] = "ref_idx";
static const char __pyx_k_ref_ref[] = "ref_ref";
static const char __pyx_k_ref_seq[] = "ref_seq";
static const char __pyx_k_rt_diff[] = "rt_diff";
static const char __pyx_k_rt_qual[] = "rt_qual";
static const char __pyx_k_setRead[] = "setRead";
static const char __pyx_k_subread[] = "subread";
static const char __pyx_k_cigarlst[] = "cigarlst";
static const char __pyx_k_end_trim[] = "end_trim";
static const char __pyx_k_gap_open[] = "gap_open";
static const char __pyx_k_is_lefty[] = "is_lefty";
static const char __pyx_k_lt_flank[] = "lt_flank";
static const char __pyx_k_mapq_lim[] = "mapq_lim";
static const char __pyx_k_qual_lim[] = "qual_lim";
static const char __pyx_k_read_end[] = "read_end";
static const char __pyx_k_read_idx[] = "read_idx";
static const char __pyx_k_read_seq[] = "read_seq";
static const char __pyx_k_rt_flank[] = "rt_flank";
static const char __pyx_k_unit_len[] = "unit_len";
static const char __pyx_k_aln_start[] = "aln_start";
static const char __pyx_k_basequals[] = "basequals";
static const char __pyx_k_consensus[] = "consensus";
static const char __pyx_k_end_cigar[] = "end_cigar";
static const char __pyx_k_end_point[] = "end_point";
static const char __pyx_k_event_len[] = "event_len";
static const char __pyx_k_indel_len[] = "indel_len";
static const char __pyx_k_indel_seq[] = "indel_seq";
static const char __pyx_k_is_target[] = "is_target";
static const char __pyx_k_read_name[] = "read_name";
static const char __pyx_k_read_qual[] = "read_qual";
static const char __pyx_k_ref_start[] = "ref_start";
static const char __pyx_k_utilities[] = "utilities";
static const char __pyx_k_cigar_list[] = "cigar_list";
static const char __pyx_k_cigar_ptrn[] = "cigar_ptrn";
static const char __pyx_k_end_offset[] = "end_offset";
static const char __pyx_k_indel_type[] = "indel_type";
static const char __pyx_k_is_covered[] = "is_covered";
static const char __pyx_k_lt_clipped[] = "lt_clipped";
static const char __pyx_k_lt_end_ref[] = "lt_end_ref";
static const char __pyx_k_mapped_ref[] = "mapped_ref";
static const char __pyx_k_mapped_seq[] = "mapped_seq";
static const char __pyx_k_mismatches[] = "mismatches";
static const char __pyx_k_mut_ref_lt[] = "mut_ref_lt";
static const char __pyx_k_mut_ref_rt[] = "mut_ref_rt";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_read_start[] = "read_start";
static const char __pyx_k_rt_clipped[] = "rt_clipped";
static const char __pyx_k_rt_end_ref[] = "rt_end_ref";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_cigarstring[] = "cigarstring";
static const char __pyx_k_get_end_pos[] = "get_end_pos";
static const char __pyx_k_lt_end_read[] = "lt_end_read";
static const char __pyx_k_lt_most_pos[] = "lt_most_pos";
static const char __pyx_k_mapped_qual[] = "mapped_qual";
static const char __pyx_k_match_score[] = "match_score";
static const char __pyx_k_mut_aligner[] = "mut_aligner";
static const char __pyx_k_mut_ref_mid[] = "mut_ref_mid";
static const char __pyx_k_ref_aligner[] = "ref_aligner";
static const char __pyx_k_repeat_unit[] = "repeat_unit";
static const char __pyx_k_report_snvs[] = "report_snvs";
static const char __pyx_k_rt_end_read[] = "rt_end_read";
static const char __pyx_k_rt_most_pos[] = "rt_most_pos";
static const char __pyx_k_start_cigar[] = "start_cigar";
static const char __pyx_k_0_9_MIDNSHPX[] = "[0-9]+[MIDNSHPX=]";
static const char __pyx_k_cigar_string[] = "cigar_string";
static const char __pyx_k_covering_end[] = "covering_end";
static const char __pyx_k_make_aligner[] = "make_aligner";
static const char __pyx_k_setReference[] = "setReference";
static const char __pyx_k_start_offset[] = "start_offset";
static const char __pyx_k_target_indel[] = "target_indel";
static const char __pyx_k_undetermined[] = "undetermined";
static const char __pyx_k_variant_type[] = "variant_type";
static const char __pyx_k_count_repeats[] = "count_repeats";
static const char __pyx_k_gap_extension[] = "gap_extension";
static const char __pyx_k_is_compatible[] = "is_compatible";
static const char __pyx_k_mut_aln_cigar[] = "mut_aln_cigar";
static const char __pyx_k_optimal_score[] = "optimal_score";
static const char __pyx_k_reference_end[] = "reference_end";
static const char __pyx_k_shiftable_pos[] = "shiftable_pos";
static const char __pyx_k_basequalthresh[] = "basequalthresh";
static const char __pyx_k_covering_start[] = "covering_start";
static const char __pyx_k_findall_indels[] = "findall_indels";
static const char __pyx_k_genome_aln_pos[] = "genome_aln_pos";
static const char __pyx_k_get_contig_seq[] = "get_contig_seq";
static const char __pyx_k_is_worth_realn[] = "is_worth_realn";
static const char __pyx_k_mut_cigar_list[] = "mut_cigar_list";
static const char __pyx_k_target_covered[] = "target_covered";
static const char __pyx_k_mut_aln_ref_end[] = "mut_aln_ref_end";
static const char __pyx_k_reference_start[] = "reference_start";
static const char __pyx_k_allow_mismatches[] = "allow_mismatches";
static const char __pyx_k_covering_subread[] = "covering_subread";
static const char __pyx_k_dist_to_left_end[] = "dist_to_left_end";
static const char __pyx_k_gap_open_penalty[] = "gap_open_penalty";
static const char __pyx_k_is_reference_seq[] = "is_reference_seq";
static const char __pyx_k_is_target_by_ssw[] = "is_target_by_ssw";
static const char __pyx_k_lt_consensus_seq[] = "lt_consensus_seq";
static const char __pyx_k_mapped_subpreads[] = "mapped_subpreads";
static const char __pyx_k_mismatch_penalty[] = "mismatch_penalty";
static const char __pyx_k_mut_aln_read_end[] = "mut_aln_read_end";
static const char __pyx_k_rt_consensus_seq[] = "rt_consensus_seq";
static const char __pyx_k_dist_to_right_end[] = "dist_to_right_end";
static const char __pyx_k_get_reference_seq[] = "get_reference_seq";
static const char __pyx_k_mut_aln_ref_start[] = "mut_aln_ref_start";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_expected_n_repeats[] = "expected_n_repeats";
static const char __pyx_k_findall_mismatches[] = "findall_mismatches";
static const char __pyx_k_mut_aln_read_start[] = "mut_aln_read_start";
static const char __pyx_k_get_mapped_subreads[] = "get_mapped_subreads";
static const char __pyx_k_with_end_mismatches[] = "with_end_mismatches";
static const char __pyx_k_generate_equivalents[] = "generate_equivalents";
static const char __pyx_k_make_insertion_first[] = "make_insertion_first";
static const char __pyx_k_aldy_indelpost_localn[] = "aldy.indelpost.localn";
static const char __pyx_k_gap_extension_penalty[] = "gap_extension_penalty";
static const char __pyx_k_is_compatible_repeats[] = "is_compatible_repeats";
static const char __pyx_k_parse_read_by_mut_aln[] = "parse_read_by_mut_aln";
static const char __pyx_k_mapped_base_cnt_thresh[] = "mapped_base_cnt_thresh";
static const char __pyx_k_to_minimal_repeat_unit[] = "to_minimal_repeat_unit";
static const char __pyx_k_aldy_indelpost_localn_pyx[] = "aldy/indelpost/localn.pyx";
static const char __pyx_k_find_by_smith_waterman_realn[] = "find_by_smith_waterman_realn";
static const char __pyx_k_is_worth_realn_locals_genexpr[] = "is_worth_realn.<locals>.genexpr";
static PyObject *__pyx_kp_u_0_9_MIDNSHPX;
static PyObject *__pyx_n_s_CIGAR;
static PyObject *__pyx_n_u_D;
static PyObject *__pyx_n_u_I;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_u_S;
static PyObject *__pyx_n_s_SSW;
static PyObject *__pyx_kp_u__3;
static PyObject *__pyx_n_s_a;
static PyObject *__pyx_n_s_aldy_indelpost_localn;
static PyObject *__pyx_kp_s_aldy_indelpost_localn_pyx;
static PyObject *__pyx_n_s_align;
static PyObject *__pyx_n_s_aligner;
static PyObject *__pyx_n_s_allow_mismatches;
static PyObject *__pyx_n_s_aln_end;
static PyObject *__pyx_n_u_aln_end;
static PyObject *__pyx_n_s_aln_start;
static PyObject *__pyx_n_u_aln_start;
static PyObject *__pyx_n_u_alt;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_basequals;
static PyObject *__pyx_n_s_basequalthresh;
static PyObject *__pyx_n_u_cigar_list;
static PyObject *__pyx_n_s_cigar_ptrn;
static PyObject *__pyx_n_u_cigar_string;
static PyObject *__pyx_n_s_cigarlst;
static PyObject *__pyx_n_s_cigarstring;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_cnt;
static PyObject *__pyx_n_s_compile;
static PyObject *__pyx_n_s_consensus;
static PyObject *__pyx_n_s_contig;
static PyObject *__pyx_n_s_count_repeats;
static PyObject *__pyx_n_s_covering_end;
static PyObject *__pyx_n_s_covering_start;
static PyObject *__pyx_n_u_covering_subread;
static PyObject *__pyx_n_s_del_pos;
static PyObject *__pyx_n_u_del_pos;
static PyObject *__pyx_n_u_del_seq;
static PyObject *__pyx_n_s_dist_to_left_end;
static PyObject *__pyx_n_s_dist_to_right_end;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_end_cigar;
static PyObject *__pyx_n_u_end_offset;
static PyObject *__pyx_n_s_end_point;
static PyObject *__pyx_n_s_end_trim;
static PyObject *__pyx_n_s_event;
static PyObject *__pyx_n_s_event_len;
static PyObject *__pyx_n_s_expected_n_repeats;
static PyObject *__pyx_n_s_find_by_smith_waterman_realn;
static PyObject *__pyx_n_s_findall;
static PyObject *__pyx_n_s_findall_indels;
static PyObject *__pyx_n_s_findall_mismatches;
static PyObject *__pyx_n_s_gap_extension;
static PyObject *__pyx_n_s_gap_extension_penalty;
static PyObject *__pyx_n_s_gap_open;
static PyObject *__pyx_n_s_gap_open_penalty;
static PyObject *__pyx_n_s_generate_equivalents;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_genome_aln_pos;
static PyObject *__pyx_n_s_get_contig_seq;
static PyObject *__pyx_n_s_get_end_pos;
static PyObject *__pyx_n_s_get_mapped_subreads;
static PyObject *__pyx_n_s_get_reference_seq;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_indel;
static PyObject *__pyx_n_s_indel_len;
static PyObject *__pyx_n_s_indel_seq;
static PyObject *__pyx_n_u_indel_seq;
static PyObject *__pyx_n_s_indel_type;
static PyObject *__pyx_n_u_indel_type;
static PyObject *__pyx_n_s_indels;
static PyObject *__pyx_n_s_is_compatible;
static PyObject *__pyx_n_s_is_compatible_repeats;
static PyObject *__pyx_n_s_is_covered;
static PyObject *__pyx_n_s_is_done;
static PyObject *__pyx_n_s_is_ins;
static PyObject *__pyx_n_s_is_left;
static PyObject *__pyx_n_s_is_lefty;
static PyObject *__pyx_n_u_is_reference_seq;
static PyObject *__pyx_n_u_is_target;
static PyObject *__pyx_n_s_is_target_by_ssw;
static PyObject *__pyx_n_s_is_worth_realn;
static PyObject *__pyx_n_s_is_worth_realn_locals_genexpr;
static PyObject *__pyx_n_s_lt_clipped;
static PyObject *__pyx_n_u_lt_clipped;
static PyObject *__pyx_n_s_lt_consensus_seq;
static PyObject *__pyx_n_s_lt_diff;
static PyObject *__pyx_n_s_lt_end_read;
static PyObject *__pyx_n_s_lt_end_ref;
static PyObject *__pyx_n_s_lt_flank;
static PyObject *__pyx_n_u_lt_flank;
static PyObject *__pyx_n_s_lt_len;
static PyObject *__pyx_n_s_lt_most_pos;
static PyObject *__pyx_n_s_lt_pos;
static PyObject *__pyx_n_s_lt_qual;
static PyObject *__pyx_n_u_lt_qual;
static PyObject *__pyx_n_s_lt_ref;
static PyObject *__pyx_n_u_lt_ref;
static PyObject *__pyx_n_s_lt_seq;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_make_aligner;
static PyObject *__pyx_n_s_make_insertion_first;
static PyObject *__pyx_n_s_mapped_base_cnt_thresh;
static PyObject *__pyx_n_s_mapped_qual;
static PyObject *__pyx_n_s_mapped_ref;
static PyObject *__pyx_n_s_mapped_seq;
static PyObject *__pyx_n_s_mapped_subpreads;
static PyObject *__pyx_n_u_mapq;
static PyObject *__pyx_n_s_mapq_lim;
static PyObject *__pyx_n_s_match_score;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_mid_seq;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_mismatch_penalty;
static PyObject *__pyx_n_s_mismatches;
static PyObject *__pyx_n_u_mismatches;
static PyObject *__pyx_n_s_mut_aligner;
static PyObject *__pyx_n_s_mut_aln;
static PyObject *__pyx_n_s_mut_aln_cigar;
static PyObject *__pyx_n_s_mut_aln_read_end;
static PyObject *__pyx_n_s_mut_aln_read_start;
static PyObject *__pyx_n_s_mut_aln_ref_end;
static PyObject *__pyx_n_s_mut_aln_ref_start;
static PyObject *__pyx_n_s_mut_cigar_list;
static PyObject *__pyx_n_s_mut_ref;
static PyObject *__pyx_n_s_mut_ref_lt;
static PyObject *__pyx_n_s_mut_ref_mid;
static PyObject *__pyx_n_s_mut_ref_rt;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_n_s_optimal_score;
static PyObject *__pyx_n_s_parse_read_by_mut_aln;
static PyObject *__pyx_n_s_pileup;
static PyObject *__pyx_n_s_pos;
static PyObject *__pyx_n_u_pos;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_q;
static PyObject *__pyx_n_s_qual_lim;
static PyObject *__pyx_n_s_quals;
static PyObject *__pyx_n_s_r;
static PyObject *__pyx_n_s_re;
static PyObject *__pyx_n_s_read;
static PyObject *__pyx_n_s_read_end;
static PyObject *__pyx_n_s_read_idx;
static PyObject *__pyx_n_u_read_idx;
static PyObject *__pyx_n_u_read_name;
static PyObject *__pyx_n_s_read_qual;
static PyObject *__pyx_n_u_read_qual;
static PyObject *__pyx_n_s_read_seq;
static PyObject *__pyx_n_u_read_seq;
static PyObject *__pyx_n_s_read_start;
static PyObject *__pyx_n_u_read_start;
static PyObject *__pyx_n_s_ref;
static PyObject *__pyx_n_u_ref;
static PyObject *__pyx_n_s_ref_aligner;
static PyObject *__pyx_n_s_ref_aln;
static PyObject *__pyx_n_s_ref_end;
static PyObject *__pyx_n_s_ref_idx;
static PyObject *__pyx_n_u_ref_idx;
static PyObject *__pyx_n_s_ref_ref;
static PyObject *__pyx_n_s_ref_seq;
static PyObject *__pyx_n_u_ref_seq;
static PyObject *__pyx_n_s_ref_start;
static PyObject *__pyx_n_s_reference_end;
static PyObject *__pyx_n_s_reference_start;
static PyObject *__pyx_n_s_repeat_unit;
static PyObject *__pyx_n_s_report_snvs;
static PyObject *__pyx_n_s_rt_clipped;
static PyObject *__pyx_n_u_rt_clipped;
static PyObject *__pyx_n_s_rt_consensus_seq;
static PyObject *__pyx_n_s_rt_diff;
static PyObject *__pyx_n_s_rt_end_read;
static PyObject *__pyx_n_s_rt_end_ref;
static PyObject *__pyx_n_s_rt_flank;
static PyObject *__pyx_n_u_rt_flank;
static PyObject *__pyx_n_s_rt_len;
static PyObject *__pyx_n_s_rt_most_pos;
static PyObject *__pyx_n_s_rt_pos;
static PyObject *__pyx_n_s_rt_qual;
static PyObject *__pyx_n_u_rt_qual;
static PyObject *__pyx_n_s_rt_ref;
static PyObject *__pyx_n_u_rt_ref;
static PyObject *__pyx_n_s_rt_seq;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_seq;
static PyObject *__pyx_n_s_setRead;
static PyObject *__pyx_n_s_setReference;
static PyObject *__pyx_n_s_shiftable_pos;
static PyObject *__pyx_n_s_snv;
static PyObject *__pyx_n_s_snvs;
static PyObject *__pyx_n_s_span;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_sswpy;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_start_cigar;
static PyObject *__pyx_n_u_start_offset;
static PyObject *__pyx_n_s_subread;
static PyObject *__pyx_n_s_target_covered;
static PyObject *__pyx_n_s_target_indel;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_to_minimal_repeat_unit;
static PyObject *__pyx_n_s_token;
static PyObject *__pyx_n_u_undetermined;
static PyObject *__pyx_n_s_unit_len;
static PyObject *__pyx_n_s_upper;
static PyObject *__pyx_n_s_utilities;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_var;
static PyObject *__pyx_n_s_variant_type;
static PyObject *__pyx_n_s_with_end_mismatches;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_pf_4aldy_9indelpost_6localn_find_by_smith_waterman_realn(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_target_indel, PyObject *__pyx_v_contig, PyObject *__pyx_v_pileup, PyObject *__pyx_v_match_score, PyObject *__pyx_v_mismatch_penalty, PyObject *__pyx_v_gap_open_penalty, PyObject *__pyx_v_gap_extension_penalty, PyObject *__pyx_v_basequalthresh, PyObject *__pyx_v_mapq_lim); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6localn_2findall_mismatches(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_end_trim); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6localn_14is_worth_realn_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6localn_14is_worth_realn_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6localn_4is_worth_realn(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_target_indel, PyObject *__pyx_v_qual_lim); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6localn_6is_target_by_ssw(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_target_indel, CYTHON_UNUSED PyObject *__pyx_v_contig, PyObject *__pyx_v_mut_ref_lt, PyObject *__pyx_v_mut_ref_mid, PyObject *__pyx_v_mut_ref_rt, PyObject *__pyx_v_mut_aligner, PyObject *__pyx_v_ref_aligner, CYTHON_UNUSED PyObject *__pyx_v_match_score, CYTHON_UNUSED PyObject *__pyx_v_mismatch_penalty, PyObject *__pyx_v_gap_open_penalty, PyObject *__pyx_v_gap_extension_penalty, CYTHON_UNUSED PyObject *__pyx_v_indel_type, CYTHON_UNUSED PyObject *__pyx_v_basequalthresh, PyObject *__pyx_v_mapq_lim, CYTHON_UNUSED PyObject *__pyx_v_mapped_base_cnt_thresh, CYTHON_UNUSED PyObject *__pyx_v_allow_mismatches); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6localn_8is_compatible_repeats(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_seq, PyObject *__pyx_v_repeat_unit, PyObject *__pyx_v_expected_n_repeats, PyObject *__pyx_v_is_left); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6localn_10make_aligner(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ref_seq, PyObject *__pyx_v_match_score, PyObject *__pyx_v_mismatch_penalty); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6localn_12align(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_aligner, PyObject *__pyx_v_read_seq, PyObject *__pyx_v_gap_open_penalty, PyObject *__pyx_v_gap_extension_penalty); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6localn_14parse_read_by_mut_aln(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_mut_aln, PyObject *__pyx_v_contig, PyObject *__pyx_v_read, PyObject *__pyx_v_indel_type); /* proto */
static PyObject *__pyx_pf_4aldy_9indelpost_6localn_16findall_indels(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ref_aln, PyObject *__pyx_v_genome_aln_pos, PyObject *__pyx_v_ref_seq, PyObject *__pyx_v_read_seq, PyObject *__pyx_v_report_snvs, PyObject *__pyx_v_basequals); /* proto */
static int __pyx_pf_7cpython_5array_5array___getbuffer__(arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_7cpython_5array_5array_2__releasebuffer__(CYTHON_UNUSED arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_23;
static PyObject *__pyx_int_40;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_slice_;
static PyObject *__pyx_slice__2;
static PyObject *__pyx_slice__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_codeobj__7;
static PyObject *__pyx_codeobj__9;
static PyObject *__pyx_codeobj__11;
static PyObject *__pyx_codeobj__13;
static PyObject *__pyx_codeobj__15;
static PyObject *__pyx_codeobj__17;
static PyObject *__pyx_codeobj__19;
static PyObject *__pyx_codeobj__21;
static PyObject *__pyx_codeobj__23;
/* Late includes */

/* "aldy/indelpost/localn.pyx":14
 * 
 * 
 * def find_by_smith_waterman_realn(             # <<<<<<<<<<<<<<
 *     target_indel,
 *     contig,
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_1find_by_smith_waterman_realn(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6localn_find_by_smith_waterman_realn[] = "Annotate if reads contain target indel\n\n    Args:\n        target_indel (Variant)\n        template (dict): indel template\n        pileup (list): list of dictized reads (dict)\n        mapq (int): minimum mapping quality\n    Returns:\n        pileup (list): annotated pileup\n    ";
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6localn_1find_by_smith_waterman_realn = {"find_by_smith_waterman_realn", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6localn_1find_by_smith_waterman_realn, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4aldy_9indelpost_6localn_find_by_smith_waterman_realn};
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_1find_by_smith_waterman_realn(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_target_indel = 0;
  PyObject *__pyx_v_contig = 0;
  PyObject *__pyx_v_pileup = 0;
  PyObject *__pyx_v_match_score = 0;
  PyObject *__pyx_v_mismatch_penalty = 0;
  PyObject *__pyx_v_gap_open_penalty = 0;
  PyObject *__pyx_v_gap_extension_penalty = 0;
  PyObject *__pyx_v_basequalthresh = 0;
  PyObject *__pyx_v_mapq_lim = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_by_smith_waterman_realn (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_target_indel,&__pyx_n_s_contig,&__pyx_n_s_pileup,&__pyx_n_s_match_score,&__pyx_n_s_mismatch_penalty,&__pyx_n_s_gap_open_penalty,&__pyx_n_s_gap_extension_penalty,&__pyx_n_s_basequalthresh,&__pyx_n_s_mapq_lim,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    values[8] = ((PyObject *)__pyx_int_1);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_target_indel)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_contig)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_by_smith_waterman_realn", 0, 8, 9, 1); __PYX_ERR(0, 14, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pileup)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_by_smith_waterman_realn", 0, 8, 9, 2); __PYX_ERR(0, 14, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_match_score)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_by_smith_waterman_realn", 0, 8, 9, 3); __PYX_ERR(0, 14, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mismatch_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_by_smith_waterman_realn", 0, 8, 9, 4); __PYX_ERR(0, 14, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gap_open_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_by_smith_waterman_realn", 0, 8, 9, 5); __PYX_ERR(0, 14, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gap_extension_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_by_smith_waterman_realn", 0, 8, 9, 6); __PYX_ERR(0, 14, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_basequalthresh)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_by_smith_waterman_realn", 0, 8, 9, 7); __PYX_ERR(0, 14, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mapq_lim);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_by_smith_waterman_realn") < 0)) __PYX_ERR(0, 14, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_target_indel = values[0];
    __pyx_v_contig = values[1];
    __pyx_v_pileup = values[2];
    __pyx_v_match_score = values[3];
    __pyx_v_mismatch_penalty = values[4];
    __pyx_v_gap_open_penalty = values[5];
    __pyx_v_gap_extension_penalty = values[6];
    __pyx_v_basequalthresh = values[7];
    __pyx_v_mapq_lim = values[8];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_by_smith_waterman_realn", 0, 8, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 14, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.localn.find_by_smith_waterman_realn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6localn_find_by_smith_waterman_realn(__pyx_self, __pyx_v_target_indel, __pyx_v_contig, __pyx_v_pileup, __pyx_v_match_score, __pyx_v_mismatch_penalty, __pyx_v_gap_open_penalty, __pyx_v_gap_extension_penalty, __pyx_v_basequalthresh, __pyx_v_mapq_lim);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6localn_find_by_smith_waterman_realn(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_target_indel, PyObject *__pyx_v_contig, PyObject *__pyx_v_pileup, PyObject *__pyx_v_match_score, PyObject *__pyx_v_mismatch_penalty, PyObject *__pyx_v_gap_open_penalty, PyObject *__pyx_v_gap_extension_penalty, PyObject *__pyx_v_basequalthresh, PyObject *__pyx_v_mapq_lim) {
  PyObject *__pyx_v_indel_type = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_indel_seq = NULL;
  PyObject *__pyx_v_mut_ref_lt = NULL;
  PyObject *__pyx_v_mut_ref_mid = NULL;
  PyObject *__pyx_v_mut_ref_rt = NULL;
  PyObject *__pyx_v_ref_ref = NULL;
  PyObject *__pyx_v_mut_ref = NULL;
  PyObject *__pyx_v_mut_aligner = NULL;
  PyObject *__pyx_v_ref_aligner = NULL;
  PyObject *__pyx_7genexpr__pyx_v_read = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_read = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_by_smith_waterman_realn", 0);
  __Pyx_INCREF(__pyx_v_pileup);

  /* "aldy/indelpost/localn.pyx":35
 *         pileup (list): annotated pileup
 *     """
 *     indel_type, indel_seq = target_indel.variant_type, target_indel.indel_seq             # <<<<<<<<<<<<<<
 * 
 *     mut_ref_lt, mut_ref_mid, mut_ref_rt = contig.get_contig_seq(split=True)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_target_indel, __pyx_n_s_variant_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_target_indel, __pyx_n_s_indel_seq); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_indel_type = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_indel_seq = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "aldy/indelpost/localn.pyx":37
 *     indel_type, indel_seq = target_indel.variant_type, target_indel.indel_seq
 * 
 *     mut_ref_lt, mut_ref_mid, mut_ref_rt = contig.get_contig_seq(split=True)             # <<<<<<<<<<<<<<
 *     ref_ref = contig.get_reference_seq()
 *     mut_ref = mut_ref_lt + mut_ref_mid + mut_ref_rt
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_contig, __pyx_n_s_get_contig_seq); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_split, Py_True) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 37, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 2; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 3) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 37, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_mut_ref_lt = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_mut_ref_mid = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_mut_ref_rt = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "aldy/indelpost/localn.pyx":38
 * 
 *     mut_ref_lt, mut_ref_mid, mut_ref_rt = contig.get_contig_seq(split=True)
 *     ref_ref = contig.get_reference_seq()             # <<<<<<<<<<<<<<
 *     mut_ref = mut_ref_lt + mut_ref_mid + mut_ref_rt
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_contig, __pyx_n_s_get_reference_seq); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_ref_ref = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "aldy/indelpost/localn.pyx":39
 *     mut_ref_lt, mut_ref_mid, mut_ref_rt = contig.get_contig_seq(split=True)
 *     ref_ref = contig.get_reference_seq()
 *     mut_ref = mut_ref_lt + mut_ref_mid + mut_ref_rt             # <<<<<<<<<<<<<<
 * 
 *     mut_aligner = make_aligner(mut_ref, match_score, mismatch_penalty)
 */
  __pyx_t_3 = PyNumber_Add(__pyx_v_mut_ref_lt, __pyx_v_mut_ref_mid); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_v_mut_ref_rt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_mut_ref = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "aldy/indelpost/localn.pyx":41
 *     mut_ref = mut_ref_lt + mut_ref_mid + mut_ref_rt
 * 
 *     mut_aligner = make_aligner(mut_ref, match_score, mismatch_penalty)             # <<<<<<<<<<<<<<
 *     ref_aligner = make_aligner(ref_ref, match_score, mismatch_penalty)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_make_aligner); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_mut_ref, __pyx_v_match_score, __pyx_v_mismatch_penalty};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_mut_ref, __pyx_v_match_score, __pyx_v_mismatch_penalty};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  {
    __pyx_t_1 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_mut_ref);
    __Pyx_GIVEREF(__pyx_v_mut_ref);
    PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_7, __pyx_v_mut_ref);
    __Pyx_INCREF(__pyx_v_match_score);
    __Pyx_GIVEREF(__pyx_v_match_score);
    PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_7, __pyx_v_match_score);
    __Pyx_INCREF(__pyx_v_mismatch_penalty);
    __Pyx_GIVEREF(__pyx_v_mismatch_penalty);
    PyTuple_SET_ITEM(__pyx_t_1, 2+__pyx_t_7, __pyx_v_mismatch_penalty);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_mut_aligner = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "aldy/indelpost/localn.pyx":42
 * 
 *     mut_aligner = make_aligner(mut_ref, match_score, mismatch_penalty)
 *     ref_aligner = make_aligner(ref_ref, match_score, mismatch_penalty)             # <<<<<<<<<<<<<<
 * 
 *     pileup = [findall_mismatches(read) for read in pileup]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_make_aligner); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_v_ref_ref, __pyx_v_match_score, __pyx_v_mismatch_penalty};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_v_ref_ref, __pyx_v_match_score, __pyx_v_mismatch_penalty};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_1) {
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1); __pyx_t_1 = NULL;
    }
    __Pyx_INCREF(__pyx_v_ref_ref);
    __Pyx_GIVEREF(__pyx_v_ref_ref);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_7, __pyx_v_ref_ref);
    __Pyx_INCREF(__pyx_v_match_score);
    __Pyx_GIVEREF(__pyx_v_match_score);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_7, __pyx_v_match_score);
    __Pyx_INCREF(__pyx_v_mismatch_penalty);
    __Pyx_GIVEREF(__pyx_v_mismatch_penalty);
    PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_7, __pyx_v_mismatch_penalty);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ref_aligner = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "aldy/indelpost/localn.pyx":44
 *     ref_aligner = make_aligner(ref_ref, match_score, mismatch_penalty)
 * 
 *     pileup = [findall_mismatches(read) for read in pileup]             # <<<<<<<<<<<<<<
 * 
 *     pileup = [
 */
  { /* enter inner scope */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 44, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (likely(PyList_CheckExact(__pyx_v_pileup)) || PyTuple_CheckExact(__pyx_v_pileup)) {
      __pyx_t_3 = __pyx_v_pileup; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_pileup); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 44, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 44, __pyx_L7_error)
    }
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 44, __pyx_L7_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 44, __pyx_L7_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_9(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 44, __pyx_L7_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_read, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_findall_mismatches); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_7genexpr__pyx_v_read) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_7genexpr__pyx_v_read);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 44, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_read); __pyx_7genexpr__pyx_v_read = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_read); __pyx_7genexpr__pyx_v_read = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __Pyx_DECREF_SET(__pyx_v_pileup, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "aldy/indelpost/localn.pyx":46
 *     pileup = [findall_mismatches(read) for read in pileup]
 * 
 *     pileup = [             # <<<<<<<<<<<<<<
 *         is_target_by_ssw(
 *             read,
 */
  { /* enter inner scope */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 46, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "aldy/indelpost/localn.pyx":64
 *             mapq_lim,
 *         )
 *         for read in pileup             # <<<<<<<<<<<<<<
 *     ]
 * 
 */
    if (likely(PyList_CheckExact(__pyx_v_pileup)) || PyTuple_CheckExact(__pyx_v_pileup)) {
      __pyx_t_3 = __pyx_v_pileup; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_pileup); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 64, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 64, __pyx_L13_error)
    }
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 64, __pyx_L13_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 64, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 64, __pyx_L13_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 64, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_9(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 64, __pyx_L13_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_read, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "aldy/indelpost/localn.pyx":47
 * 
 *     pileup = [
 *         is_target_by_ssw(             # <<<<<<<<<<<<<<
 *             read,
 *             target_indel,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_is_target_by_ssw); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "aldy/indelpost/localn.pyx":62
 *             indel_type,
 *             basequalthresh,
 *             mapq_lim,             # <<<<<<<<<<<<<<
 *         )
 *         for read in pileup
 */
      __pyx_t_5 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[16] = {__pyx_t_5, __pyx_8genexpr1__pyx_v_read, __pyx_v_target_indel, __pyx_v_contig, __pyx_v_mut_ref_lt, __pyx_v_mut_ref_mid, __pyx_v_mut_ref_rt, __pyx_v_mut_aligner, __pyx_v_ref_aligner, __pyx_v_match_score, __pyx_v_mismatch_penalty, __pyx_v_gap_open_penalty, __pyx_v_gap_extension_penalty, __pyx_v_indel_type, __pyx_v_basequalthresh, __pyx_v_mapq_lim};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 15+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L13_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[16] = {__pyx_t_5, __pyx_8genexpr1__pyx_v_read, __pyx_v_target_indel, __pyx_v_contig, __pyx_v_mut_ref_lt, __pyx_v_mut_ref_mid, __pyx_v_mut_ref_rt, __pyx_v_mut_aligner, __pyx_v_ref_aligner, __pyx_v_match_score, __pyx_v_mismatch_penalty, __pyx_v_gap_open_penalty, __pyx_v_gap_extension_penalty, __pyx_v_indel_type, __pyx_v_basequalthresh, __pyx_v_mapq_lim};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 15+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L13_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(15+__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 47, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_INCREF(__pyx_8genexpr1__pyx_v_read);
        __Pyx_GIVEREF(__pyx_8genexpr1__pyx_v_read);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_7, __pyx_8genexpr1__pyx_v_read);
        __Pyx_INCREF(__pyx_v_target_indel);
        __Pyx_GIVEREF(__pyx_v_target_indel);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_7, __pyx_v_target_indel);
        __Pyx_INCREF(__pyx_v_contig);
        __Pyx_GIVEREF(__pyx_v_contig);
        PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_7, __pyx_v_contig);
        __Pyx_INCREF(__pyx_v_mut_ref_lt);
        __Pyx_GIVEREF(__pyx_v_mut_ref_lt);
        PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_7, __pyx_v_mut_ref_lt);
        __Pyx_INCREF(__pyx_v_mut_ref_mid);
        __Pyx_GIVEREF(__pyx_v_mut_ref_mid);
        PyTuple_SET_ITEM(__pyx_t_10, 4+__pyx_t_7, __pyx_v_mut_ref_mid);
        __Pyx_INCREF(__pyx_v_mut_ref_rt);
        __Pyx_GIVEREF(__pyx_v_mut_ref_rt);
        PyTuple_SET_ITEM(__pyx_t_10, 5+__pyx_t_7, __pyx_v_mut_ref_rt);
        __Pyx_INCREF(__pyx_v_mut_aligner);
        __Pyx_GIVEREF(__pyx_v_mut_aligner);
        PyTuple_SET_ITEM(__pyx_t_10, 6+__pyx_t_7, __pyx_v_mut_aligner);
        __Pyx_INCREF(__pyx_v_ref_aligner);
        __Pyx_GIVEREF(__pyx_v_ref_aligner);
        PyTuple_SET_ITEM(__pyx_t_10, 7+__pyx_t_7, __pyx_v_ref_aligner);
        __Pyx_INCREF(__pyx_v_match_score);
        __Pyx_GIVEREF(__pyx_v_match_score);
        PyTuple_SET_ITEM(__pyx_t_10, 8+__pyx_t_7, __pyx_v_match_score);
        __Pyx_INCREF(__pyx_v_mismatch_penalty);
        __Pyx_GIVEREF(__pyx_v_mismatch_penalty);
        PyTuple_SET_ITEM(__pyx_t_10, 9+__pyx_t_7, __pyx_v_mismatch_penalty);
        __Pyx_INCREF(__pyx_v_gap_open_penalty);
        __Pyx_GIVEREF(__pyx_v_gap_open_penalty);
        PyTuple_SET_ITEM(__pyx_t_10, 10+__pyx_t_7, __pyx_v_gap_open_penalty);
        __Pyx_INCREF(__pyx_v_gap_extension_penalty);
        __Pyx_GIVEREF(__pyx_v_gap_extension_penalty);
        PyTuple_SET_ITEM(__pyx_t_10, 11+__pyx_t_7, __pyx_v_gap_extension_penalty);
        __Pyx_INCREF(__pyx_v_indel_type);
        __Pyx_GIVEREF(__pyx_v_indel_type);
        PyTuple_SET_ITEM(__pyx_t_10, 12+__pyx_t_7, __pyx_v_indel_type);
        __Pyx_INCREF(__pyx_v_basequalthresh);
        __Pyx_GIVEREF(__pyx_v_basequalthresh);
        PyTuple_SET_ITEM(__pyx_t_10, 13+__pyx_t_7, __pyx_v_basequalthresh);
        __Pyx_INCREF(__pyx_v_mapq_lim);
        __Pyx_GIVEREF(__pyx_v_mapq_lim);
        PyTuple_SET_ITEM(__pyx_t_10, 14+__pyx_t_7, __pyx_v_mapq_lim);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 46, __pyx_L13_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "aldy/indelpost/localn.pyx":64
 *             mapq_lim,
 *         )
 *         for read in pileup             # <<<<<<<<<<<<<<
 *     ]
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_read); __pyx_8genexpr1__pyx_v_read = 0;
    goto __pyx_L16_exit_scope;
    __pyx_L13_error:;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_read); __pyx_8genexpr1__pyx_v_read = 0;
    goto __pyx_L1_error;
    __pyx_L16_exit_scope:;
  } /* exit inner scope */
  __Pyx_DECREF_SET(__pyx_v_pileup, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "aldy/indelpost/localn.pyx":67
 *     ]
 * 
 *     return pileup             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pileup);
  __pyx_r = __pyx_v_pileup;
  goto __pyx_L0;

  /* "aldy/indelpost/localn.pyx":14
 * 
 * 
 * def find_by_smith_waterman_realn(             # <<<<<<<<<<<<<<
 *     target_indel,
 *     contig,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("aldy.indelpost.localn.find_by_smith_waterman_realn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_indel_type);
  __Pyx_XDECREF(__pyx_v_indel_seq);
  __Pyx_XDECREF(__pyx_v_mut_ref_lt);
  __Pyx_XDECREF(__pyx_v_mut_ref_mid);
  __Pyx_XDECREF(__pyx_v_mut_ref_rt);
  __Pyx_XDECREF(__pyx_v_ref_ref);
  __Pyx_XDECREF(__pyx_v_mut_ref);
  __Pyx_XDECREF(__pyx_v_mut_aligner);
  __Pyx_XDECREF(__pyx_v_ref_aligner);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_read);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_read);
  __Pyx_XDECREF(__pyx_v_pileup);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/localn.pyx":70
 * 
 * 
 * def findall_mismatches(read, end_trim=0):             # <<<<<<<<<<<<<<
 * 
 *     if read["is_reference_seq"]:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_3findall_mismatches(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6localn_3findall_mismatches = {"findall_mismatches", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6localn_3findall_mismatches, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_3findall_mismatches(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_read = 0;
  PyObject *__pyx_v_end_trim = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("findall_mismatches (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_read,&__pyx_n_s_end_trim,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_read)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end_trim);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "findall_mismatches") < 0)) __PYX_ERR(0, 70, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_read = values[0];
    __pyx_v_end_trim = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("findall_mismatches", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 70, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.localn.findall_mismatches", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6localn_2findall_mismatches(__pyx_self, __pyx_v_read, __pyx_v_end_trim);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6localn_2findall_mismatches(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_end_trim) {
  PyObject *__pyx_v_aln_start = NULL;
  PyObject *__pyx_v_aln_end = NULL;
  PyObject *__pyx_v_mapped_subpreads = NULL;
  PyObject *__pyx_v_mismatches = NULL;
  PyObject *__pyx_v_subread = NULL;
  PyObject *__pyx_v_start = NULL;
  PyObject *__pyx_v_end = NULL;
  PyObject *__pyx_v_span = NULL;
  PyObject *__pyx_v_cigarstring = NULL;
  PyObject *__pyx_v_cigarlst = NULL;
  PyObject *__pyx_v_read_seq = NULL;
  PyObject *__pyx_v_quals = NULL;
  PyObject *__pyx_v_lt_seq = NULL;
  PyObject *__pyx_v_rt_seq = NULL;
  PyObject *__pyx_v_lt_qual = NULL;
  PyObject *__pyx_v_rt_qual = NULL;
  PyObject *__pyx_v_lt_ref = NULL;
  PyObject *__pyx_v_rt_ref = NULL;
  PyObject *__pyx_v_mapped_seq = NULL;
  PyObject *__pyx_v_mapped_qual = NULL;
  PyObject *__pyx_v_mapped_ref = NULL;
  PyObject *__pyx_v_pos = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_v_a = NULL;
  PyObject *__pyx_v_q = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *(*__pyx_t_16)(PyObject *);
  int __pyx_t_17;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("findall_mismatches", 0);

  /* "aldy/indelpost/localn.pyx":72
 * def findall_mismatches(read, end_trim=0):
 * 
 *     if read["is_reference_seq"]:             # <<<<<<<<<<<<<<
 *         read["mismatches"] = []
 *         return read
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_is_reference_seq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "aldy/indelpost/localn.pyx":73
 * 
 *     if read["is_reference_seq"]:
 *         read["mismatches"] = []             # <<<<<<<<<<<<<<
 *         return read
 * 
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_mismatches, __pyx_t_1) < 0)) __PYX_ERR(0, 73, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "aldy/indelpost/localn.pyx":74
 *     if read["is_reference_seq"]:
 *         read["mismatches"] = []
 *         return read             # <<<<<<<<<<<<<<
 * 
 *     aln_start, aln_end = read["aln_start"], read["aln_end"]
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_read);
    __pyx_r = __pyx_v_read;
    goto __pyx_L0;

    /* "aldy/indelpost/localn.pyx":72
 * def findall_mismatches(read, end_trim=0):
 * 
 *     if read["is_reference_seq"]:             # <<<<<<<<<<<<<<
 *         read["mismatches"] = []
 *         return read
 */
  }

  /* "aldy/indelpost/localn.pyx":76
 *         return read
 * 
 *     aln_start, aln_end = read["aln_start"], read["aln_end"]             # <<<<<<<<<<<<<<
 *     mapped_subpreads = get_mapped_subreads(read["cigar_string"], aln_start, aln_end)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_aln_start = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_aln_end = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "aldy/indelpost/localn.pyx":77
 * 
 *     aln_start, aln_end = read["aln_start"], read["aln_end"]
 *     mapped_subpreads = get_mapped_subreads(read["cigar_string"], aln_start, aln_end)             # <<<<<<<<<<<<<<
 * 
 *     mismatches = []
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_get_mapped_subreads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_4, __pyx_v_aln_start, __pyx_v_aln_end};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_4, __pyx_v_aln_start, __pyx_v_aln_end};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_aln_start);
    __Pyx_GIVEREF(__pyx_v_aln_start);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_aln_start);
    __Pyx_INCREF(__pyx_v_aln_end);
    __Pyx_GIVEREF(__pyx_v_aln_end);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_v_aln_end);
    __pyx_t_4 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mapped_subpreads = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "aldy/indelpost/localn.pyx":79
 *     mapped_subpreads = get_mapped_subreads(read["cigar_string"], aln_start, aln_end)
 * 
 *     mismatches = []             # <<<<<<<<<<<<<<
 *     for subread in mapped_subpreads:
 *         start, end = subread[0], subread[1]
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_mismatches = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "aldy/indelpost/localn.pyx":80
 * 
 *     mismatches = []
 *     for subread in mapped_subpreads:             # <<<<<<<<<<<<<<
 *         start, end = subread[0], subread[1]
 *         span = end - start + 1
 */
  if (likely(PyList_CheckExact(__pyx_v_mapped_subpreads)) || PyTuple_CheckExact(__pyx_v_mapped_subpreads)) {
    __pyx_t_3 = __pyx_v_mapped_subpreads; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_mapped_subpreads); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 80, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 80, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 80, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_9(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 80, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_subread, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "aldy/indelpost/localn.pyx":81
 *     mismatches = []
 *     for subread in mapped_subpreads:
 *         start, end = subread[0], subread[1]             # <<<<<<<<<<<<<<
 *         span = end - start + 1
 * 
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_subread, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_subread, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_start, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_end, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "aldy/indelpost/localn.pyx":82
 *     for subread in mapped_subpreads:
 *         start, end = subread[0], subread[1]
 *         span = end - start + 1             # <<<<<<<<<<<<<<
 * 
 *         # trim clipped segments
 */
    __pyx_t_7 = PyNumber_Subtract(__pyx_v_end, __pyx_v_start); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_t_7, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_span, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "aldy/indelpost/localn.pyx":85
 * 
 *         # trim clipped segments
 *         cigarstring = read["cigar_string"]             # <<<<<<<<<<<<<<
 *         if "S" in cigarstring:
 *             cigarlst = read["cigar_list"]
 */
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_cigarstring, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "aldy/indelpost/localn.pyx":86
 *         # trim clipped segments
 *         cigarstring = read["cigar_string"]
 *         if "S" in cigarstring:             # <<<<<<<<<<<<<<
 *             cigarlst = read["cigar_list"]
 *             read_seq = read["read_seq"]
 */
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_S, __pyx_v_cigarstring, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 86, __pyx_L1_error)
    __pyx_t_10 = (__pyx_t_2 != 0);
    if (__pyx_t_10) {

      /* "aldy/indelpost/localn.pyx":87
 *         cigarstring = read["cigar_string"]
 *         if "S" in cigarstring:
 *             cigarlst = read["cigar_list"]             # <<<<<<<<<<<<<<
 *             read_seq = read["read_seq"]
 *             quals = read["read_qual"]
 */
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_list); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_cigarlst, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "aldy/indelpost/localn.pyx":88
 *         if "S" in cigarstring:
 *             cigarlst = read["cigar_list"]
 *             read_seq = read["read_seq"]             # <<<<<<<<<<<<<<
 *             quals = read["read_qual"]
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_seq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_read_seq, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "aldy/indelpost/localn.pyx":89
 *             cigarlst = read["cigar_list"]
 *             read_seq = read["read_seq"]
 *             quals = read["read_qual"]             # <<<<<<<<<<<<<<
 * 
 *             if "S" in cigarlst[0]:
 */
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_qual); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_quals, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "aldy/indelpost/localn.pyx":91
 *             quals = read["read_qual"]
 * 
 *             if "S" in cigarlst[0]:             # <<<<<<<<<<<<<<
 *                 cigarlst = cigarlst[1:]
 *                 read_seq = read_seq[read["start_offset"] :]
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_cigarlst, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_S, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 91, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_t_10 != 0);
      if (__pyx_t_2) {

        /* "aldy/indelpost/localn.pyx":92
 * 
 *             if "S" in cigarlst[0]:
 *                 cigarlst = cigarlst[1:]             # <<<<<<<<<<<<<<
 *                 read_seq = read_seq[read["start_offset"] :]
 *                 quals = quals[read["start_offset"] :]
 */
        __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_cigarlst, 1, 0, NULL, NULL, &__pyx_slice_, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF_SET(__pyx_v_cigarlst, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "aldy/indelpost/localn.pyx":93
 *             if "S" in cigarlst[0]:
 *                 cigarlst = cigarlst[1:]
 *                 read_seq = read_seq[read["start_offset"] :]             # <<<<<<<<<<<<<<
 *                 quals = quals[read["start_offset"] :]
 * 
 */
        __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_start_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 93, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, &__pyx_t_1, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 93, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_read_seq, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "aldy/indelpost/localn.pyx":94
 *                 cigarlst = cigarlst[1:]
 *                 read_seq = read_seq[read["start_offset"] :]
 *                 quals = quals[read["start_offset"] :]             # <<<<<<<<<<<<<<
 * 
 *             if "S" in cigarlst[-1]:
 */
        __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_start_offset); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 94, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_quals, 0, 0, &__pyx_t_7, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF_SET(__pyx_v_quals, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "aldy/indelpost/localn.pyx":91
 *             quals = read["read_qual"]
 * 
 *             if "S" in cigarlst[0]:             # <<<<<<<<<<<<<<
 *                 cigarlst = cigarlst[1:]
 *                 read_seq = read_seq[read["start_offset"] :]
 */
      }

      /* "aldy/indelpost/localn.pyx":96
 *                 quals = quals[read["start_offset"] :]
 * 
 *             if "S" in cigarlst[-1]:             # <<<<<<<<<<<<<<
 *                 cigarlst = cigarlst[:-1]
 *                 read_seq = read_seq[: -read["end_offset"]]
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_cigarlst, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_S, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 96, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_10 = (__pyx_t_2 != 0);
      if (__pyx_t_10) {

        /* "aldy/indelpost/localn.pyx":97
 * 
 *             if "S" in cigarlst[-1]:
 *                 cigarlst = cigarlst[:-1]             # <<<<<<<<<<<<<<
 *                 read_seq = read_seq[: -read["end_offset"]]
 *                 quals = quals[: -read["end_offset"]]
 */
        __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_cigarlst, 0, -1L, NULL, NULL, &__pyx_slice__2, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 97, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF_SET(__pyx_v_cigarlst, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "aldy/indelpost/localn.pyx":98
 *             if "S" in cigarlst[-1]:
 *                 cigarlst = cigarlst[:-1]
 *                 read_seq = read_seq[: -read["end_offset"]]             # <<<<<<<<<<<<<<
 *                 quals = quals[: -read["end_offset"]]
 * 
 */
        __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_end_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = PyNumber_Negative(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 98, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, NULL, &__pyx_t_7, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF_SET(__pyx_v_read_seq, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "aldy/indelpost/localn.pyx":99
 *                 cigarlst = cigarlst[:-1]
 *                 read_seq = read_seq[: -read["end_offset"]]
 *                 quals = quals[: -read["end_offset"]]             # <<<<<<<<<<<<<<
 * 
 *             cigarstring = "".join(cigarlst)
 */
        __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_end_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = PyNumber_Negative(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 99, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_quals, 0, 0, NULL, &__pyx_t_7, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF_SET(__pyx_v_quals, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "aldy/indelpost/localn.pyx":96
 *                 quals = quals[read["start_offset"] :]
 * 
 *             if "S" in cigarlst[-1]:             # <<<<<<<<<<<<<<
 *                 cigarlst = cigarlst[:-1]
 *                 read_seq = read_seq[: -read["end_offset"]]
 */
      }

      /* "aldy/indelpost/localn.pyx":101
 *                 quals = quals[: -read["end_offset"]]
 * 
 *             cigarstring = "".join(cigarlst)             # <<<<<<<<<<<<<<
 *         else:
 *             read_seq = read["read_seq"]
 */
      __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__3, __pyx_v_cigarlst); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_cigarstring, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "aldy/indelpost/localn.pyx":86
 *         # trim clipped segments
 *         cigarstring = read["cigar_string"]
 *         if "S" in cigarstring:             # <<<<<<<<<<<<<<
 *             cigarlst = read["cigar_list"]
 *             read_seq = read["read_seq"]
 */
      goto __pyx_L6;
    }

    /* "aldy/indelpost/localn.pyx":103
 *             cigarstring = "".join(cigarlst)
 *         else:
 *             read_seq = read["read_seq"]             # <<<<<<<<<<<<<<
 *             quals = read["read_qual"]
 * 
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_seq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_read_seq, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "aldy/indelpost/localn.pyx":104
 *         else:
 *             read_seq = read["read_seq"]
 *             quals = read["read_qual"]             # <<<<<<<<<<<<<<
 * 
 *         lt_seq, rt_seq = split(
 */
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_qual); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_quals, __pyx_t_1);
      __pyx_t_1 = 0;
    }
    __pyx_L6:;

    /* "aldy/indelpost/localn.pyx":107
 * 
 *         lt_seq, rt_seq = split(
 *             read_seq, cigarstring, start, aln_start, is_for_ref=False, reverse=False,             # <<<<<<<<<<<<<<
 *         )
 *         lt_qual, rt_qual = split(
 */
    if (!(likely(PyUnicode_CheckExact(__pyx_v_cigarstring))||((__pyx_v_cigarstring) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_cigarstring)->tp_name), 0))) __PYX_ERR(0, 107, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_start); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 107, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_v_aln_start); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 107, __pyx_L1_error)

    /* "aldy/indelpost/localn.pyx":106
 *             quals = read["read_qual"]
 * 
 *         lt_seq, rt_seq = split(             # <<<<<<<<<<<<<<
 *             read_seq, cigarstring, start, aln_start, is_for_ref=False, reverse=False,
 *         )
 */
    __pyx_t_1 = __pyx_f_4aldy_9indelpost_9utilities_split(__pyx_v_read_seq, ((PyObject*)__pyx_v_cigarstring), __pyx_t_6, __pyx_t_11, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(__pyx_t_1 != Py_None)) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 106, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 106, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 106, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 106, __pyx_L1_error)
    }
    __Pyx_XDECREF_SET(__pyx_v_lt_seq, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_rt_seq, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "aldy/indelpost/localn.pyx":110
 *         )
 *         lt_qual, rt_qual = split(
 *             quals, cigarstring, start, aln_start, is_for_ref=False, reverse=False,             # <<<<<<<<<<<<<<
 *         )
 *         lt_ref, rt_ref = split(
 */
    if (!(likely(PyUnicode_CheckExact(__pyx_v_cigarstring))||((__pyx_v_cigarstring) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_cigarstring)->tp_name), 0))) __PYX_ERR(0, 110, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_v_start); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_aln_start); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)

    /* "aldy/indelpost/localn.pyx":109
 *             read_seq, cigarstring, start, aln_start, is_for_ref=False, reverse=False,
 *         )
 *         lt_qual, rt_qual = split(             # <<<<<<<<<<<<<<
 *             quals, cigarstring, start, aln_start, is_for_ref=False, reverse=False,
 *         )
 */
    __pyx_t_1 = __pyx_f_4aldy_9indelpost_9utilities_split(__pyx_v_quals, ((PyObject*)__pyx_v_cigarstring), __pyx_t_11, __pyx_t_6, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(__pyx_t_1 != Py_None)) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 109, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 109, __pyx_L1_error)
    }
    __Pyx_XDECREF_SET(__pyx_v_lt_qual, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_rt_qual, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "aldy/indelpost/localn.pyx":113
 *         )
 *         lt_ref, rt_ref = split(
 *             read["ref_seq"],             # <<<<<<<<<<<<<<
 *             cigarstring,
 *             start,
 */
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_ref_seq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "aldy/indelpost/localn.pyx":114
 *         lt_ref, rt_ref = split(
 *             read["ref_seq"],
 *             cigarstring,             # <<<<<<<<<<<<<<
 *             start,
 *             aln_start,
 */
    if (!(likely(PyUnicode_CheckExact(__pyx_v_cigarstring))||((__pyx_v_cigarstring) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_cigarstring)->tp_name), 0))) __PYX_ERR(0, 114, __pyx_L1_error)

    /* "aldy/indelpost/localn.pyx":115
 *             read["ref_seq"],
 *             cigarstring,
 *             start,             # <<<<<<<<<<<<<<
 *             aln_start,
 *             is_for_ref=True,
 */
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_start); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 115, __pyx_L1_error)

    /* "aldy/indelpost/localn.pyx":116
 *             cigarstring,
 *             start,
 *             aln_start,             # <<<<<<<<<<<<<<
 *             is_for_ref=True,
 *             reverse=False,
 */
    __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_v_aln_start); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 116, __pyx_L1_error)

    /* "aldy/indelpost/localn.pyx":112
 *             quals, cigarstring, start, aln_start, is_for_ref=False, reverse=False,
 *         )
 *         lt_ref, rt_ref = split(             # <<<<<<<<<<<<<<
 *             read["ref_seq"],
 *             cigarstring,
 */
    __pyx_t_7 = __pyx_f_4aldy_9indelpost_9utilities_split(__pyx_t_1, ((PyObject*)__pyx_v_cigarstring), __pyx_t_6, __pyx_t_11, 1, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(__pyx_t_7 != Py_None)) {
      PyObject* sequence = __pyx_t_7;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 112, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 112, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 112, __pyx_L1_error)
    }
    __Pyx_XDECREF_SET(__pyx_v_lt_ref, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_rt_ref, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "aldy/indelpost/localn.pyx":121
 *         )
 * 
 *         mapped_seq = lt_seq[-1] + rt_seq[: span - 1]             # <<<<<<<<<<<<<<
 *         mapped_qual = [lt_qual[-1]] + list(rt_qual[: span - 1])
 *         mapped_ref = lt_ref[-1] + rt_ref[: span - 1]
 */
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_lt_seq, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_span, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_rt_seq, 0, 0, NULL, &__pyx_t_4, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_mapped_seq, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "aldy/indelpost/localn.pyx":122
 * 
 *         mapped_seq = lt_seq[-1] + rt_seq[: span - 1]
 *         mapped_qual = [lt_qual[-1]] + list(rt_qual[: span - 1])             # <<<<<<<<<<<<<<
 *         mapped_ref = lt_ref[-1] + rt_ref[: span - 1]
 * 
 */
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_lt_qual, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_span, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_PyObject_GetSlice(__pyx_v_rt_qual, 0, 0, NULL, &__pyx_t_4, NULL, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PySequence_List(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyNumber_Add(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_mapped_qual, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "aldy/indelpost/localn.pyx":123
 *         mapped_seq = lt_seq[-1] + rt_seq[: span - 1]
 *         mapped_qual = [lt_qual[-1]] + list(rt_qual[: span - 1])
 *         mapped_ref = lt_ref[-1] + rt_ref[: span - 1]             # <<<<<<<<<<<<<<
 * 
 *         pos = start  # pos for first elem of the rt side
 */
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_lt_ref, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_span, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_rt_ref, 0, 0, NULL, &__pyx_t_4, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_mapped_ref, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "aldy/indelpost/localn.pyx":125
 *         mapped_ref = lt_ref[-1] + rt_ref[: span - 1]
 * 
 *         pos = start  # pos for first elem of the rt side             # <<<<<<<<<<<<<<
 *         for r, a, q in zip(mapped_ref, mapped_seq, mapped_qual):
 *             if r != a:
 */
    __Pyx_INCREF(__pyx_v_start);
    __Pyx_XDECREF_SET(__pyx_v_pos, __pyx_v_start);

    /* "aldy/indelpost/localn.pyx":126
 * 
 *         pos = start  # pos for first elem of the rt side
 *         for r, a, q in zip(mapped_ref, mapped_seq, mapped_qual):             # <<<<<<<<<<<<<<
 *             if r != a:
 *                 if aln_start + end_trim <= pos <= aln_end - end_trim:
 */
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_mapped_ref);
    __Pyx_GIVEREF(__pyx_v_mapped_ref);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_mapped_ref);
    __Pyx_INCREF(__pyx_v_mapped_seq);
    __Pyx_GIVEREF(__pyx_v_mapped_seq);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_mapped_seq);
    __Pyx_INCREF(__pyx_v_mapped_qual);
    __Pyx_GIVEREF(__pyx_v_mapped_qual);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_mapped_qual);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_4 = __pyx_t_1; __Pyx_INCREF(__pyx_t_4); __pyx_t_12 = 0;
      __pyx_t_13 = NULL;
    } else {
      __pyx_t_12 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 126, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_13 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 126, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_13)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_12); __Pyx_INCREF(__pyx_t_1); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 126, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_12); __Pyx_INCREF(__pyx_t_1); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 126, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_13(__pyx_t_4);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 126, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
        PyObject* sequence = __pyx_t_1;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 3)) {
          if (size > 3) __Pyx_RaiseTooManyValuesError(3);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 126, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
          __pyx_t_14 = PyTuple_GET_ITEM(sequence, 2); 
        } else {
          __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
          __pyx_t_14 = PyList_GET_ITEM(sequence, 2); 
        }
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_14);
        #else
        __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 126, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 126, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_14 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 126, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        #endif
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_15 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 126, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_16 = Py_TYPE(__pyx_t_15)->tp_iternext;
        index = 0; __pyx_t_7 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_7)) goto __pyx_L11_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        index = 1; __pyx_t_5 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_5)) goto __pyx_L11_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        index = 2; __pyx_t_14 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_14)) goto __pyx_L11_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_14);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_15), 3) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
        __pyx_t_16 = NULL;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        goto __pyx_L12_unpacking_done;
        __pyx_L11_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_16 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 126, __pyx_L1_error)
        __pyx_L12_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_a, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_q, __pyx_t_14);
      __pyx_t_14 = 0;

      /* "aldy/indelpost/localn.pyx":127
 *         pos = start  # pos for first elem of the rt side
 *         for r, a, q in zip(mapped_ref, mapped_seq, mapped_qual):
 *             if r != a:             # <<<<<<<<<<<<<<
 *                 if aln_start + end_trim <= pos <= aln_end - end_trim:
 *                     mismatches.append((pos, r.upper(), a, q))
 */
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_r, __pyx_v_a, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 127, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_10) {

        /* "aldy/indelpost/localn.pyx":128
 *         for r, a, q in zip(mapped_ref, mapped_seq, mapped_qual):
 *             if r != a:
 *                 if aln_start + end_trim <= pos <= aln_end - end_trim:             # <<<<<<<<<<<<<<
 *                     mismatches.append((pos, r.upper(), a, q))
 * 
 */
        __pyx_t_1 = PyNumber_Add(__pyx_v_aln_start, __pyx_v_end_trim); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 128, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_14 = PyObject_RichCompare(__pyx_t_1, __pyx_v_pos, Py_LE); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 128, __pyx_L1_error)
        if (__Pyx_PyObject_IsTrue(__pyx_t_14)) {
          __Pyx_DECREF(__pyx_t_14);
          __pyx_t_5 = PyNumber_Subtract(__pyx_v_aln_end, __pyx_v_end_trim); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 128, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_14 = PyObject_RichCompare(__pyx_v_pos, __pyx_t_5, Py_LE); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 128, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 128, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (__pyx_t_10) {

          /* "aldy/indelpost/localn.pyx":129
 *             if r != a:
 *                 if aln_start + end_trim <= pos <= aln_end - end_trim:
 *                     mismatches.append((pos, r.upper(), a, q))             # <<<<<<<<<<<<<<
 * 
 *             pos += 1
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_upper); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          __pyx_t_14 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 129, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_v_pos);
          __Pyx_GIVEREF(__pyx_v_pos);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_pos);
          __Pyx_GIVEREF(__pyx_t_14);
          PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_14);
          __Pyx_INCREF(__pyx_v_a);
          __Pyx_GIVEREF(__pyx_v_a);
          PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_a);
          __Pyx_INCREF(__pyx_v_q);
          __Pyx_GIVEREF(__pyx_v_q);
          PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_q);
          __pyx_t_14 = 0;
          __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_mismatches, __pyx_t_1); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 129, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "aldy/indelpost/localn.pyx":128
 *         for r, a, q in zip(mapped_ref, mapped_seq, mapped_qual):
 *             if r != a:
 *                 if aln_start + end_trim <= pos <= aln_end - end_trim:             # <<<<<<<<<<<<<<
 *                     mismatches.append((pos, r.upper(), a, q))
 * 
 */
        }

        /* "aldy/indelpost/localn.pyx":127
 *         pos = start  # pos for first elem of the rt side
 *         for r, a, q in zip(mapped_ref, mapped_seq, mapped_qual):
 *             if r != a:             # <<<<<<<<<<<<<<
 *                 if aln_start + end_trim <= pos <= aln_end - end_trim:
 *                     mismatches.append((pos, r.upper(), a, q))
 */
      }

      /* "aldy/indelpost/localn.pyx":131
 *                     mismatches.append((pos, r.upper(), a, q))
 * 
 *             pos += 1             # <<<<<<<<<<<<<<
 * 
 *     read["mismatches"] = mismatches
 */
      __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_pos, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_pos, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "aldy/indelpost/localn.pyx":126
 * 
 *         pos = start  # pos for first elem of the rt side
 *         for r, a, q in zip(mapped_ref, mapped_seq, mapped_qual):             # <<<<<<<<<<<<<<
 *             if r != a:
 *                 if aln_start + end_trim <= pos <= aln_end - end_trim:
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "aldy/indelpost/localn.pyx":80
 * 
 *     mismatches = []
 *     for subread in mapped_subpreads:             # <<<<<<<<<<<<<<
 *         start, end = subread[0], subread[1]
 *         span = end - start + 1
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "aldy/indelpost/localn.pyx":133
 *             pos += 1
 * 
 *     read["mismatches"] = mismatches             # <<<<<<<<<<<<<<
 * 
 *     return read
 */
  if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_mismatches, __pyx_v_mismatches) < 0)) __PYX_ERR(0, 133, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":135
 *     read["mismatches"] = mismatches
 * 
 *     return read             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_read);
  __pyx_r = __pyx_v_read;
  goto __pyx_L0;

  /* "aldy/indelpost/localn.pyx":70
 * 
 * 
 * def findall_mismatches(read, end_trim=0):             # <<<<<<<<<<<<<<
 * 
 *     if read["is_reference_seq"]:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("aldy.indelpost.localn.findall_mismatches", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_aln_start);
  __Pyx_XDECREF(__pyx_v_aln_end);
  __Pyx_XDECREF(__pyx_v_mapped_subpreads);
  __Pyx_XDECREF(__pyx_v_mismatches);
  __Pyx_XDECREF(__pyx_v_subread);
  __Pyx_XDECREF(__pyx_v_start);
  __Pyx_XDECREF(__pyx_v_end);
  __Pyx_XDECREF(__pyx_v_span);
  __Pyx_XDECREF(__pyx_v_cigarstring);
  __Pyx_XDECREF(__pyx_v_cigarlst);
  __Pyx_XDECREF(__pyx_v_read_seq);
  __Pyx_XDECREF(__pyx_v_quals);
  __Pyx_XDECREF(__pyx_v_lt_seq);
  __Pyx_XDECREF(__pyx_v_rt_seq);
  __Pyx_XDECREF(__pyx_v_lt_qual);
  __Pyx_XDECREF(__pyx_v_rt_qual);
  __Pyx_XDECREF(__pyx_v_lt_ref);
  __Pyx_XDECREF(__pyx_v_rt_ref);
  __Pyx_XDECREF(__pyx_v_mapped_seq);
  __Pyx_XDECREF(__pyx_v_mapped_qual);
  __Pyx_XDECREF(__pyx_v_mapped_ref);
  __Pyx_XDECREF(__pyx_v_pos);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XDECREF(__pyx_v_a);
  __Pyx_XDECREF(__pyx_v_q);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/localn.pyx":138
 * 
 * 
 * def is_worth_realn(read, target_indel, qual_lim=23):             # <<<<<<<<<<<<<<
 * 
 *     if read["covering_subread"]:
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_5is_worth_realn(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6localn_5is_worth_realn = {"is_worth_realn", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6localn_5is_worth_realn, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_5is_worth_realn(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_read = 0;
  PyObject *__pyx_v_target_indel = 0;
  PyObject *__pyx_v_qual_lim = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_worth_realn (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_read,&__pyx_n_s_target_indel,&__pyx_n_s_qual_lim,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)__pyx_int_23);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_read)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_target_indel)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_worth_realn", 0, 2, 3, 1); __PYX_ERR(0, 138, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_qual_lim);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "is_worth_realn") < 0)) __PYX_ERR(0, 138, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_read = values[0];
    __pyx_v_target_indel = values[1];
    __pyx_v_qual_lim = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_worth_realn", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 138, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.localn.is_worth_realn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6localn_4is_worth_realn(__pyx_self, __pyx_v_read, __pyx_v_target_indel, __pyx_v_qual_lim);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_4aldy_9indelpost_6localn_14is_worth_realn_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "aldy/indelpost/localn.pyx":199
 *     if mismatches:
 *         if is_lefty:
 *             lt_most_pos = min(var[0] for var in mismatches)             # <<<<<<<<<<<<<<
 *             with_end_mismatches = abs(lt_most_pos - read["aln_start"]) < 4
 *         else:
 */

static PyObject *__pyx_pf_4aldy_9indelpost_6localn_14is_worth_realn_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr *)__pyx_tp_new_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr(__pyx_ptype_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 199, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_4aldy_9indelpost_6localn_14is_worth_realn_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_is_worth_realn_locals_genexpr, __pyx_n_s_aldy_indelpost_localn); if (unlikely(!gen)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("aldy.indelpost.localn.is_worth_realn.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_4aldy_9indelpost_6localn_14is_worth_realn_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 199, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_mismatches)) { __Pyx_RaiseClosureNameError("mismatches"); __PYX_ERR(0, 199, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_mismatches == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 199, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_mismatches; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 199, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_var);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_var, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_var, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 199, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_4aldy_9indelpost_6localn_14is_worth_realn_5generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "aldy/indelpost/localn.pyx":202
 *             with_end_mismatches = abs(lt_most_pos - read["aln_start"]) < 4
 *         else:
 *             rt_most_pos = max(var[0] for var in mismatches)             # <<<<<<<<<<<<<<
 *             with_end_mismatches = abs(rt_most_pos - read["aln_end"]) < 4
 * 
 */

static PyObject *__pyx_pf_4aldy_9indelpost_6localn_14is_worth_realn_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr *)__pyx_tp_new_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr(__pyx_ptype_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 202, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_4aldy_9indelpost_6localn_14is_worth_realn_5generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_is_worth_realn_locals_genexpr, __pyx_n_s_aldy_indelpost_localn); if (unlikely(!gen)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("aldy.indelpost.localn.is_worth_realn.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_4aldy_9indelpost_6localn_14is_worth_realn_5generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr *__pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 202, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_mismatches)) { __Pyx_RaiseClosureNameError("mismatches"); __PYX_ERR(0, 202, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_mismatches == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 202, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_mismatches; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 202, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_var);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_var, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_var, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 202, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/localn.pyx":138
 * 
 * 
 * def is_worth_realn(read, target_indel, qual_lim=23):             # <<<<<<<<<<<<<<
 * 
 *     if read["covering_subread"]:
 */

static PyObject *__pyx_pf_4aldy_9indelpost_6localn_4is_worth_realn(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_target_indel, PyObject *__pyx_v_qual_lim) {
  struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *__pyx_cur_scope;
  int __pyx_v_is_covered;
  PyObject *__pyx_v_covering_start = NULL;
  PyObject *__pyx_v_covering_end = NULL;
  PyObject *__pyx_v_dist_to_left_end = NULL;
  PyObject *__pyx_v_dist_to_right_end = NULL;
  PyObject *__pyx_v_is_lefty = NULL;
  PyObject *__pyx_v_start_cigar = NULL;
  PyObject *__pyx_v_end_cigar = NULL;
  PyObject *__pyx_v_shiftable_pos = NULL;
  PyObject *__pyx_v_lt_pos = NULL;
  PyObject *__pyx_v_rt_pos = NULL;
  PyObject *__pyx_v_lt_end_read = NULL;
  PyObject *__pyx_v_lt_end_ref = NULL;
  PyObject *__pyx_v_rt_end_read = NULL;
  PyObject *__pyx_v_rt_end_ref = NULL;
  PyObject *__pyx_v_lt_most_pos = NULL;
  PyObject *__pyx_v_with_end_mismatches = NULL;
  PyObject *__pyx_v_rt_most_pos = NULL;
  PyObject *__pyx_v_indels = NULL;
  PyObject *__pyx_8genexpr2__pyx_v_var = NULL;
  PyObject *__pyx_8genexpr3__pyx_v_v = NULL;
  PyObject *__pyx_gb_4aldy_9indelpost_6localn_14is_worth_realn_2generator = 0;
  PyObject *__pyx_gb_4aldy_9indelpost_6localn_14is_worth_realn_5generator1 = 0;
  PyObject *__pyx_8genexpr6__pyx_v_var = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_worth_realn", 0);
  __pyx_cur_scope = (struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *)__pyx_tp_new_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn(__pyx_ptype_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 138, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "aldy/indelpost/localn.pyx":140
 * def is_worth_realn(read, target_indel, qual_lim=23):
 * 
 *     if read["covering_subread"]:             # <<<<<<<<<<<<<<
 *         is_covered = True
 *         covering_start, covering_end = (
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_covering_subread); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "aldy/indelpost/localn.pyx":141
 * 
 *     if read["covering_subread"]:
 *         is_covered = True             # <<<<<<<<<<<<<<
 *         covering_start, covering_end = (
 *             read["covering_subread"][0],
 */
    __pyx_v_is_covered = 1;

    /* "aldy/indelpost/localn.pyx":143
 *         is_covered = True
 *         covering_start, covering_end = (
 *             read["covering_subread"][0],             # <<<<<<<<<<<<<<
 *             read["covering_subread"][1],
 *         )
 */
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_covering_subread); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "aldy/indelpost/localn.pyx":144
 *         covering_start, covering_end = (
 *             read["covering_subread"][0],
 *             read["covering_subread"][1],             # <<<<<<<<<<<<<<
 *         )
 *     else:
 */
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_covering_subread); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_covering_start = __pyx_t_3;
    __pyx_t_3 = 0;
    __pyx_v_covering_end = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "aldy/indelpost/localn.pyx":140
 * def is_worth_realn(read, target_indel, qual_lim=23):
 * 
 *     if read["covering_subread"]:             # <<<<<<<<<<<<<<
 *         is_covered = True
 *         covering_start, covering_end = (
 */
    goto __pyx_L3;
  }

  /* "aldy/indelpost/localn.pyx":147
 *         )
 *     else:
 *         is_covered = False             # <<<<<<<<<<<<<<
 *         if target_indel.is_ins:
 *             return False
 */
  /*else*/ {
    __pyx_v_is_covered = 0;

    /* "aldy/indelpost/localn.pyx":148
 *     else:
 *         is_covered = False
 *         if target_indel.is_ins:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_target_indel, __pyx_n_s_is_ins); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {

      /* "aldy/indelpost/localn.pyx":149
 *         is_covered = False
 *         if target_indel.is_ins:
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             covering_start = target_indel.pos
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      goto __pyx_L0;

      /* "aldy/indelpost/localn.pyx":148
 *     else:
 *         is_covered = False
 *         if target_indel.is_ins:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
    }

    /* "aldy/indelpost/localn.pyx":151
 *             return False
 *         else:
 *             covering_start = target_indel.pos             # <<<<<<<<<<<<<<
 *             covering_end = covering_start + len(target_indel.ref)
 * 
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_target_indel, __pyx_n_s_pos); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_covering_start = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "aldy/indelpost/localn.pyx":152
 *         else:
 *             covering_start = target_indel.pos
 *             covering_end = covering_start + len(target_indel.ref)             # <<<<<<<<<<<<<<
 * 
 *     dist_to_left_end = target_indel.pos - read["aln_start"]
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_target_indel, __pyx_n_s_ref); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 152, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyNumber_Add(__pyx_v_covering_start, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_covering_end = __pyx_t_3;
      __pyx_t_3 = 0;
    }
  }
  __pyx_L3:;

  /* "aldy/indelpost/localn.pyx":154
 *             covering_end = covering_start + len(target_indel.ref)
 * 
 *     dist_to_left_end = target_indel.pos - read["aln_start"]             # <<<<<<<<<<<<<<
 *     dist_to_right_end = read["aln_end"] - target_indel.pos
 *     if dist_to_left_end < 0:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_target_indel, __pyx_n_s_pos); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_start); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyNumber_Subtract(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_dist_to_left_end = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":155
 * 
 *     dist_to_left_end = target_indel.pos - read["aln_start"]
 *     dist_to_right_end = read["aln_end"] - target_indel.pos             # <<<<<<<<<<<<<<
 *     if dist_to_left_end < 0:
 *         is_lefty = True
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_target_indel, __pyx_n_s_pos); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyNumber_Subtract(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_dist_to_right_end = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "aldy/indelpost/localn.pyx":156
 *     dist_to_left_end = target_indel.pos - read["aln_start"]
 *     dist_to_right_end = read["aln_end"] - target_indel.pos
 *     if dist_to_left_end < 0:             # <<<<<<<<<<<<<<
 *         is_lefty = True
 *     elif dist_to_right_end < 0:
 */
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_dist_to_left_end, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 156, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "aldy/indelpost/localn.pyx":157
 *     dist_to_right_end = read["aln_end"] - target_indel.pos
 *     if dist_to_left_end < 0:
 *         is_lefty = True             # <<<<<<<<<<<<<<
 *     elif dist_to_right_end < 0:
 *         is_lefty = False
 */
    __Pyx_INCREF(Py_True);
    __pyx_v_is_lefty = Py_True;

    /* "aldy/indelpost/localn.pyx":156
 *     dist_to_left_end = target_indel.pos - read["aln_start"]
 *     dist_to_right_end = read["aln_end"] - target_indel.pos
 *     if dist_to_left_end < 0:             # <<<<<<<<<<<<<<
 *         is_lefty = True
 *     elif dist_to_right_end < 0:
 */
    goto __pyx_L5;
  }

  /* "aldy/indelpost/localn.pyx":158
 *     if dist_to_left_end < 0:
 *         is_lefty = True
 *     elif dist_to_right_end < 0:             # <<<<<<<<<<<<<<
 *         is_lefty = False
 *     else:
 */
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_dist_to_right_end, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 158, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "aldy/indelpost/localn.pyx":159
 *         is_lefty = True
 *     elif dist_to_right_end < 0:
 *         is_lefty = False             # <<<<<<<<<<<<<<
 *     else:
 *         is_lefty = (dist_to_left_end <= dist_to_right_end)
 */
    __Pyx_INCREF(Py_False);
    __pyx_v_is_lefty = Py_False;

    /* "aldy/indelpost/localn.pyx":158
 *     if dist_to_left_end < 0:
 *         is_lefty = True
 *     elif dist_to_right_end < 0:             # <<<<<<<<<<<<<<
 *         is_lefty = False
 *     else:
 */
    goto __pyx_L5;
  }

  /* "aldy/indelpost/localn.pyx":161
 *         is_lefty = False
 *     else:
 *         is_lefty = (dist_to_left_end <= dist_to_right_end)             # <<<<<<<<<<<<<<
 * 
 *     start_cigar, end_cigar = read["cigar_list"][0], read["cigar_list"][-1]
 */
  /*else*/ {
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_dist_to_left_end, __pyx_v_dist_to_right_end, Py_LE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 161, __pyx_L1_error)
    __pyx_v_is_lefty = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L5:;

  /* "aldy/indelpost/localn.pyx":163
 *         is_lefty = (dist_to_left_end <= dist_to_right_end)
 * 
 *     start_cigar, end_cigar = read["cigar_list"][0], read["cigar_list"][-1]             # <<<<<<<<<<<<<<
 * 
 *     # start clipped
 */
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_list); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_list); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_start_cigar = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_end_cigar = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":166
 * 
 *     # start clipped
 *     if is_lefty and covering_start < read["aln_start"] <= covering_end and int(start_cigar[:-1]) > 2:             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_is_lefty); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 166, __pyx_L1_error)
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_covering_start, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 166, __pyx_L1_error)
  if (__Pyx_PyObject_IsTrue(__pyx_t_4)) {
    __Pyx_DECREF(__pyx_t_4);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_v_covering_end, Py_LE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 166, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_start_cigar, 0, -1L, NULL, NULL, &__pyx_slice__2, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_int_2, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __pyx_t_6;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {

    /* "aldy/indelpost/localn.pyx":167
 *     # start clipped
 *     if is_lefty and covering_start < read["aln_start"] <= covering_end and int(start_cigar[:-1]) > 2:
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     # end clipped
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "aldy/indelpost/localn.pyx":166
 * 
 *     # start clipped
 *     if is_lefty and covering_start < read["aln_start"] <= covering_end and int(start_cigar[:-1]) > 2:             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  }

  /* "aldy/indelpost/localn.pyx":170
 * 
 *     # end clipped
 *     if not is_lefty and  covering_start <= read["aln_end"] < covering_end and int(end_cigar[:-1]) > 2:             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_is_lefty); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 170, __pyx_L1_error)
  __pyx_t_7 = ((!__pyx_t_6) != 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_covering_start, __pyx_t_4, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
  if (__Pyx_PyObject_IsTrue(__pyx_t_1)) {
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_4, __pyx_v_covering_end, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_end_cigar, 0, -1L, NULL, NULL, &__pyx_slice__2, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyNumber_Int(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_4, __pyx_int_2, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __pyx_t_7;
  __pyx_L11_bool_binop_done:;
  if (__pyx_t_2) {

    /* "aldy/indelpost/localn.pyx":171
 *     # end clipped
 *     if not is_lefty and  covering_start <= read["aln_end"] < covering_end and int(end_cigar[:-1]) > 2:
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     mismatches = [
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "aldy/indelpost/localn.pyx":170
 * 
 *     # end clipped
 *     if not is_lefty and  covering_start <= read["aln_end"] < covering_end and int(end_cigar[:-1]) > 2:             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  }

  /* "aldy/indelpost/localn.pyx":173
 *         return True
 * 
 *     mismatches = [             # <<<<<<<<<<<<<<
 *         var
 *         for var in read["mismatches"]
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "aldy/indelpost/localn.pyx":175
 *     mismatches = [
 *         var
 *         for var in read["mismatches"]             # <<<<<<<<<<<<<<
 *         if covering_start <= var[0] <= covering_end and var[3] > qual_lim
 *     ]
 */
    __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_mismatches); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_3 = __pyx_t_4; __Pyx_INCREF(__pyx_t_3); __pyx_t_5 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 175, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 175, __pyx_L16_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 175, __pyx_L16_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 175, __pyx_L16_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_8(__pyx_t_3);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 175, __pyx_L16_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_var, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "aldy/indelpost/localn.pyx":176
 *         var
 *         for var in read["mismatches"]
 *         if covering_start <= var[0] <= covering_end and var[3] > qual_lim             # <<<<<<<<<<<<<<
 *     ]
 * 
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_8genexpr2__pyx_v_var, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 176, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = PyObject_RichCompare(__pyx_v_covering_start, __pyx_t_4, Py_LE); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 176, __pyx_L16_error)
      if (__Pyx_PyObject_IsTrue(__pyx_t_9)) {
        __Pyx_DECREF(__pyx_t_9);
        __pyx_t_9 = PyObject_RichCompare(__pyx_t_4, __pyx_v_covering_end, Py_LE); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 176, __pyx_L16_error)
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 176, __pyx_L16_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (__pyx_t_7) {
      } else {
        __pyx_t_2 = __pyx_t_7;
        goto __pyx_L20_bool_binop_done;
      }
      __pyx_t_9 = __Pyx_GetItemInt(__pyx_8genexpr2__pyx_v_var, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 176, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_9, __pyx_v_qual_lim, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 176, __pyx_L16_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 176, __pyx_L16_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = __pyx_t_7;
      __pyx_L20_bool_binop_done:;
      if (__pyx_t_2) {

        /* "aldy/indelpost/localn.pyx":174
 * 
 *     mismatches = [
 *         var             # <<<<<<<<<<<<<<
 *         for var in read["mismatches"]
 *         if covering_start <= var[0] <= covering_end and var[3] > qual_lim
 */
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_8genexpr2__pyx_v_var))) __PYX_ERR(0, 173, __pyx_L16_error)

        /* "aldy/indelpost/localn.pyx":176
 *         var
 *         for var in read["mismatches"]
 *         if covering_start <= var[0] <= covering_end and var[3] > qual_lim             # <<<<<<<<<<<<<<
 *     ]
 * 
 */
      }

      /* "aldy/indelpost/localn.pyx":175
 *     mismatches = [
 *         var
 *         for var in read["mismatches"]             # <<<<<<<<<<<<<<
 *         if covering_start <= var[0] <= covering_end and var[3] > qual_lim
 *     ]
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_var); __pyx_8genexpr2__pyx_v_var = 0;
    goto __pyx_L22_exit_scope;
    __pyx_L16_error:;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_var); __pyx_8genexpr2__pyx_v_var = 0;
    goto __pyx_L1_error;
    __pyx_L22_exit_scope:;
  } /* exit inner scope */
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_mismatches = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":179
 *     ]
 * 
 *     shiftable_pos = [v.pos for v in target_indel.generate_equivalents()]             # <<<<<<<<<<<<<<
 *     lt_pos, rt_pos = min(shiftable_pos), max(shiftable_pos)
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L25_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_target_indel, __pyx_n_s_generate_equivalents); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 179, __pyx_L25_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L25_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 179, __pyx_L25_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 179, __pyx_L25_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 179, __pyx_L25_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 179, __pyx_L25_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_8(__pyx_t_4);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 179, __pyx_L25_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_v, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr3__pyx_v_v, __pyx_n_s_pos); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L25_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 179, __pyx_L25_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_v); __pyx_8genexpr3__pyx_v_v = 0;
    goto __pyx_L28_exit_scope;
    __pyx_L25_error:;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_v); __pyx_8genexpr3__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L28_exit_scope:;
  } /* exit inner scope */
  __pyx_v_shiftable_pos = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":180
 * 
 *     shiftable_pos = [v.pos for v in target_indel.generate_equivalents()]
 *     lt_pos, rt_pos = min(shiftable_pos), max(shiftable_pos)             # <<<<<<<<<<<<<<
 * 
 *     if lt_pos < rt_pos:
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_v_shiftable_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_v_shiftable_pos); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_lt_pos = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_rt_pos = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "aldy/indelpost/localn.pyx":182
 *     lt_pos, rt_pos = min(shiftable_pos), max(shiftable_pos)
 * 
 *     if lt_pos < rt_pos:             # <<<<<<<<<<<<<<
 *         if is_lefty:
 *             if lt_pos < read["aln_start"]:
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_lt_pos, __pyx_v_rt_pos, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 182, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_2) {

    /* "aldy/indelpost/localn.pyx":183
 * 
 *     if lt_pos < rt_pos:
 *         if is_lefty:             # <<<<<<<<<<<<<<
 *             if lt_pos < read["aln_start"]:
 *                 lt_end_read = read["read_seq"][: (rt_pos - read["aln_start"])]
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_is_lefty); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 183, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "aldy/indelpost/localn.pyx":184
 *     if lt_pos < rt_pos:
 *         if is_lefty:
 *             if lt_pos < read["aln_start"]:             # <<<<<<<<<<<<<<
 *                 lt_end_read = read["read_seq"][: (rt_pos - read["aln_start"])]
 *                 lt_end_ref = read["ref_seq"][: (rt_pos - read["aln_start"])]
 */
      __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_start); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_lt_pos, __pyx_t_4, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 184, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_2) {

        /* "aldy/indelpost/localn.pyx":185
 *         if is_lefty:
 *             if lt_pos < read["aln_start"]:
 *                 lt_end_read = read["read_seq"][: (rt_pos - read["aln_start"])]             # <<<<<<<<<<<<<<
 *                 lt_end_ref = read["ref_seq"][: (rt_pos - read["aln_start"])]
 *                 if lt_end_read == lt_end_ref:
 */
        __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_seq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_start); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = PyNumber_Subtract(__pyx_v_rt_pos, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, 0, NULL, &__pyx_t_3, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_lt_end_read = __pyx_t_4;
        __pyx_t_4 = 0;

        /* "aldy/indelpost/localn.pyx":186
 *             if lt_pos < read["aln_start"]:
 *                 lt_end_read = read["read_seq"][: (rt_pos - read["aln_start"])]
 *                 lt_end_ref = read["ref_seq"][: (rt_pos - read["aln_start"])]             # <<<<<<<<<<<<<<
 *                 if lt_end_read == lt_end_ref:
 *                     return False
 */
        __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_ref_seq); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 186, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 186, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = PyNumber_Subtract(__pyx_v_rt_pos, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_4, 0, 0, NULL, &__pyx_t_1, NULL, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 186, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v_lt_end_ref = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "aldy/indelpost/localn.pyx":187
 *                 lt_end_read = read["read_seq"][: (rt_pos - read["aln_start"])]
 *                 lt_end_ref = read["ref_seq"][: (rt_pos - read["aln_start"])]
 *                 if lt_end_read == lt_end_ref:             # <<<<<<<<<<<<<<
 *                     return False
 *         else:
 */
        __pyx_t_3 = PyObject_RichCompare(__pyx_v_lt_end_read, __pyx_v_lt_end_ref, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 187, __pyx_L1_error)
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 187, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_2) {

          /* "aldy/indelpost/localn.pyx":188
 *                 lt_end_ref = read["ref_seq"][: (rt_pos - read["aln_start"])]
 *                 if lt_end_read == lt_end_ref:
 *                     return False             # <<<<<<<<<<<<<<
 *         else:
 *             if read["aln_end"] <= rt_pos:
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(Py_False);
          __pyx_r = Py_False;
          goto __pyx_L0;

          /* "aldy/indelpost/localn.pyx":187
 *                 lt_end_read = read["read_seq"][: (rt_pos - read["aln_start"])]
 *                 lt_end_ref = read["ref_seq"][: (rt_pos - read["aln_start"])]
 *                 if lt_end_read == lt_end_ref:             # <<<<<<<<<<<<<<
 *                     return False
 *         else:
 */
        }

        /* "aldy/indelpost/localn.pyx":184
 *     if lt_pos < rt_pos:
 *         if is_lefty:
 *             if lt_pos < read["aln_start"]:             # <<<<<<<<<<<<<<
 *                 lt_end_read = read["read_seq"][: (rt_pos - read["aln_start"])]
 *                 lt_end_ref = read["ref_seq"][: (rt_pos - read["aln_start"])]
 */
      }

      /* "aldy/indelpost/localn.pyx":183
 * 
 *     if lt_pos < rt_pos:
 *         if is_lefty:             # <<<<<<<<<<<<<<
 *             if lt_pos < read["aln_start"]:
 *                 lt_end_read = read["read_seq"][: (rt_pos - read["aln_start"])]
 */
      goto __pyx_L30;
    }

    /* "aldy/indelpost/localn.pyx":190
 *                     return False
 *         else:
 *             if read["aln_end"] <= rt_pos:             # <<<<<<<<<<<<<<
 *                 rt_end_read = read["read_seq"][-(read["aln_end"] - lt_pos) :]
 *                 rt_end_ref = read["ref_seq"][-(read["aln_end"] - lt_pos) :]
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_v_rt_pos, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 190, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 190, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_2) {

        /* "aldy/indelpost/localn.pyx":191
 *         else:
 *             if read["aln_end"] <= rt_pos:
 *                 rt_end_read = read["read_seq"][-(read["aln_end"] - lt_pos) :]             # <<<<<<<<<<<<<<
 *                 rt_end_ref = read["ref_seq"][-(read["aln_end"] - lt_pos) :]
 *                 if rt_end_read == rt_end_ref:
 */
        __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_seq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyNumber_Subtract(__pyx_t_3, __pyx_v_lt_pos); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 191, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Negative(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, 0, &__pyx_t_3, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 191, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_rt_end_read = __pyx_t_4;
        __pyx_t_4 = 0;

        /* "aldy/indelpost/localn.pyx":192
 *             if read["aln_end"] <= rt_pos:
 *                 rt_end_read = read["read_seq"][-(read["aln_end"] - lt_pos) :]
 *                 rt_end_ref = read["ref_seq"][-(read["aln_end"] - lt_pos) :]             # <<<<<<<<<<<<<<
 *                 if rt_end_read == rt_end_ref:
 *                     return False
 */
        __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_ref_seq); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 192, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 192, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = PyNumber_Subtract(__pyx_t_3, __pyx_v_lt_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Negative(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 192, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_4, 0, 0, &__pyx_t_3, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_rt_end_ref = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "aldy/indelpost/localn.pyx":193
 *                 rt_end_read = read["read_seq"][-(read["aln_end"] - lt_pos) :]
 *                 rt_end_ref = read["ref_seq"][-(read["aln_end"] - lt_pos) :]
 *                 if rt_end_read == rt_end_ref:             # <<<<<<<<<<<<<<
 *                     return False
 * 
 */
        __pyx_t_1 = PyObject_RichCompare(__pyx_v_rt_end_read, __pyx_v_rt_end_ref, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 193, __pyx_L1_error)
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 193, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_2) {

          /* "aldy/indelpost/localn.pyx":194
 *                 rt_end_ref = read["ref_seq"][-(read["aln_end"] - lt_pos) :]
 *                 if rt_end_read == rt_end_ref:
 *                     return False             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(Py_False);
          __pyx_r = Py_False;
          goto __pyx_L0;

          /* "aldy/indelpost/localn.pyx":193
 *                 rt_end_read = read["read_seq"][-(read["aln_end"] - lt_pos) :]
 *                 rt_end_ref = read["ref_seq"][-(read["aln_end"] - lt_pos) :]
 *                 if rt_end_read == rt_end_ref:             # <<<<<<<<<<<<<<
 *                     return False
 * 
 */
        }

        /* "aldy/indelpost/localn.pyx":190
 *                     return False
 *         else:
 *             if read["aln_end"] <= rt_pos:             # <<<<<<<<<<<<<<
 *                 rt_end_read = read["read_seq"][-(read["aln_end"] - lt_pos) :]
 *                 rt_end_ref = read["ref_seq"][-(read["aln_end"] - lt_pos) :]
 */
      }
    }
    __pyx_L30:;

    /* "aldy/indelpost/localn.pyx":182
 *     lt_pos, rt_pos = min(shiftable_pos), max(shiftable_pos)
 * 
 *     if lt_pos < rt_pos:             # <<<<<<<<<<<<<<
 *         if is_lefty:
 *             if lt_pos < read["aln_start"]:
 */
  }

  /* "aldy/indelpost/localn.pyx":197
 * 
 * 
 *     if mismatches:             # <<<<<<<<<<<<<<
 *         if is_lefty:
 *             lt_most_pos = min(var[0] for var in mismatches)
 */
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_cur_scope->__pyx_v_mismatches) != 0);
  if (__pyx_t_2) {

    /* "aldy/indelpost/localn.pyx":198
 * 
 *     if mismatches:
 *         if is_lefty:             # <<<<<<<<<<<<<<
 *             lt_most_pos = min(var[0] for var in mismatches)
 *             with_end_mismatches = abs(lt_most_pos - read["aln_start"]) < 4
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_is_lefty); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 198, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "aldy/indelpost/localn.pyx":199
 *     if mismatches:
 *         if is_lefty:
 *             lt_most_pos = min(var[0] for var in mismatches)             # <<<<<<<<<<<<<<
 *             with_end_mismatches = abs(lt_most_pos - read["aln_start"]) < 4
 *         else:
 */
      __pyx_t_1 = __pyx_pf_4aldy_9indelpost_6localn_14is_worth_realn_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_lt_most_pos = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "aldy/indelpost/localn.pyx":200
 *         if is_lefty:
 *             lt_most_pos = min(var[0] for var in mismatches)
 *             with_end_mismatches = abs(lt_most_pos - read["aln_start"]) < 4             # <<<<<<<<<<<<<<
 *         else:
 *             rt_most_pos = max(var[0] for var in mismatches)
 */
      __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = PyNumber_Subtract(__pyx_v_lt_most_pos, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyNumber_Absolute(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_int_4, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_with_end_mismatches = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "aldy/indelpost/localn.pyx":198
 * 
 *     if mismatches:
 *         if is_lefty:             # <<<<<<<<<<<<<<
 *             lt_most_pos = min(var[0] for var in mismatches)
 *             with_end_mismatches = abs(lt_most_pos - read["aln_start"]) < 4
 */
      goto __pyx_L36;
    }

    /* "aldy/indelpost/localn.pyx":202
 *             with_end_mismatches = abs(lt_most_pos - read["aln_start"]) < 4
 *         else:
 *             rt_most_pos = max(var[0] for var in mismatches)             # <<<<<<<<<<<<<<
 *             with_end_mismatches = abs(rt_most_pos - read["aln_end"]) < 4
 * 
 */
    /*else*/ {
      __pyx_t_1 = __pyx_pf_4aldy_9indelpost_6localn_14is_worth_realn_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_rt_most_pos = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "aldy/indelpost/localn.pyx":203
 *         else:
 *             rt_most_pos = max(var[0] for var in mismatches)
 *             with_end_mismatches = abs(rt_most_pos - read["aln_end"]) < 4             # <<<<<<<<<<<<<<
 * 
 *         if with_end_mismatches:
 */
      __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = PyNumber_Subtract(__pyx_v_rt_most_pos, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyNumber_Absolute(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_int_4, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_with_end_mismatches = __pyx_t_1;
      __pyx_t_1 = 0;
    }
    __pyx_L36:;

    /* "aldy/indelpost/localn.pyx":205
 *             with_end_mismatches = abs(rt_most_pos - read["aln_end"]) < 4
 * 
 *         if with_end_mismatches:             # <<<<<<<<<<<<<<
 *             return True
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_with_end_mismatches); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 205, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "aldy/indelpost/localn.pyx":206
 * 
 *         if with_end_mismatches:
 *             return True             # <<<<<<<<<<<<<<
 *         else:
 *             if is_covered:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      goto __pyx_L0;

      /* "aldy/indelpost/localn.pyx":205
 *             with_end_mismatches = abs(rt_most_pos - read["aln_end"]) < 4
 * 
 *         if with_end_mismatches:             # <<<<<<<<<<<<<<
 *             return True
 *         else:
 */
    }

    /* "aldy/indelpost/localn.pyx":208
 *             return True
 *         else:
 *             if is_covered:             # <<<<<<<<<<<<<<
 *                 return True
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_is_covered != 0);
      if (__pyx_t_2) {

        /* "aldy/indelpost/localn.pyx":209
 *         else:
 *             if is_covered:
 *                 return True             # <<<<<<<<<<<<<<
 *             else:
 *                 return False
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_True);
        __pyx_r = Py_True;
        goto __pyx_L0;

        /* "aldy/indelpost/localn.pyx":208
 *             return True
 *         else:
 *             if is_covered:             # <<<<<<<<<<<<<<
 *                 return True
 *             else:
 */
      }

      /* "aldy/indelpost/localn.pyx":211
 *                 return True
 *             else:
 *                 return False             # <<<<<<<<<<<<<<
 * 
 *     indels = [
 */
      /*else*/ {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_False);
        __pyx_r = Py_False;
        goto __pyx_L0;
      }
    }

    /* "aldy/indelpost/localn.pyx":197
 * 
 * 
 *     if mismatches:             # <<<<<<<<<<<<<<
 *         if is_lefty:
 *             lt_most_pos = min(var[0] for var in mismatches)
 */
  }

  /* "aldy/indelpost/localn.pyx":213
 *                 return False
 * 
 *     indels = [             # <<<<<<<<<<<<<<
 *         var for var in read["I"] + read["D"] if covering_start <= var[0] <= covering_end
 *     ]
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L41_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "aldy/indelpost/localn.pyx":214
 * 
 *     indels = [
 *         var for var in read["I"] + read["D"] if covering_start <= var[0] <= covering_end             # <<<<<<<<<<<<<<
 *     ]
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_I); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 214, __pyx_L41_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_D); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 214, __pyx_L41_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = PyNumber_Add(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 214, __pyx_L41_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_9)) || PyTuple_CheckExact(__pyx_t_9)) {
      __pyx_t_4 = __pyx_t_9; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 214, __pyx_L41_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 214, __pyx_L41_error)
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_9 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_9); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 214, __pyx_L41_error)
          #else
          __pyx_t_9 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 214, __pyx_L41_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_9); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 214, __pyx_L41_error)
          #else
          __pyx_t_9 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 214, __pyx_L41_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
        }
      } else {
        __pyx_t_9 = __pyx_t_8(__pyx_t_4);
        if (unlikely(!__pyx_t_9)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 214, __pyx_L41_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_9);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_var, __pyx_t_9);
      __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_GetItemInt(__pyx_8genexpr6__pyx_v_var, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 214, __pyx_L41_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_covering_start, __pyx_t_9, Py_LE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 214, __pyx_L41_error)
      if (__Pyx_PyObject_IsTrue(__pyx_t_3)) {
        __Pyx_DECREF(__pyx_t_3);
        __pyx_t_3 = PyObject_RichCompare(__pyx_t_9, __pyx_v_covering_end, Py_LE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 214, __pyx_L41_error)
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 214, __pyx_L41_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_2) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_8genexpr6__pyx_v_var))) __PYX_ERR(0, 213, __pyx_L41_error)
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_var); __pyx_8genexpr6__pyx_v_var = 0;
    goto __pyx_L45_exit_scope;
    __pyx_L41_error:;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_var); __pyx_8genexpr6__pyx_v_var = 0;
    goto __pyx_L1_error;
    __pyx_L45_exit_scope:;
  } /* exit inner scope */
  __pyx_v_indels = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":217
 *     ]
 * 
 *     if indels:             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_indels) != 0);
  if (__pyx_t_2) {

    /* "aldy/indelpost/localn.pyx":218
 * 
 *     if indels:
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     return False
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "aldy/indelpost/localn.pyx":217
 *     ]
 * 
 *     if indels:             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  }

  /* "aldy/indelpost/localn.pyx":220
 *         return True
 * 
 *     return False             # <<<<<<<<<<<<<<
 * 
 * def is_target_by_ssw(
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "aldy/indelpost/localn.pyx":138
 * 
 * 
 * def is_worth_realn(read, target_indel, qual_lim=23):             # <<<<<<<<<<<<<<
 * 
 *     if read["covering_subread"]:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("aldy.indelpost.localn.is_worth_realn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_covering_start);
  __Pyx_XDECREF(__pyx_v_covering_end);
  __Pyx_XDECREF(__pyx_v_dist_to_left_end);
  __Pyx_XDECREF(__pyx_v_dist_to_right_end);
  __Pyx_XDECREF(__pyx_v_is_lefty);
  __Pyx_XDECREF(__pyx_v_start_cigar);
  __Pyx_XDECREF(__pyx_v_end_cigar);
  __Pyx_XDECREF(__pyx_v_shiftable_pos);
  __Pyx_XDECREF(__pyx_v_lt_pos);
  __Pyx_XDECREF(__pyx_v_rt_pos);
  __Pyx_XDECREF(__pyx_v_lt_end_read);
  __Pyx_XDECREF(__pyx_v_lt_end_ref);
  __Pyx_XDECREF(__pyx_v_rt_end_read);
  __Pyx_XDECREF(__pyx_v_rt_end_ref);
  __Pyx_XDECREF(__pyx_v_lt_most_pos);
  __Pyx_XDECREF(__pyx_v_with_end_mismatches);
  __Pyx_XDECREF(__pyx_v_rt_most_pos);
  __Pyx_XDECREF(__pyx_v_indels);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_var);
  __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_v);
  __Pyx_XDECREF(__pyx_gb_4aldy_9indelpost_6localn_14is_worth_realn_2generator);
  __Pyx_XDECREF(__pyx_gb_4aldy_9indelpost_6localn_14is_worth_realn_5generator1);
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_var);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/localn.pyx":222
 *     return False
 * 
 * def is_target_by_ssw(             # <<<<<<<<<<<<<<
 *     read,
 *     target_indel,
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_7is_target_by_ssw(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6localn_7is_target_by_ssw = {"is_target_by_ssw", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6localn_7is_target_by_ssw, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_7is_target_by_ssw(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_read = 0;
  PyObject *__pyx_v_target_indel = 0;
  CYTHON_UNUSED PyObject *__pyx_v_contig = 0;
  PyObject *__pyx_v_mut_ref_lt = 0;
  PyObject *__pyx_v_mut_ref_mid = 0;
  PyObject *__pyx_v_mut_ref_rt = 0;
  PyObject *__pyx_v_mut_aligner = 0;
  PyObject *__pyx_v_ref_aligner = 0;
  CYTHON_UNUSED PyObject *__pyx_v_match_score = 0;
  CYTHON_UNUSED PyObject *__pyx_v_mismatch_penalty = 0;
  PyObject *__pyx_v_gap_open_penalty = 0;
  PyObject *__pyx_v_gap_extension_penalty = 0;
  CYTHON_UNUSED PyObject *__pyx_v_indel_type = 0;
  CYTHON_UNUSED PyObject *__pyx_v_basequalthresh = 0;
  PyObject *__pyx_v_mapq_lim = 0;
  CYTHON_UNUSED PyObject *__pyx_v_mapped_base_cnt_thresh = 0;
  CYTHON_UNUSED PyObject *__pyx_v_allow_mismatches = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_target_by_ssw (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_read,&__pyx_n_s_target_indel,&__pyx_n_s_contig,&__pyx_n_s_mut_ref_lt,&__pyx_n_s_mut_ref_mid,&__pyx_n_s_mut_ref_rt,&__pyx_n_s_mut_aligner,&__pyx_n_s_ref_aligner,&__pyx_n_s_match_score,&__pyx_n_s_mismatch_penalty,&__pyx_n_s_gap_open_penalty,&__pyx_n_s_gap_extension_penalty,&__pyx_n_s_indel_type,&__pyx_n_s_basequalthresh,&__pyx_n_s_mapq_lim,&__pyx_n_s_mapped_base_cnt_thresh,&__pyx_n_s_allow_mismatches,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    values[15] = ((PyObject *)__pyx_int_40);
    values[16] = ((PyObject *)__pyx_int_10);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_read)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_target_indel)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, 1); __PYX_ERR(0, 222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_contig)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, 2); __PYX_ERR(0, 222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mut_ref_lt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, 3); __PYX_ERR(0, 222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mut_ref_mid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, 4); __PYX_ERR(0, 222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mut_ref_rt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, 5); __PYX_ERR(0, 222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mut_aligner)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, 6); __PYX_ERR(0, 222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ref_aligner)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, 7); __PYX_ERR(0, 222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_match_score)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, 8); __PYX_ERR(0, 222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mismatch_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, 9); __PYX_ERR(0, 222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gap_open_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, 10); __PYX_ERR(0, 222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gap_extension_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, 11); __PYX_ERR(0, 222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indel_type)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, 12); __PYX_ERR(0, 222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_basequalthresh)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, 13); __PYX_ERR(0, 222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mapq_lim)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, 14); __PYX_ERR(0, 222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mapped_base_cnt_thresh);
          if (value) { values[15] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_allow_mismatches);
          if (value) { values[16] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "is_target_by_ssw") < 0)) __PYX_ERR(0, 222, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_read = values[0];
    __pyx_v_target_indel = values[1];
    __pyx_v_contig = values[2];
    __pyx_v_mut_ref_lt = values[3];
    __pyx_v_mut_ref_mid = values[4];
    __pyx_v_mut_ref_rt = values[5];
    __pyx_v_mut_aligner = values[6];
    __pyx_v_ref_aligner = values[7];
    __pyx_v_match_score = values[8];
    __pyx_v_mismatch_penalty = values[9];
    __pyx_v_gap_open_penalty = values[10];
    __pyx_v_gap_extension_penalty = values[11];
    __pyx_v_indel_type = values[12];
    __pyx_v_basequalthresh = values[13];
    __pyx_v_mapq_lim = values[14];
    __pyx_v_mapped_base_cnt_thresh = values[15];
    __pyx_v_allow_mismatches = values[16];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_target_by_ssw", 0, 15, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 222, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.localn.is_target_by_ssw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6localn_6is_target_by_ssw(__pyx_self, __pyx_v_read, __pyx_v_target_indel, __pyx_v_contig, __pyx_v_mut_ref_lt, __pyx_v_mut_ref_mid, __pyx_v_mut_ref_rt, __pyx_v_mut_aligner, __pyx_v_ref_aligner, __pyx_v_match_score, __pyx_v_mismatch_penalty, __pyx_v_gap_open_penalty, __pyx_v_gap_extension_penalty, __pyx_v_indel_type, __pyx_v_basequalthresh, __pyx_v_mapq_lim, __pyx_v_mapped_base_cnt_thresh, __pyx_v_allow_mismatches);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6localn_6is_target_by_ssw(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_read, PyObject *__pyx_v_target_indel, CYTHON_UNUSED PyObject *__pyx_v_contig, PyObject *__pyx_v_mut_ref_lt, PyObject *__pyx_v_mut_ref_mid, PyObject *__pyx_v_mut_ref_rt, PyObject *__pyx_v_mut_aligner, PyObject *__pyx_v_ref_aligner, CYTHON_UNUSED PyObject *__pyx_v_match_score, CYTHON_UNUSED PyObject *__pyx_v_mismatch_penalty, PyObject *__pyx_v_gap_open_penalty, PyObject *__pyx_v_gap_extension_penalty, CYTHON_UNUSED PyObject *__pyx_v_indel_type, CYTHON_UNUSED PyObject *__pyx_v_basequalthresh, PyObject *__pyx_v_mapq_lim, CYTHON_UNUSED PyObject *__pyx_v_mapped_base_cnt_thresh, CYTHON_UNUSED PyObject *__pyx_v_allow_mismatches) {
  PyObject *__pyx_v_read_seq = NULL;
  PyObject *__pyx_v_ref_aln = NULL;
  PyObject *__pyx_v_mut_aln = NULL;
  PyObject *__pyx_v_mut_aln_cigar = NULL;
  PyObject *__pyx_v_mut_aln_ref_start = NULL;
  PyObject *__pyx_v_mut_aln_ref_end = NULL;
  PyObject *__pyx_v_mut_aln_read_start = NULL;
  PyObject *__pyx_v_mut_aln_read_end = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_mut_cigar_list = NULL;
  int __pyx_v_target_covered;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_target_by_ssw", 0);

  /* "aldy/indelpost/localn.pyx":243
 * 
 *     # already found
 *     if read["is_target"]:             # <<<<<<<<<<<<<<
 *         return read
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_is_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "aldy/indelpost/localn.pyx":244
 *     # already found
 *     if read["is_target"]:
 *         return read             # <<<<<<<<<<<<<<
 * 
 *     if read["is_reference_seq"] or read["mapq"] <= mapq_lim or not is_worth_realn(read, target_indel):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_read);
    __pyx_r = __pyx_v_read;
    goto __pyx_L0;

    /* "aldy/indelpost/localn.pyx":243
 * 
 *     # already found
 *     if read["is_target"]:             # <<<<<<<<<<<<<<
 *         return read
 * 
 */
  }

  /* "aldy/indelpost/localn.pyx":246
 *         return read
 * 
 *     if read["is_reference_seq"] or read["mapq"] <= mapq_lim or not is_worth_realn(read, target_indel):             # <<<<<<<<<<<<<<
 *         read["is_target"] = False
 *         return read
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_is_reference_seq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_mapq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_v_mapq_lim, Py_LE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L5_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_is_worth_realn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_read, __pyx_v_target_indel};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 246, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_read, __pyx_v_target_indel};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 246, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_read);
    __Pyx_GIVEREF(__pyx_v_read);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_read);
    __Pyx_INCREF(__pyx_v_target_indel);
    __Pyx_GIVEREF(__pyx_v_target_indel);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_target_indel);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_8 = ((!__pyx_t_3) != 0);
  __pyx_t_2 = __pyx_t_8;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_2) {

    /* "aldy/indelpost/localn.pyx":247
 * 
 *     if read["is_reference_seq"] or read["mapq"] <= mapq_lim or not is_worth_realn(read, target_indel):
 *         read["is_target"] = False             # <<<<<<<<<<<<<<
 *         return read
 * 
 */
    if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_is_target, Py_False) < 0)) __PYX_ERR(0, 247, __pyx_L1_error)

    /* "aldy/indelpost/localn.pyx":248
 *     if read["is_reference_seq"] or read["mapq"] <= mapq_lim or not is_worth_realn(read, target_indel):
 *         read["is_target"] = False
 *         return read             # <<<<<<<<<<<<<<
 * 
 *     read_seq = read["read_seq"]
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_read);
    __pyx_r = __pyx_v_read;
    goto __pyx_L0;

    /* "aldy/indelpost/localn.pyx":246
 *         return read
 * 
 *     if read["is_reference_seq"] or read["mapq"] <= mapq_lim or not is_worth_realn(read, target_indel):             # <<<<<<<<<<<<<<
 *         read["is_target"] = False
 *         return read
 */
  }

  /* "aldy/indelpost/localn.pyx":250
 *         return read
 * 
 *     read_seq = read["read_seq"]             # <<<<<<<<<<<<<<
 * 
 *     ref_aln = align(ref_aligner, read_seq, gap_open_penalty, gap_extension_penalty)
 */
  __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_seq); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_read_seq = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "aldy/indelpost/localn.pyx":252
 *     read_seq = read["read_seq"]
 * 
 *     ref_aln = align(ref_aligner, read_seq, gap_open_penalty, gap_extension_penalty)             # <<<<<<<<<<<<<<
 *     # forced to align by setting gap_open_penalty=len(read_seq)
 *     mut_aln = align(mut_aligner, read_seq, len(read_seq), gap_extension_penalty)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_align); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_ref_aligner, __pyx_v_read_seq, __pyx_v_gap_open_penalty, __pyx_v_gap_extension_penalty};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_ref_aligner, __pyx_v_read_seq, __pyx_v_gap_open_penalty, __pyx_v_gap_extension_penalty};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_ref_aligner);
    __Pyx_GIVEREF(__pyx_v_ref_aligner);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_ref_aligner);
    __Pyx_INCREF(__pyx_v_read_seq);
    __Pyx_GIVEREF(__pyx_v_read_seq);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_read_seq);
    __Pyx_INCREF(__pyx_v_gap_open_penalty);
    __Pyx_GIVEREF(__pyx_v_gap_open_penalty);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_v_gap_open_penalty);
    __Pyx_INCREF(__pyx_v_gap_extension_penalty);
    __Pyx_GIVEREF(__pyx_v_gap_extension_penalty);
    PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_6, __pyx_v_gap_extension_penalty);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ref_aln = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "aldy/indelpost/localn.pyx":254
 *     ref_aln = align(ref_aligner, read_seq, gap_open_penalty, gap_extension_penalty)
 *     # forced to align by setting gap_open_penalty=len(read_seq)
 *     mut_aln = align(mut_aligner, read_seq, len(read_seq), gap_extension_penalty)             # <<<<<<<<<<<<<<
 * 
 *     if mut_aln.optimal_score <= ref_aln.optimal_score:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_align); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = PyObject_Length(__pyx_v_read_seq); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 254, __pyx_L1_error)
  __pyx_t_5 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_mut_aligner, __pyx_v_read_seq, __pyx_t_5, __pyx_v_gap_extension_penalty};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 254, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_mut_aligner, __pyx_v_read_seq, __pyx_t_5, __pyx_v_gap_extension_penalty};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 254, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_10 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_mut_aligner);
    __Pyx_GIVEREF(__pyx_v_mut_aligner);
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_6, __pyx_v_mut_aligner);
    __Pyx_INCREF(__pyx_v_read_seq);
    __Pyx_GIVEREF(__pyx_v_read_seq);
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_6, __pyx_v_read_seq);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_6, __pyx_t_5);
    __Pyx_INCREF(__pyx_v_gap_extension_penalty);
    __Pyx_GIVEREF(__pyx_v_gap_extension_penalty);
    PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_6, __pyx_v_gap_extension_penalty);
    __pyx_t_5 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mut_aln = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "aldy/indelpost/localn.pyx":256
 *     mut_aln = align(mut_aligner, read_seq, len(read_seq), gap_extension_penalty)
 * 
 *     if mut_aln.optimal_score <= ref_aln.optimal_score:             # <<<<<<<<<<<<<<
 *         read["is_target"] = False
 *         return read
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_mut_aln, __pyx_n_s_optimal_score); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_ref_aln, __pyx_n_s_optimal_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = PyObject_RichCompare(__pyx_t_4, __pyx_t_1, Py_LE); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (__pyx_t_2) {

    /* "aldy/indelpost/localn.pyx":257
 * 
 *     if mut_aln.optimal_score <= ref_aln.optimal_score:
 *         read["is_target"] = False             # <<<<<<<<<<<<<<
 *         return read
 * 
 */
    if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_is_target, Py_False) < 0)) __PYX_ERR(0, 257, __pyx_L1_error)

    /* "aldy/indelpost/localn.pyx":258
 *     if mut_aln.optimal_score <= ref_aln.optimal_score:
 *         read["is_target"] = False
 *         return read             # <<<<<<<<<<<<<<
 * 
 *     mut_aln_cigar = mut_aln.CIGAR
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_read);
    __pyx_r = __pyx_v_read;
    goto __pyx_L0;

    /* "aldy/indelpost/localn.pyx":256
 *     mut_aln = align(mut_aligner, read_seq, len(read_seq), gap_extension_penalty)
 * 
 *     if mut_aln.optimal_score <= ref_aln.optimal_score:             # <<<<<<<<<<<<<<
 *         read["is_target"] = False
 *         return read
 */
  }

  /* "aldy/indelpost/localn.pyx":260
 *         return read
 * 
 *     mut_aln_cigar = mut_aln.CIGAR             # <<<<<<<<<<<<<<
 *     mut_aln_ref_start = mut_aln.reference_start
 *     mut_aln_ref_end = mut_aln.reference_end
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_mut_aln, __pyx_n_s_CIGAR); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_v_mut_aln_cigar = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "aldy/indelpost/localn.pyx":261
 * 
 *     mut_aln_cigar = mut_aln.CIGAR
 *     mut_aln_ref_start = mut_aln.reference_start             # <<<<<<<<<<<<<<
 *     mut_aln_ref_end = mut_aln.reference_end
 *     mut_aln_read_start = mut_aln.read_start
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_mut_aln, __pyx_n_s_reference_start); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_v_mut_aln_ref_start = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "aldy/indelpost/localn.pyx":262
 *     mut_aln_cigar = mut_aln.CIGAR
 *     mut_aln_ref_start = mut_aln.reference_start
 *     mut_aln_ref_end = mut_aln.reference_end             # <<<<<<<<<<<<<<
 *     mut_aln_read_start = mut_aln.read_start
 *     mut_aln_read_end = mut_aln.read_end
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_mut_aln, __pyx_n_s_reference_end); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_v_mut_aln_ref_end = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "aldy/indelpost/localn.pyx":263
 *     mut_aln_ref_start = mut_aln.reference_start
 *     mut_aln_ref_end = mut_aln.reference_end
 *     mut_aln_read_start = mut_aln.read_start             # <<<<<<<<<<<<<<
 *     mut_aln_read_end = mut_aln.read_end
 *     mut_cigar_list = cigar_ptrn.findall(mut_aln.CIGAR)
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_mut_aln, __pyx_n_s_read_start); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_v_mut_aln_read_start = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "aldy/indelpost/localn.pyx":264
 *     mut_aln_ref_end = mut_aln.reference_end
 *     mut_aln_read_start = mut_aln.read_start
 *     mut_aln_read_end = mut_aln.read_end             # <<<<<<<<<<<<<<
 *     mut_cigar_list = cigar_ptrn.findall(mut_aln.CIGAR)
 * 
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_mut_aln, __pyx_n_s_read_end); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_v_mut_aln_read_end = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "aldy/indelpost/localn.pyx":265
 *     mut_aln_read_start = mut_aln.read_start
 *     mut_aln_read_end = mut_aln.read_end
 *     mut_cigar_list = cigar_ptrn.findall(mut_aln.CIGAR)             # <<<<<<<<<<<<<<
 * 
 *     target_covered = is_covering_target(
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_cigar_ptrn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_findall); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mut_aln, __pyx_n_s_CIGAR); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_10 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_mut_cigar_list = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "aldy/indelpost/localn.pyx":268
 * 
 *     target_covered = is_covering_target(
 *                     read["read_name"],             # <<<<<<<<<<<<<<
 *                     read_seq,
 *                     target_indel.indel_seq,
 */
  __pyx_t_10 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_name); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_10))||((__pyx_t_10) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_10)->tp_name), 0))) __PYX_ERR(0, 268, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":269
 *     target_covered = is_covering_target(
 *                     read["read_name"],
 *                     read_seq,             # <<<<<<<<<<<<<<
 *                     target_indel.indel_seq,
 *                     mut_ref_lt,
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_read_seq))||((__pyx_v_read_seq) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_read_seq)->tp_name), 0))) __PYX_ERR(0, 269, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":270
 *                     read["read_name"],
 *                     read_seq,
 *                     target_indel.indel_seq,             # <<<<<<<<<<<<<<
 *                     mut_ref_lt,
 *                     mut_ref_mid,
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_target_indel, __pyx_n_s_indel_seq); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 270, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":271
 *                     read_seq,
 *                     target_indel.indel_seq,
 *                     mut_ref_lt,             # <<<<<<<<<<<<<<
 *                     mut_ref_mid,
 *                     mut_ref_rt,
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_mut_ref_lt))||((__pyx_v_mut_ref_lt) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_mut_ref_lt)->tp_name), 0))) __PYX_ERR(0, 271, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":272
 *                     target_indel.indel_seq,
 *                     mut_ref_lt,
 *                     mut_ref_mid,             # <<<<<<<<<<<<<<
 *                     mut_ref_rt,
 *                     mut_aln_cigar,
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_mut_ref_mid))||((__pyx_v_mut_ref_mid) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_mut_ref_mid)->tp_name), 0))) __PYX_ERR(0, 272, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":273
 *                     mut_ref_lt,
 *                     mut_ref_mid,
 *                     mut_ref_rt,             # <<<<<<<<<<<<<<
 *                     mut_aln_cigar,
 *                     len(read_seq),
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_mut_ref_rt))||((__pyx_v_mut_ref_rt) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_mut_ref_rt)->tp_name), 0))) __PYX_ERR(0, 273, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":274
 *                     mut_ref_mid,
 *                     mut_ref_rt,
 *                     mut_aln_cigar,             # <<<<<<<<<<<<<<
 *                     len(read_seq),
 *                     mut_aln_ref_start,
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_mut_aln_cigar))||((__pyx_v_mut_aln_cigar) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_mut_aln_cigar)->tp_name), 0))) __PYX_ERR(0, 274, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":275
 *                     mut_ref_rt,
 *                     mut_aln_cigar,
 *                     len(read_seq),             # <<<<<<<<<<<<<<
 *                     mut_aln_ref_start,
 *                     mut_aln_ref_end,
 */
  __pyx_t_9 = PyObject_Length(__pyx_v_read_seq); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 275, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":276
 *                     mut_aln_cigar,
 *                     len(read_seq),
 *                     mut_aln_ref_start,             # <<<<<<<<<<<<<<
 *                     mut_aln_ref_end,
 *                     mut_aln_read_start,
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_mut_aln_ref_start); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 276, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":277
 *                     len(read_seq),
 *                     mut_aln_ref_start,
 *                     mut_aln_ref_end,             # <<<<<<<<<<<<<<
 *                     mut_aln_read_start,
 *                     mut_aln_read_end,
 */
  __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_v_mut_aln_ref_end); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 277, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":278
 *                     mut_aln_ref_start,
 *                     mut_aln_ref_end,
 *                     mut_aln_read_start,             # <<<<<<<<<<<<<<
 *                     mut_aln_read_end,
 *                     target_indel.count_repeats(),
 */
  __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_mut_aln_read_start); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 278, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":279
 *                     mut_aln_ref_end,
 *                     mut_aln_read_start,
 *                     mut_aln_read_end,             # <<<<<<<<<<<<<<
 *                     target_indel.count_repeats(),
 *                    )
 */
  __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_v_mut_aln_read_end); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 279, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":280
 *                     mut_aln_read_start,
 *                     mut_aln_read_end,
 *                     target_indel.count_repeats(),             # <<<<<<<<<<<<<<
 *                    )
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_target_indel, __pyx_n_s_count_repeats); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_14 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_14 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":267
 *     mut_cigar_list = cigar_ptrn.findall(mut_aln.CIGAR)
 * 
 *     target_covered = is_covering_target(             # <<<<<<<<<<<<<<
 *                     read["read_name"],
 *                     read_seq,
 */
  __pyx_v_target_covered = __pyx_f_4aldy_9indelpost_6localn_is_covering_target(((PyObject*)__pyx_t_10), ((PyObject*)__pyx_v_read_seq), ((PyObject*)__pyx_t_4), ((PyObject*)__pyx_v_mut_ref_lt), ((PyObject*)__pyx_v_mut_ref_mid), ((PyObject*)__pyx_v_mut_ref_rt), ((PyObject*)__pyx_v_mut_aln_cigar), __pyx_t_9, __pyx_t_6, __pyx_t_11, __pyx_t_12, __pyx_t_13, __pyx_t_14);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "aldy/indelpost/localn.pyx":283
 *                    )
 * 
 *     if target_covered == 1:             # <<<<<<<<<<<<<<
 *         read["is_target"] = True
 *         return read
 */
  switch (__pyx_v_target_covered) {
    case 1:

    /* "aldy/indelpost/localn.pyx":284
 * 
 *     if target_covered == 1:
 *         read["is_target"] = True             # <<<<<<<<<<<<<<
 *         return read
 *     elif target_covered == -1:
 */
    if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_is_target, Py_True) < 0)) __PYX_ERR(0, 284, __pyx_L1_error)

    /* "aldy/indelpost/localn.pyx":285
 *     if target_covered == 1:
 *         read["is_target"] = True
 *         return read             # <<<<<<<<<<<<<<
 *     elif target_covered == -1:
 *         read["undetermined"] = True
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_read);
    __pyx_r = __pyx_v_read;
    goto __pyx_L0;

    /* "aldy/indelpost/localn.pyx":283
 *                    )
 * 
 *     if target_covered == 1:             # <<<<<<<<<<<<<<
 *         read["is_target"] = True
 *         return read
 */
    break;
    case -1L:

    /* "aldy/indelpost/localn.pyx":287
 *         return read
 *     elif target_covered == -1:
 *         read["undetermined"] = True             # <<<<<<<<<<<<<<
 *         return read
 *     else:
 */
    if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_undetermined, Py_True) < 0)) __PYX_ERR(0, 287, __pyx_L1_error)

    /* "aldy/indelpost/localn.pyx":288
 *     elif target_covered == -1:
 *         read["undetermined"] = True
 *         return read             # <<<<<<<<<<<<<<
 *     else:
 *         return read
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_read);
    __pyx_r = __pyx_v_read;
    goto __pyx_L0;

    /* "aldy/indelpost/localn.pyx":286
 *         read["is_target"] = True
 *         return read
 *     elif target_covered == -1:             # <<<<<<<<<<<<<<
 *         read["undetermined"] = True
 *         return read
 */
    break;
    default:

    /* "aldy/indelpost/localn.pyx":290
 *         return read
 *     else:
 *         return read             # <<<<<<<<<<<<<<
 * 
 * cdef int is_covering_target(
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_read);
    __pyx_r = __pyx_v_read;
    goto __pyx_L0;
    break;
  }

  /* "aldy/indelpost/localn.pyx":222
 *     return False
 * 
 * def is_target_by_ssw(             # <<<<<<<<<<<<<<
 *     read,
 *     target_indel,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("aldy.indelpost.localn.is_target_by_ssw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_read_seq);
  __Pyx_XDECREF(__pyx_v_ref_aln);
  __Pyx_XDECREF(__pyx_v_mut_aln);
  __Pyx_XDECREF(__pyx_v_mut_aln_cigar);
  __Pyx_XDECREF(__pyx_v_mut_aln_ref_start);
  __Pyx_XDECREF(__pyx_v_mut_aln_ref_end);
  __Pyx_XDECREF(__pyx_v_mut_aln_read_start);
  __Pyx_XDECREF(__pyx_v_mut_aln_read_end);
  __Pyx_XDECREF(__pyx_v_mut_cigar_list);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/localn.pyx":292
 *         return read
 * 
 * cdef int is_covering_target(             # <<<<<<<<<<<<<<
 *     str readname,
 *     str read_seq,
 */

static int __pyx_f_4aldy_9indelpost_6localn_is_covering_target(CYTHON_UNUSED PyObject *__pyx_v_readname, PyObject *__pyx_v_read_seq, PyObject *__pyx_v_indel_seq, PyObject *__pyx_v_mut_ref_lt, PyObject *__pyx_v_mut_ref_mid, PyObject *__pyx_v_mut_ref_rt, PyObject *__pyx_v_mut_aln_cigar, int __pyx_v_read_seq_len, int __pyx_v_ref_aln_start, int __pyx_v_ref_aln_end, int __pyx_v_read_aln_start, int __pyx_v_read_aln_end, int __pyx_v_n_repeats) {
  PyObject *__pyx_v_mut_cigar_list = 0;
  PyObject *__pyx_v_repeat_unit = 0;
  CYTHON_UNUSED int __pyx_v_mapped_len;
  int __pyx_v_mut_ref_lt_len;
  int __pyx_v_mut_ref_mid_len;
  CYTHON_UNUSED int __pyx_v_mut_ref_rt_len;
  int __pyx_v_lt_consumption;
  int __pyx_v_rt_consumption;
  int __pyx_v_total_consumption;
  int __pyx_v_is_lt_read_consumed;
  int __pyx_v_is_rt_read_consumed;
  PyObject *__pyx_v_lt_read = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_mid_read = NULL;
  PyObject *__pyx_v_rt_read = NULL;
  PyObject *__pyx_v_lt_check = NULL;
  PyObject *__pyx_v_rt_check = NULL;
  PyObject *__pyx_v_consumed_mid_seq = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_covering_target", 0);

  /* "aldy/indelpost/localn.pyx":307
 *     int n_repeats,
 * ):
 *     cdef list mut_cigar_list = cigar_ptrn.findall(mut_aln_cigar)             # <<<<<<<<<<<<<<
 * 
 *     if len(mut_cigar_list) > 1:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_cigar_ptrn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_findall); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_mut_aln_cigar) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_mut_aln_cigar);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 307, __pyx_L1_error)
  __pyx_v_mut_cigar_list = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":309
 *     cdef list mut_cigar_list = cigar_ptrn.findall(mut_aln_cigar)
 * 
 *     if len(mut_cigar_list) > 1:             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  if (unlikely(__pyx_v_mut_cigar_list == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 309, __pyx_L1_error)
  }
  __pyx_t_4 = PyList_GET_SIZE(__pyx_v_mut_cigar_list); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 309, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 1) != 0);
  if (__pyx_t_5) {

    /* "aldy/indelpost/localn.pyx":310
 * 
 *     if len(mut_cigar_list) > 1:
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cdef str repeat_unit = to_minimal_repeat_unit(indel_seq)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "aldy/indelpost/localn.pyx":309
 *     cdef list mut_cigar_list = cigar_ptrn.findall(mut_aln_cigar)
 * 
 *     if len(mut_cigar_list) > 1:             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "aldy/indelpost/localn.pyx":312
 *         return False
 * 
 *     cdef str repeat_unit = to_minimal_repeat_unit(indel_seq)             # <<<<<<<<<<<<<<
 *     cdef int mapped_len = int(mut_cigar_list[0][:-1])
 *     cdef int mut_ref_lt_len = len(mut_ref_lt)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_to_minimal_repeat_unit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_indel_seq) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_indel_seq);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 312, __pyx_L1_error)
  __pyx_v_repeat_unit = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":313
 * 
 *     cdef str repeat_unit = to_minimal_repeat_unit(indel_seq)
 *     cdef int mapped_len = int(mut_cigar_list[0][:-1])             # <<<<<<<<<<<<<<
 *     cdef int mut_ref_lt_len = len(mut_ref_lt)
 *     cdef int mut_ref_mid_len = len(mut_ref_mid)
 */
  if (unlikely(__pyx_v_mut_cigar_list == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 313, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_mut_cigar_list, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, -1L, NULL, NULL, &__pyx_slice__2, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mapped_len = __pyx_t_6;

  /* "aldy/indelpost/localn.pyx":314
 *     cdef str repeat_unit = to_minimal_repeat_unit(indel_seq)
 *     cdef int mapped_len = int(mut_cigar_list[0][:-1])
 *     cdef int mut_ref_lt_len = len(mut_ref_lt)             # <<<<<<<<<<<<<<
 *     cdef int mut_ref_mid_len = len(mut_ref_mid)
 *     cdef int mut_ref_rt_len = len(mut_ref_rt)
 */
  if (unlikely(__pyx_v_mut_ref_lt == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 314, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyUnicode_GET_LENGTH(__pyx_v_mut_ref_lt); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 314, __pyx_L1_error)
  __pyx_v_mut_ref_lt_len = __pyx_t_4;

  /* "aldy/indelpost/localn.pyx":315
 *     cdef int mapped_len = int(mut_cigar_list[0][:-1])
 *     cdef int mut_ref_lt_len = len(mut_ref_lt)
 *     cdef int mut_ref_mid_len = len(mut_ref_mid)             # <<<<<<<<<<<<<<
 *     cdef int mut_ref_rt_len = len(mut_ref_rt)
 *     cdef int lt_consumption
 */
  if (unlikely(__pyx_v_mut_ref_mid == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 315, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyUnicode_GET_LENGTH(__pyx_v_mut_ref_mid); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 315, __pyx_L1_error)
  __pyx_v_mut_ref_mid_len = __pyx_t_4;

  /* "aldy/indelpost/localn.pyx":316
 *     cdef int mut_ref_lt_len = len(mut_ref_lt)
 *     cdef int mut_ref_mid_len = len(mut_ref_mid)
 *     cdef int mut_ref_rt_len = len(mut_ref_rt)             # <<<<<<<<<<<<<<
 *     cdef int lt_consumption
 *     cdef int rt_consumption
 */
  if (unlikely(__pyx_v_mut_ref_rt == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 316, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyUnicode_GET_LENGTH(__pyx_v_mut_ref_rt); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 316, __pyx_L1_error)
  __pyx_v_mut_ref_rt_len = __pyx_t_4;

  /* "aldy/indelpost/localn.pyx":319
 *     cdef int lt_consumption
 *     cdef int rt_consumption
 *     cdef int total_consumption = read_aln_end - read_aln_start + 1             # <<<<<<<<<<<<<<
 *     cdef bint is_lt_indel_end_reached
 *     cdef bint is_rt_indel_end_reached
 */
  __pyx_v_total_consumption = ((__pyx_v_read_aln_end - __pyx_v_read_aln_start) + 1);

  /* "aldy/indelpost/localn.pyx":322
 *     cdef bint is_lt_indel_end_reached
 *     cdef bint is_rt_indel_end_reached
 *     cdef bint is_lt_read_consumed = (read_aln_start == 0)             # <<<<<<<<<<<<<<
 *     cdef bint is_rt_read_consumed = (read_aln_end == read_seq_len - 1)
 * 
 */
  __pyx_v_is_lt_read_consumed = (__pyx_v_read_aln_start == 0);

  /* "aldy/indelpost/localn.pyx":323
 *     cdef bint is_rt_indel_end_reached
 *     cdef bint is_lt_read_consumed = (read_aln_start == 0)
 *     cdef bint is_rt_read_consumed = (read_aln_end == read_seq_len - 1)             # <<<<<<<<<<<<<<
 * 
 *     # not covering indel
 */
  __pyx_v_is_rt_read_consumed = (__pyx_v_read_aln_end == (__pyx_v_read_seq_len - 1));

  /* "aldy/indelpost/localn.pyx":326
 * 
 *     # not covering indel
 *     if ref_aln_end < mut_ref_lt_len:             # <<<<<<<<<<<<<<
 *         return False
 *     elif mut_ref_lt_len + mut_ref_mid_len <= ref_aln_start:
 */
  __pyx_t_5 = ((__pyx_v_ref_aln_end < __pyx_v_mut_ref_lt_len) != 0);
  if (__pyx_t_5) {

    /* "aldy/indelpost/localn.pyx":327
 *     # not covering indel
 *     if ref_aln_end < mut_ref_lt_len:
 *         return False             # <<<<<<<<<<<<<<
 *     elif mut_ref_lt_len + mut_ref_mid_len <= ref_aln_start:
 *         return False
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "aldy/indelpost/localn.pyx":326
 * 
 *     # not covering indel
 *     if ref_aln_end < mut_ref_lt_len:             # <<<<<<<<<<<<<<
 *         return False
 *     elif mut_ref_lt_len + mut_ref_mid_len <= ref_aln_start:
 */
  }

  /* "aldy/indelpost/localn.pyx":328
 *     if ref_aln_end < mut_ref_lt_len:
 *         return False
 *     elif mut_ref_lt_len + mut_ref_mid_len <= ref_aln_start:             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_5 = (((__pyx_v_mut_ref_lt_len + __pyx_v_mut_ref_mid_len) <= __pyx_v_ref_aln_start) != 0);
  if (__pyx_t_5) {

    /* "aldy/indelpost/localn.pyx":329
 *         return False
 *     elif mut_ref_lt_len + mut_ref_mid_len <= ref_aln_start:
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     # ins
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "aldy/indelpost/localn.pyx":328
 *     if ref_aln_end < mut_ref_lt_len:
 *         return False
 *     elif mut_ref_lt_len + mut_ref_mid_len <= ref_aln_start:             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "aldy/indelpost/localn.pyx":332
 * 
 *     # ins
 *     if mut_ref_mid_len:             # <<<<<<<<<<<<<<
 *         if ref_aln_start < mut_ref_lt_len:
 *             lt_consumption = mut_ref_lt_len - ref_aln_start
 */
  __pyx_t_5 = (__pyx_v_mut_ref_mid_len != 0);
  if (__pyx_t_5) {

    /* "aldy/indelpost/localn.pyx":333
 *     # ins
 *     if mut_ref_mid_len:
 *         if ref_aln_start < mut_ref_lt_len:             # <<<<<<<<<<<<<<
 *             lt_consumption = mut_ref_lt_len - ref_aln_start
 * 
 */
    __pyx_t_5 = ((__pyx_v_ref_aln_start < __pyx_v_mut_ref_lt_len) != 0);
    if (__pyx_t_5) {

      /* "aldy/indelpost/localn.pyx":334
 *     if mut_ref_mid_len:
 *         if ref_aln_start < mut_ref_lt_len:
 *             lt_consumption = mut_ref_lt_len - ref_aln_start             # <<<<<<<<<<<<<<
 * 
 *             if total_consumption > lt_consumption + mut_ref_mid_len:
 */
      __pyx_v_lt_consumption = (__pyx_v_mut_ref_lt_len - __pyx_v_ref_aln_start);

      /* "aldy/indelpost/localn.pyx":336
 *             lt_consumption = mut_ref_lt_len - ref_aln_start
 * 
 *             if total_consumption > lt_consumption + mut_ref_mid_len:             # <<<<<<<<<<<<<<
 *                 lt_read = read_seq[read_aln_start : read_aln_start + lt_consumption]
 *                 mid_read = read_seq[read_aln_start + lt_consumption : read_aln_start + lt_consumption + mut_ref_mid_len]
 */
      __pyx_t_5 = ((__pyx_v_total_consumption > (__pyx_v_lt_consumption + __pyx_v_mut_ref_mid_len)) != 0);
      if (__pyx_t_5) {

        /* "aldy/indelpost/localn.pyx":337
 * 
 *             if total_consumption > lt_consumption + mut_ref_mid_len:
 *                 lt_read = read_seq[read_aln_start : read_aln_start + lt_consumption]             # <<<<<<<<<<<<<<
 *                 mid_read = read_seq[read_aln_start + lt_consumption : read_aln_start + lt_consumption + mut_ref_mid_len]
 *                 rt_read = read_seq[read_aln_start + lt_consumption + mut_ref_mid_len : read_aln_end + 1]
 */
        if (unlikely(__pyx_v_read_seq == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 337, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_PyUnicode_Substring(__pyx_v_read_seq, __pyx_v_read_aln_start, (__pyx_v_read_aln_start + __pyx_v_lt_consumption)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_v_lt_read = ((PyObject*)__pyx_t_1);
        __pyx_t_1 = 0;

        /* "aldy/indelpost/localn.pyx":338
 *             if total_consumption > lt_consumption + mut_ref_mid_len:
 *                 lt_read = read_seq[read_aln_start : read_aln_start + lt_consumption]
 *                 mid_read = read_seq[read_aln_start + lt_consumption : read_aln_start + lt_consumption + mut_ref_mid_len]             # <<<<<<<<<<<<<<
 *                 rt_read = read_seq[read_aln_start + lt_consumption + mut_ref_mid_len : read_aln_end + 1]
 * 
 */
        if (unlikely(__pyx_v_read_seq == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 338, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_PyUnicode_Substring(__pyx_v_read_seq, (__pyx_v_read_aln_start + __pyx_v_lt_consumption), ((__pyx_v_read_aln_start + __pyx_v_lt_consumption) + __pyx_v_mut_ref_mid_len)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_v_mid_read = ((PyObject*)__pyx_t_1);
        __pyx_t_1 = 0;

        /* "aldy/indelpost/localn.pyx":339
 *                 lt_read = read_seq[read_aln_start : read_aln_start + lt_consumption]
 *                 mid_read = read_seq[read_aln_start + lt_consumption : read_aln_start + lt_consumption + mut_ref_mid_len]
 *                 rt_read = read_seq[read_aln_start + lt_consumption + mut_ref_mid_len : read_aln_end + 1]             # <<<<<<<<<<<<<<
 * 
 *                 lt_check = is_compatible_repeats(lt_read, repeat_unit, n_repeats, is_left=True)
 */
        if (unlikely(__pyx_v_read_seq == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 339, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_PyUnicode_Substring(__pyx_v_read_seq, ((__pyx_v_read_aln_start + __pyx_v_lt_consumption) + __pyx_v_mut_ref_mid_len), (__pyx_v_read_aln_end + 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 339, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_v_rt_read = ((PyObject*)__pyx_t_1);
        __pyx_t_1 = 0;

        /* "aldy/indelpost/localn.pyx":341
 *                 rt_read = read_seq[read_aln_start + lt_consumption + mut_ref_mid_len : read_aln_end + 1]
 * 
 *                 lt_check = is_compatible_repeats(lt_read, repeat_unit, n_repeats, is_left=True)             # <<<<<<<<<<<<<<
 *                 rt_check = is_compatible_repeats(rt_read, repeat_unit, n_repeats, is_left=False)
 * 
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_is_compatible_repeats); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_repeats); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 341, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 341, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_v_lt_read);
        __Pyx_GIVEREF(__pyx_v_lt_read);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_lt_read);
        __Pyx_INCREF(__pyx_v_repeat_unit);
        __Pyx_GIVEREF(__pyx_v_repeat_unit);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_repeat_unit);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 341, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_is_left, Py_True) < 0) __PYX_ERR(0, 341, __pyx_L1_error)
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 341, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_lt_check = __pyx_t_7;
        __pyx_t_7 = 0;

        /* "aldy/indelpost/localn.pyx":342
 * 
 *                 lt_check = is_compatible_repeats(lt_read, repeat_unit, n_repeats, is_left=True)
 *                 rt_check = is_compatible_repeats(rt_read, repeat_unit, n_repeats, is_left=False)             # <<<<<<<<<<<<<<
 * 
 *                 if lt_check and rt_check:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_is_compatible_repeats); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 342, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_repeats); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 342, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 342, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_v_rt_read);
        __Pyx_GIVEREF(__pyx_v_rt_read);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_rt_read);
        __Pyx_INCREF(__pyx_v_repeat_unit);
        __Pyx_GIVEREF(__pyx_v_repeat_unit);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_repeat_unit);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 342, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_is_left, Py_False) < 0) __PYX_ERR(0, 342, __pyx_L1_error)
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_rt_check = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "aldy/indelpost/localn.pyx":344
 *                 rt_check = is_compatible_repeats(rt_read, repeat_unit, n_repeats, is_left=False)
 * 
 *                 if lt_check and rt_check:             # <<<<<<<<<<<<<<
 *                     #True
 *                     return 1
 */
        __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_lt_check); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 344, __pyx_L1_error)
        if (__pyx_t_8) {
        } else {
          __pyx_t_5 = __pyx_t_8;
          goto __pyx_L9_bool_binop_done;
        }
        __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_rt_check); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 344, __pyx_L1_error)
        __pyx_t_5 = __pyx_t_8;
        __pyx_L9_bool_binop_done:;
        if (__pyx_t_5) {

          /* "aldy/indelpost/localn.pyx":346
 *                 if lt_check and rt_check:
 *                     #True
 *                     return 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     #False (undetermined)
 */
          __pyx_r = 1;
          goto __pyx_L0;

          /* "aldy/indelpost/localn.pyx":344
 *                 rt_check = is_compatible_repeats(rt_read, repeat_unit, n_repeats, is_left=False)
 * 
 *                 if lt_check and rt_check:             # <<<<<<<<<<<<<<
 *                     #True
 *                     return 1
 */
        }

        /* "aldy/indelpost/localn.pyx":349
 *                 else:
 *                     #False (undetermined)
 *                     return -1             # <<<<<<<<<<<<<<
 *             else:
 *                 ### abolish heuristic for reapeat
 */
        /*else*/ {
          __pyx_r = -1;
          goto __pyx_L0;
        }

        /* "aldy/indelpost/localn.pyx":336
 *             lt_consumption = mut_ref_lt_len - ref_aln_start
 * 
 *             if total_consumption > lt_consumption + mut_ref_mid_len:             # <<<<<<<<<<<<<<
 *                 lt_read = read_seq[read_aln_start : read_aln_start + lt_consumption]
 *                 mid_read = read_seq[read_aln_start + lt_consumption : read_aln_start + lt_consumption + mut_ref_mid_len]
 */
      }

      /* "aldy/indelpost/localn.pyx":354
 *                 #if n_repeats:
 *                 #    return False
 *                 if is_rt_read_consumed:             # <<<<<<<<<<<<<<
 *                         #True
 *                         return 1
 */
      /*else*/ {
        __pyx_t_5 = (__pyx_v_is_rt_read_consumed != 0);
        if (__pyx_t_5) {

          /* "aldy/indelpost/localn.pyx":356
 *                 if is_rt_read_consumed:
 *                         #True
 *                         return 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     # aligned from left and ended in inserted seq
 */
          __pyx_r = 1;
          goto __pyx_L0;

          /* "aldy/indelpost/localn.pyx":354
 *                 #if n_repeats:
 *                 #    return False
 *                 if is_rt_read_consumed:             # <<<<<<<<<<<<<<
 *                         #True
 *                         return 1
 */
        }

        /* "aldy/indelpost/localn.pyx":359
 *                 else:
 *                     # aligned from left and ended in inserted seq
 *                     rt_consumption = total_consumption - lt_consumption             # <<<<<<<<<<<<<<
 *                     consumed_mid_seq = mut_ref_mid[:rt_consumption:]
 * 
 */
        /*else*/ {
          __pyx_v_rt_consumption = (__pyx_v_total_consumption - __pyx_v_lt_consumption);

          /* "aldy/indelpost/localn.pyx":360
 *                     # aligned from left and ended in inserted seq
 *                     rt_consumption = total_consumption - lt_consumption
 *                     consumed_mid_seq = mut_ref_mid[:rt_consumption:]             # <<<<<<<<<<<<<<
 * 
 *                     if consumed_mid_seq == read_seq[-rt_consumption:]:
 */
          __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_rt_consumption); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 360, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = PySlice_New(Py_None, __pyx_t_1, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 360, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_mut_ref_mid, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 360, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_v_consumed_mid_seq = ((PyObject*)__pyx_t_1);
          __pyx_t_1 = 0;

          /* "aldy/indelpost/localn.pyx":362
 *                     consumed_mid_seq = mut_ref_mid[:rt_consumption:]
 * 
 *                     if consumed_mid_seq == read_seq[-rt_consumption:]:             # <<<<<<<<<<<<<<
 *                         #True
 *                         return 1
 */
          if (unlikely(__pyx_v_read_seq == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 362, __pyx_L1_error)
          }
          __pyx_t_1 = __Pyx_PyUnicode_Substring(__pyx_v_read_seq, (-__pyx_v_rt_consumption), PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_consumed_mid_seq, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_8 = (__pyx_t_5 != 0);
          if (__pyx_t_8) {

            /* "aldy/indelpost/localn.pyx":364
 *                     if consumed_mid_seq == read_seq[-rt_consumption:]:
 *                         #True
 *                         return 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         #False
 */
            __pyx_r = 1;
            goto __pyx_L0;

            /* "aldy/indelpost/localn.pyx":362
 *                     consumed_mid_seq = mut_ref_mid[:rt_consumption:]
 * 
 *                     if consumed_mid_seq == read_seq[-rt_consumption:]:             # <<<<<<<<<<<<<<
 *                         #True
 *                         return 1
 */
          }

          /* "aldy/indelpost/localn.pyx":367
 *                     else:
 *                         #False
 *                         return 0             # <<<<<<<<<<<<<<
 *         else:
 *             # no left side alignment
 */
          /*else*/ {
            __pyx_r = 0;
            goto __pyx_L0;
          }
        }
      }

      /* "aldy/indelpost/localn.pyx":333
 *     # ins
 *     if mut_ref_mid_len:
 *         if ref_aln_start < mut_ref_lt_len:             # <<<<<<<<<<<<<<
 *             lt_consumption = mut_ref_lt_len - ref_aln_start
 * 
 */
    }

    /* "aldy/indelpost/localn.pyx":376
 *             #    return False
 * 
 *             if is_lt_read_consumed:             # <<<<<<<<<<<<<<
 *                 #True
 *                 return 1
 */
    /*else*/ {
      __pyx_t_8 = (__pyx_v_is_lt_read_consumed != 0);
      if (__pyx_t_8) {

        /* "aldy/indelpost/localn.pyx":378
 *             if is_lt_read_consumed:
 *                 #True
 *                 return 1             # <<<<<<<<<<<<<<
 *             else:
 *                 # aligned from right and ended in inserted seq
 */
        __pyx_r = 1;
        goto __pyx_L0;

        /* "aldy/indelpost/localn.pyx":376
 *             #    return False
 * 
 *             if is_lt_read_consumed:             # <<<<<<<<<<<<<<
 *                 #True
 *                 return 1
 */
      }

      /* "aldy/indelpost/localn.pyx":381
 *             else:
 *                 # aligned from right and ended in inserted seq
 *                 lt_consumption = mut_ref_lt_len + mut_ref_mid_len - ref_aln_start             # <<<<<<<<<<<<<<
 *                 consumed_mid_seq = mut_ref_mid[-lt_consumption:]
 * 
 */
      /*else*/ {
        __pyx_v_lt_consumption = ((__pyx_v_mut_ref_lt_len + __pyx_v_mut_ref_mid_len) - __pyx_v_ref_aln_start);

        /* "aldy/indelpost/localn.pyx":382
 *                 # aligned from right and ended in inserted seq
 *                 lt_consumption = mut_ref_lt_len + mut_ref_mid_len - ref_aln_start
 *                 consumed_mid_seq = mut_ref_mid[-lt_consumption:]             # <<<<<<<<<<<<<<
 * 
 *                 # require exact match
 */
        if (unlikely(__pyx_v_mut_ref_mid == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 382, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_PyUnicode_Substring(__pyx_v_mut_ref_mid, (-__pyx_v_lt_consumption), PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_v_consumed_mid_seq = ((PyObject*)__pyx_t_1);
        __pyx_t_1 = 0;

        /* "aldy/indelpost/localn.pyx":385
 * 
 *                 # require exact match
 *                 if consumed_mid_seq == read_seq[:lt_consumption]:             # <<<<<<<<<<<<<<
 *                     #True
 *                     return 1
 */
        if (unlikely(__pyx_v_read_seq == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 385, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_PyUnicode_Substring(__pyx_v_read_seq, 0, __pyx_v_lt_consumption); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_v_consumed_mid_seq, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 385, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_5 = (__pyx_t_8 != 0);
        if (__pyx_t_5) {

          /* "aldy/indelpost/localn.pyx":387
 *                 if consumed_mid_seq == read_seq[:lt_consumption]:
 *                     #True
 *                     return 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     #False
 */
          __pyx_r = 1;
          goto __pyx_L0;

          /* "aldy/indelpost/localn.pyx":385
 * 
 *                 # require exact match
 *                 if consumed_mid_seq == read_seq[:lt_consumption]:             # <<<<<<<<<<<<<<
 *                     #True
 *                     return 1
 */
        }

        /* "aldy/indelpost/localn.pyx":390
 *                 else:
 *                     #False
 *                     return 0             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
        /*else*/ {
          __pyx_r = 0;
          goto __pyx_L0;
        }
      }
    }

    /* "aldy/indelpost/localn.pyx":332
 * 
 *     # ins
 *     if mut_ref_mid_len:             # <<<<<<<<<<<<<<
 *         if ref_aln_start < mut_ref_lt_len:
 *             lt_consumption = mut_ref_lt_len - ref_aln_start
 */
  }

  /* "aldy/indelpost/localn.pyx":393
 * 
 *     else:
 *         lt_consumption = mut_ref_lt_len - ref_aln_start             # <<<<<<<<<<<<<<
 *         rt_consumption = total_consumption - lt_consumption
 * 
 */
  /*else*/ {
    __pyx_v_lt_consumption = (__pyx_v_mut_ref_lt_len - __pyx_v_ref_aln_start);

    /* "aldy/indelpost/localn.pyx":394
 *     else:
 *         lt_consumption = mut_ref_lt_len - ref_aln_start
 *         rt_consumption = total_consumption - lt_consumption             # <<<<<<<<<<<<<<
 * 
 *         lt_read = read_seq[read_aln_start : read_aln_start + lt_consumption]
 */
    __pyx_v_rt_consumption = (__pyx_v_total_consumption - __pyx_v_lt_consumption);

    /* "aldy/indelpost/localn.pyx":396
 *         rt_consumption = total_consumption - lt_consumption
 * 
 *         lt_read = read_seq[read_aln_start : read_aln_start + lt_consumption]             # <<<<<<<<<<<<<<
 *         rt_read = read_seq[read_aln_start + lt_consumption  : read_aln_end]
 * 
 */
    if (unlikely(__pyx_v_read_seq == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 396, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyUnicode_Substring(__pyx_v_read_seq, __pyx_v_read_aln_start, (__pyx_v_read_aln_start + __pyx_v_lt_consumption)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_lt_read = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "aldy/indelpost/localn.pyx":397
 * 
 *         lt_read = read_seq[read_aln_start : read_aln_start + lt_consumption]
 *         rt_read = read_seq[read_aln_start + lt_consumption  : read_aln_end]             # <<<<<<<<<<<<<<
 * 
 *         lt_check = is_compatible_repeats(lt_read, repeat_unit, n_repeats, is_left=True)
 */
    if (unlikely(__pyx_v_read_seq == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 397, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyUnicode_Substring(__pyx_v_read_seq, (__pyx_v_read_aln_start + __pyx_v_lt_consumption), __pyx_v_read_aln_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_rt_read = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "aldy/indelpost/localn.pyx":399
 *         rt_read = read_seq[read_aln_start + lt_consumption  : read_aln_end]
 * 
 *         lt_check = is_compatible_repeats(lt_read, repeat_unit, n_repeats, is_left=True)             # <<<<<<<<<<<<<<
 *         rt_check = is_compatible_repeats(rt_read, repeat_unit, n_repeats, is_left=False)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_is_compatible_repeats); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_repeats); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_lt_read);
    __Pyx_GIVEREF(__pyx_v_lt_read);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_lt_read);
    __Pyx_INCREF(__pyx_v_repeat_unit);
    __Pyx_GIVEREF(__pyx_v_repeat_unit);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_repeat_unit);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_is_left, Py_True) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_lt_check = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "aldy/indelpost/localn.pyx":400
 * 
 *         lt_check = is_compatible_repeats(lt_read, repeat_unit, n_repeats, is_left=True)
 *         rt_check = is_compatible_repeats(rt_read, repeat_unit, n_repeats, is_left=False)             # <<<<<<<<<<<<<<
 * 
 *         if lt_check and rt_check:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_is_compatible_repeats); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_repeats); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_rt_read);
    __Pyx_GIVEREF(__pyx_v_rt_read);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_rt_read);
    __Pyx_INCREF(__pyx_v_repeat_unit);
    __Pyx_GIVEREF(__pyx_v_repeat_unit);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_repeat_unit);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_is_left, Py_False) < 0) __PYX_ERR(0, 400, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_rt_check = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "aldy/indelpost/localn.pyx":402
 *         rt_check = is_compatible_repeats(rt_read, repeat_unit, n_repeats, is_left=False)
 * 
 *         if lt_check and rt_check:             # <<<<<<<<<<<<<<
 *             pass
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_lt_check); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 402, __pyx_L1_error)
    if (__pyx_t_8) {
    } else {
      __pyx_t_5 = __pyx_t_8;
      goto __pyx_L16_bool_binop_done;
    }
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_rt_check); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 402, __pyx_L1_error)
    __pyx_t_5 = __pyx_t_8;
    __pyx_L16_bool_binop_done:;
    if (__pyx_t_5) {
      goto __pyx_L15;
    }

    /* "aldy/indelpost/localn.pyx":406
 *         else:
 *             #False (undetermined)
 *             return -1             # <<<<<<<<<<<<<<
 * 
 *         if lt_consumption <= rt_consumption:
 */
    /*else*/ {
      __pyx_r = -1;
      goto __pyx_L0;
    }
    __pyx_L15:;

    /* "aldy/indelpost/localn.pyx":408
 *             return -1
 * 
 *         if lt_consumption <= rt_consumption:             # <<<<<<<<<<<<<<
 *             if is_lt_read_consumed:
 *                 #True
 */
    __pyx_t_5 = ((__pyx_v_lt_consumption <= __pyx_v_rt_consumption) != 0);
    if (__pyx_t_5) {

      /* "aldy/indelpost/localn.pyx":409
 * 
 *         if lt_consumption <= rt_consumption:
 *             if is_lt_read_consumed:             # <<<<<<<<<<<<<<
 *                 #True
 *                 return 1
 */
      __pyx_t_5 = (__pyx_v_is_lt_read_consumed != 0);
      if (__pyx_t_5) {

        /* "aldy/indelpost/localn.pyx":411
 *             if is_lt_read_consumed:
 *                 #True
 *                 return 1             # <<<<<<<<<<<<<<
 *             else:
 *                 if lt_consumption > 2:
 */
        __pyx_r = 1;
        goto __pyx_L0;

        /* "aldy/indelpost/localn.pyx":409
 * 
 *         if lt_consumption <= rt_consumption:
 *             if is_lt_read_consumed:             # <<<<<<<<<<<<<<
 *                 #True
 *                 return 1
 */
      }

      /* "aldy/indelpost/localn.pyx":413
 *                 return 1
 *             else:
 *                 if lt_consumption > 2:             # <<<<<<<<<<<<<<
 *                     #True
 *                     return 1
 */
      /*else*/ {
        __pyx_t_5 = ((__pyx_v_lt_consumption > 2) != 0);
        if (__pyx_t_5) {

          /* "aldy/indelpost/localn.pyx":415
 *                 if lt_consumption > 2:
 *                     #True
 *                     return 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     #False
 */
          __pyx_r = 1;
          goto __pyx_L0;

          /* "aldy/indelpost/localn.pyx":413
 *                 return 1
 *             else:
 *                 if lt_consumption > 2:             # <<<<<<<<<<<<<<
 *                     #True
 *                     return 1
 */
        }

        /* "aldy/indelpost/localn.pyx":418
 *                 else:
 *                     #False
 *                     return 0             # <<<<<<<<<<<<<<
 *         else:
 *             if is_rt_read_consumed:
 */
        /*else*/ {
          __pyx_r = 0;
          goto __pyx_L0;
        }
      }

      /* "aldy/indelpost/localn.pyx":408
 *             return -1
 * 
 *         if lt_consumption <= rt_consumption:             # <<<<<<<<<<<<<<
 *             if is_lt_read_consumed:
 *                 #True
 */
    }

    /* "aldy/indelpost/localn.pyx":420
 *                     return 0
 *         else:
 *             if is_rt_read_consumed:             # <<<<<<<<<<<<<<
 *                 #True
 *                 return 1
 */
    /*else*/ {
      __pyx_t_5 = (__pyx_v_is_rt_read_consumed != 0);
      if (__pyx_t_5) {

        /* "aldy/indelpost/localn.pyx":422
 *             if is_rt_read_consumed:
 *                 #True
 *                 return 1             # <<<<<<<<<<<<<<
 *             else:
 *                 if rt_consumption > 2:
 */
        __pyx_r = 1;
        goto __pyx_L0;

        /* "aldy/indelpost/localn.pyx":420
 *                     return 0
 *         else:
 *             if is_rt_read_consumed:             # <<<<<<<<<<<<<<
 *                 #True
 *                 return 1
 */
      }

      /* "aldy/indelpost/localn.pyx":424
 *                 return 1
 *             else:
 *                 if rt_consumption > 2:             # <<<<<<<<<<<<<<
 *                     #True
 *                     return 1
 */
      /*else*/ {
        __pyx_t_5 = ((__pyx_v_rt_consumption > 2) != 0);
        if (__pyx_t_5) {

          /* "aldy/indelpost/localn.pyx":426
 *                 if rt_consumption > 2:
 *                     #True
 *                     return 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     #False
 */
          __pyx_r = 1;
          goto __pyx_L0;

          /* "aldy/indelpost/localn.pyx":424
 *                 return 1
 *             else:
 *                 if rt_consumption > 2:             # <<<<<<<<<<<<<<
 *                     #True
 *                     return 1
 */
        }

        /* "aldy/indelpost/localn.pyx":429
 *                 else:
 *                     #False
 *                     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
        /*else*/ {
          __pyx_r = 0;
          goto __pyx_L0;
        }
      }
    }
  }

  /* "aldy/indelpost/localn.pyx":292
 *         return read
 * 
 * cdef int is_covering_target(             # <<<<<<<<<<<<<<
 *     str readname,
 *     str read_seq,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_WriteUnraisable("aldy.indelpost.localn.is_covering_target", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_mut_cigar_list);
  __Pyx_XDECREF(__pyx_v_repeat_unit);
  __Pyx_XDECREF(__pyx_v_lt_read);
  __Pyx_XDECREF(__pyx_v_mid_read);
  __Pyx_XDECREF(__pyx_v_rt_read);
  __Pyx_XDECREF(__pyx_v_lt_check);
  __Pyx_XDECREF(__pyx_v_rt_check);
  __Pyx_XDECREF(__pyx_v_consumed_mid_seq);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/localn.pyx":432
 * 
 * 
 * def is_compatible_repeats(seq, repeat_unit, expected_n_repeats, is_left):             # <<<<<<<<<<<<<<
 *     unit_len = len(repeat_unit)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_9is_compatible_repeats(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6localn_9is_compatible_repeats = {"is_compatible_repeats", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6localn_9is_compatible_repeats, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_9is_compatible_repeats(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_seq = 0;
  PyObject *__pyx_v_repeat_unit = 0;
  PyObject *__pyx_v_expected_n_repeats = 0;
  PyObject *__pyx_v_is_left = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_compatible_repeats (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_seq,&__pyx_n_s_repeat_unit,&__pyx_n_s_expected_n_repeats,&__pyx_n_s_is_left,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seq)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_repeat_unit)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_compatible_repeats", 1, 4, 4, 1); __PYX_ERR(0, 432, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expected_n_repeats)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_compatible_repeats", 1, 4, 4, 2); __PYX_ERR(0, 432, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_is_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("is_compatible_repeats", 1, 4, 4, 3); __PYX_ERR(0, 432, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "is_compatible_repeats") < 0)) __PYX_ERR(0, 432, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_seq = values[0];
    __pyx_v_repeat_unit = values[1];
    __pyx_v_expected_n_repeats = values[2];
    __pyx_v_is_left = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_compatible_repeats", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 432, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.localn.is_compatible_repeats", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6localn_8is_compatible_repeats(__pyx_self, __pyx_v_seq, __pyx_v_repeat_unit, __pyx_v_expected_n_repeats, __pyx_v_is_left);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6localn_8is_compatible_repeats(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_seq, PyObject *__pyx_v_repeat_unit, PyObject *__pyx_v_expected_n_repeats, PyObject *__pyx_v_is_left) {
  Py_ssize_t __pyx_v_unit_len;
  PyObject *__pyx_v_cnt = NULL;
  int __pyx_v_is_done;
  int __pyx_v_is_compatible;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_compatible_repeats", 0);
  __Pyx_INCREF(__pyx_v_seq);
  __Pyx_INCREF(__pyx_v_repeat_unit);

  /* "aldy/indelpost/localn.pyx":433
 * 
 * def is_compatible_repeats(seq, repeat_unit, expected_n_repeats, is_left):
 *     unit_len = len(repeat_unit)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_repeat_unit); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 433, __pyx_L1_error)
  __pyx_v_unit_len = __pyx_t_1;

  /* "aldy/indelpost/localn.pyx":436
 * 
 * 
 *     if is_left:             # <<<<<<<<<<<<<<
 *         seq = seq[::-1]
 *         repeat_unit = repeat_unit[::-1]
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_is_left); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 436, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "aldy/indelpost/localn.pyx":437
 * 
 *     if is_left:
 *         seq = seq[::-1]             # <<<<<<<<<<<<<<
 *         repeat_unit = repeat_unit[::-1]
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_seq, __pyx_slice__4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_seq, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "aldy/indelpost/localn.pyx":438
 *     if is_left:
 *         seq = seq[::-1]
 *         repeat_unit = repeat_unit[::-1]             # <<<<<<<<<<<<<<
 * 
 *     cnt = 0
 */
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_repeat_unit, __pyx_slice__4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_repeat_unit, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "aldy/indelpost/localn.pyx":436
 * 
 * 
 *     if is_left:             # <<<<<<<<<<<<<<
 *         seq = seq[::-1]
 *         repeat_unit = repeat_unit[::-1]
 */
  }

  /* "aldy/indelpost/localn.pyx":440
 *         repeat_unit = repeat_unit[::-1]
 * 
 *     cnt = 0             # <<<<<<<<<<<<<<
 *     is_done = False
 *     while not is_done:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_cnt = __pyx_int_0;

  /* "aldy/indelpost/localn.pyx":441
 * 
 *     cnt = 0
 *     is_done = False             # <<<<<<<<<<<<<<
 *     while not is_done:
 *         if not seq:
 */
  __pyx_v_is_done = 0;

  /* "aldy/indelpost/localn.pyx":442
 *     cnt = 0
 *     is_done = False
 *     while not is_done:             # <<<<<<<<<<<<<<
 *         if not seq:
 *             is_done = True
 */
  while (1) {
    __pyx_t_2 = ((!(__pyx_v_is_done != 0)) != 0);
    if (!__pyx_t_2) break;

    /* "aldy/indelpost/localn.pyx":443
 *     is_done = False
 *     while not is_done:
 *         if not seq:             # <<<<<<<<<<<<<<
 *             is_done = True
 *         elif repeat_unit == seq[: unit_len]:
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_seq); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 443, __pyx_L1_error)
    __pyx_t_4 = ((!__pyx_t_2) != 0);
    if (__pyx_t_4) {

      /* "aldy/indelpost/localn.pyx":444
 *     while not is_done:
 *         if not seq:
 *             is_done = True             # <<<<<<<<<<<<<<
 *         elif repeat_unit == seq[: unit_len]:
 *             seq = seq[unit_len :]
 */
      __pyx_v_is_done = 1;

      /* "aldy/indelpost/localn.pyx":443
 *     is_done = False
 *     while not is_done:
 *         if not seq:             # <<<<<<<<<<<<<<
 *             is_done = True
 *         elif repeat_unit == seq[: unit_len]:
 */
      goto __pyx_L6;
    }

    /* "aldy/indelpost/localn.pyx":445
 *         if not seq:
 *             is_done = True
 *         elif repeat_unit == seq[: unit_len]:             # <<<<<<<<<<<<<<
 *             seq = seq[unit_len :]
 *             cnt += 1
 */
    __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_seq, 0, __pyx_v_unit_len, NULL, NULL, NULL, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_repeat_unit, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 445, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 445, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_4) {

      /* "aldy/indelpost/localn.pyx":446
 *             is_done = True
 *         elif repeat_unit == seq[: unit_len]:
 *             seq = seq[unit_len :]             # <<<<<<<<<<<<<<
 *             cnt += 1
 *         else:
 */
      __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_seq, __pyx_v_unit_len, 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 446, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_seq, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "aldy/indelpost/localn.pyx":447
 *         elif repeat_unit == seq[: unit_len]:
 *             seq = seq[unit_len :]
 *             cnt += 1             # <<<<<<<<<<<<<<
 *         else:
 *             is_done = True
 */
      __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_v_cnt, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_cnt, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "aldy/indelpost/localn.pyx":445
 *         if not seq:
 *             is_done = True
 *         elif repeat_unit == seq[: unit_len]:             # <<<<<<<<<<<<<<
 *             seq = seq[unit_len :]
 *             cnt += 1
 */
      goto __pyx_L6;
    }

    /* "aldy/indelpost/localn.pyx":449
 *             cnt += 1
 *         else:
 *             is_done = True             # <<<<<<<<<<<<<<
 * 
 *     is_compatible = True
 */
    /*else*/ {
      __pyx_v_is_done = 1;
    }
    __pyx_L6:;
  }

  /* "aldy/indelpost/localn.pyx":451
 *             is_done = True
 * 
 *     is_compatible = True             # <<<<<<<<<<<<<<
 *     if not seq:
 *         is_compatible = False
 */
  __pyx_v_is_compatible = 1;

  /* "aldy/indelpost/localn.pyx":452
 * 
 *     is_compatible = True
 *     if not seq:             # <<<<<<<<<<<<<<
 *         is_compatible = False
 *     else:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_seq); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 452, __pyx_L1_error)
  __pyx_t_2 = ((!__pyx_t_4) != 0);
  if (__pyx_t_2) {

    /* "aldy/indelpost/localn.pyx":453
 *     is_compatible = True
 *     if not seq:
 *         is_compatible = False             # <<<<<<<<<<<<<<
 *     else:
 *         if cnt and cnt != expected_n_repeats:
 */
    __pyx_v_is_compatible = 0;

    /* "aldy/indelpost/localn.pyx":452
 * 
 *     is_compatible = True
 *     if not seq:             # <<<<<<<<<<<<<<
 *         is_compatible = False
 *     else:
 */
    goto __pyx_L7;
  }

  /* "aldy/indelpost/localn.pyx":455
 *         is_compatible = False
 *     else:
 *         if cnt and cnt != expected_n_repeats:             # <<<<<<<<<<<<<<
 *             is_compatible = False
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_cnt); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 455, __pyx_L1_error)
    if (__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_cnt, __pyx_v_expected_n_repeats, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 455, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 455, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_2 = __pyx_t_4;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_2) {

      /* "aldy/indelpost/localn.pyx":456
 *     else:
 *         if cnt and cnt != expected_n_repeats:
 *             is_compatible = False             # <<<<<<<<<<<<<<
 * 
 *     return is_compatible
 */
      __pyx_v_is_compatible = 0;

      /* "aldy/indelpost/localn.pyx":455
 *         is_compatible = False
 *     else:
 *         if cnt and cnt != expected_n_repeats:             # <<<<<<<<<<<<<<
 *             is_compatible = False
 * 
 */
    }
  }
  __pyx_L7:;

  /* "aldy/indelpost/localn.pyx":458
 *             is_compatible = False
 * 
 *     return is_compatible             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_is_compatible); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "aldy/indelpost/localn.pyx":432
 * 
 * 
 * def is_compatible_repeats(seq, repeat_unit, expected_n_repeats, is_left):             # <<<<<<<<<<<<<<
 *     unit_len = len(repeat_unit)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("aldy.indelpost.localn.is_compatible_repeats", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cnt);
  __Pyx_XDECREF(__pyx_v_seq);
  __Pyx_XDECREF(__pyx_v_repeat_unit);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/localn.pyx":463
 * 
 * 
 * def make_aligner(ref_seq, match_score, mismatch_penalty):             # <<<<<<<<<<<<<<
 *     aligner = SSW(match_score=match_score, mismatch_penalty=mismatch_penalty)
 *     aligner.setReference(ref_seq)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_11make_aligner(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6localn_11make_aligner = {"make_aligner", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6localn_11make_aligner, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_11make_aligner(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ref_seq = 0;
  PyObject *__pyx_v_match_score = 0;
  PyObject *__pyx_v_mismatch_penalty = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_aligner (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ref_seq,&__pyx_n_s_match_score,&__pyx_n_s_mismatch_penalty,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ref_seq)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_match_score)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("make_aligner", 1, 3, 3, 1); __PYX_ERR(0, 463, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mismatch_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("make_aligner", 1, 3, 3, 2); __PYX_ERR(0, 463, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "make_aligner") < 0)) __PYX_ERR(0, 463, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_ref_seq = values[0];
    __pyx_v_match_score = values[1];
    __pyx_v_mismatch_penalty = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("make_aligner", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 463, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.localn.make_aligner", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6localn_10make_aligner(__pyx_self, __pyx_v_ref_seq, __pyx_v_match_score, __pyx_v_mismatch_penalty);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6localn_10make_aligner(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ref_seq, PyObject *__pyx_v_match_score, PyObject *__pyx_v_mismatch_penalty) {
  PyObject *__pyx_v_aligner = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("make_aligner", 0);

  /* "aldy/indelpost/localn.pyx":464
 * 
 * def make_aligner(ref_seq, match_score, mismatch_penalty):
 *     aligner = SSW(match_score=match_score, mismatch_penalty=mismatch_penalty)             # <<<<<<<<<<<<<<
 *     aligner.setReference(ref_seq)
 *     return aligner
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_SSW); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_match_score, __pyx_v_match_score) < 0) __PYX_ERR(0, 464, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_mismatch_penalty, __pyx_v_mismatch_penalty) < 0) __PYX_ERR(0, 464, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_aligner = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "aldy/indelpost/localn.pyx":465
 * def make_aligner(ref_seq, match_score, mismatch_penalty):
 *     aligner = SSW(match_score=match_score, mismatch_penalty=mismatch_penalty)
 *     aligner.setReference(ref_seq)             # <<<<<<<<<<<<<<
 *     return aligner
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_aligner, __pyx_n_s_setReference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_v_ref_seq) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_ref_seq);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "aldy/indelpost/localn.pyx":466
 *     aligner = SSW(match_score=match_score, mismatch_penalty=mismatch_penalty)
 *     aligner.setReference(ref_seq)
 *     return aligner             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_aligner);
  __pyx_r = __pyx_v_aligner;
  goto __pyx_L0;

  /* "aldy/indelpost/localn.pyx":463
 * 
 * 
 * def make_aligner(ref_seq, match_score, mismatch_penalty):             # <<<<<<<<<<<<<<
 *     aligner = SSW(match_score=match_score, mismatch_penalty=mismatch_penalty)
 *     aligner.setReference(ref_seq)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("aldy.indelpost.localn.make_aligner", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_aligner);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/localn.pyx":469
 * 
 * 
 * def align(aligner, read_seq, gap_open_penalty, gap_extension_penalty):             # <<<<<<<<<<<<<<
 *     aligner.setRead(read_seq)
 *     return aligner.align(gap_open=gap_open_penalty, gap_extension=gap_extension_penalty)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_13align(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6localn_13align = {"align", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6localn_13align, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_13align(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_aligner = 0;
  PyObject *__pyx_v_read_seq = 0;
  PyObject *__pyx_v_gap_open_penalty = 0;
  PyObject *__pyx_v_gap_extension_penalty = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("align (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_aligner,&__pyx_n_s_read_seq,&__pyx_n_s_gap_open_penalty,&__pyx_n_s_gap_extension_penalty,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aligner)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_read_seq)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("align", 1, 4, 4, 1); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gap_open_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("align", 1, 4, 4, 2); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_gap_extension_penalty)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("align", 1, 4, 4, 3); __PYX_ERR(0, 469, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "align") < 0)) __PYX_ERR(0, 469, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_aligner = values[0];
    __pyx_v_read_seq = values[1];
    __pyx_v_gap_open_penalty = values[2];
    __pyx_v_gap_extension_penalty = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("align", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 469, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.localn.align", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6localn_12align(__pyx_self, __pyx_v_aligner, __pyx_v_read_seq, __pyx_v_gap_open_penalty, __pyx_v_gap_extension_penalty);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6localn_12align(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_aligner, PyObject *__pyx_v_read_seq, PyObject *__pyx_v_gap_open_penalty, PyObject *__pyx_v_gap_extension_penalty) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("align", 0);

  /* "aldy/indelpost/localn.pyx":470
 * 
 * def align(aligner, read_seq, gap_open_penalty, gap_extension_penalty):
 *     aligner.setRead(read_seq)             # <<<<<<<<<<<<<<
 *     return aligner.align(gap_open=gap_open_penalty, gap_extension=gap_extension_penalty)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_aligner, __pyx_n_s_setRead); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_read_seq) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_read_seq);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":471
 * def align(aligner, read_seq, gap_open_penalty, gap_extension_penalty):
 *     aligner.setRead(read_seq)
 *     return aligner.align(gap_open=gap_open_penalty, gap_extension=gap_extension_penalty)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_aligner, __pyx_n_s_align); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_gap_open, __pyx_v_gap_open_penalty) < 0) __PYX_ERR(0, 471, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_gap_extension, __pyx_v_gap_extension_penalty) < 0) __PYX_ERR(0, 471, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "aldy/indelpost/localn.pyx":469
 * 
 * 
 * def align(aligner, read_seq, gap_open_penalty, gap_extension_penalty):             # <<<<<<<<<<<<<<
 *     aligner.setRead(read_seq)
 *     return aligner.align(gap_open=gap_open_penalty, gap_extension=gap_extension_penalty)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("aldy.indelpost.localn.align", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/localn.pyx":474
 * 
 * 
 * def parse_read_by_mut_aln(mut_aln, contig, read, indel_type):             # <<<<<<<<<<<<<<
 *     """Decompose read based on the SSW alignment result for evaluation
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_15parse_read_by_mut_aln(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4aldy_9indelpost_6localn_14parse_read_by_mut_aln[] = "Decompose read based on the SSW alignment result for evaluation\n\n    Args:\n        mut_aln (named tuple): Alignment(CIGAR, optimal_score, sub_optimal_score, reference_start, reference_end, read_start, read_end)\n        template (dict): indel template\n        read (dict): dictized read\n        indel_type (str): \"I\" for insertion or \"D\" for deletion\n    Returns:\n        read (dict): dictized read with read seq decomposed\n    ";
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6localn_15parse_read_by_mut_aln = {"parse_read_by_mut_aln", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6localn_15parse_read_by_mut_aln, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4aldy_9indelpost_6localn_14parse_read_by_mut_aln};
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_15parse_read_by_mut_aln(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_mut_aln = 0;
  PyObject *__pyx_v_contig = 0;
  PyObject *__pyx_v_read = 0;
  PyObject *__pyx_v_indel_type = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parse_read_by_mut_aln (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mut_aln,&__pyx_n_s_contig,&__pyx_n_s_read,&__pyx_n_s_indel_type,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mut_aln)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_contig)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parse_read_by_mut_aln", 1, 4, 4, 1); __PYX_ERR(0, 474, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_read)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parse_read_by_mut_aln", 1, 4, 4, 2); __PYX_ERR(0, 474, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indel_type)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parse_read_by_mut_aln", 1, 4, 4, 3); __PYX_ERR(0, 474, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parse_read_by_mut_aln") < 0)) __PYX_ERR(0, 474, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_mut_aln = values[0];
    __pyx_v_contig = values[1];
    __pyx_v_read = values[2];
    __pyx_v_indel_type = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parse_read_by_mut_aln", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 474, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.localn.parse_read_by_mut_aln", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6localn_14parse_read_by_mut_aln(__pyx_self, __pyx_v_mut_aln, __pyx_v_contig, __pyx_v_read, __pyx_v_indel_type);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6localn_14parse_read_by_mut_aln(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_mut_aln, PyObject *__pyx_v_contig, PyObject *__pyx_v_read, PyObject *__pyx_v_indel_type) {
  PyObject *__pyx_v_lt_len = NULL;
  PyObject *__pyx_v_indel_len = NULL;
  CYTHON_UNUSED Py_ssize_t __pyx_v_rt_len;
  PyObject *__pyx_v_read_seq = NULL;
  PyObject *__pyx_v_read_qual = NULL;
  PyObject *__pyx_v_ref_start = NULL;
  PyObject *__pyx_v_ref_end = NULL;
  PyObject *__pyx_v_aln_start = NULL;
  PyObject *__pyx_v_aln_end = NULL;
  PyObject *__pyx_v_lt_flank = NULL;
  PyObject *__pyx_v_mid_seq = NULL;
  PyObject *__pyx_v_rt_flank = NULL;
  PyObject *__pyx_v_lt_qual = NULL;
  PyObject *__pyx_v_rt_qual = NULL;
  PyObject *__pyx_v_lt_diff = NULL;
  PyObject *__pyx_v_end_point = NULL;
  PyObject *__pyx_v_del_pos = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_lt_ref = NULL;
  PyObject *__pyx_v_rt_ref = NULL;
  PyObject *__pyx_v_rt_diff = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse_read_by_mut_aln", 0);

  /* "aldy/indelpost/localn.pyx":486
 *     """
 *     lt_len, indel_len, rt_len = (
 *         len(contig.lt_consensus_seq),             # <<<<<<<<<<<<<<
 *         len(contig.indel_seq),
 *         len(contig.rt_consensus_seq),
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_contig, __pyx_n_s_lt_consensus_seq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "aldy/indelpost/localn.pyx":487
 *     lt_len, indel_len, rt_len = (
 *         len(contig.lt_consensus_seq),
 *         len(contig.indel_seq),             # <<<<<<<<<<<<<<
 *         len(contig.rt_consensus_seq),
 *     )
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_contig, __pyx_n_s_indel_seq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "aldy/indelpost/localn.pyx":488
 *         len(contig.lt_consensus_seq),
 *         len(contig.indel_seq),
 *         len(contig.rt_consensus_seq),             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_contig, __pyx_n_s_rt_consensus_seq); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_lt_len = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_indel_len = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_rt_len = __pyx_t_2;

  /* "aldy/indelpost/localn.pyx":491
 *     )
 * 
 *     read_seq = read["read_seq"]             # <<<<<<<<<<<<<<
 *     read_qual = read["read_qual"]
 *     ref_start, ref_end = mut_aln.reference_start, mut_aln.reference_end
 */
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_seq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_read_seq = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "aldy/indelpost/localn.pyx":492
 * 
 *     read_seq = read["read_seq"]
 *     read_qual = read["read_qual"]             # <<<<<<<<<<<<<<
 *     ref_start, ref_end = mut_aln.reference_start, mut_aln.reference_end
 *     aln_start, aln_end = mut_aln.read_start, mut_aln.read_end
 */
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_qual); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_read_qual = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "aldy/indelpost/localn.pyx":493
 *     read_seq = read["read_seq"]
 *     read_qual = read["read_qual"]
 *     ref_start, ref_end = mut_aln.reference_start, mut_aln.reference_end             # <<<<<<<<<<<<<<
 *     aln_start, aln_end = mut_aln.read_start, mut_aln.read_end
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_mut_aln, __pyx_n_s_reference_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mut_aln, __pyx_n_s_reference_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ref_start = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_ref_end = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":494
 *     read_qual = read["read_qual"]
 *     ref_start, ref_end = mut_aln.reference_start, mut_aln.reference_end
 *     aln_start, aln_end = mut_aln.read_start, mut_aln.read_end             # <<<<<<<<<<<<<<
 * 
 *     lt_flank, mid_seq, rt_flank = "", "", ""
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mut_aln, __pyx_n_s_read_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_mut_aln, __pyx_n_s_read_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_aln_start = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_aln_end = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "aldy/indelpost/localn.pyx":496
 *     aln_start, aln_end = mut_aln.read_start, mut_aln.read_end
 * 
 *     lt_flank, mid_seq, rt_flank = "", "", ""             # <<<<<<<<<<<<<<
 *     lt_qual, rt_qual = [], []
 * 
 */
  __pyx_t_3 = __pyx_kp_u__3;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __pyx_kp_u__3;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_4 = __pyx_kp_u__3;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_lt_flank = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_mid_seq = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_rt_flank = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "aldy/indelpost/localn.pyx":497
 * 
 *     lt_flank, mid_seq, rt_flank = "", "", ""
 *     lt_qual, rt_qual = [], []             # <<<<<<<<<<<<<<
 * 
 *     if ref_start <= lt_len:
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_lt_qual = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_rt_qual = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":499
 *     lt_qual, rt_qual = [], []
 * 
 *     if ref_start <= lt_len:             # <<<<<<<<<<<<<<
 *         lt_diff = lt_len - ref_start
 *         lt_flank = read_seq[aln_start : aln_start + lt_diff]
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_ref_start, __pyx_v_lt_len, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 499, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "aldy/indelpost/localn.pyx":500
 * 
 *     if ref_start <= lt_len:
 *         lt_diff = lt_len - ref_start             # <<<<<<<<<<<<<<
 *         lt_flank = read_seq[aln_start : aln_start + lt_diff]
 *         lt_qual = read_qual[aln_start : aln_start + lt_diff]
 */
    __pyx_t_1 = PyNumber_Subtract(__pyx_v_lt_len, __pyx_v_ref_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_lt_diff = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "aldy/indelpost/localn.pyx":501
 *     if ref_start <= lt_len:
 *         lt_diff = lt_len - ref_start
 *         lt_flank = read_seq[aln_start : aln_start + lt_diff]             # <<<<<<<<<<<<<<
 *         lt_qual = read_qual[aln_start : aln_start + lt_diff]
 *         if indel_type == "I":
 */
    __pyx_t_1 = PyNumber_Add(__pyx_v_aln_start, __pyx_v_lt_diff); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, &__pyx_v_aln_start, &__pyx_t_1, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_lt_flank, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "aldy/indelpost/localn.pyx":502
 *         lt_diff = lt_len - ref_start
 *         lt_flank = read_seq[aln_start : aln_start + lt_diff]
 *         lt_qual = read_qual[aln_start : aln_start + lt_diff]             # <<<<<<<<<<<<<<
 *         if indel_type == "I":
 *             end_point = min(aln_start + lt_diff + indel_len, aln_end)
 */
    __pyx_t_4 = PyNumber_Add(__pyx_v_aln_start, __pyx_v_lt_diff); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_read_qual, 0, 0, &__pyx_v_aln_start, &__pyx_t_4, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_lt_qual, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "aldy/indelpost/localn.pyx":503
 *         lt_flank = read_seq[aln_start : aln_start + lt_diff]
 *         lt_qual = read_qual[aln_start : aln_start + lt_diff]
 *         if indel_type == "I":             # <<<<<<<<<<<<<<
 *             end_point = min(aln_start + lt_diff + indel_len, aln_end)
 *             mid_seq = read_seq[aln_start + lt_diff : end_point]
 */
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_indel_type, __pyx_n_u_I, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 503, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "aldy/indelpost/localn.pyx":504
 *         lt_qual = read_qual[aln_start : aln_start + lt_diff]
 *         if indel_type == "I":
 *             end_point = min(aln_start + lt_diff + indel_len, aln_end)             # <<<<<<<<<<<<<<
 *             mid_seq = read_seq[aln_start + lt_diff : end_point]
 *         else:
 */
      __Pyx_INCREF(__pyx_v_aln_end);
      __pyx_t_1 = __pyx_v_aln_end;
      __pyx_t_4 = PyNumber_Add(__pyx_v_aln_start, __pyx_v_lt_diff); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_v_indel_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 504, __pyx_L1_error)
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 504, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (__pyx_t_5) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1;
      } else {
        __Pyx_INCREF(__pyx_t_3);
        __pyx_t_4 = __pyx_t_3;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __pyx_t_4;
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_end_point = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "aldy/indelpost/localn.pyx":505
 *         if indel_type == "I":
 *             end_point = min(aln_start + lt_diff + indel_len, aln_end)
 *             mid_seq = read_seq[aln_start + lt_diff : end_point]             # <<<<<<<<<<<<<<
 *         else:
 *             rt_flank = read_seq[aln_start + lt_diff :]
 */
      __pyx_t_1 = PyNumber_Add(__pyx_v_aln_start, __pyx_v_lt_diff); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 505, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, &__pyx_t_1, &__pyx_v_end_point, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 505, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_mid_seq, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "aldy/indelpost/localn.pyx":503
 *         lt_flank = read_seq[aln_start : aln_start + lt_diff]
 *         lt_qual = read_qual[aln_start : aln_start + lt_diff]
 *         if indel_type == "I":             # <<<<<<<<<<<<<<
 *             end_point = min(aln_start + lt_diff + indel_len, aln_end)
 *             mid_seq = read_seq[aln_start + lt_diff : end_point]
 */
      goto __pyx_L4;
    }

    /* "aldy/indelpost/localn.pyx":507
 *             mid_seq = read_seq[aln_start + lt_diff : end_point]
 *         else:
 *             rt_flank = read_seq[aln_start + lt_diff :]             # <<<<<<<<<<<<<<
 *             rt_qual = read_qual[aln_start + lt_diff :]
 *             del_pos = get_end_pos(
 */
    /*else*/ {
      __pyx_t_4 = PyNumber_Add(__pyx_v_aln_start, __pyx_v_lt_diff); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 507, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, &__pyx_t_4, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 507, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_rt_flank, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "aldy/indelpost/localn.pyx":508
 *         else:
 *             rt_flank = read_seq[aln_start + lt_diff :]
 *             rt_qual = read_qual[aln_start + lt_diff :]             # <<<<<<<<<<<<<<
 *             del_pos = get_end_pos(
 *                 read["read_start"] + aln_start, lt_flank, read["cigar_string"]
 */
      __pyx_t_1 = PyNumber_Add(__pyx_v_aln_start, __pyx_v_lt_diff); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 508, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_read_qual, 0, 0, &__pyx_t_1, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 508, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_rt_qual, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "aldy/indelpost/localn.pyx":509
 *             rt_flank = read_seq[aln_start + lt_diff :]
 *             rt_qual = read_qual[aln_start + lt_diff :]
 *             del_pos = get_end_pos(             # <<<<<<<<<<<<<<
 *                 read["read_start"] + aln_start, lt_flank, read["cigar_string"]
 *             )
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_get_end_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "aldy/indelpost/localn.pyx":510
 *             rt_qual = read_qual[aln_start + lt_diff :]
 *             del_pos = get_end_pos(
 *                 read["read_start"] + aln_start, lt_flank, read["cigar_string"]             # <<<<<<<<<<<<<<
 *             )
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_read_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 510, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = PyNumber_Add(__pyx_t_3, __pyx_v_aln_start); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 510, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 510, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_6, __pyx_v_lt_flank, __pyx_t_3};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 509, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_6, __pyx_v_lt_flank, __pyx_t_3};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 509, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(3+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 509, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_lt_flank);
        __Pyx_GIVEREF(__pyx_v_lt_flank);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_lt_flank);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_3);
        __pyx_t_6 = 0;
        __pyx_t_3 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 509, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_del_pos = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "aldy/indelpost/localn.pyx":514
 * 
 *             lt_ref, rt_ref = split(
 *                 read["ref_seq"],             # <<<<<<<<<<<<<<
 *                 read["cigar_string"],
 *                 del_pos,
 */
      __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_ref_seq); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 514, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "aldy/indelpost/localn.pyx":515
 *             lt_ref, rt_ref = split(
 *                 read["ref_seq"],
 *                 read["cigar_string"],             # <<<<<<<<<<<<<<
 *                 del_pos,
 *                 read["aln_start"],
 */
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_cigar_string); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 515, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 515, __pyx_L1_error)

      /* "aldy/indelpost/localn.pyx":516
 *                 read["ref_seq"],
 *                 read["cigar_string"],
 *                 del_pos,             # <<<<<<<<<<<<<<
 *                 read["aln_start"],
 *                 is_for_ref=True,
 */
      __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_v_del_pos); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 516, __pyx_L1_error)

      /* "aldy/indelpost/localn.pyx":517
 *                 read["cigar_string"],
 *                 del_pos,
 *                 read["aln_start"],             # <<<<<<<<<<<<<<
 *                 is_for_ref=True,
 *                 reverse=False,
 */
      __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_v_read, __pyx_n_u_aln_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 517, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_9); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 517, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "aldy/indelpost/localn.pyx":513
 *             )
 * 
 *             lt_ref, rt_ref = split(             # <<<<<<<<<<<<<<
 *                 read["ref_seq"],
 *                 read["cigar_string"],
 */
      __pyx_t_9 = __pyx_f_4aldy_9indelpost_9utilities_split(__pyx_t_4, ((PyObject*)__pyx_t_1), __pyx_t_8, __pyx_t_10, 1, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 513, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (likely(__pyx_t_9 != Py_None)) {
        PyObject* sequence = __pyx_t_9;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 513, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        #else
        __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 513, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 513, __pyx_L1_error)
      }
      __pyx_v_lt_ref = __pyx_t_1;
      __pyx_t_1 = 0;
      __pyx_v_rt_ref = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "aldy/indelpost/localn.pyx":522
 *             )
 * 
 *             read["del_pos"] = del_pos             # <<<<<<<<<<<<<<
 *             read["del_seq"] = rt_ref[:indel_len]
 * 
 */
      if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_del_pos, __pyx_v_del_pos) < 0)) __PYX_ERR(0, 522, __pyx_L1_error)

      /* "aldy/indelpost/localn.pyx":523
 * 
 *             read["del_pos"] = del_pos
 *             read["del_seq"] = rt_ref[:indel_len]             # <<<<<<<<<<<<<<
 * 
 *     if lt_len + indel_len <= ref_end and indel_type == "I":
 */
      __pyx_t_9 = __Pyx_PyObject_GetSlice(__pyx_v_rt_ref, 0, 0, NULL, &__pyx_v_indel_len, NULL, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 523, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_del_seq, __pyx_t_9) < 0)) __PYX_ERR(0, 523, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __pyx_L4:;

    /* "aldy/indelpost/localn.pyx":499
 *     lt_qual, rt_qual = [], []
 * 
 *     if ref_start <= lt_len:             # <<<<<<<<<<<<<<
 *         lt_diff = lt_len - ref_start
 *         lt_flank = read_seq[aln_start : aln_start + lt_diff]
 */
  }

  /* "aldy/indelpost/localn.pyx":525
 *             read["del_seq"] = rt_ref[:indel_len]
 * 
 *     if lt_len + indel_len <= ref_end and indel_type == "I":             # <<<<<<<<<<<<<<
 *         rt_diff = ref_end - (lt_len + indel_len)
 *         rt_flank = read_seq[aln_end - rt_diff : aln_end]
 */
  __pyx_t_9 = PyNumber_Add(__pyx_v_lt_len, __pyx_v_indel_len); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_9, __pyx_v_ref_end, Py_LE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_11) {
  } else {
    __pyx_t_5 = __pyx_t_11;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_v_indel_type, __pyx_n_u_I, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 525, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_11;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_5) {

    /* "aldy/indelpost/localn.pyx":526
 * 
 *     if lt_len + indel_len <= ref_end and indel_type == "I":
 *         rt_diff = ref_end - (lt_len + indel_len)             # <<<<<<<<<<<<<<
 *         rt_flank = read_seq[aln_end - rt_diff : aln_end]
 *         rt_qual = read_qual[aln_end - rt_diff : aln_end]
 */
    __pyx_t_4 = PyNumber_Add(__pyx_v_lt_len, __pyx_v_indel_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = PyNumber_Subtract(__pyx_v_ref_end, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_rt_diff = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "aldy/indelpost/localn.pyx":527
 *     if lt_len + indel_len <= ref_end and indel_type == "I":
 *         rt_diff = ref_end - (lt_len + indel_len)
 *         rt_flank = read_seq[aln_end - rt_diff : aln_end]             # <<<<<<<<<<<<<<
 *         rt_qual = read_qual[aln_end - rt_diff : aln_end]
 *         end_point = max(aln_start, aln_end - rt_diff - indel_len)
 */
    __pyx_t_9 = PyNumber_Subtract(__pyx_v_aln_end, __pyx_v_rt_diff); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, &__pyx_t_9, &__pyx_v_aln_end, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_rt_flank, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "aldy/indelpost/localn.pyx":528
 *         rt_diff = ref_end - (lt_len + indel_len)
 *         rt_flank = read_seq[aln_end - rt_diff : aln_end]
 *         rt_qual = read_qual[aln_end - rt_diff : aln_end]             # <<<<<<<<<<<<<<
 *         end_point = max(aln_start, aln_end - rt_diff - indel_len)
 *         mid_seq = read_seq[end_point : aln_end - rt_diff]
 */
    __pyx_t_4 = PyNumber_Subtract(__pyx_v_aln_end, __pyx_v_rt_diff); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 528, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = __Pyx_PyObject_GetSlice(__pyx_v_read_qual, 0, 0, &__pyx_t_4, &__pyx_v_aln_end, NULL, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 528, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_rt_qual, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "aldy/indelpost/localn.pyx":529
 *         rt_flank = read_seq[aln_end - rt_diff : aln_end]
 *         rt_qual = read_qual[aln_end - rt_diff : aln_end]
 *         end_point = max(aln_start, aln_end - rt_diff - indel_len)             # <<<<<<<<<<<<<<
 *         mid_seq = read_seq[end_point : aln_end - rt_diff]
 * 
 */
    __pyx_t_9 = PyNumber_Subtract(__pyx_v_aln_end, __pyx_v_rt_diff); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = PyNumber_Subtract(__pyx_t_9, __pyx_v_indel_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_INCREF(__pyx_v_aln_start);
    __pyx_t_9 = __pyx_v_aln_start;
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_4, __pyx_t_9, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 529, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 529, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_5) {
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_1 = __pyx_t_4;
    } else {
      __Pyx_INCREF(__pyx_t_9);
      __pyx_t_1 = __pyx_t_9;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __pyx_t_1;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_end_point, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "aldy/indelpost/localn.pyx":530
 *         rt_qual = read_qual[aln_end - rt_diff : aln_end]
 *         end_point = max(aln_start, aln_end - rt_diff - indel_len)
 *         mid_seq = read_seq[end_point : aln_end - rt_diff]             # <<<<<<<<<<<<<<
 * 
 *     read["lt_flank"] = lt_flank
 */
    __pyx_t_4 = PyNumber_Subtract(__pyx_v_aln_end, __pyx_v_rt_diff); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 530, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, &__pyx_v_end_point, &__pyx_t_4, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 530, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_mid_seq, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "aldy/indelpost/localn.pyx":525
 *             read["del_seq"] = rt_ref[:indel_len]
 * 
 *     if lt_len + indel_len <= ref_end and indel_type == "I":             # <<<<<<<<<<<<<<
 *         rt_diff = ref_end - (lt_len + indel_len)
 *         rt_flank = read_seq[aln_end - rt_diff : aln_end]
 */
  }

  /* "aldy/indelpost/localn.pyx":532
 *         mid_seq = read_seq[end_point : aln_end - rt_diff]
 * 
 *     read["lt_flank"] = lt_flank             # <<<<<<<<<<<<<<
 *     read["lt_qual"] = lt_qual
 *     read["indel_seq"] = mid_seq
 */
  if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_lt_flank, __pyx_v_lt_flank) < 0)) __PYX_ERR(0, 532, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":533
 * 
 *     read["lt_flank"] = lt_flank
 *     read["lt_qual"] = lt_qual             # <<<<<<<<<<<<<<
 *     read["indel_seq"] = mid_seq
 *     read["rt_flank"] = rt_flank
 */
  if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_lt_qual, __pyx_v_lt_qual) < 0)) __PYX_ERR(0, 533, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":534
 *     read["lt_flank"] = lt_flank
 *     read["lt_qual"] = lt_qual
 *     read["indel_seq"] = mid_seq             # <<<<<<<<<<<<<<
 *     read["rt_flank"] = rt_flank
 *     read["rt_qual"] = rt_qual
 */
  if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_indel_seq, __pyx_v_mid_seq) < 0)) __PYX_ERR(0, 534, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":535
 *     read["lt_qual"] = lt_qual
 *     read["indel_seq"] = mid_seq
 *     read["rt_flank"] = rt_flank             # <<<<<<<<<<<<<<
 *     read["rt_qual"] = rt_qual
 * 
 */
  if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_rt_flank, __pyx_v_rt_flank) < 0)) __PYX_ERR(0, 535, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":536
 *     read["indel_seq"] = mid_seq
 *     read["rt_flank"] = rt_flank
 *     read["rt_qual"] = rt_qual             # <<<<<<<<<<<<<<
 * 
 *     return read
 */
  if (unlikely(PyObject_SetItem(__pyx_v_read, __pyx_n_u_rt_qual, __pyx_v_rt_qual) < 0)) __PYX_ERR(0, 536, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":538
 *     read["rt_qual"] = rt_qual
 * 
 *     return read             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_read);
  __pyx_r = __pyx_v_read;
  goto __pyx_L0;

  /* "aldy/indelpost/localn.pyx":474
 * 
 * 
 * def parse_read_by_mut_aln(mut_aln, contig, read, indel_type):             # <<<<<<<<<<<<<<
 *     """Decompose read based on the SSW alignment result for evaluation
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("aldy.indelpost.localn.parse_read_by_mut_aln", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_lt_len);
  __Pyx_XDECREF(__pyx_v_indel_len);
  __Pyx_XDECREF(__pyx_v_read_seq);
  __Pyx_XDECREF(__pyx_v_read_qual);
  __Pyx_XDECREF(__pyx_v_ref_start);
  __Pyx_XDECREF(__pyx_v_ref_end);
  __Pyx_XDECREF(__pyx_v_aln_start);
  __Pyx_XDECREF(__pyx_v_aln_end);
  __Pyx_XDECREF(__pyx_v_lt_flank);
  __Pyx_XDECREF(__pyx_v_mid_seq);
  __Pyx_XDECREF(__pyx_v_rt_flank);
  __Pyx_XDECREF(__pyx_v_lt_qual);
  __Pyx_XDECREF(__pyx_v_rt_qual);
  __Pyx_XDECREF(__pyx_v_lt_diff);
  __Pyx_XDECREF(__pyx_v_end_point);
  __Pyx_XDECREF(__pyx_v_del_pos);
  __Pyx_XDECREF(__pyx_v_lt_ref);
  __Pyx_XDECREF(__pyx_v_rt_ref);
  __Pyx_XDECREF(__pyx_v_rt_diff);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "aldy/indelpost/localn.pyx":541
 * 
 * 
 * def findall_indels(ref_aln, genome_aln_pos, ref_seq, read_seq, report_snvs=False, basequals=None):             # <<<<<<<<<<<<<<
 * 
 *     genome_aln_pos -= 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_17findall_indels(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4aldy_9indelpost_6localn_17findall_indels = {"findall_indels", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_4aldy_9indelpost_6localn_17findall_indels, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4aldy_9indelpost_6localn_17findall_indels(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ref_aln = 0;
  PyObject *__pyx_v_genome_aln_pos = 0;
  PyObject *__pyx_v_ref_seq = 0;
  PyObject *__pyx_v_read_seq = 0;
  PyObject *__pyx_v_report_snvs = 0;
  PyObject *__pyx_v_basequals = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("findall_indels (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ref_aln,&__pyx_n_s_genome_aln_pos,&__pyx_n_s_ref_seq,&__pyx_n_s_read_seq,&__pyx_n_s_report_snvs,&__pyx_n_s_basequals,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    values[4] = ((PyObject *)Py_False);
    values[5] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ref_aln)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_genome_aln_pos)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findall_indels", 0, 4, 6, 1); __PYX_ERR(0, 541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ref_seq)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findall_indels", 0, 4, 6, 2); __PYX_ERR(0, 541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_read_seq)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findall_indels", 0, 4, 6, 3); __PYX_ERR(0, 541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_report_snvs);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_basequals);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "findall_indels") < 0)) __PYX_ERR(0, 541, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_ref_aln = values[0];
    __pyx_v_genome_aln_pos = values[1];
    __pyx_v_ref_seq = values[2];
    __pyx_v_read_seq = values[3];
    __pyx_v_report_snvs = values[4];
    __pyx_v_basequals = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("findall_indels", 0, 4, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 541, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("aldy.indelpost.localn.findall_indels", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4aldy_9indelpost_6localn_16findall_indels(__pyx_self, __pyx_v_ref_aln, __pyx_v_genome_aln_pos, __pyx_v_ref_seq, __pyx_v_read_seq, __pyx_v_report_snvs, __pyx_v_basequals);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4aldy_9indelpost_6localn_16findall_indels(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ref_aln, PyObject *__pyx_v_genome_aln_pos, PyObject *__pyx_v_ref_seq, PyObject *__pyx_v_read_seq, PyObject *__pyx_v_report_snvs, PyObject *__pyx_v_basequals) {
  PyObject *__pyx_v_ref_idx = NULL;
  PyObject *__pyx_v_read_idx = NULL;
  PyObject *__pyx_v_lt_clipped = NULL;
  PyObject *__pyx_v_indels = NULL;
  PyObject *__pyx_v_snvs = NULL;
  PyObject *__pyx_v_token = NULL;
  PyObject *__pyx_v_event = NULL;
  PyObject *__pyx_v_event_len = NULL;
  PyObject *__pyx_v_indel = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_snv = NULL;
  PyObject *__pyx_v_rt_clipped = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("findall_indels", 0);
  __Pyx_INCREF(__pyx_v_genome_aln_pos);

  /* "aldy/indelpost/localn.pyx":543
 * def findall_indels(ref_aln, genome_aln_pos, ref_seq, read_seq, report_snvs=False, basequals=None):
 * 
 *     genome_aln_pos -= 1             # <<<<<<<<<<<<<<
 *     ref_idx = ref_aln.reference_start
 *     read_idx = ref_aln.read_start
 */
  __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_genome_aln_pos, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_genome_aln_pos, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":544
 * 
 *     genome_aln_pos -= 1
 *     ref_idx = ref_aln.reference_start             # <<<<<<<<<<<<<<
 *     read_idx = ref_aln.read_start
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_ref_aln, __pyx_n_s_reference_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ref_idx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":545
 *     genome_aln_pos -= 1
 *     ref_idx = ref_aln.reference_start
 *     read_idx = ref_aln.read_start             # <<<<<<<<<<<<<<
 * 
 *     lt_clipped = read_seq[:read_idx]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_ref_aln, __pyx_n_s_read_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_read_idx = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":547
 *     read_idx = ref_aln.read_start
 * 
 *     lt_clipped = read_seq[:read_idx]             # <<<<<<<<<<<<<<
 * 
 *     indels, snvs = [], []
 */
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, NULL, &__pyx_v_read_idx, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_lt_clipped = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":549
 *     lt_clipped = read_seq[:read_idx]
 * 
 *     indels, snvs = [], []             # <<<<<<<<<<<<<<
 * 
 *     for token in cigar_ptrn.findall(make_insertion_first(ref_aln.CIGAR)):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_indels = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_snvs = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "aldy/indelpost/localn.pyx":551
 *     indels, snvs = [], []
 * 
 *     for token in cigar_ptrn.findall(make_insertion_first(ref_aln.CIGAR)):             # <<<<<<<<<<<<<<
 *         event, event_len = token[-1], int(token[:-1])
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_cigar_ptrn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_findall); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_make_insertion_first); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_ref_aln, __pyx_n_s_CIGAR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 551, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 551, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 551, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 551, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 551, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 551, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_token, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "aldy/indelpost/localn.pyx":552
 * 
 *     for token in cigar_ptrn.findall(make_insertion_first(ref_aln.CIGAR)):
 *         event, event_len = token[-1], int(token[:-1])             # <<<<<<<<<<<<<<
 * 
 *         if event == "I" or event == "D":
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_token, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_token, 0, -1L, NULL, NULL, &__pyx_slice__2, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyNumber_Int(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_event, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_event_len, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "aldy/indelpost/localn.pyx":554
 *         event, event_len = token[-1], int(token[:-1])
 * 
 *         if event == "I" or event == "D":             # <<<<<<<<<<<<<<
 *             indel = {}
 * 
 */
    __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_v_event, __pyx_n_u_I, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 554, __pyx_L1_error)
    if (!__pyx_t_10) {
    } else {
      __pyx_t_9 = __pyx_t_10;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_v_event, __pyx_n_u_D, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 554, __pyx_L1_error)
    __pyx_t_9 = __pyx_t_10;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_9) {

      /* "aldy/indelpost/localn.pyx":555
 * 
 *         if event == "I" or event == "D":
 *             indel = {}             # <<<<<<<<<<<<<<
 * 
 *             indel["pos"] = genome_aln_pos
 */
      __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 555, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_indel, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "aldy/indelpost/localn.pyx":557
 *             indel = {}
 * 
 *             indel["pos"] = genome_aln_pos             # <<<<<<<<<<<<<<
 *             indel["lt_ref"] = ref_seq[:ref_idx]
 *             indel["lt_flank"] = read_seq[:read_idx]
 */
      if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_pos, __pyx_v_genome_aln_pos) < 0)) __PYX_ERR(0, 557, __pyx_L1_error)

      /* "aldy/indelpost/localn.pyx":558
 * 
 *             indel["pos"] = genome_aln_pos
 *             indel["lt_ref"] = ref_seq[:ref_idx]             # <<<<<<<<<<<<<<
 *             indel["lt_flank"] = read_seq[:read_idx]
 * 
 */
      __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_ref_seq, 0, 0, NULL, &__pyx_v_ref_idx, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 558, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_lt_ref, __pyx_t_4) < 0)) __PYX_ERR(0, 558, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "aldy/indelpost/localn.pyx":559
 *             indel["pos"] = genome_aln_pos
 *             indel["lt_ref"] = ref_seq[:ref_idx]
 *             indel["lt_flank"] = read_seq[:read_idx]             # <<<<<<<<<<<<<<
 * 
 *             if basequals:
 */
      __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, NULL, &__pyx_v_read_idx, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 559, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_lt_flank, __pyx_t_4) < 0)) __PYX_ERR(0, 559, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "aldy/indelpost/localn.pyx":561
 *             indel["lt_flank"] = read_seq[:read_idx]
 * 
 *             if basequals:             # <<<<<<<<<<<<<<
 *                 indel["lt_qual"] = basequals[:read_idx]
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_basequals); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 561, __pyx_L1_error)
      if (__pyx_t_9) {

        /* "aldy/indelpost/localn.pyx":562
 * 
 *             if basequals:
 *                 indel["lt_qual"] = basequals[:read_idx]             # <<<<<<<<<<<<<<
 * 
 *             if event == "I":
 */
        __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_basequals, 0, 0, NULL, &__pyx_v_read_idx, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 562, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_lt_qual, __pyx_t_4) < 0)) __PYX_ERR(0, 562, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "aldy/indelpost/localn.pyx":561
 *             indel["lt_flank"] = read_seq[:read_idx]
 * 
 *             if basequals:             # <<<<<<<<<<<<<<
 *                 indel["lt_qual"] = basequals[:read_idx]
 * 
 */
      }

      /* "aldy/indelpost/localn.pyx":564
 *                 indel["lt_qual"] = basequals[:read_idx]
 * 
 *             if event == "I":             # <<<<<<<<<<<<<<
 *                 indel["indel_type"] = "I"
 *                 indel["indel_seq"] = read_seq[read_idx : read_idx + event_len]
 */
      __pyx_t_9 = (__Pyx_PyUnicode_Equals(__pyx_v_event, __pyx_n_u_I, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 564, __pyx_L1_error)
      if (__pyx_t_9) {

        /* "aldy/indelpost/localn.pyx":565
 * 
 *             if event == "I":
 *                 indel["indel_type"] = "I"             # <<<<<<<<<<<<<<
 *                 indel["indel_seq"] = read_seq[read_idx : read_idx + event_len]
 *                 indel["rt_ref"] = ref_seq[ref_idx:]
 */
        if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_indel_type, __pyx_n_u_I) < 0)) __PYX_ERR(0, 565, __pyx_L1_error)

        /* "aldy/indelpost/localn.pyx":566
 *             if event == "I":
 *                 indel["indel_type"] = "I"
 *                 indel["indel_seq"] = read_seq[read_idx : read_idx + event_len]             # <<<<<<<<<<<<<<
 *                 indel["rt_ref"] = ref_seq[ref_idx:]
 *                 indel["rt_flank"] = read_seq[read_idx + event_len :]
 */
        __pyx_t_4 = PyNumber_Add(__pyx_v_read_idx, __pyx_v_event_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 566, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, &__pyx_v_read_idx, &__pyx_t_4, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 566, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_indel_seq, __pyx_t_2) < 0)) __PYX_ERR(0, 566, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "aldy/indelpost/localn.pyx":567
 *                 indel["indel_type"] = "I"
 *                 indel["indel_seq"] = read_seq[read_idx : read_idx + event_len]
 *                 indel["rt_ref"] = ref_seq[ref_idx:]             # <<<<<<<<<<<<<<
 *                 indel["rt_flank"] = read_seq[read_idx + event_len :]
 *                 indel["ref_idx"] = ref_idx
 */
        __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_ref_seq, 0, 0, &__pyx_v_ref_idx, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 567, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_rt_ref, __pyx_t_2) < 0)) __PYX_ERR(0, 567, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "aldy/indelpost/localn.pyx":568
 *                 indel["indel_seq"] = read_seq[read_idx : read_idx + event_len]
 *                 indel["rt_ref"] = ref_seq[ref_idx:]
 *                 indel["rt_flank"] = read_seq[read_idx + event_len :]             # <<<<<<<<<<<<<<
 *                 indel["ref_idx"] = ref_idx
 *                 indel["read_idx"] = read_idx
 */
        __pyx_t_2 = PyNumber_Add(__pyx_v_read_idx, __pyx_v_event_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 568, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, &__pyx_t_2, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 568, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_rt_flank, __pyx_t_4) < 0)) __PYX_ERR(0, 568, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "aldy/indelpost/localn.pyx":569
 *                 indel["rt_ref"] = ref_seq[ref_idx:]
 *                 indel["rt_flank"] = read_seq[read_idx + event_len :]
 *                 indel["ref_idx"] = ref_idx             # <<<<<<<<<<<<<<
 *                 indel["read_idx"] = read_idx
 * 
 */
        if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_ref_idx, __pyx_v_ref_idx) < 0)) __PYX_ERR(0, 569, __pyx_L1_error)

        /* "aldy/indelpost/localn.pyx":570
 *                 indel["rt_flank"] = read_seq[read_idx + event_len :]
 *                 indel["ref_idx"] = ref_idx
 *                 indel["read_idx"] = read_idx             # <<<<<<<<<<<<<<
 * 
 *                 if basequals:
 */
        if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_read_idx, __pyx_v_read_idx) < 0)) __PYX_ERR(0, 570, __pyx_L1_error)

        /* "aldy/indelpost/localn.pyx":572
 *                 indel["read_idx"] = read_idx
 * 
 *                 if basequals:             # <<<<<<<<<<<<<<
 *                     indel["rt_qual"] = basequals[read_idx + event_len :]
 * 
 */
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_basequals); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 572, __pyx_L1_error)
        if (__pyx_t_9) {

          /* "aldy/indelpost/localn.pyx":573
 * 
 *                 if basequals:
 *                     indel["rt_qual"] = basequals[read_idx + event_len :]             # <<<<<<<<<<<<<<
 * 
 *                 read_idx += event_len
 */
          __pyx_t_4 = PyNumber_Add(__pyx_v_read_idx, __pyx_v_event_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 573, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_basequals, 0, 0, &__pyx_t_4, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_rt_qual, __pyx_t_2) < 0)) __PYX_ERR(0, 573, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "aldy/indelpost/localn.pyx":572
 *                 indel["read_idx"] = read_idx
 * 
 *                 if basequals:             # <<<<<<<<<<<<<<
 *                     indel["rt_qual"] = basequals[read_idx + event_len :]
 * 
 */
        }

        /* "aldy/indelpost/localn.pyx":575
 *                     indel["rt_qual"] = basequals[read_idx + event_len :]
 * 
 *                 read_idx += event_len             # <<<<<<<<<<<<<<
 *             else:
 *                 indel["indel_type"] = "D"
 */
        __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_read_idx, __pyx_v_event_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 575, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_read_idx, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "aldy/indelpost/localn.pyx":564
 *                 indel["lt_qual"] = basequals[:read_idx]
 * 
 *             if event == "I":             # <<<<<<<<<<<<<<
 *                 indel["indel_type"] = "I"
 *                 indel["indel_seq"] = read_seq[read_idx : read_idx + event_len]
 */
        goto __pyx_L9;
      }

      /* "aldy/indelpost/localn.pyx":577
 *                 read_idx += event_len
 *             else:
 *                 indel["indel_type"] = "D"             # <<<<<<<<<<<<<<
 *                 indel["indel_seq"] = ""
 *                 indel["del_seq"] = ref_seq[ref_idx : ref_idx + event_len]
 */
      /*else*/ {
        if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_indel_type, __pyx_n_u_D) < 0)) __PYX_ERR(0, 577, __pyx_L1_error)

        /* "aldy/indelpost/localn.pyx":578
 *             else:
 *                 indel["indel_type"] = "D"
 *                 indel["indel_seq"] = ""             # <<<<<<<<<<<<<<
 *                 indel["del_seq"] = ref_seq[ref_idx : ref_idx + event_len]
 *                 indel["rt_ref"] = ref_seq[ref_idx + event_len :]
 */
        if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_indel_seq, __pyx_kp_u__3) < 0)) __PYX_ERR(0, 578, __pyx_L1_error)

        /* "aldy/indelpost/localn.pyx":579
 *                 indel["indel_type"] = "D"
 *                 indel["indel_seq"] = ""
 *                 indel["del_seq"] = ref_seq[ref_idx : ref_idx + event_len]             # <<<<<<<<<<<<<<
 *                 indel["rt_ref"] = ref_seq[ref_idx + event_len :]
 *                 indel["rt_flank"] = read_seq[read_idx:]
 */
        __pyx_t_2 = PyNumber_Add(__pyx_v_ref_idx, __pyx_v_event_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_ref_seq, 0, 0, &__pyx_v_ref_idx, &__pyx_t_2, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 579, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_del_seq, __pyx_t_4) < 0)) __PYX_ERR(0, 579, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "aldy/indelpost/localn.pyx":580
 *                 indel["indel_seq"] = ""
 *                 indel["del_seq"] = ref_seq[ref_idx : ref_idx + event_len]
 *                 indel["rt_ref"] = ref_seq[ref_idx + event_len :]             # <<<<<<<<<<<<<<
 *                 indel["rt_flank"] = read_seq[read_idx:]
 *                 indel["ref_idx"] = ref_idx
 */
        __pyx_t_4 = PyNumber_Add(__pyx_v_ref_idx, __pyx_v_event_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 580, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_ref_seq, 0, 0, &__pyx_t_4, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 580, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_rt_ref, __pyx_t_2) < 0)) __PYX_ERR(0, 580, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "aldy/indelpost/localn.pyx":581
 *                 indel["del_seq"] = ref_seq[ref_idx : ref_idx + event_len]
 *                 indel["rt_ref"] = ref_seq[ref_idx + event_len :]
 *                 indel["rt_flank"] = read_seq[read_idx:]             # <<<<<<<<<<<<<<
 *                 indel["ref_idx"] = ref_idx
 *                 indel["read_idx"] = read_idx
 */
        __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, &__pyx_v_read_idx, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 581, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_rt_flank, __pyx_t_2) < 0)) __PYX_ERR(0, 581, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "aldy/indelpost/localn.pyx":582
 *                 indel["rt_ref"] = ref_seq[ref_idx + event_len :]
 *                 indel["rt_flank"] = read_seq[read_idx:]
 *                 indel["ref_idx"] = ref_idx             # <<<<<<<<<<<<<<
 *                 indel["read_idx"] = read_idx
 * 
 */
        if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_ref_idx, __pyx_v_ref_idx) < 0)) __PYX_ERR(0, 582, __pyx_L1_error)

        /* "aldy/indelpost/localn.pyx":583
 *                 indel["rt_flank"] = read_seq[read_idx:]
 *                 indel["ref_idx"] = ref_idx
 *                 indel["read_idx"] = read_idx             # <<<<<<<<<<<<<<
 * 
 *                 if basequals:
 */
        if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_read_idx, __pyx_v_read_idx) < 0)) __PYX_ERR(0, 583, __pyx_L1_error)

        /* "aldy/indelpost/localn.pyx":585
 *                 indel["read_idx"] = read_idx
 * 
 *                 if basequals:             # <<<<<<<<<<<<<<
 *                     indel["rt_qual"] = basequals[read_idx:]
 * 
 */
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_basequals); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 585, __pyx_L1_error)
        if (__pyx_t_9) {

          /* "aldy/indelpost/localn.pyx":586
 * 
 *                 if basequals:
 *                     indel["rt_qual"] = basequals[read_idx:]             # <<<<<<<<<<<<<<
 * 
 *                 ref_idx += event_len
 */
          __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_basequals, 0, 0, &__pyx_v_read_idx, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_rt_qual, __pyx_t_2) < 0)) __PYX_ERR(0, 586, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "aldy/indelpost/localn.pyx":585
 *                 indel["read_idx"] = read_idx
 * 
 *                 if basequals:             # <<<<<<<<<<<<<<
 *                     indel["rt_qual"] = basequals[read_idx:]
 * 
 */
        }

        /* "aldy/indelpost/localn.pyx":588
 *                     indel["rt_qual"] = basequals[read_idx:]
 * 
 *                 ref_idx += event_len             # <<<<<<<<<<<<<<
 *                 genome_aln_pos += event_len
 * 
 */
        __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_ref_idx, __pyx_v_event_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 588, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_ref_idx, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "aldy/indelpost/localn.pyx":589
 * 
 *                 ref_idx += event_len
 *                 genome_aln_pos += event_len             # <<<<<<<<<<<<<<
 * 
 *             indels.append(indel)
 */
        __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_genome_aln_pos, __pyx_v_event_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 589, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_genome_aln_pos, __pyx_t_2);
        __pyx_t_2 = 0;
      }
      __pyx_L9:;

      /* "aldy/indelpost/localn.pyx":591
 *                 genome_aln_pos += event_len
 * 
 *             indels.append(indel)             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_indels, __pyx_v_indel); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 591, __pyx_L1_error)

      /* "aldy/indelpost/localn.pyx":554
 *         event, event_len = token[-1], int(token[:-1])
 * 
 *         if event == "I" or event == "D":             # <<<<<<<<<<<<<<
 *             indel = {}
 * 
 */
      goto __pyx_L5;
    }

    /* "aldy/indelpost/localn.pyx":594
 * 
 *         else:
 *             if report_snvs:             # <<<<<<<<<<<<<<
 *                 i = 0
 *                 while i < event_len:
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_report_snvs); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 594, __pyx_L1_error)
      if (__pyx_t_9) {

        /* "aldy/indelpost/localn.pyx":595
 *         else:
 *             if report_snvs:
 *                 i = 0             # <<<<<<<<<<<<<<
 *                 while i < event_len:
 *                      if ref_seq[ref_idx + i : ref_idx + i + 1] != read_seq[read_idx + i : read_idx + i + 1]:
 */
        __Pyx_INCREF(__pyx_int_0);
        __Pyx_XDECREF_SET(__pyx_v_i, __pyx_int_0);

        /* "aldy/indelpost/localn.pyx":596
 *             if report_snvs:
 *                 i = 0
 *                 while i < event_len:             # <<<<<<<<<<<<<<
 *                      if ref_seq[ref_idx + i : ref_idx + i + 1] != read_seq[read_idx + i : read_idx + i + 1]:
 *                          snv = {}
 */
        while (1) {
          __pyx_t_2 = PyObject_RichCompare(__pyx_v_i, __pyx_v_event_len, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 596, __pyx_L1_error)
          __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 596, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (!__pyx_t_9) break;

          /* "aldy/indelpost/localn.pyx":597
 *                 i = 0
 *                 while i < event_len:
 *                      if ref_seq[ref_idx + i : ref_idx + i + 1] != read_seq[read_idx + i : read_idx + i + 1]:             # <<<<<<<<<<<<<<
 *                          snv = {}
 *                          snv["pos"] = genome_aln_pos + i + 1
 */
          __pyx_t_2 = PyNumber_Add(__pyx_v_ref_idx, __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 597, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = PyNumber_Add(__pyx_v_ref_idx, __pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 597, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 597, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_ref_seq, 0, 0, &__pyx_t_2, &__pyx_t_1, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 597, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyNumber_Add(__pyx_v_read_idx, __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 597, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = PyNumber_Add(__pyx_v_read_idx, __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 597, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 597, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, &__pyx_t_1, &__pyx_t_5, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 597, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 597, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 597, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (__pyx_t_9) {

            /* "aldy/indelpost/localn.pyx":598
 *                 while i < event_len:
 *                      if ref_seq[ref_idx + i : ref_idx + i + 1] != read_seq[read_idx + i : read_idx + i + 1]:
 *                          snv = {}             # <<<<<<<<<<<<<<
 *                          snv["pos"] = genome_aln_pos + i + 1
 *                          snv["ref"] = ref_seq[ref_idx + i : ref_idx + i + 1]
 */
            __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 598, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_XDECREF_SET(__pyx_v_snv, ((PyObject*)__pyx_t_5));
            __pyx_t_5 = 0;

            /* "aldy/indelpost/localn.pyx":599
 *                      if ref_seq[ref_idx + i : ref_idx + i + 1] != read_seq[read_idx + i : read_idx + i + 1]:
 *                          snv = {}
 *                          snv["pos"] = genome_aln_pos + i + 1             # <<<<<<<<<<<<<<
 *                          snv["ref"] = ref_seq[ref_idx + i : ref_idx + i + 1]
 *                          snv["alt"] = read_seq[read_idx + i : read_idx + i + 1]
 */
            __pyx_t_5 = PyNumber_Add(__pyx_v_genome_aln_pos, __pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 599, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_5, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 599, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(PyDict_SetItem(__pyx_v_snv, __pyx_n_u_pos, __pyx_t_2) < 0)) __PYX_ERR(0, 599, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "aldy/indelpost/localn.pyx":600
 *                          snv = {}
 *                          snv["pos"] = genome_aln_pos + i + 1
 *                          snv["ref"] = ref_seq[ref_idx + i : ref_idx + i + 1]             # <<<<<<<<<<<<<<
 *                          snv["alt"] = read_seq[read_idx + i : read_idx + i + 1]
 * 
 */
            __pyx_t_2 = PyNumber_Add(__pyx_v_ref_idx, __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 600, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_5 = PyNumber_Add(__pyx_v_ref_idx, __pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 600, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_t_5, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 600, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_ref_seq, 0, 0, &__pyx_t_2, &__pyx_t_4, NULL, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 600, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(PyDict_SetItem(__pyx_v_snv, __pyx_n_u_ref, __pyx_t_5) < 0)) __PYX_ERR(0, 600, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

            /* "aldy/indelpost/localn.pyx":601
 *                          snv["pos"] = genome_aln_pos + i + 1
 *                          snv["ref"] = ref_seq[ref_idx + i : ref_idx + i + 1]
 *                          snv["alt"] = read_seq[read_idx + i : read_idx + i + 1]             # <<<<<<<<<<<<<<
 * 
 *                          snvs.append(snv)
 */
            __pyx_t_5 = PyNumber_Add(__pyx_v_read_idx, __pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 601, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_4 = PyNumber_Add(__pyx_v_read_idx, __pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 601, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 601, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, &__pyx_t_5, &__pyx_t_2, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 601, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(PyDict_SetItem(__pyx_v_snv, __pyx_n_u_alt, __pyx_t_4) < 0)) __PYX_ERR(0, 601, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "aldy/indelpost/localn.pyx":603
 *                          snv["alt"] = read_seq[read_idx + i : read_idx + i + 1]
 * 
 *                          snvs.append(snv)             # <<<<<<<<<<<<<<
 * 
 *                      i += 1
 */
            __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_snvs, __pyx_v_snv); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 603, __pyx_L1_error)

            /* "aldy/indelpost/localn.pyx":597
 *                 i = 0
 *                 while i < event_len:
 *                      if ref_seq[ref_idx + i : ref_idx + i + 1] != read_seq[read_idx + i : read_idx + i + 1]:             # <<<<<<<<<<<<<<
 *                          snv = {}
 *                          snv["pos"] = genome_aln_pos + i + 1
 */
          }

          /* "aldy/indelpost/localn.pyx":605
 *                          snvs.append(snv)
 * 
 *                      i += 1             # <<<<<<<<<<<<<<
 * 
 *             ref_idx += event_len
 */
          __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 605, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF_SET(__pyx_v_i, __pyx_t_4);
          __pyx_t_4 = 0;
        }

        /* "aldy/indelpost/localn.pyx":594
 * 
 *         else:
 *             if report_snvs:             # <<<<<<<<<<<<<<
 *                 i = 0
 *                 while i < event_len:
 */
      }

      /* "aldy/indelpost/localn.pyx":607
 *                      i += 1
 * 
 *             ref_idx += event_len             # <<<<<<<<<<<<<<
 *             read_idx += event_len
 *             genome_aln_pos += event_len
 */
      __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_ref_idx, __pyx_v_event_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 607, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_ref_idx, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "aldy/indelpost/localn.pyx":608
 * 
 *             ref_idx += event_len
 *             read_idx += event_len             # <<<<<<<<<<<<<<
 *             genome_aln_pos += event_len
 * 
 */
      __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_read_idx, __pyx_v_event_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 608, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_read_idx, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "aldy/indelpost/localn.pyx":609
 *             ref_idx += event_len
 *             read_idx += event_len
 *             genome_aln_pos += event_len             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_genome_aln_pos, __pyx_v_event_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 609, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_genome_aln_pos, __pyx_t_4);
      __pyx_t_4 = 0;
    }
    __pyx_L5:;

    /* "aldy/indelpost/localn.pyx":551
 *     indels, snvs = [], []
 * 
 *     for token in cigar_ptrn.findall(make_insertion_first(ref_aln.CIGAR)):             # <<<<<<<<<<<<<<
 *         event, event_len = token[-1], int(token[:-1])
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "aldy/indelpost/localn.pyx":612
 * 
 * 
 *     rt_clipped = read_seq[read_idx:]             # <<<<<<<<<<<<<<
 *     for indel in indels:
 *         indel["lt_clipped"] = lt_clipped
 */
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_read_seq, 0, 0, &__pyx_v_read_idx, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_rt_clipped = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "aldy/indelpost/localn.pyx":613
 * 
 *     rt_clipped = read_seq[read_idx:]
 *     for indel in indels:             # <<<<<<<<<<<<<<
 *         indel["lt_clipped"] = lt_clipped
 *         indel["rt_clipped"] = rt_clipped
 */
  __pyx_t_3 = __pyx_v_indels; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
  for (;;) {
    if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 613, __pyx_L1_error)
    #else
    __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_indel, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "aldy/indelpost/localn.pyx":614
 *     rt_clipped = read_seq[read_idx:]
 *     for indel in indels:
 *         indel["lt_clipped"] = lt_clipped             # <<<<<<<<<<<<<<
 *         indel["rt_clipped"] = rt_clipped
 * 
 */
    if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_lt_clipped, __pyx_v_lt_clipped) < 0)) __PYX_ERR(0, 614, __pyx_L1_error)

    /* "aldy/indelpost/localn.pyx":615
 *     for indel in indels:
 *         indel["lt_clipped"] = lt_clipped
 *         indel["rt_clipped"] = rt_clipped             # <<<<<<<<<<<<<<
 * 
 *     if report_snvs:
 */
    if (unlikely(PyObject_SetItem(__pyx_v_indel, __pyx_n_u_rt_clipped, __pyx_v_rt_clipped) < 0)) __PYX_ERR(0, 615, __pyx_L1_error)

    /* "aldy/indelpost/localn.pyx":613
 * 
 *     rt_clipped = read_seq[read_idx:]
 *     for indel in indels:             # <<<<<<<<<<<<<<
 *         indel["lt_clipped"] = lt_clipped
 *         indel["rt_clipped"] = rt_clipped
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "aldy/indelpost/localn.pyx":617
 *         indel["rt_clipped"] = rt_clipped
 * 
 *     if report_snvs:             # <<<<<<<<<<<<<<
 *         return indels, snvs
 *     else:
 */
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_report_snvs); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 617, __pyx_L1_error)
  if (__pyx_t_9) {

    /* "aldy/indelpost/localn.pyx":618
 * 
 *     if report_snvs:
 *         return indels, snvs             # <<<<<<<<<<<<<<
 *     else:
 *         return indels
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 618, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_indels);
    __Pyx_GIVEREF(__pyx_v_indels);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_indels);
    __Pyx_INCREF(__pyx_v_snvs);
    __Pyx_GIVEREF(__pyx_v_snvs);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_snvs);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "aldy/indelpost/localn.pyx":617
 *         indel["rt_clipped"] = rt_clipped
 * 
 *     if report_snvs:             # <<<<<<<<<<<<<<
 *         return indels, snvs
 *     else:
 */
  }

  /* "aldy/indelpost/localn.pyx":620
 *         return indels, snvs
 *     else:
 *         return indels             # <<<<<<<<<<<<<<
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_indels);
    __pyx_r = __pyx_v_indels;
    goto __pyx_L0;
  }

  /* "aldy/indelpost/localn.pyx":541
 * 
 * 
 * def findall_indels(ref_aln, genome_aln_pos, ref_seq, read_seq, report_snvs=False, basequals=None):             # <<<<<<<<<<<<<<
 * 
 *     genome_aln_pos -= 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("aldy.indelpost.localn.findall_indels", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ref_idx);
  __Pyx_XDECREF(__pyx_v_read_idx);
  __Pyx_XDECREF(__pyx_v_lt_clipped);
  __Pyx_XDECREF(__pyx_v_indels);
  __Pyx_XDECREF(__pyx_v_snvs);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XDECREF(__pyx_v_event);
  __Pyx_XDECREF(__pyx_v_event_len);
  __Pyx_XDECREF(__pyx_v_indel);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_snv);
  __Pyx_XDECREF(__pyx_v_rt_clipped);
  __Pyx_XDECREF(__pyx_v_genome_aln_pos);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":93
 *             __data_union data
 * 
 *         def __getbuffer__(self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_7cpython_5array_5array_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_7cpython_5array_5array_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7cpython_5array_5array___getbuffer__(((arrayobject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7cpython_5array_5array___getbuffer__(arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info, CYTHON_UNUSED int __pyx_v_flags) {
  PyObject *__pyx_v_item_count = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  char __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "array.pxd":98
 *             # In particular strided access is always provided regardless
 *             # of flags
 *             item_count = Py_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *             info.suboffsets = NULL
 */
  __pyx_t_1 = PyInt_FromSsize_t(Py_SIZE(((PyObject *)__pyx_v_self))); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_item_count = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "array.pxd":100
 *             item_count = Py_SIZE(self)
 * 
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.buf = self.data.as_chars
 *             info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "array.pxd":101
 * 
 *             info.suboffsets = NULL
 *             info.buf = self.data.as_chars             # <<<<<<<<<<<<<<
 *             info.readonly = 0
 *             info.ndim = 1
 */
  __pyx_t_2 = __pyx_v_self->data.as_chars;
  __pyx_v_info->buf = __pyx_t_2;

  /* "array.pxd":102
 *             info.suboffsets = NULL
 *             info.buf = self.data.as_chars
 *             info.readonly = 0             # <<<<<<<<<<<<<<
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 */
  __pyx_v_info->readonly = 0;

  /* "array.pxd":103
 *             info.buf = self.data.as_chars
 *             info.readonly = 0
 *             info.ndim = 1             # <<<<<<<<<<<<<<
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 *             info.len = info.itemsize * item_count
 */
  __pyx_v_info->ndim = 1;

  /* "array.pxd":104
 *             info.readonly = 0
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)             # <<<<<<<<<<<<<<
 *             info.len = info.itemsize * item_count
 * 
 */
  __pyx_t_3 = __pyx_v_self->ob_descr->itemsize;
  __pyx_v_info->itemsize = __pyx_t_3;

  /* "array.pxd":105
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 *             info.len = info.itemsize * item_count             # <<<<<<<<<<<<<<
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 */
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_info->itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_1, __pyx_v_item_count); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_info->len = __pyx_t_5;

  /* "array.pxd":107
 *             info.len = info.itemsize * item_count
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)             # <<<<<<<<<<<<<<
 *             if not info.shape:
 *                 raise MemoryError()
 */
  __pyx_v_info->shape = ((Py_ssize_t *)PyObject_Malloc(((sizeof(Py_ssize_t)) + 2)));

  /* "array.pxd":108
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 */
  __pyx_t_6 = ((!(__pyx_v_info->shape != 0)) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "array.pxd":109
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 *             info.shape[0] = item_count      # constant regardless of resizing
 *             info.strides = &info.itemsize
 */
    PyErr_NoMemory(); __PYX_ERR(1, 109, __pyx_L1_error)

    /* "array.pxd":108
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 */
  }

  /* "array.pxd":110
 *             if not info.shape:
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing             # <<<<<<<<<<<<<<
 *             info.strides = &info.itemsize
 * 
 */
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_v_item_count); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 110, __pyx_L1_error)
  (__pyx_v_info->shape[0]) = __pyx_t_5;

  /* "array.pxd":111
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 *             info.strides = &info.itemsize             # <<<<<<<<<<<<<<
 * 
 *             info.format = <char*> (info.shape + 1)
 */
  __pyx_v_info->strides = (&__pyx_v_info->itemsize);

  /* "array.pxd":113
 *             info.strides = &info.itemsize
 * 
 *             info.format = <char*> (info.shape + 1)             # <<<<<<<<<<<<<<
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0
 */
  __pyx_v_info->format = ((char *)(__pyx_v_info->shape + 1));

  /* "array.pxd":114
 * 
 *             info.format = <char*> (info.shape + 1)
 *             info.format[0] = self.ob_descr.typecode             # <<<<<<<<<<<<<<
 *             info.format[1] = 0
 *             info.obj = self
 */
  __pyx_t_7 = __pyx_v_self->ob_descr->typecode;
  (__pyx_v_info->format[0]) = __pyx_t_7;

  /* "array.pxd":115
 *             info.format = <char*> (info.shape + 1)
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0             # <<<<<<<<<<<<<<
 *             info.obj = self
 * 
 */
  (__pyx_v_info->format[1]) = 0;

  /* "array.pxd":116
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(self, Py_buffer* info):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "array.pxd":93
 *             __data_union data
 * 
 *         def __getbuffer__(self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cpython.array.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_item_count);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":118
 *             info.obj = self
 * 
 *         def __releasebuffer__(self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             PyObject_Free(info.shape)
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_7cpython_5array_5array_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_7cpython_5array_5array_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_7cpython_5array_5array_2__releasebuffer__(((arrayobject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7cpython_5array_5array_2__releasebuffer__(CYTHON_UNUSED arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "array.pxd":119
 * 
 *         def __releasebuffer__(self, Py_buffer* info):
 *             PyObject_Free(info.shape)             # <<<<<<<<<<<<<<
 * 
 *     array newarrayobject(PyTypeObject* type, Py_ssize_t size, arraydescr *descr)
 */
  PyObject_Free(__pyx_v_info->shape);

  /* "array.pxd":118
 *             info.obj = self
 * 
 *         def __releasebuffer__(self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             PyObject_Free(info.shape)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "array.pxd":130
 * 
 * 
 * cdef inline array clone(array template, Py_ssize_t length, bint zero):             # <<<<<<<<<<<<<<
 *     """ fast creation of a new array, given a template array.
 *     type will be same as template.
 */

static CYTHON_INLINE arrayobject *__pyx_f_7cpython_5array_clone(arrayobject *__pyx_v_template, Py_ssize_t __pyx_v_length, int __pyx_v_zero) {
  arrayobject *__pyx_v_op = 0;
  arrayobject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);

  /* "array.pxd":134
 *     type will be same as template.
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)             # <<<<<<<<<<<<<<
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 */
  __pyx_t_1 = ((PyObject *)newarrayobject(Py_TYPE(((PyObject *)__pyx_v_template)), __pyx_v_length, __pyx_v_template->ob_descr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "array.pxd":135
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:             # <<<<<<<<<<<<<<
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op
 */
  __pyx_t_3 = (__pyx_v_zero != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (((PyObject *)__pyx_v_op) != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "array.pxd":136
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 *     return op
 * 
 */
    (void)(memset(__pyx_v_op->data.as_chars, 0, (__pyx_v_length * __pyx_v_op->ob_descr->itemsize)));

    /* "array.pxd":135
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:             # <<<<<<<<<<<<<<
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op
 */
  }

  /* "array.pxd":137
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op             # <<<<<<<<<<<<<<
 * 
 * cdef inline array copy(array self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_op));
  __pyx_r = __pyx_v_op;
  goto __pyx_L0;

  /* "array.pxd":130
 * 
 * 
 * cdef inline array clone(array template, Py_ssize_t length, bint zero):             # <<<<<<<<<<<<<<
 *     """ fast creation of a new array, given a template array.
 *     type will be same as template.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.array.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":139
 *     return op
 * 
 * cdef inline array copy(array self):             # <<<<<<<<<<<<<<
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 */

static CYTHON_INLINE arrayobject *__pyx_f_7cpython_5array_copy(arrayobject *__pyx_v_self) {
  arrayobject *__pyx_v_op = 0;
  arrayobject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "array.pxd":141
 * cdef inline array copy(array self):
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)             # <<<<<<<<<<<<<<
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)
 *     return op
 */
  __pyx_t_1 = ((PyObject *)newarrayobject(Py_TYPE(((PyObject *)__pyx_v_self)), Py_SIZE(((PyObject *)__pyx_v_self)), __pyx_v_self->ob_descr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "array.pxd":142
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 *     return op
 * 
 */
  (void)(memcpy(__pyx_v_op->data.as_chars, __pyx_v_self->data.as_chars, (Py_SIZE(((PyObject *)__pyx_v_op)) * __pyx_v_op->ob_descr->itemsize)));

  /* "array.pxd":143
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)
 *     return op             # <<<<<<<<<<<<<<
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_op));
  __pyx_r = __pyx_v_op;
  goto __pyx_L0;

  /* "array.pxd":139
 *     return op
 * 
 * cdef inline array copy(array self):             # <<<<<<<<<<<<<<
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.array.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":145
 *     return op
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:             # <<<<<<<<<<<<<<
 *     """ efficient appending of new stuff of same type
 *     (e.g. of same array type)
 */

static CYTHON_INLINE int __pyx_f_7cpython_5array_extend_buffer(arrayobject *__pyx_v_self, char *__pyx_v_stuff, Py_ssize_t __pyx_v_n) {
  Py_ssize_t __pyx_v_itemsize;
  Py_ssize_t __pyx_v_origsize;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extend_buffer", 0);

  /* "array.pxd":149
 *     (e.g. of same array type)
 *     n: number of elements (not number of bytes!) """
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)
 */
  __pyx_t_1 = __pyx_v_self->ob_descr->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "array.pxd":150
 *     n: number of elements (not number of bytes!) """
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize
 *     cdef Py_ssize_t origsize = Py_SIZE(self)             # <<<<<<<<<<<<<<
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 */
  __pyx_v_origsize = Py_SIZE(((PyObject *)__pyx_v_self));

  /* "array.pxd":151
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)             # <<<<<<<<<<<<<<
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 *     return 0
 */
  __pyx_t_1 = resize_smart(__pyx_v_self, (__pyx_v_origsize + __pyx_v_n)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 151, __pyx_L1_error)

  /* "array.pxd":152
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  (void)(memcpy((__pyx_v_self->data.as_chars + (__pyx_v_origsize * __pyx_v_itemsize)), __pyx_v_stuff, (__pyx_v_n * __pyx_v_itemsize)));

  /* "array.pxd":153
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef inline int extend(array self, array other) except -1:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "array.pxd":145
 *     return op
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:             # <<<<<<<<<<<<<<
 *     """ efficient appending of new stuff of same type
 *     (e.g. of same array type)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cpython.array.extend_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":155
 *     return 0
 * 
 * cdef inline int extend(array self, array other) except -1:             # <<<<<<<<<<<<<<
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 */

static CYTHON_INLINE int __pyx_f_7cpython_5array_extend(arrayobject *__pyx_v_self, arrayobject *__pyx_v_other) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extend", 0);

  /* "array.pxd":157
 * cdef inline int extend(array self, array other) except -1:
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:             # <<<<<<<<<<<<<<
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 */
  __pyx_t_1 = ((__pyx_v_self->ob_descr->typecode != __pyx_v_other->ob_descr->typecode) != 0);
  if (__pyx_t_1) {

    /* "array.pxd":158
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 *         PyErr_BadArgument()             # <<<<<<<<<<<<<<
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 */
    __pyx_t_2 = PyErr_BadArgument(); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(1, 158, __pyx_L1_error)

    /* "array.pxd":157
 * cdef inline int extend(array self, array other) except -1:
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:             # <<<<<<<<<<<<<<
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 */
  }

  /* "array.pxd":159
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))             # <<<<<<<<<<<<<<
 * 
 * cdef inline void zero(array self):
 */
  __pyx_t_2 = __pyx_f_7cpython_5array_extend_buffer(__pyx_v_self, __pyx_v_other->data.as_chars, Py_SIZE(((PyObject *)__pyx_v_other))); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(1, 159, __pyx_L1_error)
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "array.pxd":155
 *     return 0
 * 
 * cdef inline int extend(array self, array other) except -1:             # <<<<<<<<<<<<<<
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cpython.array.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":161
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

static CYTHON_INLINE void __pyx_f_7cpython_5array_zero(arrayobject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zero", 0);

  /* "array.pxd":163
 * cdef inline void zero(array self):
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 */
  (void)(memset(__pyx_v_self->data.as_chars, 0, (Py_SIZE(((PyObject *)__pyx_v_self)) * __pyx_v_self->ob_descr->itemsize)));

  /* "array.pxd":161
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *__pyx_freelist_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn[8];
static int __pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn = 0;

static PyObject *__pyx_tp_new_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn)))) {
    o = (PyObject*)__pyx_freelist_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn[--__pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn];
    memset(o, 0, sizeof(struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn(PyObject *o) {
  struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *p = (struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_mismatches);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn)))) {
    __pyx_freelist_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn[__pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn++] = ((struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *p = (struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *)o;
  if (p->__pyx_v_mismatches) {
    e = (*v)(p->__pyx_v_mismatches, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *p = (struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn *)o;
  tmp = ((PyObject*)p->__pyx_v_mismatches);
  p->__pyx_v_mismatches = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn = {
  PyVarObject_HEAD_INIT(0, 0)
  "aldy.indelpost.localn.__pyx_scope_struct__is_worth_realn", /*tp_name*/
  sizeof(struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn, /*tp_traverse*/
  __pyx_tp_clear_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr *__pyx_freelist_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr[--__pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_var);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr[__pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_var) {
    e = (*v)(p->__pyx_v_var, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "aldy.indelpost.localn.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr *__pyx_freelist_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr[8];
static int __pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr = 0;

static PyObject *__pyx_tp_new_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr)))) {
    o = (PyObject*)__pyx_freelist_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr[--__pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr(PyObject *o) {
  struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_var);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr)))) {
    __pyx_freelist_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr[__pyx_freecount_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr++] = ((struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_var) {
    e = (*v)(p->__pyx_v_var, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "aldy.indelpost.localn.__pyx_scope_struct_2_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_localn(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_localn},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "localn",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_0_9_MIDNSHPX, __pyx_k_0_9_MIDNSHPX, sizeof(__pyx_k_0_9_MIDNSHPX), 0, 1, 0, 0},
  {&__pyx_n_s_CIGAR, __pyx_k_CIGAR, sizeof(__pyx_k_CIGAR), 0, 0, 1, 1},
  {&__pyx_n_u_D, __pyx_k_D, sizeof(__pyx_k_D), 0, 1, 0, 1},
  {&__pyx_n_u_I, __pyx_k_I, sizeof(__pyx_k_I), 0, 1, 0, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_u_S, __pyx_k_S, sizeof(__pyx_k_S), 0, 1, 0, 1},
  {&__pyx_n_s_SSW, __pyx_k_SSW, sizeof(__pyx_k_SSW), 0, 0, 1, 1},
  {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
  {&__pyx_n_s_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 0, 1, 1},
  {&__pyx_n_s_aldy_indelpost_localn, __pyx_k_aldy_indelpost_localn, sizeof(__pyx_k_aldy_indelpost_localn), 0, 0, 1, 1},
  {&__pyx_kp_s_aldy_indelpost_localn_pyx, __pyx_k_aldy_indelpost_localn_pyx, sizeof(__pyx_k_aldy_indelpost_localn_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_align, __pyx_k_align, sizeof(__pyx_k_align), 0, 0, 1, 1},
  {&__pyx_n_s_aligner, __pyx_k_aligner, sizeof(__pyx_k_aligner), 0, 0, 1, 1},
  {&__pyx_n_s_allow_mismatches, __pyx_k_allow_mismatches, sizeof(__pyx_k_allow_mismatches), 0, 0, 1, 1},
  {&__pyx_n_s_aln_end, __pyx_k_aln_end, sizeof(__pyx_k_aln_end), 0, 0, 1, 1},
  {&__pyx_n_u_aln_end, __pyx_k_aln_end, sizeof(__pyx_k_aln_end), 0, 1, 0, 1},
  {&__pyx_n_s_aln_start, __pyx_k_aln_start, sizeof(__pyx_k_aln_start), 0, 0, 1, 1},
  {&__pyx_n_u_aln_start, __pyx_k_aln_start, sizeof(__pyx_k_aln_start), 0, 1, 0, 1},
  {&__pyx_n_u_alt, __pyx_k_alt, sizeof(__pyx_k_alt), 0, 1, 0, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_basequals, __pyx_k_basequals, sizeof(__pyx_k_basequals), 0, 0, 1, 1},
  {&__pyx_n_s_basequalthresh, __pyx_k_basequalthresh, sizeof(__pyx_k_basequalthresh), 0, 0, 1, 1},
  {&__pyx_n_u_cigar_list, __pyx_k_cigar_list, sizeof(__pyx_k_cigar_list), 0, 1, 0, 1},
  {&__pyx_n_s_cigar_ptrn, __pyx_k_cigar_ptrn, sizeof(__pyx_k_cigar_ptrn), 0, 0, 1, 1},
  {&__pyx_n_u_cigar_string, __pyx_k_cigar_string, sizeof(__pyx_k_cigar_string), 0, 1, 0, 1},
  {&__pyx_n_s_cigarlst, __pyx_k_cigarlst, sizeof(__pyx_k_cigarlst), 0, 0, 1, 1},
  {&__pyx_n_s_cigarstring, __pyx_k_cigarstring, sizeof(__pyx_k_cigarstring), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_cnt, __pyx_k_cnt, sizeof(__pyx_k_cnt), 0, 0, 1, 1},
  {&__pyx_n_s_compile, __pyx_k_compile, sizeof(__pyx_k_compile), 0, 0, 1, 1},
  {&__pyx_n_s_consensus, __pyx_k_consensus, sizeof(__pyx_k_consensus), 0, 0, 1, 1},
  {&__pyx_n_s_contig, __pyx_k_contig, sizeof(__pyx_k_contig), 0, 0, 1, 1},
  {&__pyx_n_s_count_repeats, __pyx_k_count_repeats, sizeof(__pyx_k_count_repeats), 0, 0, 1, 1},
  {&__pyx_n_s_covering_end, __pyx_k_covering_end, sizeof(__pyx_k_covering_end), 0, 0, 1, 1},
  {&__pyx_n_s_covering_start, __pyx_k_covering_start, sizeof(__pyx_k_covering_start), 0, 0, 1, 1},
  {&__pyx_n_u_covering_subread, __pyx_k_covering_subread, sizeof(__pyx_k_covering_subread), 0, 1, 0, 1},
  {&__pyx_n_s_del_pos, __pyx_k_del_pos, sizeof(__pyx_k_del_pos), 0, 0, 1, 1},
  {&__pyx_n_u_del_pos, __pyx_k_del_pos, sizeof(__pyx_k_del_pos), 0, 1, 0, 1},
  {&__pyx_n_u_del_seq, __pyx_k_del_seq, sizeof(__pyx_k_del_seq), 0, 1, 0, 1},
  {&__pyx_n_s_dist_to_left_end, __pyx_k_dist_to_left_end, sizeof(__pyx_k_dist_to_left_end), 0, 0, 1, 1},
  {&__pyx_n_s_dist_to_right_end, __pyx_k_dist_to_right_end, sizeof(__pyx_k_dist_to_right_end), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_end_cigar, __pyx_k_end_cigar, sizeof(__pyx_k_end_cigar), 0, 0, 1, 1},
  {&__pyx_n_u_end_offset, __pyx_k_end_offset, sizeof(__pyx_k_end_offset), 0, 1, 0, 1},
  {&__pyx_n_s_end_point, __pyx_k_end_point, sizeof(__pyx_k_end_point), 0, 0, 1, 1},
  {&__pyx_n_s_end_trim, __pyx_k_end_trim, sizeof(__pyx_k_end_trim), 0, 0, 1, 1},
  {&__pyx_n_s_event, __pyx_k_event, sizeof(__pyx_k_event), 0, 0, 1, 1},
  {&__pyx_n_s_event_len, __pyx_k_event_len, sizeof(__pyx_k_event_len), 0, 0, 1, 1},
  {&__pyx_n_s_expected_n_repeats, __pyx_k_expected_n_repeats, sizeof(__pyx_k_expected_n_repeats), 0, 0, 1, 1},
  {&__pyx_n_s_find_by_smith_waterman_realn, __pyx_k_find_by_smith_waterman_realn, sizeof(__pyx_k_find_by_smith_waterman_realn), 0, 0, 1, 1},
  {&__pyx_n_s_findall, __pyx_k_findall, sizeof(__pyx_k_findall), 0, 0, 1, 1},
  {&__pyx_n_s_findall_indels, __pyx_k_findall_indels, sizeof(__pyx_k_findall_indels), 0, 0, 1, 1},
  {&__pyx_n_s_findall_mismatches, __pyx_k_findall_mismatches, sizeof(__pyx_k_findall_mismatches), 0, 0, 1, 1},
  {&__pyx_n_s_gap_extension, __pyx_k_gap_extension, sizeof(__pyx_k_gap_extension), 0, 0, 1, 1},
  {&__pyx_n_s_gap_extension_penalty, __pyx_k_gap_extension_penalty, sizeof(__pyx_k_gap_extension_penalty), 0, 0, 1, 1},
  {&__pyx_n_s_gap_open, __pyx_k_gap_open, sizeof(__pyx_k_gap_open), 0, 0, 1, 1},
  {&__pyx_n_s_gap_open_penalty, __pyx_k_gap_open_penalty, sizeof(__pyx_k_gap_open_penalty), 0, 0, 1, 1},
  {&__pyx_n_s_generate_equivalents, __pyx_k_generate_equivalents, sizeof(__pyx_k_generate_equivalents), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_genome_aln_pos, __pyx_k_genome_aln_pos, sizeof(__pyx_k_genome_aln_pos), 0, 0, 1, 1},
  {&__pyx_n_s_get_contig_seq, __pyx_k_get_contig_seq, sizeof(__pyx_k_get_contig_seq), 0, 0, 1, 1},
  {&__pyx_n_s_get_end_pos, __pyx_k_get_end_pos, sizeof(__pyx_k_get_end_pos), 0, 0, 1, 1},
  {&__pyx_n_s_get_mapped_subreads, __pyx_k_get_mapped_subreads, sizeof(__pyx_k_get_mapped_subreads), 0, 0, 1, 1},
  {&__pyx_n_s_get_reference_seq, __pyx_k_get_reference_seq, sizeof(__pyx_k_get_reference_seq), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_indel, __pyx_k_indel, sizeof(__pyx_k_indel), 0, 0, 1, 1},
  {&__pyx_n_s_indel_len, __pyx_k_indel_len, sizeof(__pyx_k_indel_len), 0, 0, 1, 1},
  {&__pyx_n_s_indel_seq, __pyx_k_indel_seq, sizeof(__pyx_k_indel_seq), 0, 0, 1, 1},
  {&__pyx_n_u_indel_seq, __pyx_k_indel_seq, sizeof(__pyx_k_indel_seq), 0, 1, 0, 1},
  {&__pyx_n_s_indel_type, __pyx_k_indel_type, sizeof(__pyx_k_indel_type), 0, 0, 1, 1},
  {&__pyx_n_u_indel_type, __pyx_k_indel_type, sizeof(__pyx_k_indel_type), 0, 1, 0, 1},
  {&__pyx_n_s_indels, __pyx_k_indels, sizeof(__pyx_k_indels), 0, 0, 1, 1},
  {&__pyx_n_s_is_compatible, __pyx_k_is_compatible, sizeof(__pyx_k_is_compatible), 0, 0, 1, 1},
  {&__pyx_n_s_is_compatible_repeats, __pyx_k_is_compatible_repeats, sizeof(__pyx_k_is_compatible_repeats), 0, 0, 1, 1},
  {&__pyx_n_s_is_covered, __pyx_k_is_covered, sizeof(__pyx_k_is_covered), 0, 0, 1, 1},
  {&__pyx_n_s_is_done, __pyx_k_is_done, sizeof(__pyx_k_is_done), 0, 0, 1, 1},
  {&__pyx_n_s_is_ins, __pyx_k_is_ins, sizeof(__pyx_k_is_ins), 0, 0, 1, 1},
  {&__pyx_n_s_is_left, __pyx_k_is_left, sizeof(__pyx_k_is_left), 0, 0, 1, 1},
  {&__pyx_n_s_is_lefty, __pyx_k_is_lefty, sizeof(__pyx_k_is_lefty), 0, 0, 1, 1},
  {&__pyx_n_u_is_reference_seq, __pyx_k_is_reference_seq, sizeof(__pyx_k_is_reference_seq), 0, 1, 0, 1},
  {&__pyx_n_u_is_target, __pyx_k_is_target, sizeof(__pyx_k_is_target), 0, 1, 0, 1},
  {&__pyx_n_s_is_target_by_ssw, __pyx_k_is_target_by_ssw, sizeof(__pyx_k_is_target_by_ssw), 0, 0, 1, 1},
  {&__pyx_n_s_is_worth_realn, __pyx_k_is_worth_realn, sizeof(__pyx_k_is_worth_realn), 0, 0, 1, 1},
  {&__pyx_n_s_is_worth_realn_locals_genexpr, __pyx_k_is_worth_realn_locals_genexpr, sizeof(__pyx_k_is_worth_realn_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_lt_clipped, __pyx_k_lt_clipped, sizeof(__pyx_k_lt_clipped), 0, 0, 1, 1},
  {&__pyx_n_u_lt_clipped, __pyx_k_lt_clipped, sizeof(__pyx_k_lt_clipped), 0, 1, 0, 1},
  {&__pyx_n_s_lt_consensus_seq, __pyx_k_lt_consensus_seq, sizeof(__pyx_k_lt_consensus_seq), 0, 0, 1, 1},
  {&__pyx_n_s_lt_diff, __pyx_k_lt_diff, sizeof(__pyx_k_lt_diff), 0, 0, 1, 1},
  {&__pyx_n_s_lt_end_read, __pyx_k_lt_end_read, sizeof(__pyx_k_lt_end_read), 0, 0, 1, 1},
  {&__pyx_n_s_lt_end_ref, __pyx_k_lt_end_ref, sizeof(__pyx_k_lt_end_ref), 0, 0, 1, 1},
  {&__pyx_n_s_lt_flank, __pyx_k_lt_flank, sizeof(__pyx_k_lt_flank), 0, 0, 1, 1},
  {&__pyx_n_u_lt_flank, __pyx_k_lt_flank, sizeof(__pyx_k_lt_flank), 0, 1, 0, 1},
  {&__pyx_n_s_lt_len, __pyx_k_lt_len, sizeof(__pyx_k_lt_len), 0, 0, 1, 1},
  {&__pyx_n_s_lt_most_pos, __pyx_k_lt_most_pos, sizeof(__pyx_k_lt_most_pos), 0, 0, 1, 1},
  {&__pyx_n_s_lt_pos, __pyx_k_lt_pos, sizeof(__pyx_k_lt_pos), 0, 0, 1, 1},
  {&__pyx_n_s_lt_qual, __pyx_k_lt_qual, sizeof(__pyx_k_lt_qual), 0, 0, 1, 1},
  {&__pyx_n_u_lt_qual, __pyx_k_lt_qual, sizeof(__pyx_k_lt_qual), 0, 1, 0, 1},
  {&__pyx_n_s_lt_ref, __pyx_k_lt_ref, sizeof(__pyx_k_lt_ref), 0, 0, 1, 1},
  {&__pyx_n_u_lt_ref, __pyx_k_lt_ref, sizeof(__pyx_k_lt_ref), 0, 1, 0, 1},
  {&__pyx_n_s_lt_seq, __pyx_k_lt_seq, sizeof(__pyx_k_lt_seq), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_make_aligner, __pyx_k_make_aligner, sizeof(__pyx_k_make_aligner), 0, 0, 1, 1},
  {&__pyx_n_s_make_insertion_first, __pyx_k_make_insertion_first, sizeof(__pyx_k_make_insertion_first), 0, 0, 1, 1},
  {&__pyx_n_s_mapped_base_cnt_thresh, __pyx_k_mapped_base_cnt_thresh, sizeof(__pyx_k_mapped_base_cnt_thresh), 0, 0, 1, 1},
  {&__pyx_n_s_mapped_qual, __pyx_k_mapped_qual, sizeof(__pyx_k_mapped_qual), 0, 0, 1, 1},
  {&__pyx_n_s_mapped_ref, __pyx_k_mapped_ref, sizeof(__pyx_k_mapped_ref), 0, 0, 1, 1},
  {&__pyx_n_s_mapped_seq, __pyx_k_mapped_seq, sizeof(__pyx_k_mapped_seq), 0, 0, 1, 1},
  {&__pyx_n_s_mapped_subpreads, __pyx_k_mapped_subpreads, sizeof(__pyx_k_mapped_subpreads), 0, 0, 1, 1},
  {&__pyx_n_u_mapq, __pyx_k_mapq, sizeof(__pyx_k_mapq), 0, 1, 0, 1},
  {&__pyx_n_s_mapq_lim, __pyx_k_mapq_lim, sizeof(__pyx_k_mapq_lim), 0, 0, 1, 1},
  {&__pyx_n_s_match_score, __pyx_k_match_score, sizeof(__pyx_k_match_score), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_mid_seq, __pyx_k_mid_seq, sizeof(__pyx_k_mid_seq), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_mismatch_penalty, __pyx_k_mismatch_penalty, sizeof(__pyx_k_mismatch_penalty), 0, 0, 1, 1},
  {&__pyx_n_s_mismatches, __pyx_k_mismatches, sizeof(__pyx_k_mismatches), 0, 0, 1, 1},
  {&__pyx_n_u_mismatches, __pyx_k_mismatches, sizeof(__pyx_k_mismatches), 0, 1, 0, 1},
  {&__pyx_n_s_mut_aligner, __pyx_k_mut_aligner, sizeof(__pyx_k_mut_aligner), 0, 0, 1, 1},
  {&__pyx_n_s_mut_aln, __pyx_k_mut_aln, sizeof(__pyx_k_mut_aln), 0, 0, 1, 1},
  {&__pyx_n_s_mut_aln_cigar, __pyx_k_mut_aln_cigar, sizeof(__pyx_k_mut_aln_cigar), 0, 0, 1, 1},
  {&__pyx_n_s_mut_aln_read_end, __pyx_k_mut_aln_read_end, sizeof(__pyx_k_mut_aln_read_end), 0, 0, 1, 1},
  {&__pyx_n_s_mut_aln_read_start, __pyx_k_mut_aln_read_start, sizeof(__pyx_k_mut_aln_read_start), 0, 0, 1, 1},
  {&__pyx_n_s_mut_aln_ref_end, __pyx_k_mut_aln_ref_end, sizeof(__pyx_k_mut_aln_ref_end), 0, 0, 1, 1},
  {&__pyx_n_s_mut_aln_ref_start, __pyx_k_mut_aln_ref_start, sizeof(__pyx_k_mut_aln_ref_start), 0, 0, 1, 1},
  {&__pyx_n_s_mut_cigar_list, __pyx_k_mut_cigar_list, sizeof(__pyx_k_mut_cigar_list), 0, 0, 1, 1},
  {&__pyx_n_s_mut_ref, __pyx_k_mut_ref, sizeof(__pyx_k_mut_ref), 0, 0, 1, 1},
  {&__pyx_n_s_mut_ref_lt, __pyx_k_mut_ref_lt, sizeof(__pyx_k_mut_ref_lt), 0, 0, 1, 1},
  {&__pyx_n_s_mut_ref_mid, __pyx_k_mut_ref_mid, sizeof(__pyx_k_mut_ref_mid), 0, 0, 1, 1},
  {&__pyx_n_s_mut_ref_rt, __pyx_k_mut_ref_rt, sizeof(__pyx_k_mut_ref_rt), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_n_s_optimal_score, __pyx_k_optimal_score, sizeof(__pyx_k_optimal_score), 0, 0, 1, 1},
  {&__pyx_n_s_parse_read_by_mut_aln, __pyx_k_parse_read_by_mut_aln, sizeof(__pyx_k_parse_read_by_mut_aln), 0, 0, 1, 1},
  {&__pyx_n_s_pileup, __pyx_k_pileup, sizeof(__pyx_k_pileup), 0, 0, 1, 1},
  {&__pyx_n_s_pos, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 0, 1, 1},
  {&__pyx_n_u_pos, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 1, 0, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_q, __pyx_k_q, sizeof(__pyx_k_q), 0, 0, 1, 1},
  {&__pyx_n_s_qual_lim, __pyx_k_qual_lim, sizeof(__pyx_k_qual_lim), 0, 0, 1, 1},
  {&__pyx_n_s_quals, __pyx_k_quals, sizeof(__pyx_k_quals), 0, 0, 1, 1},
  {&__pyx_n_s_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 1, 1},
  {&__pyx_n_s_re, __pyx_k_re, sizeof(__pyx_k_re), 0, 0, 1, 1},
  {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
  {&__pyx_n_s_read_end, __pyx_k_read_end, sizeof(__pyx_k_read_end), 0, 0, 1, 1},
  {&__pyx_n_s_read_idx, __pyx_k_read_idx, sizeof(__pyx_k_read_idx), 0, 0, 1, 1},
  {&__pyx_n_u_read_idx, __pyx_k_read_idx, sizeof(__pyx_k_read_idx), 0, 1, 0, 1},
  {&__pyx_n_u_read_name, __pyx_k_read_name, sizeof(__pyx_k_read_name), 0, 1, 0, 1},
  {&__pyx_n_s_read_qual, __pyx_k_read_qual, sizeof(__pyx_k_read_qual), 0, 0, 1, 1},
  {&__pyx_n_u_read_qual, __pyx_k_read_qual, sizeof(__pyx_k_read_qual), 0, 1, 0, 1},
  {&__pyx_n_s_read_seq, __pyx_k_read_seq, sizeof(__pyx_k_read_seq), 0, 0, 1, 1},
  {&__pyx_n_u_read_seq, __pyx_k_read_seq, sizeof(__pyx_k_read_seq), 0, 1, 0, 1},
  {&__pyx_n_s_read_start, __pyx_k_read_start, sizeof(__pyx_k_read_start), 0, 0, 1, 1},
  {&__pyx_n_u_read_start, __pyx_k_read_start, sizeof(__pyx_k_read_start), 0, 1, 0, 1},
  {&__pyx_n_s_ref, __pyx_k_ref, sizeof(__pyx_k_ref), 0, 0, 1, 1},
  {&__pyx_n_u_ref, __pyx_k_ref, sizeof(__pyx_k_ref), 0, 1, 0, 1},
  {&__pyx_n_s_ref_aligner, __pyx_k_ref_aligner, sizeof(__pyx_k_ref_aligner), 0, 0, 1, 1},
  {&__pyx_n_s_ref_aln, __pyx_k_ref_aln, sizeof(__pyx_k_ref_aln), 0, 0, 1, 1},
  {&__pyx_n_s_ref_end, __pyx_k_ref_end, sizeof(__pyx_k_ref_end), 0, 0, 1, 1},
  {&__pyx_n_s_ref_idx, __pyx_k_ref_idx, sizeof(__pyx_k_ref_idx), 0, 0, 1, 1},
  {&__pyx_n_u_ref_idx, __pyx_k_ref_idx, sizeof(__pyx_k_ref_idx), 0, 1, 0, 1},
  {&__pyx_n_s_ref_ref, __pyx_k_ref_ref, sizeof(__pyx_k_ref_ref), 0, 0, 1, 1},
  {&__pyx_n_s_ref_seq, __pyx_k_ref_seq, sizeof(__pyx_k_ref_seq), 0, 0, 1, 1},
  {&__pyx_n_u_ref_seq, __pyx_k_ref_seq, sizeof(__pyx_k_ref_seq), 0, 1, 0, 1},
  {&__pyx_n_s_ref_start, __pyx_k_ref_start, sizeof(__pyx_k_ref_start), 0, 0, 1, 1},
  {&__pyx_n_s_reference_end, __pyx_k_reference_end, sizeof(__pyx_k_reference_end), 0, 0, 1, 1},
  {&__pyx_n_s_reference_start, __pyx_k_reference_start, sizeof(__pyx_k_reference_start), 0, 0, 1, 1},
  {&__pyx_n_s_repeat_unit, __pyx_k_repeat_unit, sizeof(__pyx_k_repeat_unit), 0, 0, 1, 1},
  {&__pyx_n_s_report_snvs, __pyx_k_report_snvs, sizeof(__pyx_k_report_snvs), 0, 0, 1, 1},
  {&__pyx_n_s_rt_clipped, __pyx_k_rt_clipped, sizeof(__pyx_k_rt_clipped), 0, 0, 1, 1},
  {&__pyx_n_u_rt_clipped, __pyx_k_rt_clipped, sizeof(__pyx_k_rt_clipped), 0, 1, 0, 1},
  {&__pyx_n_s_rt_consensus_seq, __pyx_k_rt_consensus_seq, sizeof(__pyx_k_rt_consensus_seq), 0, 0, 1, 1},
  {&__pyx_n_s_rt_diff, __pyx_k_rt_diff, sizeof(__pyx_k_rt_diff), 0, 0, 1, 1},
  {&__pyx_n_s_rt_end_read, __pyx_k_rt_end_read, sizeof(__pyx_k_rt_end_read), 0, 0, 1, 1},
  {&__pyx_n_s_rt_end_ref, __pyx_k_rt_end_ref, sizeof(__pyx_k_rt_end_ref), 0, 0, 1, 1},
  {&__pyx_n_s_rt_flank, __pyx_k_rt_flank, sizeof(__pyx_k_rt_flank), 0, 0, 1, 1},
  {&__pyx_n_u_rt_flank, __pyx_k_rt_flank, sizeof(__pyx_k_rt_flank), 0, 1, 0, 1},
  {&__pyx_n_s_rt_len, __pyx_k_rt_len, sizeof(__pyx_k_rt_len), 0, 0, 1, 1},
  {&__pyx_n_s_rt_most_pos, __pyx_k_rt_most_pos, sizeof(__pyx_k_rt_most_pos), 0, 0, 1, 1},
  {&__pyx_n_s_rt_pos, __pyx_k_rt_pos, sizeof(__pyx_k_rt_pos), 0, 0, 1, 1},
  {&__pyx_n_s_rt_qual, __pyx_k_rt_qual, sizeof(__pyx_k_rt_qual), 0, 0, 1, 1},
  {&__pyx_n_u_rt_qual, __pyx_k_rt_qual, sizeof(__pyx_k_rt_qual), 0, 1, 0, 1},
  {&__pyx_n_s_rt_ref, __pyx_k_rt_ref, sizeof(__pyx_k_rt_ref), 0, 0, 1, 1},
  {&__pyx_n_u_rt_ref, __pyx_k_rt_ref, sizeof(__pyx_k_rt_ref), 0, 1, 0, 1},
  {&__pyx_n_s_rt_seq, __pyx_k_rt_seq, sizeof(__pyx_k_rt_seq), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_seq, __pyx_k_seq, sizeof(__pyx_k_seq), 0, 0, 1, 1},
  {&__pyx_n_s_setRead, __pyx_k_setRead, sizeof(__pyx_k_setRead), 0, 0, 1, 1},
  {&__pyx_n_s_setReference, __pyx_k_setReference, sizeof(__pyx_k_setReference), 0, 0, 1, 1},
  {&__pyx_n_s_shiftable_pos, __pyx_k_shiftable_pos, sizeof(__pyx_k_shiftable_pos), 0, 0, 1, 1},
  {&__pyx_n_s_snv, __pyx_k_snv, sizeof(__pyx_k_snv), 0, 0, 1, 1},
  {&__pyx_n_s_snvs, __pyx_k_snvs, sizeof(__pyx_k_snvs), 0, 0, 1, 1},
  {&__pyx_n_s_span, __pyx_k_span, sizeof(__pyx_k_span), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_sswpy, __pyx_k_sswpy, sizeof(__pyx_k_sswpy), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_start_cigar, __pyx_k_start_cigar, sizeof(__pyx_k_start_cigar), 0, 0, 1, 1},
  {&__pyx_n_u_start_offset, __pyx_k_start_offset, sizeof(__pyx_k_start_offset), 0, 1, 0, 1},
  {&__pyx_n_s_subread, __pyx_k_subread, sizeof(__pyx_k_subread), 0, 0, 1, 1},
  {&__pyx_n_s_target_covered, __pyx_k_target_covered, sizeof(__pyx_k_target_covered), 0, 0, 1, 1},
  {&__pyx_n_s_target_indel, __pyx_k_target_indel, sizeof(__pyx_k_target_indel), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_to_minimal_repeat_unit, __pyx_k_to_minimal_repeat_unit, sizeof(__pyx_k_to_minimal_repeat_unit), 0, 0, 1, 1},
  {&__pyx_n_s_token, __pyx_k_token, sizeof(__pyx_k_token), 0, 0, 1, 1},
  {&__pyx_n_u_undetermined, __pyx_k_undetermined, sizeof(__pyx_k_undetermined), 0, 1, 0, 1},
  {&__pyx_n_s_unit_len, __pyx_k_unit_len, sizeof(__pyx_k_unit_len), 0, 0, 1, 1},
  {&__pyx_n_s_upper, __pyx_k_upper, sizeof(__pyx_k_upper), 0, 0, 1, 1},
  {&__pyx_n_s_utilities, __pyx_k_utilities, sizeof(__pyx_k_utilities), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_var, __pyx_k_var, sizeof(__pyx_k_var), 0, 0, 1, 1},
  {&__pyx_n_s_variant_type, __pyx_k_variant_type, sizeof(__pyx_k_variant_type), 0, 0, 1, 1},
  {&__pyx_n_s_with_end_mismatches, __pyx_k_with_end_mismatches, sizeof(__pyx_k_with_end_mismatches), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 126, __pyx_L1_error)
  __pyx_builtin_min = __Pyx_GetBuiltinName(__pyx_n_s_min); if (!__pyx_builtin_min) __PYX_ERR(0, 180, __pyx_L1_error)
  __pyx_builtin_max = __Pyx_GetBuiltinName(__pyx_n_s_max); if (!__pyx_builtin_max) __PYX_ERR(0, 180, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(1, 109, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "aldy/indelpost/localn.pyx":92
 * 
 *             if "S" in cigarlst[0]:
 *                 cigarlst = cigarlst[1:]             # <<<<<<<<<<<<<<
 *                 read_seq = read_seq[read["start_offset"] :]
 *                 quals = quals[read["start_offset"] :]
 */
  __pyx_slice_ = PySlice_New(__pyx_int_1, Py_None, Py_None); if (unlikely(!__pyx_slice_)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice_);
  __Pyx_GIVEREF(__pyx_slice_);

  /* "aldy/indelpost/localn.pyx":97
 * 
 *             if "S" in cigarlst[-1]:
 *                 cigarlst = cigarlst[:-1]             # <<<<<<<<<<<<<<
 *                 read_seq = read_seq[: -read["end_offset"]]
 *                 quals = quals[: -read["end_offset"]]
 */
  __pyx_slice__2 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__2)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__2);
  __Pyx_GIVEREF(__pyx_slice__2);

  /* "aldy/indelpost/localn.pyx":437
 * 
 *     if is_left:
 *         seq = seq[::-1]             # <<<<<<<<<<<<<<
 *         repeat_unit = repeat_unit[::-1]
 * 
 */
  __pyx_slice__4 = PySlice_New(Py_None, Py_None, __pyx_int_neg_1); if (unlikely(!__pyx_slice__4)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__4);
  __Pyx_GIVEREF(__pyx_slice__4);

  /* "aldy/indelpost/localn.pyx":11
 * from .utilities cimport split, count_lowqual_non_ref_bases
 * 
 * cigar_ptrn = re.compile(r"[0-9]+[MIDNSHPX=]")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_u_0_9_MIDNSHPX); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "aldy/indelpost/localn.pyx":14
 * 
 * 
 * def find_by_smith_waterman_realn(             # <<<<<<<<<<<<<<
 *     target_indel,
 *     contig,
 */
  __pyx_tuple__6 = PyTuple_Pack(20, __pyx_n_s_target_indel, __pyx_n_s_contig, __pyx_n_s_pileup, __pyx_n_s_match_score, __pyx_n_s_mismatch_penalty, __pyx_n_s_gap_open_penalty, __pyx_n_s_gap_extension_penalty, __pyx_n_s_basequalthresh, __pyx_n_s_mapq_lim, __pyx_n_s_indel_type, __pyx_n_s_indel_seq, __pyx_n_s_mut_ref_lt, __pyx_n_s_mut_ref_mid, __pyx_n_s_mut_ref_rt, __pyx_n_s_ref_ref, __pyx_n_s_mut_ref, __pyx_n_s_mut_aligner, __pyx_n_s_ref_aligner, __pyx_n_s_read, __pyx_n_s_read); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(9, 0, 20, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_localn_pyx, __pyx_n_s_find_by_smith_waterman_realn, 14, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(0, 14, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":70
 * 
 * 
 * def findall_mismatches(read, end_trim=0):             # <<<<<<<<<<<<<<
 * 
 *     if read["is_reference_seq"]:
 */
  __pyx_tuple__8 = PyTuple_Pack(27, __pyx_n_s_read, __pyx_n_s_end_trim, __pyx_n_s_aln_start, __pyx_n_s_aln_end, __pyx_n_s_mapped_subpreads, __pyx_n_s_mismatches, __pyx_n_s_subread, __pyx_n_s_start, __pyx_n_s_end, __pyx_n_s_span, __pyx_n_s_cigarstring, __pyx_n_s_cigarlst, __pyx_n_s_read_seq, __pyx_n_s_quals, __pyx_n_s_lt_seq, __pyx_n_s_rt_seq, __pyx_n_s_lt_qual, __pyx_n_s_rt_qual, __pyx_n_s_lt_ref, __pyx_n_s_rt_ref, __pyx_n_s_mapped_seq, __pyx_n_s_mapped_qual, __pyx_n_s_mapped_ref, __pyx_n_s_pos, __pyx_n_s_r, __pyx_n_s_a, __pyx_n_s_q); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __pyx_codeobj__9 = (PyObject*)__Pyx_PyCode_New(2, 0, 27, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_localn_pyx, __pyx_n_s_findall_mismatches, 70, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__9)) __PYX_ERR(0, 70, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":138
 * 
 * 
 * def is_worth_realn(read, target_indel, qual_lim=23):             # <<<<<<<<<<<<<<
 * 
 *     if read["covering_subread"]:
 */
  __pyx_tuple__10 = PyTuple_Pack(29, __pyx_n_s_read, __pyx_n_s_target_indel, __pyx_n_s_qual_lim, __pyx_n_s_is_covered, __pyx_n_s_covering_start, __pyx_n_s_covering_end, __pyx_n_s_dist_to_left_end, __pyx_n_s_dist_to_right_end, __pyx_n_s_is_lefty, __pyx_n_s_start_cigar, __pyx_n_s_end_cigar, __pyx_n_s_mismatches, __pyx_n_s_shiftable_pos, __pyx_n_s_lt_pos, __pyx_n_s_rt_pos, __pyx_n_s_lt_end_read, __pyx_n_s_lt_end_ref, __pyx_n_s_rt_end_read, __pyx_n_s_rt_end_ref, __pyx_n_s_lt_most_pos, __pyx_n_s_with_end_mismatches, __pyx_n_s_rt_most_pos, __pyx_n_s_indels, __pyx_n_s_var, __pyx_n_s_v, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_var); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_codeobj__11 = (PyObject*)__Pyx_PyCode_New(3, 0, 29, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_localn_pyx, __pyx_n_s_is_worth_realn, 138, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__11)) __PYX_ERR(0, 138, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":222
 *     return False
 * 
 * def is_target_by_ssw(             # <<<<<<<<<<<<<<
 *     read,
 *     target_indel,
 */
  __pyx_tuple__12 = PyTuple_Pack(27, __pyx_n_s_read, __pyx_n_s_target_indel, __pyx_n_s_contig, __pyx_n_s_mut_ref_lt, __pyx_n_s_mut_ref_mid, __pyx_n_s_mut_ref_rt, __pyx_n_s_mut_aligner, __pyx_n_s_ref_aligner, __pyx_n_s_match_score, __pyx_n_s_mismatch_penalty, __pyx_n_s_gap_open_penalty, __pyx_n_s_gap_extension_penalty, __pyx_n_s_indel_type, __pyx_n_s_basequalthresh, __pyx_n_s_mapq_lim, __pyx_n_s_mapped_base_cnt_thresh, __pyx_n_s_allow_mismatches, __pyx_n_s_read_seq, __pyx_n_s_ref_aln, __pyx_n_s_mut_aln, __pyx_n_s_mut_aln_cigar, __pyx_n_s_mut_aln_ref_start, __pyx_n_s_mut_aln_ref_end, __pyx_n_s_mut_aln_read_start, __pyx_n_s_mut_aln_read_end, __pyx_n_s_mut_cigar_list, __pyx_n_s_target_covered); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  __pyx_codeobj__13 = (PyObject*)__Pyx_PyCode_New(17, 0, 27, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_localn_pyx, __pyx_n_s_is_target_by_ssw, 222, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__13)) __PYX_ERR(0, 222, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":432
 * 
 * 
 * def is_compatible_repeats(seq, repeat_unit, expected_n_repeats, is_left):             # <<<<<<<<<<<<<<
 *     unit_len = len(repeat_unit)
 * 
 */
  __pyx_tuple__14 = PyTuple_Pack(8, __pyx_n_s_seq, __pyx_n_s_repeat_unit, __pyx_n_s_expected_n_repeats, __pyx_n_s_is_left, __pyx_n_s_unit_len, __pyx_n_s_cnt, __pyx_n_s_is_done, __pyx_n_s_is_compatible); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_codeobj__15 = (PyObject*)__Pyx_PyCode_New(4, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__14, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_localn_pyx, __pyx_n_s_is_compatible_repeats, 432, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__15)) __PYX_ERR(0, 432, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":463
 * 
 * 
 * def make_aligner(ref_seq, match_score, mismatch_penalty):             # <<<<<<<<<<<<<<
 *     aligner = SSW(match_score=match_score, mismatch_penalty=mismatch_penalty)
 *     aligner.setReference(ref_seq)
 */
  __pyx_tuple__16 = PyTuple_Pack(4, __pyx_n_s_ref_seq, __pyx_n_s_match_score, __pyx_n_s_mismatch_penalty, __pyx_n_s_aligner); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(3, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_localn_pyx, __pyx_n_s_make_aligner, 463, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(0, 463, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":469
 * 
 * 
 * def align(aligner, read_seq, gap_open_penalty, gap_extension_penalty):             # <<<<<<<<<<<<<<
 *     aligner.setRead(read_seq)
 *     return aligner.align(gap_open=gap_open_penalty, gap_extension=gap_extension_penalty)
 */
  __pyx_tuple__18 = PyTuple_Pack(4, __pyx_n_s_aligner, __pyx_n_s_read_seq, __pyx_n_s_gap_open_penalty, __pyx_n_s_gap_extension_penalty); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_localn_pyx, __pyx_n_s_align, 469, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 469, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":474
 * 
 * 
 * def parse_read_by_mut_aln(mut_aln, contig, read, indel_type):             # <<<<<<<<<<<<<<
 *     """Decompose read based on the SSW alignment result for evaluation
 * 
 */
  __pyx_tuple__20 = PyTuple_Pack(24, __pyx_n_s_mut_aln, __pyx_n_s_contig, __pyx_n_s_read, __pyx_n_s_indel_type, __pyx_n_s_lt_len, __pyx_n_s_indel_len, __pyx_n_s_rt_len, __pyx_n_s_read_seq, __pyx_n_s_read_qual, __pyx_n_s_ref_start, __pyx_n_s_ref_end, __pyx_n_s_aln_start, __pyx_n_s_aln_end, __pyx_n_s_lt_flank, __pyx_n_s_mid_seq, __pyx_n_s_rt_flank, __pyx_n_s_lt_qual, __pyx_n_s_rt_qual, __pyx_n_s_lt_diff, __pyx_n_s_end_point, __pyx_n_s_del_pos, __pyx_n_s_lt_ref, __pyx_n_s_rt_ref, __pyx_n_s_rt_diff); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(4, 0, 24, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_localn_pyx, __pyx_n_s_parse_read_by_mut_aln, 474, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 474, __pyx_L1_error)

  /* "aldy/indelpost/localn.pyx":541
 * 
 * 
 * def findall_indels(ref_aln, genome_aln_pos, ref_seq, read_seq, report_snvs=False, basequals=None):             # <<<<<<<<<<<<<<
 * 
 *     genome_aln_pos -= 1
 */
  __pyx_tuple__22 = PyTuple_Pack(18, __pyx_n_s_ref_aln, __pyx_n_s_genome_aln_pos, __pyx_n_s_ref_seq, __pyx_n_s_read_seq, __pyx_n_s_report_snvs, __pyx_n_s_basequals, __pyx_n_s_ref_idx, __pyx_n_s_read_idx, __pyx_n_s_lt_clipped, __pyx_n_s_indels, __pyx_n_s_snvs, __pyx_n_s_token, __pyx_n_s_event, __pyx_n_s_event_len, __pyx_n_s_indel, __pyx_n_s_i, __pyx_n_s_snv, __pyx_n_s_rt_clipped); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(6, 0, 18, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_aldy_indelpost_localn_pyx, __pyx_n_s_findall_indels, 541, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_23 = PyInt_FromLong(23); if (unlikely(!__pyx_int_23)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_40 = PyInt_FromLong(40); if (unlikely(!__pyx_int_40)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn.tp_dictoffset && __pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn = &__pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct__is_worth_realn;
  if (PyType_Ready(&__pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 199, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr = &__pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct_1_genexpr;
  if (PyType_Ready(&__pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 202, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr.tp_dictoffset && __pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr = &__pyx_type_4aldy_9indelpost_6localn___pyx_scope_struct_2_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(3, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(4, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("array"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_5array_array = __Pyx_ImportType(__pyx_t_1, "array", "array", sizeof(arrayobject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_5array_array) __PYX_ERR(1, 58, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pysam.libchtslib"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 2680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pysam_10libchtslib_HTSFile = __Pyx_ImportType(__pyx_t_1, "pysam.libchtslib", "HTSFile", sizeof(struct __pyx_obj_5pysam_10libchtslib_HTSFile), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_10libchtslib_HTSFile) __PYX_ERR(5, 2680, __pyx_L1_error)
  __pyx_vtabptr_5pysam_10libchtslib_HTSFile = (struct __pyx_vtabstruct_5pysam_10libchtslib_HTSFile*)__Pyx_GetVtable(__pyx_ptype_5pysam_10libchtslib_HTSFile->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_10libchtslib_HTSFile)) __PYX_ERR(5, 2680, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pysam.libcfaidx"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pysam_9libcfaidx_FastaFile = __Pyx_ImportType(__pyx_t_1, "pysam.libcfaidx", "FastaFile", sizeof(struct __pyx_obj_5pysam_9libcfaidx_FastaFile), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_9libcfaidx_FastaFile) __PYX_ERR(6, 37, __pyx_L1_error)
  __pyx_vtabptr_5pysam_9libcfaidx_FastaFile = (struct __pyx_vtabstruct_5pysam_9libcfaidx_FastaFile*)__Pyx_GetVtable(__pyx_ptype_5pysam_9libcfaidx_FastaFile->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_9libcfaidx_FastaFile)) __PYX_ERR(6, 37, __pyx_L1_error)
  __pyx_ptype_5pysam_9libcfaidx_FastqProxy = __Pyx_ImportType(__pyx_t_1, "pysam.libcfaidx", "FastqProxy", sizeof(struct __pyx_obj_5pysam_9libcfaidx_FastqProxy), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_9libcfaidx_FastqProxy) __PYX_ERR(6, 45, __pyx_L1_error)
  __pyx_vtabptr_5pysam_9libcfaidx_FastqProxy = (struct __pyx_vtabstruct_5pysam_9libcfaidx_FastqProxy*)__Pyx_GetVtable(__pyx_ptype_5pysam_9libcfaidx_FastqProxy->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_9libcfaidx_FastqProxy)) __PYX_ERR(6, 45, __pyx_L1_error)
  __pyx_ptype_5pysam_9libcfaidx_FastxRecord = __Pyx_ImportType(__pyx_t_1, "pysam.libcfaidx", "FastxRecord", sizeof(struct __pyx_obj_5pysam_9libcfaidx_FastxRecord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_9libcfaidx_FastxRecord) __PYX_ERR(6, 52, __pyx_L1_error)
  __pyx_vtabptr_5pysam_9libcfaidx_FastxRecord = (struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxRecord*)__Pyx_GetVtable(__pyx_ptype_5pysam_9libcfaidx_FastxRecord->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_9libcfaidx_FastxRecord)) __PYX_ERR(6, 52, __pyx_L1_error)
  __pyx_ptype_5pysam_9libcfaidx_FastxFile = __Pyx_ImportType(__pyx_t_1, "pysam.libcfaidx", "FastxFile", sizeof(struct __pyx_obj_5pysam_9libcfaidx_FastxFile), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_9libcfaidx_FastxFile) __PYX_ERR(6, 61, __pyx_L1_error)
  __pyx_vtabptr_5pysam_9libcfaidx_FastxFile = (struct __pyx_vtabstruct_5pysam_9libcfaidx_FastxFile*)__Pyx_GetVtable(__pyx_ptype_5pysam_9libcfaidx_FastxFile->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_9libcfaidx_FastxFile)) __PYX_ERR(6, 61, __pyx_L1_error)
  __pyx_ptype_5pysam_9libcfaidx_FastqFile = __Pyx_ImportType(__pyx_t_1, "pysam.libcfaidx", "FastqFile", sizeof(struct __pyx_obj_5pysam_9libcfaidx_FastqFile), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_9libcfaidx_FastqFile) __PYX_ERR(6, 73, __pyx_L1_error)
  __pyx_vtabptr_5pysam_9libcfaidx_FastqFile = (struct __pyx_vtabstruct_5pysam_9libcfaidx_FastqFile*)__Pyx_GetVtable(__pyx_ptype_5pysam_9libcfaidx_FastqFile->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_9libcfaidx_FastqFile)) __PYX_ERR(6, 73, __pyx_L1_error)
  __pyx_ptype_5pysam_9libcfaidx_Fastafile = __Pyx_ImportType(__pyx_t_1, "pysam.libcfaidx", "Fastafile", sizeof(struct __pyx_obj_5pysam_9libcfaidx_Fastafile), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_9libcfaidx_Fastafile) __PYX_ERR(6, 78, __pyx_L1_error)
  __pyx_vtabptr_5pysam_9libcfaidx_Fastafile = (struct __pyx_vtabstruct_5pysam_9libcfaidx_Fastafile*)__Pyx_GetVtable(__pyx_ptype_5pysam_9libcfaidx_Fastafile->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_9libcfaidx_Fastafile)) __PYX_ERR(6, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pysam.libcbcf"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pysam_7libcbcf_VariantHeader = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantHeader", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantHeader), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantHeader) __PYX_ERR(7, 38, __pyx_L1_error)
  __pyx_vtabptr_5pysam_7libcbcf_VariantHeader = (struct __pyx_vtabstruct_5pysam_7libcbcf_VariantHeader*)__Pyx_GetVtable(__pyx_ptype_5pysam_7libcbcf_VariantHeader->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_7libcbcf_VariantHeader)) __PYX_ERR(7, 38, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantHeaderRecord = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantHeaderRecord", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantHeaderRecord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantHeaderRecord) __PYX_ERR(7, 46, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantHeaderRecords = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantHeaderRecords", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantHeaderRecords), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantHeaderRecords) __PYX_ERR(7, 51, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantHeaderContigs = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantHeaderContigs", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantHeaderContigs), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantHeaderContigs) __PYX_ERR(7, 55, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantHeaderSamples = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantHeaderSamples", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantHeaderSamples), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantHeaderSamples) __PYX_ERR(7, 59, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantContig = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantContig", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantContig), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantContig) __PYX_ERR(7, 63, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantMetadata = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantMetadata", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantMetadata), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantMetadata) __PYX_ERR(7, 68, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantHeaderMetadata = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantHeaderMetadata", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantHeaderMetadata), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantHeaderMetadata) __PYX_ERR(7, 74, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantRecord = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantRecord", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantRecord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantRecord) __PYX_ERR(7, 79, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantRecordFilter = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantRecordFilter", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantRecordFilter), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantRecordFilter) __PYX_ERR(7, 84, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantRecordFormat = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantRecordFormat", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantRecordFormat), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantRecordFormat) __PYX_ERR(7, 88, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantRecordInfo = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantRecordInfo", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantRecordInfo), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantRecordInfo) __PYX_ERR(7, 92, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantRecordSamples = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantRecordSamples", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantRecordSamples), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantRecordSamples) __PYX_ERR(7, 96, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantRecordSample = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantRecordSample", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantRecordSample), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantRecordSample) __PYX_ERR(7, 100, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_BaseIndex = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "BaseIndex", sizeof(struct __pyx_obj_5pysam_7libcbcf_BaseIndex), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_BaseIndex) __PYX_ERR(7, 105, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_BCFIndex = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "BCFIndex", sizeof(struct __pyx_obj_5pysam_7libcbcf_BCFIndex), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_BCFIndex) __PYX_ERR(7, 110, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_TabixIndex = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "TabixIndex", sizeof(struct __pyx_obj_5pysam_7libcbcf_TabixIndex), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_TabixIndex) __PYX_ERR(7, 115, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_BaseIterator = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "BaseIterator", sizeof(struct __pyx_obj_5pysam_7libcbcf_BaseIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_BaseIterator) __PYX_ERR(7, 119, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_BCFIterator = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "BCFIterator", sizeof(struct __pyx_obj_5pysam_7libcbcf_BCFIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_BCFIterator) __PYX_ERR(7, 124, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_TabixIterator = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "TabixIterator", sizeof(struct __pyx_obj_5pysam_7libcbcf_TabixIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_TabixIterator) __PYX_ERR(7, 128, __pyx_L1_error)
  __pyx_ptype_5pysam_7libcbcf_VariantFile = __Pyx_ImportType(__pyx_t_1, "pysam.libcbcf", "VariantFile", sizeof(struct __pyx_obj_5pysam_7libcbcf_VariantFile), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5pysam_7libcbcf_VariantFile) __PYX_ERR(7, 133, __pyx_L1_error)
  __pyx_vtabptr_5pysam_7libcbcf_VariantFile = (struct __pyx_vtabstruct_5pysam_7libcbcf_VariantFile*)__Pyx_GetVtable(__pyx_ptype_5pysam_7libcbcf_VariantFile->tp_dict); if (unlikely(!__pyx_vtabptr_5pysam_7libcbcf_VariantFile)) __PYX_ERR(7, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("aldy.indelpost.variant"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4aldy_9indelpost_7variant_NullVariant = __Pyx_ImportType(__pyx_t_1, "aldy.indelpost.variant", "NullVariant", sizeof(struct __pyx_obj_4aldy_9indelpost_7variant_NullVariant), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4aldy_9indelpost_7variant_NullVariant) __PYX_ERR(8, 6, __pyx_L1_error)
  __pyx_ptype_4aldy_9indelpost_7variant_Variant = __Pyx_ImportType(__pyx_t_1, "aldy.indelpost.variant", "Variant", sizeof(struct __pyx_obj_4aldy_9indelpost_7variant_Variant), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4aldy_9indelpost_7variant_Variant) __PYX_ERR(8, 13, __pyx_L1_error)
  __pyx_vtabptr_4aldy_9indelpost_7variant_Variant = (struct __pyx_vtabstruct_4aldy_9indelpost_7variant_Variant*)__Pyx_GetVtable(__pyx_ptype_4aldy_9indelpost_7variant_Variant->tp_dict); if (unlikely(!__pyx_vtabptr_4aldy_9indelpost_7variant_Variant)) __PYX_ERR(8, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("aldy.indelpost.utilities"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "split", (void (**)(void))&__pyx_f_4aldy_9indelpost_9utilities_split, "PyObject *(PyObject *, PyObject *, int, int, int, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "count_lowqual_non_ref_bases", (void (**)(void))&__pyx_f_4aldy_9indelpost_9utilities_count_lowqual_non_ref_bases, "int (PyObject *, PyObject *, arrayobject *, PyObject *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initlocaln(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initlocaln(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_localn(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_localn(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_localn(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'localn' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_localn(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("localn", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_aldy__indelpost__localn) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "aldy.indelpost.localn")) {
      if (unlikely(PyDict_SetItemString(modules, "aldy.indelpost.localn", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "aldy/indelpost/localn.pyx":3
 * #cython: profile=False
 * 
 * import re             # <<<<<<<<<<<<<<
 * import numpy as np
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_re, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_re, __pyx_t_1) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":4
 * 
 * import re
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * from .sswpy import SSW
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":6
 * import numpy as np
 * 
 * from .sswpy import SSW             # <<<<<<<<<<<<<<
 * from .consensus import is_compatible
 * from .utilities import get_mapped_subreads, get_end_pos, make_insertion_first, to_minimal_repeat_unit
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_SSW);
  __Pyx_GIVEREF(__pyx_n_s_SSW);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_SSW);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_sswpy, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_SSW); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SSW, __pyx_t_1) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/localn.pyx":7
 * 
 * from .sswpy import SSW
 * from .consensus import is_compatible             # <<<<<<<<<<<<<<
 * from .utilities import get_mapped_subreads, get_end_pos, make_insertion_first, to_minimal_repeat_unit
 * from .utilities cimport split, count_lowqual_non_ref_bases
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_is_compatible);
  __Pyx_GIVEREF(__pyx_n_s_is_compatible);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_is_compatible);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_consensus, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_is_compatible); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_compatible, __pyx_t_2) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "aldy/indelpost/localn.pyx":8
 * from .sswpy import SSW
 * from .consensus import is_compatible
 * from .utilities import get_mapped_subreads, get_end_pos, make_insertion_first, to_minimal_repeat_unit             # <<<<<<<<<<<<<<
 * from .utilities cimport split, count_lowqual_non_ref_bases
 * 
 */
  __pyx_t_1 = PyList_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_get_mapped_subreads);
  __Pyx_GIVEREF(__pyx_n_s_get_mapped_subreads);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_get_mapped_subreads);
  __Pyx_INCREF(__pyx_n_s_get_end_pos);
  __Pyx_GIVEREF(__pyx_n_s_get_end_pos);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_get_end_pos);
  __Pyx_INCREF(__pyx_n_s_make_insertion_first);
  __Pyx_GIVEREF(__pyx_n_s_make_insertion_first);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_make_insertion_first);
  __Pyx_INCREF(__pyx_n_s_to_minimal_repeat_unit);
  __Pyx_GIVEREF(__pyx_n_s_to_minimal_repeat_unit);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_s_to_minimal_repeat_unit);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_utilities, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_get_mapped_subreads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_mapped_subreads, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_get_end_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_end_pos, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_make_insertion_first); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_make_insertion_first, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_to_minimal_repeat_unit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_to_minimal_repeat_unit, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/localn.pyx":11
 * from .utilities cimport split, count_lowqual_non_ref_bases
 * 
 * cigar_ptrn = re.compile(r"[0-9]+[MIDNSHPX=]")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_re); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compile); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cigar_ptrn, __pyx_t_2) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/localn.pyx":14
 * 
 * 
 * def find_by_smith_waterman_realn(             # <<<<<<<<<<<<<<
 *     target_indel,
 *     contig,
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6localn_1find_by_smith_waterman_realn, NULL, __pyx_n_s_aldy_indelpost_localn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_find_by_smith_waterman_realn, __pyx_t_2) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/localn.pyx":70
 * 
 * 
 * def findall_mismatches(read, end_trim=0):             # <<<<<<<<<<<<<<
 * 
 *     if read["is_reference_seq"]:
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6localn_3findall_mismatches, NULL, __pyx_n_s_aldy_indelpost_localn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_findall_mismatches, __pyx_t_2) < 0) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/localn.pyx":138
 * 
 * 
 * def is_worth_realn(read, target_indel, qual_lim=23):             # <<<<<<<<<<<<<<
 * 
 *     if read["covering_subread"]:
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6localn_5is_worth_realn, NULL, __pyx_n_s_aldy_indelpost_localn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_worth_realn, __pyx_t_2) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/localn.pyx":222
 *     return False
 * 
 * def is_target_by_ssw(             # <<<<<<<<<<<<<<
 *     read,
 *     target_indel,
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6localn_7is_target_by_ssw, NULL, __pyx_n_s_aldy_indelpost_localn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_target_by_ssw, __pyx_t_2) < 0) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/localn.pyx":432
 * 
 * 
 * def is_compatible_repeats(seq, repeat_unit, expected_n_repeats, is_left):             # <<<<<<<<<<<<<<
 *     unit_len = len(repeat_unit)
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6localn_9is_compatible_repeats, NULL, __pyx_n_s_aldy_indelpost_localn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_compatible_repeats, __pyx_t_2) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/localn.pyx":463
 * 
 * 
 * def make_aligner(ref_seq, match_score, mismatch_penalty):             # <<<<<<<<<<<<<<
 *     aligner = SSW(match_score=match_score, mismatch_penalty=mismatch_penalty)
 *     aligner.setReference(ref_seq)
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6localn_11make_aligner, NULL, __pyx_n_s_aldy_indelpost_localn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_make_aligner, __pyx_t_2) < 0) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/localn.pyx":469
 * 
 * 
 * def align(aligner, read_seq, gap_open_penalty, gap_extension_penalty):             # <<<<<<<<<<<<<<
 *     aligner.setRead(read_seq)
 *     return aligner.align(gap_open=gap_open_penalty, gap_extension=gap_extension_penalty)
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6localn_13align, NULL, __pyx_n_s_aldy_indelpost_localn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_align, __pyx_t_2) < 0) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/localn.pyx":474
 * 
 * 
 * def parse_read_by_mut_aln(mut_aln, contig, read, indel_type):             # <<<<<<<<<<<<<<
 *     """Decompose read based on the SSW alignment result for evaluation
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6localn_15parse_read_by_mut_aln, NULL, __pyx_n_s_aldy_indelpost_localn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_parse_read_by_mut_aln, __pyx_t_2) < 0) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/localn.pyx":541
 * 
 * 
 * def findall_indels(ref_aln, genome_aln_pos, ref_seq, read_seq, report_snvs=False, basequals=None):             # <<<<<<<<<<<<<<
 * 
 *     genome_aln_pos -= 1
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_4aldy_9indelpost_6localn_17findall_indels, NULL, __pyx_n_s_aldy_indelpost_localn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_findall_indels, __pyx_t_2) < 0) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "aldy/indelpost/localn.pyx":1
 * #cython: profile=False             # <<<<<<<<<<<<<<
 * 
 * import re
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "array.pxd":161
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init aldy.indelpost.localn", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init aldy.indelpost.localn");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* py_abs */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *n) {
    if (likely(Py_SIZE(n) == -1)) {
        return PyLong_FromLong(((PyLongObject*)n)->ob_digit[0]);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    {
        PyObject *copy = _PyLong_Copy((PyLongObject*)n);
        if (likely(copy)) {
            __Pyx_SET_SIZE(copy, -Py_SIZE(copy));
        }
        return copy;
    }
#else
    return PyNumber_Negative(n);
#endif
}
#endif

/* PyUnicode_Substring */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Substring(
            PyObject* text, Py_ssize_t start, Py_ssize_t stop) {
    Py_ssize_t length;
    if (unlikely(__Pyx_PyUnicode_READY(text) == -1)) return NULL;
    length = __Pyx_PyUnicode_GET_LENGTH(text);
    if (start < 0) {
        start += length;
        if (start < 0)
            start = 0;
    }
    if (stop < 0)
        stop += length;
    else if (stop > length)
        stop = length;
    if (stop <= start)
        return __Pyx_NewRef(__pyx_empty_unicode);
#if CYTHON_PEP393_ENABLED
    return PyUnicode_FromKindAndData(PyUnicode_KIND(text),
        PyUnicode_1BYTE_DATA(text) + start*PyUnicode_KIND(text), stop-start);
#else
    return PyUnicode_FromUnicode(PyUnicode_AS_UNICODE(text)+start, stop-start);
#endif
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
